/**
 * @file
 * @brief Memory allocation polymorph.
 *
 * Some functions must allocate memory for its results. One time we
 * need an allocation in the pool, another time a direct allocation
 * in the memory is necessary. As it is inconvient to write two 
 * versions of such function we can use the @link lib::tAllocator
 * tAllocator@endlink class
 * and write only one common function.  
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */  
#ifndef _ECHELOG_ALLOCATOR_H
#define _ECHELOG_ALLOCATOR_H

namespace lib {

/**
 * @brief Class for polymorph memory allocations.
 *
 * Some functions must allocate memory for its results. One time we 
 * need an allocation in the pool, another time a direct allocation
 * in the memory. As it is inconvient to write two versions of such 
 * function we can use this class and write only one common function.
 * 
 * When a instance of this class is created, we can either supply 
 * a pool instance to the constructor or leave it blank (NULL). 
 * In the first case all allocations ordered to this instance are 
 * made in the pool, in the second case all allocations will be made  
 * directly by @link lib::xmalloc() xmalloc()@endlink call. 
 * The same holds for the deallocating
 * routine. 
 *
 * @note The way of allocations and deallocations cannot be changed 
 * during the instance's lifetime. 
 * 
 */ 
class tAllocator {

    protected:

    /**
     * @brief This attribute holds a pool for allocations and deallocations.
     * If it is NULL, then allocations are made via @link lib::xmalloc()
     * xmalloc()@endlink call.
     */
    class tMemPool *pool;

    public:

    /**
     * @brief Copy constructor for the object.
     */
    tAllocator (tMemPool &ipool) : pool (&ipool) {};

    /**
     * @brief Constructor for the object. If <tt>ipool</tt> parameter is
     * supplied, all allocations will be made in this pool, otherwise 
     * all allocations will be made directly by @link lib::xmalloc()
     * xmalloc()@endlink call.
     */
    tAllocator (tMemPool *ipool = NULL) : pool (ipool) {};

    /**
     * @brief Allocates memory either in pool or directly in memory.
     */
    inline void * p_alloc (size_t size)
    {
        if (pool)
            return pool->alloc (size);
        return xmalloc (size);
    }

    /**
     * @brief Deallocates memory either from pool or directly from memory.
     */
    inline void p_free (void *ptr)
    {
        if (!pool && ptr)
            xfree (ptr);
    }

};

}

#endif

