/**
 * @file
 * @brief Compressed memory bitmap.
 *
 * This provides a simple interface for a table space management.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_CBITMAP_H
#define ECHELOG_CBITMAP_H

/** @brief Character printed for used bit in debug printouts */
#define CBITMAP_DEBUG_UsedBit '#'
/** @brief Character printed for free bit in debug printouts */
#define CBITMAP_DEBUG_FreeBit '.'

namespace lib {

/** 
 * @brief Class for a simple table space management.
 * 
 * This class can be used for a simple table space management. It provides
 * basic mechanism for seting/reseting/testing bits in bit arrays. 
 * It uses a compressed bitmap to reduce memory requirements.
 */
class tCBitmap {

    protected:
        /** @brief Number of bits in the storage. */
        long       num_of_bits;
        /** @brief Number of real items. */
        long       num_of_items;        
        /** @brief Bit storage. */
        uint32_t *mem_bitmap;

        /** 
         * @brief Clear an allocated bit with number <tt>n</tt>.
         */
        inline void clear_bit (long n);

        /** 
         * @brief Allocate a bit with number <tt>n</tt>.
         */
        inline void set_bit (long n);

        /** 
         * @brief Query whether the bit with number <tt>n</tt> is set.
         */
        inline bool is_set (long n);

        /** 
         * @brief Query whether the bit with number <tt>n</tt> is set
         * is free.
         */
        inline bool is_free (long n);

        /** 
         * @brief Test if a continuous block of free bits is available 
         * at the particular place.
         *
         * This function tests if a continuous block of free bits is available
         * at the place starting on bit <tt>start_bit</tt>i with lenght 
         * <tt>block_length</tt>. Fuction return zero on success, maximum
         * number of available free bits in block on failure.
         *          
         */
        inline long test_block (long start_bit, long block_length);

        /** 
         * @brief Skip all allocated bits from start_bit position.
         *
         * This function skips all allocated bits from <tt>start_bit</tt>
         * position. The function returns the number of skipped bits.
         */
        inline long skip_allocated (long start_bit);

        // Skip over all false bits up to the limit
        inline long skip_false_bits (long start_pos, long length_limit);

        // Skip over all true bits up to the limit
        inline long skip_true_bits (long start_pos, long length_limit);

    public:
        /** 
         * @brief Create a compressed bitmap and set the maximum bit storage
         * to <tt>total_bits</tt>.
         */
        tCBitmap (long total_bits);	

        /**
         * @brief Destroy the compressed bitmap.
         */
        ~tCBitmap ();

        /**
         * @brief Clear the whole bitmap.
         */
        void clear_all ();

        /** 
         * @brief Try to allocate a free block of bits at least block_length
         * long. The function uses find first method.
         *
         * This function tries to allocate a free block of bits at least 
         * <tt>block_length</tt> long. The function returns the 
         * number of the first bit of the allocated block. 
         * On failure the function returns <tt>-1</tt>.
         */
        long get_block (long block_length);

        /**
         * @brief Get the first single free bit
         *
         * This function gets the first single free bit. It returns the number
         * of the allocated bit. On failure the function returns <tt>-1</tt>.
         */
        long get_bit ();

        /**
         * @brief Free an allocated continuous block of bits.
         *
         * This function frees a continuous block of bits starting with
         * bit <tt>start_bit</tt>. The allocated block size is
         * <tt>block_length</tt>.
         */
        void free_block (long start_bit, long block_length);	

        /** 
         * @brief Free an allocated bit.
         *
         * Frees a bit with <tt>bit_num</tt> number (i.e. marks it as free).
         */
        void free_bit (long bit_num)
            { free_block (bit_num, 1); }

        /**
         * @brief Allocate a continuous block of bits.
         *
         * This function allocates a continuous block of bits starting with 
         * bit <tt>start_bit</tt>. The allocated block size is
         * <tt>block_length</tt>. 
         */
        void allocate_block (long start_bit, long block_length);

        /** 
         * @brief Allocate a bit.
         *
         * This function allocates a bit with <tt>bit_num</tt> number (i.e.
         * marks it as used).
         */
        void allocate_bit (long bit_num)
            { allocate_block (bit_num, 1); }

        /**
         * @brief A CBitmap printout for debugging purposes.
         */
        void print_bitmap();
};

}

#endif
