/**
 * @file
 * @brief Reading of configuration files.
 *
 * This file contains implementation of configuration files reading.
 * 
 * @author Robert Spalek <robert@ucw.cz>
 * @author Ales Privetivy <ales@ucw.cz>
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef _ECHELOG_CONF_H
#define _ECHELOG_CONF_H

#include <unistd.h>

/**
 * @namespace lib
 * @brief This namespace contains all classes and functions in lib library. 
 */
namespace lib {

/**
 * @brief This variable holds filename of default configuration file.
 */
extern const char *gb_default_config_filename;

/** 
 * @brief Constant indicating that this key is mandatory. 
 */ 
#define ITEM_REQUIRED 1    
/**
 * @brief Constant indicating that this was filled in (it does not hold
 * implicit value).
 */
#define ITEM_LOADED   2    
 
/**
 * @brief This enumeration distinguishes type of keys.
 *
 * This enumeration distinguishes type of keys in configuration description.
 * Key can have either string value, integer value or be a list of string
 * values. There are two special key types, which mark start and end of the
 * configuration section.
 */
enum cftype { 
    CT_STOP = 0,    ///< End of a section.
    CT_SECTION,     ///< Begin of a section.
    CT_INT,         ///< Identifies key with integer value.
    CT_STRING,      ///< Identifies key with string value.
    CT_FUNCTION,    ///< Identifies key which value is analyzed by function.
    CT_LIST         ///< Identifies keys which forms lists. 
};

struct cfitem;

/**
 * @brief Type of functions analyzing data (for <tt>#CT_FUNCTION</tt> records).
 */
typedef char *(*ci_func)(struct cfitem *, char *);

/**
 * @brief Type of functions reinitializing data structures 
 * (for <tt>#CT_FUNCTION</tt> records). 
 */
typedef void (*ci_reset_func)(struct cfitem *);

/**
 * @brief Structure for the description of section items.
 *
 * This is a structure for the description of section item. Each configuration
 * section is described by array of this structures. First item of this array
 * must be <tt>#CT_SECTION</tt> and determines section name. Last item must be 
 * <tt>#CT_STOP</tt>. The keys declared in this section must be enclosed in this
 * two items.
 */
struct cfitem {

    /**
     * @brief Name of section/key.
     */
	char *name;
    /**
     * @brief Type of section item record.
     */
	enum cftype type;
    /**
     * @brief Pointer where should be read data stored.
     */
    void *data;

    /**
     * @brief Pointer to registration function (valid only for
     * <tt>#CT_FUNCTION</tt> record).
     */
    ci_func varf;

    /**
     * @brief Pointer to reinitialization function (valid only for
     * <tt>#CT_FUNCTION</tt> record).
     */
    ci_reset_func varf_reset;

    /**
     * @brief Bitfield indicating status of item (e.g. <tt>#ITEM_REQUIRED</tt>
     * or <tt>#ITEM_LOADED</tt>). 
     */
    int flags;
};

#ifndef CF_SHORT_OPTS
/** 
 * @brief Prefix containing command line switches of configuration system.
 *
 * When using cf_getopt, you must prefix your own short/long options by the
 * <tt>#CF_SHORT_OPTS</tt>. This is because the configuration system needs
 * some switches for its functionality. Namely it is <tt>-S</tt> for setting
 * value of a key from command line, <tt>-C</tt> for determining of the
 * configuration file, <tt>-X</tt> for debugging purposes and <tt>-h</tt>
 * for help print out.
 *
 * If you want the standard system help message to be appended to your
 * help message, define <tt>CF_HELP_MSG</tt>. If you don't want this define
 * your help message as <tt>CF_COMPLETE_HELP_MSG</tt>.
 *
 */
#define	CF_SHORT_OPTS	"S:C:X:h"
#endif
#ifndef CF_LONG_OPTS
/**
 * @brief Prefix containing command line switches of configuration system.
 */ 
#define	CF_LONG_OPTS	\
        {"set",     1, 0, 'S'},\
        {"config",  1, 0, 'C'},\
        {"debug",   1, 0, 'X'},\
        {"help",    0, 0, 'h'}
#endif

#ifndef DOXY_SKIP        
        
#ifndef CF_HELP_MSG
/// default help message printout
#define CF_HELP_MSG 
#endif
#ifndef CF_COMPLETE_HELP_MSG
#	if DEBUG > 1
#	define CF_X_HELP_OPT "  -X, --debug\t\t\tSet new debugging options\n"
#	else
#	define CF_X_HELP_OPT	
#	endif
#define CF_COMPLETE_HELP_MSG CF_HELP_MSG \
	"Standard configuration parameters:\n"\
	"  -S, --set\t\t\tSet a new key value\n"\
	"  -C, --config\t\t\tSet a new configuration file\n"\
	CF_X_HELP_OPT \
	"  -h, --help\t\t\tThis help\n"
#endif

/// No long options field
#define CF_NO_LONG_OPTS (const struct option []){ CF_LONG_OPTS, { NULL, 0, 0, 0 } }

#endif    

/**
 * @brief This function parses one item of the configuration file and
 * command line parameters.
 */  
int cf_getopt(int argc,char * const argv[],
		const char *shortopts,const struct option *longopts,
		int *longindex, tMemPool &pool);

/**
 * @brief Parses the configuration file and command line parameters.
 *
 * This function parses the configuration file and command line parameters and
 * reads key values of registered keys from them. 
 *
 * @return Returns zero on success, otherwise <tt>1</tt> on error.
 */
int cf_process(int argc, char **argv, tMemPool &pool);

/**
 * @brief Registers a configuration section.
 *
 * This function adds section <tt>items</tt> into the current configuration 
 * system. 
 */
void cf_register(struct cfitem *items);

/**
 * @brief Unregisters a configuration section. 
 *
 * When a section is no more needed, we may remove it from the configuration
 * system. If the configuration is reread later unregistered sections are
 * unaffected.
 */
void cf_unregister(const char *section_name);

/**
 * @brief Duplicates string within the configuration parsing pool.
 */
char *cf_strdup(const char *s);

/**
 * @brief Allocates space in configuration parsing pool.
 */
void *cf_malloc(unsigned long size);

/**
 * @brief Sets a new configuration file filename.
 */
void cf_set_config_filename (const char *new_config_name);

/**
 * @brief Rereads a configuration from the actual configuration file.
 *
 * This function rereads a configuration from the actual configuration file.
 * All data allocations during read are made in the <tt>pool</tt> memory pool.
 * This file becomes then implicit, i.e. next configuration rereads are done
 * from it.
 *
 * @return Returns zero on success, otherwise <tt>-1</tt> on error.
 */
int cf_reread(tMemPool &pool);

/** 
 * @brief Returns configuration file name in a printable form.
 */
const char* cf_get_config_filename ();

/**
 * @brief Reads a new configuration from a given file.
 *
 * This function reads a new configuration from a file <tt>filename</tt>. All
 * data allocations during read are made in the <tt>pool</tt> memory pool. 
 * This file becomes then implicit, i.e. next configuration rereads are done 
 * from it.
 *
 * @return Returns zero on success, otherwise <tt>-1</tt> on error.
 */
int cf_reread(const char *filename, tMemPool &pool);

/** 
 * @brief Invalidates all key values in all registered sections to allow 
 * a configuration reload 
 *
 * This function invalidates all key values in all registered sections to
 * allow a configuration reload.
 */
void cf_cleanup_registered_sections();

/**
 * @brief Clears the flag that item is loaded for all items and sections.
 *
 * This function clears the flag that item is loaded for all items and
 * sections. This is done especially before file reread.
 */
void cf_clear_loaded_flags();

/**
 * @brief Checks if every required parameter was loaded.
 *
 * This function checks if every required parameter is set. 
 *
 * @return Returns zero if every required parameter was loaded, otherwise 
 * <tt>-1</tt>.
 */
int cf_check_required();

/**
 * @brief Returns global pool used for parsing information.
 */
tMemPool &get_cfpool ();

/**
 * @brief Performs a preprocess of a configuration file.
 *
 * This function preprocesses the configuration file <tt>in</tt> and puts the
 * output to file <tt>out</tt>. All memory allocations are done in 
 * <tt>cf_pool</tt>, so ensure clearing of it after each preprocessing. 
 *
 * @return Returns zero on success, otherwise <tt>-1</tt> on error.
 */
int cf_preprocess(const char *in, const char *out);

/**
 * @brief Prints error passed from preprocessor.
 */
void cf_report_error();

}

#endif
