/**
 * @file
 * @brief Debugging macros with output to logfile.
 *
 * This file contains some useful macros used for debug outputs.
 * 
 * @author Kamil Toman <ktoman@email.cz>
 *	
 */

#ifndef __ECHELOG_DEBUG_H
#define __ECHELOG_DEBUG_H

#include <stdlib.h>

namespace lib {

/** @brief This is bit array indicating which debug messages should be
 * written to log and which not. */
extern long gb_debug_sections;    

/** @brief Function returning always false. */
inline bool always_false() { return false; }
/** @brief Function returning always true. */
inline bool always_true() { return true; }


/** @brief Clears debugging for all sections. */
void debug_section_vwclear_all ();
/** @brief Sets debugging for all sections. */
void debug_section_vwset_all ();
/** @brief Clears debugging for one specific section. */
void debug_section_vwclear (int section);
/** @brief Sets debugging for one specific section. */
void debug_section_vwset (int section);
/** @brief Sets multiple sections for debugging. */
void debug_section_vwset (const char *sections);
/** @brief Clears debugging for multiple sections. */
void debug_section_vwclear (const char *sections);
/** @brief Test if debugging for one specific section is set. */
bool debug_section_vwis_set (int section);

/** @brief You can supply your own printout function for debug-logs. */
void call_a_fake_im_a_stupid_cpp (int section, char *msg, ...);

}

#ifndef DOXY_SKIP

#ifdef DEBUG
#define ASSERT(x) \
    do { using lib::vwlog; \
        if (!(x)) \
        { \
            vwlog( lib::L_FATAL, "Assertion `%s' failed at %s:%d", #x, \
                    __FILE__, __LINE__); \
            abort (); \
        } \
    } while(0)
#else
#define ASSERT(x) do { } while(0)
#endif /* DEBUG */


#ifdef DEBUG_LINES
#    ifdef __GNUC__
#       define log(type, args...) do { \
            vwlog (type, "FILE:%s LINE:%d", __FILE__, __LINE__); \
            vwlog(type, args); \
        } while (0) 
#       define die(args...) do { \
            vwlog (L_FATAL, "FILE:%s LINE:%d", __FILE__, __LINE__); \
            vwdie(args); \
        } while (0) 
#    elif __USE_ISOC99
#       define log(type, ...) do { \
            vwlog (type, "FILE:%s LINE:%d", __FILE__, __LINE__); \
            vwlog(type, __VA_ARGS__); \
        }  while (0) 
#       define die(format, ...) do { \
            vwlog (L_FATAL, "FILE:%s LINE:%d", __FILE__, __LINE__); \
            vwdie(format, __VA_ARGS__); \
        } while (0) 
#    else
#        define log vwlog
#        define die vwdie
#    endif              
#else   /* no DEBUG_LINES */
#    define log vwlog
#    define die vwdie
#endif  /* DEBUG_LINES */


#if DEBUG > 1
#    define debug_log debug_vwlog
#    define debug_section_warn_if_not_compiled_in()      
#    define debug_section_clear_all() debug_section_vwclear_all()
#    define debug_section_set_all() debug_section_vwset_all()
#    define debug_section_clear(x) debug_section_vwclear (x)
#    define debug_section_set(x) debug_section_vwset(x)
#    define debug_section_is_set(x) debug_section_vwis_set(x)
#else
#    ifdef __GNUC__
#	define debug_log(section, args...) do {} while(0)
#    elif __USE_ISOC99
#       define debug_log(section, ...) do {} while(0)
#    else
#       define debug_log call_a_fake_im_a_stupid_cpp
#    endif
#    define debug_section_clear_all() do {} while (0)
#    define debug_section_set_all() do {} while (0)
#    define debug_section_clear(x) do {} while (0)
#    define debug_section_set(x) do {} while (0)
#    define debug_section_is_set(x) always_false()        
#    define debug_section_warn_if_not_compiled_in() \
        vwlog (L_WARN, "Section debugging is not compiled in")     
#endif  /* DEBUG > 1 */

#endif
        
#endif
