/**
 * @file
 * @brief Miscellaneous Functions
 *
 * This file contains basic functions which are needed everywhere.
 * This file should be included as the very first include in all
 * source files, especially before all OS includes since it sets
 * up libc feature macros.
 *
 * @author Martin Mares <mj@ucw.cz>
 * @author Ales Privetivy <ales@ucw.cz> 
 * @author Kamil Toman <ktoman@email.cz>
 *
 */


#ifndef _ECHELOG_LIB_H
#define _ECHELOG_LIB_H

#include "lib/archdep.h"
#include "echelog_config.h"

#include <stdlib.h>

namespace lib {

/** @brief Constant determining address alignment of allocated data. */    
#define DATA_ALIGN 4

/** @brief Function returning aligned address. */
inline char * align(char * p)
{
   return  (char *) (((unsigned long)(p) + DATA_ALIGN - 1 ) 
            & ~( DATA_ALIGN - 1) ); 
}

/** @brief Just return a minimal value. */
inline long min(long a, long b)
{
    return (a < b) ? a : b;
}

/** @brief Just return a maximal value. */
inline long max(long a, long b)
{
    return (a > b) ? a : b;
}

/** @brief Just return an absolute value of <tt>x</tt>. */
inline long abs(long x)
{
    return (x < 0) ? -x : x;
}

/** @brief Use a value of <tt>x</tt> but keep it in given bounds. */
inline long clamp(long x, long min_value, long max_value)
{
    if (x < min_value) 
        return min_value;
    else if (x > max_value);
        return max_value;

    return x;
}

/** @brief Divide and round down. */
inline long div_round_down (long n, long s)
{
    return n / s;
}

/** @brief Divide and round up. */
inline long div_round_up (long n, long s)
{
    return (n / s) + (((n % s) > 0) ? 1 : 0);
}

/** @brief Return the number of elements. Use it only for statically 
 * allocated arrays! */
#define ARRAY_SIZE(a) (sizeof(a)/sizeof(*(a)))

/** 
 * @brief This enumeration distinguishies log entries.
 *
 * This enumeration distinguishies log entries. 
 */
enum lStatus { 
    	L_DEBUG='D',    ///< Debugging messages.
        L_INFO='I',     ///< Informational messages.
        L_WARN='W',     ///< Warning messages.
        L_ERROR='E',    ///< Error messages.
        L_INFO_R='i',	///< Informational messages caused by external events.
        L_WARN_R='w',   ///< Warning messages caused by external events.
        L_ERROR_R='e',  ///< Error messages caused by external events.
        L_FATAL='!' 	///< Fatal errors (used by die()).
};

/** @brief Logging system default semaphore key */
#define LOG_SEMKEY 0x2703

/**
 * @brief Outut of printf-like message to log.
 *
 * This function prints message to the log file. We can specify the 
 * condition on which was this message generated (@see enum lib::lStatus).
 * The message text is passed in <tt>msg</tt> attribute.
 *
 * @note This function should never be called from interrupt handlers, since
 * it could cause mess in log, in case that other printout of the log message 
 * was interrupted.
 */
#ifndef DOXY_SKIP
void vwlog(enum lStatus, const char *msg, ...) __attribute__((format(printf,2,3)));
#elif
void vwlog(enum lStatus, const char *msg, ...);
#endif

/**
 * @brief Output of printf-like message to log and terminate.
 * 
 * This function prints message to the log file and then terminates the 
 * program. The message text is passed in <tt>msg</tt> attribute.
 *
 * @note This function should never be called from interrupt handlers, since
 * it could cause mess in log, in case that other printout of the log message
 * was interrupted.
 */
void vwdie(const char *, ...) NONRET;

/**
 * @brief Output of printf-like message to log for debugging puropses.
 *
 * This function prints message to the log file. We must specify the
 * debugging section to which section the message belongs via
 * the <tt>section</tt> parameter. The message text is passed in 
 * <tt>msg</tt> attribute. 
 *
 * @note This function should never be called from interrupt handlers, since
 * it would cause mess in log, in case that other printout of the log message
 * was interrupted.
 */
void debug_vwlog (int section, const char *msg, ...);

/**
 * @brief Closes the log-file.
 */
void log_close();

/**
 * @brief Initialization of logging.
 *
 * This function initializates logging system. For better readability of log
 * entries program name should be supplied. It is passed as <tt>argv0</tt>
 * parameter. This function exits, if the initialization was unsuccessful.
 */ 
void log_init(const char *argv0);

/**
 * @brief Initialization of logging.
 *
 * This function initializates logging system. For better readability of log
 * entries program name should be supplied. It is passed as <tt>argv0</tt>
 * parameter. This function returns zero on success, <tt>-1</tt> on error. 
 */ 
int vlog_init(const char *argv0);

/**
 * @brief Redirects output to a filename.
 *
 * This function redirects the output to file identified by <tt>name</tt>. It
 * returns zero on success, <tt>-1</tt> on error (possibly cannot create
 * log-file).
 */
int log_file(const char *name);

/**
 * @brief Locks log file.
 */
void log_lock();

/**
 * @brief Unlocks log file.
 */
void log_unlock();

/**
 * @brief Initialization of logging for programs running in multiple instances.
 */
void log_fork(void);

/**
 * @brief Malloc with die-ing when not enough space.
 *
 * This function tries to allocate space of size <tt>size</tt>.
 * On success, it returns pointer to allocated space. On error
 * it dies.
 */
void *xmalloc(unsigned long size);

/**
 * @brief Malloc and zero data with die-ing when not enough space.
 *
 * This function tries to allocate space of size <tt>size</tt>
 * and fill it with zeroes. On success, it returns pointer to 
 * allocated space. On error it dies.
 */
void *xmalloc_zero(unsigned long size);

/**
 * @brief Realloc with die-ing when not enough space.
 *
 * This function tries to reallocate space pointed by <tt>ptr</tt> pointer.
 * It must be pointer to space previously allocated with <tt>xmalloc()</tt> or
 * NULL. The new size of the space is determined by <tt>new_size</tt>
 * parameter. The contents will be  unchanged  to the minimum of the old and
 * new sizes; newly allocated memory will be uninitialized.  On success, it
 * returns pointer to newly allocated space. On error it dies.
 */
void *xrealloc(void *ptr, unsigned long new_size);

/**
 * @brief Installs new signal handler.
 *
 * This function installs new signal handler. The <tt>signum</tt> determines
 * for which signal is the handler for. Address of the handler is given in
 * <tt>handler</tt> parameter. If the handler cannot be installed, it dies.
 */
void set_signal_handler(int signum, void (*handler)(int));

/**
 * @brief Unallocates memory allocated by <tt>xmalloc()</tt>.
 */
inline void xfree(void *ptr) { if (ptr) free (ptr); }

/**
 * @brief Comparison functor type.
 *
 * This is type for comparison functors.
 * The functor should return an integer less than, equal to, or greater
 * to zero if the first parameter is considered less than, equal to, 
 * or greater to the second, respectively.
 */
typedef int (*tCompFunc) (const void *, const void *);

/**
 * @brief Converts string specifying time to unixtime.
 *
 * This function converts string specifying time to unixtime.
 * For allowed time formats see
 * @ref parse_time_format "Format of the time specification".
 * This function returns number of seconds since Epoch
 * (00:00:00 1.1.1970), if the <tt>timestr</tt> contains
 * valid time. If the time is invalid or lies out of range
 * <tt>-1</tt> is returned.
 */
time_t parsetime(char *timestr);

}

#include "lib/cbitmap.h"
#include "lib/mutex.h"
#include "lib/debug.h"
#include "lib/pools.h"
#include "lib/allocator.h"
#include "lib/list.h"
#include "lib/dlist.h"
#include "lib/string.h"
#include "lib/listqueue.h"
#include "lib/liststack.h"
#include "lib/s_io.h"
#include "lib/conf.h"
#include "lib/strptime.h"
#include "lib/timestring.h"
#include "lib/xml.h"

#endif
