/**
 * @file
 * @brief Single Linked Lists
 *
 * This file contains single linked list object.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef _ECHELOG_LIST_H
#define _ECHELOG_LIST_H

#include "property.h"

namespace lib {

/**
 * @brief Structure for one element of the single linked list.
 */ 
struct tNode {

    /// Pointer to next element.
    struct tNode *next;
    /// Pointer to data associated with this element.
    void *data;
};


/** 
 * @brief Class for a single linked list manipulation.
 *
 * This class is implementation of single linked list. It supports
 * appending, prepending a removing of an item. The item descriptors
 * can be allocated in a heap or in some supplied pool.
 * 
 */
class tList : public tAllocator {

    protected:
        /** @brief Number of items in the list. */
        long num_of_items;

        /** @brief <em>Read-only property:</em> Pointer to the first element 
         * of link-list. */
        SIMPLE_R_PROPERTY (struct tNode *, first);        
        /** @brief <em>Read-only property:</em> Pointer to the last element 
         * of link-list */
        SIMPLE_R_PROPERTY (struct tNode *, last);         

    public:

        /**
         * @brief Constructor for link-list.
         *
         * The list structure is created on the standard heap.
         */
        inline tList ();

        /**
         * @brief Constructor for link-list.
         *
         * The list structure is created in the pool <tt>ipool</tt>.
         * The item descriptors are allocated from this pool.
         */
        inline tList (tMemPool &ipool);

        /**
         * @brief Destructor for link-list.
         *
         * The list structure is destroyed by calling <tt>reset()</tt>
         * function. The data themselves are not freed.
         */
        virtual inline ~tList();
        
        /**
         * @brief Appends all items from another list.
         *
         * The function appends all data from the linked list <tt>list</tt>.
         * The data are not duplicated (shallow copy).
         */
        inline void concat (const tList &list);
        
        /**
         * @brief Makes a shallow copy of link-list.
         *
         * The function makes a shallow copy of the linked list <tt>list</tt>.
         * If the list already contains some items they are deleted.
         */
        inline void shallow_copy (const tList &list);

        /**
         * @brief Reinitialize the link-list (removes all items).
         */
        inline void reset ();

        /**
         * @brief Creates element descriptor, associates it with data and 
         * inserts to the link-list beginning. 
         *
         * Creates element descriptor, associates it with <tt>data</tt> and
         * inserts it to the link-list beginning.
         */
        inline void prepend (void *data);

        /**
         * @brief Removes the first element from the link-list.
         *
         * This function returns the pointer to data of the first element 
         * in the list or NULL if there is none. The item is also removed
         * from list.
         */
        inline void *remove_first ();

        /**
         * @brief  Creates element descriptor, associates it with data and 
         * inserts to the end of the link-list
         *
         * Creates element descriptor, associates it with <tt>data</tt> and
         * appends it to the link-list end.
         */
        inline void append (void *data);

        /**
         * @brief Removes the last element from the link-list.
         *
         * This function returns the pointer to data of the last element
         * in the list or NULL if there is none. The item is also removed
         * from list.
         *
         * @note Implementation of this method is linear to number of items in
         * the list.
         */
        inline void *remove_last ();

        /**
         * @brief Removes the first matching element from the link-list.
         *
         * This function returns the pointer to data of the first matching
         * element in the list or NULL if there is none such. The item is
         * matched via pointer to its data.
         *
         * @note Implementation of this method is linear to number of items in
         * the list.
         */
        inline void *remove_match (void *);

        /**
         * @brief Removes the first matching element from the link-list using
         * user supplied comparator.
         *
         * This function returns the pointer to data of the first matching
         * element in the list or NULL if there is none such. The item is
         * matched via user supplied comparator function <tt>f</tt>.
         *
         * @note Implementation of this method is linear to number of items in
         * the list.
         */
        inline void *remove_match (void *, tCompFunc f);

        /**
         * @brief Performs operation on each element of the link-list.
         *
         * This function performs operation <tt>fnc</tt> on each element
         * of the link-list. It returns sum of values returned from the
         * function applied to each element.
         */
        inline int walk (int (*fnc)(void *)) const;

        /**
         * @brief Counts the number of elements in the list.
         */
        inline int count () const;
};       

}


/**
 * @brief Expand iteration cycle control.
 *
 * This macro iterates through list <tt>list</tt> (pointer to list is
 * expected) and in each step <tt>n</tt> has value of the actual
 * item descriptor.
 */
#define FOR_ALL_LIST_ITEMS(n, list) \
    for ((n) = (list)->get_first (); (n); (n) = (n)->next)


/* We actually define all inline functions in a separate file */
#include "lib/list.hxx"

#endif
