/**
 * @file
 * @brief Echelog Library -- Single Linked Lists
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef _ECHELOG_LIST_HXX
#define _ECHELOG_LIST_HXX

namespace lib {


// Constructor for link-list
tList::tList() : tAllocator(), num_of_items (0)
{ 
    last = first = NULL;
}


// Constructor for link-list
tList::tList (tMemPool &ipool) : tAllocator (ipool), num_of_items (0)
{ 
    last = first = NULL;
}

// Destructor for link-list
tList::~tList ()
{
    reset ();
}


// Appends all items from another list
void tList::concat (const tList &list)
{
    tNode *c;
    for (c = list.first; c; c = c->next)
        append (c->data);
}


// Makes a shallow copy of link-list
void tList::shallow_copy (const tList &list)
{
    if (num_of_items > 0)
        reset ();
    concat (list);
}


// Reinitialize the link-list
void tList::reset ()
{ 
    if (!pool)
    {
        /* Remove all elements from the standard heap */
        long i;
        struct tNode *old, *p = first;

        for (i = num_of_items; i > 0; i--)
        {
            old = p;
            p = p->next;
            p_free (old);
        }
    }

    /* Reset the list */
    num_of_items = 0;
    last = first = NULL;
}


// Creates element descriptor, associates it with data and 
void tList::prepend (void *data) 
{
    struct tNode *p;

    p = static_cast<struct tNode *> (p_alloc (sizeof (struct tNode)));	
    p->next = first;
    p->data = data;
    first = p;	
    if (num_of_items == 0)
        last = first;
    num_of_items++;
}


// Removes the first element from the link-list
void *tList::remove_first ()
{
    void *ret = NULL;

    if (num_of_items <= 0)
        return NULL;
    else if (num_of_items == 1)
    {
        ret = first->data;
        p_free (first);
        last = first = NULL;
        num_of_items--;
    }
    else
    {
        struct tNode *p = first;

        ret = first->data;
        first = first->next;
        p_free (p);
        num_of_items--;
    }

    return ret;
}


// Creates element descriptor, associates it with data and 
void tList::append (void *data) 
{
    struct tNode *p;

    if (num_of_items == 0)
            prepend (data);
    else
    {
        p = static_cast<struct tNode *> 
            (p_alloc (sizeof (struct tNode)));	
        p->data = data;
        p->next = NULL;
        last->next = p;
        last = p;
        num_of_items++;
    }
}


// Removes the last element from the link-list
void *tList::remove_last ()
{
    void *ret = NULL;

    if (num_of_items <= 0)
        return NULL;
    else if (num_of_items == 1)
    {
        ret = first->data;
        p_free (first);
        last = first = NULL;
        num_of_items--;
    }
    else 
    {
        struct tNode *p;

        last = first;
        for (p = first->next; p->next; p = p->next)
            last = p;

        ret = p->data;
        p_free (p);
        last->next = NULL;
        num_of_items--;
    }

    return ret;
}


// Removes the first matching element from the link-list
void *tList::remove_match (void *data)
{
    if (num_of_items <= 0)
        return NULL;

    if (first && first->data == data)
    {
        struct tNode *p;

        p = first;
        first = first->next;
        if (p == last)
            last = NULL;
        p_free (p);
        num_of_items--;
        return data;
    }
    else if (num_of_items > 1)
    {
        struct tNode *p, *p2;

        p2 = first;
        for (p = first->next; p && p->data != data; p = p->next)
            p2 = p;

        if (p)
        {
            p2->next = p->next;
            if (p == last)
                last = p2;
            p_free (p);
            num_of_items--;
            return data;
        }
    }

    return NULL;
}


// Removes the first matching element from the link-list
void *tList::remove_match (void *data, tCompFunc f)
{
    if (num_of_items <= 0)
        return NULL;

    if (first && !f (first->data, data))
    {
        struct tNode *p;

        p = first;
        first = first->next;
        if (p == last)
            last = NULL;
        p_free (p);
        num_of_items--;
        return data;
    }
    else if (num_of_items > 1)
    {
        struct tNode *p, *p2;

        p2 = first;
        for (p = first->next; p && f (p->data, data); p = p->next)
            p2 = p;

        if (p)
        {
            p2->next = p->next;
            if (p == last)
                last = p2;
            p_free (p);
            num_of_items--;
            return data;
        }
    }

    return NULL;
}


// Performs operation on each element of the link-list
int tList::walk (int (*fnc)(void *)) const
{
    struct tNode *c;
    int sum=0;

    for(c = first; c; c = c->next) 
        sum += fnc (c->data);

    return sum;
}


// Counts the number of elements in the list
inline int tList::count () const
{
    return num_of_items;
}

}

#endif
