/**
 * @file
 * @brief Queues (using link-lists).
 *
 * This file contains simple queue object implemented via link-list.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef _ECHELOG_LISTQUEUE_H
#define _ECHELOG_LISTQUEUE_H

#include "property.h"

namespace lib {


/** 
 * @brief Class for a queue manipulation.
 *
 * This class is simple queue object supporting functions like insert,
 * remove, peek, etc. It is implemented via link-list object (@see lib::tList).
 * It can be created either in a heap or in a pool.
 */
class tListQueue : private tList
{

    public:

        /**
         * @brief Constructor for the queue.
         *
         * The queue structure is created on the standard heap.
         */
        inline tListQueue () : tList () {}

        /**
         * @brief Constructor for the queue
         *
         * The queue structure is created in the pool <tt>ipool</tt>,
         * i.e. item descriptors are created there.
         */
        inline tListQueue (tMemPool &ipool) : tList (ipool) {}

        /**
         * @brief Destructor for the queue
         *
         * The queue structure is deallocated. The items themselves
         * are left untouched.
         */
        inline ~tListQueue () { reset(); }

        /**
         * @brief Removes all items from the queue.
         */
        inline void reset ()
            { tList::reset (); }
            
        /**
         * @brief Insert a new item to the queue (to the end).
         *
         * This function inserts item pointer <tt>data</tt> to the end of the
         * queue.
         */
        inline void insert (void *data)
            { append (data); }

        /**
         * @brief Insert a new item to the queue (to the beginning).
         *
         * This function inserts item pointer <tt>data</tt> to the beginning
         * of the queue.
         */
        inline void insert_first (void *data)
            { prepend (data); }

        /**
         * @brief Return the item pointer from the beginning of the queue (does no
         * removal).
         *
         * This functions returns the item pointer from the queue beginning. 
         * It does not modify the queue. If the queue is empty, NULL is 
         * returned.
         */
        inline void * peek ()
            { if (first) return first->data; else return NULL; }

        /**
         * @brief Removes the item pointer from the beginning of the queue. 
         *
         * This functions removes the item pointer from the queue beginning.
         * If the queue is empty, NULL is returned.
         */
        inline void *remove ()
            { return remove_first (); }

        /**
         * @brief Counts the number of items in the queue.
         */
        inline int count () const
            { return tList::count (); }

        /**
         * @brief Queries if the queue is empty.
         */
        inline bool is_empty () const
            { return count() ? false : true ; }
};       

}

#endif
