/**
 * @file
 * @brief Memory Pools
 * @author Martin Mares <mj@ucw.cz>
 * @author Ales Privetivy <ales@ucw.cz> 
 *
 * This file contains class for memory allocation from pools.  
 *
 */

#ifndef _ECHELOG_POOLS
#define _ECHELOG_POOLS

#include <string.h>

namespace lib {

struct tMemChunk; 

/**
 * @brief Class for a memory allocation from pools.
 * 
 * This object holds link-lists of blocks of memory from which is the memory
 * to be allocated. There are two link-lists. One for common memory
 * allocation. It consists of blocks of fixed size. Space is taken from
 * last block and if the block is full, a new one is allocated. Second
 * link-list is link-list of large blocks, i.e. if we want to allocate
 * more than one third of common block size than separate block is allocated
 * and added to the special list.
 * 
 * @note Mempools only allocate memory, only way for freeing it is to free the
 * whole mempool.
 */
class tMemPool {
    private:
        
        /** @brief Pointer to first free byte in actuall mempool block. */
        char *pfree;
        /** @brief Pointer to limit of current block. */
        char *last;
        /** @brief Pointer to first block of common link-list. */
        struct tMemChunk *first;
        /** @brief Pointer to current block in link-list. */
        struct tMemChunk *current;
        /** @brief Pointer to pointer on the last block of common link-list. */
        struct tMemChunk **plast;
        /** @brief Pointer to first block of large block link-list. */
        struct tMemChunk *first_large;
        /** @brief Fixed length of common block. */
        unsigned long chunk_size;
        /** @brief One third of fixed length of common block. */
        unsigned long threshold;

    public:

        /**
         * @brief Constructor for tMemPool class
         * 
         * Constructor initializes the <tt>tMemPool</tt> instance. Its only 
         * parameter <tt>init_size</tt> determines the fixed size of common
         * block.
         */
        tMemPool(unsigned long init_size);

        /**
         * @brief Destructor for tMemPool class. Frees all the memory.
         */
        ~tMemPool();

        /**
         * @brief Allocates memory from new block.
         * 
         * This function forces a new block creation and the requested 
         * <tt>size</tt> amount of memory is given from its beginning.
         * If the last block was not full, it can lead wasting of the
         * space. We should use @link lib::tMemPool::alloc() 
         * alloc()@endlink instead. Function @link lib::tMemPool::alloc()
         * alloc()@endlink first checks if the space couldn't be taken
         * from the old block and only if it is not possible calls
         * @link lib::tMemPool::alloc_slow() alloc_slow@endlink.
         * If the allocation id successful, a pointer to allocated data 
         * block is returned.  On error (not enough memory available)
         * calls <tt>die()</tt> with appropriate error message.
         */
        void *alloc_slow(unsigned long size); 

        /**
         * @brief Frees all the data from mempool.   
         */
        void free_pool();

        /**
         * @brief Allocates memory.   
         *
         * This function returns pointer to allocated data block of size
         * <tt>size</tt>. On error (not enough memory available)
         * calls <tt>die()</tt> with appropriate error message.
         */
        void *alloc(unsigned long size)
        {
            char *f  = pfree;
           
            // enlarge the size to ensure new address is aligned too
            size = (size + DATA_ALIGN - 1) & ~( DATA_ALIGN - 1 );

            if ( pfree + size > last)
                return alloc_slow( size );
            pfree += size;
            return f;
        }

        /**
         * @brief Allocates memory and fills it with zeroes.
         *
         * This function returns pointer to allocated data block of 
         * size <tt>size</tt>. After successful allocation the block
         * is blanked with zeroes. On error (not enough memory available)
         * calls <tt>die()</tt> with appropriate error message.
         */
        void *alloc_zero(unsigned long size)
        {
            void *x = alloc( size );
            memset(x, 0, size);
            return x;
        }
};

}

#endif
