/**
 * @file
 * @brief Syscall wrappers
 *
 * This part simplify the usage of various syscalls.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef __ECHELOG_S_IO_H
#define __ECHELOG_S_IO_H

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>

#ifndef O_SYNC
#  ifdef O_FSYNC
#    define O_SYNC  O_FSYNC
#  else
#    error O_SYNC must be defined
#  endif  /* O_FSYNC */
#endif  /* O_SYNC */

/** @brief Number of microseconds to wait before the next retry */
#define S_WAIT_BEFORE_RETRY 10  

/** @brief Maximal size of file io buffer. */
#define MAX_FILE_IOBUF_SIZE 64000

/** @brief Simple typedef to avoid <tt>struct stat</tt> and <tt>stat()</tt>
 * name clash. */
typedef struct stat struct_stat;

namespace lib {

/** 
 * @brief Write a buffer to a descriptor. 
 *
 * @param fd
 * File descriptor to write in.
 *
 * @param buf
 * Buffer to be written.
 *
 * @param count
 * Size of the buffer in bytes. Size must be greater or equal to zero.
 *
 * This function uses POSIX <tt>write()</tt> to write exactly count 
 * <tt>bytes</tt> to the given file descriptor from the buffer starting
 * at <tt>buf</tt>. All requested bytes from <tt>buf</tt> must be written,
 * otherwise an error  is reported.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 * 
 * @return
 * On success, <tt>count</tt> bytes is returned. On error, <tt>-1</tt> is
 * returned.
 */
ssize_t s_write (int fd, const void *buf, ssize_t count);


/** 
 * @brief Write a buffer to a descriptor with die on error.
 *
 * @param fd
 * File descriptor to write in.
 *
 * @param buf
 * Buffer to be written.
 *
 * @param count
 * Size of the buffer in bytes. Size must be greater or equal to zero.
 *
 * This function uses POSIX <tt>write()</tt> to write exactly <tt>count</tt>
 * bytes to the given file descriptor from the buffer starting at <tt>buf</tt>.
 * All requested bytes from <tt>buf</tt> must be written. On error, the
 * function <tt>die()</tt>s.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 */
void sd_write (int fd, const void *buf, ssize_t count);


/** @brief Read from a descriptor to a buffer.
 *
 * @param fd
 * File descriptor to read from.
 *
 * @param buf
 * Buffer to be read in. The buffer must be allocated and large enough
 * to hold read data.
 *
 * @param count
 * Number of bytes to read. It must be greater or equal to zero.
 *
 * This function uses POSIX <tt>read()</tt> to read at most <tt>count</tt>
 * bytes from the given file descriptor to the buffer starting at 
 * <tt>buf</tt>. 
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 * 
 * @return 
 * On success, number of bytes read is returned. This number should be
 * equal to <tt>count</tt>, but may be smaller if fewer bytes are actually
 * available in the file. Zero indicates there are no data available (EOF,
 * socket closed etc.). On error, <tt>-1</tt> is returned.
 */
ssize_t s_read (int fd, void *buf, ssize_t count);


/** @brief Read from a descriptor to a buffer with die on a error.
 *
 * @param fd
 * File descriptor to read from.
 *
 * @param buf
 * Buffer to be read in. The buffer must be allocated and large enough
 * to hold read data.
 *
 * @param count
 * Number of bytes to read. It must be greater or equal to zero.
 *
 * This function uses POSIX <tt>read()</tt> to read at most count bytes 
 * from the given file descriptor to the buffer starting at <tt>buf</tt>.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return
 * On success, number of bytes read is returned. This number should be equal
 * to <tt>count</tt>, but may be smaller if fewer bytes are actually available
 * in the file. Zero indicates there are no data available (EOF, socket closed
 * etc.). On error, the function <tt>die()</tt>s. 
 */
ssize_t sd_read (int fd, void *buf, ssize_t count);


/** 
 * @brief Recv from a descriptor to a buffer.
 *
 * @param fd
 * File descriptor to receive data from.
 *
 * @param buf
 * Buffer to be received. The buffer must be allocated and large enough
 * to hold received data.
 *
 * @param count
 * Number of bytes to receive. It must be greater or equal to zero.
 *
 * @param options One of the following bitwise or'd with zero (see man page for
 * details): <tt>MSG_OOB</tt>, <tt>MSG_PEEK</tt>, <tt>MSG_ERRQUEUE</tt>,
 * <tt>MSG_NOSIGNAL</tt>, <tt>MSG_ERRQUEUE</tt>.
 *
 * This function uses BSD <tt>recv()</tt> to receive at most <tt>count</tt>
 * bytes from the given file descriptor to the buffer starting at <tt>buf</tt>.
 *
 * @note Variable <tt>errno</tt> is set by the system call. Variable
 * <tt>ee_errno</tt> for queued errors is also set.
 * 
 * @return 
 * On success, number of bytes received is returned. This number should be
 * equal to <tt>count</tt>, but may be smaller if fewer bytes (of requested
 * type) are actually available in the <tt>fd</tt>. Zero indicates that there
 * are no data available (EOF, socket closed etc.). On error, <tt>-1</tt> is
 * returned.
 */
ssize_t s_recv (int fd, void *buf, ssize_t count, unsigned int options);


/** 
 * @brief Recv from a descriptor to a buffer with die on error.
 *
 * @param fd
 * File descriptor to receive data from.
 *
 * @param buf
 * Buffer to be received. The buffer must be allocated and large enough
 * to hold received data.
 *
 * @param count
 * Number of bytes to receive. It must be greater or equal to zero.
 *
 * @param options One of the following bitwise or'd with zero (see man page for
 * details): <tt>MSG_OOB</tt>, <tt>MSG_PEEK</tt>, <tt>MSG_ERRQUEUE</tt>,
 * <tt>MSG_NOSIGNAL</tt>, <tt>MSG_ERRQUEUE</tt>.
 *
 * This function uses BSD <tt>recv()</tt> to receive exactly <tt>count</tt>
 * bytes from the given file descriptor to the buffer starting at <tt>buf</tt>.
 *
 * @note Variable <tt>errno</tt> is set by the system call. Variable
 * <tt>ee_errno</tt> for queued errors is also set.
 *
 * @return 
 * On success, number of bytes received is returned. This number should
 * be equal to <tt>count</tt>, but may be smaller if fewer bytes
 * (of requested type) are actually available in the <tt>fd</tt>. Zero indicates
 * that there are no data available (EOF, socket closed etc.). On error, the
 * function <tt>die()</tt>s. 
 */
ssize_t sd_recv (int fd, void *buf, ssize_t count, unsigned int options);


/** 
 * @brief Reposition read/write file offset.
 *
 * @param fd
 * File descriptor to seek in.
 *
 * @param offset
 * Absolute position in the file.
 *
 * This function uses POSIX <tt>lseek()</tt> to set the new file position.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 * 
 * @return On success, the function returns <tt>offset</tt> bytes seek
 * from the beginning of the file. On failure, <tt>-1</tt> is returned.
 */
off_t s_seek (int fd, off_t offset);


/**
 * @brief Reposition read/write file offset with die on error.
 *
 * @param fd
 * File descriptor to seek in.
 *
 * @param offset
 * Absolute position in the file.
 *
 * This function uses POSIX <tt>lseek()</tt> to set a new file position.
 * The function <tt>die()</tt>s on failure.
 */
void sd_seek (int fd, off_t offset);

/** 
 * @brief Open/create a file.
 *
 * @param name
 * A name which refers to opened/created file.
 *
 * @param flags The parameter flags is one of <tt>O_RDONLY</tt>,
 * <tt>O_WRONLY</tt> or <tt>O_RDWR</tt> which  request opening  the file
 * read-only, write-only or read/write, respectively, bitwise-or'd with zero
 * or  more of the following: <tt>O_CREAT</tt>, <tt>O_EXCL</tt>,
 * <tt>O_NOCTTY</tt>, <tt>O_TRUNC</tt>, <tt>O_APPEND</tt>, <tt>O_NONBLOCK</tt>,
 * <tt>O_NDELAY</tt>, <tt>O_SYNC</tt>, <tt>O_NOFOLLOW</tt>,
 * <tt>O_DIRECTORY</tt>, <tt>O_LARGEFILE</tt>. See <tt>open()</tt> manual page
 * for more details.
 *
 * This function uses POSIX <tt>open()</tt> to open/create a file.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 * 
 * @return On success, the function returns the new file descriptor. 
 * On failure, <tt>-1</tt> is returned.
 */
int s_open (const char *name, int flags);

/** 
 * @brief Open/create a file with die on a error.
 *
 * @param name
 * A name which refers to opened/created file.
 *
 * @param flags The parameter flags is one of <tt>O_RDONLY</tt>,
 * <tt>O_WRONLY</tt> or <tt>O_RDWR</tt> which  request opening  the file
 * read-only, write-only or read/write, respectively, bitwise-or'd with zero or
 * more of the following: <tt>O_CREAT</tt>, <tt>O_EXCL</tt>, <tt>O_NOCTTY</tt>,
 * <tt>O_TRUNC</tt>, <tt>O_APPEND</tt>, <tt>O_NONBLOCK</tt>, <tt>O_NDELAY</tt>,
 * <tt>O_SYNC</tt>, <tt>O_NOFOLLOW</tt>, <tt>O_DIRECTORY</tt>,
 * <tt>O_LARGEFILE</tt>. See <tt>open()</tt> manual page for more details.
 *
 * This function uses POSIX <tt>open()</tt> to open/create a file. 
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return The function returns the new file descriptor. On error, the function 
 * <tt>die()</tt>s.
 */
int sd_open (const char *name, int flags);


/** 
 * @brief Open/create a file.
 *
 * @param name
 * A name which refers to opened/created file.
 *
 * @param flags The parameter flags is one of <tt>O_RDONLY</tt>,
 * <tt>O_WRONLY</tt> or <tt>O_RDWR</tt> which request opening  the file
 * read-only, write-only or read/write, respectively, bitwise-or'd with zero or
 * more of the following: <tt>O_CREAT</tt>, <tt>O_EXCL</tt>, <tt>O_NOCTTY</tt>,
 * <tt>O_TRUNC</tt>, <tt>O_APPEND</tt>, <tt>O_NONBLOCK</tt>, <tt>O_NDELAY</tt>,
 * <tt>O_SYNC</tt>, <tt>O_NOFOLLOW</tt>, <tt>O_DIRECTORY</tt>,
 * <tt>O_LARGEFILE</tt>. See <tt>open()</tt> manual page for more details.
 *
 * @param mode This arguments specifies the permissions to use in case a new
 * file is created. It is modified by the process's umask in the usual way
 * (i.e. the permissions of the created file are (mode & ~umask). However,
 * newly set permissions take effect for future access only (i.e. on next
 * <tt>open()</tt>). This parameter should be always set when <tt>O_CREAT</tt>
 * is set. Otherwise it is ignored.
 *      
 * This function uses POSIX <tt>open()</tt> to open/create a file.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 * 
 * @return On success, the function returns the new file descriptor. On failure,
 * <tt>-1</tt> is returned.
 *
 */
int s_open (const char *name, int flags, mode_t mode);


/**
 * @brief Open/create a file with die on error.
 *
 * @param name
 * A name which refers to opened/created file.
 *
 * @param flags The parameter flags is one of <tt>O_RDONLY</tt>,
 * <tt>O_WRONLY</tt> or <tt>O_RDWR</tt> which  request opening  the file
 * read-only, write-only or read/write, respectively, bitwise-or'd with zero
 * or  more of the following: <tt>O_CREAT</tt>, <tt>O_EXCL</tt>,
 * <tt>O_NOCTTY</tt>, <tt>O_TRUNC</tt>, <tt>O_APPEND</tt>, <tt>O_NONBLOCK</tt>,
 * <tt>O_NDELAY</tt>, <tt>O_SYNC</tt>, <tt>O_NOFOLLOW</tt>,
 * <tt>O_DIRECTORY</tt>, <tt>O_LARGEFILE</tt>. See <tt>open()</tt> manual page
 * for more details.
 *
 * @param mode This arguments specifies the permissions to use in case a new
 * file is created. It is modified by the process's umask in the usual way
 * (i.e. the permissions of the created file are (mode & ~umask). However,
 * newly set permissions take effect for future access only (i.e. on next
 * <tt>open()</tt>). This parameter should be always set when <tt>O_CREAT</tt>
 * is set. Otherwise it is ignored.
 *
 * This function uses POSIX <tt>open()</tt> to open/create a file.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return The function returns the new file descriptor. On failure, the
 * function <tt>die()</tt>s.
 */
int sd_open (const char *name, int flags, mode_t mode);


/** 
 * @brief Close a file descriptor.
 *
 * @param fd
 * File descriptor to be closed.
 *
 * This function uses POSIX <tt>close()</tt> to close a file descriptor.
 * After the descriptor is closed the system resources are freed and the
 * descriptor may be reused. Any locks held on the file are removed.
 *
 * @note Not checking for a return value may result in silent ignorance 
 * of data loss (<tt>write()</tt> may succeed even if the data has not been
 * written at the time). When closing the file it is guaranteed that all
 * required bytes are successfully written (flushed) or an error is reported.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 * 
 * @return On success, the function returns zero. On failure, <tt>-1</tt> is
 * returned.
 */
int s_close (int fd);


/** 
 * @brief Close a file descriptor with die on error.
 *
 * @param fd
 * File descriptor to be closed.
 *
 * This function uses POSIX <tt>close()</tt> to close a file descriptor.
 * After the descriptor is closed the system resources are freed and the
 * descriptor may be reused. Any locks held on the file are removed.
 * The function <tt>die()</tt>s on failure.
 *
 * @note Not checking for a return value may result in silent ignorance
 * of data loss (<tt>write()</tt> may succeed even if the data has not been
 * written at the time). When closing the file it is guaranteed that all
 * required bytes are successfully written (flushed) or an error is reported.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 */
void sd_close (int fd);


/** 
 * @brief Get a status of a file descriptor.
 *
 * @param fd
 * File descriptor to be examined. 
 * 
 * @param buf
 * This parameter is modified to contain the real result of <tt>fstat()</tt>
 * operation and as such must not be null. See <tt>struct stat</tt> definition
 * for detailed description.
 *
 * This function uses POSIX <tt>fstat()</tt> to get a status of a file
 * descriptor. See the manual page for more detailed description.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 * 
 * @return On success, the function returns zero. On failure, <tt>-1</tt> is
 * returned.
 */ 
int s_fstat (int fd, struct_stat *buf);


/**
 * @brief Get a status of a file descriptor with die on error.
 *
 * @param fd
 * File descriptor to be examined.
 * 
 * @param buf
 * This parameter is modified to contain the real result of <tt>fstat()</tt>
 * operation and as such must not be null. See <tt>struct stat</tt> definition
 * for detailed description.
 *
 * This function uses POSIX <tt>fstat()</tt> to get a status of a file 
 * descriptor. See the manual page for more detailed description. The function 
 * <tt>die()</tt>s  on failure.
 */ 
void sd_fstat (int fd, struct_stat *buf);


/** 
 * @brief Get a status of a file.
 *
 * @param filename
 * File name to be examined.
 * 
 * @param buf
 * This parameter is modified to contain the real result of <tt>stat</tt>
 * operation and as such must not be null. See <tt>struct stat</tt> definition
 * for detailed description.
 *
 * This function uses POSIX <tt>stat()</tt> to get a status of a file
 * descriptor. See the manual page for more detailed description.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 * 
 * @return On success, the function returns zero. On failure, <tt>-1</tt> is
 * returned.
 */ 
int s_stat (const char *filename, struct_stat *buf);


/** 
 * @brief Get a status of a file with die on error.
 *
 * @param filename
 * File name to be examined.
 * 
 * @param buf
 * This parameter is modified to contain the real result of <tt>stat</tt>
 * operation and as such must not be null. See <tt>struct stat</tt> definition
 * for detailed description.
 *
 * This function uses POSIX <tt>stat()</tt> to get a status of a file
 * descriptor. See the manual page for more detailed description. The function
 * <tt>die()</tt>s  on failure.
 */ 
void sd_stat (const char *filename, struct_stat *buf);

/** 
 * @brief Get file length.
 *
 * @param fd
 * File descriptor to be examined.
 *
 * This function uses POSIX <tt>fstat()</tt> to get a file length.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return On success, the file length in bytes is returned. On failure,
 * <tt>-1</tt> is returned.
 */
long s_filelen (int fd);


/** 
 * @brief Get file length with die on error.
 *
 * @param fd
 * File descriptor to be examined.
 *
 * This function uses POSIX <tt>fstat()</tt> to get a file length.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return On success, the file length in bytes is returned. On failure, the 
 * function <tt>die()</tt>s. 
 */
long sd_filelen (int fd);


/** 
 * @brief Get file length.
 *
 * @param filename
 * File name to be examined.
 *
 * This function uses POSIX <tt>stat()</tt> to get a file length. Note that this
 * function returns the size of the original file (ie. it may be a size
 * of symlink). If you want the size of the actual file use s_filelen()
 * instead.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return On success, the file length in bytes is returned. On failure,
 * <tt>-1</tt> is returned.
 */
long s_filelen (const char *filename);


/** 
 * @brief Get file length with die on error.
 *
 * @param filename
 * File name to be examined.
 *
 * This function uses POSIX <tt>stat()</tt> to get a file length. Note that this
 * function returns the size of the original file (ie. it may be a size
 * of symlink). If you want the size of the actual file use sd_filelen()
 * instead.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return On success, the file length in bytes is returned. On failure, the
 * function <tt>die()</tt>s.
 */
long sd_filelen (const char *filename);


/**
 * @brief Check user's permissions for a file.
 *
 * @param filename
 * Filename to be examined.
 *
 * @param mode Parameter <tt>mode</tt> is a mask consisting of one or more
 * flags <tt>R_OK</tt>, <tt>W_OK</tt>, <tt>X_OK</tt>, <tt>F_OK</tt>
 * bitwise-or'd with zero.
 *
 * This function uses POSIX <tt>access()</tt> to check whether the process 
 * would be allowed to read, write or execute the file (or if it merely exists).
 * If pathname is a symbolic link permissions of the file referred to by this
 * symbolic link are tested. Only access bits are checked.
 *
 * The check is done with the process's real uid and gid, rather than with the
 * effective ids as is done when actually attempting an operation. This is to
 * allow set-UID programs to easily determine the invoking user's authority.
 *
 * @note Using <tt>access()</tt> to check, if a user is authorized to e.g. open 
 * a file, before actually doing so using <tt>open()</tt>, creates a security
 * hole, because the user might exploit the short time interval between checking
 * and opening the file to manipulate it.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return 
 * If all requested permissions are granted zero is returned, if
 * permissions were denied <tt>1</tt> is returned. On failure, <tt>-1</tt> is
 * returned.
 */
int s_access (const char *filename, int mode);


/** 
 * @brief Check user's permissions for a file with die on failure.
 *
 * @param filename
 * Filename to be examined.
 *
 * @param mode Parameter <tt>mode</tt> is a mask consisting of one or more
 * flags <tt>R_OK</tt>, <tt>W_OK</tt>, <tt>X_OK</tt>, <tt>F_OK</tt>
 * bitwise-or'd with zero.
 *
 * This function uses POSIX <tt>access()</tt> to check whether the process
 * would be allowed to read, write or execute the file (or if it merely exists).
 * If pathname is a symbolic link permissions of the file referred to by this
 * symbolic link are tested. Only access bits are checked.
 *
 * The check is done with the process's real uid and gid, rather than with the
 * effective ids as is done when actually attempting an operation. This is to
 * allow set-UID programs to easily determine the invoking user's authority.
 *
 * The function <tt>die()</tt>s on failure or if the requested permissions 
 * were denied.
 *
 * @note Using <tt>access()</tt> to check, if a user is authorized to e.g. open
 * a file, before actually doing so using <tt>open()</tt>, creates a security
 * hole, because the user might exploit the short time interval between checking
 * and opening the file to manipulate it.
 */
void sd_access (const char *filename, int mode);


/** 
 * @brief Check effective user's permissions for a file.
 *
 * @param st
 * <tt>stat</tt> structure containing filesystem permissions of the file.
 *
 * @param mode Parameter <tt>mode</tt> is a mask consisting of one or more
 * flags <tt>R_OK</tt>, <tt>W_OK</tt>, <tt>X_OK</tt>, <tt>F_OK</tt>
 * bitwise-or'd with zero.
 *
 * This function tests the filesystem access bits to check whether the process
 * would be allowed to read, write or execute the file (or if it merely
 * exists). If pathname is a symbolic link permissions of the file referred to
 * by this symbolic link are tested.
 *
 * Unlike <tt>s_access()</tt>, this function does the checks with the
 * process's effective uid and process's effective gid. 
 *
 * @note The function merely checks the file permission bits. Trying to open
 * a file for writing may fail in unexpected situations, for example, if the
 * filesystem is mounted read-only.
 *
 * @note Using <tt>access()</tt> to check, if a user is authorized to e.g. open
 * a file, before actually doing so using <tt>open()</tt>, creates a security
 * hole, because the user might exploit the short time interval between checking
 * and opening the file to manipulate it.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return 
 * If all requested permissions are granted, zero is returned. If permissions
 * were denied <tt>1</tt> is returned. On failure, <tt>-1</tt> is returned.
 */
int s_eaccess (struct_stat *st, int mode);


/** 
 * @brief Check effective user's permissions for a file with die on error.
 *
 * @param st
 * <tt>stat</tt> structure containing filesystem permissions of the file.
 *
 * @param mode Parameter <tt>mode</tt> is a mask consisting of one or more
 * flags <tt>R_OK</tt>, <tt>W_OK</tt>, <tt>X_OK</tt>, <tt>F_OK</tt>
 * bitwise-or'd with zero.
 *
 * This function tests the filesystem access bits to check whether the process
 * would be allowed to read, write or execute the file (or if it merely
 * exists). If pathname is a symbolic link permissions of the file referred to
 * by this symbolic link are tested.
 *
 * Unlike <tt>sd_access()</tt>, this function does the checks with the
 * process's effective uid and process's effective gid.
 *
 * The function <tt>die()</tt>s on failure or if the requested permissions 
 * were denied.
 *
 * @note The function merely checks the file permission bits. Trying to open
 * a file for writing may fail in unexpected situations, for example, if the
 * filesystem is mounted read-only.
 *
 * @note Using <tt>access()</tt> to check, if a user is authorized to e.g. open
 * a file, before actually doing so using <tt>open()</tt>, creates a security
 * hole, because the user might exploit the short time interval between checking
 * and opening the file to manipulate it.
 */
void sd_eaccess (struct_stat *st, int mode);


/** 
 * @brief Check effective user's permissions for a file.
 *
 * @param filename
 * Filename to be examined.
 *
 * @param mode Parameter <tt>mode</tt> is a mask consisting of one or more
 * flags <tt>R_OK</tt>, <tt>W_OK</tt>, <tt>X_OK</tt>, <tt>F_OK</tt>
 * bitwise-or'd with zero.
 *
 * This function tests the filesystem access bits to check whether the process
 * would be allowed to read, write or execute the file (or if it merely
 * exists). If pathname is a symbolic link permissions of the file referred to
 * by this symbolic link are tested.
 *
 * Unlike <tt>s_access()</tt>, this function does the checks with the
 * process's effective uid and process's effective gid. 
 *
 * @note The function merely checks the file permission bits. Trying to open
 * a file for writing may fail in unexpected situations, for example, if the
 * filesystem is mounted read-only.
 *
 * @note Using <tt>access()</tt> to check, if a user is authorized to e.g. open
 * a file, before actually doing so using <tt>open()</tt>, creates a security
 * hole, because the user might exploit the short time interval between checking
 * and opening the file to manipulate it.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return If all requested permissions are granted, zero is returned.
 * If permissions were denied <tt>1</tt> is returned. On failure, <tt>-1</tt> is returned.
 */
int s_eaccess (const char *filename, int mode);


/** 
 * @brief Check effective user's permissions for a file with die on error.
 *
 * @param filename
 * Filename to be examined.
 *
 * @param mode Parameter <tt>mode</tt> is a mask consisting of one or more
 * flags <tt>R_OK</tt>, <tt>W_OK</tt>, <tt>X_OK</tt>, <tt>F_OK</tt>
 * bitwise-or'd with zero.
 *
 * This function tests the filesystem access bits to check whether the process
 * would be allowed to read, write or execute the file (or if it merely
 * exists). If pathname is a symbolic link permissions of the file referred to
 * by this symbolic link are tested.
 *
 * Unlike <tt>sd_access()</tt>, this function does the checks with the
 * process's effective uid and process's effective gid.
 *
 * The function <tt>die()</tt>s on failure or if the requested permissions
 * were denied.
 *
 * @note The function merely checks the file permission bits. Trying to open
 * a file for writing may fail in unexpected situations, for example, if the
 * filesystem is mounted read-only.
 *
 * @note Using <tt>access()</tt> to check, if a user is authorized to e.g. open
 * a file, before actually doing so using <tt>open()</tt>, creates a security
 * hole, because the user might exploit the short time interval between checking
 * and opening the file to manipulate it.
 */
void sd_eaccess (const char *filename, int mode);

/** 
 * @brief Safely create a unique temporary file.
 *
 * @param filename 
 * Filename template - the last six characters must be XXXXXX which will be
 * REPLACED with a string which makes the file unique. Thus this string
 * must not be a string constant.
 *
 * The function calls <tt>mkstemp()</tt> to generate a unique temporary file
 * name from a template. The file is then created with mode read/write and
 * permissions <tt>0600</tt>. The  file  is  opened  with the <tt>O_EXCL</tt>
 * flag, guaranteeing that when <tt>mkstemp()</tt> returns successfully we are
 * the only user.
 *
 * @note Variable <tt>errno</tt> is set by the system call. 
 *
 * @return On success, the function returns the new file descriptor. 
 * The <tt>filename</tt> parameter is modified to contain the temporary 
 * file's name. On failure, <tt>-1</tt> is returned.
 */
int s_mkstemp (char *filename);

/** 
 * @brief Safely create a unique temporary file with die on error.
 *
 * @param filename
 * Filename template - the last six characters must be XXXXXX which will be
 * REPLACED with a string which makes the file unique. Thus this string
 * <em>must not be a string constant</em>.
 *
 * The function calls <tt>mkstemp()</tt> to generate a unique temporary file
 * name from a template. The file is then created with mode read/write and
 * permissions <tt>0600</tt>. The  file  is  opened  with the <tt>O_EXCL</tt>
 * flag, guaranteeing that when <tt>mkstemp()</tt> returns successfully we are
 * the only user.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return On success, the function returns the new file descriptor.
 * The <tt>filename</tt> parameter is modified to contain the temporary
 * file's name. On failure, the function <tt>die()</tt>s.
 */
int sd_mkstemp (char *filename);

/** 
 * @brief Check if the a group belongs to program's supplementary groups.
 *
 * @param n_gid
 * Group number we are asking about.
 *
 * The function uses POSIX <tt>getgroups()</tt> to find out whether the 
 * specified group belongs or not to process's supplementary groups.
 *
 * @note The current process's euid/egid is cached from previous library calls.
 * To ensure that euid/egid holds the proper value after a direct uid/gid
 * manipulation (not using echelog library functions) you need to call
 * <tt>s_reset_my_euid()</tt>/<tt>s_reset_my_egid()</tt>.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return If the given group is a process's supplementary group <tt>1</tt> is
 * returned. If it isn't, zero is returned. On failure, <tt>-1</tt> is
 * returned.
 */
int s_is_my_group (gid_t n_gid);


/** 
 * @brief Check if the a group belongs to program's supplementary groups with
 * die on the failure.
 *
 * @param n_gid
 * Group number we are asking about.
 *
 * The function uses POSIX <tt>getgroups()</tt> to find out whether the
 * specified group belongs or not to process's supplementary groups.
 *
 * @note The current process's euid/egid is cached from previous library calls.
 * To ensure that euid/egid holds the proper value after a direct uid/gid
 * manipulation (not using echelog library functions) you need to call
 * <tt>s_reset_my_euid()</tt>/<tt>s_reset_my_egid()</tt>.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return
 * If the given group is a process's supplementary group, <tt>true</tt> is 
 * returned, if it isn't, <tt>false</tt> is returned. The function 
 * <tt>die()</tt>s on failure.
 */
bool sd_is_my_group (gid_t n_gid);


/** 
 * @brief Clear the library cached euid.
 *
 * The current process's euid is cached from previous library calls.
 * To ensure that euid used for echelog library calls holds the
 * proper value after a direct uid manipulation (not using echelog library
 * functions) you need to call this function.The function <tt>die()</tt>s 
 * on failure.
 */
void sd_reset_my_euid ();


/** 
 * @brief Clear the library cached egid.
 *
 * The current process's egid is cached from previous library calls.
 * To ensure that egid used for echelog library calls holds the
 * proper value after a direct gid manipulation (not using echelog library
 * functions) you need to call this function. The function <tt>die()</tt>s
 * on failure.
 */
void sd_reset_my_egid ();


/** 
 * @brief Set a new user's identity.
 * 
 * @param n_uid
 * User's ID to be set.
 *
 * This function uses POSIX <tt>setuid()</tt> to set the effective user 
 * ID of the current process. If the effective userid of the caller is root,
 * the real and saved user ID's are also set.
 *
 * If the user is root or the program is setuid root, special care must be
 * taken. The <tt>setuid()</tt> function checks the effective uid of the 
 * caller and if it is the superuser, all process related user ID's are 
 * set to uid. After this  has occurred,  it is impossible for the program
 * to regain root privileges.
 *
 * @note A setuid-root program wishing to temporarily drop root privileges
 * does have to use non-POSIX (but likely implemented) BSD call 
 * <tt>seteuid()</tt>.
 *
 * @note Only superuser may use this function.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 * 
 * @return On success, zero is returned. On failure, <tt>-1</tt> is returned.
 */
int s_setuid (uid_t n_uid);


/** 
 * @brief Set a new user's identity with die on error.
 * 
 * @param n_uid
 * User's ID to be set.
 *
 * This function uses POSIX <tt>setuid()</tt> to set the effective user
 * ID of the current process. If the effective userid of the caller is root,
 * the real and saved user ID's are also set.
 *
 * If the user is root or the program is setuid root, special care must be
 * taken. The <tt>setuid()</tt> function checks the effective uid of the
 * caller and if it is the superuser, all process related user ID's are
 * set to uid. After this  has occurred,  it is impossible for the program
 * to regain root privileges.
 *
 * The function <tt>die()</tt>s on failure.
 *
 * @note A setuid-root program wishing to temporarily drop root privileges
 * does have to use non-POSIX (but likely implemented) BSD call
 * <tt>seteuid()</tt>.
 *
 * @note Only superuser may use this function.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 */
void sd_setuid (uid_t n_uid);


/** 
 * @brief Set a new user's group identity.
 * 
 * @param n_gid
 * User's group ID to be set.
 *
 * This function uses POSIX <tt>setgid()</tt> to set the effective group 
 * ID of the current process. If the effective caller is root, the real 
 * and saved user's group ID's are also set.
 *
 * If the user is root or the program is setuid root, special care must be
 * taken. The <tt>setgid()</tt> function checks the effective gid of the 
 * caller and if it is the superuser, all process related user's group 
 * ID's are set to gid. After this has occurred, it is impossible for 
 * the program to regain root privileges.
 *
 * @note A setgid-root program wishing to temporarily drop root privileges
 * does have to use non-POSIX (but likely implemented) BSD  
 * <tt>setegid()</tt> call.
 *
 * @note Only superuser may use this function.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 * 
 * @return On success, the function returns zero. On failure, <tt>-1</tt> is returned.
 */
int s_setgid (gid_t n_gid);


/** 
 * @brief Set a new user's group identity with die on error.
 * 
 * @param n_gid
 * User's group ID to be set.
 *
 * This function uses POSIX <tt>setgid()</tt> to set the effective group
 * ID of the current process. If the effective caller is root, the real
 * and saved user's group ID's are also set.
 *
 * If the user is root or the program is setuid root, special care must be
 * taken. The <tt>setgid()</tt> function checks the effective gid of the
 * caller and if it is the superuser, all process related user's group
 * ID's are set to gid. After this has occurred, it is impossible for
 * the program to regain root privileges.
 *
 * The function <tt>die()</tt>s on failure.
 *
 * @note A setgid-root program wishing to temporarily drop root privileges
 * does have to use non-POSIX (but likely implemented) BSD
 * <tt>setegid()</tt> call.
 *
 * @note Only superuser may use this function.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 */
void sd_setgid (gid_t n_gid);


/** 
 * @brief Set a new effective user's identity.
 * 
 * @param n_uid
 * User's ID to be set.
 *
 * This function uses non-POSIX (but likely implemented) BSD call 
 * <tt>seteuid()</tt> to set the effective user ID of the current process.
 *  
 * Unprivileged user processes may only set the effective user ID to the real
 * user ID, the effective user ID or the saved user ID. 
 *
 * This function can be used to temporarily drop and then regain superuser
 * privileges.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return On success, the function returns zero. On failure, <tt>-1</tt> is returned.
 */
int s_seteuid (uid_t n_uid);


/** 
 * @brief Set a new effective user's identity with die on error.
 * 
 * @param n_uid
 * User's ID to be set.
 *
 * This function uses non-POSIX (but likely implemented) BSD call
 * <tt>seteuid()</tt> to set the effective user ID of the current process.
 *
 * Unprivileged user processes may only set the effective user ID to the real
 * user ID, the effective user ID or the saved user ID.
 *
 * This function can be used to temporarily drop and then regain superuser
 * privileges.
 *
 * This function <tt>die()</tt>s on failure.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 */
void sd_seteuid (uid_t n_uid);


/** 
 * @brief Set a new effective user's group identity.
 *
 * @param n_uid
 * User's group ID to be set.
 *
 * This function uses non-POSIX (but likely implemented) BSD call 
 * <tt>setguid()</tt> to set the effective user's group ID of the 
 * current process.
 *  
 * Unprivileged user processes may only set the effective user's group ID to
 * the real user's group ID, the effective user's group ID or the saved user's
 * group ID. 
 *
 * This function can be used to temporarily drop and then regain superuser
 * privileges.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return On success, the function returns zero. On failure, <tt>-1</tt> is
 * returned.
 */
int s_setegid (uid_t n_uid);


/** 
 * @brief Set a new effective user's group identity with die on error.
 *
 * @param n_uid
 * User's group ID to be set.
 *
 * This function uses non-POSIX (but likely implemented) BSD call
 * <tt>setguid()</tt> to set the effective user's group ID of the
 * current process.
 *
 * Unprivileged user processes may only set the effective user's group ID to
 * the real user's group ID, the effective user's group ID or the saved user's
 * group ID.
 *
 * This function can be used to temporarily drop and then regain superuser
 * privileges.
 *
 * This function <tt>die()</tt>s on failure.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 */
void sd_setegid (uid_t n_uid);


/** 
 * @brief Drop all process's supplementary groups.
 *
 * This function uses non-POSIX (but likely implemented) BSD call 
 * <tt>setgroups()</tt> to drop all process's supplementary groups.
 *
 * This is used for security reasons to relinquish unneeded permissions.
 *
 * @note Only superuser may use this function.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * On success, the function returns zero. On failure, <tt>-1</tt> is returned.
 */
int s_drop_supp_groups ();


/** 
 * @brief Drop all process's supplementary groups with die on error.
 *
 * This function uses non-POSIX (but likely implemented) BSD call 
 * <tt>setgroups()</tt> to drop all process's supplementary groups.
 *
 * This is used for security reasons to relinquish unneeded permissions.
 *
 * This function <tt>die()</tt>s on failure.
 *
 * @note Only superuser may use this function.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 */
void sd_drop_supp_groups ();


/** 
 * @brief Set a new single user's group identity.
 *
 * @param n_gid
 * User's group ID to be set.
 *
 * This function uses POSIX <tt>setgid()</tt> and non-POSIX (but 
 * likely implemented) BSD call <tt>setgroups()</tt> to set a new 
 * user's group identity and simultaneously drop all other 
 * (supplementary) group id's.
 *
 * This is used for security reasons to relinquish unneeded permissions.
 *
 * @note Only superuser may use this function.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 *
 * @return On success, the function returns zero. On failure, <tt>-1</tt> is
 * returned.
 */
int s_set_single_gid (gid_t n_gid);


/** 
 * @brief Set a new single user's group identity with die on error.
 *
 * @param n_gid
 * User's group ID to be set.
 *
 * This function uses POSIX <tt>setgid()</tt> and non-POSIX (but
 * likely implemented) BSD call <tt>setgroups()</tt> to set a new
 * user's group identity and simultaneously drop all other
 * (supplementary) group id's.
 *
 * This is used for security reasons to relinquish unneeded permissions.
 *
 * This function <tt>die()</tt>s on failure.
 *
 * @note Only superuser may use this function.
 *
 * @note Variable <tt>errno</tt> is set by the system call.
 */
void sd_set_single_gid (gid_t n_gid);

/** 
 * @brief Translate the username into the system uid.
 *
 * @param username
 * Username to be translated.
 *
 * @param uid
 * A reference to an object where to save the result.
 *
 * This function uses <tt>getpwent()</tt> call to translate <tt>username</tt>
 * to <tt>uid</tt>.
 *
 * @note Call <tt>getpwent()</tt> is supposed to return a pointer to an
 * internal static object so the returned pointer is not deallocated 
 * thereafter.
 *
 * @return
 * The function return zero on success. If the <tt>username</tt> is not found
 * or other error occurres, <tt>-1</tt> is returned.
 */
int s_name2uid (const char *username, uid_t &uid);


/** 
 * @brief Translate the the system uid into user's name.
 *
 * @param uid
 * System's uid to be translated.
 *
 * @param username
 * A reference to an object where to save the result.
 *
 * This function uses <tt>getpwent()</tt> call to translate <tt>uid</tt> to
 * <tt>username</tt>. The memory is automatically allocated for 
 * <tt>username</tt> and can be released with <tt>xfree()</tt>.
 *
 * @note Call <tt>getpwent()</tt> is supposed to return a pointer to an
 * internal static object so the returned pointer is not deallocated
 * thereafter.
 *
 * @return
 * The function return zero on success. If the <tt>username</tt> is not found
 * or other error occurres, <tt>-1</tt> is returned.
 */
int s_uid2name (const uid_t uid, char *&username);


/** 
 * @brief Translate the the system uid into user's name (result is put into
 * a pool).
 *
 * @param uid
 * System's uid to be translated.
 *
 * @param username
 * A reference to an object where to save the result.
 *
 * @param pool
 * Memory pool for a new string allocation.
 *
 * This function uses <tt>getpwent()</tt> call to translate <tt>uid</tt>
 * to <tt>username</tt>. The memory is automatically allocated for 
 * <tt>username</tt> and can be released only by freeing the particular
 * memory pool.
 *
 * @note Call <tt>getpwent()</tt> is supposed to return a pointer to an
 * internal static object so the returned pointer is not deallocated
 * thereafter.
 *
 * @return
 * The function return zero on success. If the username is not found
 * or other error occurres, <tt>-1</tt> is returned.
 */
int s_uid2name (const uid_t uid, char *&username, tMemPool &pool);


/** 
 * @brief Detach from a controlling terminal.
 *
 * This function uses a multiple <tt>fork()</tt> technique to detach from
 * a controlling terminal. Upon a successful return the controlling
 * terminal can never be regained.
 *
 * @return
 * The function return zero on success. On failure, <tt>-1</tt> is returned.
 */
int s_detach_from_ctrl_terminal ();


/** 
 * @brief Detach from a controlling terminal with die on error.
 *
 * This function uses a multiple <tt>fork()</tt> technique to detach from
 * a controlling terminal. Upon a successful return the controlling
 * terminal can never be regained. This function <tt>die()</tt>s on failure.
 */
void sd_detach_from_ctrl_terminal ();


/** 
 * @brief Copy a file.
 *
 * @param srcpath
 * Path to file to be copied.
 *
 * @param destpath
 * Destination file path.
 *
 * This function copies the source to destination file in
 * <tt>MAX_FILE_IOBUF_SIZE</tt> bytes long blocks.
 *
 * @return
 * The function return zero on success. On failure, <tt>-1</tt> is returned.
 */
int s_copy_file (const char *destpath, const char *srcpath);


/** 
 * @brief Copy a file with die on error.
 *
 * @param srcpath
 * Path to file to be copied.
 *
 * @param destpath
 * Destination file path.
 *
 * This function copies the source to destination file in
 * <tt>MAX_FILE_IOBUF_SIZ</tt> bytes long blocks. The function <tt>die()</tt>s
 * on failure.
 */
void sd_copy_file (const char *destpath, const char *srcpath);


/** 
 * @brief Move a file.
 *
 * @param srcpath
 * Path to file to be moved.
 *
 * @param destpath
 * Destination file path.
 *
 * This function moves the source to destination file in <tt>MAX_FILE_IOBUF_SIZ</tt>
 * bytes long blocks.
 *
 * @return The function return zero on success. On failure, <tt>-1</tt> is
 * returned.
 */
int s_move_file (const char *destpath, const char *srcpath);


/** 
 * @brief Move a file with die on error.
 *
 * @param srcpath
 * Path to file to be moved.
 *
 * @param destpath
 * Destination file path.
 *
 * This function moves the source to destination file in
 * <tt>MAX_FILE_IOBUF_SIZ</tt> bytes long blocks. The function <tt>die()</tt>s
 * on failure.
 */
void sd_move_file (const char *destpath, const char *srcpath);


/**
 * @brief Tests if program accepts signals.
 * 
 * Tests if a signal can be delivered to the program of PID <tt>pid</tt>. 
 * The function clears out the PID entry if the process is a zombie. The
 * program's exit status is written to <tt>status</tt> memory location. 
 * If this parameter is a NULL pointer the returned status is discarded.
 *
 * @return The function returns <tt>0</tt> if signals can be successfully
 * send or <tt>1</tt> if the process already terminated.
 * On error the function returns <tt>-1</tt>.
 */
int s_test_signal_delivery (pid_t pid, int *status);

/** 
 * @brief Removes a directory with files and subdirectories within it.
 * 
 * @param dir_name 
 * Path of directory to remove.
 *
 * This function removes directory with files and subdirectories within it
 * recursively. It does work correctly, when there is a cycle in the directory
 * subtree for delete (on the other hand it is safe, when we use it on 
 * directories made by us).
 *
 * @return Returns zero on success, otherwise <tt>-1</tt>.
 */
int remove_directory (const char *dir_name);

/** 
 * @brief Renames directory. 
 *
 * @param dir_name
 * Path of directory to move.
 *
 * @param new_dir
 * Path of new directory.
 *
 * This function moves directory with files and subdirectories within it to
 * the new path.
 *
 * @return Returns zero on success, otherwise <tt>-1</tt>.
 */
int move_directory (const char *dir_name, const char *new_dir);

}

#endif
