/**
 * @file
 * @brief String manipulation.
 *
 * This file contains functions which simplify the manipulation 
 * with strings.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef __ECHELOG_STRING_H
#define __ECHELOG_STRING_H

#include <stdarg.h>
#include <stdlib.h>

namespace lib {

/** 
 * @brief Get string length of a limited-size string.
 *
 * @param s
 * Pointer to a fixed-size string to be examined.
 *
 * @param max
 * The fixed string boundary.
 *
 * This function returns the number of characters in the string
 * pointed to by <tt>s</tt>, not including the  terminating <tt>'\0'</tt>
 * character, but at most <tt>maxlen</tt>.
 *
 * @return 
 * This function returns <tt>strlen(s)</tt>, if that is less than 
 * <tt>maxlen</tt>, or maxlen if there is no <tt>'\0'</tt> character among 
 * the first <tt>maxlen</tt> characters pointed to by <tt>s</tt>.
 *
 * @see strlen()
 */
size_t s_strnlen (const char *s, size_t max);

/** @brief Duplicate a string.
 *
 * @param s
 * Pointer to string to be duplicated.
 *
 * This  function  returns a pointer to a new string
 * which is a duplicate of the string <tt>s</tt>. The allocated memory
 * can be released with <tt>xfree()</tt>.
 * 
 * The function <tt>die()</tt>s on failure.
 *
 * @return
 * This function returns a pointer to the duplicated string.
 *
 * @see <tt>s_strdup_portion()</tt>, <tt>s_strndup()</tt>
 */
char *s_strdup (const char *s);

/** 
 * @brief Duplicate a string in memory pool.
 *
 * @param s
 * Pointer to string to be duplicated
 *
 * @param pool
 * Memory pool for a new string allocation
 *
 * This  function  returns a pointer to a new string
 * which is a duplicate of the string <tt>s</tt>. The allocated memory
 * can be released only by freeing the particular memory pool.
 * 
 * The function <tt>die()</tt>s on failure.
 *
 * @return
 * This function returns a pointer to the duplicated string.
 *
 * @see <tt>s_strdup_portion()</tt>, <tt>s_strndup()</tt>
 */
char *s_strdup (const char *s, lib::tMemPool &pool);

/** @brief 
 * Duplicate a string portion.
 *
 * @param s
 * Pointer to string portion to be duplicated
 *
 * @param len
 * String portion length
 *
 * This function returns a pointer to a new string which is a duplicate 
 * of the string portion <tt>s</tt> of <tt>len</tt> characters. The allocated
 * memory can be released with <tt>xfree()</tt>.
 * 
 * The function <tt>die()</tt>s on failure.
 *
 * @return
 * This function returns a pointer to the duplicated string.
 *
 * @see <tt>s_strdup()</tt>, <tt>s_strndup()</tt>
 */
char *s_strdup_portion (const char *s, const long len);

/** 
 * @brief Duplicate a string portion in memory pool.
 *
 * @param s
 * Pointer to string to be duplicated
 *
 * @param len
 * String portion length
 *
 * @param pool
 * Memory pool for a new string allocation
 *
 * This function returns a pointer to a new string which is a duplicate of
 * the string portion <tt>s</tt> of <tt>len</tt> characters. The allocated
 * memory can be released only by freeing the particular memory pool.
 * 
 * The function <tt>die()</tt>s on failure.
 *
 * @return
 * This function returns a pointer to the  duplicated string.
 *
 * @see <tt>s_strdup()</tt>, <tt>s_strndup()</tt>
 */
char *s_strdup_portion (const char *s, const long len, lib::tMemPool &pool);

/** 
 * @brief Duplicate a string portion limited by maxlen.
 *
 * @param s
 * Pointer to string portion to be duplicated
 *
 * @param maxlen
 * Maximal string portion length
 *
 * This function  returns a pointer to a new string
 * which is a duplicate of the string <tt>s</tt> and which has at most
 * <tt>maxlen</tt> characters. The allocated memory can be released 
 * with <tt>xfree()</tt>.
 * 
 * The function <tt>die()</tt>s on failure.
 *
 * @return
 * This function returns a pointer to the duplicated string.
 *
 * @see <tt>s_strdup()</tt>, <tt>s_strdup_portion()</tt>
 */
char *s_strndup (const char *s, const long maxlen);

/** 
 * @brief Duplicate a string portion limited by maxlen in memory pool.
 *
 * @param s
 * Pointer to string to be duplicated
 *
 * @param maxlen
 * Maximal string portion length
 *
 * @param pool
 * Memory pool for a new string allocation
 *
 * This function  returns a pointer to a new string
 * which is a duplicate of the string portion <tt>s</tt> and which has at most
 * <tt>maxlen</tt> characters. The allocated memory can be released only by 
 * freeing the particular memory pool.
 * 
 * The function <tt>die()</tt>s on failure.
 *
 * @return
 * This function returns a pointer to the duplicated string.
 *
 * @see <tt>s_strdup()</tt>, <tt>s_strdup_portion()</tt>
 */
char *s_strndup (const char *s, const long maxlen, lib::tMemPool &pool);

/** 
 * @brief Trim a string (result is a new copy in pool).
 * 
 * @param s
 * Pointer to string to be trimmed
 * 
 * @param pool
 * Memory pool for a new string allocation
 * 
 * This function returns a pointer to a new string which is a 
 * trimmed duplicate of the string <tt>s</tt>. Blank characters 
 * (as reported by <tt>isspace()</tt> call at the beginning and 
 * at the end of the input string <tt>s</tt> are cut off. The 
 * resulting string is null-terminated. The allocated memory 
 * can be released only by freeing the particular memory pool.
 *
 * @return
 * The function returns a pointer to the trimmed string duplicate.
 */
char *s_trim (const char *s, tMemPool &pool);

/** @brief Trim a string.
 * 
 * @param s
 * Pointer to string to be trimmed
 * 
 * This function returns a pointer to a new string which is a 
 * trimmed duplicate of the string <tt>s</tt>. Blank characters 
 * (as reported by <tt>isspace()</tt> at the beginning and at the end 
 * of the input string <tt>s</tt> are cut off. Space for the resulting
 * string is allocated automatically, the string is null-terminated.
 * The allocated memory can be released with <tt>xfree()</tt>.
 *
 * @return
 * The function returns a pointer to the trimmed string duplicate.
 */
char *s_trim (const char *s);


/** @brief Locate a substring.
 *
 * @param s
 * String portion to find in
 * 
 * @param s_len
 * String portion s length
 * 
 * @param p
 * Substring to be found
 * 
 * @param p_len
 * Substring p length (without terminating null)
 *
 * This function finds the first occurrence of the substring in the 
 * specified string portion <tt>s</tt>.
 *
 * @return
 * This function returns a pointer to the beginning
 * of the substring, or NULL if the substring is not found.
 */
const char *s_strnstr(const char *s, size_t s_len, const char *p, size_t p_len);

/** @brief Locate a substring
 *
 * @param s
 * String to find in
 * 
 * @param p
 * Substring to be found
 * 
 * @param maxlen
 * Maximal length of string portion of <tt>s</tt>
 *
 * The function finds the first occurrence of the substring in the 
 * specified string (uses at most maxlen first characters of <tt>s</tt>).
 *
 * @return
 * The function returns a pointer to the beginning
 * of the substring, or NULL if the substring is not found in the first
 * maxlen characters of <tt>s</tt>.
 */
const char *s_strnstr (const char *s, const char *p, const size_t maxlen);

/** 
 * @brief Replace all occurrences of substring in a given string.
 *
 * @param str
 * The reference string
 *
 * @param to_replace
 * String pattern to be replaced
 *
 * @param replace_string
 * String to be replaced with
 *
 * @param maxlen
 * Maximal length of string portion of <tt>s</tt>
 *
 * The function replaces all occurrences in <tt>str</tt> within <tt>maxlen</tt> 
 * characters of the substring <tt>to_replace</tt> with a 
 * <tt>replace_string</tt>. Space for the resulting string is allocated 
 * automatically, the string is null-terminated. The allocated memory can be
 * released with <tt>xfree()</tt>.
 *
 * @return
 * The function returns a copy of reference string with all string patterns
 * replaced by the substitute string.
 */
char *s_strnreplace (const char *str, const char *to_replace, 
        const char *replace_string, const long maxlen);

/** 
 * @brief Replace all occurrences of substring in a given string (result is put
 * to a memory pool)
 *
 * @param str
 * The reference string
 *
 * @param to_replace
 * String pattern to be replaced
 *
 * @param replace_string
 * String to be replaced with
 *
 * @param maxlen
 * Maximal length of string portion of <tt>s</tt>
 *
 * @param pool
 * Memory pool for a new string allocation
 *
 * This function replaces all occurrences in <tt>str</tt> within <tt>maxlen</tt>
 * characters of the substring <tt>to_replace</tt> with a 
 * <tt>replace_string</tt>. Space for the resulting string is allocated 
 * automatically in the designated memory pool, the string is null-terminated. 
 * The allocated memory can be released only by freeing the
 * particular memory pool.
 *
 * @return
 * The function returns a copy of reference string with all string patterns
 * replaced by the substitute string.
 */
char *s_strnreplace (const char *str, const char *to_replace, 
        const char *replace_string, const long maxlen, tMemPool &pool);


/** @brief Concatenate strings.
 *
 * @param str1
 * First argument to be concatenated
 *
 * @param ...
 * Other arguments to be concatenated, last parameter must be NULL.
 *
 * This function can be used to concatenate up to <tt>#MAX_CONCAT</tt> strings
 * specified in the function call. Space for the resulting string
 * is allocated automatically, the string is null-terminated. 
 * The allocated memory can be released with <tt>xfree()</tt>.
 *
 * The function <tt>die()</tt>s on failure.
 *
 * @return
 * This function returns a string which is a concatenation of all given 
 * parameters.
 */
char *s_strconcat (const char *str1, ...);

/** @brief Concatenate strings (result is put to a pool).
 *
 * @param pool
 * Memory pool for a new string allocation
 * 
 * @param str1
 * First argument to be concatenated
 *
 * @param ...
 * Other arguments to be concatenated, last parameter must be NULL.
 *
 * This function can be used to concatenate up to <tt>#MAX_CONCAT</tt> strings
 * specified in the function call. Space for the resulting string
 * is allocated automatically in the designated memory pool, the string 
 * is null-terminated. The allocated memory can be released only by freeing
 * the particular memory pool.
 *
 * The function <tt>die()</tt>s on failure.
 *
 * @return
 * This function returns a string which is a concatenation of all given 
 * parameters.
 */
char *s_strconcat (tMemPool &pool, const char *str1, ...);

/** 
 * @brief Separate phrases from a string.
 *
 * @param list 
 * List container for the result.
 *
 * @param str
 * String argument to be tokenized. It is expected to be well-formed.
 *
 * This function can be used to separate phrases from a string. A phrase
 * is a single word or a one or more words enclosed in a pair of double
 * quotes. Blank characters which do not appear as a part of a phrase
 * are eliminated. The resulting null terminated strings are placed 
 * into <tt>list</tt>.
 *
 * Space for the resulting strings is allocated automatically. The 
 * allocated memory can be released with <tt>xfree()</tt>.
 *
 * @return
 * This function returns zero on success. On failure, <tt>-1</tt> is returned.
 */
int s_strpsep (tList &list, const char *str);

/**
 * @brief Separate phrases from a string (result is put into memory pool).
 *
 * @param pool
 * Memory pool for a new string allocation
 *
 * @param list 
 * List container for the result.
 * 
 * @param str
 * String argument to be tokenized. It is expected to be well-formed.
 *
 * This function can be used to separate phrases from a string. A phrase
 * is a single word or a one or more words enclosed in a pair of double
 * quotes. Blank characters which do not appear as a part of a phrase
 * are eliminated. The resulting null terminated strings are placed 
 * into <tt>list</tt>.
 *
 * Space for resulting strings is allocated automatically in the 
 * designated memory pool. The allocated memory can be released only by
 * freeing the particular memory pool.
 *
 * @return
 * This function returns zero on success. On failure, <tt>-1</tt> is returned.
 */
int s_strpsep (tList &list, const char *str, tMemPool &pool);

/** 
 * @brief Separate tokens from a string.
 *
 * @param pool
 * Memory pool for a new string allocation
 *
 * @param list 
 * List container for the result.
 * 
 * @param str
 * String argument to be tokenized. It is expected to be well-formed.
 *
 * @param delim
 * String symbol delimiters. It is expected to be a non-null string.
 *
 * This function can be used to separate tokens from a string. A `token' 
 * is a nonempty string of characters not occurring in the string 
 * <tt>delim</tt>. If the string <tt>str</tt> is NULL the function
 * does nothing and returns zero. Otherwise, it appends all strings 
 * separated from <tt>str</tt> to the <tt>list</tt>.
 * The string sequence is preserved. All resulting strings in the list
 * are guaranteed to be null-terminated.
 *
 * Space for resulting strings is allocated automatically
 * in the designated memory pool. The allocated memory can be released only by
 * freeing the particular memory pool.
 *
 * @return
 * This function returns zero on success. On failure, <tt>-1</tt> is returned.
 */
int s_strdsep (tList &list, const char *str, const char *delim, tMemPool &pool);

/**
 * @brief Format a string. 
 *
 * @param ret
 * Pointer set to the result string
 *
 * @param format
 * String format description
 * 
 * @param ...
 * Parameters to be formated
 *
 * This function is a safer version of <tt>sprintf()</tt> which does all
 * memory allocations itself and thus doesn't force a user
 * to make (sometimes impossible) guesses of the buffer size needed
 * to hold formated data.
 *
 * The function <tt>die()</tt>s on failure.
 *
 * @return
 * This function returns the number of bytes written.
 *
 * @see <tt>printf()</tt> family
 */
int s_asprintf (char **ret, const char *format, ...);

/** @brief Format a string (result is put into a memory pool).
 *
 * @param ret
 * Pointer set to the result string
 *
 * @param pool
 * Memory pool for a new string allocation
 *
 * @param format
 * String format description
 * 
 * @param ...
 * Parameters to be formated
 *
 * This function is a safer version of <tt>sprintf()</tt> which does all
 * memory allocations itself and thus doesn't force a user
 * to make (sometimes impossible) guesses of the buffer size needed
 * to hold formated data. The allocated memory can be released only by freeing
 * the particular memory pool.
 *
 * The function <tt>die()</tt>s on failure.
 *
 * @return
 * This function returns the number of bytes written.
 *
 * @see <tt>printf()</tt> family
 */
int s_asprintf (char **ret, lib::tMemPool &pool, const char *format, ...);

/**
 * @brief Format a string.
 *
 * @param ret
 * Pointer set to the result string
 *
 * @param format
 * String format description
 * 
 * @param ap
 * Parameter list to be formated
 *
 * This function is a safer version of <tt>sprintf()</tt> which does all
 * memory allocations itself and thus doesn't force a user
 * to make (sometimes impossible) guesses of the buffer size needed
 * to hold formated data. The function does not call the <tt>va_end()</tt>
 * macro thus the value of <tt>ap</tt> is undefined after the call. 
 * The routine should call <tt>va_end(ap)</tt> itself afterwards.
 *
 * The function <tt>die()</tt>s on failure.
 *
 * @return
 * This function returns the number of bytes written.
 *
 * @see <tt>vprintf()</tt> family
 */
int s_vasprintf (char **ret, const char *format, va_list ap);

/** 
 * @brief Format a string (result is put into a memory pool).
 *
 * @param ret
 * Pointer set to the result string
 *
 * @param pool
 * Memory pool for a new string allocation
 *
 * @param format
 * String format description
 * 
 * @param ap
 * Parameter list to be formated
 *
 * This function is a safer version of <tt>sprintf()</tt> which does all
 * memory allocations itself and thus doesn't force a user
 * to make (sometimes impossible) guesses of the buffer size needed
 * to hold formated data. The allocated memory can be released only by freeing
 * the particular memory pool. The function does not call the <tt>va_end()</tt>
 * macro thus the value of <tt>ap</tt> is undefined after the call. The routine
 * should call <tt>va_end(ap)</tt> itself afterwards.
 *
 * The function <tt>die()</tt>s on failure.
 *
 * @return
 * This function returns the number of bytes written.
 *
 * @see <tt>vprintf()</tt> family
 */
int s_vasprintf(char **ret, lib::tMemPool &pool, const char *format,va_list ap);

/** 
 * @brief Read positive integer value from string.
 *
 * @param buf
 * Pointer to char pointer to string where num is stored
 *
 * @param dest
 * Pointer to integer where the result value will be stored on success
 *
 * @param llim
 * Lower limit of allowed range
 *
 * @param ulim
 * Upper limit of allowed range
 *
 * Happily stolen from NetBSD's <tt>strptime.c</tt>.
 *
 * @return
 * <tt>1</tt> on success, <tt>0</tt> when string not a number or out of allowed
 * range
 */
int s_conv_num(const char **buf, int *dest, int llim, int ulim);

/** 
 * @brief Read long int from string.
 *
 * @param buf
 * String where the number is stored
 *
 * @param llim
 * Lower limit of allowed range
 *
 * @param ulim
 * Upper limit of allowed range
 *
 * @param dest
 * Result long of the conversion.
 *
 * The function tries to convert the supplied string a to long integer.
 * The whitespace characters (as reported by <tt>isspace()</tt>) 
 * at the beginning and at the end are skipped.
 *
 * @return
 * The function returns zero on success, non-zero if the string is 
 * not a number or if it is out of range.
 */
int s_stol (const char *buf, long *dest, long llim, long ulim);

}

#endif
