/**
 * @file
 * @brief XML supplementary functions.
 *
 * This functions simplifies the manipulation with xml snippets and
 * provide easy mechanism for xml printout.
 *
 * @author Kamil Toman <ktoman@email.cz>
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#ifndef __ECHELOG_XML_H
#define __ECHELOG_XML_H

namespace lib {

/**
 * @brief Class for XML printouts to the file.
 *
 * This class provide easy mechanism for xml printout. A output file must 
 * be set to the instance, because all output is written to some file. 
 * There are member functions printing starting, ending tags and value
 * enclosed in tags. The indentation changes automatically, when starting
 * and end tags are printed.  
 * 
 */
class tXMLOutput {

    protected:

        /** 
         * @brief File descriptor of the file, where should be output 
         * written.
         */
        SIMPLE_PROPERTY (int, fd);
        /**
         * @brief Indentation step.
         */
        SIMPLE_PROPERTY (long, indent_step);
        /**
         * @brief Actual size of indentation.
         */
        PROPERTY (long, cur_indent);
        /** 
         * @brief Flag controlling behaviour on write error. When set 
         * program dies.
         */
        SIMPLE_PROPERTY (bool, die_on_err);
        
        /**
         * @brief Indentation string, filled with <tt>cur_indent</tt> spaces.
         */
        char *indstr;

    public:

        /**
         * @brief Constructor for XMLOutput. Initializes object parameters. 
         */
        inline tXMLOutput () : fd(-1), indent_step (4), die_on_err(false),
             indstr(NULL) { set_cur_indent(0); }

	virtual ~tXMLOutput() {} 
        
        /**
         * @brief This function prints xmlized string to current
         * cursor position in output xml file. 
         *
         * @param str
         * Pointer to string to be written.
         *
         * Certain characters have special significance in XML, and should be
         * represented by XML entities if they are to preserve their meanings.
         * This function writes a string with some of these conversions made.
         *
         * @note
         * If the input string is NULL, the function fails due to ASSERT. 
         *
         * @note
         * The function <tt>die()</tt>s on failure, i.e. when there's not
         * enough memory for the temporary result string.
         *
         * @return The function returns zero on success, otherwise <tt>-1</tt>.
         * If property <tt>die_on_err</tt> is set to true and error occurs,
         * function terminates whole program.
         */
        int print_data(const char *str);

        /**
         * @brief This function prints starting tag and changes current 
         * indentation.
         *
         * @param tagname
         * Pointer to tag name.
         *
         * This function prints <tt><tagname></tt> to current
         * cursor position in output xml file. This text is indented
         * according to current indentation and newline is printed at the
         * end. After printing current indentation is enlarged by indent_step
         * property value.
         *
         * @return The function returns zero on success, otherwise <tt>-1</tt>.
         * If property <tt>die_on_err</tt> is set to true and error occurs,
         * function terminates whole program.
         */
        int print_start_tag(const char *tagname);

        /**
         * @brief This function prints ending tag and and changes current
         * indentation.
         *
         * @param tagname 
         * Pointer to tag name.
         *
         * This function prints <tt></tagname></tt> to current
         * cursor position in output xml file. This text is indented
         * according to current indentation and newline is printed at the
         * end. Before printing current indentation is shrunk by indent_step
         * property value.
         *
         * @return The function returns zero on success, otherwise <tt>-1</tt>.
         * If property <tt>die_on_err</tt> is set to true and error occurs,
         * function terminates whole program.
         */
        int print_end_tag(const char *tagname);

        /**
         * @brief This function prints tag value enclosed in given tags
         *
         * @param tagname 
         * Pointer to tag name.
         *
         * @param tagvalue
         * Pointer to tag value text.
         *
         * This function prints <tt><tagname>tagvalue</tagname></tt> to current
         * cursor position in output xml file. This text is indented 
         * according to current indentation and newline is printed at the
         * end.
         *
         * @return The function returns zero on success, otherwise <tt>-1</tt>.
         * If property <tt>die_on_err</tt> is set to true and error occurs,
         * function terminates whole program.
         */
        int print_tag(const char *tagname, const char *tagvalue);

        /**
         * @brief This function xml header 
         *
         * @param stylesheet 
         * Pointer to stylesheet name.
         *
         * This function prints xml header to current cursor position in 
         * output xml file. If the stylesheet parameter is not null, line
         * describing stylesheet to use for interpreting xml document is
         * added.
         * 
         * @return The function returns zero on success, otherwise <tt>-1</tt>.
         * If property <tt>die_on_err</tt> is set to true and error occurs,
         * function terminates whole program.
         */
        int print_header(const char *stylesheet = NULL);

}; 
         


/**
 * @brief Convert all applicable characters to XML entities 
 *
 * @param str
 * Pointer to string to be converted
 *
 * @param replace_spaces
 * If the parameter is set to true the function also replaces all spaces
 * with a non-breakable space html entity. Default is false, no such conversion
 * is performed.
 *
 * Certain characters have special significance in XML, and should be
 * represented by XML entities if they are to preserve their meanings. This
 * function returns a string with some of these conversions made. Space for
 * the resulting string is allocated automatically, the string is
 * null-terminated. The allocated memory can be released with <tt>xfree()</tt>.
 *
 * @note 
 * If the input string is NULL, the function returns NULL and does nothing
 * else.
 *
 * @note 
 * The function <tt>die()</tt>s on failure, i.e. when there's not enough memory
 * for the resulting string.
 *
 * @return 
 * The function returns a pointer to string with all special codes
 * substituted by relevant XML entities.
 *
 */
char * xmlentities (const char *str, bool replace_spaces = false);

/**
 * @brief Convert all applicable characters to XML entities 
 *
 * @param str
 * Pointer to string to be converted
 *
 * @param replace_spaces
 * If the parameter is set to true the function also replaces all spaces
 * with a non-breakable space html entity. Default is false, no such conversion
 * is performed.
 * 
 * @param pool
 * Memory pool for a new string allocation
 *
 * Certain characters have special significance in XML, and should be
 * represented by XML entities if they are to preserve their meanings. This
 * function returns a string with some of these conversions made. Space for the
 * resulting string is allocated automatically in the designated memory pool,
 * the string is null-terminated. The allocated memory can be released only by
 * freeing the particular memory pool.
 *
 * @note 
 * If the input string is NULL, the function returns NULL and does nothing
 * else.
 *
 * @note 
 * The function <tt>die()</tt>s on failure, i.e. when there's not enough memory
 * for the resulting string.
 *
 * @return 
 * The function returns a pointer to string with all special codes
 * substituted by relevant XML entities.
 *
 */
char * xmlentities (const char *str, tMemPool &pool, 
        bool replace_spaces = false);
    
}

#endif
