/**
 * @file
 * @brief Echelog miniagent support library
 *
 * This abstract class provides an interface used by
 * the tSimpleLex class.
 *
 * Any source information buffering class should 
 * include this as a minimal subset of its functionality.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_INFOBUFFER_H
#define ECHELOG_INFOBUFFER_H

#include "lib/archdep.h"

namespace magents {

/** An abstract information buffering class */
class tInfoBuffer {
    public:
        /** @brief Read the next character from the buffer.
         *
         * @return
         * The function returns an unsigned char cast to an integer.
         * If no character is available -1 is returned. 
         * On error, -2 is returned.
         */
        virtual inline int get_byte () = 0;

        /** @brief Push the last read byte back to buffer
         *
         * This is an inverse operation to get_byte(). The backed 
         * characters are naturally reverse-ordered so they appear 
         * as if they were not read from the buffer. They may 
         * be regained by ordinary read or peek operations. 
         * If there is no character to back an error is reported.
         *
         * Only one pushback is guaranteed but in most buffer implementaions
         * up to `cache size' characters may be backed.
         *
         * @return
         * The function return zero on success. On error, -1 is returned.
         */
        virtual inline int unget_byte () = 0;

        /** @brief Peek at the first byte available in the buffer
         *
         * This function may be used for a look ahead. It is
         * functionally equivalent to a pair of get_byte(), unget_byte()
         * calls (but more effective).
         *
         * @return
         * The function returns an unsigned char cast to an integer.
         * If no character is available -1 is returned. 
         * On error, -2 is returned.
         */
        virtual inline int peek_byte () = 0;

        /** 
         * @brief Destroy the buffer 
         */
        virtual ~tInfoBuffer() {};
};

}

#endif
