/**
 * @file
 * @brief Contains declaration of tMagent class
 *
 * @author Kamil Toman <ktoman@email.cz>, Marek Zoth <marek.zoth@volny.cz>
 *
 */

#ifndef ECHELOG_MAGENT_H
#define ECHELOG_MAGENT_H

#include "mframes/echelogmodule.h"
#include "magents/msgframe.h"

/**
 * @namespace magents
 * @brief Echelog Miniagent support library
 * TODOC:
 */
namespace magents {

/** 
 * This class implements a typical miniagent's functionality.
 * The agent periodically reads informational chunks
 * from a supplied parser and forms them into messages which 
 * are passed to the next module. It also covers the basic miniagent's
 * configuration and reconfiguration.
 *
 * This class should not have multiple instances in the program.
 */
class tMagent : public mframes::tEchelogModule
{
    public:
        //-------------------------------
        // Interface
        //-------------------------------

        /** Create a miniagent and bind the supplied parsing code
         *
         * @param mp
         * A mempool which can be used for internal magent's structures
         *
         * @param p
         * A parser which should be used to gain magent's information
         *
         * @param id
         * Miniagent's message identificator
         */
        tMagent (tCmdLine& cmdline, const char* cfg_filename);

        /** Install miniagent's signal handling function */
        bool init (const char *id);

        /**
         * TODOC:
         */
        virtual void cleanup ();

    protected:
        //-------------------------------
        // Internal methods
        //-------------------------------

        /// Initialization reconfiguration dependent part
        virtual bool dynamic_init ();

        /// Reverse method for dynamic_init()
        virtual void dynamic_cleanup ();

        /// Installs signal handlers
        virtual void install_sig_handlers ();

        /// Handles installed signals
        virtual SigHandlingRet handle_signals ();

        /// Handles signal SIGPIPE
        virtual SigHandlingRet on_sigpipe ();

        /// Disables standard SIGHUP handler
        virtual SigHandlingRet on_sighup ();

        /// Returns true if SIGALRM has been catched
        inline bool sigalrm_received ()
        { return s_sigalrm_received; }

        /// Returns true if SIGPIPE has been catched
        inline bool sigpipe_received ()
        { return s_sigpipe_received; }

        /**
         * @brief TODOC:
         */
        virtual int process ();

        /** @brief Form the information to messages and send them
         *
         * This function is called once in every run_interval seconds
         * from the main miniagent's loop. It calls externally supplied
         * information parser translates its input to message chunks.
         * These message chunks are formed into messages and then they're
         * sent to the next module.
         *
         * @return  TODOC:
         * True if it finished its work on time; false otherwise.
         */
        virtual bool process_messages (bool& on_time) = 0;


        //------------------------------------------------
        // Implementation of tConfigurable interface
        //------------------------------------------------

        /**
         * @brief TODOC:
         */
        virtual bool register_sections (lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void unregister_sections ();

        /**
         * @brief TODOC:
         */
        virtual bool process_config (lib::tMemPool& conf_mp,
                lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void cleanup_config ();

        //-------------------------------
        // Owned components
        //-------------------------------
        
        /// Miniagent's message frame
        tMsgFrame* mp_mf;

        /// Flag to set when SIGALRM has been catched
        volatile static bool s_sigalrm_received;

        /// Flag to set when SIGPIPE has been catched
        volatile static bool s_sigpipe_received;

        //-------------------------------
        // Object state attributes
        //-------------------------------

        /// Miniagent's internal state
        bool m_inited;

    private:
        //-------------------------------
        // Internal methods
        //-------------------------------

        /// A trivial signal transfering function
        static void sig_handling_f (int signum);

        //-------------------------------
        // Owned components
        //-------------------------------

        /** 
         * @brief Interval -- how often information
         * should be retrieved
         */
        time_t m_run_interval;         

        static struct lib::cfitem s_magentcommon_section[];
        static long s_run_min;
        static long s_run_hour;
        static long s_run_sec;
        static char* s_inst_name;
};

}

#endif
