/**
* @file
 * @brief Echelog miniagent support library
 *
 * This class covers the message handling and a communication
 * with next echelog modules. It's primarily used by tMagent
 * for a low-level operations.
 *
 * @author Kamil Toman <ktoman@email.cz>, Marek Zoth <marek.zoth@volny.cz>
 *
 */

#ifndef  ECHELOG_MSGFRAME_H
#define  ECHELOG_MSGFRAME_H

#include "messages/msgchunk.h"
#include "mframes/unixsocketioframe.h"

#define SIMPLE_LOG_MSG  "SimpleLog"


namespace magents {

/// Message handling frame
class tMsgFrame : public mframes::tUNIXSocketOFrame
{
    public:
        /** Set up a message frame */ 
        tMsgFrame (const char *id);

        bool init_alert ();

        void cleanup_alert ();

        /** Reset the message frame for future reuse */
        virtual inline void reset ();

        /** Get current total message length (in bytes) */
        virtual inline long get_store_length () const
        { return length; }

        /** Add a chunk to current message */
        virtual inline bool add_chunk (messages::tMsgChunk &c);

        /** @brief Finish the message msg_id and send it to output 
         *
         * @param msg_id
         * Message identificator number
         *
         * The function finishes the message of msg_id type writes it
         * to output buffer and then sends it to the next echelog module.
         * For empty messages nothing happens. The message frame is reset()
         * after a successful send.
         *
         * @note The function die()s on failure.
         */
        virtual bool send_message (long msg_id);

        /** @brief Send an alert to the next module
         *
         * @param alert_string
         * The alert message to be sent
         *
         * This function forms and immmediately sends a security alert
         * message.
         *
         * @note The function die()s on failure.
         */
        virtual bool send_alert (char *alert_string);

        inline bool is_broken ()
        { return m_broken; }

    protected:
        /// pool used by alert chunk
        lib::tMemPool m_pool;
        /// messages' raw data storage
        char message_buffer[MAX_MESSAGE_SIZE];
        /// frame message
        messages::tMsgBuf message;
        /// alert chunk
        messages::tMsgChunk* alert_chunk;
        /// current message length (in bytes)
        long length;
        /// length of an empty message (in bytes)
        long empty_message_length;

        SIMPLE_R_PROPERTY (long, num_of_records); //< number of msg records
        SIMPLE_PROPERTY (bool, incremental_message); //< inc flag
        PROPERTY (char *, inst_name);   //< Agent instance name

    private:
        //false if everything in constructor was OK
        bool m_broken;
};

}

/* Here we actually define all tMsgFrame inline functions */
#include "magents/msgframe.hxx"

#endif
