/**
 * @file
 * @brief Echelog Miniagent support library
 *
 * This abstract class provides an interface used by
 * the tMagent class. A typical miniagent should
 * implement these functions as a minimal subset of
 * its functionality.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_PARSER_H
#define ECHELOG_PARSER_H

#include "messages/msgchunk.h"
#include "mframes/configurable.h"

namespace magents {


/** Parser states handled by tMagent */
typedef enum {
    PR_OK = 0,
    PR_EMPTY,
    PR_ERROR,
    PR_END
} tPARSER_STATE;


/** Parser response handled by tMagent */
struct tParserResponse {
    tPARSER_STATE state;           //< Must be one of tPARSER_STATE values
    messages::tMsgChunk *chunk;    /**< This value is used only if
                                    * the parser state is equal to PR_OK */
};

/** An abstract information parsing class */
class tParser : mframes::tConfigurable
{
    public:
        /// Constructor
        tParser () {};

        /*
         * @brief TODOC:
         */
        virtual bool init () = 0;

        /*
         * @brief TODOC:
         */
        virtual void cleanup () = 0;


        /* @brief Init parser state
         *
         * Unlike the parser constructor this function is called 
         * on the beginning of every run interval from tMagent main loop.
         * In this call most of context dependent initalizations should
         * take place.
         *
         * On failure, an alert is reported by miniagent to the next module.
         * 
         * @return
         * On success the function returns zero. On error, a non-zero 
         * number is returned.
         */
        virtual int start () = 0;

        /* @brief Stop and uninitialize parser
         *
         * Unlike the parser destructor this function is called on the end
         * of every run interval from tMagent main loop. In this call most
         * of needed parser cleanups should take place.
         *
         */
        virtual void stop () = 0;

        /* @brief Clean up parser structures 
         *
         * This function is called after a message has been sucessfully
         * passed to next module. The parser may safely release all
         * structures to returned chunks.
         */
        virtual void parse_cleanup () = 0;

        /* @brief Get another informational chunk 
         *
         * @param r
         * The parser response structure to be filled in
         *
         * This function should provide all information collected
         * be the particular miniagent. It is called in a sequence 
         * from miniagent's main loop until the response's state 
         * is not PR_EMPTY.
         *
         * The message chunks need not to be of the same type. If the
         * message chunk type differs from the previous ones the
         * message is automatically sent and a the chunk is then
         * stored into a brand new message.
         */
        virtual void get_chunk (tParserResponse &r) = 0;

        /* @brief Get a chunk containing error information
         *
         * @param r
         * The parser response structure to be filled in
         *
         * If the get_chunk() function returned a PR_ERROR
         * an alert is sent and then the miniagent calls this 
         * function to get more information about what happened.
         *
         * The call is looped until the response's state
         * is not PR_EMPTY. Any non-empty response must be
         * of the type PR_OK and must contain a valid message 
         * chunk. Other types of responses are not allowed.
         *
         * When the information is complete the miniagent passes
         * the information to the next module and then gets
         * back to main loop.
         */ 
        virtual void get_error_info (tParserResponse &r) = 0;

        /* @brief Get miscellaneous information
         *
         * @param r
         * The parser response structure to be filled in
         *
         * On the end of every run interval, after the parser 
         * is stopped, the miniagent retrieves any type
         * of miscellananeous information which may have
         * piled up during the time.
         *
         * The call is looped until the response's state
         * is not PR_EMPTY. Any non-empty response must be
         * of the type PR_OK and must contain a valid message 
         * chunk. Other types of responses are not allowed.
         */
        virtual void get_misc_info (tParserResponse &r) = 0;

        virtual bool register_sections (lib::tMemPool& trash_pool)
            { return false; }
        virtual void unregister_sections () 
            { }
        virtual bool process_config (lib::tMemPool& conf_mp,
                lib::tMemPool& trash_pool)
            { return false; }
        virtual void cleanup_config ()
            { }

};

}

#endif
