/**
 * @file
 * @brief Contains declaration of tParsingMagent class
 *
 * @author Kamil Toman <ktoman@email.cz>
 */

#ifndef ECHELOG_PARSINGMAGENT_H
#define ECHELOG_PARSINGMAGENT_H

#include "magents/parser.h"
#include "magents/magent.h"

/**
 * @namespace magents
 * @brief Echelog Miniagent support library
 * TODOC:
 */
namespace magents {

/** 
 * This class implements a typical miniagent's functionality.
 * The agent periodically reads informational chunks
 * from a supplied parser and forms them into messages which 
 * are passed to the next module. It also covers the basic miniagent's
 * configuration and reconfiguration.
 *
 * This class should not have multiple instances in the program.
 */
class tParsingMagent : public tMagent
{
    public:
        //-------------------------------
        // Interface
        //-------------------------------

        /** Create a miniagent and bind the supplied parsing code
         *
         * @param mp
         * A mempool which can be used for internal magent's structures
         *
         * @param p
         * A parser which should be used to gain magent's information
         *
         * @param type
         * Miniagent's message identificator
         */
        tParsingMagent (tCmdLine& cmdline, const char* cfg_filename,
            tParser &p) : tMagent (cmdline, cfg_filename), m_parser (p) {};

    protected:

        //-------------------------------
        // Internal methods
        //-------------------------------

        /// Initialization reconfiguration independent part
        virtual bool static_init ();

        /// Reverse method for static_init()
        virtual void static_cleanup ();
  
        /** @brief Handle a new chunk
         *
         * A new chunk is appended to current message. If the message
         * is too long or the previous chunks differ from the new
         * one the message is sent and the chunk is put to a brand new one.
         */
        virtual bool handle_chunk (messages::tMsgChunk &chunk, long &prev_type,
                bool info = true);

        /** @brief Form the information to messages and send them
         *
         * This function is called once in every run_interval seconds
         * from the main miniagent's loop. It calls externally supplied
         * information parser translates its input to message chunks.
         * These message chunks are formed into messages and then they're
         * sent to the next module.
         *
         * @return  TODOC:
         * True if it finished its work on time; false otherwise.
         */
        virtual bool process_messages (bool& on_time);

        //------------------------------------------------
        // Implementation of tConfigurable interface
        //------------------------------------------------

        /**
         * @brief TODOC:
         */
        virtual bool register_sections (lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void unregister_sections ();

        /**
         * @brief TODOC:
         */
        virtual bool process_config (lib::tMemPool& conf_mp, 
                lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void cleanup_config ();

    private:

        //-------------------------------
        // Used-only components
        //-------------------------------

        /// Supplied external parser
        tParser& m_parser;
};

}

#endif
