/**
 * @file
 * @brief Echelog miniagent support library
 *
 * This class implements an external command source 
 * buffering interface. It can be used to exec and gain 
 * the standard output of a system utility.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_RUNBUFFER_H
#define ECHELOG_RUNBUFFER_H

#include "property.h"
#include "lib/lib.h"
#include "magents/infobuffer.h"

#include <sys/stat.h>

/* Buffer size (in bytes) */
#define CACHE_SIZE 16384

#define SIMPLE_LOG_MSG     "SimpleLog"

namespace magents {

/**
 * This class implements a superset of tInfoBuffer functionality. 
 * It can be used to get individual bytes from the standard output 
 * of a particular system program. The utility's standard error output 
 * can be handled via a safe temporary file.
 */
class tRunBuffer : public tInfoBuffer {
    protected:
        SIMPLE_R_PROPERTY (int, temp_fd); //< Filedescriptor of the temp file
        char *temp_filename;              //< Temporary file's name
        SIMPLE_R_PROPERTY (pid_t, pid);   //< Spawned program's PID
        pid_t pgid;                       //< Current process group

        char cache_start[CACHE_SIZE];     //< File cache
        char *cache_pos;                  //< Current in-cache position
        char *cache_end;                  //< Actual end of file cache
        bool data_available;              //< True if data are already loaded
        unsigned long data_to_unget;      //< True if there something to unget

        int bfd;                          //< Input filedescriptor

        int fetch ();                     //< Load further data to cache

    public:
        /** Create a buffer and associate a new temporary file */
        tRunBuffer ();
        /** Destroy the buffer and remove the temporary file */
        virtual ~tRunBuffer ();         

        /** @brief Init buffer state and exec a system utility
         *
         * @param command
         * The system utility to be run
         *
         * @param args
         * Command line arguments of the system utility
         *
         * @param env
         * Program environment of the system utility
         *
         * This function execs a system utility `command' with
         * supplied arguments and environment. Then it establishes
         * a pipe to its standard output and redirects its standard
         * error output to class' temporary file.
         * 
         * @return
         * The program returns zero on success. It the spawned program
         * exits immediately its exit return value is returned.
         * On error, -1 is returned.
         */
        virtual int init (const char *command, char * const args[], 
                char * const env[]);

        /** @brief Clean up the buffer state from previous runs
         *
         * This function closes any pending open pipes to a spawned
         * program and kills it.
         */
        virtual void cleanup ();

        /** @brief This is a tInfoBuffer::get_byte() implementation */
        virtual inline int get_byte ();
        /** @brief This is a tInfoBuffer::unget_byte() implementation.
         *
         * At most CACHE_SIZE-1 pushbacks are guaranteed. 
         */
        virtual inline int unget_byte ();
        /** @brief This is a tInfoBuffer::peek_byte() implementation */
        virtual inline int peek_byte ();
};

}

/* Here we actually define all tRunBuffer inline functions */
#include "magents/runbuffer.hxx"

#endif
