/**
 * @file
 * @brief Echelog Miniagent support library
 *
 * This class can be used for a simple lexical analysis
 * above an instance of tInfoBuffer.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_SIMPLELEX_H
#define ECHELOG_SIMPLELEX_H

#include "lib/lib.h"
#include "magents/infobuffer.h"

#define DEFAULT_MAX_TOKEN_SIZE 4096

namespace magents {

/// All possible types of simplelex tokens (all terminal types):
typedef enum {
    T_STRING,   //< string type
    T_NUMBER,   //< number
    T_NEWLINE   //< newline character
} tSIMPLELEX_TOKEN_TYPE;


/// Tokens returned by simplelex analyzis
struct tToken {
    tSIMPLELEX_TOKEN_TYPE type; //< type of the token

    unsigned long num;          //< numeric value (valid only for T_NUMBER)
    bool sign;                  //< sign of the value (true for negative nums)
    
    char *str;                  //< string portion representing the value
    long len;                   /**< length of the string (excluding
                                 * the termination null) */
                        
};


/** This class implements a very simple (one and two dimensional) lexical 
 * analyzis. Although it's very simple it may suffice in many cases. 
 * A few extra functions are included to add a bit of flexibility. 
 *
 * It uses an tInfoBuffer interface to gain individual characters. 
 */
class tSimpleLex {
    protected:
        tInfoBuffer &buf;               //< infobuffer instance to be used
        lib::tMemPool &pool;            //< mempool to alloc lex token in

        char *token_buf;                //< (fixed length) token buffer
        long old_len;                   //< previous token length
        long len;                       //< current token length
        const long max_token_size;      //< maximal token size

        SIMPLE_PROPERTY (long, column); //< current column (first column is #0)
        SIMPLE_PROPERTY (bool, append_mode); /**< True if the next token is 
                                               to be appended to previous one */

    protected:
        /** @brief Get a null-terminated string 
         *
         * Reads non-blank characters from the underlying infobuffer
         * and places them into the token_buf. The null-termination
         * is guaranteed. If the string is longer than maximal
         * token size it's shortened accordingly.
         *
         * @return
         * The function return zero if the string portion was has been
         * successfully placed to the token buffer. If the string
         * is longer than maximal token size a non-zero is returned.
         */
        inline int get_string ();

        /* @brief Store actual token 
         *
         * @param r
         * The token structure to be filled
         *
         * Store current token buffer state as a real token. If the
         * trimmed string (without blank characters or newline characters
         * at the beginning and at the end) corresponds to a numeric value 
         * in decimal, hexadecimal or octal base its value is also stored to r.
         */
        inline void tSimpleLex::store_token (tToken &r);
        
    public:

        /** @brief Create a tSimpleLex object and bind a buffer to it
         *
         * @param b
         * A buffer to read input from
         *
         * @param mp
         * A mempool which would be used for tokens' space allocation
         *
         * @param max_token
         * Specifies the maximal length of each token
         */
        tSimpleLex (tInfoBuffer &b, lib::tMemPool &mp, 
                const long max_token = DEFAULT_MAX_TOKEN_SIZE);

        /** Destroy the tSimpleLex object and release 
         * the space used for the token buffer (the mp is left untouched) */
        virtual ~tSimpleLex ();

        /** @brief Skip all blank characters 
         *
         * The function reads and throws away all blank characters
         * from the underlying buffer.
         */
        virtual void skip_blanks ();

        /** @brief Skip all characters till the first occurence of stop_char
         *
         * @param stop_char
         * An unsigned char value cast to int 
         *
         * The function skips all characters up to specified stop_char
         * (including the stop_char).
         *
         * @return
         * The function returns zero on success. If the stop_char was
         * not found in the buffer -1 is returned.
         */
        virtual int skip_to (const int stop_char);

        /** @brief Skip all characters till one of stop_chars is found
         *
         * @param stop_char
         * An unsigned char value cast to int 
         *
         * The function skips all characters until one of specified
         * stop_chars is found (including the actual stop character).
         *
         * @return
         * The function returns zero on success. If the stop_char was
         * not found in the buffer -1 is returned.
         */
        virtual int skip_to (const char *stop_chars);

        /** @brief Skip all characters till the first occurence of stop_char
         *
         * @param stop_char
         * An unsigned char value cast to int 
         *
         * The function skips all characters up to specified stop_char
         * (NOT including the stop_char).
         *
         * @return
         * The function returns zero on success. If the stop_char was
         * not found in the buffer -1 is returned.
         */
        virtual int skip_upto (const int stop_char);

        /** @brief Skip all characters till one of stop_chars is found
         *
         * @param stop_char
         * An unsigned char value cast to int 
         *
         * The function skips all characters until one of specified
         * stop_chars is found (NOT including the actual stop character).
         *
         * @return
         * The function returns zero on success. If the stop_char was
         * not found in the buffer -1 is returned.
         */
        virtual int skip_upto (const char *stop_chars);

        /** @brief Skip all characters up to the specified column
         *
         * @param col
         * Column number where to stop
         *
         * The function skips all characters up to the specified 
         * column (i.e. the cursor is positioned at it).
         *
         * @return
         * The function returns zero on success. If the line was
         * too short -1 is returned.
         */
        virtual int skip_upto_col (const long col);

        /** @brief Get all characters up to stop_char
         *
         * @param r
         * The token structure to be filled
         *
         * @param stop_char
         * An unsigned char value cast to int 
         *
         * Read and place all characters from the underlying buffer
         * up to (including) the stop_char to string r token.
         * The string is guaranteed to be null-terminated. If the string
         * is longer than max maximal token size it's shortened accordingly.
         *
         * @return
         * The function returns zero on success. If the string
         * has been shortened 1 is returned. On error, -1 is returned.
         */
        virtual int get_to (tToken &r, const int stop_char);

       /** @brief Get all characters up to any of stop_chars is found
         *
         * @param r
         * The token structure to be filled
         *
         * @param stop_char
         * An unsigned char value cast to int 
         *
         * Read and place all characters from the underlying buffer
         * up to (including) the actual stop character to string r token.
         * The string is guaranteed to be null-terminated. If the string
         * is longer than max maximal token size it's shortened accordingly.
         *
         * @return
         * The function returns zero on success. If the string
         * has been shortened 1 is returned. On error, -1 is returned.
         */
        virtual int get_to (tToken &r, const char *stop_chars);

        /** @brief Get all characters up to stop_char
         *
         * @param r
         * The token structure to be filled
         *
         * @param stop_char
         * An unsigned char value cast to int 
         *
         * Read and place all characters from the underlying buffer
         * up to (NOT including) the stop_char to string r token.
         * The string is guaranteed to be null-terminated. If the string
         * is longer than max maximal token size it's shortened accordingly.
         *
         * @return
         * The function returns zero on success. If the string
         * has been shortened 1 is returned. On error, -1 is returned.
         */
        virtual int get_upto (tToken &r, const int stop_char);

       /** @brief Get all characters up to any of stop_chars is found
         *
         * @param r
         * The token structure to be filled
         *
         * @param stop_char
         * An unsigned char value cast to int 
         *
         * Read and place all characters from the underlying buffer
         * up to (NOT including) the actual stop character to string r token.
         * The string is guaranteed to be null-terminated. If the string
         * is longer than max maximal token size it's shortened accordingly.
         *
         * @return
         * The function returns zero on success. If the string
         * has been shortened 1 is returned. On error, -1 is returned.
         */
        virtual int get_upto (tToken &r, const char *stop_chars);

        /** @brief Get all characters up to the specified column
         *
         * @param r
         * The token structure to be filled
         *
         * @param col
         *
         * The function reads and places all characters up to
         * (NOT including) the specified column from the underlying buffer 
         * to string r token. If the string is longer than the maximal
         * token size it is shortened accordingly.
         *
         * @return
         * The function returns zero on success. If the string
         * has been shortened 1 is returned. If the line didn't have
         * enough columns or some other error occured -1 is returned.
         */
        virtual int get_upto_col (tToken &r, const long col);

        /** @brief Get the next token 
         *
         * @param r
         * The token structure to be filled
         *
         * This function gets the next token as defined above (for example,
         * it ignores all blank characters). If the gained string is a signed
         * (preceded by '-' sign) or an unsigned (optionally preceded by '+'
         * sign) 32-bit integer it's value is also placed to num field
         * in the token structure (the sign field is set to true for negative
         * integer values). Even other than decimal based integers are 
         * recognized. If the number is preceded by '0x' the number is
         * considered to be written in the hexadecimal base. If the first 
         * character of the number is zero the number is considered
         * to be octal-based. Anyway, the string value is always returned.
         *
         * Token value must be always shorter than the maximal token size is.
         * Otherwise the string is shortened accordingly. The null-termination
         * is guaranteed. The string length is also written to len field
         * of the token structure (excluding the termination null).
         *
         * @return
         * The function returns zero on success. If the token size has been
         * exceeded the return value is 1. If there is no token available
         * -1 is returned.
         */
        virtual int get_token (tToken &r);
};


}

#endif
