/**
 * @file
 * @brief Message buffer handling
 *
 * This file defines an interface for a manipulation with standard echelog 
 * messages. 
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef _ECHELOG_MSGBUF_H
#define _ECHELOG_MSGBUF_H

#include "messages/msgheader.h"

namespace messages {

/**
 * @brief Message buffer container
 * 
 * This class can be used to operate with the complete echelog message. 
 * It is a container class for a message header, message delimiters 
 * and a message body.
 */
class tMsgBuf
{

    protected:
        
        bool r_prepared;    ///< True if the message is ready for reading.
        bool w_prepared;    ///< True if the message is ready for writing.
        
        /// Declared total message length (including all delimiters).
        int32_t declared_length;    

        /**
         * @brief <em>External resource:</em> Message header information.
         */
        SIMPLE_R_EXT_RESOURCE (tMsgHeader *, header);
        /** 
         * @brief <em>External resource:</em> Message body (NULL for empty 
         * or uninitialized messages).
         */
        SIMPLE_R_EXT_RESOURCE (tPlainBuf *, body);

        /// Delimiter record holding length of message body.
        tPlainBuf *body_length_field;
        /// <em>Read only property:</em> Maximal length of message body.
        R_PROPERTY (int32_t, max_body_len);
        
        /// Get body length field size in bytes.
        inline int32_t get_body_field_size() const;
        
        /// Return size of message (excluding total message length field).
        inline int32_t get_len () const;

        
    public:
        
        /// Constructor---creates empty message.
        tMsgBuf () : r_prepared (false), w_prepared (false),
        declared_length (-1), header (0), body (0), body_length_field (0), 
        max_body_len (0) { }
        /// Destructor---resets message and releases allocated space.
        virtual ~tMsgBuf ();       

        /** 
         * @brief Returns message's underlying raw buffer.
         *
         * The raw buffer, given at actual message initialization,
         * (see <tt>prepare_to_read()</tt>, <tt>prepare_to_write()</tt>)
         * is returned. If the message is empty or yet uninitializes the
         * function returns NULL. 
         *
         * The underlying buffer should never be touched. This function
         * exposes it only for sake of i/o system functions like 
         * <tt>read()</tt>, <tt>write()</tt>, <tt>recv()</tt> and similar.
         */
        inline const char * get_buffer() const;

        /** 
         * @brief Return true if buffer is ready for reading.
         *
         * After a new tMsgBuf is created and the basic message
         * structure is successfully read from the underlying raw buffer
         * (see <tt>prepare_to_read()</tt>), it's possible to simply
         * access message's header information or to load chunks.
         *
         * @return 
         * The function returns true if the message is correctly loaded and
         * prepared for reading, false otherwise.
         */
        inline bool is_r_prepared () const;
        
        /**
         * @brief Return true if buffer is ready for writing.
         *
         * Before we can store information to a new message we have
         * to set up its basic structure and associate a raw buffer
         * big enough to hold the complete message including headers
         * (see <tt>prepare_to_write()</tt>). 
         *
         * @return
         * The function returns true if the message is ready
         * for writing, false otherwise.
         */
        inline bool is_w_prepared () const;
        
        /** 
         * @brief Return declared size of the complete message 
         * (including all message delimiters).
         *
         * The declared message length is the length specified in the
         * message header. It is automatically loaded upon
         * <tt>prepare_to_read()</tt>. If storing, the declared length
         * denotes the message length after a successful <tt>close()</tt>.
         *
         * @return
         * The declared message length (in bytes) is returned. On error,
         * <tt>-1</tt> is returned;
         */
        inline const int32_t get_declared_length () const;
        
        /**
         * @brief Return the size of the message in bytes.
         *
         * It can be used after all chunks has been stored to find out how
         * large raw buffer we need to manipulate with.
         *
         * @return
         * This function returns number of bytes in the underlying raw 
         * buffer actually used by the message.
         */
        inline const int32_t get_store_length () const; 
        
        /**
         * @brief Prepare a blank message for writing.
         *
         * Before we can write on a newly created or <tt>reset()</tt>ed
         * message we need to call this function to setup the basic message
         * structure and to associate it with a raw buffer. The
         * specified raw buffer <tt>new_buffer</tt> must be at least long
         * enough to hold an empty message. The parameter <tt>new_length</tt>
         * states the maximal length of underlying raw buffer---writing
         * operations which would exceed this limit will yield an error.
         *
         * @note
         * This operation should be called only once after initialization or
         * after <tt>reset()</tt>. A single tMsgBuf can't be read and
         * written simultaneously. We need to <tt>close()</tt> the message at
         * first, <tt>reset()</tt> it (or to create a new tMsgBuf) and then to
         * prepare the message for reading.
         *
         * @return
         * On successful completion, the function returns zero. On error,
         * non-zero is returned.
         */
        int prepare_to_write (const char *new_buffer, int32_t new_length);

        /**
         * @brief Prepare for reading a message.
         *
         * If we want to read a message, at first, we need to call this
         * function to load the basic message structure and the message
         * header from the raw buffer <tt>new_buffer</tt>. The parameter
         * <tt>new_length</tt> gives the loaded raw buffer length. Any
         * attempt to read the message information behind this limit
         * will yield an error.
         *
         * @note
         * This operation should be called only once after tMsgBuf is created
         * or after <tt>reset()</tt>. A single tMsgBuf can't be read and
         * written simultaneously. We need to <tt>close()</tt> the message at
         * first, <tt>reset()</tt> it (or to create a new tMsgBuf) and then to
         * prepare the message for writing.
         *
         * @return
         * On successful completion, the function returns zero. On error,
         * non-zero is returned.
         */
        int prepare_to_read (const char *new_buffer, int32_t new_length);

        /** 
         * @brief Put the cursor to the beginning of message body.
         *
         * The current reading/writing position in message body is
         * reset to the beginning.
         *
         * @return
         * On successful completion, the function returns zero. On error,
         * non-zero is returned.
         */
        inline int rewind ();

        /** 
         * @brief Reinitialize the message buffer internal state.
         *
         * This functions discards the message contents completely so it can
         * be reused. If we want to do further message processing, we have to
         * initialize again (see <tt>prepare_to_read()</tt>,
         * <tt>prepare_to_write()</tt>).
         *
         * @return
         * On successful completion, the function returns zero. On error,
         * non-zero is returned.
         */
        int reset ();

        /**
         * @brief Flush the buffer and write headers (commit).
         *
         * This function flushes all message information and finalizes
         * the underlying raw buffer if necessary. It should be called
         * every time after all operations on the message are finished.
         *
         * @return
         * On successful completion, the function returns zero. On error,
         * non-zero is returned.
         */
        int close ();
};

}

/* Inline functions we actually define in here */
#include "messages/msgbuf.hxx"

#endif


