/**
 * @file
 * @brief Message chunks
 *
 * This file contains an interface for accessing standalone message chunks
 * and their attributes.
 *
 * @author Ales Privetivy <ales@ucw.cz>
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef _ECHELOG_MESSAGECHUNK_H
#define _ECHELOG_MESSAGECHUNK_H

#include "messages/types/echelogtype.h"

namespace messages {

/**
 * @brief Message chunk container
 *
 * tMsgChunk is a container class which can be used to retreive or store
 * message data contents accordingly to echelog message configuration.
 */
class tMsgChunk {

    private:
        /// Array of chunk attributes.
        class tEchelogType **attribute;
        /// Memory pool for msg-chunk allocation purposes.
        class lib::tMemPool &mempool;
        /// Internal handle identifying message.
        int32_t message_handle;
        /// Message type.
        SIMPLE_R_PROPERTY (int32_t, type);
        /// Number of chunk attributes.
        SIMPLE_R_PROPERTY (int32_t, attr_count);
        /// Status of message chunk.
        SIMPLE_PROPERTY (tMSG_STATUS, status);

    public:
        /**
         * @brief Constructor for tMsgChunk.
         *
         * It creates a new message chunk of given <tt>message_name</tt> using
         * <tt>mpool</tt> for all allocations. It also creates instances of all
         * chunk attributes.
         */
        tMsgChunk(char *message_name, class lib::tMemPool &mpool);

        /**
         * @brief Constructor for tMsgChunk.
         *
         * It creates a new message chunk of given <tt>message_type</tt> using
         * <tt>mpool</tt> for all allocations. It also creates instances of all
         * chunk attributes.
         */
        tMsgChunk(int32_t message_type, class lib::tMemPool &mpool);

        /**
         * @brief Destructor for tMsgChunk.
         *
         * It removes chunks structures and destructs all its attributes.
         */
        virtual ~tMsgChunk();

        /**
         * @brief Get attribute instance of given name within this chunk.
         *
         * This function can be used to retrieve with name <tt>attr_name</tt>
         * from within a chunk. The attribute stays associated with the chunk.
         *
         * @note
         * The function die's if <tt>attr_name</tt> is not a correct name of
         * one of the chunk's attributes.
         *
         * @return
         * The function returns a pointer of chunk's attribute with name
         * <tt>attr_name</tt>.
         */
        class tEchelogType * get_attr(const char *attr_name) const;

        /**
         * @brief Get attribute instance of given handle within this chunk.
         *
         *
         * The function retrieves the attributte with the internal
         * identificator <tt>attr_handle</tt>. The attribute stays associated
         * with the chunk. 
         *
         * This function can be used for example for iterating over all
         * attributes in the chunk. The order of attributes in the chunk is not
         * guaranteed. Use <tt>get_attr()</tt> with attribute name if you need
         * to access a particular chunk attribute.
         *
         * @note
         * The function die's if <tt>attr_handle</tt> is not a correct 
         * handle of a chunk's atttribute.
         *
         * @return
         * The function returns a pointer of chunk's attribute with handle
         * <tt>attr_handle</tt>.
         */
        class tEchelogType * get_attr(int32_t attr_handle) const;

        /** 
         * @brief Queries if the key attributes of messages chunks match. 
         *
         * <tt>key_match()</tt> checks if all key attributes in reference chunk
         * <tt>mc</tt> match with key attributes of this chunk, other
         * attributes are ignored. Matching attributes are such attributes
         * which have relation @link messages::tTYPE_RELATION::TP_IDENTICAL
         * TP_IDENTICAL@endlink or @link messages::tTYPE_RELATION::TP_EQUAL
         * TP_EQUAL@endlink (see @link messages::type_order()
         * type_order()@endlink function).

         * @return
         * The function returns true if both chunks have the same type and
         * all key attributes match, false otherwise.
         */
        bool tMsgChunk::key_match (tMsgChunk &mc);

        /**
         * @brief Stores message chunk into PlainBuf.
         * 
         * This function can be used to store (serialize) a chunk
         * to the given plain buffer <tt>buf</tt>. All chunk
         * internal structures are attributes are written. Note that
         * this function behaves atomically. If it fails for some reasons,
         * the used plain buffer <tt>buf</tt> remains untouched.
         *
         * @return
         * The function returns number of bytes stored to the plainbuffer.
         * On failure, a negative number is returned.
         */
        int32_t put (tPlainBuf *buf) const;

        /**
         * @brief Loads message chunk from PlainBuf.
         *
         * This function can be used to retrieve a serialized chunk from the
         * given plain buffer <tt>buf</tt>. It loads all chunk
         * structures and attributes. Note that
         * this function behaves atomically. If it fails for some reasons,
         * the used plain buffer <tt>buf</tt> remains untouched, i.e. as if
         * unread.
         * 
         * @return
         * The function returns number of bytes read from the plainbuffer.
         * On failure, a negative number is returned.
         */
        int32_t get (tPlainBuf *buf);        

        /**
         * @brief Get number of bytes needed for chunk serialization.
         *
         * The function returns number of bytes needed to store current
         * contents of the chunk including all attributes and internal 
         * structures.
         *
         * @return
         * Number of bytes needed to store the chunk.
         */
        int32_t get_store_length () const;

};

}

#endif


