/**
 * @file
 * @brief Message header container
 *
 * The message header can be used to fetch or set envelope information
 * of echelog messages. It is not needed to create and maintain it in separate,
 * see @link messages::tMsgBuf tMsgBuf::@endlink for overall interface.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef _ECHELOG_MSGHEADER_H
#define _ECHELOG_MSGHEADER_H

#include "messages/plainbuf.h"

#define ID_LENGTH 4

#define HDR_FLAG_INCREMENTAL 0x1
#define HDR_FLAG_AGGREGABLE 0x2

namespace messages {

/**
 * @brief Message header container.
 * 
 * This class can be used to fetch or set envelope information
 * of echelog messages. This class needn't to be created directly.
 * See See @link messages::tMsgBuf::get_header()
 * tMsgBuf::get_header()@endlink for message header retrieval.
 */
class tMsgHeader
{
        friend class tMsgBuf;

    protected:

        tPlainBuf *buffer;      ///< Header plain buffer
        bool r_prepared;        ///< True if the header is ready for reading
        bool w_prepared;        ///< True if the header is ready for writing

        /**
         * @brief <em>Read-only property:</em> Total message length field.
         */
        SIMPLE_R_PROPERTY (int32_t, total_msg_length);

        /**
         * @brief <em>Property:</em> Message type field.
         */
        SIMPLE_PROPERTY (int32_t, msg_type);

        /**
         * @brief <em>Property:</em> Message timestamp fielf.
         */
        SIMPLE_PROPERTY (int32_t, msg_time);

        /**
         * @brief <em>Property:</em> Number of subrecords (ie. message chunks)
         * field.
         */
        SIMPLE_PROPERTY (int32_t, subrecords);      
	
        /**
         * @brief <em>Property:</em> Agent identificator field. The field is
         * copied out as null-terminated string on request.
         */
        ARRAY_PROPERTY (char, ID_LENGTH, agent_id);
	
        /**
         * @brief <em>Property:</em> Agent instance identificator field. The
         * field is copied out as null-terminated string on request.
         */
        ARRAY_PROPERTY (char, ID_LENGTH, agent_instance_id);

        /**
         * @brief <em>Property:</em> Host identificator field. The field is
         * copied out as null-terminated string on request.
         */
        ARRAY_PROPERTY (byte, ID_LENGTH, host_id);

        /**
         * @brief <em>Property:</em> Message flags bit-field.
         */
        SIMPLE_PROPERTY (int32_t, flags);

        /**
         * @brief <em>Read-only property:</em> Total header length exluding
         * all header delimiters.
         */
        SIMPLE_R_PROPERTY (int32_t, len);

        /// Gets pure length of header (without all delimiters).
        inline int32_t pure_header_size () const;

        /// Gets header length field size in bytes.
        inline int32_t get_header_length_field_size() const;

        /// Gets length of total message length field in bytes.
        inline int32_t get_total_msg_length_field_size () const;

        /// Gets header length + initial message delimiter in bytes.
        inline int32_t get_store_length () const;

        /// Returns the associated buffer raw buffer.
        inline const char * get_buffer() const; 

        /// Returns true if the header is ready for reading.
        inline bool is_r_prepared () const;

        /// Returns true if the header is ready for writing.
        inline bool is_w_prepared () const;
        
        
        // In functions below a non-zero return status states an error 
        
        /**
         * @brief Prepares header for writing.
         *
         * <tt>prepare_to_write()</tt> creates a new plain buffer above the raw
         * buffer <tt>new_buffer</tt>. The raw buffer size <tt>max_length</tt>
         * must be big enough to hold the complete buffer structure including
         * intial message delimiters.
         *
         * @return 
         * On success, the function returns zero. On error, a negative number
         * is returned.
         */
        int prepare_to_write (const char *new_buffer, int32_t max_length);

        /**
         * @brief Prepares header for reading.
         *
         * <tt>prepare_to_read()</tt> creates a new plainbuffer and loads
         * the basic message delimiters from the underlying raw buffer
         * <tt>new_buffer</tt> of size <tt>max_length</tt>.  The given raw
         * buffer must contain a complete header.
         *
         * @return 
         * On success, the function returns zero. On error, a negative number
         * is returned.
         */ 
        int prepare_to_read (const char *new_buffer, int32_t max_length);

        /**
         * @brief Loads header information
         *
         * This function loads the complete header information from the
         * associated raw buffer to the prepared message header.
         * 
         * @return 
         * On success, the function returns zero. On error, a negative number
         * is returned.
         */
        int get (); 

        /**
         * @brief Stores header information
         *
         * This function stores the complete header information from the
         * prepared message header the associated raw buffer.
         * 
         * @return 
         * On success, the function returns zero. On error, a negative number
         * is returned.
         */
        int put ();
        
        /**
         * @brief Flushes and close message header.
         *
         * This function properly flushes header information to underlying
         * raw buffer and finalizes it.
         *
         * @return 
         * On success, the function returns zero. On error, a negative number
         * is returned.
         */
        int close ();

        /**
         * @brief Resets message header
         *
         * All allocated structures are released, the header is restored to
         * initial state (as if just newly created).
         *
         * @return 
         * On success, the function returns zero. On error, a negative number
         * is returned.
         */
        int reset ();

        
    public:

        /// Constructor---creates empty message header
        tMsgHeader () : buffer (0), r_prepared (false), 
        w_prepared (false), flags (0), len (EE_BUF_NOT_READY) { };

        /// Destructor---resets header and releases allocated space.
        virtual ~tMsgHeader ();

        /// Get host id in human readable string reprezentation 
        char *get_host_str () const;
        
        /**
         * @brief More convenient method to set the host id.
         *
         * This function sets a new host identificator but allows
         * to pass ordinary string as a parameter. It handles the 
         * neccessary type conversion automatically.
         */
        virtual void set_host_id (const char *id);

        /**
         * @brief Marks message incremental 
         *
         * This function enables the incremental flag in 
         * @link messages::tMsgHeader::flags flags@endlink field.
         */
        inline void set_flag_incremental (bool fl);

        /**
         * @brief Gets message incremental flag 
         *
         * This function retrieves the incremental flag status from
         * @link messages::tMsgHeader::flags flags@endlink field.
         *
         * @return
         * The function returns true if the flag is set, false otherwise.
         */
        inline bool get_flag_incremental () const;

        /**
         * @brief Marks message aggregable 
         *
         * This function enables the aggregable flag in 
         * @link messages::tMsgHeader::flags flags@endlink field.
         */
        inline void set_flag_aggregable (bool fl);

        /**
         * @brief Gets message aggregable flag 
         *
         * This function retrieves the aggregable flag status from
         * @link messages::tMsgHeader::flags flags@endlink field.
         *
         * @return
         * The function returns true if the flag is set, false otherwise.
         */
        inline bool get_flag_aggregable () const;
        
        /**
         * @brief Makes a carbon copy of header information
         *
         * <tt>copy_evenlope()</tt> retrives envelope information from the 
         * message header <tt>header</tt> and saves it to current header.
         *
         * Following fields are copied: message type, message timestamp, 
         * agent identificator, agent instance identificator, host
         * identificator, number of subrecors and message flags.
         */
        void copy_envelope (tMsgHeader &header);
};

}

/* Inline functions we actually define in here */
#include "messages/msgheader.hxx"

#endif
