/**
 * @file
 * @brief Message manager
 *
 * This file provides and interface for accessing features of registered 
 * messages and attributes.
 *
 * @author Ales Privetivy <ales@ucw.cz>
 */

#ifndef _ECHELOG_MESSAGEMGR_H
#define _ECHELOG_MESSAGEMGR_H

#include "messages/types/typemgr.h"

namespace messages {

/**
 * @brief Namespace containing classes and structures which are private to
 * %messages library implementation.
 */
namespace private_messages {

    /// Attribute definition holder.
    struct tAttributeDefinition {
        char *attr_name;            ///< Name of message attribute
        char *type_name;            ///< Echelog type name of attribute
        bool  key;                  ///< Key flag, true for key attributes
    };

    /// Message definition holder.
    struct tMsgDefinition {
        char *message_name;         ///< Message name
        int32_t message_type;       ///< Message type
        int32_t attr_count;         ///< Number of defined message attributes
        
        /// Array of subordinate attribute definitions
        struct tAttributeDefinition *attribute;
    };

    /**
     * @brief Temporary holder for message definitions (used for parsing).
     */
    struct tAttrBag {
        int32_t count;                  ///< Number of parsed attributes
        class tAttributeDefinition *al; ///< Parsed attribute definitions
    };

}

/**
 * @brief Message and attribute management
 *
 * Messages and their inner attributes are mostly referenced by string names.
 * For an effective use, we rather need an index of all defined attributes
 * and messages. This class provides a mechanism for translation between
 * various identificators. We can also ask at tMsgMgr about a handle to
 * particular message or attribute. This handle can be then used for faster
 * access to other features of requested objects. Note that these handles
 * may differ between configurations.
 */
class tMsgMgr {
        
        friend class tMsgChunk;

    protected:

        /// Internal structure for translation message and attribute names.
        struct private_messages::tMsgDefinition **message_def_list;
        /// Count of message definitions.
        int32_t message_def_count;

        /// Checks if fiven <tt>message handle</tt> is within its limits.
        void check_message_handle (int32_t message_handle);
        
        /// Checks if the <tt>attribute_handle</tt> is within its limits.
        void check_attribute_handle (private_messages::tMsgDefinition *md,
                int32_t attribute_handle);

        /**
         * @brief Creates array of messages' attributes.
         *
         * This function creates an array of all instances of the message
         * attributes with respective message handle <tt>message_handle</tt>.
         * The space is allocated on memory pool <tt>pool</tt>.
         *
         * @note
         * The function <tt>die()s</tt> on failure.
         *
         * @return
         * It returns a pointer to array of constructed echelog types.
         */
        class tEchelogType **create_attr_list(int32_t message_handle, 
                class lib::tMemPool &mpool);

        /**
         * @brief Makes deep copy of message definition.
         *
         * The function makes a deep copy of message definition <tt>def</tt>
         * within the def within <tt>msgdef_pool</tt> (see
         * @link messages::get_msgdef_pool() get_msgdef_pool()@endlink).
         *
         * @note
         * The function <tt>die()s</tt> on failure.
         *
         * @return
         * It returns a pointer to a message definition duplicate.
         */
        private_messages::tMsgDefinition *def_dup 
            (private_messages::tMsgDefinition *def);

    public:

        /**
         * @brief Constructor---creates empty tMsgMgr object.
         */
        tMsgMgr();  
        
        /**
         * @brief Destructor---releases all subordinate structures.
         */
        virtual ~tMsgMgr() {}


        /**
         * @brief Registers list of messages.
         *
         * This function iterates over the message definition list
         * <tt>msg_def_l</tt>, and sets up all neccessary structures for
         * operations above them. A unique handle is assigned to each
         * definition.
         */
        void register_message_list(class lib::tList *msg_def_l);	

        /**
         * @brief Gets message handle by its name.
         *
         * The function searches for a handle of the message with name
         * <tt>message_name</tt>.
         *
         * @return
         * On success, the function returns the message handle. On failure,
         * <tt>-1</tt> is returned.
         */
        int32_t get_handle_by_name(const char *message_name);

        /**
         * @brief Gets message handle by its type.
         *
         * The function searches for a handle of the message with type
         * <tt>message_type</tt>.
         *
         * @return
         * On success, the function returns the message handle. On failure,
         * <tt>-1</tt> is returned.
         */
        int32_t get_handle_by_type (int32_t type);

        /**
         * @brief Get echelog type name of given attribute.
         *
         * The function finds an echelog type name identified by a pair of the
         * attribute name <tt>attr</tt> and a message of the given handle
         * <tt>message_handle</tt>. For echelog type construction see @link
         * messages::tTypeMgr::get_type_instance() get_type_instance()@endlink.
         *
         * @return
         * On success the funtion returns the requsted echelog type name.
         * On failure, NULL is returned.
         */ 
        char *get_type_of_attr(int32_t message_handle, const char *attr);

        /**
         * @brief Checks if given attribute is key attribute.
         *
         * The function checks if the attribute identified by a pair
         * of message handle <tt>message_handle</tt> and attribute name
         * <tt>attr</tt> is a message key attribute.
         *
         * @return
         * If the given attribute is a key attribute of the message, the
         * function returns a non-zero integer. If the attribute isn't a key
         * attribute or if it doesn't exist in the given message, zero is
         * returned.
         */
        bool is_key_attr(int32_t message_handle, const char *attr);

        /**
         * @brief Checks if given attribute is key attribute.
         *
         * The function checks if the attribute identified by a pair
         * of message handle <tt>message_handle</tt> and attribute handle
         * <tt>attr_handle</tt> is a message key attribute.
         *
         * @return
         * If the given attribute is a key attribute of the message, the
         * function returns a non-zero integer. If the attribute isn't a key
         * attribute, zero is returned.
         */
        bool is_key_attr_by_handle(int32_t message_handle, int32_t attr_handle);

        /**
         * @brief Checks if given message type can be aggregated.
         *
         * The function checks for existence of key attributes in the message
         * definition idenitified by handle <tt>message_handle</tt>. If there 
         * are such attributes, the given message type can be aggregated.
         *
         * @return
         * The function returns true if there are key attributes in the given
         * message definition, otherwise it returns false.
         */
        bool is_aggregable(int32_t message_handle); 
        
        /**
         * @brief Gets number of attributes of given message type identified 
         * by <tt>message_handle</tt>.
         */
        int32_t get_attr_count(int32_t message_handle);

        /**
         * @brief Gets attribute name of given attribute.
         *
         * The function searches for the attribute identified by a pair
         * of handles <tt>message_handle</tt> and <tt>attr_handle</tt>
         * and reports it attribute name. This function can be useful
         * when iterating over all attributes of a certain message type.
         *
         * @return
         * The function returns the attribute name of the given attribute.
         */
        char * get_attr_name (int32_t message_handle, int32_t attr_handle);

        /**
         * @brief Gets attribute handle of given attribute.
         *
         * The function searches for the attribute identified by message
         * handle <tt>message_handle</tt> and attribute name <tt>attr</tt>
         * and reports its handle.
         * 
         * @return
         * The function returns the handle of the given attribute. If the
         * attribute doesn't exist in the given message, -1 is returned.
         */
        int32_t get_handle_of_attr(int32_t message_handle, const char *attr);

        /**
         * @brief Gets message type of given message.
         *
         * The function searches and reports the type of the message
         * identified by handle <tt>message_handle</tt>.
         *
         * @return
         * The respective message type identificator is returned.
         */
        int32_t get_message_type_by_handle(int32_t message_handle);

        /**
         * @brief Get message name of given message.
         *
         * The function searches and reports the name of the message
         * identified by handle <tt>message_handle</tt>.
         *
         * @return
         * The respective message name is returned.
         */
        char *get_message_name_by_handle(int32_t message_handle);

};

/// Global pool used for message definitions information
lib::tMemPool &get_msgdef_pool ();

/// Global instance of Message Manager 
extern class tMsgMgr gb_message_manager;

}

#endif


