/**
 * @file
 * @brief Plain buffer handling
 *
 * The plain buffer provides a simple interface for low-level message data 
 * manipulation. It is mostly used by other components in %messages library.
 *      
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef _ECHELOG_PLAINBUF_H
#define _ECHELOG_PLAINBUF_H

#include "property.h"
#include "lib/lib.h"
#include "messages/msgstatus.h"

/// Default plain buffer size if constructed with no parameters
#define DEFAULT_BUF_SIZE 4096

/// General plain buffer error
#define EE_BUF_GENERAL_ERROR -1
/// Plain buffer is not ready for reading
#define EE_BUF_R_NOT_READY -2
/// Plain buffer is not ready for writing
#define EE_BUF_W_NOT_READY -3
/// Plain buffer is not ready for any i/o operations
#define EE_BUF_NOT_READY -4
/// Attempt to write behind the buffer limit
#define EE_BUF_EXCEEDED -5
/// Buffer contents corrupted
#define EE_BUF_READ_ERROR -6
/// Can't complete write operation, buffer is too small 
#define EE_BUF_STORE_ERROR -7

/// Failed to read the message chunk structure
#define EE_BUF_CHUNK_R_FAILED -8
/// Failed to write the message chunk structure
#define EE_BUF_CHUNK_W_FAILED -9

/// Operation with plain buffer has been successfully completed
#define SUCCESS 0

/**
 * @namespace messages
 * @brief This namespace contains all classes and functions of %messages 
 * library.
 */
namespace messages {

/**
 * tPlainBuf covers the I/O message buffering. It's used mainly by
 * message buffer and attribute types get()/put() functions.
 * 
 * Plain message buffer class holding buffer internal state
 *
 * An allocated or loaded buffer must be passed when tMsgBuf is to be created.
 * If constructed without any parameters the buffer is enlarged when needed.
 *
 * The byte order in the message buffer is guaranteed to be the network order.
 * The store functions return number of bytes written to buffer. The negative
 * number denotes an error.
 *
 * The read/store functions return the number of bytes read/written. The 
 * negative value denotes an error.
 *
 * Note: all lengths are without their headers.
 */
class tPlainBuf
{
    protected:
        SIMPLE_R_PROPERTY (char *, buffer);     ///< main store

        int32_t max_length;     //< buffer length
        char *cursor;           //< current buffer cursor position
        bool foreign_buffer;    //< true if the buffer is given from outside
        
    public:
        /// The buffer will be reallocated on the fly
        inline tPlainBuf();
        /// The buffer must be allocated and large enough
        inline tPlainBuf(const char *new_buffer, int32_t length);
        /// The buffer space is deallocated in case of internal buffer
        virtual inline ~tPlainBuf();

        /// Read/store the submessage's status field
        inline int32_t read_state (tMSG_STATUS &st);
        inline int32_t store_state (tMSG_STATUS st);

        /// Read/write a single character
        inline int32_t read_char (char &val);
        inline int32_t store_char (char val);
        
        /// Read/write a byte
        inline int32_t read_byte (byte &val);
        inline int32_t store_byte (byte val);

        /// Read/write int16 to buffer
        inline int32_t read_int16 (int16_t &val);
        inline int32_t store_int16 (int16_t val);

        /// Read/write int32 integer to buffer
        inline int32_t read_int32 (int32_t &val);
        inline int32_t store_int32 (int32_t val);

        /** Read/write compressed 32bit integer to buffer
         *
         * Mostly the length is lesser than 255 and can be stored
         * as a single byte. If field's length is 
         * greater the first byte's stutus is set to 0xff
         * and the actual length is written thereafter as 32bit integer.
         */
        inline int32_t read_cint32 (int32_t &val);
        inline int32_t store_cint32 (int32_t val);

        /** Read a string to buffer
         *
         * The length should be exactly number of bytes stored 
         * in buffer (thus not counting the terminating null)
         *
         * The string is newly allocated and copied to a given memory pool
         * (the terminating null is added). The 'string' parameter is modified
         * to point to that copy. 
         *
         * Function returns the number of bytes actually read
         */
        inline int32_t read_string (lib::tMemPool &pool, char *&string, 
                int32_t length);

        /** Read a string to buffer
         *
         * The string is newly allocated and copied to mempool 'pool'
         * (the terminating null is added). The string length is read
         * from buffer in the form of cint32.
         *
         * The 'string' parameter is modified
         * to point to that copy. 
         *
         * Function returns the number of bytes actually read
         */
        inline int32_t read_cstring (lib::tMemPool &pool, char *&string);

        /** Read a string to buffer
         *
         * The string is newly allocated and copied to standard heap
         * (the terminating null is added). The string length is read
         * from buffer in the form of cint32.
         *
         * The 'string' parameter is modified
         * to point to that copy. 
         *
         * Function returns the number of bytes actually read
         */
        inline int32_t read_cstring (char *&string);

        /** Write string to buffer
         *
         * Store: String is stored to buffer excluding its terminating null.
         * Function returns the number of bytes actually written.
         */
        inline int32_t store_string (const char *string);

        /** Write string portion to buffer
         *
         * Store: String portion of length bytes is stored to buffer.
         * It should not contain a termination null.
         *
         * Function returns the number of bytes actually written.
         */
        inline int32_t store_string_portion(const char *buffer, int32_t length);

        /// Get actual length of buffer in bytes
        inline int32_t get_store_length ();

        /// Get actual cursor position 
        inline int32_t get_cursor_position () 
        { return cursor - buffer; }

        /// Set new relative cursor position
        inline void seek (int32_t relative_shift) 
        { cursor += relative_shift; }

        /// Absolute position seek
        inline void abs_seek (int32_t position)
            { ASSERT (position >= 0); cursor = buffer + position; }

        /// Reset cursor to the beginning of the buffer
        inline void rewind () { cursor = buffer; }

        /// Return true if there is not enough space in the buffer
        inline bool check_spare_space (int32_t wanted_space)
            { return (foreign_buffer ? 
                    ((max_length < wanted_space + get_cursor_position ()) ? 
                        true : false) 
                    : false); }
};


/* PlainBuf error messages. Stored separately to avoid bloating the code. */
extern char *gb_plainbuf_errors[];

/* Get sizeof cint32 number */
inline int32_t cint32_size (int32_t number_to_be_stored_as_cint32);

}

/* We actually define all inline functions in a separate file */
#include "messages/plainbuf.hxx"

#endif
