/**
 * @brief Echelog Library -- Plain buffer handling
 * This class covers the I/O message buffering. 
 * It's used mainly by message buffer, header and "types'"
 * get()/put() functions.
 *
 * These are all inline functions.
 *
 * Important note: This code is arch-dependent. Though, it does work
 * for most Intel and non-Intel architectures. One of the assumptions
 * is that integer types may be directly stored to char locations.
 *      
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef __ECHELOG_PLAINBUF_HXX
#define __ECHELOG_PLAINBUF_HXX

#include <netinet/in.h>
#include <limits.h>


/*
 * We don't want to type all of the checks again and again 
 */
#define __CHECKLOWER_LIMIT() \
    do { using namespace lib; \
        if (buffer - cursor > 0) \
        { \
            log (L_ERROR, \
            gb_plainbuf_errors[0]); \
            return EE_BUF_EXCEEDED; \
        } \
    } while (0) 

#define __R_CHECKUPPER_LIMIT(len) \
    do { using namespace lib; \
        if (static_cast<int32_t> (cursor - buffer + (len)) > max_length) \
        { \
                log (L_ERROR, \
                gb_plainbuf_errors[1]); \
                return EE_BUF_EXCEEDED; \
        } \
    } while (0)
        
#define __W_CHECKUPPER_LIMIT(len) \
    do { using namespace lib; \
        if (static_cast<int32_t> (cursor - buffer + (len)) > max_length) \
        { \
            if (foreign_buffer == true) \
            { \
                log (L_ERROR, \
                gb_plainbuf_errors[2]); \
                return EE_BUF_EXCEEDED; \
            } \
            else \
            { \
                max_length <<= 2; \
                buffer = static_cast<char *> (realloc (buffer, max_length)); \
                if (!buffer) \
                    die(gb_plainbuf_errors[3], max_length); \
            } \
        } \
    } while (0)

/**
 * This code supports architectures which require int locations to 
 * be memory-aligned. The network order must be assured before/after.
 *
 * Often, the compiler is smart enough to unroll the loop to individual
 * byte assignments.
 *
 * Even then, since the c++ norm forbids it but it works everywhere,
 * we use casting technique to read/write individual bytes.
 * For this reason -fno-strict-aliasing option is required in GNU C.
 */
#define __STORE_INTEGER(which, buffer, value) \
 do { \
    char *__individual_bytes = reinterpret_cast<char *> (&value); \
    for (unsigned __st_idx = 0; __st_idx < sizeof (which); __st_idx++) \
        (buffer)[__st_idx] = __individual_bytes[__st_idx]; \
    buffer += sizeof (which); \
 } while (0)

#define __READ_INTEGER(which, buffer, value) \
 do { \
    char *__individual_bytes = reinterpret_cast<char *> (&value); \
    for (unsigned __st_idx = 0; __st_idx < sizeof (which); __st_idx++) \
        __individual_bytes[__st_idx] = (buffer)[__st_idx]; \
    buffer += sizeof (which); \
 } while (0)


namespace messages {

    
/* Get sizeof cint32 number */
inline int32_t cint32_size (int32_t number_to_be_stored_as_cint32)
{
  if (number_to_be_stored_as_cint32 < static_cast<int32_t> (UCHAR_MAX))
      return 1;

  return static_cast<int32_t> (sizeof (int32_t) + 1);
}


inline tPlainBuf::tPlainBuf ()
{
    using namespace lib;

    buffer = static_cast<char *> (malloc (DEFAULT_BUF_SIZE));
    cursor = buffer;
    max_length = DEFAULT_BUF_SIZE;
    foreign_buffer = false;
}


inline tPlainBuf::tPlainBuf (const char *new_buffer, int32_t length)
{
    using namespace lib;
    if (!new_buffer)
        die (gb_plainbuf_errors[4]);

    buffer = const_cast<char *> (new_buffer);
    cursor = buffer;
    max_length = length;
    foreign_buffer = true;

    if (length < 0)
    {
        log (L_ERROR, gb_plainbuf_errors[5]);
        max_length = 0;
    }
}


inline tPlainBuf::~tPlainBuf ()
{
    if (!foreign_buffer && buffer)
        free (buffer);
}


inline int32_t tPlainBuf::read_state (tMSG_STATUS &st)
{
    __CHECKLOWER_LIMIT ();
    __R_CHECKUPPER_LIMIT (1);

    st = static_cast<tMSG_STATUS> (*cursor++);
    return 1;
}


inline int32_t tPlainBuf::store_state (tMSG_STATUS st)
{
    __CHECKLOWER_LIMIT ();
    __W_CHECKUPPER_LIMIT (1);

    *cursor++ = static_cast<tMSG_STATUS> (st);
    return 1;
}


inline int32_t tPlainBuf::read_char (char &val)
{
    __CHECKLOWER_LIMIT ();
    __R_CHECKUPPER_LIMIT (1);

    val = *cursor++;
    return 1;
}


inline int32_t tPlainBuf::store_char (char val)
{
    __CHECKLOWER_LIMIT ();
    __W_CHECKUPPER_LIMIT (1);

    *cursor++ = val;
    return 1;
}


inline int32_t tPlainBuf::read_byte (byte &val)
{
    __CHECKLOWER_LIMIT ();
    __R_CHECKUPPER_LIMIT (1);

    val = static_cast<byte> (*cursor++);
    return 1;
}


inline int32_t tPlainBuf::store_byte (byte val)
{
    __CHECKLOWER_LIMIT ();
    __W_CHECKUPPER_LIMIT (1);

    *cursor++ = static_cast <char> (val);
    return 1;
}


inline int32_t tPlainBuf::read_int16 (int16_t &val) 
{
    int16_t tmp;

    __CHECKLOWER_LIMIT ();
    __R_CHECKUPPER_LIMIT (sizeof(int16_t));

    __READ_INTEGER(int16_t, cursor, tmp);
    val = ntohs (tmp);

    return sizeof(int16_t);
}


inline int32_t tPlainBuf::store_int16 (int16_t val)
{
    int16_t tmp;

    __CHECKLOWER_LIMIT ();
    __W_CHECKUPPER_LIMIT (sizeof(int16_t));

    tmp = htons (val);
    __STORE_INTEGER(int16_t, cursor, tmp);

    return sizeof(int16_t);
}


inline int32_t tPlainBuf::read_int32 (int32_t &val)
{
    int32_t tmp;

    __CHECKLOWER_LIMIT ();
    __R_CHECKUPPER_LIMIT (sizeof(int32_t));

    __READ_INTEGER(int32_t, cursor, tmp);
    val = ntohl (tmp);

    return sizeof(int32_t);
}


inline int32_t tPlainBuf::store_int32 (int32_t val)
{
    int32_t tmp;

    __CHECKLOWER_LIMIT ();
    __W_CHECKUPPER_LIMIT (sizeof(int32_t));

    tmp = htonl (val);
    __STORE_INTEGER(int32_t, cursor, tmp);

    return sizeof(int32_t);
}


inline int32_t tPlainBuf::read_cint32 (int32_t &val)
{
    __CHECKLOWER_LIMIT ();
    __R_CHECKUPPER_LIMIT (1);

    val = static_cast<byte> (*cursor);

    if (val == 0xff)
    {
        int32_t tmp;

        __R_CHECKUPPER_LIMIT (sizeof(int32_t) + 1);
        cursor++;
        __READ_INTEGER(int32_t, cursor, tmp);
        val = static_cast<int32_t> (ntohl (tmp));

        return sizeof(int32_t) + 1;
    }
    else
        cursor++;
    
    return 1;
}


inline int32_t tPlainBuf::store_cint32 (int32_t val)
{
    int32_t tmp;

    __CHECKLOWER_LIMIT ();
    
    if (val < static_cast<int32_t> (UCHAR_MAX))
    {
        __W_CHECKUPPER_LIMIT (1);
        *cursor++ = static_cast<char> (val);

        return 1;
    }

    __W_CHECKUPPER_LIMIT (sizeof(int32_t) + 1);
    *cursor++ = static_cast<char> (0xff);
    tmp = htonl(static_cast<int32_t>(val));
    __STORE_INTEGER (int32_t, cursor, tmp);

    return sizeof(int32_t) + 1;
}


inline int32_t tPlainBuf::read_string (lib::tMemPool &pool, char *&string, 
    int32_t length) 
{
    using namespace lib;
    __CHECKLOWER_LIMIT ();

    if (length <= 0)
    {
        log (L_ERROR, gb_plainbuf_errors[6]);
        return EE_BUF_GENERAL_ERROR;
    }

    __R_CHECKUPPER_LIMIT (length);
    string = static_cast<char *> (pool.alloc (length + 1));
    memcpy (string, cursor, length);
    string[length] = 0;

    cursor += length;
    return length;
}


inline int32_t tPlainBuf::read_cstring (lib::tMemPool &pool, char *&string) 
{
    using namespace lib;
    char *old_cursor = cursor;
    int32_t length;
    int32_t bytes_read;

    bytes_read = read_cint32 (length);

    if (length <= 0)
    {
        log (L_ERROR, gb_plainbuf_errors[6]);
        cursor = old_cursor;
        return EE_BUF_GENERAL_ERROR;
    }

    if (static_cast<int32_t> (cursor - buffer + (length)) > max_length) 
    { 
        if (foreign_buffer == true) 
        { 
            cursor = old_cursor;
            log (L_ERROR, gb_plainbuf_errors[1]); 
            return EE_BUF_EXCEEDED; 
        } 
        else 
        { 
            max_length <<= 2; 
            buffer = static_cast<char *> (realloc (buffer, max_length));
            if (!buffer) 
                die(gb_plainbuf_errors[3], max_length); 
        } 
    } 

    string = static_cast<char *> (pool.alloc (length + 1));
    memcpy (string, cursor, length);
    (string)[length] = 0;

    cursor += length;
    return length + bytes_read;
}


inline int32_t tPlainBuf::read_cstring (char *&string) 
{
    using namespace lib;
    char *old_cursor = cursor;
    int32_t length;
    int32_t bytes_read;

    bytes_read = read_cint32 (length);

    if (length <= 0)
    {
        log (L_ERROR, gb_plainbuf_errors[6]);
        cursor = old_cursor;
        return EE_BUF_GENERAL_ERROR;
    }

    if (static_cast<int32_t> (cursor - buffer + (length)) > max_length) 
    { 
        if (foreign_buffer == true) 
        { 
            cursor = old_cursor;
            log (L_ERROR, gb_plainbuf_errors[1]);
            return EE_BUF_EXCEEDED; 
        } 
        else 
        { 
            max_length <<= 2; 
            buffer = static_cast<char *> (realloc (buffer, max_length));
            if (!buffer) 
                die(gb_plainbuf_errors[3], max_length); 
        } 
    } 

    string = static_cast<char *> (xmalloc (length + 1));
    memcpy (string, cursor, length);
    (string)[length] = 0;

    cursor += length;
    return length + bytes_read;
}


inline int32_t tPlainBuf::store_string (const char *string)
{
    using namespace lib;
    int32_t length;

    __CHECKLOWER_LIMIT ();

    if (!string)
    {
        log (L_ERROR, gb_plainbuf_errors[7]);
        return EE_BUF_GENERAL_ERROR;
    }

    length = strlen (string);
    __W_CHECKUPPER_LIMIT (length);
    memcpy (cursor, string, length);
    cursor += length;

    return length;
}


inline int32_t tPlainBuf::store_string_portion (const char *buf, int32_t length)
{
    using namespace lib;

    __CHECKLOWER_LIMIT ();

    if (!buf)
    {
        log (L_ERROR, gb_plainbuf_errors[7]);
        return EE_BUF_GENERAL_ERROR;
    }

    __W_CHECKUPPER_LIMIT (length);
    memcpy (cursor, buf, length);
    cursor += length;

    return length;
}


inline int32_t tPlainBuf::get_store_length ()
{
    __CHECKLOWER_LIMIT ();

    return cursor - buffer;
}

}
#endif
