/**
 * @file
 * @brief Echelog Library -- "Type" handling
 * This type define the basic opertaions for handling "complex" echelog types.
 *      
 * @author Kamil Toman <ktoman@email.cz>
 */
#ifndef _ECHELOG_COMPLEX_H
#define _ECHELOG_COMPLEX_H

#include "messages/testbuf.h"
#include "messages/types/echelogtype.h"

namespace messages {

/**
 * tComplex is an automatically generated echelog "complex type"
 */
class tComplex : public tEchelogType
{
    protected:
        const char **components_names;
        tEchelogType **components;
        
        /** Get a component */
        virtual tEchelogType *get_component_by_path (lib::tList &path) const;

    public:
        /** The value is explicitly set null on creation */
        tComplex(lib::tMemPool &mem_pool_buffer) 
            : tEchelogType(mem_pool_buffer) {};
        /** Complex type destructor */
        virtual ~tComplex() {};

        /** Get real store length in bytes */
        virtual int32_t get_store_length ();
        /** Return the pure type len */
        virtual int32_t get_len ();
        /** Set the whole complex type to null value */
        inline virtual void set_null ()
        {
            for (int32_t i = 0; i < get_num_of_components(); i++)
                components[i]->set_null();
        }

        /** As always we have to at least implement loading from a buffer */
        virtual int32_t get (tPlainBuf *buf); 
        /** As always we have to at least implement storing to a buffer */
        virtual int32_t put (tPlainBuf *buf);

        /** Common debugging print outs */
        virtual long sprint (lib::tList &list, lib::tMemPool &pool, 
            const tFmtNode *node = NULL);

        /** Get the number of components */
        virtual int32_t get_num_of_components () const = 0;

        /** Get a component */
        virtual tEchelogType *get_component_by_path (const lib::tList &path) 
            const;

        /** Get a component */
        virtual tEchelogType *get_component_by_num (int32_t comp_num) const
            { if (comp_num >= 0 && comp_num < get_num_of_components ())
                  return components[comp_num];
              else
                  return NULL;
            }

        /** Get a component name */
        virtual const char *get_component_name_by_num (int32_t comp_num) const
            { if (comp_num >= 0 && comp_num < get_num_of_components ())
                  return components_names[comp_num];
              else
                  return NULL;
            }

        /** Get the plain formating tree */
        virtual const tFmtNode * get_plain_format()
        { return get_default_format(); }

        /** Scan the string list */
        virtual int sscan (lib::tList &string_list);

        /** Type assignment */
        virtual tTYPE_RELATION assign (tEchelogType &tp);

        /** Type comparison '==' operator */
        virtual tTYPE_RELATION order (tEchelogType &tp);

        /** Merge extra information from tp */
        virtual int merge (tEchelogType &tp);
};

}

#endif
