/**
 * @file
 * @brief Echelog Library -- "Type" handling
 *
 * Any new basic or complex echelog "type" should be inherited 
 * from this abstract base class.
 *
 * Since we're not willing to use templates (as they are still buggy
 * and/or expensive feature in most compilers) it misses some other
 * general methods like 'get_value()' or 'set_value()' which should
 * be also implemented for each "type".
 *      
 * @author Kamil Toman <ktoman@email.cz>
 *
 */
#ifndef _ECHELOG_ECHELOGTYPE_H
#define _ECHELOG_ECHELOGTYPE_H

#include "lib/lib.h"
#include "messages/printout.h"

#include <limits.h>

namespace messages {

typedef int32_t tTypeNumber;

typedef enum {
    TP_IDENTICAL = 1,       ///< Types have an identical value
    TP_EQUAL = 0,           ///< Types are considered equal (match, no conflict)

    TP_LESS = -1,           ///< The first type is less than the second one
    TP_GREATER = -2,        ///< The first type is greater to the second one
    TP_CONFLICT = -10,      /**< Types have conflicting (i.e. not equal
                                  non-empty) components */
    TP_INCOMPARABLE = -20,  ///< Types are incomparable
    TP_INCOMPATIBLE = -100  ///< Incompatible types in relation
} tTYPE_RELATION;


#define TEST_ASSIGN() \
do { \
    if (!is_compatible (tp)) \
        return TP_INCOMPATIBLE; \
    if (this == &tp) \
        return TP_IDENTICAL; \
}\
while (0)


#define COMMON_ORDER() \
do { \
    if (!is_compatible (tp)) \
        return TP_INCOMPATIBLE; \
    if (is_null() && !tp.is_null()) \
        return TP_LESS; \
    else if (!is_null() && tp.is_null()) \
        return TP_GREATER; \
    else if (is_null() && tp.is_null()) \
        return TP_IDENTICAL; \
}\
while (0) 

#define NUMERIC_ORDER() \
do { \
    if (value == tpc.value) \
        return TP_IDENTICAL; \
    else if (value < tpc.value) \
        return TP_LESS; \
    else \
        return TP_GREATER; \
} \
while (0) 

/// TODOC:
class tEchelogType
{
    protected:
        lib::tMemPool &mempool;         ///< placeholder
        int32_t len;                       ///< pure type length

        SIMPLE_R_PROPERTY (long , last_printout_len); //< without the term null

        /** Get type info */
        virtual tTypeNumber get_type () const = 0;

    public:
        /** Create a new "type" and assign it a null value */
        tEchelogType (lib::tMemPool &mem_pool_buffer) 
            : mempool (mem_pool_buffer),
                len (0), last_printout_len (0) {};

	virtual ~tEchelogType() {}

        /** get() and put() methods should return the real number of bytes
         * read/stored on success, zero or negative (with the value 
         * of EE_BUF_... errno set) on failure 
         */

        /** Load "type" contents from the given message buffer */
        virtual int32_t get (tPlainBuf *buf) = 0;
        /** Store "type" contents to the message buffer */
        virtual int32_t put (tPlainBuf *buf) = 0;

        /** Return the number of bytes needed to store "type" to a buffer */
        virtual int32_t get_store_length () = 0;

        /** Return the pure type len */
        virtual int32_t get_len () { return len; }

        /** Set "type" value to null */
        inline virtual void set_null () { len = 0; }

        /** Return true if the value is null */
        inline bool is_null ()
            { return (get_len() == 0) ? true : false; }

        /** Get the number of type components */
        virtual int32_t get_num_of_components () const
            { return 0; }

        /** Get a component */
        virtual tEchelogType *get_component (const char *comp_name) const
            { lib::tMemPool p(1024); lib::tList l(p);
              if (s_strdsep (l, comp_name, ".", p) < 0)
                  return NULL;
              else
                  return get_component_by_path (l); }

        /** Get a component */
        virtual tEchelogType *get_component_by_path (const lib::tList &path) 
            const
            { return NULL; }

        /** Get a component */
        virtual tEchelogType *get_component_by_num (int32_t comp_num) const
            { return NULL; }

        /** Get a component name */
        virtual const char *get_component_name_by_num (int32_t comp_num) const
        { return NULL; }

        /** Common debugging print outs */
        virtual long sprint (lib::tList &list, lib::tMemPool &pool, 
                const tFmtNode *node = NULL) = 0;

        /** Get the default formating tree */
        virtual const tFmtNode * get_default_format() = 0;

        /** Get the plain formating tree */
        virtual const tFmtNode * get_plain_format() = 0;

        /** Check if the type is of a compatible type */
        inline bool is_compatible (tEchelogType &tp) const
            { return get_type() == tp.get_type(); }

        /** Scan the string list */
        virtual int sscan (lib::tList &string_list) = 0;

        /** Type assignment */
        virtual tTYPE_RELATION assign (tEchelogType &tp) = 0;
        
        /** Type comparison '==' operator */
        virtual tTYPE_RELATION order (tEchelogType &tp) = 0;

        /** Merge extra information from tp */
        virtual int merge (tEchelogType &tp);
};


inline tTYPE_RELATION type_order (tEchelogType &tp1, tEchelogType &tp2)
{
    return tp1.order (tp2);
}


}

#endif
