/**
 * @file
 * @brief Echelog Library -- "Type" handling
 * This type can be used for handling variable-length strings.
 *      
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef _ECHELOG_VARCHAR_H
#define _ECHELOG_VARCHAR_H

#include "messages/types/echelogtype.h"

namespace messages {

/**
 * TODOC:
 */
class tVarchar : public tEchelogType
{
    typedef enum { 
        A_NORMAL,             // standard property value
        A_DIRECT,                 // directly assigned
        A_DIRECT_NON_TERMINATED   // directly assigned string portion
    } tStringAssignment;

    protected:
        PROPERTY (char *, value);               ///< string value

        tStringAssignment assign_type;          ///< allocation flag
        /** string length indicates the real string length without
         * its terminating null */

        /** Get type info */
        virtual tTypeNumber get_type () const;

    public:
        /** The value is explicitly set null on creation */
        tVarchar (lib::tMemPool &mem_pool_buffer)
            : tEchelogType (mem_pool_buffer), value (0), 
            assign_type(A_DIRECT) { }

        /** Create the "type" and assign the given value */
        tVarchar (lib::tMemPool &mem_pool_buffer, const char *val);

        /** Must do a copy deallocation if the value wasn't set directly */
        virtual tVarchar::~tVarchar ();

        /** Direct set/get. Unlike true property functions
         * no extra allocation/copying is done. 
         * The letter set function can be used to assign various
         * string portions directly when no additional manipulation is needed.
         * Use all these functions with extra care. */
        const char * get_direct_value () { return value; }
        void set_direct_value (const char * val);
        void set_direct_value (const char * val, int32_t val_length);

        /** As always we have to at least implement buffer load */
        virtual int32_t get (tPlainBuf *buf);
        /** As always we have to at least implement buffer store */
        virtual int32_t put (tPlainBuf *buf);

        /** Get real store length in bytes */
        virtual int32_t get_store_length ()
           { return len + cint32_size (len); }
        
        /** Common debugging print outs */
        virtual long sprint (lib::tList &list, lib::tMemPool &pool, 
                const tFmtNode *node = NULL);

        /** Get the default formating tree */
        virtual const tFmtNode * get_default_format();

        /** Get the plain formating tree */
        virtual const tFmtNode * get_plain_format();

        /** Scan the string list */
        virtual int sscan (lib::tList &string_list);

        /** Type assignment */
        virtual tTYPE_RELATION assign (tEchelogType &tp);
        
        /** Type comparison '==' operator */
        virtual tTYPE_RELATION order (tEchelogType &tp);
};

}

#endif
