/**
 * @file
 * @brief Contains declaration of tEchelogModule class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#ifndef ECHELOG_ECHELOGMODULE_H
#define ECHELOG_ECHELOGMODULE_H

#include "mframes/ioframe.h"
#include "mframes/reconfigurablemodule.h"
#include "mframes/socketio.h"

namespace mframes {

/**
 * TODOC:
 * @brief Base class for common echelog modules.
 *
 * This class encapsulates all basic echelog filters functionality.
 * Reads common configuration and establishes socket communication, handles
 * signals, reconfiguration, reads messages and passes them to the filter
 * specified in the initialization for further processing.
 *
 * Typical instance lifecycle is initialization by
 * init(), work performed by run() and cleanup by cleanup().
 */
class tEchelogModule : public tReconfigurableModule
{
    public:
        //-------------------------------
        // Interface
        //-------------------------------

        /**
         * TODOC:
         * @brief Filter module's cleanup
         *
         * Call this function to clean and release all resources. Should
         * be called after successful initialization by init().
         */
        virtual void cleanup ();

    protected:
        /**
         * TODOC:
         * @brief Filter module's initialization
         *
         * Call this function to initialize whole filter module.
         *
         * @return
         * On success returns false. On failure returns true.
         */
        bool init (tSocketIO* sockio, tIFrame* iframe, tOFrame* oframe);

        /**
         * @brief Constructor
         */
        tEchelogModule (tCmdLine& cmdline, const char* cfg_filename);

        /// Initialization reconfiguration independent part
        virtual bool static_init ();

        /// Initialization reconfiguration dependent part
        virtual bool dynamic_init ();

        /// Reverse method for static_init()
        virtual void static_cleanup ();

        /// Reverse method for dynamic_init()
        virtual void dynamic_cleanup ();

        /// Handles installed signals
        virtual SigHandlingRet handle_signals ();

        /// Handles socket error
        virtual SigHandlingRet on_sockerr ();

        /// Returns true if socket error has occured
        inline bool sockerr_received ()
        { return s_sockerr_received; }

        //-------------------------------
        // Implementation of tConfigurable interface
        //-------------------------------

        /**
         * @brief TODOC:
         */
        virtual bool register_sections (lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void unregister_sections ();

        /**
         * @brief TODOC:
         */
        virtual bool process_config (lib::tMemPool& conf_mp, 
                lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void cleanup_config ();

    private:
        //-------------------------------
        // Internal types
        //-------------------------------

        /// TODOC: Enum for filter module's state identification
        enum EchelogModuleState
        {
            EM_CREATED,
            EM_STAT_INITIALIZED,
            EM_DYN_INITIALIZED
        };

        //-------------------------------
        // Object state attributes
        //-------------------------------

        /// Actual state of filter module
        EchelogModuleState m_emstate;

        /// Flag to set when error in socket communication occured
        volatile static bool s_sockerr_received;

        //-------------------------------
        // Used-only components
        //-------------------------------

        /// Pointer to common io
        tSocketIO* mp_sio;

        /// Pointer to input interface
        tIFrame* mp_iframe;

        /// Pointer to output interface
        tOFrame* mp_oframe;
};

} //end of namespace mframes

#endif //ifdef ECHELOG_ECHELOGMODULE_H

