/**
 * @file
 * @brief Contains declaration of tModule class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#ifndef ECHELOG_MODULE_H
#define ECHELOG_MODULE_H

#include "mframes/configurable.h"

/**
 * TODOC:
 * @namespace mframes
 * @brief Echelog Filters support library namespace
 *
 * All parts of the %mframes library are contained in this namespace. If you
 * wanna use this library you should use directive <tt>using namespace
 * mframes;</tt> or fully qualified names beginning with <tt>mframes::</tt> in
 * your code.
 */
namespace mframes {

/**
 * @brief TODOC:
 */
class tModule : protected tConfigurable
{
    public:
        //-------------------------------
        // Interface types
        //-------------------------------

        /**
         * @brief Simple wrapper for command line arguments
         *
         * Intended for easier using and passing of command line arguments.
         */
        class tCmdLine
        {
            public:
                /**
                 * @brief Constructor
                 *
                 * Non-parametrical constructor.
                 */
                inline tCmdLine ()
                { m_argc = 0; m_argv = NULL; }

                /**
                 * @brief Constructor
                 *
                 * @param argc
                 * Number of arguments on command line eg. number of stored
                 * string values in argv
                 *
                 * @param argv
                 * Pointer to array of string values of cmd-line
                 */
                inline tCmdLine (int argc, char* argv[])
                { m_argc = argc; m_argv = argv; }

                /**
                 * @brief Argument count retrieval method
                 *
                 * @return
                 * Number of arguments on command line
                 */
                inline int get_argc ()
                { return m_argc; }

                /**
                 * @brief Argument value retrieval method
                 *
                 * @param num
                 * Index of the value which should be returned, starting from
                 * zero.
                 *
                 * @return
                 * String value of the argument on the command line
                 */
                inline char* get_argv (int num)
                {
                    ASSERT (num >= 0 && num < m_argc);
                    return m_argv[num];
                }

                /**
                 * @brief Value array retrieval method
                 *
                 * @return
                 * Pointer to array of string values passed on command line
                 */
                inline char** get_argv ()
                {
                    ASSERT (m_argc >= 0);
                    return m_argv;
                }

            private:
                /// argument count
                int m_argc;

                /// argument values
                char** m_argv;
        };

        //-------------------------------
        // Interface
        //-------------------------------

        /**
         * TODOC:
         * @brief Main filter module loop
         *
         * Call this function to start filter's work. Module must be
         * previously successfully initialized by init().
         *
         * @return
         * Exit status of the module
         */
        virtual bool run () = 0;

        /**
         * TODOC:
         * @brief Filter module's cleanup
         *
         * Call this function to clean and release all resources. Should
         * be called after successful initialization by init().
         */
        virtual void cleanup ();

        /**
         * @brief Destructor
         */
        virtual ~tModule ();

    protected:
        /**
         * @brief Destructor
         */
        enum SigHandlingRet
        {
            SHR_BREAK_OK,
            SHR_BREAK_ERR,
            SHR_CONTINUE
        };

        //-------------------------------
        // Internal methods
        //-------------------------------

        /**
         * @brief Constructor
         */
        tModule (tCmdLine& cmdline, const char* cfg_filename);

        bool init ();

        /**
         * TODOC:
         * @brief Loads configuration of the module
         *
         * Configuration is read and prepared to the form suitable for
         * further use.
         */
        bool read_config (bool accept_cmdline = false);

        /**
         * TODOC:
         * @brief Reverse function for read config
         *
         * Every member allocated or initialized by read_config() should be
         * destroyed and reset.
         */
        void release_config ();

        /// Installs signal handlers
        virtual void install_sig_handlers ();

        /// Handles installed signals
        virtual SigHandlingRet handle_signals ();

        /// Handles signal SIGINT
        virtual SigHandlingRet on_sigint ();

        /// Handles signal SIGQUIT
        virtual SigHandlingRet on_sigquit ();

        /// Returns true if SIGINT has been catched
        inline bool sigint_received ()
        { return s_sigint_received; }

        /// Returns true if SIGQUIT has been catched
        inline bool sigquit_received ()
        { return s_sigquit_received; }

        //-------------------------------
        // Implementation of tConfigurable interface
        //-------------------------------

        /**
         * @brief TODOC:
         */
        virtual bool register_sections (lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void unregister_sections ();

        /**
         * @brief TODOC:
         */
        virtual bool process_config (lib::tMemPool& conf_mp, 
                lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void cleanup_config ();

        //-------------------------------
        // Owned components
        //-------------------------------

        /// Place where command line arguments are stored
        tCmdLine m_cmdline;

        /// Name of the log file to log to
        char* m_log_name;

    private:
        /// A trivial signal transfering function
        static void sig_handling_f (int signum);

        /// Rereads configuration - core function
        bool reread_configuration (lib::tMemPool& cf_pool,
            bool apply_cmdline);

        //-------------------------------
        // Object state
        //-------------------------------

        bool m_configured;

        //-------------------------------
        // Owned components
        //-------------------------------

        /// Memory pool which holds values parsed from configuration
        lib::tMemPool m_conf_mp;

        /// Description of "Common" section in configuration
        lib::cfitem* m_common_section;

        /// Flag to set when SIGINT has been catched
        volatile static bool s_sigint_received;

        /// Flag to set when SIGQUIT has been catched
        volatile static bool s_sigquit_received;
};

} //end of namespace mframes

#endif //ifdef ECHELOG_MODULE_H

