/**
 * @file
 * @brief Contains declaration of tMsgProcessingModule class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#ifndef ECHELOG_MSGPROCESSINGMODULE_H
#define ECHELOG_MSGPROCESSINGMODULE_H

#include "mframes/echelogmodule.h"
#include "mframes/unixsocketioframe.h"

namespace mframes {

/**
 * TODOC:
 * @brief Base class for echelog filters.
 *
 * This class encapsulates all basic echelog filters functionality.
 * Reads common configuration and establishes socket communication, handles
 * signals, reconfiguration, reads messages and passes them to the filter
 * specified in the initialization for further processing.
 *
 * Typical instance lifecycle is initialization by
 * init(), work performed by run() and cleanup by cleanup().
 *
 * This class behaves as a singleton eg. there is only one instance
 * of this class and it isn't possible create more of them. To obtain
 * a instance call static get_filter_module() method.
 */
class tMsgProcessingModule : public tEchelogModule
{
    public:
        //-------------------------------
        // Interface
        //-------------------------------

        /**
         * @brief Filter module's initialization
         *
         * Call this function to initialize whole filter module.
         *
         * @param cmdline
         * Command line of the module.
         *
         * @param cfg_filename
         * Name of the file which contains the configuration for the filter
         * module.
         *
         * @param filter
         * Created filter instance. This instance will be initialized by
         * module. Messages will be passed there.
         *
         * @return
         * On success returns false. On failure returns true.
         */
        bool init (tBufferTarget* target, bool enable_out = true);

        /**
         * @brief Filter module's cleanup
         *
         * Call this function to clean and release all resources. Should
         * be called after successful initialization by init().
         */
        void cleanup ();

        /**
         * @brief Constructor
         */
        tMsgProcessingModule (tCmdLine& cmdline, const char* cfg_filename);

    protected:
        //-------------------------------
        // Internal methods
        //-------------------------------

        /**
         * @brief TODOC:
         */
        virtual int process ();

    private:
        //-------------------------------
        // Object state attributes
        //-------------------------------

        /// Actual state of filter module
        bool m_inited;

        //-------------------------------
        // Owned components
        //-------------------------------

        /// IO frame
        tUNIXSocketIFrame* mp_ioframe;
};

} //end of namespace mframes

#endif //ifdef ECHELOG_MSGPROCESSINGMODULE_H

