/**
 * @file
 * @brief Contains declaration of tMsgTarget interface
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#ifndef ECHELOG_MSGTARGET_H
#define ECHELOG_MSGTARGET_H

#include "messages/msgbuf.h"
#include "mframes/configurable.h"

namespace mframes {

/**
 * TODOC:
 * @brief Basic interface of the filter
 *
 * This is the description of the filter interface only. No functionality
 * is covered by this class. All instantiatable filters should be derived
 * directly or indirectly from this interface. Filter should implement all
 * pure virtual methods in the meaning as they are described here.
 */
class tBufferTarget : public tConfigurable
{
    public:
        //-------------------------------
        // Interface
        //-------------------------------

        /**
         * TODOC:
         * @brief Initialization of the filter
         *
         * Filter should read its configuration and prepare itself for
         * use by pass_msg() method.
         *
         * @return
         * On success returns false. On failure returns true.
         */
        virtual bool init () = 0;

        /**
         * TODOC:
         * @brief Cleanup of the filter
         *
         * Reverse method for init(), it should release and reset all
         * the contents of the filter.
         */
        virtual void cleanup () = 0;

        /**
         * TODOC:
         * @brief Message processing function
         *
         * Message should be processed by the needs of the filter and then
         * the message or newly created or modified message should be passed
         * to output(s) via filters::tFilterFrame::send_message().
         *
         * @param message
         * Message to be processed. Must be prepared for reading.
         *
         * @return
         * On success returns false. On failure returns true.
         */
        virtual bool process_buffer (char* buf, int len) = 0;

        /**
         * @brief Constructor
         */
        inline tBufferTarget () {};
};

class tMsgTarget : public tBufferTarget
{
    public:
        //-------------------------------
        // Interface
        //-------------------------------

        virtual bool process_buffer (char* buf, int len);

        /**
         * TODOC:
         * @brief Message processing function
         *
         * Message should be processed by the needs of the filter and then
         * the message or newly created or modified message should be passed
         * to output(s) via filters::tFilterFrame::send_message().
         *
         * @param message
         * Message to be processed. Must be prepared for reading.
         *
         * @return
         * On success returns false. On failure returns true.
         */
        virtual bool process_msg (messages::tMsgBuf& message) = 0;

        /**
         * @brief Constructor
         */
        inline tMsgTarget () {};

    protected:
        //-------------------------------
        // Interface
        //-------------------------------

        /**
         * @brief TODOC:
         */
        inline int get_read_len ()
        { return m_read_len; }

        //-------------------------------
        // Owned components
        //-------------------------------

        int m_read_len;
};



} //end of namespace mframes

#endif //ifdef ECHELOG_MSGTARGET_H

