/**
 * @file
 * @brief Contains declaration of tReconfigurableModule class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#ifndef ECHELOG_RECONFIGURABLEMODULE_H
#define ECHELOG_RECONFIGURABLEMODULE_H

#include "mframes/ioframe.h"
#include "mframes/module.h"

namespace mframes {

/**
 * TODOC:
 * @brief Base class for common echelog modules.
 *
 * This class encapsulates all basic echelog filters functionality.
 * Reads common configuration and establishes socket communication, handles
 * signals, reconfiguration, reads messages and passes them to the filter
 * specified in the initialization for further processing.
 *
 * Typical instance lifecycle is initialization by
 * init(), work performed by run() and cleanup by cleanup().
 */
class tReconfigurableModule : public tModule
{
    public:
        //-------------------------------
        // Interface
        //-------------------------------

        /**
         * TODOC:
         * @brief Filter module's cleanup
         *
         * Call this function to clean and release all resources. Should
         * be called after successful initialization by init().
         */
        virtual void cleanup ();

        virtual bool run ();

    protected:
        //-------------------------------
        // Internal types
        //-------------------------------

        /**
         * TODOC:
         */
        enum ProcessRet
        {
            PR_BREAK_OK,
            PR_BREAK_ERR,
            PR_CONTINUE
        };

        /**
         * TODOC:
         * @brief Filter module's initialization
         *
         * @return
         * On success returns false. On failure returns true.
         */
        bool init ();

        /**
         * @brief Constructor
         */
        tReconfigurableModule (tCmdLine& cmdline, const char* cfg_filename);

        /// TODOC: Initialization reconfiguration independent part
        virtual bool static_init ();

        /// TODOC: Initialization reconfiguration dependent part
        virtual bool dynamic_init ();

        /// TODOC: Reverse method for static_init()
        virtual void static_cleanup ();

        /// TODOC: Reverse method for dynamic_init()
        virtual void dynamic_cleanup ();

        /// Does the hard work
        virtual int process () = 0;

        /// Installs signal handlers
        virtual void install_sig_handlers ();

        /// Handles installed signals
        virtual SigHandlingRet handle_signals ();

        /// Handles signal SIGHUP
        virtual SigHandlingRet on_sighup ();

        /// Returns true if SIGHUP has been catched
        inline bool sighup_received ()
        { return s_sighup_received; }

    private:
        //-------------------------------
        // Internal types
        //-------------------------------

        /// TODOC: Enum for filter module's state identification
        enum ReconfModuleState
        {
            RM_CREATED,
            RM_STAT_INITIALIZED,
            RM_DYN_INITIALIZED
        };

        //-------------------------------
        // Internal methods
        //-------------------------------

        /**
         * TODOC:
         * @brief Reconfiguration method
         *
         * Called from SIGHUP handler to perform all operations needed
         * by module reconfiguration. Module is cleaned - dynamic part.
         * All the configuration items are freed and reset and loaded again.
         * Then module is dynamicly initalized again.
         */
        bool reinit ();

        /// A trivial signal transfering function
        static void sig_handling_f (int signum);

        //-------------------------------
        // Object state attributes
        //-------------------------------

        /// TODOC:
        ReconfModuleState m_rmstate;

        /// Flag to set when SIGHUP has been catched
        volatile static bool s_sighup_received;
};

} //end of namespace mframes

#endif //ifdef ECHELOG_RECONFIGURABLEMODULE_H

