/**
 * @file
 * @brief Control command layer
 *
 * The control component provides an interface to send and receive short
 * commands between echelog modules.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef _ECHELOG_CONTROL_H
#define _ECHELOG_CONTROL_H

#include "socks/ctrl_cmd.h"

/**
 * @brief Control commands internal protocol version 
 * (see private_socks::tCmdHeader::protocol_version)
 */
#define ECHELOG_CTRL_PROTOCOL_VERSION 1


namespace socks {


/**
 * @brief Primitive for sending and receiving control commands. 
 *
 * The control layer is mostly used for communication with the main echelog
 * deamon to interchange short local impulses, commands or notifications.
 * Though it would be possible to use standard echelog messaging it wouldn't be
 * neither effective nor comfortable to do so. %tControl provides much simple
 * interface for this purpose.
 *
 * Currently the interface is implemented in extended versions of echelog
 * unix domain sockets @link socks::tCtrlUNIXSocket tCtrlUNIXSocket@endlink and
 * @link socks::tCtrlUNIXSocketNB tCtrlUNIXSocketNB@endlink.
 */
class tControl {

    protected:

        /// <em>Constant property:</em> Name of control channel
        CONST_PROPERTY (char *, ctrl_name);
        /// <em>Read-only property:</em> Length of channel name
        SIMPLE_R_PROPERTY (long, ctrl_namelen);
        /// Retry counter
        int retry_count;

        /// Buffer used to receive commands
        void * cmd_in_buffer;
        /// Buffer used to send commands
        void * cmd_out_buffer;

        /** 
         * @brief Converts the buffer contents to a tCommand structure.
         *
         * The function reads the command frame buffer and creates a new
         * @link socks::tCommand tCommand@endlink holding a restructured copy
         * of received data.
         *
         * @return
         * The function returns a received command structure.
         */
        tCommand *buf2cmd (void *buffer) const;

    public:
    
        /// Constructor---creates and empties control structure.
        tControl ();
        /// Destructor---releases associated buffers.
        virtual ~tControl ();

        /**
         * @brief Stores a command into the control output buffer.
         *
         * The function creates a structure holding a command
         * with ancillary data passed in <tt>ac_buf</tt> of length
         * <tt>ac_buflen</tt>
         * and stores it into an internal buffer.
         */
        virtual void store_cmd (const tCMD_TYPE command, 
                const void * ac_buf = NULL, long ac_buflen = 0);
        
        /**
         * @brief Stores command into control output buffer.
         *
         * The function copies the passed command structure <tt>c</tt> into
         * an internal buffer.
         */
        virtual void store_cmd (const tCommand *c);

        /**
         * @brief Retrieves command from control input buffer.
         *
         * The function retrieves a command structure from an internal
         * in-buffer and tests it for validity.
         *
         * @return
         * On success, a pointer to the received command is returned.
         * If the data can't be retrieved or if the command is corrupted,
         * null is returned.
         */
        virtual tCommand *get_cmd ();

        /**
         * @brief Checks command in input buffer for correctness.
         *
         * The function checks the contents of a received command and tests it
         * for validity.
         *
         * @return
         * If the command is well-structured the function return zero.
         * Otherwise, <tt>-1</tt> is returned.
         */
        inline int test_received_cmd () const;

        /**
         * @brief Checks command in input buffer for correctness and type.
         *
         * The function checks the contents and type of a received command and 
         * tests it for validity.
         *
         * @return
         * The function returns zero if the buffer is well-structured and has
         * the specified type. Otherwise, <tt>-1</tt> is returned.
         * 
         */
        inline int test_received_cmdtype (tCMD_TYPE command_type) const;

        /**
         * @brief Checks if retry is allowed.
         *
         * A simple retry counter is provided by the interface. This function
         * is mainly intended for internal use in default handlers.
         *
         * @return
         * The function tests the retry counter and returns zero if an
         * operation can be retried. Otherwise, non-zero number is returned.
         */
        inline int test_retries ();

        /**
         * @brief Sets up control structure defaults.
         *
         * The function should set the connection name, default handlers and
         * parameters. Actual class implementing @link
         * socks::tSocketNB::tSocket tSocket@endlink interface should redefine
         * it as neccessary.
         */
        virtual void setup_defaults (const char *name) = 0;

        /** 
         * @brief Sends command structure.
         *
         * This command should initiate the command transfer. Note that
         * the <tt>tControl</tt> has to contain all neccessary data supplied by
         * <tt>store_cmd()</tt>.
         *
         * @return
         * The function returns number of bytes actually transfered. 
         * On error, <tt>-1</tt> is returned.
         */
        virtual int send_cmd () = 0;

        /**
         * @brief Receives command structure.
         *
         * This command should initiate the command receival. The 
         * received data can be actually retrieved by <tt>get_cmd()</tt>.
         *
         * @return
         * The function returns number of bytes actually read. 
         * On error, <tt>-1</tt> is returned.
         */
        virtual int receive_cmd () = 0;

};


}

/* We actually define all inline functions in a separate file */
#include "socks/control.hxx"


#endif
