/**
 * @file
 * @brief Control commands supplementary functions
 *
 * This file contains structures and functions for tCommand manipulation.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef _ECHELOG_CTRL_CMD_H
#define _ECHELOG_CTRL_CMD_H


namespace socks {


/** 
 * @brief This enumeration covers all known echelog command types.
 *
 * This is the complete listing of echelog command types. The commands
 * can be divided into two main groups: privileged (denoted by (p) in the
 * listing) and non-privileged commands. Privileged commands should be accepted
 * only from a module running under echelog administrator's PID.
 *
 * Additionaly, commands can possess some flags. The commands which are
 * expected to be confirmed by sending of other commands are denoted by (r).
 * The commands marked as system (denoted by (s)) have a special meaning for
 * the echelog deamon and, under normal circumstances, we shouldn't need to use
 * any of them.
 */
typedef enum {

/* Standalone commands */
    CMD_NOP = 0,                        ///< No operation
    CMD_PONG = 0x00000001,              ///< Pong command---response to ping
    CMD_LOGIN_OK= 0x00000002,           ///< Successful login notification

/* Standalone commands that need a confirmation */
    CMD_IDENTIFY = 0x00000F01,          ///< Authentication request (r)
    CMD_PING = 0x00000F02,              ///< Ping command (r)


/* Privileged commands */    
    CMD_START = 0xFF000001,             ///< Standalone module start (p)
    CMD_STOP = 0xFF000002,              ///< Standalone module stop (p)
    CMD_RECONFIGURE = 0xFF000003,       ///< Standalone module reconfigure (p)
    CMD_TOGGLE = 0xFF000004,            ///< Standalone flowswitch toggle (p)

/* Privileged system commands */    
    CMD_SYS_HALT = 0xFFF00000,          ///< System halt command (ps)
    CMD_SYS_START = 0xFFF00001,         ///< System start command (ps)
    CMD_SYS_STOP = 0xFFF00002,          ///< System stop command (ps)
    CMD_SYS_RECONFIGURE = 0xFFF00003,   ///< System reconfigure command (ps)
    CMD_SYS_RESTART_ALL = 0xFFF00004,   ///< Restart all module command (ps)
    CMD_CHAIN_START = 0xFFF00005,       ///< Start all chained modules (ps)
    CMD_CHAIN_STOP = 0xFFF00006,        ///< Stop all chained modules (ps)
    CMD_CHAIN_RESTART = 0xFFF00007      ///< Restart all chained modules (ps)
} tCMD_TYPE;


/**
 * @brief Structure holding command information
 *
 * A command is described by its type and associated ancillary data.
 * The data length must be less than @link socks::MAX_CMD_DATA_LENGTH
 * MAX_CMD_DATA_LENGTH@endlink.
 * The total length including the command structure must not exceed
 * @link socks::MAX_CMD_LENGTH MAX_CMD_LENGTH@endlink.
 */
struct tCommand {
    tCMD_TYPE command;      ///< Command type
    long d_len;             ///< Ancillary command data length
    void *d_ptr;            ///< Pointer to ancillary command data
};


/**
 * @brief Tests whether the command type is privileged or not.
 *
 * The function checks the desired flag of the specified command type.
 *
 * @return
 * If the command type is privileged the function returns true, otherwise
 * false.
 */
inline bool cmd_is_privileged (const tCMD_TYPE type);

/** 
 * @brief Tests whether the command type is a system command type or not.
 *
 * The function checks the desired flag of the specified command type.
 *
 * @return
 * If the command is a system command the function returns true, otherwise
 * false.
 */
inline bool cmd_is_system (const tCMD_TYPE type);

/** 
 * @brief Tests whether the command need a confirmation reply.
 *
 * The function checks the desired flag of the specified command type.
 *
 * @return
 * If the command type needs a confirmation the function returns true,
 * otherwise false.
 */
inline bool cmd_needs_confirmation (const tCMD_TYPE type);

/** 
 * @brief Creates a simple command type (with no ancillary data)
 *
 * The function creates a new command of the given type with no ancillary data.
 * The command is newly allocated via <tt>xmalloc()</tt> call.
 *
 * @return
 * The function returns a pointer to the newly created command.
 */
inline tCommand *create_simple_command (const tCMD_TYPE command);

/**
 * @brief Releases space allocated for the command structure and data.
 *
 * This function releases memory allocated for the command structure
 * and associated ancillary buffer. The command is expected to be allocated
 * on the standard heap.
 */
inline void free_cmd (tCommand *c);


/**
 * @brief Namespace containing classes and structures which are private to
 * %socks library implementation.
 */
namespace private_socks {

    /**
     * @brief Command structure header format
     *
     * This structure describes an internal format of command buffers
     * transfered via unix domain sockets. See @link socks::tCtrlUNIXSocket
     * tCtrlUNIXSocket@endlink and @link socks::tCtrlUNIXSocketNB
     * tCtrlUNIXSocketNB@endlink.
     */
    struct tCmdHeader {
        int protocol_version;   ///< Command protocol version (see control.h)
        tCMD_TYPE command;      ///< Command type
        int ad_len;             ///< Ancillary data length
        char ad_buffer[0];      ///< Ancillary data frame
    };

}


/// Maximal size of a control command 
#define MAX_CMD_LENGTH 2048

/// Maximal size of a control command data frame
#define MAX_CMD_DATA_LENGTH \
    (MAX_CMD_LENGTH - sizeof (private_socks::tCmdHeader))

}

/* We actually define all functions in a separate file */
#include "socks/ctrl_cmd.hxx"


#endif
