/**
 * @file
 * @brief Blocking UNIX domain sockets with support for control commands
 *
 * This file extends an interface of blocking unix domain sockets with
 * a control command layer. <tt>tCtrlUNIXSocket</tt> covers all common
 * attributes, buffers etc. and supplies reasonable default handlers for the
 * control layer.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef _ECHELOG_CTRL_UNIXSOCKET_H
#define _ECHELOG_CTRL_UNIXSOCKET_H

#include "socks/control.h"
#include "socks/unixsocket.h"


namespace socks {


/** 
 * @brief Sending control commands over blocking unix domain sockets.
 *
 * This class supplies a general blocking unix domain socket functionality (see
 * @link socks::tUNIXSocket tUNIXSocket@endlink plus extra functions and
 * default handlers for the control layer.
 *
 * Generally, on the client's side we won't need to redefine the standard
 * behaviour. If this is the case then we must take only three steps:
 * - Create <tt>tCtrlUNIXSocket</tt> object.
 * - Setup the control layer defaults.
 * - Run the socket manager, it will take care of the rest.
 *
 * If we need to use the control layer more generally we usually have to
 * manage control commands. After the control socket is set a control command
 * (created by @link socks::create_simple_command()
 * create_simple_command()@endlink or manually) can be 
 * sent by calling:
 * - @link socks::tControl::store_cmd() store_cmd()@endlink,
 * - @link socks::tControl::send_cmd() send_cmd()@endlink.
 * Similary, we can receive a command by calling:
 * - @link socks::tControl::receive_cmd() receive_cmd@endlink,
 * - @link socks::tControl::get_cmd() receive_cmd@endlink.
 *
 * Control commands (see @link socks::tCommand tCommand@endlink) can be
 * manipulated using auxiliary functions defined in <tt>ctrl_cmd.h</tt>.
 */
class tCtrlUNIXSocket : public tControl, public tUNIXSocket {

    protected:

       /// Creates new blocking control unix domain socket
       virtual tUNIXSocket *create_new_socket()
            { return new tCtrlUNIXSocket; }

    public:

       /// Creates socket and allocates communication buffers.
       tCtrlUNIXSocket () : tControl (), tUNIXSocket () {}

       /**
        * @brief Registers echelog module at overseer.
        *
        * The function registers the socket at the specified socket manager
        * and sends an authentification request of module <tt>name</tt>
        * to echelogd and waits for a reply. The parameter
        * <tt>max_registration_time</tt> is ignored since reading and writing
        * operation above the socket can block.
        *
        * @return
        * On success, zero is returned. On error, <tt>-1</tt> is returned.
        */
       virtual int register_echelog_module (tSocketMgr &smanager, 
               const char *name = NULL, 
               long max_registration_time = DEFAULT_REGISTRATION_TIMEOUT);

       /**
        * @brief Sets up control structure and socket handler defaults.
        *
        * The function sets the control connection name, default handlers and
        * default socket parameters. The default error and hangup handlers
        * are set to a routine which closes the broken control socket
        * and unregisters it from the associated socket manager.
        */
       virtual void setup_defaults (const char *name);

       /** 
         * @brief Sends command structure to socket.
         *
         * The function transfers the control command stored in the 
         * in-buffer to another echelog module.
         *
         * @return
         * The number of bytes actually written is returned. On error, 
         * the function returns <tt>-1</tt>.
         */
       virtual int send_cmd ();

       /** 
         * @brief Sends command structure from socket.
         *
         * The function receives a control command from another echelog
         * module to the in-buffer.
         *
         * @return
         * The number of bytes actually read is returned. On error, the
         * function returns <tt>-1</tt>.
         */
       virtual int receive_cmd ();
};


}

#endif
