/**
 * @file
 * @brief Non-blocking UNIX domain sockets with support for control commands
 *
 * This file extends an interface of non-blocking unix domain sockets with
 * a control command layer. tCtrlUNIXSocketNB
 * covers all common attributes, buffers etc. and
 * supplies reasonable default handlers for the control layer.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef _ECHELOG_CTRL_UNIXSOCKETNB_H
#define _ECHELOG_CTRL_UNIXSOCKETNB_H

#include "socks/control.h"
#include "socks/unixsocket_nb.h"


namespace socks {


/** 
 * @brief Sending control commands over non-blocking unix domain sockets.
 *
 * This class supplies a general non-blocking unix domain socket functionality
 * (see @link socks::tUNIXSocket tUnixSocket@endlink plus extra functions and
 * default handlers for the control layer.
 *
 * Generally, on the client's side we won't need to redefine the standard
 * behaviour. If this is the case then we must take only three steps:
 * - Create @link socks::tCtrlUNIXSocket tCtrlUNIXSocket@endlink object.
 * - Setup the control layer defaults.
 * - Run the socket manager, it will take care of the rest.
 *
 * If we need to use the control layer more generally we usually have to
 * manage control commands. After the control socket is set a control command
 * (created by @link socks::create_simple_command()
 * create_simple_command()@endlink or manually) can be 
 * sent by calling:
 * - @link socks::tControl::store_cmd() store_cmd()@endlink,
 * - @link socks::tControl::send_cmd() send_cmd()@endlink.
 * Similary, we can receive a command by calling:
 * - @link socks::tControl::receive_cmd() receive_cmd()@endlink,
 * - @link socks::tControl::get_cmd() get_cmd()@endlink.
 *
 * Control commands (see @link socks::tCommand tCommand@endlink) can be
 * manipulated using auxiliary functions defined in <tt>ctrl_cmd.h</tt>.
 * 
 */
class tCtrlUNIXSocketNB : public tControl, public tUNIXSocketNB {

    protected:
        
        /// Creates new non-blocking control unix domain socket
        virtual tUNIXSocketNB *create_new_socket()
            { return new tCtrlUNIXSocketNB; }        

    public:
    
        /// Creates socket and allocates communication buffers.
        tCtrlUNIXSocketNB () : tControl (), tUNIXSocketNB () {}

        /**
        * @brief Registers echelog module at overseer.
        *
        * The function registers the socket at the specified socket manager
        * and sends an authentification request of module <tt>name</tt>
        * to echelogd and waits for a reply. If an answer doesn't come in
        * <tt>max_registration_time</tt> then the registration is considered
        * unsuccessful.
        *
        * @return
        * On success, zero is returned. If the operation timed out or if there
        * was an error, <tt>-1</tt> is returned.
        */
        virtual int register_echelog_module (tSocketMgr &smanager, 
                const char *name = NULL, 
                long max_registration_time = DEFAULT_REGISTRATION_TIMEOUT);

        /**
        * @brief Sets up a control structure and socket handler defaults.
        *
        * The function sets the control connection name, default handlers and
        * default socket parameters. The default error and hangup handlers
        * are set to a routine which closes the broken control socket
        * and unregisters it from the associated socket manager.
        */
        virtual void setup_defaults (const char *name);

        /** 
         * @brief Sends a command structure to the socket.
         *
         * This function should initiate a data write from the out-buffer to
         * the socket. The function returns immediately, the result of the
         * operation is either a success (@link
         * socks::tSocketNB::on_write_finished on_write_finished@endlink event
         * is raised upon completion) or a failure (@link
         * socks::tSocketNB::on_err on_err@endlink event is raised).
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual int send_cmd ();

        /** 
         * @brief Sends a command structure from the socket.
         *
         * This function should initiate a data read from socket. This function
         * returns immediately, the result of the operation is either a success
         * @link socks::tSocketNB::on_write_finished on_write_finished@endlink
         * event is raised upon completion) or a failure (@link
         * socks::tSocketNB::on_err on_err@endlink event is raised).
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual int receive_cmd ();

};


}

#endif
