/**
 * @file
 * @brief Socket operation states
 *
 * @author Ales Privetivy <ales@ucw.cz>
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_OPERATION_H_
#define ECHELOG_OPERATION_H_

namespace socks {

/**
 * @brief Last operation on the socket.
 *
 * This enumeration is for non-blocking sockets only. Since every 
 * operation could be processed in parts, it is necessary to remember, what
 * operation was in progress, to be able continue in it.
 * See also @link socks::tSocketNB::last_operation
 * tSocketNB::last_operation@endlink attribute.
 */
enum tLastOperation 
{
    /* Non-blocking operation states */
        OP_NO_OPERATION,                  ///< No operation is in progress.
        OP_IO,                            ///< I/O operation in progress.
        OP_HANDSHAKE,                     ///< Handshake in progress.

    /* SSL operation states */
        OP_SHUTDOWN                       ///< Shutdown in progress (only SSL).
};


/**
 * @brief Last io operation.
 *
 * This enumeration is for non-blocking sockets only. It
 * extends the information of enum @link socks::tLastOperation
 * tLastOperation@endlink for value
 * <tt>OP_IO</tt> giving more precise identification of the operation
 * performed. See also @link socks::tOpState::last_op tOpState::last_op@endlink.
 */
enum tLastIO {
        IO_OP_READING,                       ///< Reading is in progress.
        IO_OP_WRITING,                       ///< Writing is in progress.
        IO_OP_ATOM_READING,                  ///< Atomic reading in progress.
        IO_OP_ATOM_WRITING,                  ///< Atomic writing in progress.
        IO_OP_NO_OPERATION                   ///< No io operation in progress.
};

/**
 * @brief Atomic operation state.
 *
 * This enumeration is for non-blocking sockets only. When enum 
 * @link socks::tLastOperation tLastOperation@endlink is set to value
 * <tt>OP_IO</tt> and atomic operation is currently in progress, it determines
 * phase of atomic operation.
 * 
 * Atomic data transfer consists of three steps:
 *
 * - beginning synchro frame is being sent/received.
 * - data are being sent/received. 
 * - ending synchro frame is being sent/received.
 * 
 * Since atomic data transfer could processed in parts, each 
 * non-blocking socket derived from class @link socks::tSocketNB
 * tSocketNB@endlink has
 * attributes @link socks::tSocketNB::r_io tSocket::r_io@endlink and 
 * @link socks::tSocketNB::w_io tSocketNB::w_io@endlink holding the details of
 * atomic data receiving/sending. Both these attributes are of
 * @link socks::tOpState tOpState@endlink type. The actual step of data
 * transfer is stored in @link socks::tOpState::atom_op_state
 * tOpState::atom_op_state@endlink.
 */
enum tAtomOpState
{
        AT_BEGINFRAME,        ///< read/write of beginning frame phase
        AT_DATA,              ///< read/write of data phase
        AT_ENDFRAME           ///< read/write of end frame phase
};

}

#endif

