/**
 * @file socket.h
 * @brief General socket class 
 *
 * This file defines an interface for the common abstract ancestor for all
 * %socks socket classes. @link socks::tSocket tSocket@endlink covers all
 * common attributes of sockets (i.e. file descriptor, state of socket, socket
 * role, socket specific data) and additional information for the socket
 * manager (see @link socks::tSocketMgr tSocketMgr@endlink).
 *
 * @author Ales Privetivy <ales@ucw.cz>
 */

#ifndef ECHELOG_SOCKET_H_
#define ECHELOG_SOCKET_H_

#include "lib/lib.h"
#include "property.h"

/**
 * @namespace socks
 * @brief Namespace containing all public structures and interfaces.
 */
namespace socks {

/**
 * @brief This enumeration holds actual socket state.
 *
 * This is an enumeration type for @link socks::tSocket::status
 * tSocket::status@endlink attribute. It is set by various calls and it should
 * be checked to better identify the result of the last operation on a socket.
 *
 * @note In general, negative values mean unrecoverable states. Socket should
 * be released then (i.e. call @link tSocket::close_socket()
 * close_socket()@endlink). 
 */
typedef enum {
    SS_UNDEFINED=100,           ///< Undefined state.

    SS_EMPTY=8,                 ///< Socket not created. 
    SS_CONNECTING=7,            ///< Connect in progress (for nonblocking IO).
    SS_ACCEPTING=6,             ///< Accept in progress (for nonblocking IO).
    SS_BROKEN=5,                ///< Connection broken.
    SS_DEFAULT=4,               ///< Correctly created socket.
    SS_BOUND=2,                 ///< Bound socket.
    SS_LISTENING=1,             ///< Listening on socket.
    SS_CONNECTED=0,             ///< Socket connected.

    SS_ERRBIND=-1,              ///< Cannot bind socket.
    SS_ERRLISTENING=-2,         ///< Socket is not listening.
    SS_ERRCONNECT=-3,           ///< Cannot connect.
    SS_ERRSOCK=-4,              ///< Invalid socket.
    SS_ERRRESOURCES=-5,         ///< Cannot gain/free socket resources.
    SS_ERROR=-100               ///< General error.
} tSOCK_STAT;

/**
 * @brief This enumeration distinguishes server and client socket side.
 *
 * This enumeration type is widely used in @link socks::tSocket tSocket@endlink
 * and its derivatives. See also @link socks::tSocket::role role@endlink
 * attribute.
 */
typedef enum {
    AS_SOCK_CLIENT=0,           ///< Client socket side
    AS_SOCK_SERVER=1            ///< Server socket side
} tSOCK_TYPE;

/**
 * @brief Bit indicating that atomic data were compressed (see struct 
 * @link socks::tAtomicDesc tAtomicDesc::flags@endlink).
 */
#define FL_ZLIB_COMPRESSED      0x1 

/**
 * @brief Synchronization frame used for atomic data transfers. 
 *
 * Atomic data are sent ``fortified'' by two frames. Both of them have this
 * structure. Sending routines fill in the magic number (see @link
 * socks::V_MAGIC V_MAGIC@endlink), the number of bytes of transfered data (for
 * compressed data it is the compressed length). The compression is signalized
 * by a bit in <tt>flags</tt> field.
 * 
 * @note Now, the compression is available for ssl sockets only.
 */
struct tAtomicDesc {
    /** @brief Magic number (see @link socks::V_MAGIC V_MAGIC@endlink).
     */
    int32_t magic;
    int32_t length;          ///< Length of data in frame.
    /** @brief Flags (see 
     * @link socks::FL_ZLIB_COMPRESSED FL_ZLIB_COMPRESSED@endlink).
     */
    int32_t flags;            
};

/**
 * @brief A magic number for at_read/at_write calls
 *
 * A magic number used to identify echelog atomical reads/writes.
 * @see struct @link socks::tAtomicDesc tAtomicDesc@endlink */
#define V_MAGIC 154905066L

/**
 * @brief Default sync framebuffer length
 *
 * The default sync framebuffer constant. It must be wide enough to hold struct
 * @link socks::tAtomicDesc tAtomicDesc@endlink.
 */
#define AT_FRAMEBUF_LEN (sizeof(struct tAtomicDesc))

class tSocket;
class tSocketMgr;

/**
 * @var tSocketHandlerFunc
 * @brief This is the type for event handlers (callbacks) for tSocket class.
 */
typedef void (*tSocketHandlerFunc)(class tSocket *); 

/**
 * @brief Abstract ancestor of all %socks socket classes.
 *
 * tSocket covers all common attributes of sockets (i.e. file descriptor, state
 * of socket, socket role, socket specific data) and additional information for
 * the socket manager (see @link socks::tSocketMgr tSocketMgr@endlink).
 *
 * After a socket is created and successfully connected, it can be used in a
 * straightforward way using only <tt>sock_read()</tt> and <tt>sock_write()</tt>
 * calls. In this approach we must take care of events occurring on sockets
 * ourselves (i.e. incoming data, clear to send data, incoming client
 * connection, error on socket, otherside hangup). This is usually done via
 * <tt>select()</tt> or <tt>poll()</tt> syscalls.
 *
 * The socket manager provides a more comfortable way to handle this. All what
 * we need to do is to register a new socket to the socket manager and to
 * supply callback routines for the five events mentioned above plus events
 * which occur when a socket is registered resp. unregistered to the socket
 * manager. It's not necessary to define all socket event handlers but at least
 * error and hangup callbacks should be defined to ensure proper program
 * functionality. Both hangup and error event handlers should always unregister
 * and release the relevant socket from the socket manager to avoid polling an
 * invalid or closed socket descriptor.
 *
 * Since the socket communication is client/server based, there is attribute
 * @link socks::tSocket::role role@endlink distinguishing whether socket is
 * server or client. Server side usually accepts connection requests and
 * creates new sockets for them. If we use socket manager, we do not need
 * to do anything. Whenever socket manager finds out that there is connection
 * request on the socket, it automatically calls server method 
 * <tt>accept_socket()</tt>. This method
 * creates new socket, registers it under the same socket manager, where the
 * request was caught. Event handlers (e.g. incoming data, possibility to write
 * data, error on socket, socket hangup etc.) for the new socket are supplied
 * by server socket. This means that server socket shares event handlers with
 * its child client sockets, and thus in such handler attribute @link
 * socks::tSocket::role role@endlink should be tested to determine if the event
 * occurred on server
 * or client type socket. This does not affect incoming data and possibility to
 * write data events, since they are never called for server type socket. The
 * newly created socket has set @link socks::tSocket::created_by_smgr
 * created_by_smgr@endlink flag, indicating
 * that this instance was created while accepting the socket by socket manager.
 * The new socket also holds internal socket identificator (@link
 * socks::tSocket::sid sid@endlink)
 * of its parent server socket in attribute @link socks::tSocket::psid
 * psid@endlink. This way we can
 * find out which socket was created by which server. If we do not use socket
 * manager, it is up to us to handle all connection requests ourselves. 
 *
 * Once is socket registered in a socket manager, it holds pointer to socket
 * manager instance in attribute @link socks::tSocket::smgr smgr@endlink. 
 * Null value of @link socks::tSocket::smgr smgr@endlink 
 * attribute means, that socket is not registered anywhere. For socket manager
 * purposes all registered sockets are given unique identifier which is held in
 * <tt>@ref smgr_handle</tt> attribute. Each registered socket has its own 
 * <tt>@ref event_mask</tt> attribute determining which events are to be
 * caught. For instances of client side it has its obvious meaning. For
 * instances of server side the event mask determines event mask of created
 * child client instances, events to be caught for server socket are not given
 * by this mask, since server is always registered in order to accept client
 * connections and handle errors.
 *
 * The socket state is reflected in attribute @link tSocket::status
 * status@endlink. Possible states are listed in enum @link socks::tSOCK_STAT
 * tSOCK_STAT@endlink. General rule is that negative values are always
 * unrecoverable error states (socket should be closed then). This attribute
 * is set by various methods operating on socket instance and should be checked
 * after each of them.  
 *
 * Sometimes it is useful to bind specific data with specific socket. There
 * is <tt>@ref app_specific_data</tt> attribute for this purpose. It is pointer
 * to data which should be allocated by <tt>malloc()</tt>, since they are
 * unallocated via <tt>free()</tt> in destructor (if the <tt>@ref 
 * app_specific_data</tt> attribute is not null). Be careful with this---it
 * can lead to segfaults.
 * 
 */
class tSocket 
{
        /**
         * @brief <em>Property:</em> This attribute holds a file descriptor
         * associated with the socket. If its value is negative then the
         * instance has no socket assigned.  
         */
        SIMPLE_PROPERTY(long, fd);

        /** @brief <em>Property:</em> It reflects the socket state. Possible
         * socket states are listed in enum @link socks::tSOCK_STAT
         * tSOCK_STAT@endlink. The general rule is that negative values are
         * always unrecoverable error states (socket should be closed then).
         * Almost every operation on socket can fail, this attribute can be
         * checked then to find out the reason of a failure.
         */
        SIMPLE_PROPERTY(tSOCK_STAT, status);

        /**
         * @brief <em>Property:</em> The type of a socket. The type of a socket
         * is either @link socks::AS_SOCK_CLIENT AS_SOCK_CLIENT@endlink or 
         * @link socks::AS_SOCK_SERVER AS_SOCK_SERVER@endlink indicating if the
         * socket represents a server or a client side of connection.
         */
        SIMPLE_PROPERTY(tSOCK_TYPE, role);

        /**
         * @brief <em>Property:</em> An attribute determining which events
         * should be caught. On the client side, it is a bitmask of events to
         * be caught ---namely <tt>POLLIN</tt> activated on new data arrival,
         * <tt>POLLPRI</tt> activated on new urgent data arrival and
         * <tt>POLLOUT</tt> activated when non-blocking write becomes
         * possible. On the server side the event mask is not actually used
         * but it determines an event mask for newly created child socket
         * instances (on incoming connections).
         */
        SIMPLE_PROPERTY(long, event_mask); 

        /**
         * @brief <em>External resource:</em> A pointer to the socket manager
         * where the socket instance is registered. This attribute is filled
         * via <tt>tSocketMgr::register_socket()</tt> call. It should be never
         * set directly. Note that socket manager object is typically shared
         * between sockets and it's internal state may change between socket
         * operations.
         */
        SIMPLE_EXT_RESOURCE (tSocketMgr *,smgr);
        
        /**
         * @brief <em>Property:</em> An internal identifier of the socket in
         * the socket manager in which the socket is registered. This value
         * should NOT be modified directly.
         */
        SIMPLE_PROPERTY(long, smgr_handle);

        /**
         * @brief <em>Property:</em> A flag indicating whether the socket
         * instance was created as a child client of the server socket.
         */
        SIMPLE_PROPERTY(bool, created_by_smgr);

        /**
         * @brief <em>Property:</em> Unique socket instance identifier of
         * socket.
         */
        SIMPLE_PROPERTY(long, sid);

        /**
         * @brief <em>Property:</em> Parent socket instance identifier.
         * This is relevant only if the socket was created as a child client of 
         * the server socket.
         */
        SIMPLE_PROPERTY(long, psid);

        /**
         * @brief <em>External resource:</em> A pointer to the socket specific
         * data. The data should be allocated by <tt>xmalloc()</tt> call
         * , since they are unallocated via <tt>xfree()</tt> in destructor (if
         * the <tt>@ref app_specific_data</tt> attribute is not null). Be
         * careful with this---it can lead to segfaults.
         */
        SIMPLE_EXT_RESOURCE(void *, app_specific_data);

        /**
         * @brief <em>Property:</em> A handler called when the socket is
         * registered to the socket manager. This handler is called at the end
         * of <tt>tSocketMgr::register_socket()</tt> call only for client side
         * sockets.
         */
        SIMPLE_PROPERTY(tSocketHandlerFunc, on_register);

        /**
         * @brief <em>Property:</em> A handler called when the socket is
         * unregistered from the socket manager. This handler is called at the
         * end of <tt>tSocketMgr::unregister_socket()</tt> call only for client
         * side sockets.
         */
        SIMPLE_PROPERTY(tSocketHandlerFunc, on_unregister);

        /**
         * @brief <em>Property:</em> A handler called when there are data on
         * the socket to be read. It is called only for client side sockets.
         */
        SIMPLE_PROPERTY(tSocketHandlerFunc, on_data_read);

        /** 
         * @brief <em>Property:</em> A handler called when it becomes possible
         * to write to the socket without blocking. It is called only for
         * client side sockets.
         */
        SIMPLE_PROPERTY(tSocketHandlerFunc, on_data_write);

        /** @brief <em>Property:</em> A handler called when otherside ended the
         * connection. This callback should unregister the socket to ensure
         * that the socket is released from the socket manager (otherwise the
         * socket manager would endlessly keep on calling error handler trying
         * to get rid of the closed socket).
         */
        SIMPLE_PROPERTY(tSocketHandlerFunc, on_data_hup);

        /**
         * @brief <em>Property:</em> A handler called when an error occurred.
         * This callback should unregister the socket to ensure that the socket
         * is released from the socket manager (otherwise the socket manager
         * would endlessly keep on calling error handler trying to get rid of
         * the broken socket). This callback can be also called from internal
         * methods to clear out the socket state if an error occured during
         * their execution.
         */
        SIMPLE_PROPERTY(tSocketHandlerFunc, on_data_err);

    protected:

        /**
         * @brief Creates new socket  
         *
         * This function creates a new socket via <tt>socket()</tt> syscall 
         * requesting a particular socket type. After this, the socket can be
         * set to the blocking/nonblocking state depending on purpose for which
         * the socket is created.
         *
         * @return Returns zero on success. If <tt>socket()</tt> syscall fails
         * or the socket state cannot be changed, it returns <tt>-1</tt>.
         */ 
        virtual int new_socket ()=0;

    public:
        /**
         * @brief Constructor for tSocket objects
         * 
         * This constructor ensures resetting of all tSocket class attributes.
         * Attributes of newly created <tt>tSocket</tt> instances are set this
         * way: @link tSocket::fd fd@endlink to <tt>-1</tt>, @link 
         * tSocket::status status@endlink to <tt>@ref SS_UNDEFINED</tt>,
         * @link tSocket::smgr smgr@endlink and all event handlers are set
         * to NULL.
         */
        tSocket() : fd (-1), status (SS_UNDEFINED), smgr(NULL), 
        smgr_handle(-1), created_by_smgr(false), sid (-1), psid (-1), 
        app_specific_data(NULL), on_register(NULL), on_unregister(NULL),
        on_data_read(NULL), on_data_write(NULL), on_data_hup(NULL),
        on_data_err(NULL) {};

        /**
         * @brief Destructor for tSocket objects
         *
         * This destructor does nothing in this class. In inherited classes
         * it should contain a cleanup of all resources used in the socket
         * instance. 
         */ 
        virtual ~tSocket() {};

        /**
         * @brief Closes socket  
         *
         * This is a pure method. In inherited classes this method should close 
         * the file descriptor of socket and reset all attributes to initial
         * state (see <tt>@ref tSocket()</tt> ).
         *
         * @note
         * It is quite possible that errors on a previous write
         * operation can be first  reported  at  the  final  close.  Not
         * checking the return value when closing the file may thus lead 
         * to silent loss of data.
         *
         * @return Returns zero on success. If <tt>free()</tt>
         * syscall or anything else fails, it returns <tt>-1</tt>.
         */
        virtual int close_socket()=0;

        /**
         * @brief Creates new socket for connection request and register it 
         * to the socket manager
         *
         * This is a pure method. In inherited classes this method 
         * should be called when connection request occurs 
         * on listening socket. It is called automatically if the involved
         * listening socket is registered in the active socket manager.
         * It consists of two steps: 
         * <ol>
         * <li>Create a new socket for an incoming request via <tt>accept()</tt>
         * syscall.</li>
         * <li>Register the new socket to the socket manager given by the
         * parameter <tt>sock_mgr</tt> with an event mask inherited from the
         * parent listening socket.</li>
         * </ol>
         *
         * @note This function may invoke <tt>err_handler()</tt> setting 
         * @link tSocket::status status@endlink accordingly.
         *
         * @return Returns zero on success---the socket was successfully
         * created and registered or proccesing was interrupted and
         * <tt>accept_socket()</tt> should be called again. If something fails,
         * it returns <tt>-1</tt> (check @link tSocket::status status@endlink
         * for a reason).
         */
        virtual int accept_socket(class tSocketMgr &sock_mgr)=0;

        /**
         * @brief Writes data to socket
         *
         * This is a pure method. In inherited classes this method 
         * should write <tt>len</tt> bytes of data stored in <tt>buffer</tt>
         * to the socket. The exact behavior of this depends on the particular 
         * socket class (i.e. if the operation is non/blocking etc.). This 
         * operation could write only portion of data---in this case we are 
         * informed how large portion was written by the return value.
         *
         * @note This function could invoke <tt>err_handler()</tt> setting 
         * @link tSocket::status status@endlink accordingly.
         *
         * @return A positive return value means the number of bytes really
         * written. On error, <tt>-1</tt> is returned.
         */
        virtual ssize_t sock_write(void *buffer, ssize_t len)=0;

        /**
         * @brief Read data from socket
         *
         * This is a pure method. In inherited classes this method should try to
         * read len <tt>bytes</tt> from the socketinto <tt>buffer</tt>.
         * The exact behavior of this depends on the particular socket class
         * (i.e. if the operation is non/blocking etc.). This operation
         * could read only portion of data---in this case we are informed
         * how large portion was read by the return value. 
         *
         * @note This function could invoke <tt>err_handler()</tt> setting 
         * @link tSocket::status status@endlink accordingly.
         *
         * @return A positive return value means the number of bytes really
         * read. On error, <tt>-1</tt> is returned.
         */
        virtual ssize_t sock_read(void *buffer, ssize_t len)=0;

        /**
         * @brief Read data from socket (additional options possible)
         *
         * This is a pure method. In inherited classes this method should try
         * to read len <tt>bytes</tt> from the socketinto <tt>buffer</tt>.
         * Difference from <tt>sock_read()</tt> is that developer can specify
         * additional options (look at <tt>recv()</tt> syscall for
         * <tt>flags</tt> field). The exact behavior of this depends on the
         * particular socket class (i.e. if the operation is non/blocking
         * etc.). This operation could read only portion of data---in this case
         * we are informed how large portion was read by the return value. 
         *
         * @note This function could invoke <tt>err_handler()</tt> setting 
         * @link tSocket::status status@endlink accordingly.
         *
         * @return A positive return value means the number of bytes really
         * read. On error, <tt>-1</tt> is returned.
         */
        virtual ssize_t sock_recv (void *buffer, ssize_t len, 
                unsigned int options)=0;

        /**
         * @brief Peek at incoming data (if possible) but don't block or retry
         *
         * This is a pure method. In inherited classes this method should try to
         * read len <tt>bytes</tt> from the socketinto <tt>buffer</tt> leaving 
         * the data in the buffer for the next read.
         * The exact behavior of this depends on the particular socket class
         * (i.e. if the operation is non/blocking etc.). This operation
         * could read only portion of data---in this case we are informed
         * how large portion was read by the return value. 
         *
         * @note This function could invoke <tt>err_handler()</tt> setting 
         * @link tSocket::status status@endlink accordingly.
         *
         * @return A positive return value means the number of bytes really
         * read. Zero denotes EOF or closed fd. On error, <tt>-errno</tt> value
         * is returned.
         */
        virtual ssize_t sock_poke (void *buffer, ssize_t len)=0;

        /**
         * @brief Atomic read of data from socket
         *
         * This is a pure method. In inherited classes this method should try to
         * read len <tt>bytes</tt> from the socketinto <tt>buffer</tt>.
         * In an atomic transfer, data are enclosed between two synchronization
         * frames. The start and the end frame have the same structure (i.e.
         * magic number and length of data---see struct <tt>@ref 
         * tAtomicDesc</tt>). If
         * <tt>len</tt> is less than the length of atomic data, nothing is read.
         * This function is available only for blocking sockets and therefore it
         * may block. 
         *
         * @return A positive return value means the number of bytes really
         * read. On error, -2 means that data does not fit in the supplied
         * buffer and <tt>-1</tt> means a general transfer error.
         */
        virtual ssize_t at_read(void *buffer, ssize_t len)=0;

        /**
         * @brief Atomic write of data to socket
         *
         * This is a pure method. In inherited classes this method should try to
         * write len <tt>bytes</tt> from <tt>buffer</tt> to socket.
         * In an atomic transfer, data are enclosed between two synchronization
         * frames. The start and the end frame have the same structure (i.e.
         * the magic number and the length of data---see struct <tt>@ref
         * tAtomicDesc</tt>). This function is available only for blocking
         * sockets and therefore it may block. 
         *
         * @return A positive return value means the number of bytes really
         * written. On error, the function returns <tt>-1</tt>.
         */
        virtual ssize_t at_write(void *buffer, ssize_t len)=0;

        /**
         * @brief Returns human readable identification of connection 
         * (e.g. ip:port, filename, ...)
         */
        virtual char * get_address()=0;
        
        /**
         * @brief Wrapper calling supplied hangup event handler
         */
        virtual void hup_handler()           
            { if (on_data_hup) on_data_hup(this); }

        /**
         * @brief Wrapper calling supplied error event handler
         */
        virtual void err_handler()          
            { if (on_data_err) on_data_err(this); }

        /**
         * @brief Wrapper calling supplied read possible event handler
         */
        virtual void read_handler()  
            { if (on_data_read) on_data_read(this); }
        

        /**
         * @brief Wrapper calling supplied write possible event handler
         */
        virtual void write_handler()        
            { if (on_data_write) on_data_write(this); }

};

}

#endif
