/**
 * @file
 * @brief Non-blocking sockets
 *
 * @author Ales Privetivy <ales@ucw.cz>
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_SOCKET_NB_H
#define ECHELOG_SOCKET_NB_H

#include "socks/operation.h"
#include "socks/socketmgr.h"


namespace socks {

/**
 * @var tSocketNBHandlerFunc
 * @brief This is the type for event handlers (callbacks) for tSocketNB class.
 */
typedef void (*tSocketNBHandlerFunc)(class tSocketNB *);

/**
 * @brief Structure of state information for atomic data transfers over
 * non-blocking sockets 
 *
 * An operation involving the data transfer over a non-blocking socket can be
 * processed in parts. To be able to continue it is necessary to somehow
 * remember a state of a socket when the operation stopped. This structure holds
 * all neccessary information for it. Attributes of this structure are purely
 * internal and should be modified or direcly used only by methods implemeting
 * atomic data transfers. 
 *
 * The only exception is the attribute @link socks::tOpState::last_op_bytes
 * last_op_bytes@endlink
 * which may be check to determine the amount of data processed during the
 * last operation (especially after @link socks::tSocketNB::at_read()
 * at_read()@endlink and
 * @link socks::tSocketNB::sock_read() sock_read()@endlink calls).
 */
struct tOpState {
    char *last_op_buf;               /**< Buffer where last operation data 
                                          are held. */
    char *last_op_buf_cur;           ///< Cursor in the buffer.
    long last_op_bytes;              ///< Size of data in buffer in bytes.
    enum tAtomOpState atom_op_state; ///< State of atom operation. 
    long already_processed;          ///< Number of bytes already processed.
    struct tAtomicDesc atom_frame;   ///< Frames for atomic operation. 
    char framebuf[AT_FRAMEBUF_LEN];  ///< Buffer for frames.
    enum tLastIO last_op;            ///< Type of the last operation.
};


/**
 * @brief Base class for all non-blocking sockets.
 *
 * Sometimes we need to use sockets in environment where we can't
 * block on a socket I/O operation or where it's objectionable to wait longer
 * time than it is indispensable. Echelog %socks library thus provide a simple
 * interface for non-blocking sockets which don't have these unpleasant
 * characteristics. 
 *
 * Handling non-blocking sockets is far more complex than their blocking
 * counterparts because practically any operation may be incomplete at the time
 * the socket manager returns the control to the application. We need to ensure
 * that all operations will continue from the point where they stopped. The
 * state information is stored in @link socks::tSocketNB::last_operation
 * last_operation@endlink, @link socks::tSocketNB::r_io r_io@endlink and 
 * @link socks::tSocketNB::w_io w_io@endlink attributes.
 *
 * Every non-blocking operation consists of three steps: 
 *
 * - <em>Scheduling the operation</em>---Function which initiates the operation
 *   is called, socket state structures are filled accordingly. Control is
 *   returned immediately from the function. - <em>Transparent processing by
 *   socket manager</em>---The operation is being processed by the socket
 *   manager using sockets' state information. For this reason inherited event
 *   handlers of @link socks::tSocket tSocket@endlink class are implemented
 *   directly in the library and we should avoid changing them (we really don't
 *   need to). 
 * - <em>Final notify</em>---Once the operation has been successfully finished,
 *   the appropriate handler is called. These notifiers are what we need to set
 *   in an application.
 * 
 * There is a common set of functions which all non-blocking sockets must
 * implement: 
 * - <em>Init functions for I/O operations</em>---Functions which can be used
 *   to schedule a start of a new socket I/O operation. Their basic usage is
 *   approximately same as the usage of non-blocking @link socks::tSocket
 *   tSocket@endlink
 *   functions <tt>sock_read()</tt>, <tt>at_read()</tt>, <tt>at_write()</tt>,
 *   etc. with the exception that all operations are actually done later in
 *   socket manager's main processing loop. Unlike their blocking counterparts,
 *   they should immediately return the control back to the application. 
 * - <em>The processing functions</em>---Functions which are called from the
 *   socket manager to handle sockets' states and continue processing of
 *   current operations (see <tt>connect_again()</tt>, <tt>accept_again()</tt>,
 *   <tt>read_again()</tt> etc.).
 * - <em>Event handlers</em>---Event handlers which are able to select proper
 *   processing functions accordingly to current operations. - <em>Final notify
 *   events</em>---A set of events exported to the application to allow actual
 *   data processing. See @link socks::tSocketNB::on_read_available
 *   on_read_available@endlink,
 *   @link socks::tSocketNB::on_read_finished on_read_finished@endlink,
 *   @link socks::tSocketNB::on_write_finished on_write_finished@endlink etc.).
 *
 *  The operations on sockets are:
 *  - <tt>reconnect()</tt>---schedules connecting.
 *  - <tt>sock_read()</tt>---schedules reading from socket.
 *  - <tt>sock_write()</tt>---schedules writing to socket.
 *  - <tt>at_read()</tt>---schedules atomic reading from socket.
 *  - <tt>at_write()</tt>---schedules atomic writing to socket.
 *
 *  The events notifying end of operation are:
 *  - @link socks::tSocketNB::on_connected on_connected@endlink---this event is
 *  raised after successful connecting.
 *  - @link socks::tSocketNB::on_read_finished on_read_finished@endlink---this
 *  event is raised after the successfully completed reading operation
 *  (operation started by <tt>sock_read()</tt> or <tt>at_read()</tt>). Check
 *  the @link socks::tOpState::last_op_bytes r_io::last_op_bytes@endlink
 *  attribute for the number of bytes actually read.
 *  - @link socks::tSocketNB::on_write_finished
 *  on_write_finished@endlink---this event is raised after the successfully
 *  completed writing operation (operation started by <tt>sock_write()</tt> or
 *  <tt>at_write()</tt>). Check the @link socks::tOpState::last_op_bytes
 *  w_io::last_op_bytes@endlink attribute for the number of bytes actually
 *  written (if we are using a compression this means the number of compressed
 *  bytes sent).
 *  - @link socks::tSocketNB::on_err on_err@endlink---this event is raised in
 *  the case that an operation failed.
 *
 *  @note Inherited event handlers @link socks::tSocketNB::on_data_read
 *  on_data_read@endlink, @link socks::tSocketNB::on_data_write
 *  on_data_write@endlink, @link socks::tSocketNB::on_data_err
 *  on_data_err@endlink and @link socks::tSocketNB::on_data_hup
 *  on_data_hup@endlink should not be modified.
 *
 *  Other events:
 *  - @link socks::tSocketNB::on_err on_err@endlink---this event is raised when
 *  an error occurs (either on the socket or during some operation).
 *  - @link socks::tSocketNB::on_hup on_hup@endlink---this event is raised when
 *  the other side ends the communication.
 *  - @link socks::tSocketNB::on_read_available
 *  on_read_available@endlink---this event is raised when the data arrive and
 *  they are not part of any running operation.
 *  - @link socks::tSocketNB::on_register on_register@endlink---this event is
 *  raised when the socket is registered to a socket manager.
 *  - @link socks::tSocketNB::on_unregister on_unregister@endlink---this event
 *  is raised when the socket is unregistered from a socket manager.
 *
 */
class tSocketNB : public tSocket 
{
    
    protected:

        /** @brief <em>Property:</em> Event handler raised when an error occurs
         * on the socket. */
        SIMPLE_PROPERTY( tSocketNBHandlerFunc, on_err );
        
        /** @brief <em>Property:</em> Event handler raised when the other side
         * ends the communication. */
        SIMPLE_PROPERTY( tSocketNBHandlerFunc, on_hup );

        /** @brief <em>Property:</em> Event handler raised when finished
         * connecting operation. */
        SIMPLE_PROPERTY( tSocketNBHandlerFunc, on_connected );

        /** @brief <em>Property:</em> Event handler raised when there are new
         * data on the socket. */
        SIMPLE_PROPERTY( tSocketNBHandlerFunc, on_read_available );
        
        /** @brief <em>Property:</em> Event handler raised when reading
         * operation was finished. */
        SIMPLE_PROPERTY( tSocketNBHandlerFunc, on_read_finished );
        
        /** @brief <em>Property:</em> Event handler raised when writing
         * operation was finished. */
        SIMPLE_PROPERTY( tSocketNBHandlerFunc, on_write_finished );
        
        /// <em>Property:</em> Type of the last operation. 
        SIMPLE_PROPERTY(tLastOperation, last_operation);

        /** @brief <em>Read-only property:</em> State information for reading
         * operations. */
        SIMPLE_R_PROPERTY( tOpState, r_io ); 

        /** @brief <em>Read-only property:</em> State information for writing
         * operations. */
        SIMPLE_R_PROPERTY( tOpState, w_io );
        
        /// Routine for connecting---continues from the actual state.
        virtual int  connect_again() = 0;             
        /// Routine for accepting connection---continues from the actual state.
        virtual int  accept_again() = 0;
        /// Routine for reading---continues from the actual state.
        virtual int  read_again() = 0;
        /// Routine for writing---continues from the actual state.
        virtual int  write_again() = 0;
        /// Routine for atomic reading---continues from the actual state.
        virtual int  atom_read_again() = 0;
        /// Routine for atomic writing---continues from the actual state.
        virtual int  atom_write_again() = 0;

        /// Set blocking socket behavior.
        virtual int set_blocking();

        /// Set non-blocking socket behavior.
        virtual int set_non_blocking();

    public:
       
        /// Create the socket and invalidate all its items.
        tSocketNB() : tSocket(), on_err (NULL), on_hup (NULL), 
                      on_connected (NULL), on_read_available (NULL),
                      on_read_finished (NULL), on_write_finished (NULL),
                      last_operation (OP_NO_OPERATION) {}
        
        /// Stop reading operation.
        inline void stop_r_io ();
        /// Stop writing operation.
        inline void stop_w_io ();
        /// Clear both io operations.
        inline void clear_io ();


        /// Handler for hangup. 
        virtual void hup_handler();
        /// Handler for error.
        virtual void err_handler();
        /// Handler for reading.
        virtual void read_handler();
        /// Handler for writing.
        virtual void write_handler();


        /**
         * @brief Real connection initiation.
         *
         * This function should initiate connecting to the other side. The
         * function returns immediately, the result of the operation is either
         * a success (@link socks::tSocketNB::on_connected on_connected@endlink
         * event is raised) or a failure (@link socks::tSocketNB::on_err
         * on_err@endlink event is raised).
         *
         * @return
         * The function returns zero on success, <tt>-1</tt> on error.
         */
        virtual int reconnect() = 0;

        /**
         * @brief Initiate a data write on socket.
         *
         * This function should initiate a data write on socket. This function
         * returns immediately, the result of the operation is either a success
         * (@link socks::tSocketNB::on_write_finished on_write_finished@endlink
         * event is raised) or a failure (@link socks::tSocketNB::on_err
         * on_err@endlink event is raised).
         *
         * @return
         * The function returns zero on success, <tt>-1</tt> on error.
         */
        virtual ssize_t sock_write(void *buffer, ssize_t len);


        /**
         * @brief Initiate a data read on socket.
         *
         * This function should initiate a data read from the socket. This
         * function returns immediately, the result of the operation is either
         * a success (@link socks::tSocketNB::on_read_finished
         * on_read_finished@endlink event is raised) or a failure (@link
         * socks::tSocketNB::on_err on_err@endlink event is raised).
         *
         * @return
         * The function returns zero on success, <tt>-1</tt> on error.
         */
        virtual ssize_t sock_read(void *buffer, ssize_t len);

        /**
         * @brief Initiate an atomic data read on socket.
         *
         * This function should initiate an atomic data read from the socket.
         * This function returns immediately, the result of the operation is
         * either a success (@link socks::tSocketNB::on_read_finished
         * on_read_finished@endlink event is raised) or a failure (@link
         * socks::tSocketNB::on_err on_err@endlink event is raised).
         *
         * @return
         * The function returns zero on success, <tt>-1</tt> on error.
         */
        virtual ssize_t at_read(void *buffer, ssize_t len);

        /**
         * @brief Initiate an atomic data write on socket.
         *
         * This function should initiate an atomic data write from the socket.
         * This function returns immediately, the result of the operation is
         * either a success (@link socks::tSocketNB::on_write_finished
         * on_write_finished@endlink event is raised) or a failure (@link
         * socks::tSocketNB::on_err on_err@endlink event is raised).
         *
         * @return
         * The function returns zero on success, <tt>-1</tt> on error.
         */
        virtual ssize_t at_write(void *buffer, ssize_t len);
};

}

/* We actually define all inline functions in a separate file */
#include "socks/socket_nb.hxx"


#endif

