/**
 * @file
 * @brief Socket manager 
 * 
 *
 * A socket manager provides a primitive for polling of events on
 * various sockets.
 *      
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#ifndef ECHELOG_SOCKETMGR_H_
#define ECHELOG_SOCKETMGR_H_

#include "socks/socket.h"
#include <sys/poll.h>


namespace socks {

/**
 * @brief Socket manager provides a primitive for polling of events on
 * various sockets.
 *
 * This primitive consists of a list of sockets which are to be monitored.
 * Sockets are added to this list via a registration method and removed via
 * an unregistration method. There are six events which can be raised for 
 * each socket:
 * - <em>registration of a socket to the socket manager</em>---We may want 
 *   to perform some actions when a socket is registered to the
 *   socket manager. This event allows us to hook a callback on it. This is
 *   necessary especially when the socket was registered to the socket manager
 *   as part of <tt>accept_socket()</tt> call. This event is raised only for
 *   non-listening (i.e. client) sockets.
 * - <em>unregistration of a socket from the socket manager</em>---We may
 *   want to perform some actions when a socket is unregistered 
 *   from the socket manager. This event allows us to hook a callback on it.
 *   Although na unregistration from the socket manager is never done
 *   automatically, it is more safe and we do not need to remember to take the
 *   post-unregister actions, if we use this callback. This event is raised
 *   only for non-listening sockets.
 * - <em>data available for reading</em>---This callback is called for sockets
 *   in ther client mode only. Listening sockets interprets available data 
 *   as connection requests and try to establish connections using
 *   <tt>accept_socket()</tt> call automatically. 
 * - <em>writing possible</em>---This callback is only for for sockets in the
 *   client mode only, since listening (i.e. server) sockets don't need
 *   writing capabilities. For a client it means that data can be written to 
 *   the socket. We must define this event handler very carefully. It is quite
 *   easy to end up in an endless loop if the data are not read from the socket
 *   properly.
 * - <em>otherside hangup</em>---When other side ends the communication, this
 *   event is raised. It's a really good idea to set this handler containing
 *   a cleanup code for all sockets, otherwise the socket manager will
 *   endlessly keep on calling the error handler in order to get rid of the
 *   closed socket descriptor.
 * - <em>error on socket</em>---Occasionally, an error can occur on the socket.
 *   We should be ready to handle this. All sockets should have this event
 *   handler, containing at least unregistering code, otherwise the socket
 *   manager will endlessly keep on calling the error handler, trying to get
 *   rid of the closed socket. Possible
 *   sources of an error can be also a connection refusal by the other side or
 *   a failure a higher layer (i.e. an error event can be raised also by calling
 *   <tt>err_handler()</tt>).
 *   
 * These events can be masked and unmasked via <tt>set_event_mask()</tt> calls. 
 * Events to be caught are held in @link socks::tSocket::event_mask
 * tSocket::event_mask@endlink attribute.
 *
 * @note This attribute for listening sockets determines an event_mask
 * for accepted connections created from this socket. Server sockets
 * always listen and accept incoming connnections.
 *
 * @note The number of sockets which can be registered in the socket manager at 
 * the same time is limited. Check out the <tt>#SERVING_MAXSOCKS</tt> constant
 * in echelog_config.h file.
 * 
 */
class tSocketMgr 
{
    protected:
        /// Number of sockets currently registered. 
        int sockcnt;
        /// Internal array for <tt>poll()</tt> syscall.
        struct pollfd ufds[SERVING_MAXSOCKS];
        /// Array of the registered sockets.
        class tSocket *socket[SERVING_MAXSOCKS];
        /// Bitmap of free socket identifiers.
        class lib::tCBitmap map;
        /// Polling index (last processed socket).
        long pidx;

    public:
        /// Constructor---creates empty socket manager.
        tSocketMgr() : sockcnt (0), map (SERVING_MAXSOCKS) { }
        /// Destructor---releases all sockets from socket manager.
        ~tSocketMgr();

        /**
         * @brief Registers socket with given event mask.
         *
         * Adds the socket to the socket manager's list. After that socket
         * receives unique socket id (sid), which loses when unregistered. If
         * the socket has been created from an accepted connection, parameter
         * <tt>parent</tt> should be passed with <tt>sid</tt> of the listening
         * socket. The socket manager tries to hold only sockets in following
         * states:
         * @link socks::tSOCK_STAT SS_CONNECTED@endlink,
         * @link socks::tSOCK_STAT  SS_CONNECTING@endlink, 
         * @link socks::tSOCK_STAT  SS_LISTENNING@endlink. If a socket is any
         * other state, the error event is raised during the next process()
         * call. For non-listening sockets a registration event is called 
         * at the end.
         *
         * @return
         * The function returns zero on success, otherwise <tt>-1</tt>.
         */
        int register_socket(class tSocket *s, int ev_mask = POLLIN | POLLPRI,
                long parent = -1);
        
        /**
         * @brief Unregisters socket.
         *
         * Removes the socket from the socket manager's list. For non-listening
         * sockets an unregistration event is called at the end.
         *
         * @return
         * The function returns zero on success, otherwise <tt>-1</tt>.
         */
        int unregister_socket (class tSocket *s);

        /**
         * @brief Poll on sockets.
         *
         * This is the main method of the socket manager. It polls the list of 
         * sockets for events. If an event occurs and is not masked and there 
         * is a supplied event handler, the event handler is called. A timeout
         * can be specified---if no event occur within a given time, the polling
         * ends. Timeout <tt>-1</tt> is considered as infinite.
         * 
         * @return
         * The function returns zero, when no error occurred. If the function 
         * was interrupted by some signal, it returns 1. If there was error
         * during polling, it returns <tt>-1</tt>.
         */
        int process (int timeout = -1);

        /**
         * @brief Sets new event mask for given socket.
         *
         * Parameter <tt>ev_mask</tt> gives the new mask for the socket
         * <tt>s</tt>. A bitmask may be a combination of following constants:
         * - <tt>POLLIN</tt>---Raise an event when data are available.
         * - <tt>POLLPRI</tt>---Raise an event when urgent data are available.
         * - <tt>POLLOUT</tt>---Raise an event when writing becomes possible.
         */
        void set_event_mask (class tSocket *s, int ev_mask);

        /**
         * @brief Unregisters and closes all registered sockets.
         *
         * This function unregisters all sockets at first and then it closes
         * all sockets managed by this socket manager. If a socket was created
         * by the socket manager as an accepted connection it is destructed.
         */
        void cleanup ();

        /**
         * @brief Unregisters and closes registered sockets which were created
         * by socket manager as accepted connection.
         *
         * This function unregisters all sockets at first and then it closes
         * all sockets managed by this socket manager. If a socket was created
         * by the socket manager as an accepted connection it is destructed.
         */
        void cleanup_created ();
              
};

}

#endif

