/**
 * @file
 * @brief Parameters of SSL layer
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#ifndef ECHELOG_SSLCONTEXT_H_
#define ECHELOG_SSLCONTEXT_H_

#include "lib/lib.h"

#include <openssl/ssl.h>

namespace socks {

/**
 * @brief Mechanism for setting parameters of SSL layer. 
 *
 * A communication over SSL dependends on many parameters. Thus, we want to
 * separate settings of communication details from the actual communication.
 * This class provides a simple interface for it.
 * 
 * The side which wants to use the SSL communication needs to set these
 * parameters:
 * - signed certificate 
 * - self signed root certificate
 * - its own private key
 * - Diffie Hellmann parameters (for server side only)
 *
 * Note that the SSL context must be set prior to any SSL socket manipulation.
 * For a client side it suffices to create tSSLContext object
 * and to set the first three parameters only (using <tt>set_key_files()</tt>).
 * For a server side all parameters must be supplied (using
 * set_key_files_with_dh).
 *
 * Rules for creating private keys and signing certificates for our purposes
 * are described in 
 * <A href="http://www.tldp.org/HOWTO/SSL-Certificates-HOWTO">SSL Certificates
 * HOWTO</a>.
 */
class tSSLContext  
{
    
    private:
        /**
         * @brief Internal SSL structure holding communication parameters.
         */
        SSL_CTX *context;
      
    public:

        /// Creates new SSL context (version 3).
        tSSLContext();
            
        /// Frees the object's SSL context.
        ~tSSLContext();

        /**
         * @brief Sets the client side SSL parameters.
         *
         * This function assigns a private key, a certificate and a self 
         * signed root certificate to the SSLContext instance. The private
         * key will be loaded from the file with the <tt>pk_file</tt> 
         * filename. This file must be in the PEM format. The same holds 
         * for the certificate (loaded from the file with the 
         * <tt>cert_file</tt> filename) and the self signed root certificate
         * (loaded from the file with the <tt>CA_cert_file</tt> filename).
         *
         * @return
         * If any file is not present or if it is corrupted the function
         * returns <tt>-1</tt>. Otherwise, zero is returned. 
         */
        int set_key_files(const char *pk_file, const char *cert_file,
                          const char *CA_cert_file);
        
        /**
         * @brief Sets the server side SSL parameters.
         *
         * This function assigns a private key, a certificate, a self signed
         * root certificate nad a file with the Diffie Hellman parameters to
         * the SSLContext instance. The private key will be loaded from the
         * file with the <tt>pk_file</tt> filename. This file must be in the
         * PEM format. The same holds for the certificate (loaded from the
         * file with the <tt>cert_file</tt> filename), the self signed root
         * certificate (loaded from the file with the <tt>CA_cert_file</tt>
         * filename) and the Diffie Hellman parameters (loaded from the file
         * with the <tt>dh_file</tt> filename).
         *
         * @return
         * If any file is not present or if it is corrupted the function
         * returns <tt>-1</tt>. Otherwise, zero is returned. 
         */
        int set_key_files_with_dh(const char *pk_file, const char *cert_file,
                          const char *CA_cert_file, const char *dh_file);

        /// Returns pointer to the SSL_CTX structure.
        SSL_CTX *get_context() { return context; }

};

}

#endif

