/**
 * @file
 * @brief SSL TCP sockets
 *
 * @author Ales Privetivy <ales@ucw.cz>
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_SSLSOCKET_H_
#define ECHELOG_SSLSOCKET_H_

#include "socks/socket_nb.h"
#include "socks/sslcontext.h"

#include <netinet/in.h>
#include <openssl/ssl.h>


/// The length of the queue for listening socket
#define LISTENCNT 5

namespace socks {

/**
 * @brief This enumeration holds actual state of SSL layer.
 *
 * If TCP layer is connected, then the SSL layer could be activated. This enum 
 * holds all possible states of the SSL layer. For more information see 
 * @link socks::tSSLSocket::ssl_status tSSLSocket::ssl_status@endlink.
 */
enum tSSLStatus 
{
        SL_EMPTY,             ///< SSL layer is not activated.
        SL_CONNECTING,        ///< SSL layer connecting in progress.
        SL_CONNECTED,         ///< SSL layer connected.
        SL_SHUTDOWN_INIT,     ///< SSL layer shutdown in progress (init phase).
        SL_SHUTDOWN_FINAL     ///< SSL layer shutdown in progress (final phase).
};

/**
 * @var tSSLHandlerFunc
 * @brief This is the type of event handlers for tSSLSocket class.
 */
typedef void (*tSSLHandlerFunc)(class tSSLSocket *);

/**
 * @brief Communication over non-blocking TCP sockets with added openssl layer. 
 *
 * This is an implementation of non-blocking TCP sockets with added openssl
 * layer. This class needs a correctly initialized @link socks::tSSLContext
 * tSSLContext@endlink object to begin a SSL communication.
 *
 * To set up the client side we must take these steps:
 * - Create tSSLContext object and set correctly its parameters.
 * - Create tSSLSocket object and set its event handlers. We must set at least
 *   @link socks::tSocketNB::on_err on_err@endlink, @link
 *   socks::tSocketNB::on_hup handlers on_hup handlers@endlink. We can set up
 *   additional @link socks::tSocketNB::on_connected on_connected@endlink, @link
 *   socks::tSocketNB::on_read_finished on_read_finished@endlink, @link
 *   socks::tSocketNB::on_write_finished on_write_finished@endlink, @link
 *   socks::tSocketNB::on_read_available on_read_available@endlink, @link
 *   socks::tSSLSocket::on_ssl_shutdown_init on_ssl_shutdown_init@endlink,
 *   @link socks::tSSLSocket::on_ssl_shutdown_finished
 *   on_ssl_shutdown_finished@endlink, @link socks::tSocketNB::on_register
 *   on_register@endlink and @link socks::tSocketNB::on_unregister
 *   on_unregister@endlink handlers. For an explanation of their functionality
 *   see @link socks::tSocket tSocket@endlink and @link socks::tSocketNB
 *   tSocketNB@endlink reference. The @link
 *   socks::tSSLSocket::on_ssl_shutdown_init on_ssl_shutdown_init@endlink is
 *   called when the shutdown handshake is started, @link
 *   socks::tSSLSocket::on_ssl_shutdown_finished
 *   on_ssl_shutdown_finished@endlink is called when the shutdown handshake is
 *   done.
 * - Call <tt>create_client()</tt> with already created @link
 *   socks::tSSLContext tSSLContext@endlink object as a parameter.
 * - Initiate connecting to the other side calling <tt>reconnect()</tt>.
 * - Run the socket manager, it will take care of the rest.
 *
 * To set up the server side we must take these steps:
 * - Create tSSLContext object and set correctly its parameters.
 * - Create tSSLSocket object and set its event handlers. We must set at least
 *   @link socks::tSocketNB::on_err on_err@endlink, @link
 *   socks::tSocketNB::on_hup on_hup@endlink handlers. We can set up additional
 *   @link socks::tSocketNB::on_connected on_connected@endlink, @link
 *   socks::tSocketNB::on_read_finished on_read_finished@endlink, @link
 *   socks::tSocketNB::on_write_finished on_write_finished@endlink, @link
 *   socks::tSocketNB::on_read_available on_read_available@endlink, @link
 *   socks::tSSLSocket::on_ssl_shutdown_init on_ssl_shutdown_init@endlink,
 *   @link socks::tSSLSocket::on_ssl_shutdown_finished
 *   on_ssl_shutdown_finished@endlink, @link socks::tSocketNB::on_register
 *   on_register@endlink and @link socks::tSocketNB::on_unregister
 *   on_unregister@endlink handlers. For an explanation of their functionality
 *   see @link socks::tSocket tSocket@endlink and @link socks::tSocketNB
 *   tSocketNB@endlink reference. The @link
 *   socks::tSSLSocket::on_ssl_shutdown_init on_ssl_shutdown_init@endlink is
 *   called when the shutdown handshake is started, @link
 *   socks::tSSLSocket::on_ssl_shutdown_finished
 *   on_ssl_shutdown_finished@endlink is called when the shutdown handshake is
 *   done.
 * - Call <tt>create_server()</tt> with already created
 *   @link socks::tSSLContext tSSLContext@endlink object as a  parameter.
 * - Run the socket manager, it will take care of the rest.
 *
 * For an explanation how operations are processed see @link socks::tSocketNB
 * tSocketNB@endlink reference.
 *
 * There is one remarkable extension of @link socks::tSocketNB
 * tSocketNB@endlink class.  A SSL connection should end with the ssl shutdown
 * handshake. It means that the side which wants to end the communication sends
 * some data to the other and the other side must answer again with some data
 * (see the reference documentation for a bidirectional handshake). In our
 * implementation @link socks::tSSLSocket::on_ssl_shutdown_init
 * on_ssl_shutdown_init@endlink event is raised in the beginning of the
 * operation. This ensures that the the application is acknowledged that the
 * connection is about to be terminated and stops sending further data. The
 * @link socks::tSSLSocket::on_ssl_shutdown_finished
 * on_ssl_shutdown_finished@endlink event is called, when the handshake is
 * finished, telling the application that the socket can be closed and
 * released.
 * 
 */
class tSSLSocket : public tSocketNB
{
    private:
        /// IP address of other side.
        struct sockaddr_in addr;      
        /// Hostname of other side.
        char *hostname;                   
        /// Port of the otherside.
        int port;                

    protected:

        /// <em>Property:</em> State of SSL layer.
        SIMPLE_PROPERTY( tSSLStatus, ssl_status );    

        /** @brief <em>Property:</em> This is event handler activated when
         * shutdown handshake is started. */
        SIMPLE_PROPERTY( tSSLHandlerFunc, on_ssl_shutdown_init );
        /** @brief <em>Property:</em> This is event handler activated when
         * shutdown handshake is done. */
        SIMPLE_PROPERTY( tSSLHandlerFunc, on_ssl_shutdown_finished );

        /**
         * @brief <em>Property:</em> SSL context---parameters determining SSL behavior.
         */
        SIMPLE_PROPERTY( SSL_CTX *, context );   
        /// <em>Property:</em> SSL object---internal structure.
        SSL *ssl;
 
        /// Creation of new non-blocking TCP/IP socket. 
        virtual int new_socket ();

        /// Creation of new non-blocking TCP/IP socket.
        virtual tSSLSocket *create_new_socket()
            { return new tSSLSocket; }

        /// Internal routine for reading
        int safe_SSL_read (char *buffer, long bytes);
        /// Internal routine for writing
        int safe_SSL_write (char *buffer, long bytes);

        /** @brief Routine for connecting---see 
         * @link socks::tSocketNB tSocketNB@endlink for explanation.
         */
        virtual int  connect_again();             
        /** @brief Routine for accepting connection---see
         * @link socks::tSocketNB tSocketNB@endlink for
         * explanation. */
        virtual int  accept_again();
        /** @brief Routine for data reading---see 
         * @link socks::tSocketNB tSocketNB@endlink for explanation.
         */
        virtual int  read_again();
        /** @brief Routine for atomic data reading---see
         * @link socks::tSocketNB tSocketNB@endlink 
         * for explanation. */
        virtual int  atom_read_again();
        /** @brief Routine for data writing---see 
         * @link socks::tSocketNB tSocketNB@endlink for explanation.
         */
        virtual int  write_again();
        /** @brief Routine for atomic data writing---see
         * @link socks::tSocketNB tSocketNB@endlink 
         * for explanation. */
        virtual int  atom_write_again();

        /// Routine for shutdown handshake---continues from actual state. 
        virtual int  ssl_shutdown_again();


    public:
       
        /// Creates the socket structure and invalidates all its attributes.
        tSSLSocket();

        /// Closes the socket and the SSL-layer. 
        virtual ~tSSLSocket();

        /**
         * @brief Closes the socket.
         *
         * Closes the TCP/IP socket, frees the ssl layer object (if necessary),
         * and invalidates attributes (sets fd to <tt>-1</tt>).
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual int close_socket();

        /**
         * @brief Frees the ssl layer object
         *
         * This function frees the ssl layer object, sets it to NULL. 
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        int tSSLSocket::close_ssl();

        /**
         * @brief Creates the client side of a connection.
         *
         * This call prepares all structures for establishing a connection. The
         * hostname and the port of the server where the client should connect
         * to, correctly initialized @link socks::tSSLContext
         * tSSLContext@endlink object and the socket manager intended for
         * further usage must be supplied as parameters. 
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        int create_client( const char *ahostname, const int aport,
                           class tSSLContext *ctx, class tSocketMgr *smgr );
         
        /**
         * @brief Creates the server side of a connection.
         *
         * This call creates a listening server socket with the given hostname
         * and port (the hostname must be the same as the hostname of host
         * where the server runs). The @link socks::tSSLContext
         * tSSLContext@endlink object must be also passed, determining ssl
         * parameters for accepted connections. The fourth parameter is the
         * socket manager to which this socket should be registered.
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        int create_server( const char *ahostname, const int aport,
                           class tSSLContext *ctx, class tSocketMgr *smgr );


        /**
         * @brief Verifies the common name of the other side
         *
         * The other side must present a certificate, in order to pass through 
         * the connection phase. After a connection has been established
         * we can use this common name to check the supplied parameter
         * to ensure if we are really talking to whom we wanted to.
         *
         * @return
         * The function returns zero if common names match, <tt>-1</tt> otherwise.
         */
        int verify_common_name(const char *common_name);

        /**
         * @brief Accepts an incoming connection.
         *
         * This function is used by the socket manager. When socket manager
         * detects a connection request on a listening socket it call 
         * this routine to accept the connection. The routine accepts the
         * connection, creates a new socket, registers it to the socket manager
         * and initiates the ssl handshake.
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual int accept_socket(class tSocketMgr &);

        /**
         * @brief Initiate a connection to the other side.
         *
         * This function initiates connecting to the other side. The first
         * resolved ip-address of the host where the client is connecting to
         * is used.
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual int reconnect();

        /**
         * @brief Initiate a ssl shutdown handshake.
         *
         * This function should initiate a ssl shutdown handshake. The function
         * returns immediately, the result of the operation is either a success
         * (@link socks::tSSLSocket::on_ssl_shutdown_finished
         * on_ssl_shutdown_finished@endlink event is raised) or a failure
         * (@link socks::tSocketNB::on_err on_err @endlink event is
         * raised).
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        int shutdown_ssl ();

        /**
         * @brief Initiate a data write on socket.
         *
         * This function should initiate a data write on the socket. The
         * function returns immediately, the result of the operation is either
         * a success (@link socks::tSocketNB::on_write_finished
         * on_write_finished@endlink event is raised) or a failure (@link
         * socks::tSocketNB::on_err on_err@endlink event is
         * raised).
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual ssize_t sock_write(void *buffer, ssize_t len);

        /**
         * @brief Initiate a data read on socket.
         *
         * This function should initiate a data read from the socket. The
         * function returns immediately, the result of the operation is either
         * a success (@link socks::tSocketNB::on_read_finished
         * on_read_finished@endlink event is raised) or a failure (@link
         * socks::tSocketNB::on_err on_err@endlink event is raised).
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual ssize_t sock_read(void *buffer, ssize_t len);
  
        /**
         * @brief Initiate an atomic data write on socket.
         *
         * This function should initiate an atomic data write to the socket.
         * The function returns immediately, the result of the operation is
         * either a success (@link socks::tSocketNB::on_read_finished
         * on_read_finished@endlink event is raised) or a failure (@link
         * socks::tSocketNB::on_err on_err@endlink event is raised).
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual ssize_t at_write(void *buffer, ssize_t len);

        /**
         * @brief Initiate an atomic data write on socket.
         *
         * This function should initiate an atomic data write to the socket.
         * Additional flags may be specified (at this time only compression).
         * The function returns immediately, the result of the operation is
         * either a success (@link socks::tSocketNB::on_read_finished
         * on_read_finished@endlink event is raised) or a failure (@link
         * socks::tSocketNB::on_err on_err@endlink event is raised).
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual ssize_t at_write(void *buffer, ssize_t len, long flags);

        /**
         * @brief Initiate an atomic data read from socket.
         *
         * This function should initiate an atomic data read from the socket.
         * The function returns immediately, the result of the operation is
         * either a success (@link socks::tSocketNB::on_write_finished
         * on_write_finished@endlink event is raised) or a failure (@link
         * socks::tSocketNB::on_err on_err@endlink event is raised).
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual ssize_t at_read(void *buffer, ssize_t len);

        /// This function is not implemented.
        virtual ssize_t sock_recv(void *buffer, ssize_t len, unsigned int
                options);

        /**
         * @brief Checks if there are some data to be read. 
         *
         * Parameters passed to function are not relevant, this function
         * only checks if there are any data waiting on the socket.
         *
         * @return
         * The function returns 1 if there are data available on the socket,
         * zero if the data have not yet arrived.
         */
        virtual ssize_t sock_poke(void *buffer, ssize_t len);

        /// Handler for hangup.
        virtual void hup_handler();
        /// Handler for error.
        virtual void err_handler();
        /// Handler for reading.
        virtual void read_handler();
        /// Handler for writing.
        virtual void write_handler();

        /// Returns the hostname of the other side.
        virtual char * get_address() 
            { return hostname; } 

};

}

#endif

