/**
 * @file
 * @brief Blocking UNIX domain sockets
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#ifndef ECHELOG_UNIXSOCKET_H_
#define ECHELOG_UNIXSOCKET_H_

#include "socks/socketmgr.h"
#include "socks/unixsocket_auth.h"

#include <unistd.h>

namespace socks {

/**
 * @brief Communication over blocking unix domain sockets. 
 *
 * This is an implementation of blocking UNIX domain sockets with
 * authentication. This class is derived from @link socks::tSocket
 * tSocket@endlink class and is written according to its concepts.
 * 
 * To set up the client side we must take these steps:
 * - Create <tt>tUNIXSocket</tt> object.
 * - Create the client side and connect to the server via
 *   <tt>create_client()</tt> call.
 * 
 * To set up the server side we must take these steps:
 * - Create <tt>tUNIXSocket</tt> object.
 * - Create the listening unix domain socket via <tt>create_server()</tt> call.
 * - Supply event handlers for accepted connections.
 * - Register the listening socket at the socket manager.
 *
 * This class also supports authentication. It is done automatically during
 * the connecting phase. 
 * 
 * See also @link socks::tSocket tSocket@endlink.
 * 
 */
class tUNIXSocket : public tSocket, public tUNIXSocketAuth
{
    
    protected:

        /// Creates new unix domain blocking socket.
        virtual int new_socket ();

        /// Sends authentication information.
        virtual int send_auth_info ();
        /// Receives authentication information.
        virtual int receive_auth_info ();
        /// Perform the authentication handshake.
        virtual int handshake ();

        /// Creates new unix domain blocking socket.
        virtual tUNIXSocket *create_new_socket()
            { return new tUNIXSocket; }
        
        /// <em>Read-only property:</em> Client authentication id.
        SIMPLE_R_PROPERTY (uid_t, auth_id);

    public:
        /// Creates socket structure and invalidates attributes.
        tUNIXSocket() : tSocket (), tUNIXSocketAuth () {};

        /// Closes socket, frees application specific data.
        virtual ~tUNIXSocket();

        /**
         * @brief Creates client side of connection.
         *
         * This function tries to connect to the unix domain socket of the given
         * name. Upon successful connecting, authentication handshake is
         * performed.
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        int create_client(const char *name);

        /**
         * @brief Creates server side of connection.
         *
         * This function tries to establish a listening unix domain socket 
         * of the given name. 
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        int create_server(const char *name);

        /// Closes the socket, for server removes the socket file.
        virtual int close_socket();
        
        /**
         * @brief Accepts incoming connection.
         *
         * This function is used by the socket manager. When the socket manager
         * detects an incoming connection on a listening socket it calls this
         * routine to accept the connection. The routine accepts the
         * connection, creates a new socket and registers it to the socket
         * manager. The authentication handshake is started afterwards.
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual int accept_socket(class tSocketMgr &smanager);

        /**
         * @brief Write of data.
         *
         * This function sends len bytes data from the buffer.
         *
         * @return 
         * Returns number of bytes sent on success (=<tt>len</tt>), <tt>-1</tt>
         * otherwise.
         */
        virtual ssize_t sock_write(void *buffer, ssize_t len);

        /**
         * @brief Read of data.
         *
         * This function receives len bytes data to the buffer.
         *
         * @return 
         * Returns number of bytes sent on success (=<tt>len</tt>), <tt>-1</tt>
         * otherwise.
         */
        virtual ssize_t sock_read(void *buffer, ssize_t len);

        /**
         * @brief Read of data with additional options.
         *
         * This function reads len bytes data to the buffer in a special way
         * described by the parameter <tt>options</tt> (see <tt>recv()</tt>
         * syscall manpage for possible values).
         *
         * @return 
         * Returns number of bytes sent on success (=<tt>len</tt>), <tt>-1</tt>
         * otherwise.
         */
        virtual ssize_t sock_recv(void *buffer, ssize_t len, 
                unsigned int options);

        /**
         * @brief Read of data with leaving the data for next reading.
         *
         * This function reads len bytes data to the buffer, but does not
         * remove the incoming data from the socket queue.
         *
         * @return 
         * Returns number of bytes sent on success (=<tt>len</tt>), <tt>-1</tt>
         * otherwise.
         */
        virtual ssize_t sock_poke(void *buffer, ssize_t len);
  
        /**
         * @brief Atomic read of data.
         *
         * This function receives atomic data to the buffer (i.e.
         * the data are enclosed within two frames which describe them).
         * If the data are larger than the size of the buffer given by
         * parameter len, no data are received (data are left on the socket for
         * a new receipt try) and <tt>-1</tt> is returned.
         *
         * @return 
         * Returns number of bytes sent on success (=<tt>len</tt>), <tt>-1</tt>
         * otherwise.
         */
        virtual ssize_t at_read(void *buffer, ssize_t len);

        /**
         * @brief Atomic write of data.
         *
         * This function sends len bytes data from the buffer atomically (i.e.
         * the data are enclosed within two frames which describe them).
         *
         * @return 
         * Returns number of bytes sent on success (=<tt>len</tt>), <tt>-1</tt>
         * otherwise.
         */
    	virtual ssize_t at_write(void *buffer, ssize_t len);

        /// Returns printable identificator of connection. 
        virtual char * get_address()
        { return get_addr(); }


};

}

#endif

