/**
 * @file
 * @brief Authentication for UNIX domain sockets 
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_UNIXSOCKET_AUTH_H_
#define ECHELOG_UNIXSOCKET_AUTH_H_

#include "socks/socket.h"
#include <sys/un.h>
#include <sys/socket.h>

namespace socks {


/// Length of the queue for listening socket.
#define LISTENCNT 5

/// Maximal length of filename for unix domain sockets.
#define MAX_SUN_PATH sizeof(((struct sockaddr_un *) NULL)->sun_path)

/// Length of sockaddr structure for unix domain sockets macro.
#ifndef SUN_LEN
#define SUN_LEN(su) \
        (sizeof(*(su)) - sizeof((su)->sun_path) + strlen((su)->sun_path))
#endif
    


/**
 * @brief Interface for basic unix-domain socket authentication
 *
 * If using unix domain sockets it is always good to know who is the ``man with
 * the hat'' on the other side. Moreover, we would like to to be sure that the
 * ``man with the hat'' will not lie about his identity (e.g. pid). Hopefully,
 * most modern systems support unix-domain socket authentication at the kernel
 * level. 
 *
 * Upon connection, each unix-domain socket performs a handshake and (as
 * as a part of it) sends its unix credentials via the socket to the other
 * side. Then the application can easily accept or refuse connections
 * accordingly to system process identificator associated with the remote
 * socket end. 
 *
 * @note This works only on Linux and BSD platforms.
 *
 */
class tUNIXSocketAuth
{
    
    protected:

        /// Name of the file to which socket is bound.
        struct sockaddr_un addr;

        /// Client authentication id.
        SIMPLE_R_PROPERTY (uid_t, auth_id);

    public:
	
	virtual ~tUNIXSocketAuth() {}
		
        /// Send authentication information.
        virtual int send_auth_info () = 0;
        /// Receive authentication information.
        virtual int receive_auth_info () = 0;

        /**
         * @brief Performs the authentication handshake
         *
         * This routine first sends authentication information and then 
         * receives authentication info from the other side.
         * 
         * @return
         * If the authentication handshake is not yet completed the function
         * returns <tt>0</tt>, and call should be repeated. Upon a successful
         * handshake it returns <tt>1</tt>. On error, <tt>-1</tt> is returned.
         */
        virtual int handshake () = 0;

        /// Set address where to connect.
        void set_addr(const char *name)
        { 
          memset (&addr, 0, sizeof (addr));
          addr.sun_family = AF_UNIX; 
          strncpy(addr.sun_path, name, MAX_SUN_PATH-1);
          addr.sun_path[MAX_SUN_PATH-1]='\0'; 
        }

        /// Gets address where to connect.
        char * get_addr()
            { return addr.sun_path; }

};

}

#endif

