/**
 * @file
 * @brief Non-blocking UNIX domain sockets
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_UNIXSOCKET_NB_H_
#define ECHELOG_UNIXSOCKET_NB_H_

#include "socks/socket_nb.h"
#include "socks/unixsocket_auth.h"

#include <unistd.h>
#include <string.h>


namespace socks {

/**
 * @brief This enumeration is for non-blocking sockets only. It holds the actual
 * state of authentication.
 *
 * This enum holds actual authentication state. See also 
 * @link socks::tUNIXSocketNB::auth_status tUNIXSocketNB::auth_status@endlink.
 *
 * The unix socket authentication consists of this parts:
 * - sending authentication credentials
 * - receiving of authentication credentials about the other side
 *
 * Each of these steps can be interrupted, since the socket is non-blocking. 
 */
typedef enum tAUTH_NB
{
  AU_NONE,                ///< Socket is not yet authenticated.
  AU_AUTHENTICATING,      ///< Authentication in progress---sending authdata.
  AU_AUTHENTICATION_SENT, ///< Authentication in progress---receiving authdata. 
  AU_AUTHENTICATED,       ///< Socket was correctly authenticated.
  AU_BROKEN               ///< Unknown authentication state.
};


/**
 * @brief Communication over non-blocking unix domain sockets. 
 *
 * This is an implementation of non-blocking unix domain sockets with 
 * authentication.
 *
 * To set up the client side we must take these steps:
 * - Create tUNIXSocketNB object and set its event handlers. We must set at
 *   least @link socks::tSocketNB::on_err on_err@endlink, @link
 *   socks::tSocketNB::on_hup on_hup@endlink handlers. We can set up additional
 *   @link socks::tSocketNB::on_connected on_connected@endlink, @link
 *   socks::tSocketNB::on_read_finished on_read_finished@endlink, @link
 *   socks::tSocketNB::on_write_finished on_write_finished@endlink, @link
 *   socks::tSocketNB::on_read_available on_read_available@endlink, @link
 *   socks::tSocketNB::on_register on_register@endlink and  @link
 *   socks::tSocketNB::on_unregister on_unregister@endlink handlers. For an
 *   explanation of their functionality see @link socks::tSocket
 *   tSocket@endlink and @link socks::tSocketNB tSocketNB@endlink reference.
 * - Call <tt>create_client()</tt>.
 * - Initiate connecting to other side calling <tt>reconnect()</tt>.
 * - Run the socket manager, it will take care of the rest.
 *
 * To set up the server side we must take these steps:
 * - Create tUNIXSocketNB object and set its event handlers. We must set at
 *   least @link socks::tSocketNB::on_err on_err@endlink, @link
 *   socks::tSocketNB::on_hup on_hup@endlink handlers. We can set up additional
 *   @link socks::tSocketNB::on_connected on_connected@endlink, @link
 *   socks::tSocketNB::on_read_finished on_read_finished@endlink, @link
 *   socks::tSocketNB::on_write_finished on_write_finished@endlink, @link
 *   socks::tSocketNB::on_read_available on_read_available@endlink, @link
 *   socks::tSocketNB::on_register on_register@endlink and @link
 *   socks::tSocketNB::on_unregister on_unregister@endlink handlers. For an
 *   explanation of their functionality  see @link socks::tSocket
 *   tSocket@endlink and @link socks::tSocketNB tSocketNB@endlink reference.
 * - Call <tt>create_server()</tt>.
 * - Run the socket manager, it will take care of the rest.
 *
 * For an explanation how are operations processed see @link socks::tSocketNB
 * tSocketNB@endlink reference.
 *
 */
class tUNIXSocketNB : public tSocketNB, public tUNIXSocketAuth
{
    
    protected:

        /** @brief Routine for connecting---see @link socks::tSocketNB
         * tSocketNB@endlink for explanation.
         */
        virtual int  connect_again();
        /** @brief Routine for accepting connection---see
         * @link socks::tSocketNB tSocketNB@endlink 
         * for explanation. */
        virtual int  accept_again();
        /** @brief Routine for data reading---see @link socks::tSocketNB
         * tSocketNB@endlink for explanation.
         */
        virtual int  read_again();
        /** @brief Routine for atomic data reading---see
         * @link socks::tSocketNB tSocketNB@endlink for 
         * explanation. */
        virtual int  atom_read_again();
        /** @brief Routine for data writing---see @link socks::tSocketNB
         * tSocketNB@endlink for explanation.
         */
        virtual int  write_again();
        /** @brief Routine for atomic data writing---see
         * @link socks::tSocketNB tSocketNB@endlink 
         * for explanation. */
        virtual int  atom_write_again();

        /// Send authentication information.
        virtual int send_auth_info ();
        /// Receive authentication information.
        virtual int receive_auth_info ();
        /// Perform the authentication handshake.
        virtual int handshake ();

        /// Creates new non-blocking unix domain socket. 
        virtual int new_socket ();

        /// Creates new non-blocking unix domain socket. 
        virtual tUNIXSocketNB *create_new_socket()
            { return new tUNIXSocketNB; }

        /// <em>Read-only property:</em> Client authentication progress status.
        SIMPLE_R_PROPERTY (tAUTH_NB, auth_status);

        /// <em>Read-only property:</em> Client authentication id.
        SIMPLE_R_PROPERTY (uid_t, auth_id);

    public:
       
        /// Creates socket structure and invalidates all attributes.
        tUNIXSocketNB();

        /// Closes socket and frees application data.
        virtual ~tUNIXSocketNB();

        /**
         * @brief Closes the socket.
         *
         * Closes the unix domain socket, frees the application data
         * and invalidates attributes (sets fd to <tt>-1</tt>).
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual int close_socket();

        /**
         * @brief Creates client side of connection.
         *
         * This call prepares all structures for needed to establish 
         * a connection. The unix socket pathname, where the client
         * should connect to, and the socket manager intended for future
         * usage must be supplied as parameters.
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        int create_client(const char *name, tSocketMgr &smgr);

        /**
         * @brief Creates server side of connection.
         *
         * This call creates a listening server socket with the given filename.
         * The second parameter is the socket manager, to which
         * this socket should be registered. 
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        int create_server(const char *name, tSocketMgr &smgr);

        /**
         * @brief Accepts incoming connection
         *
         * This function is for the socket manager. When the socket manager
         * detects an incoming connection on a listening socket it calls this
         * routine to accept the connection. This routine accepts the
         * connection, creates a new socket, registers it to the socket manager
         * and initiates the authentication handshake. 
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual int accept_socket(class tSocketMgr &);

        /**
         * @brief Initiate a connection to the other side.
         *
         * This function initiates connecting to other side. 
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual int reconnect();

        /**
         * @brief Initiate a data read from socket with additional parameters.
         *
         * This function should initiate a data read from the socket. This
         * function returns immediately, the result of the operation is either
         * a success (@link socks::tSocketNB::on_write_finished
         * on_write_finished@endlink event is raised upon completion) or
         * a failure (@link socks::tSocketNB::on_err on_err@endlink event is
         * raised).
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual ssize_t sock_recv(void *buffer, ssize_t len, unsigned int
                options);

        /**
         * @brief Initiate a data read from socket leaving the data on socket.
         *
         * This function should initiate a data read from the socket. This
         * function returns immediately, the result of the operation is either
         * a success (@link socks::tSocketNB::on_write_finished
         * on_write_finished@endlink event is raised upon completion) or
         * a failure (@link socks::tSocketNB::on_err on_err@endlink event is
         * raised).
         *
         * @return 
         * On success zero is returned, <tt>-1</tt> otherwise.
         */
        virtual ssize_t sock_poke(void *buffer, ssize_t len);

        /// Return connection id.
        virtual char * get_address() 
            { return get_addr (); } 

};

}

#endif

