/**
 * @file binstore.h
 * @brief Bin implementation 
 * @author Ales Privetivy <ales@ucw.cz>
 *
 * This file contains tBinStore class and structures needed for
 * storing of data items into the bin.
 */
 
#ifndef _ECHELOG_BINSTORE
#define _ECHELOG_BINSTORE

namespace spool {

/**
 * @var tBinIdentificator
 * @brief This is a type for identificators of bin items.
 *
 * Each item receives upon its insertion into bin an unique identificator of
 * <tt>spool::tBinIdentificator</tt> type. Zero identificator means an invalid
 * item, so valid item identificator is at least <tt>1</tt>. 
 */
typedef unsigned long tBinIdentificator;

/**
 * @var tBinItemOffset
 * @brief This is a type for item offsets into bin data file.
 *
 * Each item is located somewhere in the bin data file. Its position
 * should be stored in variables of tBinItemOffset type.
 */
typedef off_t tBinItemOffset;

/**
 * @brief This is a type describing the bin item data.
 *
 * This is a type describing the bin item data. A structure of this 
 * type is prepended before each item data when the item is inserted
 * into the bin.
 */
struct tBinItem {

        /**
         * @brief Magic number, which must be filled with certain constant.
         * It is tested to detect, whether the pagefile is corrupted.
         */
        unsigned long magic; 

        /**
         * @brief Item identificator - always greater than zero.
         */
        tBinIdentificator id;   
        
        /**
         * @brief Length of the item data.
         */
        int length;
};

/**
 * @class tBinStore
 * @brief This class implements the bin as was described in the introduction.
 * 
 * The bin consists of items placed one right after another like in a pagefile.
 * The difference from the pagefile is, that the bin has not limited its size.
 * Every bin item has an item description at its beginning followed by the item
 * data. The item description has this attributes:
 *
 * - <em>magic number</em> The magic number must be filled with a certain
 *   constant. It is tested to detect, whether the pagefile is corrupted.
 * - <em>identificator</em> The item identificator (valid values must be
 *   always greater than zero).
 * - <em>length</em> The length of the item data.
 *
 * For a quick access to items, an index file is bound to the bin file. 
 * The index file is a list of the item beginnings in the bin  file.
 * If we need to access an item, we look to the index list under the item
 * identificator. This gives us the offset into the bin file, where the
 * desired item starts. We can only append items to the end of the
 * bin, but no removing or other operations are allowed.
 *
 * The bin can be opened either in read-only or write-only mode, but not
 * in both of these modes.
 * 
 */
class tBinStore {

    protected:

        /** @brief <em>Property:</em> Access mode (<tt>0</tt> means cached
         * mode, <tt>1</tt> means writethrough).
         */
        SIMPLE_PROPERTY(int, write_through);
        
        /**
         * @brief <em>Property:</em> indicating that the bin is opened.
         */
        SIMPLE_PROPERTY(int, present);

        /**
         * @brief <em>Property:</em> Access mode 
         * (<tt>1</tt> = read/<tt>2</tt> = write).
         */
        SIMPLE_PROPERTY(int, access_mode);
        
        /**
         * @brief <em>Property:</em> Identificator of the last inserted item.
         */
        SIMPLE_PROPERTY(tBinIdentificator, last_id);
        
        /**
         * @brief <em>Property:</em> Offset in the bin file, where the next
         * item should be inserted.
         */
        SIMPLE_PROPERTY(tBinItemOffset, tail);

        /**
         * @brief <em>Property:</em> Filename of the bin data file.
         */
        SIMPLE_PROPERTY(char *, data_filename);

        /**
         * @brief <em>Property:</em> File descriptor of then bin data file.
         */
        SIMPLE_PROPERTY(int, data_fd);

        /**
         * @brief <em>Property:</em> Filename of the bin index file.
         */
        SIMPLE_PROPERTY(char *, index_filename);

        /**
         * @brief <em>Property:</em> File descriptor of then bin index file.
         */
        SIMPLE_PROPERTY(int, index_fd);

    public:
        /**
         * @brief Initializes the tBinStore structure 
         *
         * This constructor initializes the bin object, sets whether the 
         * access to bin is cached (<tt>wr_thr</tt> parameter is <tt>0</tt>) or 
         * write-through (<tt>wr_thr</tt> parameter is <tt>1</tt>).
         */
        tBinStore(int wr_thr);

        /** 
         * @brief Creates a new bin files and opens it for writing only.
         *
         * This function creates the bin files. The created bin file
         * has <tt>data_fname</tt> filename and and its index file has
         * <tt>index_fname</tt> filename. Returns zero on success,
         * otherwise <tt>-1</tt>.
         */
        int create(char *data_fname, char *index_fname);

        /** 
         * @brief Opens the bin for reading.
         *
         * This function opens the bin for reading. The opened bin
         * has <tt>data_fname</tt> filename and and its index file has
         * <tt>index_fname</tt> filename. Returns zero on success,
         * otherwise <tt>-1</tt>.
         */
        int open_r(char *data_fname, char *index_fname); 

        /**
         * @brief Opens the bin for writing.
         *
         * This function opens the bin for writing. The opened bin
         * has <tt>data_fname</tt> filename and and its index file has
         * <tt>index_fname</tt> filename. Returns zero on success,
         * otherwise <tt>-1</tt>.
         */
        int open_w(char *data_fname, char *index_fname);

        /** 
         * @brief Closes the bin.  
         *
         * Closes the bin file and the index file. Returns zero on success,
         * otherwise <tt>-1</tt>.         
         */
        int close();

        /** 
         * @brief Inserts item to the end of the bin. 
         *
         * This function creates a new item and stores it at the end of the 
         * bin. The new item is created from data from <tt>buffer</tt> of
         * size <tt>length</tt>. Returns zero on success, otherwise <tt>-1</tt>.
         *
         * This function works only with the bin opened for writing.
         * 
         * @return Zero on error, otherwise item id of currently inserted item.
         */
        tBinIdentificator add_item(void * buffer, int length);

        /** 
         * @brief Reads the data of requested item into the buffer.
         *
         * Reads the data of requested item with <tt>item_id</tt> identificator
         * into the buffer <tt>buffer</tt>. Only first <tt>buf_size</tt> bytes
         * will be stored into buffer. Returns zero on success, otherwise
         * <tt>-1</tt> and sets <tt>buf_size</tt> to the real data size of the
         * read item.
         *
         * This function works only with the bin opened for reading.
         */
        int get_item(tBinIdentificator item_id, void *buffer, int *buf_size);

        /**
         * @brief Returns the identificator of the last item in the bin. 
         *
         * Returns the identificator of the last item in the bin.
         * If there is no item in the bin, it returns <tt>0</tt>.
         */
        tBinIdentificator get_last_id();
            
};

} // namespace spool

#endif
