/**
 * @file pagefile.h
 * @brief Pagefile implementation.
 * 
 * This file contains tPagefile class and structures needed for
 * storing of data items into the pagefile.
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */
 
#ifndef _ECHELOG_PAGEFILE
#define _ECHELOG_PAGEFILE

#include "lib/lib.h"
#include <sys/mman.h>

/**
 * @namespace spool
 * @brief This namespace contains all classes and structures in Echelog Spool
 * Library implementation.
 */
namespace spool {

/**
 * @var tIdentificator
 * @brief This is a type for identificators of pagefile items.
 *
 * Each item receives upon its insertion an unique identificator of 
 * <tt>spool::tIdentificator</tt> type. Zero identificator means an invalid item,
 * so valid item identificator is at least <tt>1</tt>. There is no cycling -
 * after reaching the upper limit (given by the tIdentficator type), 
 * we cannot continue inserting. It must be ensured, that we will never
 * reach this limit (e.g. via renumbering in the case of the %spool). 
 */
typedef unsigned long tIdentificator;

/**
 * @var ITEMSTATUS
 * @brief This is a type in which the status of an item is stored.
 * 
 * This is a type in which the status of an item is stored.
 * Item can be either present (the item is in the pagefile and it was not 
 * extracted yet), extracted (the item is marked as extracted) or missing
 * (when the pagefile is empty, first attribute is set as missing).
 */       
enum ITEMSTATUS {
        MISSING=0,     ///< Requested item is not present. 
        PRESENT=1,     ///< Item is present and not extracted.
        EXTRACTED=2    ///< Item is present but is marked as extracted.
};

/**
 * @brief This is a type describing the pagefile item data.
 *
 * This is a type describing the pagefile item data. A structure of this type is
 * prepended before each item data when the item is inserted into the 
 * pagefile.
 */
struct tItem {
        unsigned long magic;    ///< Magic number, which must be filled with certain constant. It is tested to detect, whether the pagefile is corrupted. 
        tIdentificator id;      ///< Item identificator - always greater than zero. 
        int length;             ///< Length of the item data. 
        enum ITEMSTATUS status; ///< Status of the item. 
        char data[0];           ///< Data of the item. 
};

/**
 * @class tPageFile
 * @brief This class implements simple pagefile as was described in the 
 * introduction.
 *
 * Pagefile consists of items placed one right after another. Every item has 
 * added a description at its beginning, then the item data follows. The item
 * description has this attributes:
 *
 * - <em>magic number</em> Magic number, which must be filled with certain
 *   constant. It is tested to detect, whether the pagefile is corrupted.
 * - <em>identificator</em> Item identificator - always greater than zero.
 * - <em>length</em> Length of item data.
 * - <em>item status</em> Status of item. The item can be either present (the
 *   item is in the pagefile and it was not extracted yet), extracted (the item
 *   is marked as extracted) or missing (when the pagefile is empty, the first
 *   attribute is set as missing).
 *   
 * The pagefile is implemented as one file containing the items with their
 * descriptions. If we want work with the pagefile, it must be present (mapped)
 * in the memory.  If we do not need the pagefile for some time, we can 
 * unmap it from the memory via <tt>detach_pagefile()</tt> call. When we need it again,
 * we can map it again via <tt>attach_pagefile()</tt> call. For faster implementation 
 * of pagefile operations, upon opening of the pagefile a list of item offsets
 * is built by traversing the whole pagefile. This list is present in the memory
 * during the whole timelife of the pagefile instance (i.e. it is not affected
 * by * <tt>detach_pagefile()</tt> and <tt>attach_pagefile()</tt> calls).
 */
class tPageFile {

    private:
        /**
         * @brief <em>Property:</em> Access mode (<tt>0</tt> means cached mode,
         * <tt>1</tt> means writethrough).
         */
        SIMPLE_PROPERTY(int, write_through);      
        /**
         * @brief <em>Property:</em> Indicating if pagefile is mapped into
         * memory.
         */
        SIMPLE_PROPERTY(int, present);      
        /**
         * @brief <em>Property:</em> Indicating that pagefile is closed.
         */
        SIMPLE_PROPERTY(int, closed);    

        /**
         * @brief <em>Property:</em> Pointer to mapped file in memory. 
         */
        SIMPLE_PROPERTY(char *,pg_data);
        /** 
         * @brief <em>Property:</em> Pointer into mapped memory used
         * internally for traversing. 
         */
        SIMPLE_PROPERTY(char *,pg_cursor); 
        /**
         * @brief <em>Property:</em> Pointer to end of mapped memory. 
         */
        SIMPLE_PROPERTY(char *,pg_limit); 
        /**
         * @brief <em>Property:</em> List of item offsets. 
         */
        SIMPLE_PROPERTY(int  *,pg_index); 

        /**
         * @brief <em>Property:</em> Maximal page file size. 
         */
        SIMPLE_PROPERTY(int, pg_max_page_size);
        /**
         * @brief <em>Property:</em> Maximal number of items. 
         */
        SIMPLE_PROPERTY(int, pg_max_items);
        /**
         * @brief <em>Property:</em> Number of items in pagefile. 
         */
        SIMPLE_PROPERTY(int, pg_items);  
        /**
         * @brief <em>Property:</em> Number of non-extracted items. 
         */
        SIMPLE_PROPERTY(int, pg_nonextracted);  

        /**
         * @brief <em>Property:</em> Filename of the pagefile file. 
         */
        SIMPLE_PROPERTY(char *,filename);
        /** 
         * @brief <em>Property:</em> File descriptor of the opened pagefile
         * file. 
         */
        SIMPLE_PROPERTY(int, fd);              
        
        /**
         * @brief <em>Property:</em> Identificator value of the first item. 
         */
        SIMPLE_PROPERTY(tIdentificator, first_id);
        /**
         * @brief <em>Property:</em> Identificator value of the last item plus
         * one. 
         */
        SIMPLE_PROPERTY(tIdentificator, last_id);
        /** 
         * @brief <em>Property:</em> Identificator value of the first
         * non-extracted item. 
         */
        SIMPLE_PROPERTY(tIdentificator, first_nonextracted_id);

    private:
        /* forces synchro of mapped file */
        inline int synchro_pagefile();
       
    public:
        /**
         * @brief Initializes the pagefile structure.
         *
         * Initializes the pagefile structure.
         * Invalidates all pagefile attributes, and sets access mode
         * according to <tt>wr_thr</tt> parameter, maximal pagefile
         * size to <tt>max_page_size</tt> and maximal number of items
         * to <tt>max_items</tt>. It also allocates memory for the
         * item offset list.
         */
        tPageFile(int wr_thr, int max_page_size, int max_items);

        /**
         * @brief Creates a new pagefile and maps it into the memory.
         *
         * Creates new pagefile file, and maps it into memory. Resets
         * all pagefile attributes to make pagefile empty, blanks 
         * the whole pagefile with zeroes, and creates empty list
         * of item offsets. First item of the pagefile will have
         * <tt>first</tt> identificator value. Returns zero on success,
         * otherwise <tt>-1</tt>.
         */
        int create_pagefile(char *afilename, tIdentificator first);
        
        /**
         * @brief Loads an existing pagefile into the memory.
         *
         * Opens existing pagefile file, maps it into memory. Walks 
         * through it, looking for beginnings of items. This functions 
         * constructs list of item offsets and counts other pagefile
         * attributes. Then removes all items with <tt>item_id</tt>
         * lesser than <tt>first</tt>.
         * Returns zero on success, otherwise <tt>-1</tt>.
         */ 
        int assay_pagefile(char *afilename, tIdentificator first);

        /**
         * @brief Attaches the pagefile.
         * 
         * Maps pagefile into memory assuming that the other internal
         * structures are already filled in. 
         * Returns zero on success, otherwise <tt>-1</tt>.
         */
        int attach_pagefile();

        /**
         * @brief Detaches the pagefile.
         * 
         * Unmaps the pagefile, but leaves other internal structures in
         * the memory (i.e. array of item offsets item in the pagefile).
         * Returns zero on success, otherwise <tt>-1</tt>.
         */
        int detach_pagefile();

        /**
         * @brief Closes the pagefile.
         *
         * Unmaps the pagefile, saves it, and frees all internal structures.
         * Returns zero on success, otherwise <tt>-1</tt>.
         */
        int close_pagefile();

        /**
         * @brief Completely removes the pagefile.
         *
         * This works only for non-active pagefiles (i.e. pagefile should
         * not be mapped in to the memory - see <tt>attach_pagefile()</tt>). The
         * method removes the pagefile (deletes its file).
         * Returns zero on success, otherwise <tt>-1</tt>. 
         */
        int destroy_pagefile();

        /**
         * @brief Inserts the item into the pagefile.
         *
         * This function creates a new item and inserts it into the pagefile.
         * It must be ensured, that item will fit in pagefile (see
         * <tt>is_space()</tt>).
         * Returns the item_id, which was assigned to the newly inserted item,
         * otherwise zero (on error).
         */
        tIdentificator insert_item(char *buffer, int length);

        /**
         * @brief Marks an item as extracted.
         *
         * This method marks an item with <tt>item_id</tt> as extracted. If all
         * items in the whole segment are ``extracted'' and the pagefile is not 
         * closed, the pagefile is emptied and new items could be inserted.
         */
        void extract_item(tIdentificator item_id);

        /**
         * @brief Returns how much of free space is still in pagefile.
         *
         * Returns how much of free space is still in pagefile.
         */
        inline int get_free_space_size();

        /**
         * @brief Checks if there is enough space for insertion of an item 
         * with given length.
         *
         * This function returns <tt>1</tt>, if there is enough space for
         * insertion of an item with <tt>length</tt> size, otherwise
         * <tt>0</tt>. This function must be called every time before
         * <tt>insert_item()</tt>, since <tt>insert_item()</tt> does not checks
         * this.
         */
        inline int is_space(int length);

        /**
         * @brief Sets the status field of an item.
         *
         * This function changes the status field of an item with
         * <tt>item_id</tt> identificator to the desired value.
         */
        void set_status(tIdentificator item_id, enum ITEMSTATUS new_st);

        /**
         * @brief Returns the status field of an item.
         *
         * This function returns the status field of item with <tt>item_id</tt>
         * identificator. If item is not in pagefile
         * <tt>socks::ITEMSTATUS::MISSING</tt>         
         * is returned.
         */
        enum ITEMSTATUS get_status(tIdentificator item_id);

        /**
         * @brief Returns the pointer to the item data.
         *
         * This function returns pointer to item data (pointer to
         * beginning of data, the item description is skipped) for item
         * with the <tt>item_id</tt> identificator.
         * If the item is not present in the pagefile, returns NULL,
         * otherwise the pointer to the memory where the item data are held.
         */ 
        inline char *get_item(tIdentificator item_id);

        /**
         * @brief Returns the item description structure.
         *
         * This function returns the item description structure into the
         * supplied <tt>it_desc</tt> structure (must be already allocated)
         * from item with <tt>item_id</tt> identificator.
         * If the item is not present in the pagefile, returns NULL,
         * otherwise the pointer to the user supplied <tt>it_desc</tt> 
         * structure.
         */ 
        inline struct tItem *get_item_description(tIdentificator item_id,
               struct tItem *it_desc); 

        /**
         * @brief Sets the pagefile closed for inserting.
         *
         * Sets the pagefile closed. This means that no more items can be 
         * inserted into it.
         */
        inline void set_closed();

        /**
         * @brief Checks if all items are extracted and the pagefile is
         * set closed.
         * 
         * When the pagefile is closed, no more items can be inserted into it.
         * In the moment, when all items are marked as extracted, the pagefile
         * can be removed. 
         * Returns zero if conditions above are not met, otherwise <tt>1</tt>.
         */
        inline int to_remove();

};

#include "spool/pagefile.hxx"

} // namespace spool

#endif

