/**
 * @file queue.h
 * @brief Queue implementation
 * 
 * This file contains tQueue class and structures needed for
 * storing of data items into the queue.
 * 
 * @author Ales Privetivy <ales@ucw.cz>
 */

#ifndef _ECHELOG_QUEUE
#define _ECHELOG_QUEUE

#include "spool/pagefile.h"

/** This constant defines the pagefile size used in queue. */
#define MAX_PAGE_SIZE 400000
/** This constant defines the maximal number of items in queue. */
#define MAX_ITEMS       4096 
/** This constant defines the maximal number of pagefiles in queue. */
#define MAX_PAGES       8192

namespace spool {
	
/**
 * @class tQueue
 * @brief This class implements the queue as was described in the introduction.
 *
 * The queue is of a chain of pagefiles. Items can be inserted only
 * into the tail pagefile and extracted only from the head pagefile. 
 * This is forced by the rule that the whole queue must be formed
 * from items with consecutive identificators.
 *
 * We need to keep information, which pagefiles and in which order form the
 * queue. This is ensured by the index file which keeps values of first 
 * identificators for each pagefile. Since the pagefile filenames are
 * values of their first item identificators, this suffices for reconstruction
 * of the whole queue.
 *
 * There are three constants #MAX_PAGE_SIZE, #MAX_ITEMS and #MAX_PAGES
 * determining, how large the pagefiles are, how much items can be inserted in 
 * one of them and how many pagefiles we can have. We will be able to hold only
 * #MAX_ITEMS * #MAX_PAGES items of total size less
 * than #MAX_PAGE_SIZE * #MAX_PAGES bytes.
 * 
 */
class tQueue {

    private:
        /** 
         * @brief <em>Property:</em> Access mode (<tt>0</tt> means cached mode,
         * <tt>1</tt> means writethrough). 
         */
        SIMPLE_PROPERTY(int, write_through);  
        
        /** 
         * @brief <em>Property:</em> Indicates that the instance has loaded the
         * index file. 
         */
        SIMPLE_PROPERTY(int, present);   

        /**
         * @brief <em>Property:</em> Number of pagefiles in the queue. 
         */
        SIMPLE_PROPERTY(tIdentificator, pages);
        
        /**
         * @brief <em>Property:</em> The mmaped index file. 
         */
        SIMPLE_PROPERTY(tIdentificator *,idx);  

        /**
         * @brief <em>External resource:</em> Array of pagefiles in the queue. 
         */
        SIMPLE_EXT_RESOURCE(class tPageFile **,pg_file);
       
        /**
         * @brief <em>Property:</em> Path to spool dir. 
         */
        SIMPLE_PROPERTY(char *,spool_dir);  
        /**
         * @brief <em>Property:</em> Filename of the queue index file. 
         */
        SIMPLE_PROPERTY(char *,filename);    
        /**
         * @brief <em>Property:</em> File descriptor of the associated index
         * file. 
         */ 
        SIMPLE_PROPERTY(int, fd);              

    private:
        int attach_queue();        /* loads the queue into memory */
        int detach_queue();        /* removes the queue from memory */
        inline int synchro_idx_file(); /* forces synchro of index file */
 
    public:

        /**
         * @brief Creates a queue object.
         *
         * This constructor creates a queue object, sets its access mode 
         * accordingly to <tt>wr_thr</tt> parameter, sets the queue working
         * directory to <tt>sp_dir</tt> and looks for the queue index file
         * with filename <tt>fname</tt>, if it does not exists it creates
         * new one. After this the index file is mapped into memory.
         */
        tQueue(int wr_thr, char *sp_dir, char *fname);

        /**
         * @brief Destructs the queue object.
         *
         * This destructor calls only <tt>close_queue()</tt> method.
         */
        virtual ~tQueue();

        /**
         * @brief Opens a queue and makes it ready for use.
         *
         * This function looks into the queue's index file and tries to 
         * open all the pagefiles in queue. The items with item_id 
         * lesser than <tt>limit</tt> are extracted. All the pagefiles 
         * except of the head and the tail pagefile are detached then.
         * Returns zero on success, otherwise <tt>-1</tt>.
         */
        int open_queue(tIdentificator limit); 

        /**
         * @brief Closes the queue.
         *
         * Synchronizes all data and frees all resources.
         * Returns zero on success, otherwise <tt>-1</tt>.
         */
        int close_queue();

        /**
         * @brief Destroys the whole queue.
         *
         * Deletes all pagefiles and index file. The queue must be closed
         * before this action.
         * Returns zero on success, otherwise <tt>-1</tt>.
         */
        int destroy_queue();

        /**
         * @brief Extracts items from the queue with item_id < <tt>limit</tt>.
         *
         * Extracts items from the queue with item_id < <tt>limit</tt>. The
         * queue must be closed before this action. Returns zero on success,
         * otherwise <tt>-1</tt>.
         */
        int strip_queue(tIdentificator limit);   

        /**
         * @brief Inserts the item into the queue.
         *
         * This function creates a new item and inserts it into the tail 
         * pagefile. If the item won't fit in pagefile, the tail pagefile
         * is set closed and new one is created and appended to the queue.
         * Returns the item_id, which was assigned to the newly inserted item,
         * otherwise zero on error.
         */
        tIdentificator insert_item(char *buffer, int length);

        /**
         * @brief Marks the item as extracted.
         *
         * This method marks the item with <tt>item_id</tt> as extracted. 
         * If all items in the head pagefile are ``extracted'' and the pagefile
         * is also tail, the pagefile is emptied. If all items in the head
         * pagefile are ``extracted'' and the head pagefile is not also tail,
         * it is removed (deleted). 
         */
        void extract_item(tIdentificator item_id);
        
        /**
         * @brief Returns the pointer to the item data.
         *
         * This function returns pointer to the item data (pointer to
         * beginning of data, item description is skipped) for the item
         * with <tt>item_id</tt> identificator.
         * If the item is not present in the head pagefile, 
         * returns NULL, otherwise pointer to the memory where the 
         * item data are held is returned.
         */
        inline char *get_item(tIdentificator item_id);

        /**
         * @brief Returns the item description structure.
         *
         * This function returns the item description structure into the
         * supplied <tt>it_desc</tt> structure (must be already allocated)
         * for item with <tt>item_id</tt> identificator.
         * If the item is not present in the head pagefile, returns 
         * NULL, otherwise the pointer to the user supplied 
         * <tt>it_desc</tt> structure is returned.
         */
        inline struct tItem *get_item_description(tIdentificator item_id, 
               struct tItem *it_desc);

        /**
         * @brief Returns the identificator of the last item in the queue
         * plus one.
         *
         * Returns the identificator of the last item in the queue plus one.
         * If there is no item in the queue, it returns the identificator of
         * the last inserted item plus one.
         */
        inline tIdentificator get_last_id();

        /**
         * @brief Returns the identificator of the first item in the queue.
         *
         * Returns the identificator of the first item in the queue. If 
         * there is no item in the queue, it returns the identificator of
         * the last inserted item plus one.
         */
        inline tIdentificator get_first_id();

        /**
         * @brief Returns the identificator of the first non-extracted item 
         * in the queue.
         *
         * Returns the identificator of the first non-extracted item in 
         * the queue. If there is no non-extracted item in the queue, 
         * it returns the identificator of the last inserted item plus one.
         */
        inline tIdentificator get_first_nonextracted_id();

        /**
         * @brief Returns the number of items currently held in the queue.
         *
         * Returns the number of both (extracted and non-extracted) items in
         * the queue.
         */
        inline tIdentificator get_item_count();

        /**
         * @brief Returns the status field of an item.
         *
         * This function returns the status field of an item with 
         * <tt>item_id</tt> identificator. If the item is not in the pagefile 
         * <tt>socks::ITEMSTATUS::MISSING</tt> is returned.
         */
        inline enum ITEMSTATUS get_status(tIdentificator item_id);
        
};

#include "spool/queue.hxx"

} // namespace spool

#endif
