/**
 * @file
 * @brief Echelog library - Compressed memory bitmap
 *
 * This provides a simple interface for a table space management
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/lib.h"
#include <sys/types.h>
#include <limits.h>
#include <stdio.h>


namespace lib {

// Private inline functions

// Clear an allocated bit
inline void tCBitmap::clear_bit (long n)
{ 
    ASSERT(n >= 0 && n < num_of_bits);      
    mem_bitmap[n / UINT32_T_BITS] &= ~(1 << (n % UINT32_T_BITS));
}


// Allocate a bit
inline void tCBitmap::set_bit (long n)
{ 
    ASSERT(n >= 0 && n < num_of_bits);      
    mem_bitmap[n / UINT32_T_BITS] |= 1 << (n % UINT32_T_BITS);
}


// Query whether the bit is set
inline bool tCBitmap::is_set (long n)
{
    ASSERT(n >= 0 && n < num_of_bits);      

    return mem_bitmap[n / UINT32_T_BITS] & (1 << (n % UINT32_T_BITS));
}


// Query whether the bit is free
inline bool tCBitmap::is_free (long n)  
{ 
    return (!is_set (n)); 
}


// Skip over all false bits up to the limit
inline long tCBitmap::skip_false_bits (long start_pos, long length_limit)
{
    register long start_bit;
    register long j;
    uint32_t current_block;

    long max_length = min (start_pos + length_limit, num_of_bits);
    long max_block_count = div_round_up (max_length, UINT32_T_BITS);
    
    ASSERT (start_pos >= 0 && start_pos < max_length);
    
    start_bit = start_pos % UINT32_T_BITS;
    for (j = start_pos / UINT32_T_BITS; j < max_block_count - 1; j++)
    {
        current_block = mem_bitmap[j];
        
        // Skip all items filled only with zeroes
        if (current_block)
        {
            register long i;
            long mask = ((uint32_t) 1) << start_bit;
            for (i = start_bit; i <= UINT32_T_BITS - 1; i++)
            {
                if (current_block & mask)
                    return (j * UINT32_T_BITS + i) - start_pos;
                mask = mask << 1;
            }
        }
        start_bit = 0;
    }

    current_block = mem_bitmap[j];
    if (current_block)
    {
        long end_bit = (max_length - 1 ) % UINT32_T_BITS;
        long mask = ((uint32_t) 1) << start_bit;
        register long i;
        for (i = start_bit; i <= end_bit; i++)
        {
            if (current_block & mask)
                return (j * UINT32_T_BITS + i) - start_pos;
            mask = mask << 1;
        }
    }
 
    // not found
    return max_length - start_pos;
}


// Skip over all true bits up to the limit
inline long tCBitmap::skip_true_bits (long start_pos, long length_limit)
{
    register long start_bit;
    register long j;
    uint32_t current_block;

    long max_length = min (start_pos + length_limit, num_of_bits);
    long max_block_count = div_round_up (max_length, UINT32_T_BITS);
    
    ASSERT (start_pos >= 0 && start_pos < max_length);
    
    start_bit = start_pos % UINT32_T_BITS;
    for (j = start_pos / UINT32_T_BITS; j < max_block_count - 1; j++)
    {
        current_block = ~(mem_bitmap[j]);
        
        // Skip all items filled only with zeroes
        if (current_block)
        {
            register long i;
            long mask = ((uint32_t) 1) << start_bit;
            for (i = start_bit; i <= UINT32_T_BITS - 1; i++)
            {
                if (current_block & mask)
                    return (j * UINT32_T_BITS + i) - start_pos;
                mask = mask << 1;
            }
        }
        start_bit = 0;
    }

    current_block = ~(mem_bitmap[j]);
    if (current_block)
    {
        long end_bit = (max_length - 1 ) % UINT32_T_BITS;
        long mask = ((uint32_t) 1) << start_bit;
        register long i;
        for (i = start_bit; i <= end_bit; i++)
        {
            if (current_block & mask)
                return (j * UINT32_T_BITS + i) - start_pos;
            mask = mask << 1;
        }
    }
 
    // not found
    return max_length - start_pos;
}


// Test if a continuous block of free bits is available 
inline long tCBitmap::test_block (long start_bit, long block_length)
{
    // sanity check
    ASSERT (start_bit >= 0 && start_bit < num_of_bits 
            && block_length > 0 && start_bit + block_length <= num_of_bits );

    // check for a continuous block
    long res;
    if ((res = skip_false_bits (start_bit, block_length)) != block_length)
        return res;

    return 0;
}


// Private inline functions end

// Set up a bit storage and clear it out
tCBitmap::tCBitmap (long total_bits)
{
    ASSERT (total_bits > 0);

    /* Allocate a storage big enough */
    num_of_bits = total_bits;
    num_of_items = div_round_up (num_of_bits, UINT32_T_BITS);
    mem_bitmap = static_cast<uint32_t *> 
        (xmalloc (num_of_items * sizeof (uint32_t)));

    clear_all ();
}


// Clean-up used memory
tCBitmap::~tCBitmap ()
{ 
    xfree (mem_bitmap);
}


// Clear the whole bitmap
void tCBitmap::clear_all ()
{
    memset (mem_bitmap, 0, num_of_items * sizeof (uint32_t));
}


// Try to allocate the first block of bits at least block_length long
long tCBitmap::get_block (long block_length)
{
    long skip, i = 0;

    ASSERT (block_length > 0);

    while (i < num_of_bits)
    {
        /* Skip allocated bits */
        i += skip_true_bits (i, num_of_bits - i);

        if (i >= num_of_bits - block_length)
            break;

        skip = test_block (i, block_length);
        if (!skip)
        {
            // Okay, we found a block large enough - mark it used
            long j, m = i + block_length;

            for (j = i; j < m; j++)
                set_bit (j); 

            return i;
        }
        else
            i += skip;
    }

    /* Failed to find the block */
    return -1;    
}


// Get the first single free bit
long tCBitmap::get_bit ()
{
    long i = skip_true_bits (0, num_of_bits);

    if (i < num_of_bits && is_free (i))
    {
        // Okay, mark it used
        set_bit (i);
        return i;
    }

    /* Failed to find the bit */
    return -1;
}


// Free an allocated continuous block of bits
void tCBitmap::free_block (long start_bit, long block_length)
{
    long i, end = start_bit + block_length;

    ASSERT (start_bit >= 0 && block_length > 0 && end <= num_of_bits);

    for (i = start_bit; i < end; i++)
    {
        // Do not try to clear out unallocated bits
        ASSERT (is_set (i));
        clear_bit (i);
    }
}


// Allocate a continuous block of bits
void tCBitmap::allocate_block (long start_bit, long block_length)
{
    long i, end = start_bit + block_length;

    ASSERT (start_bit >= 0 && block_length > 0 && end <= num_of_bits);

    for (i = start_bit; i < end; i++)
    {
        // Do not try to allocate bits which are already allocated
        ASSERT (is_free (i));
        set_bit (i);
    }
}


// A CBitmap printout for debugging purposes
void tCBitmap::print_bitmap() 
{
    long i;

    printf ("CBitmap contents (total %li bits):\n", num_of_bits);

    for (i = 0; i < num_of_bits; i++)
        if (is_set(i))
            printf ("%c", CBITMAP_DEBUG_UsedBit);
        else
            printf ("%c", CBITMAP_DEBUG_FreeBit);

    printf("\n"); 
}

}
