/**
 * @brief Echelog Library -- Reading configuration files and debug facility
 * @author Robert Spalek <robert@ucw.cz>
 * @author Ales Privetivy <ales@ucw.cz>
 * @author Kamil Toman <ktoman@email.cz>
 *
 */


#include "lib/archdep.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

#include "lib/lib.h"
#include "lib/conflexer.h"

namespace lib {
    
static struct cfitem *cfsection = NULL;
static char *cf_current_config_file = NULL;
static bool config_read_status = false;
const char *gb_default_config_filename = DEFAULT_CONFIG;

/** on demand cfpool creation */
tMemPool &get_cfpool ()
{
    static tMemPool cfpool (4096);
    return cfpool;
}


/** Buffer allocation within cfpool */
void * cf_malloc(unsigned long size)
{
    return get_cfpool().alloc(size);
}


/** String duplication within cfpool */
char * cf_strdup(const char *s)
{
    return s_strdup (s, get_cfpool());
}


/** Registering of section with keys to be read */
void cf_register(struct cfitem *items)
{
    if (!items)
        die ("Registering a null section");

    if(items[0].type!=CT_SECTION)
        die("Invalid configuration section, first item must be of type "
            "CT_SECTION");
    items[0].data=cfsection;
    cfsection=items;
}


/** Unregister an section */
void cf_unregister(const char *section_name)
{
    struct cfitem *item, *prev = NULL;

    if (!section_name)
        die ("Unregistering a null section");

    item = cfsection;
    // traverse through sections looking for a match
    while(item && strcasecmp(item->name,section_name))
        prev = item, item = static_cast<struct cfitem *>(item->data);

    if(!item)	/* unknown section */
        return;

    if (!prev)  /* removing the first section */
        cfsection = static_cast<cfitem *> (item->data);
    else        /* removing from the middle or from the end */
        prev->data = item->data;
}


/** Look-up for key name in section sect */
static struct cfitem *cf_get_item(char *sect, char *name)
{
    struct cfitem *item;

    item = cfsection;
    // traverse through sections looking for a match
    while(item && strcasecmp(item->name,sect))
        item = static_cast<struct cfitem *>(item->data);

    if(!item)	/* unknown section */
        return NULL;
    
    for(++item; (item->type != CT_STOP) && strcasecmp(item->name,name); item++)
        ;

    return item;	/* item->type == CT_STOP if not found */
}


/** Clean up all registered sections to allow a configuration reload */
void cf_cleanup_registered_sections()
{
    struct cfitem *item,*subitem;

    item = cfsection;
    // traverse through all currently registered sections
    while(item)
    { 
        /* traverse for all keys in current section */
        for(subitem = item; subitem->type != CT_STOP; subitem++)
            switch (subitem->type)
            {
                case CT_STRING: /* reset string pointers */
                    ASSERT (subitem->data);
                    *static_cast<char **>(subitem->data) = NULL;
                    break;
                case CT_FUNCTION: /* if the item is loaded call reinit func */
                    if ((subitem->flags | ITEM_LOADED) && subitem->varf_reset)
                        subitem->varf_reset (subitem);
                    break;
                case CT_LIST:   /* if the item is loaded call reinit func */
                    if (subitem->flags | ITEM_LOADED)
                        static_cast<class tList *>(subitem->data)->reset();
                    break;
                default:
                    ; /* no reinitialization is needed */
            }

        /* advance to next section */
        item = static_cast<struct cfitem *>(item->data);
    }
    
}


/** Clears loaded flags - especially before file reread */
void cf_clear_loaded_flags()
{
    struct cfitem *item,*subitem;

    item = cfsection;
    // traverse through sections 
    while(item)
    {
        for(subitem = item; subitem->type != CT_STOP; subitem++)
            subitem->flags &= ~ITEM_LOADED;
        item = static_cast<struct cfitem *>(item->data);
    }
    
}


/** Checks if every required parameter is set - 0 if ok, 1 if missing */
int cf_check_required()
{
    struct cfitem *item,*subitem;
    int fail=0;

    item = cfsection;
    // traverse through sections 
    while(item)
    {
        for(subitem = item; subitem->type != CT_STOP; subitem++)
        {

            if ((subitem->flags & (ITEM_REQUIRED | ITEM_LOADED)) == 1)
            {
               fail=1;
               log(L_ERROR,"Missing parameter value %s.%s.",
                            item->name, subitem->name);
            }
        }

        item = static_cast<struct cfitem *>(item->data);
    }

    return fail; 
}


/** Setting value of key in given section */ 
static char *cf_set_item(char *sect, char *name, char *value, tMemPool &pool)
{
    struct cfitem *item;
    char *msg = NULL;

    item=cf_get_item(sect,name);
    if(!item)	/* ignore unknown section */
        return NULL; 

    if (item->type!=CT_STOP && item->type!=CT_LIST && 
        item->type!=CT_FUNCTION && item->flags&ITEM_LOADED)
    {
        log (L_INFO, "Value of key \"%s.%s\" redefined to \"%s\".",
                sect, name, value);
    }
    
    switch(item->type){
        case CT_INT:
            {
                char *end;
                if(!*value)
                    msg="Missing number";
                else{
                    *static_cast<unsigned long *>(item->data) 
                        = strtoul(value, &end, 0);
                    if (end && *end)
                        msg = "Invalid number";
                }
                item->flags |= ITEM_LOADED;
                break;
            }
        case CT_STRING:
            item->flags |= ITEM_LOADED;
            *static_cast<char **>(item->data) = s_strdup(value, pool);
            break;
        case CT_FUNCTION:
            ASSERT (item->varf);
            item->flags |= ITEM_LOADED;
            msg = item->varf(item, s_strdup(value, pool));
            break;
        case CT_LIST:
            item->flags |= ITEM_LOADED;
            static_cast<class tList*>
                (item->data)->append(s_strdup(value, pool));
            break;
        default:
            msg = "Unknown key ";
            break;
    }

    return msg;
}


/** Printing of error message with place of encouter */
static void cf_error_conf(char *error_msg)
{
    using namespace conflex_lib;
    int i;

    if (gb_include_stack_ptr==0)
        log(L_ERROR,"In file %s line %d: %s",gb_conf_stack[0].filename,
                gb_conf_stack[0].line_no, error_msg);
    else
    {
        log(L_ERROR,"In file %s line %d: included from",
                gb_conf_stack[gb_include_stack_ptr].filename,
                gb_conf_stack[gb_include_stack_ptr].line_no);

        for(i=gb_include_stack_ptr-1;i;i--)
        {
            log(L_ERROR,"file %s line %d included from",
                    gb_conf_stack[i].filename, gb_conf_stack[i].line_no);
        }

        log(L_ERROR,"file %s line %d: %s",
                gb_conf_stack[0].filename, gb_conf_stack[0].line_no, error_msg);
    }

}


void cf_set_config_filename (const char *new_config_name)
{
    if (cf_current_config_file)
        xfree(cf_current_config_file);
    cf_current_config_file = s_strdup (new_config_name);
}


const char* cf_get_config_filename ()
{
    if (cf_current_config_file)
        return cf_current_config_file;
    else
        return "NULL";
}


/* Performs a (re)read of the current configuration file */
int cf_reread(tMemPool &pool)
{
   if (!cf_current_config_file)
       cf_current_config_file = s_strdup (gb_default_config_filename);

   return cf_reread(cf_current_config_file, pool);
}


/* Performs a preprocess of a configuration file in to file out */
/* Memory allocations are done in cf_pool, so ensure clearing */
int cf_preprocess(const char *in, const char *out)
{
    using namespace conflex_lib;

    ASSERT(in);
    ASSERT(out);
    
    expandin=fopen(in,"r");
    if(!expandin){
        log(L_ERROR,"Cannot open configuration file %s: %m",in);
        return -1;
    }

    expandout=fopen(out,"w+");
    if(!expandout){
        log(L_ERROR,"Cannot create output file %s: %m",out);
        return -1;
    }

    gb_conf_stack[0].filename = const_cast<char *>(in);
    gb_conf_stack[0].line_no = 1;
    gb_include_stack_ptr = 0;
    gb_actual_line = &(gb_conf_stack[0].line_no);

    return expandlex();

}


/* prints error passed from preprocessor */
void cf_report_error()
{
    using namespace conflex_lib;

    if(gb_conf_error)
                cf_error_conf(gb_conf_error);
}


/* Performs a (re)read of a configuration file */
int cf_reread(const char *filename, tMemPool &pool)
{
    using namespace conflex_lib;

    if (!filename)
        die ("Trying to load a null configuration filename");
    
    char key[MAXDATALEN + 1], def_section[MAXDATALEN + 1];
    char *msg = NULL;
    enum LX_RESULT token;

    // Now we supress loading of default conf-file
    config_read_status = true;

    // Remember the configuration file 
    if (filename != cf_current_config_file)
        cf_set_config_filename (filename);

    debug_log ('c', "Using configuration file '%s'", filename);
    confin=fopen(filename,"r");
    if(!confin){
        log(L_ERROR,"Cannot open configuration file %s: %m",filename);
        return 1;
    }

    gb_conf_stack[0].filename = const_cast<char *>(filename);
    gb_conf_stack[0].line_no = 1;
    gb_include_stack_ptr = 0;
    gb_actual_line = &(gb_conf_stack[0].line_no);

    confout = NULL;  

    def_section[0]=0;

    while(1){

        token=static_cast<enum LX_RESULT>(conflex());	
        switch(token)
        {
            case LX_EOF: 
                fclose( confin );
                return 0;

            case LX_SECTION: 
                strncpy(def_section,gb_conf_data,MAXDATALEN);
                def_section[MAXDATALEN] = '\0';
                break;

            case LX_IDENT:
                strncpy(key,gb_conf_data,MAXDATALEN);
                key[MAXDATALEN] = '\0';
                break;

            case LX_DATA:
                msg=cf_set_item(def_section,key,gb_conf_data,pool);
                break;

            case LX_ERROR:
                msg=gb_conf_error;
                break;
        }

        if (msg) break;

    }	

    if(msg)
        cf_error_conf(msg);
    log(L_ERROR, "There were errors while reading configuration file");

    return 1;
}


/** Reads parameters from conf-file and program arguments */
int cf_getopt(int argc,char * const argv[],
        const char *shortopts,const struct option *longopts,
        int *longindex, tMemPool &pool)
{
    int res;
    bool already_read = config_read_status;

    optind = 0;
    do{
        res=getopt_long(argc,argv,shortopts,longopts,longindex);
        switch (res)
        {
            case 'S':
                {
                    char *sect,*name,*value;
                    char *c;
                    char *msg=NULL;

                    name=optarg;
                    c=strchr(name,'=');
                    if(!c)
                    {
                        msg="Missing argument";
                        sect=value="";
                    }
                    else
                    {
                        *c++=0;
                        value=c;

                        c=strchr(name,'.');
                        if(!c)
                            sect="";
                        else{
                            sect=name;
                            *c++=0;
                            name=c;
                        }

                        if (!config_read_status)
                            if (cf_reread(pool))
                                return -2;

                        msg=cf_set_item(sect,name,value,pool);
                        debug_log ('c', "Explicitly setting %s='%s' in "
                                "section [%s]", name, value, sect);
                        *--value='=';
                        if (sect != "")
                            *--name='.';
                    }
                    if(msg)
                    {
                        log (L_ERROR, "Invalid command line argument %s '%s=%s"
                                ": %s", sect, name, value, msg);
                        return -2;
                    }
                }
                break;
            case 'C':
                if (!config_read_status)
                {
                    if (cf_reread(optarg, pool))
                        return -2;
                }
                else
                {
                    if (!already_read)
                        log (L_WARN, "Option -C specified after -S option, "
                            "therefore -C is ignored.");
                }
                break;
            case 'X':
                debug_section_set (optarg);
                break;
	        case 'h':
    	    	fprintf (stderr, "%s", CF_COMPLETE_HELP_MSG);
	    	    exit (0);
        		break;
            case -1:
                /* end of options, read the configuration if not read */
                if (!config_read_status)
                    if (cf_reread(pool))
                        return -2;
                return -1;
            default:
                /* unhandled option or missing parameters */
                log (L_ERROR, "Unknown option or missing parameters on the "
                    "command line.");
                return -2;
        }
    }while(1);
}


/** Wrapper for cf_getopt for simple programs */
int cf_process(int argc, char **argv, tMemPool &pool)
{
    int c;

    while(1){
        c=cf_getopt(argc, argv, CF_SHORT_OPTS, CF_NO_LONG_OPTS, NULL, pool);
        if(c==-1)
            break;
        else if (c==-2)
            return 1;
    }

    if (optind < argc)
    {
        log(L_ERROR, "non-option ARGV-elements");
        return 1;
    }

    return 0;
}

}
