/**
 * @brief Echelog Library -- Lexical parser for confoguration files 
 * @author Ales Privetivy <ales@ucw.cz>
 *      
 */


%option noyywrap
%option nounput
%option prefix="conf"

%x clx                
/* indicates we are in block (i.e. { ... } )  */
%x iden         
/* indicates that last token was identifier  */
%x sect                
/* indicates that we are in a section  */
%x incl                
/* include from outside of section  */
%x sincl        
/* include from inside of section  */

%{
     #include <string.h>
     #include <stdio.h>
     #include <ctype.h>
     #include "lib/lib.h"
     #include "lib/conflexer.h"

     namespace lib {
         
     namespace conflex_lib {
         
        /* Variables holding file nesting */ 
         struct tConfNest gb_conf_stack[CONF_MAX_INCLUDE_DEPTH];
         int gb_include_stack_ptr;
         int *gb_actual_line;

        /* Variables for passing output */

         char *gb_conf_data;
         char *gb_conf_error;
        }
     }

     /* Some local stuff */
    
     static char conf_tmp_buf[MAXDATALEN];
     static char * cfil,*dfil,*cfl,*efil;
     static char * CE_2K = "data exceeds 2k size";
     static char * CE_SYNTAX = "syntax error";
     static char * CE_UNEXPEOF = "unexpected end of file";
     static char * CE_DEEP = "includes nested too deeply";
     static char * CE_CINCL = "error in include syntax";
     static char * CE_NOTEXISTS = "file for inclusion does not exists";

     /* start of conflex */
%}

IDENT       [_a-zA-Z][_A-Za-z0-9]*
ESCAPE      \\[^\n]

%%
        using namespace lib;
        using namespace conflex_lib;
        char *temporal=conf_tmp_buf, *limit=conf_tmp_buf+MAXDATALEN;

<sect>[ \t]*include     {
                                BEGIN(sincl);
                        }

<INITIAL>[ \t]*include  {
                                BEGIN(incl);
                        }

<incl,sincl>[ \t]*\n    {
                                gb_conf_error = CE_CINCL;
                                return LX_ERROR;
                        }

<incl>[ \t]*\"({ESCAPE}|[^\n\"\\])*\"[ \t]*        {

                BEGIN(INITIAL);

                // check if maximal nesting depth reached
                if ( gb_include_stack_ptr >= CONF_MAX_INCLUDE_DEPTH )
                {
                    gb_conf_error = CE_DEEP;
                    return LX_ERROR;
                }

                /* look for left " */
                for(cfil=yytext;*cfil && *cfil!='"';cfil++) ; 
                cfil++;

                /* look for right " */
                for(dfil=cfil;*dfil;dfil++) ;
                for(;dfil>cfil && *dfil!='"';dfil--) ;
                *dfil = 0;

                /* strip slashes */
                for(efil=dfil=cfil;*dfil;dfil++)
                {
                    if (*dfil=='\\') 
                        if (dfil[1]) dfil++; 
                    *efil ++= *dfil;         
                }
                *efil = 0 ;

                // backup current buffer
                gb_conf_stack[gb_include_stack_ptr].buff = YY_CURRENT_BUFFER;
                gb_include_stack_ptr++;

                // extract filename in which is new file included
                dfil = gb_conf_stack[gb_include_stack_ptr-1].filename;
                // allocated enough space for new file-name
                cfl = gb_conf_stack[gb_include_stack_ptr].filename = 
                (char *)cf_malloc(strlen(cfil)+strlen(dfil)+1);

                // copy the first part into string
                strcpy(cfl,dfil); 

                // leave only path from it
                for(dfil=cfl+strlen(cfl);dfil>cfl && *dfil!='/';dfil--) ;
                if (*dfil=='/') *(++dfil)=0; else *dfil = 0;

                // and add the relative filename
                if (*cfil!='/') strcpy(dfil,cfil); else strcpy(cfl,cfil);

                // add the info about included  to stack
                gb_conf_stack[gb_include_stack_ptr].line_no = 1; 
                gb_actual_line = &(gb_conf_stack[gb_include_stack_ptr].line_no);

                // try to open it
                yyin = fopen( cfl, "r" );

                // file not exist or another problem ?
                if ( ! yyin )
                {
                     gb_conf_error = CE_NOTEXISTS; 
                     gb_include_stack_ptr--;
                     return LX_ERROR;
                }

                // switch the processing to it
                yy_switch_to_buffer( yy_create_buffer( yyin, YY_BUF_SIZE ) ); 

}


<sincl>[ \t]*\"({ESCAPE}|[^\n\"\\])*\"[ \t]*        {

                BEGIN(sect);

                // check if maximal nesting depth reached
                if ( gb_include_stack_ptr >= CONF_MAX_INCLUDE_DEPTH )
                {
                    gb_conf_error = CE_DEEP;
                    return LX_ERROR;
                }

                /* look for left " */
                for(cfil=yytext;*cfil && *cfil!='"';cfil++) ; 
                cfil++;

                /* look for right " */
                for(dfil=cfil;*dfil;dfil++) ;
                for(;dfil>cfil && *dfil!='"';dfil--) ;
                *dfil = 0;

                /* strip slashes */
                for(efil=dfil=cfil;*dfil;dfil++)
                {
                    if (*dfil=='\\') 
                        if (dfil[1]) dfil++; 
                    *efil ++= *dfil;         
                }
                *efil = 0 ;

                // backup current buffer
                gb_conf_stack[gb_include_stack_ptr].buff = YY_CURRENT_BUFFER;
                gb_include_stack_ptr++;

                // extract filename in which is new file included
                dfil = gb_conf_stack[gb_include_stack_ptr-1].filename;
                // allocated enough space for new file-name
                cfl = gb_conf_stack[gb_include_stack_ptr].filename = 
                (char *)cf_malloc(strlen(cfil)+strlen(dfil)+1);

                // copy the first part into string
                strcpy(cfl,dfil); 

                // leave only path from it
                for(dfil=cfl+strlen(cfl);dfil>cfl && *dfil!='/';dfil--) ;
                if (*dfil=='/') *(++dfil)=0; else *dfil = 0;

                // and add the relative filename
                if (*cfil!='/') strcpy(dfil,cfil); else strcpy(cfl,cfil);

                // add the info about included  to stack
                gb_conf_stack[gb_include_stack_ptr].line_no = 1; 
                gb_actual_line = &(gb_conf_stack[gb_include_stack_ptr].line_no);

                // try to open it
                yyin = fopen( cfl, "r" );

                // file not exist or another problem ?
                if ( ! yyin )
                {
                     gb_conf_error = CE_NOTEXISTS; 
                     gb_include_stack_ptr--;
                     return LX_ERROR;
                }

                // switch the processing to it
                yy_switch_to_buffer( yy_create_buffer( yyin, YY_BUF_SIZE ) ); 
               
}

<sect,INITIAL,clx>^#[^\n]*                

<sect,INITIAL,clx>\n        {
                                // line counter
                                (*gb_actual_line)++;
                            }

<sect,INITIAL>[ \t]

<sect,INITIAL>.             {
                                 gb_conf_error=CE_SYNTAX;return LX_ERROR; 
                            }        

<sect,INITIAL>"["{IDENT}"]" {
                                // we are inside new section
                                BEGIN(sect);
                                strncpy(conf_tmp_buf,yytext,MAXDATALEN);
                                gb_conf_data = conf_tmp_buf + 1;
                                conf_tmp_buf[strlen(conf_tmp_buf)-1]=0;
                                return LX_SECTION;
                            }


<sect>{IDENT}               {
                                // new key name encountered
                                BEGIN(iden);
                                gb_conf_data = yytext;
                                return LX_IDENT;
                            }

<iden>[ \t]+"{"             {
                                // key value would be multiline
                                BEGIN(clx);
                                conf_tmp_buf[0]=0;
                                temporal = conf_tmp_buf;
                            }

<iden>([ \t]*|[ \t]+[^{\n][^\n]*)                {
                                // next item is section or new key
                                BEGIN(sect);

                                // strip whitespace from beginning and end
                                for(cfil=yytext;*cfil && isspace(*cfil);cfil++) ;
                                for(dfil=cfil+strlen(cfil);dfil>cfil && isspace(*(dfil-1));dfil--) ;
                                *dfil =0;

                                // return the singleline key value
                                gb_conf_data = cfil;
                                return LX_DATA;
                        }

<clx>[^}\n]*            {
                                // check if limit is not exceeded
                                if(temporal + strlen(yytext) + 2 >= limit)
                                {
                                        gb_conf_error = CE_2K;
                                        return LX_ERROR;
                                }        

                                // add line to multiline key value
                                strncpy( temporal , yytext, limit - temporal);
                                temporal += strlen(yytext);
                                *temporal++ = ' ';
                                *temporal = 0; 
                        }

<clx>"}"                {
                                // next item is section or new key
                                BEGIN(sect);
                                
                                // strip whitespace from beginning and end and '}'
                                for(cfil=conf_tmp_buf;*cfil && isspace(*cfil);cfil++) ;
                                for(dfil=cfil+strlen(cfil);dfil>cfil && isspace(*(dfil-1));dfil--) ;
                                *dfil =0;

                                gb_conf_data = cfil;
                                return LX_DATA;
                        }        
        
<sect,INITIAL><<EOF>>   {
                                // we are at end of last file
                                if ( --gb_include_stack_ptr < 0 )
                                {
                                        return LX_EOF; 
                                }
                                else
                                {
                                        // leaving nested file
                                        fclose( yyin );
                                        yy_delete_buffer( YY_CURRENT_BUFFER );
                                        yy_switch_to_buffer( (YY_BUFFER_STATE)
                                             gb_conf_stack[gb_include_stack_ptr].buff );
                                        gb_actual_line = &(gb_conf_stack[gb_include_stack_ptr].line_no);
                                }

                        }

<clx,iden,sincl,incl><<EOF>>        {
                                // unexpected end of file
                                gb_conf_error = CE_UNEXPEOF;
                                return LX_ERROR;
                        }

%%

namespace lib {
    namespace conflex_lib {
        /* Namespace wrapper function */
        int conflex (void) { return ::yylex(); }
    }
}
