/**
 * @brief Echelog Library -- Logging
 * @author Martin Mares <mj@ucw.cz>
 * @author Ales Privetivy <ales@ucw.cz> 
 *	
 */

#include "lib/lib.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>


namespace lib {

/** program name for printing in log messages */
static const char *log_progname;
/** pid for printing in log messages */
static pid_t log_pid;
/** fd of log-file - is always stderr or it was replaced */
static const int log_fd=2;
/** fd of original stderr */
static int log_fd_orig;
/* write to log file denied */
static int log_deny = 0;

/** forked programs now log also their pids */
void log_fork(void)
{
    log_pid = getpid();
}


void log_lock()
{
    WRITELOCK_FILE(log_fd);
}


void log_unlock()
{
    UNLOCK_FILE(log_fd);
}


/** outputs log message with given status */
static inline void vlog(enum lStatus cat, int section, 
        const char *msg, va_list args)
{
    time_t tim = time(NULL);
    struct tm *log_tm = localtime(&tim);
    char time_buf[32];
    char head_buf[MAX_LOG_LINE_LENGTH];
    char msg_buf[MAX_LOG_LINE_LENGTH];
    char out_buf[MAX_LOG_LINE_LENGTH];
    int headlen, msglen;
   
    if (!log_deny) 
    { 
        strftime (time_buf, sizeof(time_buf), "%Y-%m-%d %H:%M:%S", log_tm);

        if (section)
        {
            if (log_progname)
            {
                if (log_pid)
                {
                    headlen = snprintf (head_buf, sizeof(head_buf), "%c:%c: %s "
                            "[%s (%d)] ", (char)cat, (char)section, time_buf,
                            log_progname, log_pid);
                } else
                {
                    headlen = snprintf (head_buf, sizeof(head_buf), "%c:%c: %s "
                            "[%s] ", (char)cat, (char)section, time_buf,
                            log_progname);
                }
            } else
            {
                if (log_pid)
                {
                    headlen = snprintf (head_buf, sizeof(head_buf), "%c:%c: %s "
                            "[%d] ", (char)cat,(char)section,time_buf,log_pid); 
                } else
                {
                    headlen = snprintf (head_buf, sizeof(head_buf),"%c:%c: %s ",
                            (char)cat, (char)section, time_buf);
                }
            }
        } else
        {
            if (log_progname)
            {
                if (log_pid)
                {
                    headlen = snprintf (head_buf, sizeof(head_buf), "%c %s "
                            "[%s (%d)] ", (char)cat, time_buf,
                            log_progname, log_pid);
                } else
                {
                    headlen = snprintf (head_buf, sizeof(head_buf), "%c %s "
                            "[%s] ", (char)cat, time_buf,
                            log_progname);
                }
            } else
            {
                if (log_pid)
                {
                    headlen = snprintf (head_buf, sizeof(head_buf), "%c %s "
                            "[%d] ", (char)cat, time_buf,log_pid);
                } else
                {
                    headlen = snprintf (head_buf, sizeof(head_buf),"%c %s ",
                            (char)cat, time_buf);
                }
            }
        }             

        msglen = vsnprintf (msg_buf, sizeof(msg_buf), msg, args);
        snprintf (out_buf, sizeof(out_buf)-1, "%s%s", head_buf, msg_buf);
        strncat (out_buf, "\n", 1);

        char *errormsg = "Error printing log line.\n";

        WRITELOCK_FILE(log_fd);

        if (msglen <0 || headlen <0)
            s_write (log_fd, errormsg, strlen (errormsg));
        else
            s_write (log_fd, out_buf, strlen (out_buf));

        UNLOCK_FILE(log_fd);

    }

}


void log_close()
{
    log_deny = 1;
    s_close(log_fd_orig);
}


/** wrapper for vlog */
void vwlog(enum lStatus cat, const char *msg, ...)
{
    va_list args;

    va_start(args, msg);
    vlog(cat, 0, msg, args);
    va_end(args);
}


/** putput message to log and exit */
void vwdie(const char *msg, ...)
{
    va_list args;

    va_start(args, msg);
    vlog(L_FATAL, 0, msg, args);
    va_end(args);
    log_close();
    exit(1);
}


/** returns basename of file */
static inline const char * basename(const char *n)
{
    const char *p = n;

    while (*n)
        if (*n++ == '/')
            p = n;
    return p;
}


/** initilaization of logging */
int vlog_init(const char *argv0)
{
    if (argv0)
        log_progname = basename (argv0);

    if ((log_fd_orig = dup (log_fd)) < 0)
          return -1;
    
    return 0;
}


/** The same as vlog_init but exit()s with non-zero status on failure */
void log_init (const char *argv0)
{
    if (vlog_init (argv0) < 0)
    {
        char tm_str[32];
        time_t tim = time(NULL);
        struct tm *log_tm = localtime(&tim);

        strftime(tm_str, sizeof(tm_str), "%Y-%m-%d %H:%M:%S", log_tm);
        fprintf (stderr, "%c %s [%s] Can't initialize logging\n", L_FATAL,
                tm_str, log_progname);
        fflush(stderr);
        exit (1);
    }
}


/** redirect of log to file */
int log_file(const char *name)
{
    if (name)
    {
        int fd = s_open (name, O_WRONLY | O_CREAT | O_APPEND, 0600);
        if (fd < 0)
        {
            log(L_ERROR, "Unable to open log file %s: %m", name);
            return -1;
        }
        if (dup2 (fd, log_fd) < 0)
        {
            log(L_ERROR,"Unable to duplicate log file descriptor");
            return -1;
        }
        if (s_close(fd) < 0)
        {
            log(L_ERROR,"Unable to close log file fd duplicate.");
            return -1;
        }
    } 
    else 
    {
        if (dup2 (log_fd_orig, log_fd) < 0)
        {
            log(L_ERROR,"Unable to duplicate the original log file descriptor");
            return -1;
        }
    }

    return 0;
}


/** output to log for debugging */
void debug_vwlog (int section, const char *msg, ...)
{
    va_list args;

    if (debug_section_is_set (section))
    {
        va_start(args, msg);
        vlog(L_DEBUG, section, msg, args);
        va_end(args);
    }
}


/** service function */
void call_a_fake_im_a_stupid_cpp (int section, const char *msg, ...)
{
    /* do nothing, we don't want any output */
}

}
