/**
 * @brief Echelog Library -- Syntax parser for time strings
 * @author Ales Privetivy <ales@ucw.cz>
 *      
 * The BNF-like specification of the time syntax parsed is below:
 *
 * As usual, [ X ] means that X is optional, { X } means that X may
 * be either omitted or specified as many times as needed,
 * alternatives are separated by |, brackets are used for grouping.
 * (# marks the beginning of comment that extends to the end of line)
 *
 * TIME-SPECIFICATION ::= TIME-REFERENCE [ OFFSET-SPEC ] |
 *                             OFFSET-SPEC   |
 *             ( START | END ) OFFSET-SPEC
 *
 * TIME-REFERENCE ::= NOW | TIME-OF-DAY-SPEC [ DAY-SPEC-1 ] |
 *                        [ TIME-OF-DAY-SPEC ] DAY-SPEC-2
 *
 * TIME-OF-DAY-SPEC ::= NUMBER ':' NUMBER [am|pm]     # HH:MM
 *                      NUMBER ':' NUMBER ':' NUMBER  # HH:MM:SS
 *
 * DAY-SPEC-1 ::= NUMBER '/' NUMBER '/' NUMBER |  # MM/DD/[YY]YY
 *                NUMBER '.' NUMBER '.' NUMBER    # DD.MM.[YY]YY
 *
 * DAY-SPEC-2 ::= MONTH-NAME NUMBER [NUMBER] |    # Month DD [YY]YY
 *                'yesterday' | 'today' | 'tomorrow' |
 *                DAY-OF-WEEK
 *
 *
 * OFFSET-SPEC ::= '+'|'-' NUMBER TIME-UNIT 
 *
 * TIME-UNIT ::= SECONDS | MINUTES | HOURS |
 *               DAYS | WEEKS | MONTHS | YEARS
 *
 * NOW ::= 'now' 
 *
 * START ::= 'start' 
 * END   ::= 'end' 
 *
 * SECONDS ::= 'seconds' | 'second' | 'sec' | 's'
 * MINUTES ::= 'minutes' | 'minute' | 'min' | 'm'
 * HOURS   ::= 'hours' | 'hour' | 'hr' | 'h'
 * DAYS    ::= 'days' | 'day' | 'd'
 * WEEKS   ::= 'weeks' | 'week' | 'wk' | 'w'
 * MONTHS  ::= 'months' | 'month'  
 * YEARS   ::= 'years' | 'year' | 'yr' | 'y'
 *
 * MONTH-NAME ::= 'jan' | 'january' | 'feb' | 'february' | 'mar' | 'march' |
 *                'apr' | 'april' | 'may' | 'jun' | 'june' | 'jul' | 'july' |
 *                'aug' | 'august' | 'sep' | 'september' | 'oct' | 'october' |
 *                'nov' | 'november' | 'dec' | 'december'
 *
 * DAY-OF-WEEK ::= 'sunday' | 'sun' | 'monday' | 'mon' | 'tuesday' | 'tue' |
 *                 'wednesday' | 'wed' | 'thursday' | 'thu' | 'friday' | 'fri' |
 *                 'saturday' | 'sat'
 */

%{

#include "lib/archdep.h"
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include "lib/lib.h"

using namespace lib;

%}

%union {
    struct tm TimeValue;
    time_t NumericValue;      
}

%{

static struct tm zero_time = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };
static struct tm current_time;
static time_t p_time;

extern int yylex();
extern FILE *parsetimein,*parsetimeout;

int yyparse();

int yyerror (char *s)
{
    return 0;
}

namespace lib {

    extern char *gb_parsetimestr;

    time_t parsetime(char *timestr)
    {

        time_t now = time(NULL);

        current_time = *localtime(&now);

        gb_parsetimestr = timestr;
        parsetimein = NULL;
        parsetimeout = NULL;

        if (yyparse())
           return -1;
          
        return p_time;
    }

}

%}

%start time_specification 

%token ':'
%token '.'
%token '/'
%token '+'
%token '-'
%token START 
%token END
%token NOW
%token AM
%token PM
%token YESTERDAY
%token TODAY
%token TOMORROW
%token ERROR
%token <NumericValue> NUMBER
%token <NumericValue> MONTH_NAME
%token <NumericValue> DAY_OF_WEEK
%token <NumericValue> TIME_UNIT

%type  <NumericValue> time_specification 
%type  <TimeValue>    time_reference offset_spec time_of_day_spec day_spec_1 day_spec_2

%%

time_specification:     time_reference offset_spec {
                            $1.tm_sec  += $2.tm_sec;
                            $1.tm_min  += $2.tm_min;
                            $1.tm_hour += $2.tm_hour;
                            $1.tm_mday += $2.tm_mday;
                            $1.tm_mon  += $2.tm_mon;
                            $1.tm_year += $2.tm_year;
                            $1.tm_isdst = -1;
                            p_time = mktime (&$1);
                        }
                      | time_reference {
                            $1.tm_isdst = -1;
                            p_time = mktime (&$1);
                        }
                      | offset_spec  {
                            current_time.tm_sec  += $1.tm_sec;
                            current_time.tm_min  += $1.tm_min;
                            current_time.tm_hour += $1.tm_hour;
                            current_time.tm_mday += $1.tm_mday;
                            current_time.tm_mon  += $1.tm_mon;
                            current_time.tm_year += $1.tm_year;
                            current_time.tm_isdst = -1;
                            p_time = mktime (&current_time);
                        }
                      | START {
                            p_time = 0;
                        }   
                      | END {
                            p_time = 0x7FFFFFFF;
                        } 
;

time_reference:         NOW {
                            $$ = current_time;
                        } 
                      | time_of_day_spec {
                            $$ = $1;
                            $$.tm_mday = current_time.tm_mday;
                            $$.tm_mon = current_time.tm_mon;
                            $$.tm_year = current_time.tm_year;
                        }
                      | time_of_day_spec day_spec_1 {
                            $$ = $1;
                            $$.tm_mday = $2.tm_mday;
                            $$.tm_mon = $2.tm_mon;
                            $$.tm_year = $2.tm_year;
                        }
                      | day_spec_2 {
                            $$ = $1;
                        }
                      | time_of_day_spec day_spec_2 {
                            $$ = $1;
                            $$.tm_mday = $2.tm_mday;
                            $$.tm_mon = $2.tm_mon;
                            $$.tm_year = $2.tm_year;
                        }
; 

time_of_day_spec:       NUMBER ':' NUMBER {
                            $$ = zero_time;
                            $$.tm_hour = $1;
                            $$.tm_min = $3;
                        } 
                      | NUMBER ':' NUMBER AM {
                            $$ = zero_time;
                            $$.tm_hour = $1;
                            $$.tm_min = $3;
                        }
                      | NUMBER ':' NUMBER PM {
                            $$ = zero_time;
                            $$.tm_hour = $1 + 12;
                            $$.tm_min = $3;
                        }
                      | NUMBER ':' NUMBER ':' NUMBER {
                            $$ = zero_time;
                            $$.tm_hour = $1;
                            $$.tm_min = $3;
                            $$.tm_sec = $5;
                        }
;

day_spec_1:             NUMBER '/' NUMBER '/' NUMBER {
                            $$ = zero_time;
                            $$.tm_mday = $3;
                            $$.tm_mon = $1 - 1;
                            if ($5<100) $$.tm_year = 100 + $5; else
                                        $$.tm_year = $5 - 1900;
                        }
                      | NUMBER '.' NUMBER '.' NUMBER {
                            $$ = zero_time;
                            $$.tm_mday = $1;
                            $$.tm_mon = $3 - 1;
                            if ($5<100) $$.tm_year = 100 + $5; else
                                        $$.tm_year = $5 - 1900;
                        }
;

day_spec_2:             MONTH_NAME NUMBER {
                            $$ = zero_time; 
                            $$.tm_mon = $1; 
                            $$.tm_mday = $2;
                            $$.tm_year = current_time.tm_year;
                        }
                      | MONTH_NAME NUMBER NUMBER {
                            $$ = zero_time; 
                            $$.tm_mon = $1; 
                            $$.tm_mday = $2;
                            if ($3<100) $$.tm_year = 100 + $3; else 
                                        $$.tm_year = $3 - 1900;
                        }
                      | YESTERDAY {
                            $$ = current_time;
                            $$.tm_mday = current_time.tm_mday - 1;
                            $$.tm_mon = current_time.tm_mon;
                            $$.tm_year = current_time.tm_year;
                        }
                      | TODAY {
                            $$ = current_time;
                            $$.tm_mday = current_time.tm_mday;
                            $$.tm_mon = current_time.tm_mon;
                            $$.tm_year = current_time.tm_year;
                        }
                      | TOMORROW {
                            $$ = current_time;
                            $$.tm_mday = current_time.tm_mday + 1;
                            $$.tm_mon = current_time.tm_mon;
                            $$.tm_year = current_time.tm_year;
                        }                          
                      | DAY_OF_WEEK {
                            $$ = current_time;
                            $$.tm_mday = current_time.tm_mday + 
                                 ($1 - current_time.tm_wday);
                            $$.tm_mon = current_time.tm_mon;
                            $$.tm_year = current_time.tm_year;
                        }
;

offset_spec:            '+' NUMBER TIME_UNIT {
                            $$ = zero_time;
                            switch ($3)
                            {
                                case 1: $$.tm_sec  = $2;break;   // seconds
                                case 2: $$.tm_min  = $2;break;   // minutes
                                case 3: $$.tm_hour = $2;break;   // hours
                                case 4: $$.tm_mday = $2;break;   // days
                                case 5: $$.tm_mday = $2*7;break; // weeks
                                case 6: $$.tm_mon  = $2;break;   // months
                                case 7: $$.tm_year = $2;break;   // years
                                default: 
                                     log(L_WARN, "Error parsing time (offset "
                                              "part) - ignoring offset part.");
                            }
                        }
                      | '-' NUMBER TIME_UNIT {
                            $$ = zero_time;
                            switch ($3)
                            {
                                case 1: $$.tm_sec  = $2*-1;break;   // seconds
                                case 2: $$.tm_min  = $2*-1;break;   // minutes
                                case 3: $$.tm_hour = $2*-1;break;   // hours
                                case 4: $$.tm_mday = $2*-1;break;   // days
                                case 5: $$.tm_mday = $2*-7;break;   // weeks
                                case 6: $$.tm_mon  = $2*-1;break;   // months
                                case 7: $$.tm_year = $2*-1;break;   // years
                                default: 
                                     log(L_WARN, "Error parsing time (offset "
                                              "part) - ignoring offset part.");
                            }
                        }
;
%%
