/**
 * @brief Echelog Library -- Memory Pools (One-Time Allocation)
 * @author Martin Mares <mj@ucw.cz>
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include <stdlib.h>
#include <string.h>

#include "lib/lib.h"

namespace lib {

struct tMemChunk {
    struct tMemChunk *next;
    char data[0];
};

/** initialization of mempool, settting to empty */
tMemPool::tMemPool(unsigned long init_size)
{
    pfree = last = NULL;
    first = current = first_large = NULL; 
    plast = &first;
    chunk_size = init_size;
    threshold = init_size / 3;
}


/** freeing of mempool, freeing both lists */ 
tMemPool::~tMemPool()
{
    free_pool();

    pfree = last = NULL;
    first = current = first_large = NULL;
    plast = &first;
}


void tMemPool::free_pool()
{
    struct tMemChunk *c, *d;

    // free common allocation list
    for(d=first; d; d = c)
    {
        c = d->next;
        xfree(d);
    }

    // free large block list */
    for(d=first_large; d; d = c)
    {
        c = d->next;
        xfree(d);
    }

    pfree = last = NULL;
    first = current = first_large = NULL;
    plast = &first;
}


// alocation routine
void * tMemPool::alloc_slow(unsigned long size)
{
    // if size <= threshold then we allocate in common block list 
    // otherwise in large block list
    if (size <= threshold)
    {

        char *x = pfree; 
        
        // enlarge the size to ensure new address is aligned too
        size = (size + DATA_ALIGN - 1) & ~( DATA_ALIGN - 1 );

        // if there is not enough space in current block, we must create
        // new and append it to beginning of list
        if (x + size > last)
        {
            struct tMemChunk *c;

            c = static_cast<struct tMemChunk *>
                (xmalloc(sizeof(struct tMemChunk) + chunk_size));
            *plast = c;
            plast = &c->next;
            c->next = NULL; 
            x = c->data;
            last = x + chunk_size;

        }

        // allocation in current block
        pfree = x + size;
        return x;

    } else {

        // large block allocation

        struct tMemChunk *c;

        c = static_cast<struct tMemChunk *>
            (xmalloc(sizeof(struct tMemChunk) + size));

        c->next = first_large;
        first_large = c;
        return c->data;
    }
}

}
