/**
 * @brief Echelog Library -- Syscall wrappers
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/archdep.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <grp.h>
#include <pwd.h>
#include <dirent.h>

#include "lib/lib.h"

namespace lib {

static long my_egid = -1;
static long my_euid = -1;


ssize_t s_write (int fd, const void *buf, ssize_t count)
{
    ssize_t bytes_written = 0;

    /* kick off nonsential writes */
    ASSERT (buf && count >= 0);

    /* try to write everything */
    while (bytes_written < count)
    {
        /* write the rest */
        int b = write (fd, (static_cast<const char *>(buf) + bytes_written), 
                count - bytes_written);
        if (b <= 0)     /* error */
        {
            if (b == 0 || errno != EINTR) /* bad error */
                return -1;
            else 
            {
                /* not that bad, try again after a while */
                usleep (S_WAIT_BEFORE_RETRY);
                continue;
            }
        }
        else    /* new data */
            bytes_written += b;
    }
        
    if (bytes_written != count)
        return -1;

    /* seems okay */
    return bytes_written;
}


void sd_write (int fd, const void *buf, ssize_t count)
{
    if (s_write (fd, buf, count) < 0)
        die ("Can't write to file");
}


ssize_t s_read (int fd, void *buf, ssize_t count)
{
    ssize_t bytes_read = 0;

    /* kick off nonsential reads */
    ASSERT (buf && count >= 0);

    /* try to read everything */
    while (bytes_read < count)
    {
        /* read the rest */
        int b = read (fd, static_cast<char *>(buf) + bytes_read, 
                count - bytes_read);
        if (b < 0)     /* error */
        {
            if (errno == EINTR)
            {
                /* not that bad, try again after a while */
                usleep (S_WAIT_BEFORE_RETRY);
                continue;
            }
            else /* bad error */
                return -1;
        }
        else if (b == 0) /* EOF, short of data */
            return bytes_read;
        else    /* new data */
            bytes_read += b;
    }
        
    if (bytes_read != count)
        return -1;

    return bytes_read;
}


ssize_t sd_read (int fd, void *buf, ssize_t count)
{
    ssize_t bytes_read;
    if ((bytes_read = s_read (fd, buf, count)) < 0)
        die ("Can't read from file");

    return bytes_read;
}


ssize_t s_recv (int fd, void *buf, ssize_t count, unsigned int options)
{
    ssize_t bytes_received = 0;

    /* kick off nonsential receives */
    ASSERT (buf && count >= 0);

    /* try to receive everything */
    while (bytes_received < count)
    {
        /* received the rest */
        int b = recv (fd, static_cast<char *>(buf) + bytes_received, 
                count - bytes_received, options | MSG_WAITALL);
        if (b < 0)     /* error */
        {
            if (errno == EINTR)
            {
                /* not that bad, try again after a while */
                usleep (S_WAIT_BEFORE_RETRY);
                continue;
            }
            else /* bad error */
                return -1;
        }
        else if (b == 0)
            return bytes_received; /* EOF, short of data */
        else if (options & MSG_PEEK) /* just peeking */
            return b;
        else 
            bytes_received += b;
    }
        
    if (bytes_received != count)
        return -1;

    return bytes_received;
}


ssize_t sd_recv (int fd, void *buf, ssize_t count, unsigned int options)
{
    ssize_t bytes_received;
    if ((bytes_received = s_recv (fd, buf, count, options)) < 0)
        die ("Can't receive from fd");

    return bytes_received;
}


off_t s_seek (int fd, off_t offset)
{
    off_t seeked = 0;

    /* absolute seek */
    ASSERT (offset >= 0);
    seeked = lseek (fd, offset, SEEK_SET);
    if (seeked != offset)
        return -1;

    return offset;
}


void sd_seek (int fd, off_t offset)
{
    if (s_seek (fd, offset))
        die ("I/O error, can't seek to offset %li", static_cast<long> (offset));
}


int s_open (const char *name, int flags)
{
    return s_open (name, flags, 0);
}


int sd_open (const char *name, int flags)
{
    return sd_open (name, flags, 0);
}


int s_open (const char *name, int flags, mode_t mode)
{
    int new_fd;

    ASSERT (name);
    new_fd = open (name, flags, mode);
    return new_fd;
}


int sd_open (const char *name, int flags, mode_t mode)
{
    int new_fd;

    ASSERT (name);
    new_fd = open (name, flags, mode);
    if (new_fd < 0)
        switch (errno)
        {
            case EEXIST:
                die ("Can't create file '%s', file already exists.", name);
            case EISDIR:
                die ("Can't open file '%s' for writing, it refers to "
                        "a directory", name);
            case ENAMETOOLONG:
                die ("Can't open file '%s', file name too long", name);
            case ENOENT:
            case ENOTDIR:
                die ("Can't open file '%s', invalid path", name);
            case EROFS:
                die ("Can't open file '%s' for writing, read-only filesystem", 
                        name);
            case ENOSPC:
                die ("Can't open file '%s', device full", name);
            case ENFILE:
            case EMFILE:
                die ("Can't open file '%s', the limit has been reached", name);
            default:
                die ("Can't open file '%s'", name);
        }

    return new_fd;
}


int s_close (int fd)
{
    do {
        int r = close (fd);
        if (!r)
           return 0;      /* ok, closed */
        else if (errno == EINTR)
        {
            /* not that bad, try it again after a while */
            usleep (S_WAIT_BEFORE_RETRY);
        }
        else
            break;
    }
    while (1);
        
    /* we had no luck */
    return -1;
}


void sd_close (int fd)
{
    if (s_close (fd))
        die ("Can't close file");
}


int s_fstat (int fd, struct_stat *buf)
{
    ASSERT (buf);
    return fstat (fd, buf);
}


void sd_fstat (int fd, struct_stat *buf)
{
    if (fstat (fd, buf))
        die ("Can't fstat file");
}


int s_stat (const char *filename, struct_stat *buf)
{
    ASSERT (buf);
    return ::stat (filename, buf);
}


void sd_stat (const char *filename, struct_stat *buf)
{
    if (s_stat (filename, buf))
        die ("Can't stat file '%s'", filename);
}


long s_filelen (int fd)
{
    struct_stat st;

    if (fstat (fd, &st))
        return -1;
    else
        return static_cast<long> (st.st_size);
}


long sd_filelen (int fd)
{
    int res = s_filelen (fd);
    if (res < 0)
        die ("Can't get file size");

    return res;
}


long s_filelen (const char *filename)
{
    struct_stat st;

    if (s_stat (filename, &st))
        return -1;
    else
        return static_cast<long> (st.st_size);
}


long sd_filelen (const char *filename)
{
    int res = s_filelen (filename);
    if (res < 0)
        die ("Can't get file size of '%s'", filename);

    return res;
}


int s_is_my_group (gid_t n_gid)
{
    int res = 0;
    int i, n_groups;
    gid_t *grp_list;
    
    /* remember my egid */
    if (my_egid == -1)
        my_egid = getegid ();

    /* efective group is my group */
    if (n_gid == static_cast<gid_t> (my_egid))
        return 1;

    /* get number of supplementary groups and allocate space for them */
    if ((n_groups = getgroups (0, NULL)) == -1)
        return -1;
    
    grp_list = static_cast<gid_t *> (xmalloc (n_groups * sizeof (gid_t)));

    /* get all supplementary groups */
    if (getgroups (n_groups, grp_list) != n_groups || !grp_list)     
    {
        /* strange error, user had no luck */
        xfree (grp_list);
        return -1;
    }

    /* traverse and compare with desired n_gid */
    for (i = 0; i < n_groups; i++)
        if (n_gid == grp_list[i])
        {
            res = 1;
            break;
        }

    xfree (grp_list);
    return res;
}


bool sd_is_my_group (gid_t n_gid)
{
    int r;
    if ((r = s_is_my_group (n_gid)) == -1)
        die ("Can't get process's supplementary groups");

    return r;
}


int s_access (const char *filename, int mode)
{
    int r;
    ASSERT (filename);

    if ((r = access (filename, mode)) == 0)
        return 0;

    if (errno == EACCES)
        return 1;

    return -1;
}


void sd_access (const char *filename, int mode)
{
    if (s_access (filename, mode))
        die ("Can't access file '%s'", filename);
}


int s_eaccess (struct_stat *st, int mode)
{
    int r;
    ASSERT (st);
    
    if (mode == F_OK)
        return 0;       /* F_OK only */
    
    /* remember my euid */
    if (my_euid == -1)
        my_euid = geteuid ();

    /* root may everyhing */
    if (my_euid == 0)
    {
        /* but execute nonexecutable files ;) */
        if (mode != X_OK)
            return 0;

        /* needs X_OK, let's do the additional check */
        if (st->st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))
            return 0;
        else
            return -1;
    }

    if (st->st_uid == static_cast<uid_t> (my_euid))      /* owner */
        mode <<= 6;
    else 
    {
        r = s_is_my_group (st->st_gid);

        if (r == -1)                           /* sgid error */
            return -1;
        if (r == 1)                            /* group owner */
            mode <<= 3;
    }
                                               /* otherwise (mode <<= 0) */

    /* now just compare the rights */
    if (st->st_mode & mode)
        return 0;

    /* unlucky */
    return 1;
}


int s_mkstemp (char *filename)
{
    int fd;

    ASSERT(filename);
    fd = mkstemp (filename);

#if HAVE_BROKEN_MKSTEMP    
    if (fchmod (fd, 00600))
    {
        close (fd);
        fd = -1;
    }
#endif
   
    return fd;
}


int sd_mkstemp (char *filename)
{
    int fd;

    ASSERT(filename);
    fd = mkstemp (filename);

#if HAVE_BROKEN_MKSTEMP    
    if (fchmod (fd, 00600))
    {
        close (fd);
        fd = -1;
    }
#endif

    if (fd == -1)
        die ("Can't create a temporary file");
   
    return fd;
}


void sd_eaccess (struct_stat *st, int mode)
{
    if (s_eaccess (st, mode))
        die ("Can't access file");
}


int s_eaccess (const char *filename, int mode)
{
    struct_stat st;
    
    if (s_stat (filename, &st))
        return -1;

    return s_eaccess (&st, mode);
}


void sd_eaccess (const char *filename, int mode)
{
    if (s_eaccess (filename, mode))
        die ("Can't access file '%s'", filename);
}


void sd_reset_my_euid ()
{
    my_euid = -1;
}


void sd_reset_my_egid ()
{
    my_egid = -1;
}


int s_setuid (uid_t n_uid)
{
    my_euid = -1;
    if (setuid (n_uid))
        return -1;
    my_euid = n_uid;
    return 0;
}


int s_setgid (gid_t n_gid)
{
    my_egid = -1;
    if (setgid (n_gid))
        return 1;
    my_egid = n_gid;
    return 0;
}


void sd_setuid (uid_t n_uid)
{
    if (s_setuid (n_uid))
        die ("Can't set uid %i", n_uid);
}


void sd_setgid (gid_t n_gid)
{
    if (s_setgid (n_gid))
        die ("Can't set gid %i", n_gid);
}


int s_seteuid (uid_t n_uid)
{
    my_euid = -1;
    if (seteuid (n_uid))
        return -1;
    my_euid = n_uid;
    return 0;
}


int s_setegid (gid_t n_gid)
{
    my_egid = -1;
    if (setegid (n_gid))
        return 1;
    my_egid = n_gid;
    return 0;
}


void sd_seteuid (uid_t n_uid)
{
    if (s_seteuid (n_uid))
        die ("Can't set euid %i", n_uid);
}


void sd_setegid (gid_t n_gid)
{
    if (s_setegid (n_gid))
        die ("Can't set egid %i", n_gid);
}


int s_drop_supp_groups ()
{
    return setgroups (static_cast<size_t>(0), static_cast<const gid_t *>(NULL));
}


void sd_drop_supp_groups ()
{
    if (s_drop_supp_groups ())
        die ("Can't drop supplementary groups");
}


int s_set_single_gid (gid_t n_gid)
{
    my_egid = -1;
    if (s_setgid (n_gid))
        return -1;
    if (s_drop_supp_groups ())
        return -1;

    my_egid = n_gid;
    return 0;
}


void sd_set_single_gid (gid_t n_gid)
{
    if (s_set_single_gid (n_gid))
        die ("Can't set to single group %i", n_gid);
}


int s_name2uid (const char *username, uid_t &uid)
{
    struct passwd *p;

    ASSERT (username);
    p = getpwnam(username);

    if (!p)
    {
        log (L_ERROR, "Username '%s' not found in password file.", username);
        return 1;
    }

    uid = p->pw_uid;
    return 0;
}


int s_uid2name (const uid_t uid, char *&username)
{
    struct passwd *p;

    p = getpwuid(uid);

    if (!p)
    {
        log (L_ERROR, "UID '%u' not found in password file.", uid);
        return 1;
    }

    username = s_strdup (p->pw_name);
    return 0;
}


int s_uid2name (const uid_t uid, char *&username, tMemPool &pool)
{
    struct passwd *p;

    p = getpwuid(uid);

    if (!p)
    {
        log (L_ERROR, "UID '%u' not found in password file.", uid);
        return 1;
    }

    username = s_strdup (p->pw_name, pool);
    return 0;
}


int s_detach_from_ctrl_terminal ()
{
    pid_t pid, apid;

    pid = fork ();
    if (pid == 0)
    {
        /* This is the child process. Become a session group leader */
        setsid ();

        apid = fork ();
        if (apid < 0)   /* Nobody is perfect */
        {
            log (L_ERROR, "fork failed.");
            return -1;
        }
        else if (apid > 0)
            /* This is the parent process. Just exit */
            _exit (0);

    }
    else if (pid < 0)   /* Bad luck this time */
    {
        log (L_ERROR, "fork failed.");
        return -1;
    }
    else
        /* This is the parent process. Just exit */
        _exit (0);

    /* 
     * Okay, finally we're a non-session group leader
     * and we can never regain a controlling terminal.
     */
    return 0;
}


void sd_detach_from_ctrl_terminal ()
{
    if (s_detach_from_ctrl_terminal ())
        die ("Can't detach from controlling terminal");
}


int s_copy_file (const char *destpath, const char *srcpath)
{
    struct_stat st;
    int src_fd, fd, r = 0;
    char buffer[MAX_FILE_IOBUF_SIZE];
    ssize_t loop, rest, i;

    if ((src_fd = s_open (srcpath, O_RDONLY)) < 0)
    {
        close (src_fd);
        return -1;
    }

    if (s_fstat (src_fd, &st) < 0)
    {
        close (src_fd);
        return -1;
    }

    if ((fd = s_open (destpath, O_WRONLY | O_CREAT | O_TRUNC, st.st_mode)) < 0)
    {
        close (src_fd);
        return -1;
    }

    loop = st.st_size / MAX_FILE_IOBUF_SIZE;
    rest = st.st_size % MAX_FILE_IOBUF_SIZE;

    for (i = 0; i < loop; i++)
    {
        if (s_read (src_fd, buffer, MAX_FILE_IOBUF_SIZE) < 0)
        {
            close (src_fd);
            close (fd);
            return -1;
        }

        if (s_write (fd, buffer, MAX_FILE_IOBUF_SIZE) < 0)
        {
            close (src_fd);
            close (fd);
            return -1;
        }
    }

    if (rest > 0)
    {
        if (s_read (src_fd, buffer, rest) < 0)
        {
            close (src_fd);
            close (fd);
            return -1;
        }

        if (s_write (fd, buffer, rest) < 0)
        {
            close (src_fd);
            close (fd);
            return -1;
        }
    }

    if (s_close (fd) < 0)
        r = -1;
    if (s_close (src_fd) < 0)
        r = -1;

    return r;
}


void sd_copy_file (const char *destpath, const char *srcpath)
{
    struct_stat st;
    int src_fd, fd;
    char buffer[MAX_FILE_IOBUF_SIZE];
    ssize_t loop, rest, i;

    src_fd = sd_open (srcpath, O_RDONLY);
    sd_fstat (src_fd, &st);

    fd = sd_open (destpath, O_WRONLY | O_CREAT | O_TRUNC, st.st_mode);

    loop = st.st_size / MAX_FILE_IOBUF_SIZE;
    rest = st.st_size % MAX_FILE_IOBUF_SIZE;

    for (i = 0; i < loop; i++)
    {
        sd_read (src_fd, buffer, MAX_FILE_IOBUF_SIZE);
        sd_write (fd, buffer, MAX_FILE_IOBUF_SIZE);
    }

    if (rest > 0)
    {
        sd_read (src_fd, buffer, rest);
        sd_write (fd, buffer, rest);
    }

    sd_close (fd);
    sd_close (src_fd);
}


int s_move_file (const char *destpath, const char *srcpath)
{
    if (s_copy_file (destpath, srcpath) < 0)
        return -1;
    if (unlink (srcpath) < 0)
    {
        log (L_ERROR, "Can't move file '%s', copying instead.", srcpath);
        return -1;
    }

    return 0;
}


void sd_move_file (const char *destpath, const char *srcpath)
{
    sd_copy_file (destpath, srcpath);
    if (unlink (srcpath) < 0)
        die ("Can't move file '%s'", srcpath);
}


int s_test_signal_delivery (pid_t pid, int *status)
{
    if (kill (pid, 0) < 0)
        switch (errno)
        {
            case ESRCH: /* The process does not exist or might be a zombie */
                while (true)
                {
                    /* Get the PID exit status to de-zombifie it.
                     * This should never block. */
                    if (waitpid (pid, status, 0) < 0)
                        switch (errno)
                        {
                            case EINTR:  /* The syscall was interrupted */
                                continue;

                            case ECHILD: /* This is weird, process is gone */
                            default:    
                                return -1;
                        }
                    return 1;
                }
                break;
            case EPERM: /* Not an echelog process */
            default:
                return -1;
        }
    return 0;
}

int remove_directory (const char *dir_name)
{
    DIR *dir_info;
    struct dirent *dir_entry;
    struct_stat stat_buf;

    ASSERT (dir_name);
    
    /* try to open directory information file */
    if (!(dir_info = opendir (dir_name)))
    {
        log (L_ERROR, "Failed to read directory info file for \"%s\": %m",
                dir_name);
        return -1;
    }    

    /* for each entry in directory */
    while ((dir_entry = readdir(dir_info)))
    {
        int len = strlen (dir_name) + strlen (dir_entry->d_name) + 2;
        char *filename;
    
        /* skip . and .. entries */
        if (!strcmp (".", dir_entry->d_name)) 
                continue;
        if (!strcmp ("..", dir_entry->d_name))                
                continue;
       
        /* create filename */
        filename = static_cast<char *> (xmalloc (len));
        memcpy (filename, dir_name, strlen (dir_name));
        memcpy (filename + strlen (dir_name) + 1, dir_entry->d_name, 
                strlen (dir_entry->d_name));
        filename [strlen (dir_name)] = '/'; filename [len - 1] = 0;
        
        /* determine file type and remove it */
        if (s_stat (filename, &stat_buf))
            log (L_ERROR, "Failed to stat %s: %m", filename);

        if (S_ISREG (stat_buf.st_mode) && unlink (filename))
            log (L_ERROR, "Failed to unlink %s: %m", filename);

        if (S_ISDIR(stat_buf.st_mode) && remove_directory (filename))
            log (L_ERROR, "Failed to unlink %s: %m", filename);

        xfree (filename);
    }
    
    /* try to close directory information file */
    if (closedir (dir_info))
    {
        log (L_ERROR, "Failed to close directory info filefor \"%s\": %m",
                dir_name);
        return -1;
    }

    /* remove the directory */
    if (rmdir (dir_name))
    {
        log (L_ERROR, "Failed to unlink %s: %m", dir_name);
        return -1;
    }    
    
    return 0;
     
}

int move_directory (const char *dir_name, const char *new_dir)
{
    ASSERT (dir_name);
    ASSERT (new_dir);

    /* move the directory */
    if (rename (dir_name, new_dir))
    {
        log (L_ERROR, "Failed to move dir %s to %s: %m", dir_name, new_dir);
        return -1;
    }

    return 0;

}

}
