/**
 * @brief Echelog Library -- String manipulation
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/lib.h"

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <ctype.h>

#define MAX_CONCAT 100


namespace lib {

size_t s_strnlen (const char *s, size_t maxlen)
{
    size_t len = 0;

    ASSERT (s);
    while (*s++ && len < maxlen)
        len++;

    return len;
}

char *s_strdup (const char *s)
{
    char *k;

    ASSERT (s);
    k = strdup (s);
    if (!k)
        die ("Insufficient memory available to allocate duplicate string");
    return k;
}


char *s_strdup (const char *s, tMemPool &pool)
{
    unsigned l;
    char *k;

    ASSERT (s);
    l = strlen(s);
    k = static_cast<char *>(pool.alloc(l + 1));
    memcpy(k, s, l+1);
    return k;
}


char *s_strdup_portion (const char *s, const long len)
{
    char *k;

    ASSERT (s);
    k = static_cast<char *> (xmalloc (len + 1));
    memcpy (k, s, len);
    k[len] = 0;

    return k;
}


char *s_strdup_portion (const char *s, const long len, tMemPool &pool)
{
    char *k;

    ASSERT (s);
    k = static_cast<char *> (pool.alloc (len + 1));
    memcpy (k, s, len);
    k[len] = 0;

    return k;
}


char *s_strndup (const char *s, const long maxlen)
{
    size_t size = s_strnlen (s, maxlen);

    return s_strdup_portion (s, size);
}


char *s_strndup (const char *s, const long maxlen, tMemPool &pool)
{
    size_t size = s_strnlen (s, maxlen);

    return s_strdup_portion (s, size, pool);
}


char *s_trim (const char *s, tMemPool &pool)
{
    const char *first, *last;

    if (!s) 
        return NULL;

    /* Skip leading blanks */
    while (*s != 0 && isspace (*s))
        s++;
    last = first = s;

    /* Find the last non-blank character */
    while (*s != 0)
        if (!isspace (*s++))
                last = s;

    return s_strdup_portion (first, last - first, pool);
}


char *s_trim (const char *s)
{
    const char *first, *last;

    if (!s) 
        return NULL;

    /* Skip leading blanks */
    while (*s != 0 && isspace (*s))
        s++;
    last = first = s;

    /* Find the last non-blank character */
    while (*s != 0)
        if (!isspace (*s++))
                last = s;

    return s_strdup_portion (first, last - first);
}


const char *s_strnstr(const char *s, size_t s_len, const char *p, size_t p_len)
{
    /* 
     * This is simple and inefficient but it may be easily replaced
     * it there's someone with an intention to use it for 
     * large strings (like mmapped files)
     */
    while (s_len >= p_len) {
        if (!strncmp(s, p, p_len))
            return s;
        s++, s_len--;
    }
    return NULL;
}


const char *s_strnstr (const char *s, const char *p, const size_t maxlen)
{
    size_t s_len = s_strnlen (s, maxlen);
    size_t p_len = s_strnlen (p, maxlen);

    return s_strnstr (s, s_len, p, p_len);
}


char *s_strnreplace (const char *str, const char *to_replace, 
        const char *replace_string, const long maxlen)
{
    ASSERT (to_replace);

    const char *s, *ls;
    char *ret, *r;
    long n, m;
    long tr_len = strlen (to_replace);

    if (!str)
        return NULL;

    /* If the string which is to be replaced is too long 
     * we may safely return a copy of the original string */
    if (maxlen <= 0 || tr_len >= maxlen)
        return s_strndup (str, maxlen);

    if (!replace_string)
        replace_string = "";

    long str_len = s_strnlen (str, maxlen);
    long r_len = strlen (replace_string); 
    long diff_len = r_len - tr_len;

    
    /* Find the number of substitution places */
    n = 0;
    m = str_len;
    ls = s = str;
    while ((s = s_strnstr (s, m, to_replace, tr_len)))
    {
        s += tr_len;
        m -= s - ls;
        ls = s;
        n++;
    }

    
    /* Alloc enough space */
    ret = static_cast<char *> (xmalloc (str_len + (n * diff_len) + 1));

    /* Copy and substitute */
    for (m = str_len, s = str, r = ret; n > 0; n--)
    {
        long diff;

        ls = s;
        s = s_strnstr (s, m, to_replace, tr_len);
        diff = s - ls;

        /* Copy next part up to matching pattern */
        memcpy (r, ls, diff);

        /* Substitute */
        r += diff;
        s += tr_len;
        memcpy (r, replace_string, r_len);
        r += r_len;

        /* Update bounds */
        m -= tr_len + diff;
    }
    
    /* Append the rest and dd the terminating null */
    memcpy (r, s, m);
    *(ret + (str_len) + (n * diff_len)) = 0;

    return ret;
}


char *s_strnreplace (const char *str, const char *to_replace, 
        const char *replace_string, const long maxlen, tMemPool &pool)
{
    ASSERT (to_replace);

    const char *s, *ls; 
    char *ret, *r;
    long n, m;
    long tr_len = strlen (to_replace);

    if (!str)
        return NULL;

    /* If the string which is to be replaced is too long 
     * we may safely return a copy of the original string */
    if (maxlen <= 0 || tr_len >= maxlen)
        return s_strndup (str, maxlen, pool);

    if (!replace_string)
        replace_string = "";

    long str_len = s_strnlen (str, maxlen);
    long r_len = strlen (replace_string); 
    long diff_len = r_len - tr_len;

    
    /* Find the number of substitution places */
    n = 0;
    m = str_len;
    ls = s = str;
    while ((s = s_strnstr (s, m, to_replace, tr_len)))
    {
        s += tr_len;
        m -= s - ls;
        ls = s;
        n++;
    }
    

    /* Alloc enough space */
    ret = static_cast<char *> (pool.alloc (str_len + (n * diff_len) + 1));

    /* Copy and substitute */
    for (m = str_len, s = str, r = ret; n > 0; n--)
    {
        long diff;

        ls = s;
        s = s_strnstr (s, m, to_replace, tr_len);
        diff = s - ls;

        /* Copy next part up to matching pattern */
        memcpy (r, ls, diff);

        /* Substitute */
        r += diff;
        s += tr_len;
        memcpy (r, replace_string, r_len);
        r += r_len;

        /* Update bounds */
        m -= tr_len + diff;
    }
    
    /* Append the rest and dd the terminating null */
    memcpy (r, s, m);
    *(ret + (str_len) + (n * diff_len)) = 0;

    return ret;
}


char *s_strconcat (const char *str1, ...)
{
    static long lengths[MAX_CONCAT];
    char *r, *s = NULL;
    va_list args, args2;
    long total, i, n = 1;

    if (!str1)
	return NULL;

    /* Count parameters and their lengths */
    total = lengths[0] = strlen (str1);
    va_start (args, str1);
    va_copy(args2, args);
    while ((s = va_arg (args, char *)) != NULL)
    {
    	ASSERT (n < MAX_CONCAT);
	total += lengths[n++] = strlen (s);
    }
    va_end (args);

    r = s = static_cast<char *> (xmalloc (total + 1));

    /* Concatenate them */
    memcpy (s, str1, lengths[0]);
    s += lengths[0];
    for (i = 1; i < n; i++)
    {
	memcpy (s, va_arg (args2, char *), lengths[i]);
	s += lengths[i];
    }
    /* And add a termination null */
    *s = 0;
    va_end (args2);

    return r;
}


char *s_strconcat (tMemPool &pool, const char *str1, ...)
{
    static long lengths[MAX_CONCAT];
    char *r, *s = NULL;
    va_list args, args2;
    long total, i, n = 1;

    if (!str1)
	return NULL;

    /* Count parameters and their lengths */
    total = lengths[0] = strlen (str1);
    va_start (args, str1);
    va_copy(args2, args);
    while ((s = va_arg (args, char *)) != NULL)
    {
    	ASSERT (n < MAX_CONCAT);
	total += lengths[n++] = strlen (s);
    }
    va_end (args);

    r = s = static_cast<char *> (pool.alloc (total + 1));

    /* Concatenate them */
    memcpy (s, str1, lengths[0]);
    s += lengths[0];
    for (i = 1; i < n; i++)
    {
	memcpy (s, va_arg (args2, char *), lengths[i]);
	s += lengths[i];
    }
    /* And add a termination null */
    *s = 0;
    va_end (args2);

    return r;
}


int s_strpsep (tList &list, const char *str)
{
    const char *beg, *s = str;

    /* For an empty string do nothing */
    if (!str)
        return 0;

    /* Count number of phrases and detect syntax errors */
    while (*s != 0)
    {
        /* Skip leading blanks */
        while (*s != 0 && isspace (*s))
            s++;

        /* Is it a single word or a phrase? */
        if (*s == '"')
        {
            /* Phrase */ 
            beg = ++s;

            /* Go over a phrase */
            while (*s != 0 && *s != '"')
            {
                if (*s == '\\' && *(s+1) == '"')
                        s++;    /* go over backslashed double quotes */
                s++;
            }

            /* Oops, no right quote */
            if (*s++ != '"')
                return -1;

            /* Add the phrase to the list */
            list.append (s_strnreplace (beg, "\\\"", "\"", (s - beg - 1)));
        }
        else if (*s != 0)
        {
            /* It's only a single word */
            beg = s++;

            while (*s != 0 && !isspace (*s) && *s != '"')
            {
                if (*s == '\\' && *(s+1) == '"')
                        s++;    /* go over backslashed double quotes */
                s++;
            }
            
            /* Add the phrase to the list */
            list.append (s_strnreplace (beg, "\\\"", "\"", (s - beg)));
        }
    }

    return 0;
}


int s_strpsep (tList &list, const char *str, tMemPool &pool)
{
    const char *beg, *s = str;

    /* For an empty string do nothing */
    if (!str)
        return 0;

    /* Count number of phrases and detect syntax errors */
    while (*s != 0)
    {
        /* Skip leading blanks */
        while (*s != 0 && isspace (*s))
            s++;

        /* Is it a single word or a phrase? */
        if (*s == '"')
        {
            /* Phrase */ 
            beg = ++s;

            /* Go over a phrase */
            while (*s != 0 && *s != '"')
            {
                if (*s == '\\' && *(s+1) == '"')
                        s++;    /* go over backslashed double quotes */
                s++;
            }

            /* Oops, no right quote */
            if (*s++ != '"')
                return -1;

            /* Add the phrase to the list */
            list.append (s_strnreplace (beg, "\\\"","\"", (s - beg - 1), pool));
        }
        else if (*s != 0)
        {
            /* It's only a single word */
            beg = s++;

            while (*s != 0 && !isspace (*s) && *s != '"')
            {
                if (*s == '\\' && *(s+1) == '"')
                        s++;    /* go over backslashed double quotes */
                s++;
            }
            
            /* Add the phrase to the list */
            list.append (s_strnreplace (beg, "\\\"", "\"", (s - beg), pool));
        }
    }

    return 0;
}


int s_strdsep (tList &list, const char *str, const char *delim, tMemPool &pool)
{
    ASSERT (delim);

    if (!str)
        return 0;

    const char *os, *s = str;
    long len;

    do 
    {
        os = s;
        len = strcspn (s, delim);
        s += len;
        list.append (s_strdup_portion (os, len, pool));
    }
    while (*s++);

    return 0;
}


int s_asprintf (char **ret, const char *format, ...)
{
    va_list args;
    int real_length;

    va_start(args, format);
    real_length = s_vasprintf (ret, format, args);
    va_end (args);

    return real_length;
}


int s_vasprintf (char **ret, const char *format, va_list ap)
{
    int real_length;
    size_t buf_len;
    char fake_string;

    /* We need a copy of ap since we can't traverse the same stack twice.
     * Each traversal needs to be backed by va_end.
     */
    va_list ap2;
    va_copy(ap2, ap);

    /* Most systems have a nice, sane and pretty buggy ;)
     * implementation of vsnprintf(). But if you keep the buffer non-NULL and
     * set its limit to (non-zero) positive value you can expect quite sane
     * results.
     *
     * (v)asprintf() seems to be only a bourgeois yield...
     */
    buf_len = vsnprintf (&fake_string, sizeof(fake_string), format, ap);
    *ret = static_cast<char *> (xmalloc (buf_len + 1));
    real_length = vsnprintf (*ret, buf_len + 1, format, ap2);
    va_end(ap2);
    if (real_length < 0)
        die ("Insufficient memory available to allocate a string");

    return real_length;
}


int s_asprintf (char **ret, tMemPool &pool, const char *format, ...)
{
    va_list args;
    int real_length;

    va_start(args, format);
    real_length = s_vasprintf (ret, pool, format, args);
    va_end (args);

    return real_length;
}


int s_vasprintf (char **ret, tMemPool &pool, const char *format, va_list ap)
{
    int real_length;
    size_t buf_len;
    char fake_string;

    /* We need a copy of ap since we can't traverse the same stack twice.
     * Each traversal needs to be backed by va_end.
     */
    va_list ap2;
    va_copy(ap2, ap);

    /* Most systems have a nice, sane and pretty buggy ;)
     * implementation of vsnprintf(). But if you keep the buffer non-NULL and
     * set its limit to (non-zero) positive value you can expect quite sane
     * results.
     *
     * (v)asprintf() seems to be only a bourgeois yield...
     */
    buf_len = vsnprintf (&fake_string, sizeof(fake_string), format, ap);
    *ret = static_cast<char *> (pool.alloc (buf_len + 1));
    real_length = vsnprintf (*ret, buf_len + 1, format, ap2);
    if (real_length < 0)
        die ("Insufficient memory available to allocate a string");

    return real_length;
}


int s_conv_num(const char **buf, int *dest, int llim, int ulim)
{
	int result = 0;

	/* The limit also determines the number of valid digits. */
	int rulim = ulim;

	if (**buf < '0' || **buf > '9')
		return (0);

	do {
		result *= 10;
		result += *(*buf)++ - '0';
		rulim /= 10;
	} 
        while ((result * 10 <= ulim) && rulim && **buf >= '0' && **buf <= '9');

	if (result < llim || result > ulim)
		return (0);

	*dest = result;
	return (1);
}


int s_stol (const char *buf, long *dest, long llim, long ulim)
{
    char *end, *nend;
    long n;

    /* Try to read it as a number */
    n = strtol (buf, &end, 0);
    if (!(n == LONG_MAX && errno == ERANGE))
    {
        /* Check the ending */
        for (nend = end; *nend; nend++)
            if (!isspace (*nend))
                return -1;
        
        /* Okay, it looks like a long int. Check the limits. */
        if (n < llim || n > ulim)
            return -1;

        /* Successfully converted */
        *dest = n;
        return 0;
    }

    /* Not a long integer */
    return -1;
}


}
