/**
 * @brief Echelog strptime() implementation borrowed from NetBSD
 *
 * This library converts time from string representation to
 * computer representation, does the opposite of strftime.
 * This works with time periods in opposite to ordinary strptime()
 *
 * @author Martin Macok <martin.macok@underground.cz>
 *
 * Based on strptime() borrowed from NetBSD 1.5 libc code (strptime.c,v 1.19),
 * which was originally written by Klaus Klein and contributed to NetBSD.
 *
 */

#include "lib/lib.h"

#ifdef HAVE_LANGINFO_H
#define NL_LANGINFO(a) nl_langinfo(a)
#else
// this is almost only cause of freebsd... could be workarounded later...
#define NL_LANGINFO(a) NULL
#endif

#include <locale.h>
#ifdef HAVE_LANGINFO_H
#include <langinfo.h>
#endif
#include <ctype.h>
#include <string.h>
#include <time.h>
//#include <tzfile.h>

#ifndef TM_YEAR_BASE
#define TM_YEAR_BASE 1900
#endif

/*
 * We do not implement alternate representations. However, we always
 * check whether a given modifier is allowed for a certain conversion.
 */
#define ALT_E			0x01
#define ALT_O			0x02
#define	LEGAL_ALT(x)		{ if (alt_format & ~(x)) return (0); }

using namespace lib;
 
namespace lib {

char *ech_strptime_p(const char *buf, const char *fmt, struct tm *tim)
{
	char c;
	const char *bp;
	int alt_format, i;
    
	bp = buf;

	while ((c = *fmt) != '\0') {
		/* Clear `alternate' modifier prior to new conversion. */
		alt_format = 0;

		/* Eat up white-space. */
		if (isspace(c)) {
			while (isspace(*bp))
				bp++;

			fmt++;
			continue;
		}
				
		if ((c = *fmt++) != '%')
			goto literal;


again:		switch (c = *fmt++) {
		case '%':	/* "%%" is converted to "%". */
literal:
			if (c != *bp++)
				return (0);
			break;

		/*
		 * "Alternative" modifiers. Just set the appropriate flag
		 * and start over again.
		 */
		case 'E':	/* "%E?" alternative conversion modifier. */
			LEGAL_ALT(0);
			alt_format |= ALT_E;
			goto again;

		case 'O':	/* "%O?" alternative conversion modifier. */
			LEGAL_ALT(0);
			alt_format |= ALT_O;
			goto again;
			
		/*
		 * "Complex" conversion rules, implemented through recursion.
		 */
		case 'R':	/* The time as "%H:%M". */
			LEGAL_ALT(0);
			if (!(bp = ech_strptime_p(bp, "%H:%M", tim)))
				return (0);
			break;

		case 'T':	/* The time as "%H:%M:%S". */
			LEGAL_ALT(0);
			if (!(bp = ech_strptime_p(bp, "%H:%M:%S", tim)))
				return (0);
			break;

		case 'd':	/* The days. */
		case 'e':
		case 'j':
			LEGAL_ALT(ALT_O);
			if (!(s_conv_num(&bp, &tim->tm_mday, 0, (INT_MAX-1)/86400)))
				return (0);
			break;

		case 'k':	/* The hours. */
			LEGAL_ALT(0);
			/* FALLTHROUGH */
		case 'H':
			LEGAL_ALT(ALT_O);
			if (!(s_conv_num(&bp, &tim->tm_hour, 0, (INT_MAX-1)/3600)))
				return (0);
			break;

		case 'M':	/* The minute. */
			LEGAL_ALT(ALT_O);
			if (!(s_conv_num(&bp, &tim->tm_min, 0, (INT_MAX-1)/60)))
				return (0);
			break;

		case 'S':	/* The seconds. */
			LEGAL_ALT(ALT_O);
			if (!(s_conv_num(&bp, &tim->tm_sec, 0, INT_MAX-1)))
				return (0);
			break;

		case '-':	/* 0 seconds (thx freebsd "w") */
			if (*bp++ != '-')
				return (0);
            tim->tm_sec = 0;
			break;

		case 'Y':	/* The year. */
		case 'y':
			LEGAL_ALT(ALT_E);
			if (!(s_conv_num(&bp, &i, 0, 100))) // reasonal?
				return (0);

			tim->tm_year = i;
			break;

		/*
		 * Miscellaneous conversions.
		 */
		case 'n':	/* Any kind of white-space. */
		case 't':
			LEGAL_ALT(0);
			while (isspace(*bp))
				bp++;
			break;

		default:	/* Unknown/unsupported conversion. */
			return (0);
		}


	}

	/* LINTED functional specification */
	return const_cast<char *>(bp);
} // end of ech_strptime_p()

} // end of namespace lib

