/**
 * @brief Echelog timestring and time conversion library
 *
 * This library converts time from string representation to
 * computer representation, does the opposite of strftime (like strptime).
 * It can handle multiple alternative format string representations and
 * it can guess the time from string without "hinting" format specification
 *
 * @author Martin Macok <martin.macok@underground.cz>
 *
 */

/*
 * TODO list:
 * 1) implement _guess() function, and make it better, faster and sweeter
 * 2) check how it handles locales and TZ env settings (maybe we should
 *    unset locale settings for echelog processes)
 * 3) test it tough and hard
 */

#include "lib/archdep.h"

#include <locale.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <limits.h>

#include "lib/strptime.h"
#include "lib/timestring.h"

#define Y_IS_LEAP(y) (!((((y)+300) % 400) % 4)) // OK for 1901-2099

using namespace lib;

namespace lib {

void reset_tm(struct tm *tim)
{
    tim->tm_sec  = TM_NOTSET;
    tim->tm_min  = TM_NOTSET;
    tim->tm_hour = TM_NOTSET;
    tim->tm_mday = TM_NOTSET;
    tim->tm_mon  = TM_NOTSET;
    tim->tm_year = TM_NOTSET;
    tim->tm_wday = TM_NOTSET;
    tim->tm_yday = TM_NOTSET;
    tim->tm_isdst = TM_NOTSET;
}

// this is really obvious, aren't you so simple you don't understand?
int day_of_week(const struct tm *tim)
{
    int d,m,c,y;
    if (tim->tm_mday==TM_NOTSET || tim->tm_mon==TM_NOTSET ||
            tim->tm_year==TM_NOTSET)
        return -1; // we need date to compute it, sweet
    d = tim->tm_mday;
    c = 19 + (tim->tm_year / 100);
    y = tim->tm_year % 100;
    m = tim->tm_mon - 1;
    if (m<1) {
        m += 12;
        y--;
        if (y<0) {
            y = 99;
            c--;
        }
    }
    return (((26*m - 2)/10 + d + y + y/4 + c/4 + 5*c) % 7); // 0=sunday
}


int days_in_month(int month, int year)
{
    switch(month) {
        case 0:
        case 2:
        case 4:
        case 6:
        case 7:
        case 9:
        case 11:
            return 31;
            break; // <- this is VERY VERY VERY important, you see
        case 1:
            return (28 + Y_IS_LEAP(year));
            break; // again! :)
        default:
            return 30;
            break;
    }
}


void correct_tm(struct tm *tim)
{
    while (tim->tm_mon < 0) {
       tim->tm_year--; 
       tim->tm_mon += 12;
    }

    while (tim->tm_mday < 0) {
        tim->tm_mon -= 1;
        while (tim->tm_mon < 0) {
           tim->tm_year--;
           tim->tm_mon += 12; 
        }
        tim->tm_mday += days_in_month(tim->tm_mon, tim->tm_year);
    }

    // and the reverse order
    while (tim->tm_mon > 11) {
       tim->tm_year++;
       tim->tm_mon -= 12;
    }
    
    while (tim->tm_mday > days_in_month(tim->tm_mon, tim->tm_year)) {
        tim->tm_mon += 1;
        while (tim->tm_mon > 11) {
           tim->tm_year++;
           tim->tm_mon -= 12; 
        }
        tim->tm_mday -= days_in_month(tim->tm_mon, tim->tm_year);
    }
    
    // congratulation, you have reached the very end of the world!
}


char *str2timef(const char *buf, char **fmt, struct tm *tim, int guess)
{
    char *done;
    struct tm temp_tm;

    temp_tm = *tim; // backup

    if(fmt==NULL && guess>0)
        return str2timef_guess(buf, tim, guess);

    // try all format strings
    while(*fmt!=NULL) {
        done = strptime(buf,*fmt,tim);
        if (done!=NULL)
            return (done);
        fmt++;
        *tim = temp_tm; // rollback
    }

    // format strings failed, try guess function
    if (guess > 0)
        return str2timef_guess(buf, tim, guess);

    // DONT DO THIS ->    
    // TODO if one of day/month/year was not set, clear wday (NO!!!!)
    // looks like it's not really needed, but some implementations (strptime in
    // glibc for example) set this even if only one of d/m/y is read and
    // this workaround makes sure wday isn't abused...
    //
    // MAYBE this is the reason to use our strptime (from netbsd)
    // but
    // 1) nl_langinfo() is not in freebsd
    // 2) when we don't know mday nor wday, we are clueless anyway
    //    (see str2timef_rel())
    //
    //    looks like workarounds and test conditions in
    //    str2timef_rel() should work around this should be sufficient
    
	return NULL; // we completely failed, try crystal ball
} // end of str2timef()


int count_tm_p(struct tm *tim)
{
    int res=0;

    if(tim->tm_year == tim->tm_mday == tim->tm_hour == tim->tm_min
            == tim->tm_sec == TM_NOTSET)
        return -1;              // no value

    if(tim->tm_year!=TM_NOTSET)
        res += tim->tm_year;
    res *= 365;
    if(tim->tm_year!=TM_NOTSET)
        res += tim->tm_year / 4; // leap years

    if(tim->tm_mday!=TM_NOTSET)
        res += tim->tm_mday;
    res *= 24;

    if(tim->tm_hour!=TM_NOTSET)
        res += tim->tm_hour;
    res *= 60;

    if(tim->tm_min!=TM_NOTSET)
        res += tim->tm_min;
    res *= 60;

    if(tim->tm_sec!=TM_NOTSET)
        res += tim->tm_sec;

    return res;
}


int str2timef_p(const char *buf, char **fmt, struct tm *tim, int guess)
{
    char *done;
    struct tm temp_tm;

    temp_tm = *tim; // backup

    // try all format strings
    while(*fmt!=NULL) {
        done = ech_strptime_p(buf,*fmt,tim);
        if (done!=NULL)
            return (count_tm_p(tim));
        fmt++;
        *tim = temp_tm; // rollback
    }

	return -1; // we completely failed, try crystal ball
} // end of str2timef()


// unused (unset) fields of "tm" need to be reseted (see reset_tm())
char *str2timef_rel(const char *buf, char **fmt, const struct tm *rel_tm,
        int rel, struct tm *tim, int guess)
{
    char *ret;
    int tm_weekday, mday_was_set=1;

    if (rel!=-1 && rel!=1)
        return NULL;    // bad argument;

    // check if rel_tm is complete
    if (rel_tm==NULL || rel_tm->tm_mday==TM_NOTSET || \
            rel_tm->tm_mon==TM_NOTSET || rel_tm->tm_year==TM_NOTSET)
        return NULL;    // bad argument;

    ret = str2timef(buf, fmt, tim, guess);
    if (ret==NULL)
        return NULL;
   
    // if we don't know year yet, copy it from rel_tm
    if (tim->tm_year==TM_NOTSET)
        tim->tm_year = rel_tm->tm_year;
   
    // if we don't know month yet, copy it from rel_tm
    if (tim->tm_mon==TM_NOTSET)
        tim->tm_mon = rel_tm->tm_mon;
   
    // if we don't know mday yet, copy it from rel_tm
    if (tim->tm_mday==TM_NOTSET) {
        tim->tm_mday = rel_tm->tm_mday;
        mday_was_set = 0;
    }
    
    // jestlize jsme se dozvedeli tm->tm_wday (a NE tm_mday), tak poopravime
    // pomoci rel_tm->(datum) (tam eventualne wday dopocitame)
    if ((!mday_was_set) && tim->tm_wday!=TM_NOTSET) {
        tm_weekday = day_of_week(tim); // the wday from rel_tm
        if (tm_weekday==-1)
            return NULL;    // rel_tm doesn't contain all relevant info!
        tim->tm_mday += rel*((tm_weekday - tim->tm_wday) % 7);
        correct_tm(tim); // things like negative mday etc...
    }

    // hours, minutes, seconds - reset to zero if not specified
    if (tim->tm_sec==TM_NOTSET)
        tim->tm_sec = 0;
    else { // seconds were set, copy minutes and hours (if not set too)
        if (tim->tm_min==TM_NOTSET)
            tim->tm_min = rel_tm->tm_min;
        if (tim->tm_hour==TM_NOTSET)
            tim->tm_hour = rel_tm->tm_hour;
    }
        
    if (tim->tm_min==TM_NOTSET)
        tim->tm_min = 0;
    else // minutes were set, copy hours (if not set too)
        if (tim->tm_hour==TM_NOTSET)
            tim->tm_hour = rel_tm->tm_hour;

    if (tim->tm_hour==TM_NOTSET)
        tim->tm_hour = 0;

    // what to do with DST? this shouldn't hurt:
    if (tim->tm_isdst==TM_NOTSET)
        tim->tm_isdst = rel_tm->tm_isdst;

    return ret;
}


char *str2timef_guess(const char *buf, struct tm *tim, int guess_hint)
{
    // TODO: plz implement me, I will be your baby tonight
    return NULL; // we're really smart now
}

} // end of namespace lib
