/**
 * @brief Echelog Library -- XML supplementary functions
 *
 * This part simplifies the manipulation with XML snippets.
 * 
 * @author Kamil Toman <ktoman@email.cz>
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include "lib/lib.h"

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>


namespace lib {

/// Char codes to be translated to xml entities
struct tTranslation {
    const char code;                 ///< Char code
    const char * const xml;          ///< XML entity for the code
    const int  size_diff;            ///< Subst size difference
};


/// Char codes to XML translation table
static tTranslation trans_table[] = 
{
    { ' ',    "&nbsp;", 5 },
    { '\xa4', "&euro;", 5 },
    { '\x3c', "&lt;", 3   },
    { '\x3e', "&gt;", 3   },
    { '\x26', "&amp;", 4  },
    { '\x22', "&quot;", 5 },
    { '\x27', "&apos;", 5 }
};


/// Number of items int the translation table
static const long trans_table_size = ARRAY_SIZE (trans_table);


char * xmlentities (const char *str, bool replace_spaces)
{
    long i, j, len = 1, new_len = 1, start = 1;
    char *ret, *r;
    const char *s;

    /* Return null for empty strings */
    if (!str)
        return NULL;

    /* '&nbsp;' is the first hardwired code */
    if (replace_spaces)
        start = 0;

    /* Calculate the size of the XMLized string */
    for (s = str; *s; s++)
    {
        for (i = start; i < trans_table_size; i++)
            if (*s == trans_table[i].code)
            {
                new_len += trans_table[i].size_diff;
                break;
            }
        len++, new_len++;
    }

    ret = static_cast<char *> (xmalloc (new_len));

    /* If the size of both versions is the same just return a copy */
    if (new_len == len)
    {
        memcpy (ret, str, len);   
        return ret;
    }

    /* Replace each relevant code with its xml entity */
    for (s = str, r = ret; *s; s++)
    {
        tTranslation *t = NULL;
        const char *c;

        for (i = start; i < trans_table_size; i++)
            if (*s == trans_table[i].code)
            {
                t = &trans_table[i];
                break;
            }

        if (t)
            /* We found a replacement, substitute the char code */
            for (j = 0, c = t->xml; j <= t->size_diff; j++)
                *r++ = *c++;
        else
            /* Just copy the char code */
            *r++ = *s;
    }

    /* Add the termination null */
    *r = '\0';

    return ret;
}


char * xmlentities (const char *str, tMemPool &pool, 
        bool replace_spaces)
{
    long i, j, len = 1, new_len = 1, start = 1;
    char *ret, *r;
    const char *s;

    /* Return null for empty strings */
    if (!str)
        return NULL;

    /* '&nbsp;' is the first hardwired code */
    if (replace_spaces)
        start = 0;

    /* Calculate the size of the XMLized string */
    for (s = str; *s; s++)
    {
        for (i = start; i < trans_table_size; i++)
            if (*s == trans_table[i].code)
            {
                new_len += trans_table[i].size_diff;
                break;
            }
        len++, new_len++;
    }

    ret = static_cast<char *> (pool.alloc (new_len));

    /* If the size of both versions is the same just return a copy */
    if (new_len == len)
    {
        memcpy (ret, str, len);   
        return ret;
    }

    /* Replace each relevant code with its xml entity */
    for (s = str, r = ret; *s; s++)
    {
        tTranslation *t = NULL;
        const char *c;

        for (i = start; i < trans_table_size; i++)
            if (*s == trans_table[i].code)
            {
                t = &trans_table[i];
                break;
            }

        if (t)
            /* We found a replacement, substitute the char code */
            for (j = 0, c = t->xml; j <= t->size_diff; j++)
                *r++ = *c++;
        else
            /* Just copy the char code */
            *r++ = *s;
    }

    /* Add the termination null */
    *r = '\0';

    return ret;
}


int tXMLOutput::print_data(const char *str)
{
    long i, j, len = 1, new_len = 1;
    char *tmpstr, *r;
    const char *s;

    /* fallout for empty strings */
    ASSERT (str);

    /* Calculate the size of the XMLized string */
    for (s = str; *s; s++)
    {
        for (i = 1; i < trans_table_size; i++)
            if (*s == trans_table[i].code)
            {
                new_len += trans_table[i].size_diff;
                break;
            }
        len++, new_len++;
    }

    /* alloc space for output string */
    tmpstr = static_cast<char *> (xmalloc (new_len));

    /* Replace each relevant code with its xml entity */
    for (s = str, r = tmpstr; *s; s++)
    {
        tTranslation *t = NULL;
        const char *c;

        for (i = 1; i < trans_table_size; i++)
            if (*s == trans_table[i].code)
            {
                t = &trans_table[i];
                break;
            }

        if (t)
            /* We found a replacement, substitute the char code */
            for (j = 0, c = t->xml; j <= t->size_diff; j++)
                *r++ = *c++;
        else
            /* Just copy the char code */
            *r++ = *s;
    }

    /* Add the termination null */
    *r = '\0';

    /* And write the string to output file */
    if (s_write (get_fd (), tmpstr, new_len - 1) < 0)
    {
        xfree (tmpstr);
        if (die_on_err) 
            die("Error writing XML output.");  
        else
            return -1;
    }

    /* Success */
    xfree (tmpstr);
    return 0;
}


int tXMLOutput::print_start_tag(const char *tagname)
{
    int err = 0;

    /* write indentation */
    err = err || (s_write (get_fd (), indstr, cur_indent) < 0);

    /* write beginning '<' */
    err = err || (s_write (get_fd (), "<", 1 ) < 0);

    /* write tagname */
    err = err || (s_write (get_fd (), tagname, strlen (tagname)) < 0);

    /* write ending '>' */
    err = err || (s_write (get_fd (), ">\n", 2 ) < 0);

    /* change indentation */
    set_cur_indent (get_cur_indent () + get_indent_step ());

    /* handle errors */
    if (err && die_on_err)
        die("Error writing XML output.");

    return err;
}


int tXMLOutput::print_end_tag(const char *tagname)
{
    int err = 0;

    /* change indentation */
    set_cur_indent (get_cur_indent () - get_indent_step ());

    /* write indentation */
    err = err || (s_write (get_fd (), indstr, cur_indent) < 0);

    /* write beginning '<' */
    err = err || (s_write (get_fd (), "</", 2 ) < 0);

    /* write tagname */
    err = err || (s_write (get_fd (), tagname, strlen (tagname)) < 0);

    /* write ending '>' */
    err = err || (s_write (get_fd (), ">\n", 2 ) < 0);

    /* handle errors */
    if (err && die_on_err)
        die("Error writing XML output.");

    return err;
}


int tXMLOutput::print_tag(const char *tagname, const char *tagvalue)
{
    int err = 0;

    /* write indentation */
    err = err || (s_write (get_fd (), indstr, cur_indent) < 0);

    /* write beginning tag */
    err = err || (s_write (get_fd (), "<", 1 ) < 0);
    err = err || (s_write (get_fd (), tagname, strlen (tagname)) < 0);
    err = err || (s_write (get_fd (), ">", 1 ) < 0);

    /*write xmlized value */
    err = err || (print_data(tagvalue) < 0);

    /* write ending tag */
    err = err || (s_write (get_fd (), "</", 2 ) < 0);
    err = err || (s_write (get_fd (), tagname, strlen (tagname)) < 0);
    err = err || (s_write (get_fd (), ">\n", 2 ) < 0);

    /* handle errors */
    if (err && die_on_err)
        die("Error writing XML output.");

    return err;

}


int tXMLOutput::print_header(const char *stylesheet)
{
    int err = 0;
    static char *xml_version = "<?xml version=\"1.0\"?>\n";
    static char *xml_slt_beg = "<?xml-stylesheet type=\"text/xsl\" href=\"";
    static char *xml_slt_end = "\"?>\n";

    /* write xml version clause */
    err = err || (s_write (get_fd (), xml_version, strlen (xml_version)) < 0);

    /* if requested print stylesheet clause */
    if (stylesheet)
    {
        /* print the stylesheet clause */
        err = err || (s_write (get_fd (), xml_slt_beg, strlen (xml_slt_beg))<0);
        err = err || (s_write (get_fd (), stylesheet, strlen (stylesheet)) < 0);
        err = err || (s_write (get_fd (), xml_slt_end, strlen (xml_slt_end))<0);
    }

    /* handle errors */
    if (err && die_on_err)
        die("Error writing XML output.");

    return err;
    
}


const long tXMLOutput::get_cur_indent () const
{   
    return cur_indent;
}


void tXMLOutput::set_cur_indent (long prop_value)
{
    ASSERT (prop_value >= 0);

    /* relaese old indentation string */
    if (indstr) 
        xfree (indstr);

    /* alloc new space for indentation string */
    indstr = static_cast<char *> (xmalloc (prop_value + 1));

    /* fill in spaces */
    memset (indstr, ' ', prop_value);
    indstr[prop_value] = 0;

    /* change current indentation */
    cur_indent = prop_value;
}

}
