/**
 * @file
 * @brief Contains implementation of tMagent class
 *
 * This class implements a typical miniagent's functionality.
 * The agent periodically reads informational chunks
 * from a supplied parser and forms them into messages which 
 * are passed to the next module. It also covers the basic miniagent's
 * configuration and reconfiguration.
 *
 * @author Kamil Toman <ktoman@email.cz>, Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "mframes/unixsocketioframe.h"
#include "magents/magent.h"
#include <signal.h>
#include <time.h>
#include <sys/time.h>


namespace magents {

using namespace lib;
using namespace mframes;

//------------------------
// Static data definition
//------------------------

/* Internal variables used for configuration */
long tMagent::s_run_min = 0;
long tMagent::s_run_hour = 0;
long tMagent::s_run_sec = 0;
char* tMagent::s_inst_name = NULL;


/* Common miniagents' configuration structure */
lib::cfitem tMagent::s_magentcommon_section[]=
{
    {
        name: "MagentCommon",
        type: CT_SECTION,
        data: NULL
    },
    {
        name: "RunIntervalSec",
        type: CT_INT,
        data: &tMagent::s_run_sec,
        varf: NULL,
        varf_reset: NULL,
        flags: ITEM_REQUIRED
    },
    {
        name: "RunIntervalMin",
        type: CT_INT,
        data: &tMagent::s_run_min,
        varf: NULL,
        varf_reset: NULL,
        flags: ITEM_REQUIRED
    },
    {
        name: "RunIntervalHour",
        type: CT_INT,
        data: &tMagent::s_run_hour,
        varf: NULL,
        varf_reset: NULL,
        flags: ITEM_REQUIRED
    },
    {
        name: "InstanceName",
        type: CT_STRING,
        data: &tMagent::s_inst_name,
        varf: NULL,
        varf_reset: NULL,
        flags: ITEM_REQUIRED
    },
    {
        name: NULL,
        type: CT_STOP,
        data: NULL
    }
};

/* Signal states */
volatile bool tMagent::s_sigalrm_received = false;
volatile bool tMagent::s_sigpipe_received = false;


/* Create a miniagent */
tMagent::tMagent (tCmdLine& cmdline, const char* cfg_filename)
    : tEchelogModule (cmdline, cfg_filename)
{
    m_run_interval = 0;
    m_inited = false;
    mp_mf = NULL;
}


bool tMagent::init (const char *id)
{
    ASSERT (!m_inited);

    mp_mf = new tMsgFrame (id);
    if (mp_mf == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        goto fail;
    }

    /* check if message frame is OK */
    if (mp_mf->is_broken ())
        goto mf_clean;

    if (tEchelogModule::init (mp_mf, NULL, mp_mf))
        goto mf_clean;

    if (mp_mf->init_alert ())
        goto anc_clean;

    m_inited = true;

    return false;

anc_clean:
    tEchelogModule::cleanup ();
mf_clean:
    delete mp_mf;
    mp_mf = NULL;
fail:
    return true;
}


void tMagent::cleanup ()
{
    ASSERT (m_inited);

    mp_mf->cleanup_alert ();
    tEchelogModule::cleanup ();
    delete mp_mf;
    mp_mf = NULL;

    m_inited = false;
}


bool tMagent::dynamic_init ()
{
    if (tEchelogModule::dynamic_init ())
        goto fail;

    m_run_interval = s_run_sec + s_run_min * 60 + s_run_hour * 3600;
    debug_log ('c', "Setting running interval to %li seconds",
        m_run_interval);
    /* Set a new instance name */
    mp_mf->set_inst_name (s_inst_name);

    return false;

fail:
    return true;
}


void tMagent::dynamic_cleanup ()
{
    m_run_interval = 0;
    mp_mf->set_inst_name (NULL);
    tEchelogModule::dynamic_cleanup ();
}


bool tMagent::register_sections (lib::tMemPool& trash_pool)
{
    if (tEchelogModule::register_sections (trash_pool))
        goto fail;

    cf_register (s_magentcommon_section);
    return false;

fail:
    return true;
}


void tMagent::unregister_sections ()
{
    cf_unregister (s_magentcommon_section[0].name);
    tEchelogModule::unregister_sections ();
}


bool tMagent::process_config (lib::tMemPool& conf_mp, lib::tMemPool& trash_pool)
{
    /* call ancestor's processing */
    if (tEchelogModule::process_config (conf_mp, trash_pool))
        goto fail;

    if (s_run_sec < 0)
    {
        log (L_ERROR, "Error in configuration file. "
             "RunIntervalSec can't be less than zero.");
        goto anc_clean;
    }

    if (s_run_min < 0)
    {
        log (L_ERROR, "Error in configuration file. "
             "RunIntervalMin can't be less than zero.");
        goto anc_clean;
    }

    if (s_run_hour < 0 || s_run_hour > 8800)
    {
        log (L_ERROR, "Error in configuration file. "
             "RunIntervalHour can't be less than zero or greater to 8800.");
        goto anc_clean;
    }

    if (!s_inst_name || strlen (s_inst_name) != ID_LENGTH)
    {
        log (L_ERROR, "Agent instance name should consist only "
            "of %d characters.", ID_LENGTH);
        goto anc_clean;
    }

    /* duplicate the strings to conf_mp*/
    s_inst_name = s_strdup (s_inst_name, conf_mp);

    return false;

anc_clean:
    tEchelogModule::cleanup_config ();
fail:
    return true;
}


void tMagent::cleanup_config ()
{
    s_inst_name = NULL;
    s_run_min = 0;
    s_run_hour = 0;
    s_run_sec = 0;

    /* call ancestor's cleanup */
    tEchelogModule::cleanup_config ();
}


void tMagent::install_sig_handlers ()
{
    tReconfigurableModule::install_sig_handlers ();

    set_signal_handler (SIGALRM, sig_handling_f);
    set_signal_handler (SIGPIPE, sig_handling_f);
}


/* A trivial signal transfering function */
void tMagent::sig_handling_f (int signum)
{
    switch (signum)
    {
        case SIGALRM:
            s_sigalrm_received = true;
            break;
        case SIGPIPE:
            s_sigpipe_received = true;
            break;
    }
}


tModule::SigHandlingRet tMagent::handle_signals ()
{
    SigHandlingRet ret = tEchelogModule::handle_signals ();

    if (ret == SHR_BREAK_OK || ret == SHR_BREAK_ERR)
        return ret;

    if (s_sigpipe_received)
        ret = on_sigpipe ();

    return ret;
}


tModule::SigHandlingRet tMagent::on_sigpipe ()
{
    log (L_FATAL, "Broken output channel. Can't send further "
            "information, aborting.");
    s_sigpipe_received = false;

    return SHR_BREAK_OK;
}


/* redefine standard SIGHUP hander to do nothin' cause we will handle
   SIGHUP specially */
tModule::SigHandlingRet tMagent::on_sighup ()
{
    return SHR_CONTINUE;
}


int tMagent::process ()
{
    ASSERT (m_inited);

    static time_t my_otime;
    static long my_rtime;
    static bool on_time;

    /* Get start time */
    my_otime = time(NULL);

    /* Set initial alarm */
    alarm (m_run_interval);

    /* Do what you can */
    if (process_messages (on_time))
        return PR_BREAK_ERR;

    /* Do we have some time left? */
    my_rtime = m_run_interval - (time(NULL) - my_otime);

sleep_loop:

    if (on_time && my_rtime > 0) /* Okay, sleep for the rest
                                    of the given interval */
    {
        /* If there's time left set alarm value more precisely */
        alarm (my_rtime);
        mp_mf->process (my_rtime);
    }

    /* if aborting signal received let's handle it */
    if (sigpipe_received () ||
        sigquit_received () ||
        sigint_received () ||
        sockerr_received ())
    {
        /* Cancel all alarms (if scheduled) */
        alarm (0);
        return PR_CONTINUE;
    }

    /* if sighup received do the reload */
    if (sighup_received ())
    {
        if (tReconfigurableModule::on_sighup () != SHR_CONTINUE)
            return PR_BREAK_ERR;

        /* reconfigured properly, cancel sleeping */
        s_sigalrm_received = true;
    }

    /* We don't want to wake up on an unknown signal */
    if (!sigalrm_received () && my_rtime > 0)
    {
        my_rtime = m_run_interval - (time(NULL) - my_otime);
        goto sleep_loop;
    }
    else
        s_sigalrm_received = false;

    return PR_CONTINUE;
}

} //end of magents namespace
