/**
 * @file
 * @brief Echelog miniagent support library
 *
 * This class covers the message handling and a communication
 * with next echelog modules. It's primarily used by tMagent
 * for a low-level operations.
 *
 * @author Kamil Toman <ktoman@email.cz>, Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "messages/msgmgr.h"
#include "magents/msgframe.h"
#include <netdb.h>


namespace magents {

using namespace lib;
using namespace messages;
using namespace mframes;


/* Set a new instance name */
void tMsgFrame::set_inst_name (char *str)
{
    inst_name = str;
}


/* Get the instance name */
const char *tMsgFrame::get_inst_name () const
{
    return inst_name;
}


/** Set up a message frame and fill in basic information */
tMsgFrame::tMsgFrame (const char *magent_id)
    :  m_pool (4096), num_of_records (0),
    incremental_message(false), inst_name (NULL)
{
    char my_hostname[1024];
    struct hostent *host_entry;
    tMsgHeader *header;

    m_broken = false;

    /* Get host information */
    if (gethostname (my_hostname, 1024))
    {
        log (L_ERROR, "Can't get hostname of the current processor.");
        m_broken = true;
        return;
    }
    if ((host_entry = gethostbyname (my_hostname)) == NULL)
    {
        log (L_ERROR, "Can't get ip-address of the current processor.");
        m_broken = true;
        return;
    }

    /* Set up a message */
    if (message.prepare_to_write (message_buffer, MAX_MESSAGE_SIZE))
    {
    	log (L_ERROR, "The message failed to prepare for writing.");
    	m_broken = true;
        return;
    }

    if ((empty_message_length = message.get_store_length()) <= 0)
    {
        log (L_ERROR, "General message buffer error.");
        m_broken = true;
        return;
    }
    length = empty_message_length;
    
    /* Fill in some basic header values */
    header = message.get_header ();
    header->set_agent_id (const_cast<char *> (magent_id));
    header->set_host_id (host_entry->h_addr_list[0]);

    alert_chunk = NULL;
}


bool tMsgFrame::init_alert ()
{
    alert_chunk = new tMsgChunk (SIMPLE_LOG_MSG, m_pool);
    if (alert_chunk == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        goto fail;
    }

    /* Alert chunk setup */
    alert_chunk->set_status (ST_ITEM);

    return false;

fail:
    return true;
}


void tMsgFrame::cleanup_alert ()
{
    delete alert_chunk;
    alert_chunk = NULL;
}


/** Finish the message and send it to output */
bool tMsgFrame::send_message (long msg_type)
{
    tMsgHeader *header = message.get_header ();
    int handle = gb_message_manager.get_handle_by_type (msg_type);

    /* Set type of the message */
    header->set_msg_type (msg_type);
    header->set_agent_instance_id (inst_name);
    header->set_flag_incremental (incremental_message);
    header->set_flag_aggregable (gb_message_manager.is_aggregable (handle));

    /* Place a time stamp and finish the message */
    header->set_subrecords (num_of_records);
    header->set_msg_time (time (NULL));
    if (message.close ())
    {
        log (L_ERROR, "Can't store a message to message buffer, "
            "therefore can't send the message.");
        goto fail;
    }

    debug_log ('m', "Sending a message");
    ASSERT (m_output_count > 0);

    /* Write the message to the output socket */
    if (m_outputs[0]->send_buffer (message_buffer, length))
    {
        log (L_ERROR, "Can't write message to output unix socket.");
        goto fail;
    }

    /* Process IO on sockets */
    if (process (DEFAULT_RCV_CTRL_PACKET_TIMEOUT))
    {
        log (L_ERROR, "Error while processing I/O, aborting.");
        goto fail;
    }

    /* Reset the frame */
    reset ();
    return false;

fail:
    reset ();
    return true;
}


/** Send an alert */
bool tMsgFrame::send_alert (char *alert_string)
{
    using namespace lib;
    using namespace messages;

    static tVarchar *msg = static_cast<tVarchar *>
        (alert_chunk->get_attr("LINE"));

    /* make alert */
    log (L_WARN, "securityalert: %s", alert_string);
    num_of_records = 0;
    msg->set_direct_value (alert_string);

    if (add_chunk (*alert_chunk))
        goto fail;

    if (send_message (alert_chunk->get_type ()))
        goto fail;

    return false;

fail:
    return true;
}

}
