/**
 * @file
 * @brief Contains implementation of tParsingMagent class
 *
 * This class implements a typical miniagent's functionality.
 * The agent periodically reads informational chunks
 * from a supplied parser and forms them into messages which 
 * are passed to the next module. It also covers the basic miniagent's
 * configuration and reconfiguration.
 *
 * @author Kamil Toman <ktoman@email.cz>, Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "magents/parsingmagent.h"
#include "mframes/unixsocketioframe.h"
#include <signal.h>
#include <time.h>
#include <sys/time.h>

namespace magents {

using namespace lib;
using namespace mframes;


bool tParsingMagent::static_init ()
{
    if (tMagent::static_init ())
        goto fail;

    if (m_parser.init ())
        goto anc_clean;

    return false;

anc_clean:
    tMagent::static_cleanup ();
fail:
    return true;
}


void tParsingMagent::static_cleanup ()
{
    m_parser.cleanup ();
    tMagent::static_cleanup ();
}


bool tParsingMagent::register_sections (lib::tMemPool& trash_pool)
{
    if (tMagent::register_sections (trash_pool))
        goto fail;

    if (m_parser.register_sections (trash_pool))
    {
        log (L_ERROR, "Error registering parser's configuration sections.");
        goto reg_clean;
    }

    return false;

reg_clean:
    tMagent::unregister_sections ();
fail:
    return true;
}


void tParsingMagent::unregister_sections ()
{
    m_parser.unregister_sections ();
    tMagent::unregister_sections ();
}


bool tParsingMagent::process_config (lib::tMemPool& conf_mp, 
        lib::tMemPool& trash_pool)
{
    /* call ancestor's processing */
    if (tMagent::process_config (conf_mp, trash_pool))
        goto fail;

    if (m_parser.process_config (conf_mp, trash_pool))
    {
        log (L_ERROR, "Error processing parser's configuration.");
        goto anc_clean;
    }

    return false;

anc_clean:
    tMagent::cleanup_config ();
fail:
    return true;

}


void tParsingMagent::cleanup_config ()
{
    m_parser.cleanup_config ();

    /* call ancestor's cleanup */
    tMagent::cleanup_config ();
}


/* Append a chunk to current message or send the current and
 * create a new one (if neccessary) */
bool tParsingMagent::handle_chunk (messages::tMsgChunk &chunk, long &prev_type,
    bool info)
{
    if (info)
    {
        /* If the new chunk won't fit in current message
         * or if it is incompatible. */
        if (prev_type != -1 && (chunk.get_type() != prev_type ||
                MAX_MESSAGE_SIZE <= (chunk.get_store_length ()
                    + mp_mf->get_store_length()) ||
                (chunk.get_status() != messages::ST_ITEM)
                    != mp_mf->get_incremental_message()))
        {
            /* Send it and prepare a new one */
            if (mp_mf->send_message (prev_type))
                goto fail;
            if (mp_mf->add_chunk (chunk))
                goto fail;
            prev_type = chunk.get_type();
            m_parser.parse_cleanup ();
        }
        else
        {
            /* Just add a new message chunk */
            if (mp_mf->add_chunk (chunk))
                goto fail;
            prev_type = chunk.get_type ();
        }
    }
    else
    {
        /* Send previous (if any) */
        if (mp_mf->get_num_of_records() > 0)
            if (mp_mf->send_message (prev_type))
                goto fail;

        /* Send an empty message */
        mp_mf->set_incremental_message (chunk.get_status() !=
            messages::ST_ITEM);
        if (mp_mf->send_message (chunk.get_type ()))
            goto fail;
    }

    return false;

fail:
    return true;
}


/* Form and send messages */
bool tParsingMagent::process_messages (bool& on_time)
{
    using namespace messages;

    long prev_type = -1;
    bool portion_finished = false;
    tParserResponse r;

    on_time = true;

    debug_log ('m', "Preparing messages");
    if (m_parser.start ())
    {
        /* Can't init parser */
        static char *err = "Can't parse source information.";
        log (L_ERROR, "%s", err);
        if (mp_mf->send_alert (err))
            goto stop_clean;
    }
    else /* Parser init ok */
    {
        while (!portion_finished)
        {
            /* Get another informitation chunk */
            m_parser.get_chunk (r);

            switch (r.state)
            {
                case PR_OK:
                    if (handle_chunk (*r.chunk, prev_type))
                        goto stop_clean;
                    break;

                case PR_EMPTY:
                    if (handle_chunk (*r.chunk, prev_type, false))
                        goto stop_clean;
                    break;

                case PR_END:
                    /* We're done for now */
                    portion_finished = true;
                    break;

                case PR_ERROR:
                    /* Send what you can and report an error */
                    if (mp_mf->get_num_of_records() > 0)
                        if (mp_mf->send_message (prev_type))
                            goto stop_clean;

                    /* Send an alert */
                    if (mp_mf->send_alert ("Can't parse source information."))
                        goto stop_clean;

                    m_parser.parse_cleanup ();

                    /* Send more detailed information about the error */
                    while (m_parser.get_error_info (r), r.state != PR_END)
                    {
                        if (r.state != PR_OK)
                        {
                            log (L_ERROR, "Error reporting chunks corrupted.");
                            goto stop_clean;
                        }

                        if (handle_chunk (*r.chunk, prev_type))
                            goto stop_clean;
                    }
                    m_parser.parse_cleanup ();
                    break;

                default:
                    /* This should never happen */
                    log (L_ERROR, "Unknown response from parser.");
                    goto stop_clean;
            }

            /* Alarm too early? */
            if (s_sigalrm_received)
            {
                /* We're not on time, force sending a message */
                if (mp_mf->get_num_of_records() > 0)
                    if (mp_mf->send_message (prev_type))
                        goto stop_clean;

                /* And report that magent is overworked */
                if (mp_mf->send_alert ("Miniagent overworked."))
                    goto stop_clean;

                /* Return to schedule routine */
                m_parser.parse_cleanup ();
                on_time = false;
                break;
            }
            if (sighup_received ())
                log(L_INFO, "Scheduling configuration reload.");
            if (sigquit_received () ||
                sigint_received () ||
                sockerr_received () ||
                sigpipe_received ())
            {
                m_parser.parse_cleanup ();
                on_time = false;      /* not finished on time */
                break;
            }
        }

        /* Send the rest and clean up the parsing structures */
        if (mp_mf->get_num_of_records() > 0)
            if (mp_mf->send_message (prev_type))
                goto stop_clean;

        m_parser.parse_cleanup ();
    }

    /* Reset the parser */
    m_parser.stop ();

    /* Send additional information */
    while (m_parser.get_misc_info (r), r.state != PR_END)
    {
        if (r.state != PR_OK)
        {
            log (L_ERROR, "Miscellaneous information chunks corrupted.");
            goto parser_clean;
        }
        if (handle_chunk (*r.chunk, prev_type))
            goto parser_clean;
    }
    if (mp_mf->get_num_of_records() > 0)
        if (mp_mf->send_message (prev_type))
            goto parser_clean;

    m_parser.parse_cleanup ();

    /* Fine, finished */
    return false;

stop_clean:
    m_parser.parse_cleanup ();
    m_parser.stop ();
    return true;

parser_clean:
    m_parser.parse_cleanup ();
    return true;
}

} //end of magents namespace
