/**
 * @file
 * @brief Echelog Miniagent support library
 *
 * External command source buffering interface.
 * It can be used to exec and gain the standard output
 * of a system utility.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/lib.h"
#include "magents/runbuffer.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <errno.h>

namespace magents {

using namespace lib;


/* Init the buffer */
tRunBuffer::tRunBuffer () 
    : pid (-1), cache_pos(cache_start - 1), cache_end(cache_start), 
    data_available (false), data_to_unget(0), bfd (-1)
{
    char *tmpdir;

    if (!(tmpdir = getenv ("TMP")) || !(tmpdir = getenv ("TMPDIR")))
        tmpdir = TMPDIR_FALLBACK;
    
    temp_filename = s_strconcat (tmpdir, "/magent_XXXXXX", NULL);
    ASSERT (temp_filename);

    temp_fd = sd_mkstemp (temp_filename);
    pgid = getpgid (0);
}


/* Destroy the buffer */
tRunBuffer::~tRunBuffer ()
{
    /* Anihilate the last child */
    cleanup ();

    /* Remove the temporary file (if any) */
    if (temp_fd >= 0)
    {
        if (s_close (temp_fd))
            log (L_ERROR, "Can't close temporary file.");
        
        if (unlink (temp_filename))
            log (L_ERROR, "Can't remove temporary file.");
    }

    xfree (temp_filename);
}


/* Kill the previously spawn child and close the pipe to it */
void tRunBuffer::cleanup ()
{
    /* Close old input pipe */
    if (bfd != -1)
        s_close (bfd);

    if (pid > -1)
    {
        debug_log ('p', "Stopping pid: %i", pid);

        /* Kill a pending spawned child */
        if (!s_test_signal_delivery (pid, NULL))
        {
            int r;

            debug_log ('p', "Killing spawned child %i", pid);
            kill (pid, SIGKILL);

            do 
            {
                r = waitpid (pid, NULL, 0);

                if (r == -1 && errno == EINTR)
                    r = 0;
            } 
            while (!r);
        }
        pid = -1;
    }

    /* Reset the values */
    cache_pos = cache_start - 1;
    cache_end = cache_start;
}


#define FAILURE 1

/* Establish a pipe to the source of information */
int tRunBuffer::init (const char *command, char *const args[], 
        char *const env[])
{ 
    int mypipe[2];

    /* Delete error information from previous session */
    if (ftruncate (temp_fd, 0))
        die ("Can't truncate the temporary file.");
    
    /* Create the pipe */
    if (pipe (mypipe))
    {
        log (L_ERROR, "Can't create a pipe.");
        return -FAILURE;
    }

    pid = fork ();
    if (pid == 0)
    {
        /* This is the child process.  Close the other end of the pipe 
         * and let executed command to write its output into it */
        s_close (0);
        s_close (2);
        s_close (mypipe[0]);
        if (dup2 (mypipe[1], 1) == -1)
            _exit (FAILURE);
        if (dup2 (temp_fd, 2) == -1)
            _exit (FAILURE);

        execve (command, args, env);
        log (L_FATAL, "Exec failed.");
        _exit (FAILURE);
    }
    else if (pid < 0)   /* Bad luck this time */
    {
        log (L_FATAL, "Fork failed.");
        return -FAILURE;
    }
    else
    {
        /* This is the parent process. Read child's output */
        s_close (mypipe[1]);

        debug_log ('p', "Forking pid: %i", pid);
        bfd = mypipe[0];
    }

    return 0;
}


/* Load further data to cache */
int tRunBuffer::fetch ()
{
    ssize_t p1, p2, bytes = 0;

    /* Do we have a pipe connection? */
    if (bfd < 0)
    {
        log (L_ERROR, "Pipe doesn't exist.");
        return -2;      /* obviously not */
    }

    /* Load the cache but leave the cache_pos byte untouched */
    p1 = (cache_start + CACHE_SIZE) - cache_end; 
    p2 = cache_end - cache_start - 1;

    if (p1)
    {
        /* Load the <end, real cache end> part */
        ssize_t r = s_read (bfd, cache_start, p1);

        if (r < 0)
        {
            log (L_ERROR, "Can't read from a pipe.");
            return -2;
        }
        if (r == 0)
            return -1;

        bytes += r;
    }

    if (p2 > 0)
    {
        /* Load <real cache start, end-1> part */
        ssize_t r = s_read (bfd, cache_start, p1);

        if (r < 0)
        {
            log (L_ERROR, "Can't read from a pipe.");
            return -2;
        }
        if (r == 0)
            return -1;

        bytes += r;
    }

    /* Update cache delimiters */
    cache_end = (((cache_end - cache_start) + bytes) % CACHE_SIZE) 
        + cache_start;

    /* Okay, data has been successfully fetched to the cache */
    return bytes;
}

}
