/**
 * @file
 * @brief Echelog Miniagent support library
 *
 * This class can be used for a simple lexical analysis
 * above an instance of tInfoBuffer.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/lib.h"
#include "magents/simplelex.h"

#include <string.h>
#include <limits.h>
#include <errno.h>


namespace magents {


/* Create a tSimpleLex object and bind a buffer to it */
tSimpleLex::tSimpleLex (tInfoBuffer &b, lib::tMemPool &mp, 
        const long max_token) 
    : buf (b), pool (mp), max_token_size (max_token), append_mode(false)
{
    using namespace lib;

    /* Allocate a pseudo-static buffer */
    token_buf = static_cast<char *> (xmalloc (max_token + 1));
    column = 0;
}


/* Destroy the tSimpleLex object */
tSimpleLex::~tSimpleLex ()
{
    using namespace lib;

    xfree (token_buf);
}


/* Store actual token */
inline void tSimpleLex::store_token (tToken &r)
{
    char *end, *nstart, *nend;
    unsigned long n;

    if (!len)
    {
        /* Store an empty string */
        r.str = static_cast<char *> (pool.alloc (1));
        *r.str = 0, r.len = 0;
        r.type = T_STRING;
        return;
    }

    /* null-terminate the token_buffer at first */
    token_buf[len] = 0;

    /* trim for number value checking */
    nstart = token_buf;
    nend = token_buf + len - 1;
    while (nstart < token_buf + len 
            && *nstart == ' ' || *nstart == '\t' || *nstart == '\n') 
        nstart++;
    while (nend > token_buf && *nend == ' ' || *nend == '\t' || *nend == '\n') 
        nend--;

    /* Try to read it as a number */
    n = strtoul (nstart, &end, 0);
    if ((end == nend + 1) && (n != ULONG_MAX || errno != ERANGE))
    {
        /* Seems ok, write it in */
        r.type = T_NUMBER;
        r.num = n;

        if (token_buf[0] == '-')
            r.sign = true;
        else
            r.sign = false;
    }
    else
        r.type = T_STRING;

    /* Store it as a string in any case.
     * We expect a null-terminated token_buf from get_string() */
    r.str = static_cast<char *> (pool.alloc (len + 1));
    memcpy (r.str, token_buf, len + 1);
    r.len = len;
}


/* Get a null-terminated string */
inline int tSimpleLex::get_string ()
{
    int c, status = 0;

    if (append_mode)
        old_len = len;
    else
        old_len = len = 0;

    /* Is there anything to get? */
    if ((c = buf.peek_byte ()) < 0)
    {
        column = 0;
        return -1;
    }

    /* Read to buffer till the end of a string */
    while (len < max_token_size && (c = buf.get_byte ()) >= 0
            && c != ' ' && c != '\t')
    {
        /* Don't read another line */
        if (c == '\n')
            break;

        token_buf[len++] = c;
        column++;
    }

    /* Clean the ending */
    token_buf[len] = 0;
    if (c < 0)
        column = 0;
    else if (len > old_len)
        buf.unget_byte ();
    
    /* Check for overflow */
    if (len == max_token_size)
        status = 1;

    return status;
}


/* Skip all blank characters */
void tSimpleLex::skip_blanks ()
{
    int c;

    /* Is there anything to skip? */
    if ((c = buf.peek_byte ()) < 0)
    {
        column = 0;
        return;
    }
    
    /* Just skip them */
    while ((c = buf.get_byte ()) >= 0 && (c == ' ' || c == '\t'))
       column++;

    /* Mark the first non-blank character as unread */
    if (c >= 0)
        buf.unget_byte ();
}


/* Skip all characters till the first occurence of stop_char */
int tSimpleLex::skip_to (const int stop_char)
{
    int c;

    /* Read to buffer till the stop_char is found */
    while ((c = buf.get_byte ()) >= 0 && (c != stop_char))
    {
        if (c != '\n')
            column++;
        else
            column = 0;
    }

    /* Premature end? */
    if (c < 0)
    {
        column = 0;
        return -1;
    }

    /* Update column number if we are on a new line */
    if (stop_char == '\n')
        column = 0;
    else
        column++;

    /* Okay, found successfully */
    return 0;
}


/* Skip all characters till any of stop_chars accur */
int tSimpleLex::skip_to (const char *stop_chars)
{ 
    int c;

    ASSERT (stop_chars);
    
    /* Read to buffer till the stop_char isn't found */
    while ((c = buf.get_byte ()) >= 0 && index (stop_chars, c) == NULL)
    {
        if (c != '\n')
            column++;
        else
            column = 0;
    }

    /* Premature end? */
    if (c < 0)
    {
        column = 0;
        return -1;
    }

    /* Update column number if we are on a new line */
    if (c == '\n' && index (stop_chars, '\n') != NULL)
        column = 0;
    else
        column++;

    /* Okay, found successfully */
    return 0;
}


/* Skip all characters till stop_char (not including it) */
int tSimpleLex::skip_upto (const int stop_char)
{
    int c;

    /* Is there anything to skip? */
    if ((c = buf.peek_byte ()) < 0)
    {
        column = 0;
        return -1;
    }

    /* If we're on the stop_char just now bail out */
    if (c == stop_char)
        return 0;

    /* Read to buffer till the stop_char is found */
    while ((c = buf.get_byte ()) >= 0 && (c != stop_char))
    {
        if (c != '\n')
            column++;
        else
            column = 0;
    }

    /* Premature end? */
    if (c < 0)
    {
        column = 0;
        return -1;
    }
    
    /* Mark last byte as unread */
    buf.unget_byte ();

    /* Okay, found successfully */
    return 0;

}


/* Skip all characters till any of stop_chars accur (not including it) */
int tSimpleLex::skip_upto (const char *stop_chars)
{
    int c;

    ASSERT (stop_chars);

    /* Is there anything to skip? */
    if ((c = buf.peek_byte ()) < 0)
    {
        column = 0;
        return -1;
    }

    /* If we're on the stop_char just now bail out */
    if (index (stop_chars, c) != NULL)
        return 0;

    /* Read to buffer till the stop_char is found */
    while ((c = buf.get_byte ()) >= 0 && index (stop_chars, c) == NULL)
    {
        if (c != '\n')
            column++;
        else
            column = 0;
    }

    /* Premature end? */
    if (c < 0)
    {
        column = 0;
        return -1;
    }
    
    /* Mark last byte as unread */
    buf.unget_byte ();

    /* Okay, found successfully */
    return 0;

}


/* Skip all characters up to the specified column */
int tSimpleLex::skip_upto_col (const long col)
{
    int c;

    /* Is there anything to skip? */
    if ((c = buf.peek_byte ()) < 0)
    {
        column = 0;
        return -1;
    }

    /* Is the request valid? */
    if (col < column)
        return -1;

    /* Are we already there? */
    if (col == column)
        return 0;

    /* Read to buffer up to col */
    while ((c != '\n')  && column < col && (c = buf.get_byte ()) >= 0)
        column++;

    /* Premature end ? */
    if (column != col)
    {
        column = 0;
        return -1;
    }

    /* Okay, done successfully */
    return 0;
}


/* Get all characters till specified stop_char is found */
int tSimpleLex::get_to (tToken &r, const int stop_char)
{
    int c, status = 0;

    /* Is there anything to read? */
    if ((c = buf.peek_byte ()) < 0)
        return -1;

    if (append_mode)
        old_len = len;
    else
        old_len = len = 0;

    /* Read to buffer till the stop_char is found */
    while (len < max_token_size && (c = buf.get_byte ()) >= 0
            && (c != stop_char))
    {
        token_buf[len++] = c;
        if (c != '\n')
            column++;
        else
            column = 0;
    }

    if (c < 0)
    {
        column = 0;
        return -1;
    }
    else if (len == max_token_size)
        status = 1;
    else if (len == max_token_size - 1)
    {
        token_buf[len++] = c;
        status = 1;
    }
    else
        token_buf[len++] = c;

    /* Update column number if we are on a new line */
    if (c == '\n' && stop_char == '\n')
        column = 0;
    else
        column++;

    /* Store the string */
    store_token (r);

    return status;
}


/* Get all characters till specified stop_char is found */
int tSimpleLex::get_to (tToken &r, const char *stop_chars)
{
    int c, status = 0;

    ASSERT (stop_chars);
    /* Is there anything to read? */
    if ((c = buf.peek_byte ()) < 0)
        return -1;

    if (append_mode)
        old_len = len;
    else
        old_len = len = 0;

    /* Read to buffer till the stop_char is found */
    while (len < max_token_size && (c = buf.get_byte ()) >= 0
            && index (stop_chars, c) == NULL)
    {
        token_buf[len++] = c;
        if (c != '\n')
            column++;
        else
            column = 0;
    }

    if (c < 0)
    {
        column = 0;
        return -1;
    }
    else if (len == max_token_size)
        status = 1;
    else if (len == max_token_size - 1)
    {
        token_buf[len++] = c;
        status = 1;
    }
    else
        token_buf[len++] = c;
    
    /* Update column number if we are on a new line */
    if (c == '\n' && index (stop_chars, '\n') != NULL)
        column = 0;
    else
        column++;

    /* Store the string */
    store_token (r);

    return status;
}


/* Get all characters up to stop_char */
int tSimpleLex::get_upto (tToken &r, const int stop_char)
{
    int c, status = 0;

    /* Is there anything to skip? */
    if ((c = buf.peek_byte ()) < 0)
    {
        column = 0;
        return -1;
    }

    if (append_mode)
        old_len = len;
    else
        old_len = len = 0;

    /* Are we on the stop_char? */
    if (c == stop_char)
    {
        /* Yes, store an empty string */
        store_token (r);
        return 0;
    }

    /* Read to buffer till the stop_char is found */
    while (len < max_token_size && 
            (c = buf.get_byte ()) >= 0 && (c != stop_char))
    {
        if (c != '\n')
            column++;
        else
            column = 0;
        token_buf[len++] = c;
    }

    if (len == max_token_size)
        status = 1;
    else if (c < 0)
    {
        /* Premature end */
        column = 0;
        return -1;
    }

    /* Store the string */
    store_token (r);

    /* Mark last byte as unread */
    buf.unget_byte ();

    /* Okay, found successfully */
    return status;
}


/* Get all characters up to stop_char */
int tSimpleLex::get_upto (tToken &r, const char *stop_chars)
{
    int c, status = 0;

    ASSERT (stop_chars);

    /* Is there anything to skip? */
    if ((c = buf.peek_byte ()) < 0)
    {
        column = 0;
        return -1;
    }

    if (append_mode)
        old_len = len;
    else
        old_len = len = 0;

    /* Are we on the stop_char? */
    if (index(stop_chars, c) != NULL)
    {
        /* Yes, store an empty string */
        store_token (r);
        return 0;
    }

    /* Read to buffer till the stop_char is found */
    while (len < max_token_size && 
            (c = buf.get_byte ()) >= 0 && index (stop_chars, c) == NULL)
    {
        if (c != '\n')
            column++;
        else
            column = 0;
        token_buf[len++] = c;
    }

    if (len == max_token_size)
        status = 1;
    else if (c < 0)
    {
        /* Premature end */
        column = 0;
        return -1;
    }

    /* Store the string */
    store_token (r);

    /* Mark last byte as unread */
    buf.unget_byte ();

    /* Okay, found successfully */
    return status;
}


/* Get all characters up to the specified column */
int tSimpleLex::get_upto_col (tToken &r, const long col)
{ 
    int c, status = 0;

    /* Is there anything to read? */
    if ((c = buf.peek_byte ()) < 0)
    {
        column = 0;
        return -1;
    }

    /* Is the request valid? */
    if (col < column)
        return -1;

    if (append_mode)
        old_len = len;
    else
        old_len = len = 0;

    /* Are we already there? */
    if (col == column)
    {
        /* Yes, store an empty string */
        store_token (r);
        return 0;
    }

    /* Read to buffer till the stop_char is found */
    while (len < max_token_size && column < col && (c = buf.get_byte ()) >= 0
             && (c != '\n'))
    {
       token_buf[len++] = c;
       column++;
    }

    if (len == max_token_size)
        status = 1;
    else if (column < col)
    {
        /* Line is too short */
        column = 0;
        return -1;
    }

    /* Store the string */
    store_token (r);

    return status;
}


/* Get the next token */
int tSimpleLex::get_token (tToken &r)
{
    using namespace lib;

    int c, status;

    /* Go over unimportant characters */
    skip_blanks ();

    /* Peek at input */
    c = buf.peek_byte();
    if (c < 0)
    {
        column = 0;
        return -1;
    }

    if (c == '\n')
    {
        r.type = T_NEWLINE;
        buf.get_byte (); 
        column = 0;
        r.len = 1;

        /* Okay, return */
        return 0;
    }

    /* Get next continuous token */
    status = get_string ();

    /* Store the string */
    store_token (r);

    /* Okay, return */
    return status;
}


}
