/**
 * @brief Echelog Library -- Message buffer handling
 * This class covers the I/O message buffering. 
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "messages/msgbuf.h"
#include "messages/testbuf.h"

namespace messages {

using namespace lib;


tMsgBuf::~tMsgBuf () 
{ 
    using namespace lib;

    debug_log ('m', "Trying to destroy message buffer");
    reset(); 
    debug_log ('m', "Message buffer destroyed"); 
}



int tMsgBuf::prepare_to_write (const char *new_buffer, int32_t new_length)
{
    ASSERT (!r_prepared && !w_prepared);

    new_length = min (new_length, MAX_MESSAGE_SIZE);

    debug_log ('m', "Trying to prepare message buffer for writing");
    if (!new_buffer)
    {
        log (L_ERROR, "Message buffer must be pre-allocated.");
        return EE_BUF_GENERAL_ERROR;
    }

    // Let the header begin at the very beginning of the message buffer
    header = new tMsgHeader ();
    if (!header)
    {
        log (L_ERROR, "Not enough memory.");
        return EE_BUF_GENERAL_ERROR;
    }
    if (header->prepare_to_write (new_buffer, new_length))
        return EE_BUF_GENERAL_ERROR;

    // Seek after the complete header
    new_length -= header->get_store_length ();
    new_buffer += header->get_store_length ();

    // Assign the body length field
    body_length_field = new tPlainBuf (new_buffer, new_length);
    if (!body_length_field)
    {
        log (L_ERROR, "Not enough memory.");
        return EE_BUF_GENERAL_ERROR;
    }

    // and skip it
    new_length -= get_body_field_size();
    new_buffer += get_body_field_size();
    if (!new_length)
    {
        log (L_ERROR, "There is no space for message body in the buffer.");
        return EE_BUF_GENERAL_ERROR;
    }
    
    // Assign the rest of the buffer for to the message body
    body = new tPlainBuf (new_buffer, new_length);
    if (!body)
    {
        log (L_ERROR, "Not enough memory.");
        return EE_BUF_GENERAL_ERROR;
    }
    w_prepared = true;

    debug_log ('m', "Message buffer prepared for writing");
    return SUCCESS;
}


int tMsgBuf::prepare_to_read (const char *new_buffer, int32_t new_length)
{
    ASSERT (!r_prepared && !w_prepared);

    int32_t max_length = min (new_length, MAX_MESSAGE_SIZE);
    new_length = max_length;

    debug_log ('m', "Trying to prepare message buffer for reading");
    if (!new_buffer)
    {
        log (L_ERROR, "Message buffer must be pre-allocated.");
        return EE_BUF_GENERAL_ERROR;
    }

    // Give the very begining of the buffer to a header 
    header = new tMsgHeader ();
    if (!header)
    {
        log (L_ERROR, "Not enough memory.");
        return EE_BUF_GENERAL_ERROR;
    }
    if (header->prepare_to_read (new_buffer, max_length))
        return EE_BUF_GENERAL_ERROR;

    // Get the complete header
    header->get ();

    // Set position there after
    new_length -= header->get_store_length ();
    new_buffer += header->get_store_length ();
    
    // Check if there is enough space at least for empty message body
    if (new_length - get_body_field_size() < 0)
    {
        log (L_ERROR, "Buffer is too small to hold the message.");
        return EE_BUF_EXCEEDED;
    }
    
    // Init a body length buffer
    body_length_field = new tPlainBuf (new_buffer, new_length);
    if (!body_length_field)
    {
        log (L_ERROR, "Not enough memory.");
        return EE_BUF_GENERAL_ERROR;
    }
    READ_INT32 (body_length_field, max_body_len);

    if (max_body_len < 0)
    {
        log (L_ERROR, "Message indicated negative body length.");
        return EE_BUF_GENERAL_ERROR;
    }
    new_buffer += get_body_field_size();
    new_length -= get_body_field_size();
    debug_log ('m', "Message body length: %li", max_body_len);

    // Check if the basic size delimiters are correct, ie.
    // if total message size == header size + body delimiter size + body size
    if (header->total_msg_length != 
            (header->get_len() + get_body_field_size() + max_body_len ))
    {
        log (L_ERROR, "Message corrupted: invalid basic message delimiters.");
        return EE_BUF_GENERAL_ERROR;
    }
    
    // Init a new body buffer
    body = new tPlainBuf (new_buffer, new_length);
    if (!body)
    {
        log (L_ERROR, "Not enough memory.");
        return EE_BUF_GENERAL_ERROR;
    }
    r_prepared = true;
    declared_length = header->get_total_msg_length () + 
        header->get_total_msg_length_field_size();

    debug_log ('m', "Message buffer prepared for reading");
    return SUCCESS;
}


int tMsgBuf::reset ()
{
    if (header)
        delete header;
    if (body_length_field)
        delete body_length_field;
    if (body)
        delete body;

    r_prepared = w_prepared = false;
    declared_length = -1;
    header = NULL;
    body_length_field = body = NULL;

    debug_log ('m', "Message buffer reset");
    return SUCCESS;
}


int tMsgBuf::close ()
{
    debug_log ('m', "Trying to close message buffer");
    if (w_prepared)
    {
        long msg_len;

        declared_length = false;

        // Save the body length
        body_length_field->rewind ();
        STORE_INT32 (body_length_field, body->get_store_length ());

        // Save the whole message length
        msg_len = get_len ();
        if (msg_len < header->get_store_length())
            return EE_BUF_GENERAL_ERROR;
        header->total_msg_length = msg_len;  

        // Finish him!
        if (header->close ())
            return EE_BUF_GENERAL_ERROR;

        declared_length = get_store_length ();
    }

    debug_log ('m', "Message buffer closed");
    return SUCCESS;
}

}
