/**
 * @file
 * @brief Echelog Library -- Message Chunk Class
 * @author Ales Privetivy <ales@ucw.cz>
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/lib.h"
#include "messages/msgmgr.h"
#include "messages/msgchunk.h"

namespace messages {

using namespace lib;

/** constructor for msgchunk - creates the chunk ti represent the message of
 *  the given name and allocation are made in mpool mempool */
tMsgChunk::tMsgChunk(char *message_name, class lib::tMemPool &mpool)
: mempool(mpool)
{
    // get message handle
    message_handle=gb_message_manager.get_handle_by_name(message_name);

    // message of this name does not exists
    if (message_handle<0) 
        die("Trying to instantiate unknown message type %s", message_name);

    // create message of this type
    attribute = gb_message_manager.create_attr_list(message_handle, mpool);
    attr_count = gb_message_manager.get_attr_count(message_handle);
    type = gb_message_manager.get_message_type_by_handle(message_handle);

    // debug output when message instance successfully created
    debug_log ('m', "Chunk %li (%s) of total #%li attributes created",
            type, message_name, attr_count);
    
}


/** constructor for msgchunk - creates the chunk ti represent the message of
 * the given type and allocation are made in mpool mempool */
tMsgChunk::tMsgChunk(int32_t message_type, class lib::tMemPool &mpool)
: mempool(mpool)
{
    // get message handle
    message_handle=gb_message_manager.get_handle_by_type(message_type);

    // message of this type does not exists
    if (message_handle<0)
        die("Trying to instantiate unknown message type message_type=%d", 
            message_type);

    // create message of this type
    attribute = gb_message_manager.create_attr_list(message_handle, mpool);
    attr_count = gb_message_manager.get_attr_count(message_handle);
    type = message_type;

    // debug output when message instance successfully created
    debug_log ('m', "Chunk %li (%s) of total #%li attributes created",
            type,gb_message_manager.get_message_name_by_handle (message_handle),
            attr_count);

}


/** destructor for msgchunk - frees attribute list */
tMsgChunk::~tMsgChunk()
{
     int32_t i;

     // free all attributes
     for(i=0;i<attr_count;i++)
         delete attribute[i];

}


/** gets pointer to instance of the attribute with attr_name */
class tEchelogType * tMsgChunk::get_attr (const char *attr_name) const
{
    int32_t attr_handle;
    
    // get handle of attribute within message
    attr_handle = gb_message_manager.get_handle_of_attr(message_handle,
                attr_name);

    // attribute of this name does not exists
    if (attr_handle<0)
        die("Trying to get unknown attribute %s from message %s",attr_name,
            gb_message_manager.get_message_name_by_handle(message_handle));

    // return pointer to attribute instance
    return attribute[attr_handle];
}


/** gets pointer to instance of the attribute with attr_handle */
class tEchelogType * tMsgChunk::get_attr (int32_t attr_handle) const
{
    if (attr_handle < 0 || attr_handle >= attr_count)
        die("Trying to get unknown attribute (#%li) from message %s",
                attr_handle, 
                gb_message_manager.get_message_name_by_handle(message_handle));

    return attribute[attr_handle];
}


/** queries if the key attributes of messages chunks match */
bool tMsgChunk::key_match (tMsgChunk &mc)
{
    int32_t i;

    if (message_handle != mc.message_handle)
        return false;

    if (!gb_message_manager.is_aggregable(message_handle))
        return false;

    for (i = 0; i < attr_count; i++)
        if (gb_message_manager.is_key_attr_by_handle (message_handle, i))
        {
            tTYPE_RELATION r = type_order (*attribute[i], *mc.attribute[i]);
            if (r != TP_IDENTICAL && r != TP_EQUAL)
                return false;
        }

    return true;
}


/** Get the space needed to store the message chunk */
int32_t tMsgChunk::get_store_length () const
{
    int32_t i, sum;

    /* get the store length of each type in message chunk */
    for (i = 0, sum = 0; i < attr_count; i++)
        sum += attribute[i]->get_store_length ();
    /* add the status field */
    sum ++;

    /* add the message chunk size delimiter and return the sum */
    return sum + cint32_size (sum);
}


/** loads data from actual position in buf and stores it in the chunk */
int32_t tMsgChunk::get (tPlainBuf *buf)
{
    int32_t i, sum, bytes_read, real_sum, mark, tmp;

    debug_log ('m', "Trying to read message chunk %li (%s)", type,
            gb_message_manager.get_message_name_by_handle (message_handle));
    CHECK_BUFFER (buf);
    READ_CINT32 (buf, sum, tmp);
    mark = buf->get_cursor_position ();

    if (buf->read_state (status) <= 0)
    {
        // failed skip over this message chunk
        buf->abs_seek (mark + sum);
        return EE_BUF_CHUNK_R_FAILED;
    }

    for(i = 0, real_sum = 1; i < attr_count; i++)
    {
        if ((real_sum += bytes_read = attribute[i]->get(buf)) <= 0)
        {
            buf->abs_seek (mark + sum);
            return EE_BUF_CHUNK_R_FAILED;
        }
        debug_log ('m', "Chunk [%i] attribute #%i (%li bytes) read "
                "successfully", type, i, bytes_read);
    }

    // check the real number of bytes read
    if (sum != real_sum)
    {
        buf->abs_seek (mark + sum);
        log (L_WARN, "The message chunk corrupted. Skipped to next.");
        return EE_BUF_CHUNK_R_FAILED;
    }

    debug_log ('m', "Message chunk %i reading (%li bytes) sucessfully "
            "completed", type, sum);
    return sum;
}


/** stores data from chunk to the actual position to actual position in
 *  the plainbuffer */
int32_t tMsgChunk::put (tPlainBuf *buf) const
{
    int32_t i, tmp, sum = 0;

    debug_log ('m', "Trying to store message chunk %li (%s)", type,
            gb_message_manager.get_message_name_by_handle (message_handle));
    CHECK_BUFFER (buf);
    for (i = 0; i < attr_count; i++)
        sum += attribute[i]->get_store_length ();
    sum ++;     // state info

    // check the space for its contents, the length field and the status
    if (buf->check_spare_space (sum + cint32_size (sum)))        
        return EE_BUF_CHUNK_W_FAILED;

    STORE_CINT32 (buf, sum, tmp);
    STORE_STATE (buf, get_status ());

    for(i = 0; i < attr_count; i++)
    {
        int32_t bytes;
        if ((bytes = attribute[i]->put(buf)) <= 0)
            return EE_BUF_STORE_ERROR;
    }
    // return the real number of bytes stored
    sum += tmp;

    debug_log ('m', "Message chunk %i storing (%li bytes in %li in attributes) "
                    "completed sucessfully ", type, sum, attr_count);
    return sum;
}

}
