/**
 * @brief Echelog Library -- Message buffer header handling
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/archdep.h"
#include <string.h>

#include "lib/lib.h"
#include "messages/msgheader.h"
#include "messages/testbuf.h"

namespace messages {

using namespace lib;


tMsgHeader::~tMsgHeader () 
{
    debug_log('m', "Destroying header"); 
    reset(); 
}
            

char * tMsgHeader::get_agent_id () const
{
    int i;
    char * str = static_cast<char *> (xmalloc (ID_LENGTH + 1));

    for (i = 0; i < ID_LENGTH; i++)
        str[i] = agent_id[i];

    str[i] = 0; // add the termination null

    return str; // return the duplicated string
}


void tMsgHeader::set_agent_id (const char *prop_value)
{
    int i;

    if (!prop_value)
    	die ("Agent ID must not be null");

    for (i = 0; i < ID_LENGTH; i++)
        agent_id[i] = prop_value[i];
}


char * tMsgHeader::get_agent_instance_id () const
{
    int i;
    char * str = static_cast<char *> (xmalloc (ID_LENGTH + 1));

    for (i = 0; i < ID_LENGTH; i++)
        str[i] = agent_instance_id[i];

    str[i] = 0; // add the termination null

    return str; // return the duplicated string
}


void tMsgHeader::set_agent_instance_id (const char *prop_value)
{
    int i;

    if (!prop_value)
    	die ("Agent instance ID must not be null");

    for (i = 0; i < ID_LENGTH; i++)
        agent_instance_id[i] = prop_value[i];
}

    
byte * tMsgHeader::get_host_id () const
{
    int i;
    byte * str = static_cast<byte *> (xmalloc (ID_LENGTH + 1));

    for (i = 0; i < ID_LENGTH; i++)
        str[i] = host_id[i];

    str[i] = 0; // add the termination null

    return str; // return the duplicated string
}


char * tMsgHeader::get_host_str () const
{
#if ID_LENGTH != 4
#warning Identificator length not supported
#endif

    char *ret;
    s_asprintf (&ret, "%u.%u.%u.%u", 
            host_id[0], host_id[1], host_id[2], host_id[3]);

    return ret;
}


void tMsgHeader::set_host_id (const byte *prop_value)
{
    int i;

    if (!prop_value)
    	die ("Host ID must not be null");

    for (i = 0; i < ID_LENGTH; i++)
        host_id[i] = prop_value[i];
}


void tMsgHeader::set_host_id (const char *prop_value)
{
    int i;

    if (!prop_value)
    	die ("Host ID must not be null");

    for (i = 0; i < ID_LENGTH; i++)
        host_id[i] = static_cast<byte> (prop_value[i]);
}


int tMsgHeader::prepare_to_write (const char *new_buffer, int32_t max_length)
{
    len = pure_header_size() + get_header_length_field_size();
    if (len + get_total_msg_length_field_size() > max_length)
    {
        log (L_ERROR, "Buffer is too small for message header.");
        return EE_BUF_EXCEEDED;
    }

    buffer = new tPlainBuf (new_buffer, get_store_length());
    if (!buffer)
    {
        log (L_ERROR, "Not enough memory.");
        return -1;
    }

    w_prepared = true;

    debug_log ('m', "Header prepared for writing (%li bytes)", 
            get_store_length());
    return SUCCESS;
}


int tMsgHeader::prepare_to_read (const char *new_buffer, int32_t max_length)
{ 
    int32_t given_length;

    len = pure_header_size() + get_header_length_field_size();
    if (len + get_total_msg_length_field_size() > max_length)
    {
        log (L_ERROR, "Buffer is too small for message header.");
        return EE_BUF_EXCEEDED;
    }

    buffer = new tPlainBuf (new_buffer, get_store_length());
    if (!buffer)
    {
        log (L_ERROR, "Not enough memory.");
        return -1;
    }

    READ_INT32 (buffer, total_msg_length);
    debug_log ('m', "Message length: %li", total_msg_length);

    /* Check if the total message size is withing basic limits */
    if ((total_msg_length < get_store_length ()) ||
            total_msg_length > 
            (MAX_MESSAGE_SIZE - get_total_msg_length_field_size ()))
    {
        log (L_ERROR, "Message corrupted: nonsential message size: %i.",
            total_msg_length);
        return EE_BUF_GENERAL_ERROR;
    }

    READ_INT32 (buffer, given_length);
    if (given_length != pure_header_size())
    {
        log (L_ERROR, "Given header length differs from current "
            "header specification.");
        return EE_BUF_GENERAL_ERROR;
    }
    r_prepared = true;

    debug_log ('m', "Header prepared for reading (%li bytes)", 
            get_store_length());
    return SUCCESS;

}


int tMsgHeader::get ()
{ 
    int32_t i, header_length;

    CHECK_R_PREPARED();
    buffer->rewind ();

    // Total message length
    READ_INT32 (buffer, total_msg_length);
    // Skip the header length (already retrieved)
    READ_INT32 (buffer, header_length);
    if (header_length != pure_header_size())
        return EE_BUF_GENERAL_ERROR;
    // Message type
    READ_INT32 (buffer, msg_type);
    // Agent identificator
    for (i = 0; i < ID_LENGTH; i++)
        READ_CHAR (buffer, agent_id[i]);
    // Agent instance identificator
    for (i = 0; i < ID_LENGTH; i++)
        READ_CHAR (buffer, agent_instance_id[i]);
    // Host identificator
    for (i = 0; i < ID_LENGTH; i++)
        READ_BYTE (buffer, host_id[i]);
    // Flag bits
    READ_INT32 (buffer, flags);
    // Timestamp
    READ_INT32 (buffer, msg_time);
    // Number of subrecords
    READ_INT32 (buffer, subrecords);

    debug_log ('m', "Header information successfully read (%li bytes)", len);
    return SUCCESS;
}


int tMsgHeader::put ()
{
    int32_t i;

    CHECK_W_PREPARED();
    buffer->rewind ();

    // Total message length
    STORE_INT32 (buffer, total_msg_length);
    // Store the header length
    STORE_INT32 (buffer, pure_header_size());
    // Message type
    STORE_INT32 (buffer, msg_type);
    // Agent identificator
    for (i = 0; i < ID_LENGTH; i++)
        STORE_CHAR (buffer, agent_id[i]);
    // Agent identificator
    for (i = 0; i < ID_LENGTH; i++)
        STORE_CHAR (buffer, agent_instance_id[i]);
    // Host identificator
    for (i = 0; i < ID_LENGTH; i++)
        STORE_CHAR (buffer, host_id[i]);
    // Flag bits
    STORE_INT32 (buffer, flags);
    // Timestamp
    STORE_INT32 (buffer, msg_time);
    // Number of subrecords
    STORE_INT32 (buffer, subrecords);

    debug_log ('m', "Header information successfully stored (%li bytes)", len);
    return SUCCESS;
}


int tMsgHeader::reset ()
{
    if (buffer)
        delete buffer;
    len = EE_BUF_NOT_READY;
    r_prepared = false;
    w_prepared = false;

    debug_log ('m', "Header reset");
    return SUCCESS;
}


int tMsgHeader::close ()
{
    if (w_prepared)
    {
        // Store the header contents
        put ();
    }

    debug_log ('m', "Header closed");
    return SUCCESS;
}


void tMsgHeader::copy_envelope (tMsgHeader &header)
{
    int i;

    msg_type = header.msg_type;
    msg_time = header.msg_time;
    subrecords = header.subrecords;
    
    for (i = 0; i < ID_LENGTH; i++)
        agent_id[i] = header.agent_id[i];
    for (i = 0; i < ID_LENGTH; i++)
        agent_instance_id[i] = header.agent_instance_id[i];
    for (i = 0; i < ID_LENGTH; i++)
        host_id[i] = header.host_id[i];

    flags = header.flags;
}

}
