/**
 * @brief Echelog Library -- Message Manager Class
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include "lib/lib.h"
#include "messages/msgmgr.h"

namespace messages {

using namespace lib;
using namespace private_messages;


/// global variable ( we need only one instance of message manager)
class tMsgMgr gb_message_manager;


tMemPool &get_msgdef_pool ()
{
    static tMemPool msgdef_pool (4096);
    return msgdef_pool;
}


void tMsgMgr::check_message_handle (int32_t message_handle)
{
    if (message_handle < 0 || message_handle > message_def_count)
            die("Unknown message handle. Message doesn't exist?.");
}


void tMsgMgr::check_attribute_handle (tMsgDefinition *md, 
        int32_t attribute_handle)
{
    if (attribute_handle < 0 || attribute_handle >= md->attr_count)
            die("Unknown message attribute handle. Attribute doesn't exist?.");
}


/// Sets the message definition structures to empty.
tMsgMgr::tMsgMgr()
{
    message_def_list = static_cast<struct
        private_messages::tMsgDefinition **>(NULL);
    message_def_count=0;
}


/// Makes a deep copy of a message definition def within pool
tMsgDefinition *tMsgMgr::def_dup (tMsgDefinition *def)
{
    if (!def)
        return NULL;

    /* Alloc the structures */
    tMsgDefinition *copy = static_cast<tMsgDefinition *>
        (get_msgdef_pool().alloc (sizeof (tMsgDefinition)));
    tAttributeDefinition *attr_copy = static_cast<tAttributeDefinition *>
        (get_msgdef_pool().alloc 
             (sizeof (tAttributeDefinition) * def->attr_count));

    /* Duplicate the body */
    copy->message_name = s_strdup (def->message_name, get_msgdef_pool());
    copy->message_type = def->message_type;
    copy->attr_count = def->attr_count;
    copy->attribute = attr_copy;

    /* And each attribute */
    for (int32_t i = 0; i < def->attr_count; i++)
    {
        attr_copy[i].attr_name = s_strdup (def->attribute[i].attr_name, 
                get_msgdef_pool ());
        attr_copy[i].type_name = s_strdup (def->attribute[i].type_name, 
                get_msgdef_pool ());
        attr_copy[i].key = def->attribute[i].key;
    }

    return copy;
}


/// registers message list and fills in message-mapping structures
void tMsgMgr::register_message_list(class lib::tList *msg_def_l)
{
    using namespace lib;

    // tmp variable for walking through message def. list 
    struct tNode const *c;
    // one element of message definition list
    struct tMsgDefinition *md;
    // loop variables
    int32_t i,j;

    // stores the number of messages to be registered
    message_def_count = msg_def_l->count();

    debug_log ('c', "Registering message list of total #%i messages", 
               message_def_count);

    // now we allocate array for storing message info
    message_def_list = static_cast<struct tMsgDefinition **>
       (get_msgdef_pool().alloc 
            (sizeof(struct tMsgDefinition *)*message_def_count));

    // for each message definition
    i = 0;
    FOR_ALL_LIST_ITEMS (c, msg_def_l)
    {
        message_def_list[i] = def_dup (static_cast<tMsgDefinition *>(c->data));
        md = message_def_list[i];

        // we process all for message attributtes
        for(j=0;j<md->attr_count;j++)

            // does this type exist ?
            if (!gb_type_manager.check_type(md->attribute[j].type_name))
                die("Message type %s has unregistered attribute type %s "
                    "for field %s", md->message_name,
                    md->attribute[j].type_name,
                    md->attribute[j].attr_name); 

        // debug output on succesfull message registration
        debug_log ('c', "Message '%s' of %li attributes sucessfully "
                   "registered",  md->message_name, md->attr_count);

        i++;
    }
}


// return index of message definition in our structures
int32_t tMsgMgr::get_handle_by_name(const char *message_name)
{

    int32_t i;	

    // walk through message def array and look for the name
    for(i=0;i<message_def_count;i++)
        if (!strcmp(message_name,message_def_list[i]->message_name)) break;

    // name not found
    if (i==message_def_count) return -1;

    // found under this index
    return i;

}


// return index of message definition in our structures 
int32_t tMsgMgr::get_handle_by_type(int32_t type)
{
    int32_t i;

    // walk through message def array and look for the name
    for(i = 0; i < message_def_count; i++)
        if (type == message_def_list[i]->message_type) 
            break;

    // type not found
    if (i==message_def_count) return -1;

    // found under this index
    return i;
}


// return name of type of attribute identified by message_handle:attr
char * tMsgMgr::get_type_of_attr(int32_t message_handle, const char *attr)
{

    struct tMsgDefinition *md;
    int32_t i;

    check_message_handle (message_handle);

    // we take the definition of message[message_handle]
    md = message_def_list[message_handle];

    // walk through attribute list and look for attrname
    for(i=0;i<md->attr_count;i++)
        if (!strcmp(attr,md->attribute[i].attr_name)) break;

    // attr name not found
    if (i==md->attr_count) return NULL; 

    // name found and returned
    return md->attribute[i].type_name;
}


// return index of attribute in message def array identified by
// message_handle:attr
int32_t tMsgMgr::get_handle_of_attr(int32_t message_handle, const char *attr)
{
    struct tMsgDefinition *md;
    int32_t i;

    check_message_handle (message_handle);

    // we take the definition of message[message_handle]
    md = message_def_list[message_handle];

    // walk through attribute list and look for attrname
    for(i=0;i<md->attr_count;i++)
        if (!strcmp(attr,md->attribute[i].attr_name)) break;

    // attr name not found
    if (i==md->attr_count) return -1;

    // name found and returned appropriate index
    return i;
}


// return number of attributes for given message
int32_t tMsgMgr::get_attr_count(int32_t message_handle)
{
    check_message_handle (message_handle);
    return message_def_list[message_handle]->attr_count;

}

// gets a name of i-th message attribute
char * tMsgMgr::get_attr_name (int32_t message_handle, int32_t attr_handle)
{
    struct tMsgDefinition *md;

    check_message_handle (message_handle);

    // we take the definition of message[message_handle]
    md = message_def_list[message_handle];
    check_attribute_handle (md, attr_handle);

    return md->attribute[attr_handle].attr_name;
}


// message-chunk creation
class tEchelogType **tMsgMgr::create_attr_list(int32_t message_handle, 
        class lib::tMemPool &mpool)
{

    class tEchelogType **attr_array;
    struct tMsgDefinition *md;
    int32_t i;

    check_message_handle (message_handle);

    // take the message definition for the wanted message 
    md = message_def_list[message_handle];

    // allocate array for attribute instances
    attr_array = static_cast<class tEchelogType **>
        (mpool.alloc(sizeof(class tEchelogType *)* md->attr_count));

    // create each instance for each attribute
    for(i=0;i<md->attr_count;i++)
        attr_array[i] = gb_type_manager.get_type_instance
            (mpool, md->attribute[i].type_name);

    // return the instance array
    return attr_array;
}


/// Gets message_type according to message_handle
int32_t tMsgMgr::get_message_type_by_handle(int32_t message_handle)
{
    using namespace lib;

    check_message_handle (message_handle);
    return message_def_list[message_handle]->message_type;
}


/// Gets message_name according to message_handle
char * tMsgMgr::get_message_name_by_handle(int32_t message_handle)
{
    using namespace lib;

    check_message_handle (message_handle);
    return message_def_list[message_handle]->message_name;
}


/// Is aggregable ?
bool tMsgMgr::is_aggregable(int32_t message_handle)
{
    struct tMsgDefinition *md;
    int32_t i;

    check_message_handle (message_handle);

    // we take the definition of message[message_handle]
    md = message_def_list[message_handle];

    for (i=0;i<md->attr_count;i++)
        if (md->attribute[i].key) return 1;

    return 0;
}


/// Is it the key attribute ?
bool tMsgMgr::is_key_attr(int32_t message_handle, const char *attr)
{
    struct tMsgDefinition *md;
    int32_t i;

    check_message_handle (message_handle);

    // we take the definition of message[message_handle]
    md = message_def_list[message_handle];

    if (i < 0 || i >= md->attr_count)
        return 0;

    // walk through attribute list and look for attrname
    for(i=0;i<md->attr_count;i++)
        if (!strcmp(attr,md->attribute[i].attr_name)) break;

    // attr name not found
    if (i==md->attr_count) 
        return 0;
    
    return md->attribute[i].key;
}


/// Is it the key attribute ?
bool tMsgMgr::is_key_attr_by_handle(int32_t message_handle, int32_t attr_handle)
{
    struct tMsgDefinition *md;

    check_message_handle (message_handle);

    // we take the definition of message[message_handle]
    md = message_def_list[message_handle];
    check_attribute_handle (md, attr_handle);

    return md->attribute[attr_handle].key;
}

}
