/**
 * @file
 * @brief Contains implementation of tMsgPrinter class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 * @author Kamil Toman <ktoman@email.cz>
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include "messages/msgprinter.h"
#include "messages/msgmgr.h"
#include <unistd.h>

using namespace lib;
using namespace messages;

namespace messages {


bool tMsgPrinter::print_message (int logfd, tMsgBuf& msg)
{
    tMsgChunk *chunk;
    int32_t i, msg_type, msg_subrecords;

    /* Print basic information */
    if (print_header (logfd, msg))
        goto fail;

    /* Get message characteristics */
    msg_type = msg.get_header()->get_msg_type ();
    msg_subrecords = msg.get_header()->get_subrecords ();

    /* Create a new chunk */
    chunk = new tMsgChunk (msg_type, m_chunk_mp);
    if (!chunk)
    {
        log (L_ERROR, "Not enough memory.");
        return true;
    }

    lseek (logfd, 0, SEEK_END);

    /* Load one chunk at a time */
    for (i = 0; i < msg_subrecords; i++)
    {
        int32_t bytes = chunk->get (msg.get_body ());
        if (bytes > 0)
            if (print_chunk (logfd, *chunk, i))
                goto chunk_clean;
        else if (bytes < 0)
            log (L_WARN, "Invalid chunk.");
    }

    fsync (logfd);

    /* Destroy what's left */
    delete chunk;
    m_chunk_mp.free_pool ();

    return false;

chunk_clean:
    log (L_ERROR, "Message was not printed correctly.");
    delete chunk;
    m_chunk_mp.free_pool ();
fail:
    return true;
}


/* Print header information */
int tMsgPrinter::print_header (int logfd, tMsgBuf &message)
{
    return 0;
}


/* print the list of printouts */
int tMsgPrinter::printout (int logfd, tList &list)
{
    char *string;
    int32_t cl, n;
    int err = 0;

    /* check if we print a complex type, open a brace if so */
    if ((cl = list.count ()) > 1)
        err = min (err, s_write (logfd, "{ ", 2));

    /* print out the type */
    for (n = list.count (); n > 0; n--)
    {
        void *item = list.remove_first ();
        string = static_cast<char *> (item);

        if (!string)
            string = "(null)";

        err = min (err, s_write (logfd, string, strlen (string)));

        if (n > 1)
            err = min (err, s_write (logfd, ", ", 2));
    }

    /* If we print a complex type add closing brace */
    if (cl > 1)
        err = min (err, s_write (logfd, " }", 2));

    return err;
}


/* print chunk contents */
int tMsgPrinter::print_chunk (int logfd, tMsgChunk &chunk, int32_t num)
{
    static tMemPool pool (4096);
    static tList list (pool);
    int32_t i, num_of_attributes, message_handle;
    int err = 0;
    const char *prefix;

    message_handle = gb_message_manager.get_handle_by_type (chunk.get_type());
    num_of_attributes = chunk.get_attr_count ();

    prefix = str_status (chunk.get_status ());

    err = min (err, s_write (logfd, prefix, strlen (prefix)));

    /* and then each type in its default format */
    for (i = 0; i < num_of_attributes; i++)
    {
        char *name = gb_message_manager.get_attr_name (message_handle, i);
        chunk.get_attr (i)->sprint (list, pool,
               chunk.get_attr (i)->get_plain_format ());

        if (i > 0)
            err = min (err, s_write (logfd, " ", 1));

        if (num_of_attributes > 1)
        {
            err = min (err, s_write (logfd, name, strlen (name)));
            err = min (err, s_write (logfd, ": ", 2));
        }

        err = min (err, printout (logfd, list));

    }

    err = min (err, s_write (logfd, "\n", 1));

    pool.free_pool ();

    return err;
}


/* convert tMSG_STATUS to string */
const char* tMsgPrinter::str_status (tMSG_STATUS status) const
{
    switch (status)
    {
        case ST_ADDED:
            return "";
        case ST_NOCHANGE:
            return "NOCHANGE ";
        case ST_REMOVED:
            return "REMOVED ";
        case ST_CHANGED:
            return "CHANGED ";
        case ST_ITEM:
            return "ITEM ";
        default:
            return "UNKNOWN ";
    }
}

}
