/**
 * @brief Echelog Library -- "Type" handling
 * Supplementary structures and functions for type printouts
 *
 * The description string conversion is implemented 
 * as LL(1) recursive parser.
 *      
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "messages/printout.h"

namespace messages {

using namespace lib;
    
/* All possible types of tokens (all terminal types): */
typedef enum 
{
    T_LBRACE,   // '{'
    T_RBRACE,   // '}'
    T_COMMA,    // ','
    T_STRING,   // "..." (a text enclosed in a pair of double quotes)
    T_DEFAULT,  // the keyword 'default'
    T_EOS,      // end of the string      
    T_OTHER,    // an ordinary character different from above
} tTOKEN_TYPE;


// Returned token
struct tToken 
{
    tTOKEN_TYPE type;
    union {
        char * str;
        char c;
    } value;
};

/* 
 * Non-terminal types: 
 * T (a tree description: a terminal or an embraced list of trees)
 * L (a list of trees)
 */


static tMemPool *fmt_pool;      // current pool to alloc in
static const char *buf;         // current in-format string position


/* a static wrapper around the default mempool for printouts to ensure 
 * that it is created before the first use */
lib::tMemPool &get_prpool()
{
    static tMemPool prpool (4096);
    return prpool;
}


/* make a null-terminated copy in mempool and advance the position */
static inline char *dup_and_advance (const char *start, long length)
{
    char *ret = static_cast<char *> (fmt_pool->alloc (length + 1));

    memcpy (ret, start, length);
    ret[length] = 0;

    buf += length + 1;
    return ret;
}


/* read the whole string; return NULL on error  */
static inline char *read_string ()
{
   const char *s = buf; 

   if (!*s)
       return NULL;     // unterminated substring

   if (*s == '"')       // empty string
       return dup_and_advance (buf, 0);

   do // read the string
   {
       switch (*s)
       {
           case '\\':
               if (!*++s)             
                   return NULL;     // unterminated substring

               /* skip escaped char */
               s++;
               break;
           case '"':
               return dup_and_advance (buf, s - buf);
        }
        s++;
   }
   while (*s);
   
   /* unterminated string */
   return NULL;
}


/* match the keyword 'default' in format string, advance the pos on match */
static inline bool is_default ()
{
    static const char def_str[] = "default";
    static const int def_str_len = 7;
    const char *s = buf;

    /* match keyword (the first char already matches) */
    for (int i = 1; i < def_str_len; i++)
    {
        if (!*s)
            return false;       // string too short
        if (*s++ != def_str[i])
            return false;       // don't match
    }

    buf += def_str_len - 1;
    return true;
}


/* get next lexical element */
static void get_token (tToken &ret)
{
    int c;

    /* ignore whitespace */
    while (*buf && (*buf == ' ' || *buf == '\t' || *buf == '\n')) buf++;

    if (!*buf)
    {
        ret.type = T_EOS;
        return;
    }

    switch (c = *buf++)
    {
        case 0:   // end of the format string
            ret.type = T_EOS;
            break;
        case '{':
            ret.type = T_LBRACE;
            break;
        case '}':
            ret.type = T_RBRACE;
            break;
        case ',':
            ret.type = T_COMMA;
            break;
        case '"': // print format string beginning
            ret.value.str = read_string ();
            if (ret.value.str)
                ret.type = T_STRING;
            else        // parse error, unterminated substring
                ret.type = T_EOS;
            break;
        case 'd':
            if (is_default ())
            {
                ret.type = T_DEFAULT;
                break;
            }
        default:
            ret.type = T_OTHER;
            ret.value.c = c;
    }
}


/* peek at next lexical element */
static inline void look_ahead (tToken &ret)
{
    const char *old_pos = buf;
    get_token (ret);
    buf = old_pos;
}


/* create a leaf node containg the print format */
static inline tFmtNode *create_leaf (const char *format_description)
{
    tFmtNode *node;

    ASSERT (format_description);
    node = static_cast<tFmtNode *> (fmt_pool->alloc (sizeof (tFmtNode)));
    node->nodes = static_cast<tFmtLink *> (fmt_pool->alloc (sizeof (tFmtLink)));
    node->num_of_nodes = 0;
    node->nodes[0].format = const_cast<char *> (format_description);

    return node;
}


#define SUCCESS 0
#define PARSE_ERROR 1


/* create an inner node of `count' subtrees */
static inline tFmtNode *create_inner_node (tList &list, int32_t count)
{
    int32_t i;
    tFmtNode *node;
    
    /* allocate neccessary space */
    node = static_cast<tFmtNode *> (fmt_pool->alloc (sizeof (tFmtNode)));
    node->nodes = static_cast<tFmtLink *> 
        (fmt_pool->alloc (count * sizeof (tFmtLink)));

    /* add in each component */
    for (i = 0; i < count; i++)
    {
        void *item = list.remove_first ();
        node->nodes[i].link = static_cast<tFmtNode *> (item);
    }
    node->num_of_nodes = count;

    /* return the tree */
    return node;
}


static int parse_fmt_tree (tFmtNode * &ret);
    
/* 
 * Parse the inner format list
 *
 * Grammar:
 * L -> T
 * L -> T, L
 *
 */
static int parse_list (tFmtNode * &ret)
{
    tToken token, next_token;
    tList components (*fmt_pool);
    int32_t n = 0;

    for (;;) 
    {
        tFmtNode *c;

        // parse next component
        if (parse_fmt_tree (c))
            return PARSE_ERROR;
        components.append (c);
        n++;

        // is there another list component?
        look_ahead (next_token);
        if (next_token.type != T_COMMA) 
            break;      // no, bail out
        else                  
            get_token (token);  // eat comma
    } 

    ret = create_inner_node (components, n);
    return SUCCESS;
}


/* 
 * Parse the format tree 
 *
 * Grammar:
 * T -> default
 * T -> "..."
 * T -> { L }
 */
static int parse_fmt_tree (tFmtNode * &ret)
{
    tToken token;

    get_token (token);
    switch (token.type)
    {
        case T_DEFAULT:
            ret = NULL;
            break;
        case T_STRING:
            if (!token.value.str)
                return PARSE_ERROR;
            ret = create_leaf (token.value.str);
            break;
        case T_LBRACE:
            if (parse_list (ret))
                return PARSE_ERROR;

            get_token (token);
            if (token.type != T_RBRACE)
                return PARSE_ERROR;
            break;
        default:
            return PARSE_ERROR;
    }

    return SUCCESS;
}


/* Create and return a tree description of the printout format accordingly
   to the given string. Return NULL on parse error. */
tFmtNode * create_fmt_tree (lib::tMemPool &pool, const char * format)
{
    tFmtNode *ret;

    ASSERT (format);

    fmt_pool = &pool;
    buf = format;
    if (parse_fmt_tree (ret))
        return NULL;    // parse error
    
    // return the tree
    return ret;
}

}
