/**
 * @brief Echelog Library -- Standard configuration file handling
 * @author Ales Privetivy <ales@ucw.cz>
 * @author Kamil Toman <ktoman@email.cz>
 *	
 */

#include "lib/lib.h"
#include "messages/std_conf.h"
#include "messages/msgmgr.h"
#include "messages/types/typemgr.h"

namespace messages {

using namespace lib;

static tList tli (get_msgdef_pool ()),
             mli (get_msgdef_pool ());

static struct cfitem types_section[]={
    {name: "Types",             type: CT_SECTION,       data: NULL},
    {name: "RegisterType",      type: CT_LIST,          data: &tli,  
varf: NULL, varf_reset: NULL,  flags: ITEM_REQUIRED },
    {name: NULL,                type: CT_STOP,          data: NULL}
};

static struct cfitem messages_section[]={
    {name: "Messages",          type: CT_SECTION,       data: NULL},
    {name: "RegisterMessage",   type: CT_FUNCTION,      data: &mli,
varf:  cf_register_message,     varf_reset: cf_unregister_messages },
    {name: NULL,                type: CT_STOP,          data: NULL}
};


/* Registers types, messages sections and do the load from file */
tConfError cf_read_std_configuration (int argc, char *argv[])
{
    /* register section for echelog types and messages configuration */
    cf_register(types_section);
    cf_register(messages_section);

    /* process command line parameters */
    if (cf_process(argc,argv,get_msgdef_pool()))
    {
        log (L_ERROR, "Can't parse configuration or cmdline parameters.");
        return CF_E_PROCESS;
    }
    if (cf_check_required())
    {
        log (L_ERROR, "Missing parameters.");
        return CF_E_REQUIRED;
    }

    /* now registering types and messages from conf file*/
    gb_type_manager.register_type_list(&tli);
    gb_message_manager.register_message_list(&mli);

    /* we don't want to reconfigure types or messages */
    cf_unregister (messages_section[0].name);
    cf_unregister (types_section[0].name);

    return CF_OK;
}


/* Reloads configuration part */
tConfError cf_reread_configuration (lib::tMemPool &pool)
{
    /* Purge previous session parse and configuration data */
    get_cfpool().free_pool ();
    cf_cleanup_registered_sections ();
    cf_clear_loaded_flags();

    /* Load new configuration */
    if (cf_reread (pool))
    {
        log (L_ERROR, "Could not reload a configuration file.");
        return CF_E_READ;
    }

    /* Check if all required parameters are present */
    if (cf_check_required())
    {
        log (L_ERROR, "Missing parameters.");
        return CF_E_REQUIRED;
    }

    return CF_OK;
}


/* Reloads configuration section */
tConfError cf_reread_section (lib::cfitem* section, lib::tMemPool &pool)
{
    tConfError err = CF_OK;

    /* Purge previous session parse and configuration data */
    get_cfpool().free_pool ();
    cf_cleanup_registered_sections ();
    cf_clear_loaded_flags();

    /* register given section */
    cf_register (section);

    /* Load new configuration */
    if (cf_reread (pool))
    {
        log (L_ERROR, "Could not reload a configuration file");
        err = CF_E_READ;
        goto fail;
    }

    /* Check if all required parameters are present */
    if (cf_check_required())
    {
        log (L_ERROR, "Missing parameters.");
        err = CF_E_REQUIRED;
        goto fail;
    }

fail:
    cf_unregister (section->name);

    return err;
}

}

