/**
 * @file
 * @brief Echelog Library -- "Type" handling
 * This type define the basic opertaions for handling "complex" echelog types.
 *      
 * @author Kamil Toman <ktoman@email.cz>
 */

#include "messages/testbuf.h"
#include "messages/types/complex.h"

#define MIN_PRINTOUT_GUESS 50


namespace messages {

    
int32_t tComplex::get_len ()
{
    int32_t j, real_sum = 0;

    for (j = 0; j < get_num_of_components(); j++)
        real_sum += components[j]->get_store_length ();

    /* If we need only ony byte per component
       than the whole complex type is null */
    if (real_sum == get_num_of_components())
        len = 0;
    else
        len = real_sum;

    return len;
}


int32_t tComplex::get_store_length ()
{
    get_len ();
    return len + cint32_size (len);
}


int32_t tComplex::get (tPlainBuf *buf)
{
    int32_t i, bytes_read, real_sum = 0;

    CHECK_BUFFER (buf);
    READ_CINT32 (buf, len, bytes_read);

    if (!len)
        return 1;       /* We read only one byte zero length */

    for (i = 0; i < get_num_of_components(); i++)
    {
        if ((bytes_read = components[i]->get (buf))<= 0)
            return EE_BUF_GENERAL_ERROR;
        real_sum += bytes_read;
    }

    if (len != real_sum)
        return EE_BUF_READ_ERROR;
     
    /* Return number of bytes read */
    return len + cint32_size (len);
}


int32_t tComplex::put (tPlainBuf *buf)
{
    int32_t i, tmp;

    CHECK_BUFFER (buf);

    /* At first update type len field and write it */
    get_len ();
    STORE_CINT32 (buf, len, tmp);

    if (len)
    {
        int32_t bytes_written, real_sum = 0;

        if (buf->check_spare_space (len))
            return EE_BUF_STORE_ERROR;

        for (i = 0; i < get_num_of_components(); i++)
        {
            if ((bytes_written = components[i]->put (buf))<= 0)
                return EE_BUF_GENERAL_ERROR;
            real_sum += bytes_written;
        }

        if (len != real_sum)
            return EE_BUF_STORE_ERROR;
    }

    /* Return number of bytes writen */
    return len + cint32_size (len);
}


long tComplex::sprint (lib::tList &list, lib::tMemPool &pool, 
    const tFmtNode *node)
{    
    using namespace lib;
    int32_t j;

   /* return null pointer for null object */
    if (is_null ())
    {
        list.append (NULL);
        return 0;
    }

    /* use defaults for an empty tree */
    if (!node)
    {
        /* create a default tree in printout temporary pool */
        node = get_default_format ();
        ASSERT (node);

        /* sprint if to given pool */
        return sprint (list, pool, node);
    }

    ASSERT (node->num_of_nodes ==  get_num_of_components ());
    tMemPool sub_pool(get_num_of_components() * MIN_PRINTOUT_GUESS);
    
    last_printout_len = 0;

    /* print each individual component */
    for (j = 0; j < get_num_of_components(); j++)
        last_printout_len += components[j]->sprint (list, pool, 
                node->nodes[j].link);

    return last_printout_len;
}


int tComplex::sscan (lib::tList &string_list)
{
    int32_t i, j, total_scanned = 0, scanned = 0;

    if (string_list.count() < get_num_of_components())
        return 0;

    lib::tList sl;
    sl.shallow_copy (string_list);

    for (i = 0; i < get_num_of_components(); i++)
    {
        if (!(scanned = components[i]->sscan (sl)))
            return 0;
        for (j = 0; j < scanned; j++)
            sl.remove_first ();
        total_scanned += scanned;
    }
    
    return total_scanned;
}


tTYPE_RELATION tComplex::assign (tEchelogType &tp)
{
    int32_t i;
    TEST_ASSIGN ();

    tComplex &tpc = static_cast<tComplex &> (tp);
    len = tpc.len;

    for (i = 0; i < get_num_of_components(); i++)
        components[i]->assign (*tpc.components[i]);

    return TP_IDENTICAL;
}


tTYPE_RELATION tComplex::order (tEchelogType &tp)
{
    bool id_f = false, eq_f = false, less_f = false, greater_f = false;
    int32_t i;

    if (!is_compatible (tp))
        return TP_INCOMPATIBLE;

    if (is_null() && tp.is_null())
        return TP_IDENTICAL;

    tComplex &tpc = static_cast<tComplex &> (tp);

    for (i = 0; i < get_num_of_components(); i++)
    {
        if (!components[i]->is_null() && !tpc.components[i]->is_null())
        {
            /* Both components are non-empty */
            tTYPE_RELATION r = components[i]->order (*tpc.components[i]);
            switch (r)
            {
                case TP_IDENTICAL:
                    if (less_f || greater_f)
                        return TP_CONFLICT;
                    if (!eq_f)
                        id_f = true;
                    break;
                case TP_EQUAL:
                    if (less_f || greater_f)
                        return TP_CONFLICT;
                    id_f = false, eq_f = true;
                    break;
                case TP_LESS:
                    if (id_f || eq_f)
                        return TP_CONFLICT;
                    if (!greater_f)
                        less_f = true;
                    break;
                case TP_GREATER:
                    if (id_f || eq_f)
                        return TP_CONFLICT;
                    if (!less_f)
                        greater_f = true;
                    break;
                default:
                    return r;
            }
        }
        else if (!(components[i]->is_null() && tpc.components[i]->is_null()))
        {
            /* Exactly one of subcomponents is non-empty */
            tTYPE_RELATION r = components[i]->order (*tpc.components[i]);
            switch (r)
            {
                case TP_EQUAL:
                    eq_f = true;
                    if (id_f)
                        id_f = false;
                    break;
                case TP_LESS:
                    if (id_f)
                        id_f = false, eq_f = true;
                    break;
                case TP_GREATER:
                    if (id_f)
                        id_f = false, eq_f = true;
                    break;
                case TP_INCOMPARABLE:
                    if (id_f)
                        id_f = false, eq_f = true;
                    break;
                default:
                    ASSERT (false);
                    break;
            }
        }
    }

    if (id_f)
        return TP_IDENTICAL;
    if (eq_f)
        return TP_EQUAL;
    if (less_f)
        return TP_LESS;
    if (greater_f)
        return TP_GREATER;

    return TP_INCOMPARABLE;
}


int tComplex::merge (tEchelogType &tp)
{
    int32_t i;

    if (!is_compatible (tp))
        return -1; 

    if (!tp.is_null ())
    {
        if (is_null ())
            assign (tp);
        else
        {
            tComplex &tpc = static_cast<tComplex &> (tp);

            for (i = 0; i < get_num_of_components(); i++)
            {
                if (!components[i]->is_null () && !tpc.components[i]->is_null())
                {
                    if (components[i]->merge (*tpc.components[i]) < 0)
                        return -1;
                }
                else if (components[i]->is_null() && 
                        !tpc.components[i]->is_null())

                    components[i]->assign (*tpc.components[i]);
            }
        }
    }

    return 0;
}


tEchelogType *tComplex::get_component_by_path (lib::tList &path) const
{
    int32_t i;

    /* If the request is void, return null */
    if (path.count () <= 0)
        return NULL;
    char *s = static_cast<char *> (path.remove_first ());
    if (!s)
        return NULL;


    /* Try to find a component with the specified name */
    for (i = 0; i < get_num_of_components(); i++)
        if (!strcmp (components_names[i], s))
            break;

    /* If not found return null */
    if (i == get_num_of_components())
        return NULL;

    /* On match, if it was the last component return it */
    if (!path.count ())
        return components[i];

    /* Find a nested component */
    return components[i]->get_component_by_path (path);
}


tEchelogType *tComplex::get_component_by_path (const lib::tList &path) const
{
    using namespace lib;

    tMemPool p (512);
    tList sl (p);
    sl.shallow_copy (path);

    /* Get the component (the list structure is lost) */
    tEchelogType *ret = get_component_by_path (sl);

    p.free_pool();

    return ret;
}

}
