/**
 * @brief Echelog Library -- "Type" handling
 * This type can be used for handling 16bit integers.
 *      
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "messages/testbuf.h"
#include "messages/types/int16.h"

namespace messages {

int32_t tInt16::get (tPlainBuf *buf)
{
    byte tmp;

    CHECK_BUFFER (buf);
    READ_BYTE (buf, tmp);
    if (tmp)
    {
        len = static_cast<int32_t> (tmp);
        if (len != INT16_T_BYTES)
            return EE_BUF_READ_ERROR;
        READ_INT16 (buf, value);
    }
    else
        len = 0;
     
    return len + 1;
}


int32_t tInt16::put (tPlainBuf *buf)
{
    CHECK_BUFFER (buf);
    STORE_BYTE (buf, static_cast<byte> (len));

    if (len)
        STORE_INT16 (buf, value);

    return len + 1;
}

/** Get the default formating tree */
const tFmtNode * tInt16::get_default_format()
{
    static tFmtNode * node = create_fmt_tree (get_prpool (), 
            "\"int16: %i\"");
    return node;
}

/** Get the plain formating tree */
const tFmtNode * tInt16::get_plain_format()
{
    static tFmtNode * node = create_fmt_tree (get_prpool (), 
            "\"%i\"");
    return node;
}

long tInt16::sprint (lib::tList &list, lib::tMemPool &pool, 
        const tFmtNode *node)
{
    using lib::s_asprintf;

    /* return null pointer for null object */
    if (is_null ())
    {
        list.append (NULL);
        return 0;
    }

    /* use defaults for an empty tree */
    if (!node)
    {
        /* create a default tree in printout temporary pool */
        node = get_default_format ();
        ASSERT (node);

        /* sprint if to given pool */
        return sprint (list, pool, node);
    }

    ASSERT (node->num_of_nodes == 0);
    char *ret_str, *format = node->nodes[0].format;

    ASSERT (format);
    last_printout_len = s_asprintf (&ret_str, pool, format, value);
    list.append (ret_str);

    return last_printout_len;
}


int tInt16::sscan (lib::tList &string_list)
{
    using namespace lib;

    if (!string_list.count())
        return 0;

    const tNode *c = string_list.get_first();
    char *s = static_cast<char *> (c->data);

    if (s)
    {
        long v;
        
        if (s_stol (s, &v, INT16_MIN, INT16_MAX))
        {
            len = 0;
            return 0;
        }

        len = INT16_T_BYTES;
        value = static_cast<int16_t> (v);
    }
    else
        len = 0;

    return 1;
}


tTYPE_RELATION tInt16::assign (tEchelogType &tp)
{
    TEST_ASSIGN ();

    tInt16 &tpc = static_cast<tInt16 &> (tp);
    len = tpc.len;
    value = tpc.value;

    return TP_IDENTICAL;
}


tTYPE_RELATION tInt16::order (tEchelogType &tp)
{
    COMMON_ORDER();

    tInt16 &tpc = static_cast<tInt16 &> (tp);
    NUMERIC_ORDER();
}

}
