/**
 * @brief Echelog Library -- "Type" handling
 * This type can be used for handling IPv4 addresses
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "messages/testbuf.h"
#include "messages/types/ipv4addr.h"

namespace messages {

const tFmtNode * tIPv4Addr::get_default_format()
{
    static tFmtNode * node = create_fmt_tree (get_prpool (), 
            "\"IPv4 address %i.%i.%i.%i\"");
    return node;
}

const tFmtNode * tIPv4Addr::get_plain_format()
{
    static tFmtNode * node = create_fmt_tree (get_prpool (), 
            "\"%i.%i.%i.%i\"");
    return node;
}

long tIPv4Addr::sprint (lib::tList &list, lib::tMemPool &pool, 
        const tFmtNode *node)
{
    using lib::s_asprintf;
    using lib::debug_vwlog;

    /* return null pointer for null object */
    if (is_null ())
    {
        list.append (NULL);
        return 0;
    }

    /* use defaults for an empty tree */
    if (!node)
    {
        /* create a default tree in printout temporary pool */
        node = get_default_format ();
        ASSERT (node);

        /* sprint if to given pool */
        return sprint (list, pool, node);
    }

    ASSERT (node->num_of_nodes == 0);
    char *ret_str, *format = node->nodes[0].format;
    int addr[4];

    ASSERT (format);
    for (int i = 0; i < 4; i++)
        addr[i] = ((value >> (i * 8)) & 0xff);
    last_printout_len = s_asprintf (&ret_str, pool, format, 
            addr[0], addr[1], addr[2], addr[3]);
    list.append (ret_str);

    return last_printout_len;
}


int tIPv4Addr::sscan (lib::tList &string_list)
{
    using namespace lib;

    if (!string_list.count())
        return 0;

    const tNode *c = string_list.get_first();
    char *ss = NULL, *s = static_cast<char *> (c->data);

    if (s)
    {
        int i;
        char *t;

        ss = strdup (s);
        if (!(t = strtok (ss, ".")))
            goto err_scan;

        i = 0;
        value = 0;
        do
        {
            long v;

            if (i == 5)
                goto err_scan;

            if (s_stol (t, &v, 0, UCHAR_MAX))
                goto err_scan;

            value |= v << (i * 8);
            i++;
        }
        while ((t = strtok (NULL, ".")));

        if (i != 4)
            goto err_scan;

        len = INT32_T_BYTES;
    }
    else 
        len = 0;

    /* Okay, one string scanned successfully */
    return 1;

  err_scan:
    len = 0;
    xfree (ss);
    return 0;
}


tTYPE_RELATION tIPv4Addr::assign (tEchelogType &tp)
{
    TEST_ASSIGN ();

    tIPv4Addr &tpc = static_cast<tIPv4Addr &> (tp);
    len = tpc.len;
    value = tpc.value;

    return TP_IDENTICAL;
}


tTYPE_RELATION tIPv4Addr::order (tEchelogType &tp)
{
    COMMON_ORDER();

    tIPv4Addr &tpc = static_cast<tIPv4Addr &> (tp);
    NUMERIC_ORDER();
}

}
