/**
 * @brief Echelog Library -- Type Manager Class
 *
 * @author Ales Privetivy <ales@ucw.cz>
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/lib.h"
#include "messages/types/typemgr.h"


namespace messages {

class tTypeMgr gb_type_manager;

using namespace lib;
using namespace private_messages;


static int type_name_compare (const void *a, const void *b)
{
    const tTypeDesc * x = static_cast<const tTypeDesc *> (a);
    const tTypeDesc * y = static_cast<const tTypeDesc *> (b);

    return strcmp (x->name, y->name);
}


static int component_name_compare (const void *a, const void *b)
{
    const tComponentDesc * x = static_cast<const tComponentDesc *> (a);
    const tComponentDesc * y = static_cast<const tComponentDesc *> (b);

    return strcmp (x->name, y->name);
}



tTypeDesc * tTypeMgr::get_typedesc (const char *name) const
{
    tTypeDesc tmp;

    if (!name)
        return NULL;

    tmp.name = name;
    tTypeDesc *tp = static_cast<tTypeDesc *>
        (bsearch (&tmp, type_list, TOTAL_ECHELOG_TYPES, sizeof (tTypeDesc),
                  type_name_compare));

    if (!tp)
        return NULL;

    return tp;
}


void tTypeMgr::register_type(const char *name)
{
    tTypeDesc *tp = get_typedesc (name);

    if (!tp)
        die("Attempting to register unsuported type: '%s'.", name);

    tp->registered = 1;
    debug_log ('c', "Echelog type %s sucessfully registered", name);
}


void tTypeMgr::register_type_list(class lib::tList *list)
{
    int total_types_registered = 0;
    struct tNode const *c;

    FOR_ALL_LIST_ITEMS (c, list)
    {
        register_type(static_cast<char *>(c->data));
        total_types_registered++;
    }

    debug_log ('c', "Registered total #%i echelog types", 
            total_types_registered);
}


int tTypeMgr::check_type (const char *name) const
{
    tTypeDesc *tp = get_typedesc (name);

    if (!tp)
        return 0;

    return tp->registered;
}


tComponentDesc *tTypeMgr::get_component_desc (tTypeDesc &tp,
  const char* comp_name) const
{
    /* Try to find the item */
    tComponentDesc tmp, *tc;
    tmp.name = comp_name;
    tc = static_cast<tComponentDesc *>
        (bsearch (&tmp, tp.type_components, tp.num_of_components, 
                  sizeof (tComponentDesc), component_name_compare));

    /* Return it or null if not found */
    return tc;
}


const char *tTypeMgr::get_component_type_by_desc_path (tTypeDesc &tp, 
        lib::tList &path) const
{
    /* If the request is void, return null */
    if (path.count () <= 0)
        return NULL;
    char *s = static_cast<char *> (path.remove_first ());
    if (!s)
        return NULL;

    /* Try to find the component descriptor */
    tComponentDesc *tc = get_component_desc (tp, s);

    /* If not found return null */
    if (!tc)
        return NULL;

    /* On match (if it was the last component), return its type name */
    if (!path.count ())
        return tc->type_name;

    /* If subordinary descriptor table is empty return null */
    if (!tc->type_descriptor)
        return NULL;

    /* Find a nested component */
    return get_component_type_by_desc_path (*tc->type_descriptor, path);
}


const char *tTypeMgr::get_component_type_by_path (const char *type_name, 
        const lib::tList &path) const
{
    /* Get the type description */
    tTypeDesc *tp = get_typedesc (type_name);

    /* If there is no path to follow return null */
    if (!tp || !tp->registered || !tp->type_components)
        return NULL;

    tMemPool p(512);
    tList ls(p);
    ls.shallow_copy (path);
    return get_component_type_by_desc_path (*tp, ls);
}

}
