/**
 * @brief Echelog Library -- "Type" handling
 * This type can be used for handling time
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include "messages/testbuf.h"
#include "messages/types/unixtime.h"

#include <time.h>

namespace messages {

const tFmtNode * tUnixTime::get_default_format()
{
    static tFmtNode * node = create_fmt_tree (get_prpool (), 
            "\"Date: %Y-%m-%d %H:%M:%S\"");
    return node;
}

const tFmtNode * tUnixTime::get_plain_format()
{
    static tFmtNode * node = create_fmt_tree (get_prpool (), 
            "\"%Y-%m-%d %H:%M:%S\"");
    return node;
}

long tUnixTime::sprint (lib::tList &list, lib::tMemPool &pool, 
        const tFmtNode *node)
{
    using lib::s_asprintf;
    using lib::debug_vwlog;
    using lib::s_strdup;

    struct tm *b_time; 
    char str_buf[1000];
    int s_len;

    /* return null pointer for null object */
    if (is_null ())
    {
        list.append (NULL);
        return 0;
    }

    /* use defaults for an empty tree */
    if (!node)
    {
        /* create a default tree in printout temporary pool */
        node = get_default_format ();
    } 

    /* piskup the format string */
    ASSERT (node);
    ASSERT (node->num_of_nodes == 0);
    char *format = node->nodes[0].format;
    ASSERT (format);

    /* and insert the datum in it */
    b_time = localtime( (const time_t *)(&value) );
    if ((s_len = strftime( str_buf, sizeof(str_buf), format, b_time )) == 0 ) 
        str_buf[0]=0;

    /* append it to the list */
    list.append (s_strdup( str_buf, pool ) );

    return s_len;
}


int tUnixTime::sscan (lib::tList &string_list)
{
    using namespace lib;

    if (!string_list.count())
        return 0;

    const tNode *c = string_list.get_first();
    char *s = static_cast<char *> (c->data);

    if (s)
    {
        struct tm tm_val;

        /* get format string */
        const tFmtNode *node = get_plain_format();
        ASSERT (node);
        ASSERT (node->num_of_nodes == 0);
        char *format = node->nodes[0].format;
        ASSERT (format);

        /* try getting time value */
        if ( strptime (s, format, &tm_val) )
        {
            value = mktime (&tm_val);
            if ( value == -1 ) /* shouldn't happen with good strptime() */
            {
                log (L_WARN, "Unexpected error when converting time value");
                len = 0;
                return 0;
            }
            len = sizeof (time_t);
        }
        else
        {
            len = 0;
            return 0;
        }

    }
    else
        len = 0;

    return 1;
}


tTYPE_RELATION tUnixTime::assign (tEchelogType &tp)
{
    TEST_ASSIGN ();

    tUnixTime &tpc = static_cast<tUnixTime &> (tp);
    len = tpc.len;
    value = tpc.value;

    return TP_IDENTICAL;
}


tTYPE_RELATION tUnixTime::order (tEchelogType &tp)
{
    COMMON_ORDER();

    tUnixTime &tpc = static_cast<tUnixTime &> (tp);
    NUMERIC_ORDER();
}

}
