/**
 * @brief Echelog Library -- "Type" handling
 * This type can be used for handling variable-length strings.
 *      
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/archdep.h"
#include <string.h>

#include "lib/lib.h"
#include "messages/testbuf.h"
#include "messages/types/varchar.h"

namespace messages {

tVarchar::tVarchar (lib::tMemPool &mem_pool_buffer, const char *val)
    : tEchelogType (mem_pool_buffer)
{
    /* let's call the function doing the real stuff */
    set_value (const_cast<char *> (val));
}


tVarchar::~tVarchar ()
{
}


const char * tVarchar::get_value () const
{
    char * retvalue;
    if (!value)
        return 0;
    
    /* let's allocate a piece of memory on a standard heap */
    retvalue = static_cast <char *> (lib::xmalloc (len + 1));

    if (assign_type == A_DIRECT_NON_TERMINATED)
    {
        /* we have to add a termination null */
        memcpy (retvalue, value, len);
        retvalue[len] = 0;
    }
    else
        /* okay, we can just simply make a string copy */
        memcpy (retvalue, value, len + 1);

    return retvalue;
}


void tVarchar::set_value (char * val)
{
    using namespace lib;
    int32_t length;

    if (!val)
        die ("The string in assignment must not be null");

    /* let's allocate some space in the designated mempool */
    length = static_cast<int32_t> (strlen (val));
    value = static_cast<char *> (mempool.alloc (length + 1));
    
    assign_type = A_NORMAL;
    len = length; 
    memcpy (value, val, length + 1);
}


void tVarchar::set_direct_value (const char *val)
{
    using namespace lib;
    if (!val)
        die ("The string in assignment must not be null");

    assign_type = A_DIRECT;
    len = static_cast<int32_t> (strlen (val));
    value = const_cast<char *> (val);
}


void tVarchar::set_direct_value (const char *val, int32_t val_length)
{
    using namespace lib;
    if (!val || val_length <= 0)
        die ("The string in assignment must not be null or nonsential");

    assign_type = A_DIRECT;
    len = val_length;
    value = const_cast<char *> (val);
}


int32_t tVarchar::get (tPlainBuf *buf)
{
    using namespace lib;
    int32_t bytes_read;

    CHECK_BUFFER (buf);
    READ_CINT32 (buf, len, bytes_read);

    if (!len)
        return 1;       /* We read only one byte zero length */
    READ_STRING (buf, value, len, mempool);
    if (len <= 0)
        return EE_BUF_READ_ERROR;

    /* this is not strictly true but we deallocate 
       the memory within the pool deallocation */
    assign_type = A_DIRECT_NON_TERMINATED;

    return len + bytes_read;
}


int32_t tVarchar::put (tPlainBuf *buf)
{
    int32_t bytes_written, tmp = 0;

    CHECK_BUFFER (buf);
    STORE_CINT32 (buf, len, bytes_written);

    if (len && ((tmp = buf->store_string_portion (value, len)) <= 0))
            return tmp;  /* pass the error */

    return bytes_written + tmp;
}


const tFmtNode * tVarchar::get_default_format()
{
    static tFmtNode * node = create_fmt_tree (get_prpool (), 
            "\"varchar: %s\"");
    return node;
}


const tFmtNode * tVarchar::get_plain_format()
{
    static tFmtNode * node = create_fmt_tree (get_prpool (), 
            "\"%s\"");
    return node;
}


long tVarchar::sprint (lib::tList &list, lib::tMemPool &pool, 
        const tFmtNode *node)
{
    using lib::s_asprintf;

    /* return null pointer for null object */
    if (is_null ())
    {
        list.append (NULL);
        return 0;
    }

    /* use defaults for an empty tree */
    if (!node)
    {
        /* create a default tree in printout temporary pool */
        node = get_default_format ();
        ASSERT (node);

        /* sprint if to given pool */
        return sprint (list, pool, node);
    }

    ASSERT (node->num_of_nodes == 0);
    char *ret_str, *format = node->nodes[0].format;

    ASSERT (format);
    last_printout_len = s_asprintf (&ret_str, pool, format, value);
    list.append (ret_str);

    return last_printout_len;
}


int tVarchar::sscan (lib::tList &string_list)
{
    if (!string_list.count())
        return 0;

    const lib::tNode *c = string_list.get_first();
    char *s = static_cast<char *> (c->data);

    assign_type = A_NORMAL;
    if (s)
    {
        len = strlen (s);
        value = static_cast<char *> (mempool.alloc (len + 1));
        memcpy (value, s, len + 1);
    }
    else
    {
        len = 0;
        value = NULL;
    }

    return 1;
}


tTYPE_RELATION tVarchar::assign (tEchelogType &tp)
{
    using namespace lib;

    TEST_ASSIGN ();

    tVarchar &tpc = static_cast<tVarchar &> (tp);
    len = tpc.len;
    assign_type = A_NORMAL;

    if (len)
    {
        /* let's allocate some space in the designated mempool */
        value = static_cast<char *> (mempool.alloc (len + 1));
        
        memcpy (value, tpc.value, len);
        value[len] = 0;
    }
    else
        value = NULL;

    return TP_IDENTICAL;
}


tTYPE_RELATION tVarchar::order (tEchelogType &tp)
{
    using namespace lib;

    COMMON_ORDER();
 
    tVarchar &tpc = static_cast<tVarchar &> (tp);
    int32_t complen = min (len, tpc.len);
    int32_t compr = strncmp (value, tpc.value, complen);

    if (!compr)
    {
        if (len == tpc.len)
            return TP_IDENTICAL;
        else if (len < tpc.len)
            return TP_LESS;
        else
            return TP_GREATER;
    }
    else if (compr > 0)
        return TP_GREATER;
    else
        return TP_LESS;
}

}
