/**
 * @file
 * @brief Contains implementation of tEchelogModule class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "lib/lib.h"
#include "mframes/echelogmodule.h"

namespace mframes {

using namespace lib;

/* static members definition */
volatile bool tEchelogModule::s_sockerr_received = false;

/* Basic filter initialization.*/
tEchelogModule::tEchelogModule (tCmdLine& cmdline, const char* cfg_filename)
    : tReconfigurableModule (cmdline, cfg_filename)
{
    mp_sio = NULL;
    mp_oframe = NULL;
    mp_iframe = NULL;

    /* initial state is created */
    m_emstate = EM_CREATED;
}


/* module's initialization */
bool tEchelogModule::init (tSocketIO* sockio, tIFrame* iframe, tOFrame* oframe)
{
    ASSERT (m_emstate == EM_CREATED);
    ASSERT (sockio != NULL);
    ASSERT (iframe != NULL ||
            oframe != NULL);

    /* store the commio&frames */
    mp_sio = sockio;
    mp_oframe = oframe;
    mp_iframe = iframe;

    /* initialize ancestor */
    if (tReconfigurableModule::init ())
        return true;

    return false;
}


/* module's cleanup */
void tEchelogModule::cleanup ()
{
    tReconfigurableModule::cleanup ();

    mp_sio = NULL;
    mp_oframe = NULL;
    mp_iframe = NULL;
}


/* reconfiguration independent part of initialization */
bool tEchelogModule::static_init ()
{
    ASSERT (m_emstate == EM_CREATED);

    /* call ancestor's init */
    if (tReconfigurableModule::static_init ())
        goto fail;

    /* initialize socket io */
    if (mp_sio->init (&s_sockerr_received))
        goto anc_clean;

    /* if input was given, init it */
    if (mp_iframe != NULL)
    {
        if (mp_iframe->connect_in ())
            goto io_init_clean;
    }

    /* input socket is ready, so register me to let echelogd
     continue its work */
    mp_sio->register_echelog_module (m_cmdline.get_argv(0));

    /* set my iframe as actual */
    tIFrame::set_actual_iframe (mp_iframe);

    /* set my sio as actual */
    tSocketIO::set_actual_socketio (mp_sio);

    /* set my oframe as actual */
    tOFrame::set_actual_oframe (mp_oframe);

    m_emstate = EM_STAT_INITIALIZED;

    return false;

io_init_clean:
    mp_sio->cleanup ();
anc_clean:
    tReconfigurableModule::static_cleanup ();
fail:
    return true;
}


/* reverse method for static_init */
void tEchelogModule::static_cleanup()
{
    ASSERT (m_emstate == EM_STAT_INITIALIZED);

    /* if input was given, init it */
    if (mp_iframe != NULL)
    {
        mp_iframe->close_in ();
    }

    /* no iframe is actual now */
    tIFrame::set_actual_iframe (NULL);

    /* no cio is actual now */
    tSocketIO::set_actual_socketio (NULL);

    /* no oframe is actual now */
    tOFrame::set_actual_oframe (NULL);

    /* cleanup socketio */
    mp_sio->cleanup ();

    /* cleanup ancestor */
    tReconfigurableModule::static_cleanup ();

    m_emstate = EM_CREATED;
}


/* reconfiguration dependent part of initialization */
bool tEchelogModule::dynamic_init()
{
    ASSERT (m_emstate == EM_STAT_INITIALIZED);

    /* call ancestor's dyn_init */
    if (tReconfigurableModule::dynamic_init ())
        goto fail;

    /* pass output info to OFrame & connect outputs*/
    if (mp_oframe != NULL)
        if (mp_oframe->connect_out ())
            goto anc_clean;

    debug_log ('1', "Connected out.");

    /* initialize the target for IFrame*/
    if (mp_iframe != NULL)
        if (mp_iframe->get_target ()->init ())
        {
            log (L_ERROR, "Error during buffer target's initialization.");
            goto out_clean;
        }

    m_emstate = EM_DYN_INITIALIZED;

    return false;

out_clean:
    if (mp_oframe != NULL)
        mp_oframe->close_out ();
anc_clean:
    tReconfigurableModule::dynamic_cleanup ();
fail:
    return true;
}


/* reverse method for dynamic_init */
void tEchelogModule::dynamic_cleanup()
{
    ASSERT(m_emstate == EM_DYN_INITIALIZED);

    /* Close output unix sockets */
    if (mp_oframe != NULL)
       mp_oframe->close_out ();

    /* cleanup the iframe's target */
    if (mp_iframe != NULL)
        mp_iframe->get_target ()->cleanup ();

    /* call ancestor's dyn_clean */
    tReconfigurableModule::dynamic_cleanup ();

    m_emstate = EM_STAT_INITIALIZED;
}


tModule::SigHandlingRet tEchelogModule::handle_signals ()
{
    SigHandlingRet ret = SHR_CONTINUE;

    if (s_sockerr_received)
    {
        ret = on_sockerr ();

        if (ret == SHR_BREAK_OK || ret == SHR_BREAK_ERR)
            return ret;

        s_sockerr_received = false;
    }

    /* call ancestor's handling */
    return tReconfigurableModule::handle_signals ();
}


tModule::SigHandlingRet tEchelogModule::on_sockerr ()
{
    log (L_FATAL, "Fatal socket error has occured, "
        "aborting.");

    return SHR_BREAK_ERR;
}


bool tEchelogModule::register_sections (lib::tMemPool& trash_pool)
{
    /* call ancestor's registering */
    if (tReconfigurableModule::register_sections (trash_pool))
        goto fail;

    /* simply tranfer the request to my configurable components */
    if (mp_sio->register_sections (trash_pool))
        goto anc_clean;

    if (mp_iframe != NULL)
        if (mp_iframe->get_target ()->register_sections (trash_pool))
            goto sio_clean;

    return false;

sio_clean:
    mp_sio->unregister_sections ();
anc_clean:
    tReconfigurableModule::unregister_sections ();
fail:
    return true;
}


void tEchelogModule::unregister_sections ()
{
    /* simply tranfer the request to my configurable components */
    if (mp_iframe != NULL)
        mp_iframe->get_target ()->unregister_sections ();
    mp_sio->unregister_sections ();

    /* call ancestor's unregistering */
    tReconfigurableModule::unregister_sections ();
}


bool tEchelogModule::process_config (lib::tMemPool& conf_mp, 
        lib::tMemPool& trash_pool)
{
    /* call ancestor's processing */
    if (tReconfigurableModule::process_config (conf_mp, trash_pool))
        goto fail;

    /* simply tranfer the request to my configurable components */
    if (mp_sio->process_config (conf_mp, trash_pool))
        goto anc_clean;

    if (mp_iframe != NULL)
        if (mp_iframe->get_target ()->process_config (conf_mp, trash_pool))
            goto sio_clean;

    return false;

sio_clean:
    mp_sio->cleanup_config ();
anc_clean:
    tReconfigurableModule::cleanup_config ();
fail:
    return true;
}


/* reverse method for read_config resp. process_config */
void tEchelogModule::cleanup_config ()
{
    /* simply tranfer the request to my configurable components */
    if (mp_iframe != NULL)
        mp_iframe->get_target ()->cleanup_config ();
    mp_sio->cleanup_config ();

    /* call ancestor's cleanup */
    tReconfigurableModule::cleanup_config ();
}

} //end of namespace mframes

