/**
 * @file
 * @brief Contains implementation of tModule class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "messages/std_conf.h"
#include "mframes/module.h"
#include <signal.h>

namespace mframes {

using namespace lib;

/* static members definition */
volatile bool tModule::s_sigint_received = false;
volatile bool tModule::s_sigquit_received = false;

/* Basic module initialization.*/
tModule::tModule (tCmdLine& cmdline, const char* cfg_filename)
    : m_log_name (NULL), m_conf_mp (4096)
{
    ASSERT (cfg_filename != NULL);

    /* store the command line */
    m_cmdline = cmdline;

    /* set config file name */
    gb_default_config_filename = cfg_filename;

    /* init logging routines */
    log_init (m_cmdline.get_argv(0));

    /* create and fill basic description of common section */
    m_common_section = new cfitem [3];
    if (m_common_section == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        return;
    }

    int i = 0;
    m_common_section[i].name = "Common";
    m_common_section[i].type = CT_SECTION;
    m_common_section[i].data = NULL;
    i++;
    m_common_section[i].name = "ErrFile";
    m_common_section[i].type = CT_STRING;
    m_common_section[i].data = NULL;
    m_common_section[i].varf = NULL;
    m_common_section[i].varf_reset = NULL;
    m_common_section[i].flags = 0;
    i++;
    m_common_section[i].name = NULL;
    m_common_section[i].type = CT_STOP;
    m_common_section[i].data = NULL;

    m_configured = false;
}


/* Final cleanup */
tModule::~tModule ()
{
    delete [] m_common_section;

    /* logging cleanup */
    log_close ();
}


bool tModule::init ()
{
    if (m_common_section == NULL)
        return false;

    /* Configure echelog types and messages */
    if (messages::cf_read_std_configuration (m_cmdline.get_argc (),
                                    m_cmdline.get_argv ()))
    {
        log (L_ERROR, "Unable to load configuration file '%s'",
            cf_get_config_filename ());
        goto conf_clean;
    }

    /* read the whole module's configuration */
    if (read_config (true))
        goto conf_clean;

    /* release parsing data */
    get_cfpool().free_pool ();

    /* install all signal handlers */
    install_sig_handlers ();

    return false;

conf_clean:
    get_cfpool().free_pool ();

    return true;
}


void tModule::cleanup ()
{
    /* cleanup whole module's configuration */
    release_config ();
}


bool tModule::reread_configuration (tMemPool& cf_pool, bool apply_cmdline)
{
    /* Purge previous session parse and configuration data */
    cf_cleanup_registered_sections ();
    cf_clear_loaded_flags();

    /* Load new configuration */
    if (cf_reread (cf_pool))
    {
        debug_log ('c', "Could not reload a configuration file.");
        goto fail;
    }
    /* apply cmd line */
    if (apply_cmdline)
    {
        if (cf_process (m_cmdline.get_argc (),
                        m_cmdline.get_argv (),
                        cf_pool))
        {
            debug_log ('c', "Can't parse configuration "
                "or cmdline parameters.");
            goto fail;
        }
    }
    /* Check if all required parameters are present */
    if (cf_check_required ())
    {
        debug_log ('c', "Missing parameters.");
        goto fail;
    }

    return false;
fail:
    return true;
}


/* reads and prepares configuration for further use */
bool tModule::read_config (bool accept_cmdline)
{
    using namespace messages;

    /* Register all sections */
    if (register_sections (get_cfpool()))
        goto fail;

    log (L_INFO, "Using configuration file '%s'.", cf_get_config_filename ());

    /* Read filter configuration */
    if (reread_configuration (get_cfpool(), accept_cmdline))
    {
        debug_log ('c', "Unrecoverable errors in configuration, aborting.");
        goto register_clean;
    }

    if (process_config (m_conf_mp, get_cfpool()))
    {
       debug_log ('c', "Errors in configuration.");
       goto register_clean;
    }

    /* Deregister all sections */
    unregister_sections ();

    /* All has been duplicated (processed) so remove the trash */
    get_cfpool().free_pool ();

    m_configured = true;

    return false;

register_clean:
    unregister_sections ();
    get_cfpool().free_pool ();
fail:
    return true;
}


/* reverse method for read_config */
void tModule::release_config ()
{
    /* it is allowed to call this repeatedly during cleanup */
    /* ie. it is possible, that module is unconfigured during cleanup */
    if (m_configured)
    {
        cleanup_config ();

        /* Release config data */
        m_conf_mp.free_pool ();
        debug_log ('1', "Released configuration pool.");

        m_configured = false;
    }
}


bool tModule::register_sections (lib::tMemPool& trash_pool)
{
    m_common_section[1].data = &m_log_name;
    cf_register (m_common_section);

    return false;
}


void tModule::unregister_sections ()
{
    cf_unregister (m_common_section[0].name);
    m_common_section[1].data = NULL;
}


bool tModule::process_config (lib::tMemPool& conf_mp, lib::tMemPool& trash_pool)
{
    /* duplicate strings to m_conf_mp */
    if (m_log_name != NULL)
        m_log_name = s_strdup (m_log_name, conf_mp);

    return false;
}


/* reverse method for read_config resp. process_config */
void tModule::cleanup_config ()
{
    m_log_name = NULL;
}


/* signal handling function instalation */
void tModule::install_sig_handlers ()
{
    set_signal_handler (SIGINT, tModule::sig_handling_f);
    set_signal_handler (SIGQUIT, tModule::sig_handling_f);
}


/* A trivial signal transfering function */
void tModule::sig_handling_f (int signum)
{
    switch (signum)
    {
        case SIGINT:
            s_sigint_received = true;
            break;
        case SIGQUIT:
            s_sigquit_received = true;
            break;
    }
}


tModule::SigHandlingRet tModule::handle_signals ()
{
    SigHandlingRet ret = SHR_CONTINUE;

    if (s_sigint_received)
    {
        ret = on_sigint ();

        if (ret == SHR_BREAK_OK || ret == SHR_BREAK_ERR)
            return ret;
    }

    if (s_sigquit_received)
    {
        ret = on_sigquit ();

        if (ret == SHR_BREAK_OK || ret == SHR_BREAK_ERR)
            return ret;
    }

    return ret;
}


tModule::SigHandlingRet tModule::on_sigint ()
{
    log(L_INFO, "Termination request received, attempting to exit gracefully.");
    s_sigint_received = false;

    return SHR_BREAK_OK;
}


tModule::SigHandlingRet tModule::on_sigquit ()
{
    log(L_INFO, "Termination request received, attempting to exit gracefully.");
    s_sigquit_received = false;

    return SHR_BREAK_OK;
}


} //end of namespace mframes

