/**
 * @file
 * @brief Contains implementation of tReconfigurableModule class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "mframes/reconfigurablemodule.h"
#include <signal.h>

namespace mframes {

using namespace lib;

/* static members definition */
volatile bool tReconfigurableModule::s_sighup_received = false;

/* Basic filter initialization.*/
tReconfigurableModule::tReconfigurableModule (tCmdLine& cmdline, const char* cfg_filename)
    : tModule (cmdline, cfg_filename)
{
    /* initial state is created */
    m_rmstate = RM_CREATED;
}


/* module's initialization */
bool tReconfigurableModule::init ()
{
    ASSERT (m_rmstate == RM_CREATED);

    /* call ancestor's init */
    if (tModule::init ())
        goto fail;

    /* do the static inititialization */
    if (static_init ())
        goto anc_clean;
    else
        m_rmstate = RM_STAT_INITIALIZED;

    /* do the dynamic inititialization */
    if (dynamic_init ())
        goto static_clean;
    else
        m_rmstate = RM_DYN_INITIALIZED;

    return false;

static_clean:
    static_cleanup();
    m_rmstate = RM_CREATED;
anc_clean:
    tModule::cleanup ();
fail:
    return true;
}


/* module's cleanup */
void tReconfigurableModule::cleanup ()
{
    /* module's dynamic cleanup */
    if (m_rmstate == RM_DYN_INITIALIZED)
    {
        dynamic_cleanup ();
        m_rmstate = RM_STAT_INITIALIZED;
    }

    /* module's static cleanup */
    if (m_rmstate == RM_STAT_INITIALIZED)
        static_cleanup ();

    /* call ancestor's cleanup */
    tModule::cleanup ();
}


bool tReconfigurableModule::static_init ()
{
    return false;
}


void tReconfigurableModule::static_cleanup ()
{
}


bool tReconfigurableModule::dynamic_init ()
{
    /* redirect the logging */
    if (log_file (m_log_name))
    {
        log (L_ERROR, "Error while redirecting log file to '%s'.",
            m_log_name ? m_log_name : "(null)");
    }

    return false;
}


void tReconfigurableModule::dynamic_cleanup ()
{
}


/* reconfiguration implementation */
bool tReconfigurableModule::reinit()
{
    /* Announce this */
    log (L_INFO, "Reloading configuration.");

    /* at first cleanup */
    dynamic_cleanup();
    m_rmstate = RM_STAT_INITIALIZED;

    /* cleanup previous config */
    release_config ();

    /* reread config */
    if (read_config ())
        goto fail;

    /* then init */
    if (dynamic_init ())
        goto fail;

    m_rmstate = RM_DYN_INITIALIZED;
    log (L_INFO, "Successfully reconfigured");

    return false;

fail:
    return true;
}


/* signal handling function instalation */
void tReconfigurableModule::install_sig_handlers ()
{
    tModule::install_sig_handlers ();
    set_signal_handler (SIGHUP, tReconfigurableModule::sig_handling_f);
}


/* A trivial signal transfering function */
void tReconfigurableModule::sig_handling_f (int signum)
{
    switch (signum)
    {
        case SIGHUP:
            s_sighup_received = true;
            break;
    }
}


tModule::SigHandlingRet tReconfigurableModule::handle_signals ()
{
    SigHandlingRet ret = tModule::handle_signals ();

    if (ret == SHR_BREAK_OK || ret == SHR_BREAK_ERR)
        return ret;

    if (s_sighup_received)
    {
        ret = on_sighup ();

        if (ret == SHR_BREAK_OK || ret == SHR_BREAK_ERR)
            return ret;
    }

    return ret;
}


tModule::SigHandlingRet tReconfigurableModule::on_sighup ()
{
    /* do the reload */
    bool ret = reinit ();
    s_sighup_received = false;

    if (ret)
        return SHR_BREAK_ERR;
    else
        return SHR_CONTINUE;
}


bool tReconfigurableModule::run ()
{
    ASSERT (m_rmstate == RM_DYN_INITIALIZED);

    bool ret = false;
    SigHandlingRet hret = SHR_CONTINUE;
    int pret = PR_CONTINUE;

    for (;;)
    {
        pret = process ();
        if (pret == PR_BREAK_ERR)
        {
            ret = true;
            break;
        }
        if (pret == PR_BREAK_OK)
        {
            ret = false;
            break;
        }

        hret = handle_signals ();

        if (hret == SHR_BREAK_ERR)
            ret = true;

        if (hret != SHR_CONTINUE)
            break;
    }

    if (ret)
        log (L_INFO, "Terminated with non-zero status.");
    else
        log (L_INFO, "Successfully terminated.");

    return ret;
}

} //end of namespace mframes

