/**
 * @file
 * @brief Contains implementation of tSocketIO class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "mframes/socketio.h"

namespace mframes {

using namespace lib;

/* static members definition */
tSocketIO* tSocketIO::sp_actual_sockio;
const int tSocketIO::SIO_CREATED = 1;
const int tSocketIO::SIO_CONFIGURED = 2;
const int tSocketIO::SIO_INITIALIZED = 4;

/* Basic filter initialization.*/
tSocketIO::tSocketIO ()
     : m_input_name (NULL), m_oia (NULL), m_oia_count (0)
{
    /* create and fill basic description of system section */
    m_system_section = new cfitem [4];
    if (m_system_section == NULL)
        die ("Not enough memory.");

    int i = 0;
    m_system_section[i].name = "System";
    m_system_section[i].type = CT_SECTION;
    m_system_section[i].data = NULL;
    i++;
    m_system_section[i].name = "Input";
    m_system_section[i].type = CT_STRING;
    m_system_section[i].data = NULL;
    m_system_section[i].varf = NULL;
    m_system_section[i].varf_reset = NULL;
    m_system_section[i].flags = 0;
    i++;
    m_system_section[i].name = "Output";
    m_system_section[i].type = CT_LIST;
    m_system_section[i].data = NULL;
    m_system_section[i].varf = NULL;
    m_system_section[i].varf_reset = NULL;
    m_system_section[i].flags = 0;
    i++;
    m_system_section[i].name = NULL;
    m_system_section[i].type = CT_STOP;
    m_system_section[i].data = NULL;

    /* initial state is created */
    m_siostate = SIO_CREATED;
}


/* Final cleanup */
tSocketIO::~tSocketIO ()
{
    delete [] m_system_section;
}


/* module's initialization */
bool tSocketIO::init (volatile bool* socket_error)
{
    ASSERT (socket_error != NULL);
    ASSERT (m_siostate & SIO_CONFIGURED);

    /* save the address for error flag */
    mp_socket_error = socket_error;

    m_siostate |= SIO_INITIALIZED;

    return false;
}

/* module's initialization */
void tSocketIO::register_echelog_module (const char* mod_name)
{
    ASSERT (mod_name != NULL);
    ASSERT (m_siostate & SIO_INITIALIZED);

    /* Setup and register the control socket */
    m_ctrl.setup_defaults (mod_name);
    if (!m_ctrl.register_echelog_module (m_smgr))
        log (L_INFO, "Successfully registered at echelogd.");
}


/* module's cleanup */
void tSocketIO::cleanup ()
{
    ASSERT (m_siostate & SIO_INITIALIZED);

    m_ctrl.close_socket ();
    m_smgr.cleanup ();
    mp_socket_error = NULL;

    m_siostate &= ~SIO_INITIALIZED;
}


bool tSocketIO::process (int timeout)
{
    ASSERT (m_siostate & SIO_INITIALIZED);

    //TODO: just for now behaviour should be revised
    if  (m_smgr.process (timeout) == -1)
        return true;

    return false;
}


bool tSocketIO::register_sections (lib::tMemPool& trash_pool)
{
    tList* output_list = new tList (trash_pool);
    if (output_list == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        return true;
    }

    /* Init filter configuration sections */
    m_system_section[1].data = &m_input_name;
    m_system_section[2].data = output_list;

    cf_register (m_system_section);

    return false;
}


void tSocketIO::unregister_sections ()
{
    cf_unregister (m_system_section[0].name);

    delete static_cast<lib::tList *> (m_system_section[2].data);

    m_system_section[1].data = NULL;
    m_system_section[2].data = NULL;
}


bool tSocketIO::process_config (lib::tMemPool& conf_mp, 
        lib::tMemPool& trash_pool)
{
    ASSERT (m_siostate & SIO_CREATED);

    tMemPool list_pool (1024);
    tList list (list_pool);
    tNode const* n = NULL;
    int i = 0;

    /* duplicate strings to m_conf_mp */
    if (m_input_name != NULL)
        m_input_name = s_strdup (m_input_name, conf_mp);

    /* prepare buffer for output info */
    tList& output_list =
        *static_cast<tList*> (m_system_section[2].data);
    m_oia_count = output_list.count ();
    if (m_oia_count)
    {
        m_oia = new tOutputInfo [m_oia_count];
        if (m_oia == NULL)
        {
            log (L_ERROR, "Not enough memory.");
            goto fail;
        }
    }

    /* fill output info */
    FOR_ALL_LIST_ITEMS (n, &output_list)
    {
        debug_log ('1', "Parsing: %s", static_cast<char *> (n->data));
        if (s_strpsep (list, static_cast<char*> (n->data), conf_mp))
        {
            log (L_ERROR, "Configuration error in \"Output\" value.");
            goto oia_clean;
        }
        if (list.count () != 2)
        {
            log (L_ERROR, "Configuration error in \"Output\" value.");
            list.reset ();
            goto oia_clean;
        }
        m_oia[i].name = static_cast<char*> (list.get_first()->data);
        m_oia[i].path = static_cast<char*> (list.get_last()->data);
        debug_log ('1', "Output - name: %s path: %s", m_oia[i].name,
            m_oia[i].path);
        i++;
        list.reset ();
        list_pool.free_pool ();
    }

    m_siostate |= SIO_CONFIGURED;

    return false;

oia_clean:
    if (m_oia_count)
        delete [] m_oia;
    m_oia = NULL;
    m_oia_count = 0;
fail:

    return true;
}


/* reverse method for read_config resp. process_config */
void tSocketIO::cleanup_config ()
{
    ASSERT (m_siostate & SIO_CONFIGURED);

    m_input_name = NULL;
    if (m_oia_count)
        delete [] m_oia;
    m_oia = NULL;
    m_oia_count = 0;

    m_siostate &= ~SIO_CONFIGURED;
}

} //end of namespace mframes

