/**
 * @file
 * @brief Contains implementation of tUNIXSocketIFrame, tUNIXSocketOFrame
 * tUNIXSocketIOFrame classes
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "mframes/unixsocketioframe.h"

namespace mframes {

using namespace lib;
using namespace socks;

//-------------------------------
// Implementation of tUNIXSocketIFrame
//-------------------------------

/* Set up a frame and fill in basic information */
tUNIXSocketIFrame::tUNIXSocketIFrame ()
{
    m_connected = false;
}


bool tUNIXSocketIFrame::connect_in ()
{
    ASSERT (is_initialized ());
    ASSERT (mp_target != NULL);
    ASSERT (!m_connected);

    if (m_input_name == NULL)
    {
        log (L_ERROR, "Path for input socket was not found in "
            "configuration.");
        goto fail;
     }

    /* listen at input */
    m_input.set_on_data_hup (tUNIXSocketIFrame::hup_f);
    m_input.set_on_data_err (tUNIXSocketIFrame::err_f);
    m_input.set_on_data_read (tUNIXSocketIFrame::process_f);
    if (m_input.create_server (m_input_name))
    {
        log (L_ERROR, "Cannot create listenning unix domain socket "
                        "(%s).", m_input_name);
        goto fail;
    }

    /* register the listening socket */
    if (m_smgr.register_socket (&m_input))
    {
        log (L_ERROR, "Cannot register socket in socket manager.");
        goto socket_clean;
    }

    m_connected = true;

    return false;

socket_clean:
    /* close output unix socket */
    if (m_input.close_socket ())
        log (L_WARN, "Cannot close the socket.");
fail:
    return true;
}


void tUNIXSocketIFrame::close_in ()
{
    ASSERT (m_connected);

    if (m_input.get_smgr ()->unregister_socket (&m_input))
        log (L_ERROR, "Cannot unregister socket.");
    if (m_input.close_socket ())
        log (L_ERROR, "Cannot close socket.");

    m_connected = false;
}


/* socket hangup handler */
void tUNIXSocketIFrame::hup_f (class tSocket *sock)
{
    tSocketIO* sio = tSocketIO::get_actual_socketio ();

    ASSERT (sio != NULL);

    if (sock->get_role () == AS_SOCK_SERVER)
    {
        sio->set_sock_err ();
        log (L_FATAL, "Server listening socket broken. Aborting.");
    }
    else
    {
        tSocketMgr *m = sock->get_smgr();
        if (m)
             if(m->unregister_socket (sock))
                log (L_ERROR, "Error while unregistering socket.");
        if (sock->close_socket ())
            log (L_ERROR,"Error closing socket.");
        if (sock->get_created_by_smgr ())
            delete sock;
    }
}


/* socket error handler */
void tUNIXSocketIFrame::err_f (class tSocket *sock)
{
    tSocketIO* sio = tSocketIO::get_actual_socketio ();

    ASSERT (sio != NULL);

    if (sock->get_role () == AS_SOCK_SERVER)
    {
        log (L_FATAL, "Server listening socket broken. Aborting.");
        sio->set_sock_err ();
    }
    else
    {
        tSocketMgr *m = sock->get_smgr();
        if (m)
            if (m->unregister_socket (sock))
                log (L_ERROR, "Error while unregistering socket.");
        if (sock->close_socket())
            log (L_WARN,"Error closing socket.");
        if (sock->get_created_by_smgr ())
            delete sock;
        log (L_INFO,"Error on client connection socket cleared.");
    }
}


/* process incoming data */
void tUNIXSocketIFrame::process_f (socks::tSocket *sock)
{
    int len;
    tIFrame* fr = tIFrame::get_actual_iframe ();

    ASSERT (fr != NULL);
    ASSERT (fr->get_target () != NULL);

    /* load incoming message */
    if ((len = sock->at_read (s_message_buffer, MAX_MESSAGE_SIZE))<0)
    {
        log (L_WARN,"Error receiving data.");
        return;
    }    	

    debug_log ('1', "Received %d bytes.", len);

    /* pass the buffer */
    if (fr->get_target ()->process_buffer (s_message_buffer, len))
        log (L_WARN, "Error occured during the processing of received data.");
}


//-------------------------------
// Implementation of tOutputInt
//-------------------------------

bool tUNIXSocketOFrame::tOutputInt::send_buffer (char* buf, int len)
{
    if ((m_output.at_write (buf, len)) < 0)
    {
        log (L_ERROR, "Can't pass message to the next module via output '%s'",
            m_name);
        return true;
    }

    return false;
}


//-------------------------------
// Implementation of tUNIXSocketOFrame
//-------------------------------

/* Set up a frame and fill in basic information */
tUNIXSocketOFrame::tUNIXSocketOFrame ()
{
    m_connected = false;
}


/* Connects to the next modules in echelog topology */
bool tUNIXSocketOFrame::connect_out ()
{
    ASSERT (is_initialized ());
    ASSERT (!m_connected);

    if (m_oia_count < 1)
    {
        debug_log ('s', "No output socket configuration found. ");
        m_connected = true;
        return false;
    }

    ASSERT (m_oia != NULL);
    int i = 0;
    m_output_count = m_oia_count;

    /* prepare array for outputs */
    m_outputs = new tOutput* [m_output_count];
    if (m_outputs == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        goto fail;
    }
    for (i = 0; i < m_output_count; i++)
    {
        m_outputs[i] = new tOutputInt;
        if (m_outputs[i] == NULL)
        {
            log (L_ERROR, "Not enough memory.");
            goto outputs_clean;
        }
    }

    /* initialize each output */
    for (i = 0; i < m_output_count; i++)
    {
        ASSERT (m_oia[i].name != NULL);
        ASSERT (m_oia[i].path != NULL);

        tOutputInt* output =
            static_cast<tOutputInt*>(m_outputs[i]);
        output->set_name (m_oia[i].name);
        output->get_socket ().set_on_data_hup (tUNIXSocketOFrame::s_err_f);
        output->get_socket ().set_on_data_err (tUNIXSocketOFrame::s_err_f);

        /* connect output unix socket */
        if (output->get_socket ().create_client (m_oia[i].path))
        {
            log (L_ERROR, "Cannot connect to subsequent module via "
                "output '%s' (unix domain socket: '%s').",
                output->get_name (), m_oia[i].path);
            goto socket_clean;
        }

        /* register socket in socket manager to automate I/O */
        if (m_smgr.register_socket (&output->get_socket ()))
        {
            log (L_ERROR, "Cannot register socket in socket manager.");
            goto socket_close_clean;
        }
    }

    m_connected = true;

    return false;

socket_close_clean:
    /* close actual socket */
    if (static_cast<tOutputInt*>(m_outputs[i])->
        get_socket ().close_socket ())
        log (L_ERROR, "Can't close socket.");
socket_clean:
    /* close all previously opened sockets */
    for(i--; i >= 0; i--)
    {
        socks::tUNIXSocket& output =
            static_cast<tOutputInt*>(m_outputs[i])->get_socket ();

        if (m_smgr.unregister_socket (&output))
            log (L_ERROR, "Error while unregistering socket.");
        if (output.close_socket ())
            log (L_ERROR, "Can't close socket.");
    }
    i = m_output_count;
outputs_clean:
    for(i--; i >= 0; i--)
        delete static_cast<tOutputInt*> (m_outputs[i]);
    /* delete allocated memory */
    delete [] m_outputs;
    m_outputs = NULL;
    m_output_count = 0;
fail:
    return true;
}


/* closes output sockets */
void tUNIXSocketOFrame::close_out ()
{
    ASSERT (m_connected);

    for (int i = m_output_count-1; i >= 0; i--)
    {
        socks::tUNIXSocket& output =
            static_cast<tOutputInt*>(m_outputs[i])->get_socket ();

        if (output.get_smgr ()->unregister_socket (&output))
            log (L_ERROR, "Error while unregistering socket.");
        if (output.close_socket ())
            log (L_ERROR, "Can't close socket.");
    }
    for (int i = m_output_count-1; i >= 0; i--)
        delete static_cast<tOutputInt*> (m_outputs[i]);
    delete [] m_outputs;
    m_outputs = NULL;
    m_output_count = 0;

    m_connected = false;
}


/* socket error handler */
void tUNIXSocketOFrame::s_err_f (class tSocket *sock)
{
    tSocketIO* sio = tSocketIO::get_actual_socketio ();

    ASSERT (sio != NULL);

    sio->set_sock_err ();
    log (L_FATAL, "Output connection broken. Aborting.");
}

} //end of namespace mframes
