/**
 * @file
 * @brief Echelog Socks library -- Echelog Control Command Layer
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/lib.h"
#include "socks/control.h"
#include <string.h>


namespace socks {

using namespace lib;    
using namespace private_socks;


/* CONST_PROPERTY implementation */
const char *tControl::get_ctrl_name () const
{
    return ctrl_name;
}


/* CONST_PROPERTY implementation */
void tControl::set_ctrl_name (const char *name)
{
    /* Assign a only new control channel name */
    if (!ctrl_name)
    {
        ctrl_name = lib::s_strdup (name);
        ctrl_namelen = strlen (name) + 1;
    }
}

            
tCommand *tControl::buf2cmd (void *buffer) const
{
    tCmdHeader *header = static_cast<tCmdHeader *> (buffer);
    tCommand *c = static_cast<tCommand *> (lib::xmalloc (sizeof (tCommand)));

    /* Translate the command frame buffer into a tCommand structure */
    c->command = header->command;
    c->d_len = header->ad_len;
    if (c->d_len)
    {
        /* Allocate neccessary space and make a copy of ancillary data */
        c->d_ptr = lib::xmalloc (c->d_len);
        memcpy (c->d_ptr, header->ad_buffer, c->d_len);
    }
    else
        c->d_ptr = NULL;

    /* Return the resulting structure */
    return c;
}


tControl::tControl ()  
    : ctrl_name (NULL), ctrl_namelen (0), retry_count (CTRL_CONN_RETRY), 
      cmd_in_buffer (NULL), cmd_out_buffer (NULL) 
{
    /* Initialize standard ctrl communication buffers */
    cmd_in_buffer = xmalloc (MAX_CMD_LENGTH);
    cmd_out_buffer = xmalloc (MAX_CMD_LENGTH);
}


tControl::~tControl ()
{
    /* Do a small exercise to free data where a pointer to constant points */
    char *s = const_cast<char *> (ctrl_name);
    ctrl_name = NULL;
    xfree (s);

    /* Release resources used by standard ctrl communication buffers */
    xfree (cmd_in_buffer);
    xfree (cmd_out_buffer);
}


void tControl::store_cmd (const tCMD_TYPE command, 
        const void * ac_buf, long ac_buflen)
{
    tCmdHeader *dest_header = static_cast<tCmdHeader *> (cmd_out_buffer);

    /* At first fill in a protocol version number to avoid 
     * a possible misinterpretation */
    dest_header->protocol_version = ECHELOG_CTRL_PROTOCOL_VERSION;

    /* Then just translate fractioned command parameters */
    dest_header->command = command;
    dest_header->ad_len = 
        min (ac_buflen, static_cast<long>(MAX_CMD_DATA_LENGTH));

    /* And copy (at most MAX_CMD_DATA_LENGTH) ancillary data to the 
     * output command buffer */
    if (dest_header->ad_len > 0)
        memcpy (&dest_header->ad_buffer[0], ac_buf, dest_header->ad_len);
}

            
void tControl::store_cmd (const tCommand *c)
{
    tCmdHeader *dest_header = static_cast<tCmdHeader *> (cmd_out_buffer);

    /* At first fill in a protocol version number to avoid 
     * a possible misinterpretation */
    dest_header->protocol_version = ECHELOG_CTRL_PROTOCOL_VERSION;

    /* Fill in the command frame headers */
    dest_header->command = c->command;
    dest_header->ad_len = 
        min (c->d_len, static_cast<long>(MAX_CMD_DATA_LENGTH));

    /* And copy (at most MAX_CMD_DATA_LENGTH) ancillary data to the 
     * output command buffer */
    if (dest_header->ad_len > 0)
        memcpy (&dest_header->ad_buffer[0], c->d_ptr, dest_header->ad_len);
}


tCommand *tControl::get_cmd ()
{
    /* If there is no valid command structure in the input buffer
     * return a NULL pointer */
    if (test_received_cmd ())
        return NULL;
    
    /* Copy-out the command otherwise */
    return buf2cmd (cmd_in_buffer);
}

}
