/**
 * @file
 * @brief Echelog Socks library -- Echelog Control Command Layer
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/lib.h"
#include "socks/socket.h"
#include "socks/ctrl_unixsocket_nb.h"

#include <sys/types.h>
#include <sys/un.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>


namespace socks {

using namespace lib;    

    
/* This enumerates all possible registration phases */
enum tNotificationStatus {
   NS_NONE,                 // Nothing happened
   NS_OK,                   // Authentication successful
   NS_BAD,                  // Bad authentication status
   NS_SND_IDENTIFICATION,   // Authentication sent, awaiting reply
   NS_RCV_CONFIRMATION      // Receiving confirmation
};

/* This is a structure referenced by a control socket. It holds information
 * about the registration phase and status */
struct tCmdNotification
{
    tNotificationStatus op_status;  // Operation status
    tCMD_TYPE command;         // Used command type
    int ad_len;                     // Ancillary command data length
    const void *ad_ptr;             // Pointer to ancillary command data
};


/* Control socket default hangup handler */
static void default_on_ctrl_hup (class tSocketNB *sock)
{
    tCtrlUNIXSocketNB *s = static_cast<tCtrlUNIXSocketNB *> (sock);
    tSocketMgr *mgr = s->get_smgr();
    log (L_INFO, "Hangup received on control channel.");

    /* Cleanup the socket */
    s->close_socket ();
    mgr->unregister_socket (s);

    /* May we try it again? */
    if (!s->test_retries ())
    {
        log (L_INFO, "Control socket closed. Trying to reopen the connection.");
        if (s->register_echelog_module (*mgr, s->get_ctrl_name ()) < 0)
        {
            log (L_WARN, "Could not reestablish the connection on control "
                    "socket. Giving up.");
            mgr->unregister_socket (s);
            delete s;
        }
        else
            log (L_INFO, "Reconnected on control socket.");
    }
    else
        /* Print the error message and cleanup the socket */
        log (L_WARN, "Number of retries exceeded. Giving up.");
}


/* Control socket default error handler */
static void default_on_ctrl_err(class tSocketNB *sock)
{
    tCtrlUNIXSocketNB *s = static_cast<tCtrlUNIXSocketNB *> (sock);
    tSocketMgr *mgr = s->get_smgr();
    log (L_INFO, "Control socket error");

    /* Cleanup the socket */
    s->close_socket ();
    mgr->unregister_socket (s);

    /* May we try it again? */
    if (!s->test_retries ())
    {
        log(L_INFO, "Control socket cleared. Trying to reopen the connection.");
        if (s->register_echelog_module (*mgr, s->get_ctrl_name ()) < 0)
        {
            log (L_WARN, "Could not reestablish the connection on control "
                    "socket. Giving up.");
            mgr->unregister_socket (s);
            delete s;
        }
        else
            log (L_INFO, "Reconnected on control socket.");
    }
    else
        /* Print the error message and cleanup the socket */
        log (L_WARN, "Number of retries exceeded. Giving up.");
}


/* On control socket data available */
static void default_on_ctrl_read_avail (class tSocketNB *sock)
{
    tCtrlUNIXSocketNB *s = static_cast<tCtrlUNIXSocketNB *> (sock);

    /* Schedule the control command receival */
    if (s->receive_cmd () < 0)
        log (L_WARN, "Error receiving data from the control socket.");
}


/* On control socket data available */
static void default_on_ctrl_read (class tSocketNB *sock)
{
    tCtrlUNIXSocketNB *s = static_cast<tCtrlUNIXSocketNB *> (sock);
    tCommand *c;
    
    /* Try to retrieve a command from the input buffer */
    if (!(c = s->get_cmd ()))
    {
        log (L_WARN, "Invalid control packet received. Discarding it.");
        return ;
    }


    /* Execute a pre-defined action */
    switch (c->command)
    {
        case CMD_NOP:   // No operation
            break;
            
        case CMD_PING:  // Ping request
            debug_log ('t', "scheduling PONG");

            /* Store and immediately schedule the pong response */
            s->store_cmd (CMD_PONG);
            s->send_cmd ();
            break;

        default:        // No predefined action
            debug_log ('t', "No response implemented for control command %x",
                    c->command);
    }

    /* Release the received command structure */
    free_cmd (c);
}


/* Socket hangup/err handler during module registration */
static void reg_abort_registration (tSocketNB *sock)
{
    /* Retrieve the notification structure */
    tCmdNotification *n = static_cast<tCmdNotification *> 
        (sock->get_app_specific_data ());
    ASSERT (n);

    debug_log ('t', "Hangup/Error - Aborting registration");

    /* Switch to final unsuccessful status */
    n->op_status = NS_BAD;

    /* Cleanup the socket */
    sock->close_socket ();
    sock->get_smgr()->unregister_socket (sock);
}


/* Incoming data during the module registration */
static void reg_on_response_available (tSocketNB *sock)
{
    /* Retrieve the notification structure */
    tCtrlUNIXSocketNB *s = static_cast<tCtrlUNIXSocketNB *> (sock);
    tCmdNotification *n = static_cast<tCmdNotification *> 
        (s->get_app_specific_data ());
    ASSERT (n);

    /* On an unexepected operation status, switch to the final 
     * unsuccessful status */
    if (n->op_status != NS_RCV_CONFIRMATION)
        n->op_status = NS_BAD;

    /* Receive incoming command in any case */
    if (s->receive_cmd () < 0)
        log (L_WARN, "Error receiving data from the control socket.");
}


/* Data received during the module registration */
static void reg_on_response_read (tSocketNB *sock)
{
    /* Retrieve the notification structure */
    tCtrlUNIXSocketNB *s = static_cast<tCtrlUNIXSocketNB *> (sock);
    tCmdNotification *n = static_cast<tCmdNotification *> 
        (s->get_app_specific_data ());
    ASSERT (n);

    debug_log ('t', "Received a response from server.");

    /* On an unexepected operation state, switch to the final 
     * unsuccessful status */
    if (n->op_status != NS_RCV_CONFIRMATION
            || s->test_received_cmdtype (CMD_LOGIN_OK))
        n->op_status = NS_BAD;
    else 
        /* The registration has been confirmed, mark as successful */
        n->op_status = NS_OK;
}


/* Data sent notifier function during the module registration */
static void reg_on_ident_sent (tSocketNB *sock)
{
    /* Retrieve the notification structure */
    tCmdNotification *n = static_cast<tCmdNotification *> 
        (sock->get_app_specific_data ());
    ASSERT (n);

    /* On an unexepected operation state, switch to the final 
     * unsuccessful status */
    if (n->op_status != NS_SND_IDENTIFICATION)
        /* We have no limburgers, sir! */
        n->op_status = NS_BAD;
    else
        /* Okay, the whole packet has been sent. Waiting for a reply. */
        n->op_status = NS_RCV_CONFIRMATION;
};


/* Connection with overseer established notifier function */
static void reg_sched_registration_packet (tSocketNB *sock)
{
    /* Retrieve the notification structure */
    tCtrlUNIXSocketNB *s = static_cast<tCtrlUNIXSocketNB *> (sock);
    tCmdNotification *n = static_cast<tCmdNotification *> 
        (s->get_app_specific_data ());
    ASSERT (n);

    debug_log ('t', "Sending identification to echelogd.");

    /* Store and schedule the authentication request */
    s->store_cmd (n->command, n->ad_ptr, n->ad_len);
    if (s->send_cmd () < 0)
        /* Do you have any cheese at all? No, sir. Arrrgh! */
        n->op_status = NS_BAD;
    else
        /* Authentication request has been scheduled */
        n->op_status = NS_SND_IDENTIFICATION;
}


int tCtrlUNIXSocketNB::register_echelog_module (tSocketMgr &smanager,
        const char *name, long max_registration_time)
{
    int r = 0;
    set_ctrl_name (name);
    ASSERT (ctrl_name);

    /* Backup socket handlers */
    tSocketNBHandlerFunc last_err_handler = get_on_err();
    tSocketNBHandlerFunc last_hup_handler = get_on_hup();
    tSocketNBHandlerFunc last_on_connected_handler = get_on_connected();
    tSocketNBHandlerFunc last_write_finished = get_on_write_finished ();
    tSocketNBHandlerFunc last_read_available = get_on_read_available ();
    tSocketNBHandlerFunc last_read_finished = get_on_read_finished ();

    /* Set new temporary handlers */
    set_on_hup (reg_abort_registration);
    set_on_err (reg_abort_registration);
    set_on_connected (reg_sched_registration_packet);
    set_on_write_finished (reg_on_ident_sent);
    set_on_read_available (reg_on_response_available);
    set_on_read_finished (reg_on_response_read);

    /* Backup the old write notification routine and app data */
    void *old_app_data = get_app_specific_data ();

    /* Set the notification structure */
    tCmdNotification *notification = 
        static_cast<tCmdNotification *> (xmalloc (sizeof (tCmdNotification)));
    notification->op_status = NS_NONE;
    notification->ad_len = ctrl_namelen;
    notification->ad_ptr = ctrl_name;
    notification->command = CMD_IDENTIFY;
    set_app_specific_data (static_cast<void *> (notification));

    /* Try to connect to overseer */
    if (create_client (DEFAULT_CTRL_SOCKET, smanager) < 0)
    {
        log (L_WARN, "Can't create a client control socket.");
        r = -1;
        goto restore_sock;
    }
    reconnect();
    
    /* Compute the time limit */
    long io_time;
    struct timeval limit_tv, tv, dtime, ptime;

    /* Transform the time format of the desired maximal connection interval */
    dtime.tv_sec = max_registration_time / 1000;
    dtime.tv_usec = (max_registration_time - dtime.tv_sec * 1000) * 1000;

    /* Set the deadline */
    gettimeofday (&tv, NULL);
    timeradd (&tv, &dtime, &limit_tv);
    io_time = max_registration_time;
    
    /* Try to register within the specified interval */
    do 
    {
        /* Process the socket i/o. */
        smanager.process (io_time);

        /* Compute if there is time left */
        gettimeofday (&tv, NULL);
        timersub (&limit_tv, &tv, &ptime);
        io_time = ptime.tv_sec * 1000 + ptime.tv_usec / 1000;
    }
    while (notification->op_status != NS_OK &&
           notification->op_status != NS_BAD &&
           io_time > 0);

    debug_log ('t', "Registration status: %d", notification->op_status);

   restore_sock:

    /* Now fear of the reality */
    if (notification->op_status != NS_OK)
    {
        log (L_WARN, "Can't register at echelogd.");
        r = -1;
    }
    
    /* Free the callback status structure */
    xfree (notification);

    /* Write original information back to socket structure */
    set_app_specific_data (old_app_data);

    /* Restore socket callbacks */
    set_on_read_finished (last_read_finished);
    set_on_read_available (last_read_available);
    set_on_write_finished (last_write_finished);
    set_on_connected (last_on_connected_handler);
    set_on_hup (last_hup_handler);
    set_on_err (last_err_handler);


    /* Report success if any and finish the job */
    if (!r)
    {
        debug_log ('t', "Successfully registered at echelogd.");

        /* Now we consider the connection established, call user's handler */
        if (last_on_connected_handler)
            last_on_connected_handler (this);
    }

    /* Propagate the resulting status */
    return r;
}


void tCtrlUNIXSocketNB::setup_defaults (const char *name)
{
    /* Name the connection channel after the module name */
    set_ctrl_name (name);

    /* Set pre-defined command handlers and simple cleanups
     * if something goes wrong */
    set_on_hup (default_on_ctrl_hup);
    set_on_err (default_on_ctrl_err);
    set_on_read_available(default_on_ctrl_read_avail);
    set_on_read_finished (default_on_ctrl_read);
}


/* Send control command */
int tCtrlUNIXSocketNB::send_cmd ()
{
    using namespace private_socks;
    tCmdHeader *h = static_cast<tCmdHeader *> (cmd_out_buffer);

    /* Schedule a transfer of the control command */
    return at_write (cmd_out_buffer, sizeof (tCmdHeader) + h->ad_len);
}


/* Receive control command */
int tCtrlUNIXSocketNB::receive_cmd ()
{
    /* Schedule the control command receival */
    return at_read (cmd_in_buffer, MAX_CMD_LENGTH);
}


}
