/**
 * @file
 * @brief Echelog Socks library -- Non-blocking sockets
 *
 * @author Ales Privetivy <ales@ucw.cz>
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/archdep.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>

#include "socks/socket_nb.h"



namespace socks {

using namespace lib;


// Handler for hangup
void tSocketNB::hup_handler ()
{
    debug_log ('s', "Hup-handler called (sock=%d).", get_fd());

    status = SS_DEFAULT;
    
    if (on_hup) 
        on_hup(this);

}


// Socket error handler
void tSocketNB::err_handler ()
{
    debug_log ('s', "Err-handler called (sock=%d).", get_fd ());

    status = SS_BROKEN;

    if (on_err) 
        on_err (this);

}


// Socket reading handler
void tSocketNB::read_handler ()
{
    ASSERT (role == AS_SOCK_CLIENT);
    ASSERT (status == SS_CONNECTED);

    debug_log ('s', "Read-handler called (sock=%d).", get_fd ());

    // continue started operation
    switch (last_operation)
    {
        case OP_HANDSHAKE:
            debug_log ( 's', "Read handler: OP_HANDSHAKE");
            if (created_by_smgr) 
                accept_again();
            else 
                connect_again(); 
            break;
        case OP_IO:
            switch (r_io.last_op)
            {
                case IO_OP_READING:
                    debug_log ('s', "Read handler: OP_READING");
                    read_again();
                    break;
                case IO_OP_ATOM_READING:
                    debug_log ('s', "Read handler: OP_ATOM_READING");
                    atom_read_again();
                    break;
                case IO_OP_NO_OPERATION:
                    debug_log ('s', "Read handler: data available "
                            "(OP_NO_OPERATION)");
                    if (on_read_available) 
                        on_read_available(this);
                    break;
                default:
                    ;
            }
            switch (w_io.last_op)
            {
                case IO_OP_WRITING:
                    debug_log ('s', "Read handler: OP_WRITING");
                    write_again();
                    break;
                case IO_OP_ATOM_WRITING:
                    debug_log ('s', "Read handler: OP_ATOM_WRITING");
                    atom_write_again();
                    break;
                default:
                    ;
            }
            break;
        case OP_NO_OPERATION:
            debug_log ('s', "Read handler: OP_NO_OPERATION");
            if (on_read_available) 
                on_read_available(this);
            break;
        default:
            ;
    } 
}


// Socket writing handler
void tSocketNB::write_handler ()
{
    int connect_bad;
    socklen_t v_len = sizeof(int);

    ASSERT (role == AS_SOCK_CLIENT);
    ASSERT (status == SS_CONNECTED || status == SS_CONNECTING );

    debug_log ('s', "Write-handler called (sock=%d).", get_fd());

    if (status == SS_CONNECTING)
    {
        // not fully connected yet
        debug_log ('s', "Checking the connection state");

        if (getsockopt (get_fd(), SOL_SOCKET, SO_ERROR, &connect_bad, &v_len ))
        {
            // unreadable state
            log( L_ERROR, "Cannot obtain socket state: %m");
            err_handler();
            return;
        }

        if(connect_bad)
        {
            log (L_ERROR, "Error connecting to server: Connection refused.");
            err_handler();
            return;
        }

        // seems okay
        status = SS_CONNECTED;
        debug_log ('s', "Sock=%d sucesfully connected.", get_fd() );

        // finish the connection protocol
        reconnect();
        return;
    }

    if (status == SS_CONNECTED)
    {    
        // fully connected
        switch (last_operation)
        {
            case OP_HANDSHAKE:
                debug_log ( 's', "Write handler: OP_HANDSHAKE");
                if (created_by_smgr) 
                    accept_again();
                else 
                    connect_again(); 
                break;
            case OP_IO:
                switch (r_io.last_op)
                {
                    case IO_OP_READING:
                        debug_log ('s', "Write handler: OP_READING");
                        read_again();
                        break;
                    case IO_OP_ATOM_READING:
                        debug_log ('s', "Write handler: OP_ATOM_READING");
                        atom_read_again();
                        break;
                    default:
                        ;
                }
                switch (w_io.last_op)
                {
                    case IO_OP_WRITING:
                        debug_log ('s', "Write handler: OP_WRITING");
                        write_again();
                        break;
                    case IO_OP_ATOM_WRITING:
                        debug_log ('s', "Write handler: OP_ATOM_WRITING");
                        atom_write_again();
                        break;
                    default:
                        ;
                }
                break;
            case OP_NO_OPERATION:
                debug_log ('s', "Write handler: OP_NO_OPERATION");
                smgr->set_event_mask (this, POLLPRI | POLLIN);
                break;
            default:
                ;
        } 
    }
}


// Write
ssize_t tSocketNB::sock_write (void *buffer, ssize_t len)
{

    ASSERT (status == SS_CONNECTED);
    ASSERT ((last_operation == OP_NO_OPERATION) || 
            (last_operation == OP_IO && (w_io.last_op == IO_OP_NO_OPERATION)));
    ASSERT (buffer && len >= 0);

    debug_log ('s', "SOCK_WRITE sheduled on sock=%d.", get_fd());

    w_io.last_op_buf = static_cast<char *>(buffer);
    w_io.last_op_bytes = len;
    w_io.last_op = IO_OP_WRITING;
    last_operation = OP_IO;
    smgr -> set_event_mask (this, POLLPRI | POLLIN | POLLOUT);

    return 0; 

}


// Read
ssize_t tSocketNB::sock_read (void *buffer, ssize_t len)
{
    ASSERT (status == SS_CONNECTED);
    ASSERT ((last_operation == OP_NO_OPERATION) ||
            (last_operation == OP_IO && (r_io.last_op == IO_OP_NO_OPERATION)));
    ASSERT (buffer && len >= 0);

    debug_log ('s', "SOCK_READ sheduled on sock=%d.", get_fd());

    r_io.last_op_buf = static_cast<char *>(buffer);
    r_io.last_op_bytes = len;
    r_io.last_op = IO_OP_READING;
    last_operation = OP_IO;
    
    return 0;
}


// Atomic read
ssize_t tSocketNB::at_read (void *buffer, ssize_t len)
{
    ASSERT(status == SS_CONNECTED);
    ASSERT ((last_operation == OP_NO_OPERATION) ||
            (last_operation == OP_IO && (r_io.last_op == IO_OP_NO_OPERATION)));
    ASSERT (buffer && len >= 0);

    debug_log ('s', "ATOM_READ sheduled on sock=%d.", get_fd());

    r_io.last_op_buf = r_io.last_op_buf_cur = static_cast<char *>(buffer);
    r_io.last_op_bytes = len;
    r_io.already_processed = 0;
    r_io.last_op = IO_OP_ATOM_READING;
    last_operation = OP_IO;
    r_io.atom_op_state = AT_BEGINFRAME;

    return 0;
}


// Atomic write
ssize_t tSocketNB::at_write(void *buffer, ssize_t len)
{
    ASSERT(status == SS_CONNECTED);
    ASSERT ((last_operation == OP_NO_OPERATION) || 
            (last_operation == OP_IO && (w_io.last_op == IO_OP_NO_OPERATION)));
    ASSERT (buffer && len >= 0);

    debug_log ('s', "ATOM_WRITE sheduled on sock=%d.", get_fd());

    w_io.last_op_buf = w_io.last_op_buf_cur = static_cast<char *>(buffer);
    w_io.last_op_bytes = len;
    w_io.already_processed = 0;
    w_io.last_op = IO_OP_ATOM_WRITING;
    last_operation = OP_IO;
    w_io.atom_op_state = AT_BEGINFRAME;

    w_io.atom_frame.magic = V_MAGIC;
    w_io.atom_frame.length = len;
    w_io.atom_frame.flags = 0;
    smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);

    return 0;
}


// Set blocking socket behaviour
int tSocketNB::set_blocking ()
{
    ASSERT (fd >= 0);

    // make the socket blocking
    if ( fcntl( get_fd(), F_SETFL, 0 ) )
    {
        log (L_ERROR, "Can't make socket blocking: %m");
        return -1;
    }

    return 0;
}


// Set non-blocking socket behaviour
int tSocketNB::set_non_blocking ()
{
    ASSERT (fd >= 0);

    // make the socket nonblocking
    if ( fcntl( get_fd(), F_SETFL, O_NONBLOCK ) )
    {
        log (L_ERROR, "Can't make socket nonblocking: %m");
        return -1;
    }

    return 0;
}


}
