/**
 * @brief Echelog Socks library -- Socket manager
 *
 * Performs check for event on set of sockets. Each socket must be
 * registered first.
 *
 * @author Ales Privetivy <ales@ucw.cz>
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/archdep.h"
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>

#include "lib/lib.h"
#include "socks/socketmgr.h"


namespace socks {

using namespace lib;


tSocketMgr::~tSocketMgr ()
{
    cleanup();
}


/// cleanup all registered sockets
void tSocketMgr::cleanup ()
{
    long i;

    for (i = 0; i < sockcnt; i++)
    {
        tSocket *s = socket[i];
        unregister_socket (s);
        if (s->get_created_by_smgr ())
            delete s;
        else
            if (s->close_socket ())
                log (L_ERROR, "Error closing socket (fd=%ld)", s->get_fd ());
    }
}

/// cleanup all created by manager sockets
void tSocketMgr::cleanup_created ()
{
    long i;

    for (i = 0; i < sockcnt; i++)
    {
        tSocket *s = socket[i];

        if (s->get_created_by_smgr ())
        {
            if (unregister_socket (s) < 0)
                log (L_ERROR, "Can't unregister socket (fd=%ld)", s->get_fd ());
            delete s;
        }    

    }
}

// registers socket with given event mask and handler array
int tSocketMgr::register_socket(class tSocket *s, int ev_mask, long parent)
{
    using namespace lib;

    // is it a reasonable socket?
    ASSERT (s);
    ASSERT (s->get_status()==SS_CONNECTED || s->get_status()==SS_LISTENING 
            || s->get_status()==SS_CONNECTING );

    // number of polled sockets is limited
    if (sockcnt >= SERVING_MAXSOCKS) 
    {
        log (L_WARN, "Program limit on registered sockets exceeded");
        return -1;
    }

    // insert the new socket structure
    socket[sockcnt] = s;
    // the file descriptor
    ufds[sockcnt].fd = s->get_fd();

    // server part
    if (s->get_role() == AS_SOCK_SERVER) 
        // during accept we will be interested in this events
        ufds[sockcnt].events = POLLIN | POLLPRI;
    else 
        ufds[sockcnt].events = ev_mask;

    // and also update socket structures fields
    s->set_event_mask (ev_mask);
    s->set_sid (map.get_bit ());
    s->set_psid (parent);
    s->set_smgr (this);
    s->set_smgr_handle (sockcnt);

    debug_log ('s', "Registering the socket %ld (pos=%ld).", s->get_fd(),
            sockcnt);

    sockcnt ++;

    if (s->get_on_register() && s->get_role() != AS_SOCK_SERVER) 
        (s->get_on_register())(s);

    return 0;
}


// sets event mask
void tSocketMgr::set_event_mask(class tSocket *s, int ev_mask)
{
    ASSERT(s);
    long handle = s->get_smgr_handle ();
    
    if(handle >= 0 && handle < sockcnt)
        // change the mask
        ufds[handle].events = ev_mask; 
}


// unregister the socket
int tSocketMgr::unregister_socket(class tSocket *s)
{
    using namespace lib;

    ASSERT(s);
    long i = s->get_smgr_handle();
    ASSERT(i >= 0);

    debug_log ('s', "Unregistering the socket %ld (pos=%ld).", s->get_fd(), i);
    
    // free socket mapping id
    map.free_bit (s->get_sid ());
    s->set_sid (-1);
    s->set_smgr(NULL);

    // update poll index
    if (i <= pidx)
        pidx = max (0, pidx - 1);
    
    // and shift the sockets data behind one place left
    for(; i < sockcnt - 1; i++) 
    {
        ufds[i] = ufds[i+1];
        socket[i] = socket[i+1];
        socket[i]->set_smgr_handle(i);
    }

    // and now we have one socket less
    sockcnt--;

    if (s->get_on_unregister() && s->get_role() != AS_SOCK_SERVER) 
        (s->get_on_unregister())(s);
    return 0;
}


// poll the sockets with given timeout
int tSocketMgr::process(int timeout)
{
    using namespace lib;

    int e, re, hits, nchanges;

    pidx = 0;

   poll_refresh:
    if ((nchanges = poll(ufds, sockcnt, timeout)) < 0)
    {
         if (errno != EINTR)
         {
             log (L_ERROR, "SocketMgr: Poll error: %m");
             return -1;
         }

         // Processing interrupted by a signal
         return 1;
    }
    else if (!nchanges)
        // no event occured (timed out)
        return 0;
    
    // refreshing polls should not block
    timeout = 0;

    // now we check one per one socket till we find everything
    for (hits = 0; pidx < sockcnt && hits < nchanges; pidx++)
    {
        ASSERT (socket[pidx]);

        // mask of events we want to catch
        e=ufds[pidx].events;
        // mask of events occured
        re=ufds[pidx].revents;

        if (re)
        {
            debug_log ('s', "Poll events on socket %d:%s%s%s%s%s%s", 
                socket[pidx]->get_fd(),
                (re & POLLIN) ? " Incoming data," : "", 
                (re & POLLPRI) ? " Incoming priority data," : "",
                (re & POLLHUP) ? " Hangup," : "",
                (re & POLLOUT) ? " Non blocking write," : "",
                (re & POLLERR) ? " Poll error," : "", 
                (re & POLLNVAL) ? " Invalid request," : "");
            hits++;
        }

        // invalid request (fd not opened)
        if (re & POLLNVAL)
        {
            tSocket *s = socket[pidx];

            log(L_ERROR, "Polling on invalid socket %ld", s->get_fd ());
            if (s->get_on_data_err ()) 
                (s->get_on_data_err ())(s);

            return -1;
        }

        // it is listening socket
        if (socket[pidx]->get_role() == AS_SOCK_SERVER)
        {
            // error or hangup on listenning sockets shouldnt occur
            if (re & POLLERR || re & POLLHUP) 
            {
                log(L_ERROR, "Error on listenning socket.");
                return -1;
            }

            // hah! somebody wants to be accepted
            if (re & POLLPRI || re & POLLIN)
            {
                // accept connection and refresh socket information
                if (socket[pidx++]->accept_socket(*this))
                {
                    log (L_ERROR, "Error accepting socket");
                    continue;
                }

                // new socket added - new refresh
                goto poll_refresh;
            }
        }

        // find out result of currently connecting socket 
        if (socket[pidx]->get_status()==SS_CONNECTING)
        {    
                if (socket[pidx]->get_on_data_write ())
                   (socket[pidx]->get_on_data_write ()) (socket[pidx]);

                goto poll_refresh;
        }
        
        // any data to be read ?
        if ((re & POLLIN) || (re & POLLPRI))
        {
            int n;
            char b;

            /* If there is no data to recieve suppose hangup.
             * BSDs and SunOS machines don't bother to set POLLHUP
             * so this may be the only way out.
             */
            n = socket[pidx]->sock_poke (&b, 1);
            if (!n)
                goto pollhup;
            if (n < 0)
            {
                if (n == -EINTR) // Processing interrupted by a signal
                    return 1;
                else
                    goto pollerr;
            }

            // Okay, we have some prioritized data to be read
            if (e & re & POLLPRI)
                if (socket[pidx]->get_on_data_read ())
                    (socket[pidx]->get_on_data_read ()) (socket[pidx]);

            // Okay, this seems that we really have data to be read
            if (e & re & POLLIN)
                if (socket[pidx]->get_on_data_read ())
                    (socket[pidx]->get_on_data_read ()) (socket[pidx]);

            pidx++;
            goto poll_refresh;
        }

        // error occured on socket (connection broken)
        if (re & POLLERR)
        {
           pollerr:
            debug_log ('s', "Error occured on socket %d", 
                    socket[pidx]->get_fd());
            // if we were during connect it is connect error
            // otherwise it signalizes broken connection
            if(socket[pidx]->get_status()!= SS_CONNECTING)
                    socket[pidx]->set_status(SS_BROKEN);
            if (socket[pidx]->get_on_data_err ())
                (socket[pidx]->get_on_data_err ()) (socket[pidx]);
            continue;
        }

        // hungup on the other side
        if (re & POLLHUP)
        {
           pollhup:
            if (socket[pidx]->get_on_data_hup ())
                (socket[pidx]->get_on_data_hup ()) (socket[pidx]);

            continue;
        }
        else    // is socket ready to write?
        { 
            if (e & re & POLLOUT) 
                if (socket[pidx]->get_on_data_write ())
                    (socket[pidx]->get_on_data_write ()) (socket[pidx]);
        }
            

    } // for 

    return 0;
}

}
