/**
 * @brief Echelog Socks library -- SSL context
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include "lib/archdep.h"
#include "lib/lib.h"

#define ZLIB
#include "socks/sslcontext.h"
#undef ZLIB

namespace socks {

using namespace lib;

/* flag if SSL library was initialized */
static int ssl_library_initilalized=0;

/* callback for private key files passphrase */
static int pem_passwd_cb(char *buf, int size, int rwflag, void *password)
{
    return 0;
}

/* for fututre use
static int enable_rle();
static int enable_zlib();
*/


static void ssl_lib_init()
{
    /* If first executed then inicialize SSL library */
    if (!ssl_library_initilalized)
    {
        SSL_library_init();
        SSL_load_error_strings();
        ssl_library_initilalized = 1;
    }
     
}


/* creates new SSL context ver 3 */
tSSLContext::tSSLContext()
{
    SSL_METHOD *meth;

    ssl_lib_init();

    /* Create our context*/
    meth = SSLv3_method();
    context = SSL_CTX_new(meth);

    /* check if context created successfully */
    if (!context)
    {
        debug_log( 'l', "SSLCONTEXT: Cannot create SSL context.");
    }

    SSL_CTX_set_verify(context, SSL_VERIFY_PEER | 
                       SSL_VERIFY_FAIL_IF_NO_PEER_CERT, 0);
}


/* frees SSL context */
tSSLContext::~tSSLContext()
{
    /* do it only if context is still present */
    if (context)
        SSL_CTX_free(context);
}


/* set certificate and private key */
int tSSLContext::set_key_files(const char *pk_file, const char *cert_file,
                               const char *CA_cert_file)
{
    /* set callback for passphrases on private key files */
    SSL_CTX_set_default_passwd_cb( context, pem_passwd_cb);

    /* load our certificate*/
    if(!(SSL_CTX_use_certificate_file( context, cert_file, SSL_FILETYPE_PEM )))
    {
        debug_log( 'l', "SSLCONTEXT: Couldn't read certificate file (\"%s\").", 
                cert_file);
        return -1;
    }

    /* load our private key */
    if(!(SSL_CTX_use_PrivateKey_file( context, pk_file, SSL_FILETYPE_PEM )))
    {
        debug_log( 'l', "SSLCONTEXT: Couldn't read private key file (\"%s\").",
                pk_file);
        return -1;
    }

    /* Load the CAs we trust*/
    if(!(SSL_CTX_load_verify_locations( context, CA_cert_file, NULL)))
    {
        debug_log( 'l', "SSLCONTEXT: Couldn't read CA-certificate (\"%s\").",
                CA_cert_file);
        return -1;
    }

    /* we verify the certificates directly */
    SSL_CTX_set_verify_depth( context, 1 );

    return 0;
}

        
/* set certificate, private key and dh_file */
int tSSLContext::set_key_files_with_dh(const char *pk_file, 
        const char *cert_file, const char *CA_cert_file, const char *dh_file)
{
    DH *dh_ret=0;
    BIO *bio;

    /* set certificate and private key */
    if (set_key_files( pk_file, cert_file, CA_cert_file )) return -1;
        
    /* open the dh-file */
    if ((bio=BIO_new_file(dh_file,"r")) == NULL)
    {
        debug_log( 'l', "SSLCONTEXT: Couldn't read dh-paramfile (\"%s\").",
                dh_file );
        return -1;
    }

    /* create the dh object */
    dh_ret = PEM_read_bio_DHparams(bio,NULL,NULL,NULL);

    /* free the BIO */
    BIO_free(bio);
    
    /* set the dh-parameters to context */ 
    if( SSL_CTX_set_tmp_dh( context, dh_ret ) < 0 )
    {
        debug_log( 'l', "SSLCONTEXT: Couldn't set the dh-parameters.");
        return -1;
    }

    return 0;
}


/* This code is here for future use
static int enable_zlib()
{
 * This code is not "correct" for now:
 * The first argument to SSL_COMP_add_compression_method() is supposed to be
 * the TLS algorithm number.
 *  
 * Of course, since there isn't any algorithm number assigned to ZLIB for TLS
 * for now, the number you use is completely at our leasure :-).  Don't count
 * on the following code to work when such a number is assigned, however if
 * you use this library on both sides the constants will be same and  the code
 * should work.
    

    COMP_METHOD * cm = COMP_zlib();
  
    if (!cm|| cm->type == NID_undef)
    {
        debug_log('l',"SSLCONTEXT: Openssl library was not compiled with ZLIB "
                      "option.");
        return -1;
    }
    if (SSL_COMP_add_compression_method(8, cm)) return 0;
        else return -1;
          
}


static int enable_rle()
{
 * This code is not "correct" for now:
 * The first argument to SSL_COMP_add_compression_method() is supposed to be
 * the TLS algorithm number.
 *
 * Of course, since there isn't any algorithm number assigned to RLE for TLS
 * for now, the number you use is completely at our leasure :-).  Don't count
 * on the following code to work when such a number is assigned, however if
 * you use this library on both sides the constants will be same and  the code
 * should work.


   COMP_METHOD * cm = COMP_rle();

   if (!cm || cm->type == NID_undef)
   {
       debug_log('l',"SSLCONTEXT: Openssl library does not contain RLE "
                     "support.");
       return -1;
   }

   if (SSL_COMP_add_compression_method(4, cm)) return 0;
       else return -1;
}
*/

}
