/**
 * @brief Echelog Socks library - SSL TCP Socket
 *
 * @author Ales Privetivy <ales@ucw.cz> 
 * @author Kamil Toman <ktoman@email.cz>
 * 
 */

#include "lib/archdep.h"
#include "lib/lib.h"
#include "socks/sslsocket.h"
#include "socks/socketmgr.h"
#include "messages/plainbuf.h"
#include "messages/testbuf.h"
#include "socks/compress.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <openssl/err.h>
#include <arpa/inet.h>

extern int h_errno;

namespace socks {

using namespace lib;


/* handlers of events in tcp-layer */
static void _hup_f(class tSocket *sock) { sock->hup_handler (); }
static void _read_f(class tSocket *sock) { sock->read_handler (); }
static void _write_f(class tSocket *sock) { sock->write_handler (); }
static void _err_f(class tSocket *sock) { sock->err_handler (); }

// constructor - create socket and set file descriptor
tSSLSocket::tSSLSocket() : tSocketNB (), 
    on_ssl_shutdown_init (NULL), on_ssl_shutdown_finished (NULL),
    ssl (NULL)
{
    // set both (tcp and ssl) status to empty
    set_status (SS_EMPTY);
    set_ssl_status (SL_EMPTY);
}


// closes the socket
tSSLSocket::~tSSLSocket()
{

    // perform the tcp shutdown
    if (close_socket())
        log (L_ERROR, "Can't close socket (fd=%ld)", get_fd());

    // clear out of application specific data 
    xfree (app_specific_data); 

    // free allocated space for hostname
    if (hostname)
    {
        xfree (hostname);
        hostname = NULL;
    }
   
}


int tSSLSocket::new_socket ()
{
    long new_fd = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);

    // socket cannot be opened
    if (new_fd < 0)
    {
        log (L_ERROR, "SSLSOCKET: Can't create new socket: %m");
        set_status (SS_ERRSOCK);
        return -1;
    }

    // assign it to object 
    set_fd (new_fd); 

    // make the socket nonblocking
    if (set_non_blocking() )
    {
        set_status (SS_ERRSOCK);
        return -1;
    }

    // socket is created but not connected
    set_status (SS_DEFAULT);

    return 0;
}


// release ssl resources 
int tSSLSocket::close_ssl()
{
    ssl_status = SL_EMPTY;
    if (ssl)
    {
        SSL_free (ssl);
        ssl = NULL;
    }

    return 0;
}


// closes the socket and unregisters it 
int tSSLSocket::close_socket()
{
    int r = 0;
    close_ssl();
   
    // try to close the socket
    if (get_fd()>=0 && s_close (get_fd() ) )
    {
        log (L_ERROR, "SSLSOCKET: Error closing socket: %m");
        r = -1;
    }

    // inactivate the tcp layer and  invalidate the fd
    set_status (SS_EMPTY);
    set_fd(-1);
    return r;
}


// creates client side of the socket
int tSSLSocket::create_client (const char *ahostname, const int aport,
                               class tSSLContext *ctx, 
                               class tSocketMgr *smanager)
{
    ASSERT (smanager && ahostname && ctx);

    // get new socket
    if (new_socket()) 
        return -1;

    // backup hostname
    if (!(hostname = strdup (ahostname)))
    {
        log(L_ERROR,"SSLSOCKET: Not enough memory.");
        return -1;
    }

    // backup socket manager object 
    smgr = smanager; 
  
    // backup port
    port = aport;

    // backup context
    context = ctx->get_context();
    
    // mark that this is client side
    set_role(AS_SOCK_CLIENT);

    // set the event mask 
    set_event_mask (POLLPRI | POLLIN | POLLOUT);

    // set the handlers
    set_on_data_read (_read_f);
    set_on_data_write (_write_f);
    set_on_data_err (_err_f);
    set_on_data_hup (_hup_f);

    return 0;
}


// creates server side of the socket
int tSSLSocket::create_server (const char *ahostname, const int aport, 
                               class tSSLContext *ctx, 
                               class tSocketMgr *smanager)
{
    int sock_opt_val=1;
  
    ASSERT (smanager && ahostname && ctx);

    // mark that this is server side
    set_role(AS_SOCK_SERVER);

    // get new socket
    if (new_socket()) 
    {
        return -1;
    }    

    // backup hostname
    if (!(hostname = strdup (ahostname)))
    {
        log(L_ERROR,"SSLSOCKET: Not enough memory.");
        status = SS_ERRRESOURCES;
        return -1;
    }

    // backup port
    port = aport;

    // backup context
    context = ctx->get_context();

    // fill in server address structure
    addr.sin_addr.s_addr = htonl (INADDR_ANY);
    addr.sin_family = AF_INET;
    addr.sin_port = htons(aport);

    // disallow socket reusal
    setsockopt (get_fd(), SOL_SOCKET, SO_REUSEADDR, &sock_opt_val, sizeof(int));

    // bound the socket to specified port
    if (bind (get_fd(), (struct sockaddr *) &addr, sizeof(addr))) 
    {
        log(L_ERROR, "SSLSOCKET: Cannot bind socket to port %d: %m", port);
        status =  SS_ERRBIND;
        return -1;
    }

    // make the socket listenning for requests
    if (listen (get_fd(), LISTENCNT) )
    {
        log(L_ERROR, "SSLSOCKET: Cannot make socket listening: %m");
        status =  SS_ERRLISTENING;
        return -1;
    }

    // now we are listenning
    status = SS_LISTENING;

    // backup socket manager object 
    smgr = smanager; 

    // set the event mask 
    set_event_mask (POLLPRI | POLLIN);

    // set the handlers
    set_on_data_read (_read_f);
    set_on_data_write (_write_f);
    set_on_data_err (_err_f);
    set_on_data_hup (_hup_f);
   
    // register the socket
    if (smgr->register_socket (this, POLLPRI | POLLIN))
    {
        status = SS_ERRRESOURCES;
        return -1;
    }

    return 0;
}


// function for accepting connection on listening socket
int tSSLSocket::accept_socket(class tSocketMgr &smanager)
{
    struct hostent *hp;
    class tSSLSocket *sock;
    int new_sock;
    struct sockaddr_in address;
    socklen_t sock_addr_len = sizeof(struct sockaddr_in);

    if (status != SS_LISTENING && status != SS_ACCEPTING)
    {
        log (L_ERROR, "SSLSOCKET: Server cannot accept connections unless "
               "in listenning mode.");
        return -1;
    }
   
    // try to accept new socket       
    if( (new_sock = accept (get_fd(), (struct sockaddr *)(&address),
         &sock_addr_len)) < 0)
    {
        if (errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR) 
            return 0;
        log (L_ERROR, "SSLSOCKET: Error accepting connection: %m");
        return -1;
    }
    debug_log ('l', "New socket (sock=%d) accepted.", new_sock);

    set_status (SS_LISTENING);

    // try to get hostname of client
    if (!(hp = gethostbyaddr ((char *)&(address.sin_addr),
          sizeof(address.sin_addr), AF_INET)) ) 
    {
        log (L_ERROR, "SSLSOCKET: Cannot resolve client hostname." );
        s_close (new_sock);
        return -1;
    }

    // creates new SSL Socket
    sock = create_new_socket ();

    // assign it to object
    sock->set_fd (new_sock);
    sock->set_status (SS_CONNECTED);
    sock->set_role (AS_SOCK_CLIENT);
    sock->set_smgr (&smanager);
    sock->set_created_by_smgr (true);

    // let it inherit master's properties
    sock->set_on_register (get_on_register ());
    sock->set_on_unregister (get_on_unregister ());
    sock->set_on_data_read (_read_f);
    sock->set_on_data_write (_write_f);
    sock->set_on_data_hup (_hup_f);
    sock->set_on_data_err (_err_f);

    sock->set_on_err (get_on_err ());
    sock->set_on_hup (get_on_hup ());
    sock->set_on_read_available (get_on_read_available ());
    sock->set_on_write_finished (get_on_write_finished ());
    sock->set_on_read_finished (get_on_read_finished ());
    sock->set_on_connected (get_on_connected ());

    sock->set_on_ssl_shutdown_init (get_on_ssl_shutdown_init() );
    sock->set_on_ssl_shutdown_finished (get_on_ssl_shutdown_finished() );
    
    // backup hostname
    if (! (sock->hostname = s_strdup (hp->h_name)))
    {
        log(L_ERROR,"SSLSOCKET: Not enough memory.");
        delete sock;
        return -1;
    }

    // backup port
    sock->port = address.sin_port;

    // fill in the client address
    memcpy (&(sock->addr), &address , sizeof(sock->addr) );

    // backup context
    sock->context = context;

    // make the socket nonblocking
    if (sock->set_non_blocking() )
    {
        delete sock;
        return -1;
    }

    // describe the communications details
    sock->ssl = SSL_new (sock->context);
    if (! (sock->ssl))
    {
        log(L_ERROR,"SSLSOCKET: Creating SSL structure failed.");
        delete sock;
        return -1;
    }

    // tell the ssl on which socket we operate
    if (!(SSL_set_fd (sock->ssl, new_sock))) 
    {
         log(L_ERROR,"SSLSOCKET: Assigning socket to ssl object failed.");
         SSL_free (sock->ssl);
         delete sock;
         return -1;
    }
 

    // register the socket
    smgr = &smanager;
    if (smgr->register_socket (sock, POLLPRI | POLLIN, get_sid() ))
    {
        delete sock;
        return -1;
    }

    // set connection state
    sock->status = SS_CONNECTED;
    sock->set_ssl_status (SL_CONNECTING);
    sock->clear_io ();
    sock->last_operation = OP_HANDSHAKE;

    debug_log ('l', "Sheduling SSL_ACCEPT on sock=%d.", sock->get_fd() );

    return 0; 

}


int tSSLSocket::reconnect()
{
    struct hostent *hp;
    
    ASSERT(smgr);
    if (role != AS_SOCK_CLIENT)
    {
         log(L_ERROR,"SSLSOCKET: Call for reconnecting of server side.");
         return -1;
    }
   
    if (status != SS_DEFAULT && status != SS_BROKEN && status != SS_CONNECTED
            && status != SS_CONNECTING)
    {   
         log(L_ERROR,"SSLSOCKET: Cannot reconnect socket from this state.");
         status = SS_ERROR;
         err_handler();
         return -1;
    }

    if (status == SS_DEFAULT || status == SS_BROKEN)
    {

        if (!(hp = gethostbyname(hostname)) ) 
        {
           switch (h_errno)
           {
              case HOST_NOT_FOUND:
                  log(L_ERROR,"SSLSOCKET: Unknown host (\"%s\").", hostname);
                  break;
              case NO_ADDRESS:
                  log(L_ERROR,"SSLSOCKET: Host (\"%s\") has no IP.", hostname);
                  break;
              default:
                  log(L_ERROR,"SSLSOCKET: Nameserver error while resolving "
                       "(\"%s\").", hostname);
           }
 
           status = SS_BROKEN;
           err_handler();
           return -1;
        }

        // fill in the server address
        memset (&addr, 0, sizeof(addr) );
        memcpy (&(addr.sin_addr), hp->h_addr_list[0], sizeof(addr.sin_addr) );
        addr.sin_family = AF_INET;
        addr.sin_port = htons (port);
        
        status = SS_CONNECTING;

        // register the socket to ensure connection completion (needs POLLOUT)
        if (smgr->register_socket (this, POLLPRI | POLLIN | POLLOUT))
        {
            status = SS_ERRRESOURCES;
            err_handler();
            return -1;
        }
   
        debug_log ('l', "Sock=%d is going to be connected.",  get_fd() );
        if (connect (get_fd(), (struct sockaddr *) &addr, sizeof(addr)) < 0) 
        {
            if (errno == EINPROGRESS) 
                return 0;

            if (errno == EINTR || errno == EAGAIN)
            {
                status = SS_DEFAULT;
                return 0;
            }    

            if (errno == ECONNREFUSED)
            {
                log(L_ERROR, "SSLSOCKET: Connection refused by (\"%s\").", hostname);
                status = SS_BROKEN;
                err_handler ();
                return 0;            
            }
          
            log(L_ERROR, "SSLSOCKET: Cannot connect to (\"%s\"): %m", hostname);
            status = SS_BROKEN;
            err_handler ();

            return -1;
        }

        status = SS_CONNECTED;
        debug_log ('l', "Sock=%d sucesfully connected (tcp-layer).", get_fd ());

    }    

    if (status == SS_CONNECTED)
    {    
        if (ssl_status == SL_EMPTY)
        {
             // describe the communications details
             ssl = SSL_new (context);
             if (!ssl) 
             {
                 log(L_ERROR,"SSLSOCKET: Creating SSL structure failed.");
                 status = SS_ERRRESOURCES;
                 err_handler();
                 return -1;
             }
             
             // set the ssl over our socket 
             if (!SSL_set_fd(ssl, get_fd())) 
             {
                 log (L_ERROR, "SSLSOCKET: Cannot SSL_set_fd() on our socket.");
                 SSL_free (ssl);
                 ssl = NULL;
                 status = SS_ERRRESOURCES;
                 err_handler();
                 return -1;
             }

             ssl_status = SL_CONNECTING;
        }
             
        smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
        clear_io ();
        last_operation = OP_HANDSHAKE;
        debug_log ('l', "Sheduling SSL_CONNECT on sock=%d.", get_fd() );

        /* try to negotiate connecting handshake */
        return 0; 
    }
    
    return -1;
}


int tSSLSocket::verify_common_name(const char *common_name)
{
    X509 *peer;
    char peer_CN[256];

    // verify authenticity 
    if(SSL_get_verify_result(ssl)!=X509_V_OK) 
    {
        log (L_INFO, "SSLSOCKET: Peer certificate doesn't verify.");
        return -1;
    }

    /* Check the cert chain. The chain length
       is automatically checked by OpenSSL when we
       set the verify depth in the ctx 
     */

    /* Take the certificate */
    if (!(peer = SSL_get_peer_certificate(ssl)))
    {
        log (L_INFO, "SSLSOCKET: Peer certificate inavailable.");
        return -1;
    }

    /* Extract the common name field */
    X509_NAME_get_text_by_NID (X509_get_subject_name(peer),
      NID_commonName, peer_CN, 256);

    /* compare the cert common name with given */    
    if (strcasecmp (peer_CN, common_name))
    {
        log (L_INFO, "SSLSOCKET: Peer certficate common name doesn't "
                     "match: requested: %s certificate: %s.",
                        common_name, peer_CN);
        return -1;
    }

    return 0;    
}


ssize_t tSSLSocket::sock_write (void *buffer, ssize_t len)
{
    ASSERT(ssl_status == SL_CONNECTED);

    return tSocketNB::sock_write (buffer, len); 
}


ssize_t tSSLSocket::sock_read(void *buffer, ssize_t len)
{
    ASSERT(ssl_status == SL_CONNECTED);

    return tSocketNB::sock_read (buffer, len);
}


ssize_t tSSLSocket::at_read(void *buffer, ssize_t len)
{
    ASSERT(ssl_status == SL_CONNECTED);

    return tSocketNB::at_read (buffer, len);
}


ssize_t tSSLSocket::at_write(void *buffer, ssize_t len)
{
    return at_write(buffer, len, 0);
}

ssize_t tSSLSocket::at_write(void *buffer, ssize_t len, long flags)
{
    unsigned long length;

    ASSERT(ssl_status == SL_CONNECTED);
    ASSERT(status == SS_CONNECTED);
    ASSERT ((last_operation == OP_NO_OPERATION) ||
            (last_operation == OP_IO && (w_io.last_op == IO_OP_NO_OPERATION)));
    ASSERT (buffer && len >= 0);

    debug_log ('s', "ATOM_WRITE sheduled on sock=%d.", get_fd());

    length = static_cast<unsigned long>(len);
    w_io.atom_frame.magic = V_MAGIC;
    w_io.atom_frame.flags = 0;

    if (flags & FL_ZLIB_COMPRESSED)
    {
        if(compress_data(buffer, &length)==0)
            w_io.atom_frame.flags |= FL_ZLIB_COMPRESSED;
    }

    w_io.atom_frame.length = length;            

    w_io.last_op_buf = w_io.last_op_buf_cur = static_cast<char *>(buffer);
    w_io.last_op_bytes = length;
    w_io.already_processed = 0;
    w_io.last_op = IO_OP_ATOM_WRITING;
    last_operation = OP_IO;
    w_io.atom_op_state = AT_BEGINFRAME;

    smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);

    return 0;

}


int tSSLSocket::shutdown_ssl ()
{
    ASSERT(status == SS_CONNECTED);
    ASSERT(last_operation == OP_NO_OPERATION);
    ASSERT(ssl_status == SL_CONNECTED);
     
    debug_log ('l', "SHUTDOWN sheduled on sock=%d.", get_fd());
    clear_io ();
    last_operation = OP_SHUTDOWN;
    ssl_status = SL_SHUTDOWN_INIT;
    smgr->set_event_mask (this , POLLPRI | POLLIN | POLLOUT);

    return 0;
}


ssize_t tSSLSocket::sock_recv(void *buffer, ssize_t len, unsigned int options)
{
    using namespace lib;
    die("ssl / sock_recv not implemented yet.");
    return -1;
}


ssize_t tSSLSocket::sock_poke(void *buffer, ssize_t len)
{
    ssize_t r;
    ASSERT(status == SS_CONNECTED || status == SS_CONNECTING);
    ASSERT (buffer && len >= 0);

    r = recv (get_fd(), buffer, len, MSG_PEEK);
    if (r < 0)
         return -errno;

    return r;
}


int tSSLSocket::safe_SSL_read (char *buffer, long bytes)
{
    int ret;

    ret = SSL_read (ssl, buffer, bytes);
    debug_log ('l', "Safe SSLread returned (%d)", ret);

    if (ret > 0)
    {
        return ret;
    } 
    else if (ret == 0)
    {
        if (SSL_get_shutdown (ssl))
        {
             debug_log ('l', "Shutdown needed: Bitmask(%d).",
                               SSL_get_shutdown (ssl));
             set_ssl_status (SL_SHUTDOWN_FINAL);
             clear_io ();
             set_last_operation (OP_SHUTDOWN);
             smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
             if (on_ssl_shutdown_init) 
                       on_ssl_shutdown_init(this);
             return -2;
        }
        return 0;
    } 
    else
    {
        switch (SSL_get_error (ssl,ret))
        {
            case SSL_ERROR_WANT_READ:
                debug_log ('l', "Safe SSLread waits for incoming data.");
                smgr->set_event_mask (this, POLLPRI | POLLIN);
                return 0;
            case SSL_ERROR_WANT_WRITE:
                debug_log ('l', "Safe SSLread waits for writing possibility.");
                smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
                return 0;
            default:
                log (L_ERROR, "SSLSOCKET: SSL/TLS read error.");
                close_socket();
                set_status ( SS_BROKEN );
                err_handler();
                return -1;
        }    
    }
}


int tSSLSocket::safe_SSL_write (char *buffer, long bytes)
{
    int ret;

    ret = SSL_write (ssl, buffer, bytes);
    debug_log ('l', "Safe SSLwrite returned (%d)", ret);

    if (ret > 0)
    {
        return ret;
    } 
    else if (ret == 0)
    {
        if (SSL_get_shutdown (ssl))
        {
             debug_log ('l', "Shutdown needed: Bitmask(%d).",
                               SSL_get_shutdown (ssl));
             set_ssl_status (SL_SHUTDOWN_FINAL);
             clear_io ();
             set_last_operation (OP_SHUTDOWN);
             smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
             if (on_ssl_shutdown_init) 
                       on_ssl_shutdown_init(this);
             return -2;
        }
        return 0;
    } 
    else
    {
        switch (SSL_get_error (ssl,ret))
        {
            case SSL_ERROR_WANT_READ:
                debug_log ('l', "Safe SSLwrite waits for incoming data.");
                smgr->set_event_mask (this, POLLPRI | POLLIN);
                return 0;
            case SSL_ERROR_WANT_WRITE:
                debug_log ('l', "Safe SSLwrite waits for writing possibility.");
                smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
                return 0;
            default:
                log (L_ERROR, "SSLSOCKET: SSL/TLS read error.");
                close_socket();
                set_status ( SS_BROKEN );
                err_handler();
                return -1;
        }    
    }
}


int tSSLSocket::connect_again()
{
    int ret;
    time_t tim = time(NULL);
    struct tm *log_tm = localtime(&tim);
    char buf[32];
  
    ASSERT(status == SS_CONNECTED);
    ASSERT(last_operation == OP_HANDSHAKE || last_operation == OP_NO_OPERATION);
    ASSERT(ssl_status == SL_EMPTY || ssl_status == SL_CONNECTING);
                
    ret=SSL_connect(ssl);
    debug_log ('l', "SSLconnect returned (%d)", ret);

    switch (ret)
    {
        case 1:

           debug_log ('l', "SSL connection using %s.", SSL_get_cipher (ssl));
           smgr->set_event_mask (this, POLLPRI | POLLIN);
           ssl_status = SL_CONNECTED;
           clear_io ();
           last_operation = OP_NO_OPERATION;
           if (on_connected) 
               on_connected (this);
           return 0;

        case 0:
            log (L_ERROR, "SSLSOCKET: SSL/TLS handshake failed - "
                          "not conforming specs.");
            
            log_lock ();
            strftime (buf, sizeof (buf), "%Y-%m-%d %H:%M:%S", log_tm);
            fprintf (stderr, "E %s Reason: ", buf);
            ERR_print_errors_fp (stderr);
            log_unlock ();
            
            close_socket();
            status = SS_ERROR;
            err_handler();
            return -1;
        case -1:
            switch (SSL_get_error(ssl,ret) )
            {
                 case SSL_ERROR_WANT_READ:
                          debug_log('l', "SSLconnect waits for incoming data.");
                          smgr->set_event_mask (this, POLLPRI | POLLIN);
                          clear_io ();
                          last_operation = OP_HANDSHAKE;
                          return 0;
                 case SSL_ERROR_WANT_WRITE:
                          debug_log ('l', "SSLconnect waits for "
                                  "writing possibility."); 
                          smgr->set_event_mask(this,POLLPRI | POLLIN | POLLOUT);
                          clear_io ();
                          last_operation = OP_HANDSHAKE;
                          return 0;
                 default:
                     log (L_ERROR, "SSLSOCKET: SSL/TLS handshake failed. "
                                   "Otherside gave up.");
                     close_socket();
                     status = SS_ERROR;
                     err_handler();
                     return -1;
            }
        default: 
            log (L_ERROR, "SSLSOCKET: Unexpected return value "
                          "from SSL_connect().");
            close_socket();
            status = SS_ERROR;
            err_handler();
            return -1;

    }

}


int tSSLSocket::accept_again()
{
    int ret;
    time_t tim = time(NULL);
    struct tm *log_tm = localtime(&tim);
    char buf[32];

    ASSERT(status == SS_CONNECTED);
    ASSERT(last_operation == OP_HANDSHAKE || last_operation == OP_NO_OPERATION);
    ASSERT(ssl_status == SL_EMPTY || ssl_status == SL_CONNECTING);
                
    ret=SSL_accept(ssl);
    debug_log ('l', "SSLaccept returned (%d)", ret);

    switch (ret)
    {
        case 1:
            
            debug_log ('l', "SSL connection using %s.", SSL_get_cipher (ssl));
            smgr->set_event_mask (this, POLLPRI | POLLIN);
            ssl_status = SL_CONNECTED;
            clear_io ();
            last_operation = OP_NO_OPERATION;

            if (on_connected) 
                on_connected (this);

            return 0;

        case 0:
            log (L_ERROR, "SSLSOCKET: SSL/TLS handshake failed - "
                          "not conforming specs.");

            log_lock ();
            strftime (buf, sizeof (buf), "%Y-%m-%d %H:%M:%S", log_tm);
            fprintf (stderr, "E %s Reason: ", buf);
            ERR_print_errors_fp (stderr);
            log_unlock ();

            close_socket();
            status = SS_ERROR;
            err_handler();
            return -1;
        case -1:
            switch (SSL_get_error(ssl,ret) )
            {
                 case SSL_ERROR_WANT_READ:
                          debug_log ('l', "SSLaccept waits for incoming data.");
                          smgr->set_event_mask (this, POLLPRI | POLLIN);
                          clear_io ();
                          last_operation = OP_HANDSHAKE;
                          return 0;
                 case SSL_ERROR_WANT_WRITE:
                          debug_log('l', "SSLaccept waits for writing "
                                  "possibility." );
                          smgr->set_event_mask (this, 
                                  POLLPRI | POLLIN | POLLOUT);
                          clear_io ();
                          last_operation = OP_HANDSHAKE;
                          return 0;
                 default:
                     log (L_ERROR, "SSLSOCKET: SSL/TLS handshake failed. "
                                   "Other side gave up..");
                     close_socket();
                     status = SS_ERROR;
                     err_handler();
                     return -1;
            }
        default: 
            log (L_ERROR, "SSLSOCKET: Unexpected return value "
                    "from SSL_accept().");
            close_socket();
            status = SS_ERROR;
            err_handler();
            return -1;
    }

}


int tSSLSocket::ssl_shutdown_again()
{
    int ret;

    ASSERT(status == SS_CONNECTED);
    ASSERT(last_operation == OP_SHUTDOWN || last_operation == OP_NO_OPERATION);
    ASSERT(ssl_status == SL_CONNECTED || ssl_status == SL_SHUTDOWN_INIT ||
           ssl_status == SL_SHUTDOWN_FINAL);
    
    ret = SSL_shutdown (ssl);
    debug_log ('l', "SSLshutdown returned (%d)", ret);
  
    switch (ret)
    {
        case 1:
            debug_log ('l', "Bidirectional shutdown complete.");
            clear_io ();
            last_operation = OP_NO_OPERATION;
            ssl_status = SL_EMPTY;
            smgr->set_event_mask (this, POLLPRI | POLLIN);
            if (ssl) { SSL_free (ssl); ssl = NULL; }
            if (on_ssl_shutdown_finished) on_ssl_shutdown_finished (this);
            return 0;
        case 0:
            debug_log ('l', "Call SSLShutdown to complete "
                            "bidirectional shutdown.");
            clear_io ();
            last_operation = OP_SHUTDOWN;
            ssl_status = SL_SHUTDOWN_FINAL;
            smgr->set_event_mask (this, POLLPRI | POLLIN);
            return 0;
        case -1:
            switch (SSL_get_error(ssl,ret) )
            {
                 case SSL_ERROR_WANT_READ:
                     debug_log ('l', "SSLshutdown waits for incomming data.");
                     smgr->set_event_mask (this, POLLPRI | POLLIN);
                     clear_io ();
                     last_operation = OP_SHUTDOWN;
                     if (ssl_status == SL_CONNECTED)
                          ssl_status = SL_SHUTDOWN_INIT;
                     return 0;
                 case SSL_ERROR_WANT_WRITE:
                     debug_log ('l', "SSLshutdown waits for "
                             "writing possibility.");
                     smgr->set_event_mask (this, 
                             POLLPRI | POLLIN | POLLOUT);
                     clear_io ();
                     last_operation = OP_SHUTDOWN;
                     if (ssl_status == SL_CONNECTED)  
                          ssl_status = SL_SHUTDOWN_INIT;
                     return 0;
                 default:
                     log (L_ERROR, "SSLSOCKET: Shutdown handshake error.");
                     close_socket();
                     status = SS_ERROR;
                     err_handler();
                     return -1;
            }
        default: 
            log (L_ERROR, "SSLSOCKET: Unexpected return value from "
                          "SSL_shutdown().");
            close_socket();
            status = SS_ERROR;
            err_handler();
            return -1;
    }

}


int tSSLSocket::read_again()
{
    int ret;

    ASSERT(status == SS_CONNECTED);
    ASSERT((last_operation == OP_IO && r_io.last_op == IO_OP_READING) || 
            last_operation == OP_NO_OPERATION);
    ASSERT(ssl_status == SL_CONNECTED);
                
    ret = SSL_read (ssl, r_io.last_op_buf, r_io.last_op_bytes);
    debug_log ('l', "SSLread returned (%d)", ret);

    if (ret > 0)
    {
        r_io.last_op_bytes = ret;
        stop_r_io ();
        smgr->set_event_mask (this, POLLPRI | POLLIN);
        if (on_read_finished) on_read_finished (this);
        return 0;
    } else 
    if (ret == 0)
    {
        if (SSL_get_shutdown (ssl))
        {
             debug_log ('l', "Shutdown needed: Bitmask(%d).",
                               SSL_get_shutdown (ssl));
             ssl_status = SL_SHUTDOWN_FINAL;
             clear_io ();
             last_operation = OP_SHUTDOWN;
             smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
             if (on_ssl_shutdown_init) on_ssl_shutdown_init (this);
             return 0;
        }
        r_io.last_op_bytes = ret;
        stop_r_io ();
        smgr->set_event_mask (this, POLLPRI | POLLIN);
        if (on_read_finished) on_read_finished (this);
        return 0;
    } else
    {
        switch (SSL_get_error(ssl,ret) )
        {
            case SSL_ERROR_WANT_READ:
                debug_log ('l', "SSLread waits for incoming data.");
                smgr->set_event_mask (this, POLLPRI | POLLIN);
                r_io.last_op = IO_OP_READING;
                last_operation = OP_IO;
                return 0;
            case SSL_ERROR_WANT_WRITE:
                debug_log ('l', "SSLread waits for writing possibility.");
                smgr->set_event_mask (this, 
                        POLLPRI | POLLIN | POLLOUT);
                r_io.last_op = IO_OP_READING;
                last_operation = OP_IO;
                return 0;
            default:
                log (L_ERROR, "SSLSOCKET: SSL/TLS read error.");
                close_socket();
                status = SS_ERROR;
                err_handler();
                return -1;
        }    
    }
}


int tSSLSocket::write_again()
{
    int ret;

    ASSERT(status == SS_CONNECTED);
    ASSERT((last_operation == OP_IO && w_io.last_op == IO_OP_WRITING) || 
            last_operation == OP_NO_OPERATION);
    ASSERT(ssl_status == SL_CONNECTED);
                
    ret = SSL_write (ssl, w_io.last_op_buf, w_io.last_op_bytes);
    debug_log ('l', "SSLwrite returned (%d)", ret);

    if (ret > 0)
    {
        w_io.last_op_bytes = ret;
        stop_w_io ();
        smgr->set_event_mask (this, POLLPRI | POLLIN);
        if (on_write_finished) on_write_finished (this);
        return 0;
    } else 
    if (ret == 0)
    {
        if (SSL_get_shutdown (ssl))
        {
             debug_log ('l', "Shutdown needed: Bitmask(%d).", 
                               SSL_get_shutdown (ssl));
             ssl_status = SL_SHUTDOWN_FINAL;
             clear_io ();
             last_operation = OP_SHUTDOWN;
             if (on_ssl_shutdown_init) on_ssl_shutdown_init (this);
             smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
             return 0;
        }
        w_io.last_op_bytes = ret;
        stop_w_io ();
        smgr->set_event_mask (this, POLLPRI | POLLIN);
        if (on_write_finished) on_write_finished (this);
        return 0;
    } else
    {
        switch (SSL_get_error(ssl,ret) )
        {
            case SSL_ERROR_WANT_READ:
                debug_log ('l', "SSLwrite waits for incomming data.");
                smgr->set_event_mask (this, POLLPRI | POLLIN);
                w_io.last_op = IO_OP_WRITING;
                last_operation = OP_IO;
                return 0;
            case SSL_ERROR_WANT_WRITE:
                debug_log ('l', "SSLwrite waits for writing possibility.");
                smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
                w_io.last_op = IO_OP_WRITING;
                last_operation = OP_IO;
                return 0;
            default:
                log (L_ERROR, "SSLSOCKET: SSL/TLS write error.");
                close_socket();
                status = SS_ERROR;
                err_handler();
                return -1;
        }    
    }
}


int tSSLSocket::atom_read_again()
{
    using messages::tPlainBuf;
    class tPlainBuf b (r_io.framebuf, sizeof(r_io.framebuf) );
    long ret;
    int32_t x,y;
    unsigned long un_len;

    ASSERT(status == SS_CONNECTED);
    ASSERT((last_operation == OP_IO && r_io.last_op == IO_OP_ATOM_READING) || 
            last_operation == OP_NO_OPERATION);
    ASSERT(ssl_status == SL_CONNECTED);

    r_io.last_op = IO_OP_ATOM_READING;
    last_operation = OP_IO;
    switch (r_io.atom_op_state)
    {
        case AT_BEGINFRAME:
            ret = safe_SSL_read (r_io.framebuf +  r_io.already_processed, 
                               sizeof(r_io.framebuf) - r_io.already_processed);
            if (ret >= 0)
            {
                r_io.already_processed += ret;
                if (r_io.already_processed == sizeof(r_io.framebuf) ) 
                {
                     READ_INT32 (&b, r_io.atom_frame.magic);
                     READ_INT32 (&b, r_io.atom_frame.length);
                     READ_INT32 (&b, r_io.atom_frame.flags);
                     if (r_io.atom_frame.magic != V_MAGIC || 
                         r_io.atom_frame.length > r_io.last_op_bytes)
                     {
                        log (L_ERROR, "SSLSOCKET: Error while processing "
                                "start frame - shutting down ssl layer." );
                        close_socket();
                        set_status(SS_BROKEN);
                        err_handler();    
                        return 0;
                     }
                     b.rewind();
                     r_io.already_processed = 0;
                     r_io.atom_op_state = AT_DATA;
                } else return 0;
            } else return -1; 
                
        case AT_DATA:
            ret = safe_SSL_read (r_io.last_op_buf + r_io.already_processed,
                    r_io.atom_frame.length - r_io.already_processed);
            if (ret >= 0)
            {
                r_io.already_processed += ret;
                if (r_io.already_processed == r_io.atom_frame.length)
                {
                     un_len = r_io.last_op_bytes;
                     if (r_io.atom_frame.flags & FL_ZLIB_COMPRESSED)
                     {
                         if (uncompress_data(r_io.last_op_buf,
                                     r_io.atom_frame.length, &un_len)) {
                                log (L_ERROR, "SSLSOCKET: Error while "
                                        " processing start frame - shutting "
                                        "down ssl layer." );
                                close_socket();
                                set_status(SS_BROKEN);
                                err_handler();
                                return 0;
                         }
                         debug_log( 'l', "Uncompress: %ld => %ld",
                                 r_io.atom_frame.length, un_len);    
                     } else un_len = r_io.atom_frame.length;
                     r_io.already_processed = 0;
                     r_io.atom_op_state = AT_ENDFRAME;
                } else return 0;
            } else return -1;    
            
        case AT_ENDFRAME:

            ret = safe_SSL_read (r_io.framebuf +  r_io.already_processed, 
                                sizeof(r_io.framebuf) - r_io.already_processed);
            if (ret >= 0)
            {
                r_io.already_processed += ret;
                if (r_io.already_processed == sizeof(r_io.framebuf) ) 
                {
                     READ_INT32 (&b, x);
                     READ_INT32 (&b, y);
                     if (r_io.atom_frame.magic != x || 
                             r_io.atom_frame.length != y  )
                     {
                        log (L_ERROR, "SSLSOCKET: Error while processing "
                                "end frame - shutting down ssl layer." );
                        close_socket();
                        set_status(SS_BROKEN);
                        err_handler();
                        return 0;
                     }
                     r_io.last_op_bytes = un_len;
                     stop_r_io ();
                     smgr->set_event_mask (this, POLLPRI | POLLIN);
                     if (on_read_finished) on_read_finished (this);
                     return 0;
                } else return 0;
            } else return -1; 
            break;

        default:  
            log (L_ERROR, "SSLSOCKET: Unexpected state of atom-op.");
    }

    return -1;
}


int tSSLSocket::atom_write_again()
{
    using messages::tPlainBuf;
    class tPlainBuf b (w_io.framebuf, sizeof(w_io.framebuf) );
    int ret;

    ASSERT(status == SS_CONNECTED);
    ASSERT((last_operation == OP_IO && w_io.last_op == IO_OP_ATOM_WRITING) || 
            last_operation == OP_NO_OPERATION);
    ASSERT(ssl_status == SL_CONNECTED);

    STORE_INT32 (&b, w_io.atom_frame.magic);
    STORE_INT32 (&b, w_io.atom_frame.length);
    STORE_INT32 (&b, w_io.atom_frame.flags);

    w_io.last_op = IO_OP_ATOM_WRITING;
    last_operation = OP_IO;
    switch (w_io.atom_op_state)
    {
        case AT_BEGINFRAME:
            ret = safe_SSL_write(w_io.framebuf +  w_io.already_processed, 
                               sizeof(w_io.framebuf) - w_io.already_processed);
            if (ret >= 0)
            {
                w_io.already_processed += ret;
                if (w_io.already_processed == sizeof(w_io.framebuf) ) 
                {
                     w_io.already_processed = 0;
                     w_io.atom_op_state = AT_DATA;
                } else return 0;
            } else return -1; 
                
        case AT_DATA:
            ret = safe_SSL_write (w_io.last_op_buf + w_io.already_processed, 
                    w_io.atom_frame.length - w_io.already_processed);
            if (ret >= 0)
            {
                w_io.already_processed += ret;
                if (w_io.already_processed == w_io.atom_frame.length)
                {
                     w_io.already_processed = 0;
                     w_io.atom_op_state = AT_ENDFRAME;
                } else return 0;
            } else return -1;    
            
        case AT_ENDFRAME:
            ret = safe_SSL_write(w_io.framebuf +  w_io.already_processed, 
                                sizeof(w_io.framebuf) - w_io.already_processed);
            if (ret >= 0)
            {
                w_io.already_processed += ret;
                if (w_io.already_processed == sizeof(w_io.framebuf) ) 
                {
                     w_io.last_op_bytes = w_io.atom_frame.length;
                     stop_w_io ();
                     smgr->set_event_mask (this, POLLPRI | POLLIN);
                     if (on_write_finished) on_write_finished (this);
                     return 0;
                } else return 0;
            } else return -1; 
            break;

        default:  
            log (L_ERROR, "SSLSOCKET: Unexpected state of atom-op.");
    }

    return -1;
}


void tSSLSocket::hup_handler()
{
  
    debug_log ('l', "Hup-handler called (sock=%d).", get_fd());

    close_socket();
    status = SS_BROKEN;
 
    if (on_hup) 
        on_hup(this);

}


void tSSLSocket::err_handler()
{
    tSOCK_STAT tmp = get_status();
  
    debug_log ('l', "Err-handler called (sock=%d).", get_fd());

    close_socket();
    set_status( tmp );
  
    if (on_err) 
        on_err(this);
 
}


void tSSLSocket::read_handler()
{
    ASSERT (role == AS_SOCK_CLIENT);
    ASSERT (status == SS_CONNECTED);
    ASSERT (ssl_status == SL_CONNECTED || ssl_status == SL_CONNECTING ||
            ssl_status == SL_SHUTDOWN_INIT || ssl_status == SL_SHUTDOWN_FINAL);

    switch(last_operation)
    {
        case OP_SHUTDOWN:
            debug_log ('l', "Read handler: OP_SHUTDOWN");
            ssl_shutdown_again();
            break;
        default:
            tSocketNB::read_handler ();
    } 

}


void tSSLSocket::write_handler()
{
    ASSERT(role == AS_SOCK_CLIENT);
    ASSERT(status == SS_CONNECTED || status == SS_CONNECTING);

    if (status == SS_CONNECTING)
        tSocketNB::write_handler ();

    if (status == SS_CONNECTED)
    {    
         ASSERT(ssl_status == SL_CONNECTED || ssl_status == SL_CONNECTING ||
            ssl_status == SL_SHUTDOWN_INIT || ssl_status == SL_SHUTDOWN_FINAL);

         switch(last_operation)
         {
            case OP_SHUTDOWN:
                debug_log ('l', "Write handler: OP_SHUTDOWN");
                ssl_shutdown_again();
                break;
            default:
                tSocketNB::write_handler ();
          }

    }
}

}
