/**
 * @brief Echelog Socks library -- Unix domain sockets
 *
 * Actual implementation of the abstract tSocket class 
 * above Unix domain sockets.
 *
 * @author Ales Privetivy <ales@ucw.cz>
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/lib.h"
#include "messages/plainbuf.h"
#include "messages/testbuf.h"
#include "socks/unixsocket.h"
#include "socks/socketmgr.h"

#include "lib/archdep.h"
#include <sys/types.h>
#include <sys/un.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <errno.h>

#ifdef SCM_CREDS
#include <sys/uio.h> /* for struct iovec */
#include <sys/ucred.h>
#endif


namespace socks {

using namespace lib;


// This will create socket
int tUNIXSocket::new_socket ()
{
    long new_fd = socket(AF_UNIX, SOCK_STREAM, 0);

    // hmm, socket cannot be opened
    if (new_fd < 0)
    {
        log (L_ERROR, "Can't create a new socket.");
        status = SS_ERRSOCK;
        return -1;
    }
   
    set_fd(new_fd); 
    status = SS_DEFAULT;

    return 0;
}


// closes the socket
tUNIXSocket::~tUNIXSocket()
{
    if (close_socket())
        log (L_ERROR, "Error closing socket %ld.", get_fd());

    // clear out of application specific data 
    xfree (app_specific_data);
}


// perform the handshake
int tUNIXSocket::handshake ()
{
#if defined(HAVE_LINUX_CREDENTIALS) || defined(HAVE_LOCAL_CREDENTIALS)    
     
    /* Connection authentication - send auth info */

#   if !defined(HAVE_LINUX_CREDENTIALS) \
    && ( defined(HAVE_STRUCT_FCRED) || defined(HAVE_STRUCT_SOCKCRED) )

    /* For BSD/OS and NetBSD we need to set this before the authentication
     * request or the whole process fscks up */
    long on = 1;
    if (setsockopt(get_fd (), 0, LOCAL_CREDS, &on, sizeof(on)) < 0)
        log (L_ERROR, "Can't do setsockopt to socket %li: %m", get_fd ()); 
#   endif

    int r = send_auth_info ();

    // any success?
    if (r != 1)
    {
        // something is completely wrong
        log (L_ERROR, "UNIXSocket handshake failed for fatal reasons");
        close_socket ();
        status = SS_BROKEN;
        return -1;
    }

    r = receive_auth_info ();

    // any success?
    if (r != 1)
    {
        // something is completely wrong
        log (L_ERROR, "UNIXSocket handshake failed for fatal reasons");
        close_socket ();
        status = SS_BROKEN;
        return -1;
    }  

#else
#   warning No unix domain socket authentication available - disabling it.
#endif

    debug_log ('s', "Handshake successful on socket fd=%ld", get_fd ());

    return 1;
}


// creates client side of the socket
int tUNIXSocket::create_client(const char *name)
{
    // if somebody uses unopened socket he shall feel our wrath
    if (new_socket () < 0)
        die("Trying to create connection upon non-initialized socket.");

    // fill in the filename of the UD-socket
    ASSERT (name);
    set_addr(name);

    debug_log ('s', "Connecting client to socket %d (%s)", get_fd(), 
            get_addr ());

    // mark that we are in the client mode
    set_role(AS_SOCK_CLIENT);

    // and connect
    if (connect(get_fd(), (struct sockaddr *) &addr, SUN_LEN (&addr)) < 0)
        status = SS_ERRCONNECT;
    else
    {
        // try handshake
        if (handshake () != 1)
            // ahhh, a secret passage ... oops!
            return -1;
    }

    if (status == SS_ERRCONNECT)
    {
        close_socket ();
        return -1;
    }

    // okay, client connection established
    status = SS_CONNECTED;
    return 0;    
}


// creates server side of socket
int tUNIXSocket::create_server(const char *name)
{
    // ok, again, if somebody uses unopened socket he shall feel our wrath :)
    if (new_socket () < 0)
        die("Trying to create connection upon non-existing socket.");

    // fills in where to connect
    ASSERT (name);
    set_addr(name);

    // disallow socket reusal
    setsockopt (get_fd(), SOL_SOCKET, SO_REUSEADDR, 0, 0);
    
    /* There should be no such socket file */
    if (!s_eaccess (name, F_OK) && unlink (name))
        log (L_WARN,"Can't set old socket resources free");
    
    debug_log ('s', "Connecting a server to socket %d (%s)", 
            get_fd(), get_addr ());

    // mark that this is the server side
    set_role(AS_SOCK_SERVER);

    // bind with the filename 
    debug_log ('s', "Binding socket %d (%s)", get_fd(), get_addr ());
    if (bind( get_fd(), (struct sockaddr *) &addr,
        sizeof(struct sockaddr_un)))
    {
           log(L_ERROR,"Cannot bind socket.");
           status = SS_ERRBIND;
           close_socket ();
           return -1;
    }

    // and listen upon incomming events
    debug_log ('s', "Listening at socket %d (%s)", get_fd(), get_addr());
    if(listen( get_fd(), LISTENCNT))
    {
           log(L_ERROR,"Cannot set the socket into listenning state.");
           status = SS_ERRLISTENING;
           close_socket ();
           return -1;
    }

    status = SS_LISTENING;
    return 0;
}


// closing (possibly unlinking) socket
int tUNIXSocket::close_socket()
{
    int r = 0;

    // so close your socket and kiss yourself goodbye
    debug_log ('s', "Closing socket %d", get_fd());
    if (get_fd() >= 0 && s_close (get_fd()) < 0)
    {
        log (L_ERROR, "Error closing socket: %m");
        r = -1;
    }

    // and server needs always something extra 
    if (get_role() == AS_SOCK_SERVER) 
    {
        // unlinking failed ?
        if (!s_access (get_addr(), F_OK) && unlink (get_addr ()))
        {
            log (L_WARN,"Can't set socket resources free");
            status = SS_ERRRESOURCES;
            return -1;
        }
    }
    
    // remember that this socket is no longer in use
    set_fd(-1);
    status = SS_EMPTY;
    return r;
}


// function for accepting connection on listening socket
int tUNIXSocket::accept_socket(class tSocketMgr &smanager)
{
    tUNIXSocket *newsock_s;
    int newsock;

    // lets try to accept it
    debug_log ('s', "Accepting socket on %d", get_fd ());
    if ((newsock = accept( get_fd(), 0, 0))==-1)
    {
        log(L_ERROR, "Error while accepting new socket");
        return -1;         
    }

    // create structure for the new socket 
    newsock_s = create_new_socket ();
    newsock_s->set_fd (newsock);
    newsock_s->set_role (AS_SOCK_CLIENT);
    newsock_s->set_smgr (&smanager);

    // let it inherit master's properties
    newsock_s->set_on_register (get_on_register ());
    newsock_s->set_on_unregister (get_on_unregister ());
    newsock_s->set_on_data_read (get_on_data_read ());
    newsock_s->set_on_data_write (get_on_data_write ());
    newsock_s->set_on_data_hup (get_on_data_hup ());
    newsock_s->set_on_data_err (get_on_data_err ());
    newsock_s->set_addr (get_addr ());
    newsock_s->set_created_by_smgr (true);
    newsock_s->status = SS_CONNECTED;

    // and do the registration work upon it
    debug_log ('s', "Registering new socket %d", newsock_s->get_fd ());
    if (smanager.register_socket (newsock_s, get_event_mask (), get_sid ()))
    {
        newsock_s->close_socket ();
        return -1;
    }

    // try handshake
    if (newsock_s->handshake () != 1)
    {
        // ahhh, a secret passage ... oops!
        smanager.unregister_socket (newsock_s);
        return -1;
    }

    // we're lucky today
    return 0;    
}


// atomical read into buffer - with limited length to len
// if the buffer is not big enough it will leave the message unread
ssize_t tUNIXSocket::at_read(void *buffer, ssize_t len)
{
    using messages::tPlainBuf;
    char framebuf [AT_FRAMEBUF_LEN];
    class tPlainBuf b (framebuf, sizeof(framebuf));
    tAtomicDesc synb, syne;

    if (get_status () != SS_CONNECTED)
    {
        log (L_ERROR, "Reading from an invalid socket");
        return -1;
    }
    
    // ok, now let's hope it won't fsck up
    // reread the synchro part
    if (s_read (get_fd (), framebuf, sizeof (framebuf)) 
            < static_cast<ssize_t> (sizeof (framebuf)))
        return -1;

    READ_INT32 (&b, synb.magic);
    READ_INT32 (&b, synb.length);
    READ_INT32 (&b, synb.flags);
    b.rewind ();

    // on sync start?
    if (synb.magic != V_MAGIC)
        return -1;      // umm, no...   
    // do we have a buffer long enough?
    if (synb.length > len)
        return -1;      // umm, no... ;)

    // read data
    if (s_read (get_fd(), buffer, synb.length) < synb.length)
        return -1;
    
    // read the final synchro part
    if (s_read (get_fd (), framebuf, sizeof (framebuf)) 
            < static_cast<ssize_t> (sizeof (framebuf)))
        return -1;

    READ_INT32 (&b, syne.magic);
    READ_INT32 (&b, syne.length);
    READ_INT32 (&b, synb.flags);

    // on sync start?
    if (syne.magic != V_MAGIC)
        return -1;      // umm, no...   
    // do we have a buffer long enough?
    if (syne.length != synb.length)
        return -1;      // umm, no... ;)

    return synb.length;
}


// atomical write into buffer of len bytes
ssize_t tUNIXSocket::at_write(void *buffer, ssize_t len)
{
    using messages::tPlainBuf;
    char framebuf [AT_FRAMEBUF_LEN];
    class tPlainBuf b (framebuf, sizeof(framebuf));
    tAtomicDesc syn = { V_MAGIC, len, 0 };

    if (get_status () != SS_CONNECTED)
    {
        log (L_ERROR, "Writing to an invalid socket");
        return -1;
    }

    STORE_INT32 (&b, syn.magic);
    STORE_INT32 (&b, syn.length);
    STORE_INT32 (&b, syn.flags);

    // write the synchro part
    if (s_write (get_fd (), framebuf, sizeof (framebuf)) < 0)
        return -1;

    // write data
    if (s_write(get_fd (), buffer, len) < 0) 
        return -1;

    // write the synchro part again
    if (s_write (get_fd (), framebuf, sizeof (framebuf)) < 0)
        return -1;

    return len;
}


ssize_t tUNIXSocket::sock_write (void *buffer, ssize_t len)
{
    ssize_t r;

    if (get_status () != SS_CONNECTED)
    {
        log (L_ERROR, "Writing to an invalid socket");
        return -1;
    }
        
    r = s_write (get_fd(), buffer, len);
    if (r<0)
        log (L_ERROR, "Error writing to socket %s.", get_address());

    return r;
}


ssize_t tUNIXSocket::sock_read(void *buffer, ssize_t len)
{
    ssize_t r;

    if (get_status () != SS_CONNECTED)
    {
        log (L_ERROR, "Reading from an invalid socket");
        return -1;
    }
        
    r = s_read (get_fd(), buffer, len);
    if (r<0)
        log(L_ERROR, "Error reading from socket %s.", get_address());
    
    return r;
}


ssize_t tUNIXSocket::sock_recv(void *buffer, ssize_t len, unsigned int options)
{
    ssize_t r;

    if (get_status () != SS_CONNECTED)
    {
        log (L_ERROR, "Receiving from an invalid socket");
        return -1;
    }

    r = s_recv (get_fd(), buffer, len, options);
    if (r<0)
        log(L_ERROR, "Can't receive data from socket %s", get_address());

    return r;
}


ssize_t tUNIXSocket::sock_poke(void *buffer, ssize_t len)
{
    ssize_t r;

    ASSERT (buffer && len >= 0);

    r = recv (get_fd(), buffer, len, MSG_PEEK);
    if (r < 0)
        return -errno;

    return r;
}


int tUNIXSocket::send_auth_info()
{
#if !defined(HAVE_LINUX_CREDENTIALS) && defined(HAVE_LOCAL_CREDENTIALS)

    char buf;
    struct iovec iov;
    struct msghdr msg;

#   ifdef HAVE_STRUCT_CMSGCRED
    /* Point to start of first structure */
    char cbuffer[sizeof(struct cmsghdr) + sizeof(struct cmsgcred)];
    struct cmsghdr *cmsg = (struct cmsghdr *) cbuffer;
#   endif

#   ifdef HAVE_STRUCT_SOCKCRED
    /* Point to start of first structure */
    char cbuffer[sizeof(struct cmsghdr) + sizeof(struct sockcred)];
    struct cmsghdr *cmsg = (struct cmsghdr *) cbuffer;
#   endif

    /* Send something, we do really care about ancilliary message data */
    buf = 0;
    iov.iov_base = &buf;
    iov.iov_len = 1;

    memset(&msg, 0, sizeof (msg));
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

#   ifdef HAVE_STRUCT_CMSGCRED
    /* Create control header, FreeBSD */
    msg.msg_control = cmsg;
    msg.msg_controllen = sizeof(cbuffer);
    memset(cmsg, 0, sizeof(cbuffer));
    cmsg->cmsg_len = sizeof(cbuffer);
    cmsg->cmsg_level = SOL_SOCKET;
    cmsg->cmsg_type = SCM_CREDS;
#   endif

    int r = 1;
    do 
    {
        if ((r = sendmsg (get_fd (), &msg, 0)) < 0)
        { 
            switch (errno)
            {
                case EINTR:
                case EAGAIN:        // retry
                    break;
                default:            // error
                    log (L_ERROR, "Can't send credentials on socket %ld: %m.", 
                            get_fd ());
                    close_socket ();
                    status = SS_BROKEN;
                    return -1;
            }
        }
    } while (!r);

    debug_log ('c', "Credentials sent on socket %d", get_fd());

#endif

    return 1;
}


#if defined(HAVE_LINUX_CREDENTIALS)

int tUNIXSocket::receive_auth_info()
{
    /* Linux style: use getsockopt(SO_PEERCRED) */
    struct ucred peercred;
    socklen_t so_len = sizeof(peercred);

    errno = 0;
    if (getsockopt(get_fd (), SOL_SOCKET, SO_PEERCRED, &peercred, &so_len) != 0 
            || so_len != sizeof(peercred))
    {
        /* Bad luck, no valid credentials structure available */
        log (L_ERROR, "getsockopt error on socket %ld: %m.", get_fd ());
        return -1;
    }

    auth_id = peercred.uid;
    debug_log ('s', "Socket %li authentication id: %li", get_fd (), auth_id); 

    /* Report success */
    return 1;
}

#elif defined(HAVE_LOCAL_CREDENTIALS)

int tUNIXSocket::receive_auth_info()
{
    /* BSD way to gain credentials structure */
    struct msghdr msg;
    struct iovec iov;
    char buf;
    tCred *cr;

    /* Compute size of input buffer (beware of NetBSD) */
    char cbuffer[ ALIGN (sizeof (struct cmsghdr)) + ALIGN (sizeof (tCred)) ];

    /* Point to start of first structure */
    struct cmsghdr *cmsg = (struct cmsghdr *) cbuffer;

    /* Set message and iovector */
    memset(&msg, 0, sizeof (msg));
    msg.msg_name = NULL;
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;
    msg.msg_control = (char *) cmsg;
    msg.msg_controllen = sizeof (cbuffer);
    memset(cmsg, 0, sizeof (cbuffer));

    /* Receive the info character and credentials struct */
    iov.iov_base = &buf;
    iov.iov_len = 1;

    /* Get it */
    int r = 1;
    do 
    {
        if ((r = recvmsg (get_fd (), &msg, 0)) < 0)
        { 
            switch (errno)
            {
                case EINTR:
                case EAGAIN:        // retry
                    break;
                default:            // error
                    log (L_ERROR, "Recvmsg error: %m");
                    goto err_receiving_credentials;
            }
        }
    } while (!r);

    if (cmsg->cmsg_len < sizeof (cbuffer) || cmsg->cmsg_type != SCM_CREDS)
    {
      err_receiving_credentials:
        log (L_ERROR, "Can't receive credentials on socket %ld.", get_fd ());
        return -1;
    }
    
    /* It seems to be okay, save the data */
    cr = (tCred *) CMSG_DATA(cmsg);
    auth_id = cr->cr_uid;
    debug_log ('s', "Socket %li authentication id: %li", get_fd (), auth_id); 

    return 1;
}

#else

int tUNIXSocket::receive_auth_info()
{
#   warning No unix domain socket authentication available - disabling it.
    return 1;
}

#endif

}
