/**
 * @brief Echelog Socks library - Non-blocking unix domain sockets
 *
 * @author Kamil Toman <ktoman@email.cz>
 * 
 */

#include "lib/archdep.h"
#include "lib/lib.h"
#include "socks/socketmgr.h"
#include "socks/unixsocket_nb.h"
#include "messages/plainbuf.h"
#include "messages/testbuf.h"

#include "lib/archdep.h"
#include <sys/types.h>
#include <sys/un.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

#ifdef SCM_CREDS
#include <sys/uio.h> /* for struct iovec */
#include <sys/ucred.h>
#endif


extern int h_errno;

namespace socks {

using namespace lib;


/** handlers of events in usock-layer */
static void _hup_f(class tSocket *sock) { sock->hup_handler (); }
static void _read_f(class tSocket *sock) { sock->read_handler (); }
static void _write_f(class tSocket *sock) { sock->write_handler (); }
static void _err_f(class tSocket *sock) { sock->err_handler (); }


// constructor - create socket and set file descriptor
tUNIXSocketNB::tUNIXSocketNB() : tSocketNB (), tUNIXSocketAuth (),
    auth_status (AU_NONE), auth_id (0)
{
    // set both status to empty
    set_status (SS_EMPTY);
}


// closes the socket
tUNIXSocketNB::~tUNIXSocketNB()
{
    // perform the shutdown
    close_socket();

    // free ap specific data
    xfree (app_specific_data);
}


int tUNIXSocketNB::new_socket ()
{
    long new_fd = socket(AF_UNIX, SOCK_STREAM, 0);

    // hmm, socket cannot be opened
    if (new_fd < 0)
    {
        log (L_ERROR, "Can't create a new socket.");
        status = SS_ERRSOCK;
        return -1;
    }

    set_fd(new_fd); 

    // make the socket nonblocking
    if (set_non_blocking() )
    {
        set_status (SS_ERRSOCK);
        return -1;
    }

    // socket is created but not connected
    set_status (SS_DEFAULT);

    return 0;
}


// closes the socket and unregisters it 
int tUNIXSocketNB::close_socket()
{
    int r = 0;

    // try to close the socket
    if (get_fd() >= 0 && s_close (get_fd()) )
    {
        log (L_ERROR, "Error closing socket: %m");
        r = -1;
    }

    // server needs always something extra 
    if (get_role() == AS_SOCK_SERVER) 
    {
        // unlinking failed ?
        if(!s_access (get_addr (), F_OK) && unlink (get_addr ()))
        {
            log(L_WARN,"Can't set socket resources free");
            status = SS_ERRRESOURCES;
            return -1; 
        }
    }

    // inactivate the usock layer and invalidate the fd
    set_status (SS_EMPTY);
    set_fd(-1);

    return r;
}


// creates client side of the socket
int tUNIXSocketNB::create_client(const char *name, tSocketMgr &smanager)
{
    ASSERT (name);

    // get new socket
    if (new_socket()) 
        return -1;

    // set new socket address
    set_addr (name);

    // backup socket manager object 
    smgr = &smanager; 
  
    // mark that this is client side
    set_role(AS_SOCK_CLIENT);

    // set the event mask 
    set_event_mask (POLLPRI | POLLIN | POLLOUT);

    // set the handlers
    set_on_data_read (_read_f);
    set_on_data_write (_write_f);
    set_on_data_err (_err_f);
    set_on_data_hup (_hup_f);

    return 0;
}


// creates server side of the socket
int tUNIXSocketNB::create_server(const char *name, tSocketMgr &smanager)
{
    ASSERT (name);

    // get new socket
    if (new_socket()) 
        return -1;
    
    /* There should be no such socket file */
    if (!s_eaccess (name, F_OK) && unlink (name))
        log (L_WARN,"Can't set old socket resources free");
    
    // set new socket address
    set_addr (name);

    // mark that this is server side
    set_role(AS_SOCK_SERVER);

    // disallow socket reusal
    setsockopt (get_fd(), SOL_SOCKET, SO_REUSEADDR, 0, 0);

    // bind with the filename 
    debug_log ('s', "Binding socket %d (%s)", get_fd(), get_addr ());
    if (bind (get_fd(), (struct sockaddr *) &addr,
        sizeof(struct sockaddr_un)))
    {
           log(L_ERROR,"Cannot bind socket.");
           status = SS_ERRBIND;
           err_handler();
           return -1;
    }

    debug_log ('s', "Listening at socket %d (%s)", get_fd(), get_addr());
    // make the socket listenning for requests
    if (listen (get_fd(), LISTENCNT) )
    {
        log(L_ERROR, "Cannot make socket listening: %m");
        status = SS_ERRLISTENING;
        err_handler();    
        return -1;
    }

    // now we are listenning
    status = SS_LISTENING;

    // backup socket manager object 
    smgr = &smanager; 

    // set the event mask 
    set_event_mask (POLLPRI | POLLIN);

    // set the handlers
    set_on_data_read (_read_f);
    set_on_data_write (_write_f);
    set_on_data_err (_err_f);
    set_on_data_hup (_hup_f);
   
    // register the socket
    if (smgr->register_socket(this, POLLPRI | POLLIN))
    {
        log(L_ERROR, "Cannot register unix socket.");
        status = SS_ERRRESOURCES;
        err_handler();
        return -1;
    }

    return 0;
}


// function for accepting connection on listening socket
int tUNIXSocketNB::accept_socket(class tSocketMgr &smanager)
{
    class tUNIXSocketNB *sock;
    int new_sock;

    if (status != SS_LISTENING && status != SS_ACCEPTING)
    {
        log (L_ERROR, "Server cannot accept connections unless "
               "in listenning mode.");
        return -1;
    }
   
    // try to accept new socket       
    if( (new_sock = accept (get_fd(), 0, 0)) < 0)
    {
        if (errno == EAGAIN || errno == EINTR) 
            return 0;

        log (L_ERROR, "Error accepting connection: %m");
        return -1;
    }
    debug_log ('s', "New socket (sock=%d) accepted.", new_sock);

    set_status (SS_LISTENING);

    // creates new socket
    sock = create_new_socket ();
     
    // assign it to object 
    sock->set_fd (new_sock); 
    sock->set_status (SS_CONNECTED);
    sock->set_role (AS_SOCK_CLIENT);
    sock->set_smgr (&smanager);
    sock->set_created_by_smgr (true);

    // let it inherit master's properties
    sock->set_on_register (get_on_register ());
    sock->set_on_unregister (get_on_unregister ());
    sock->set_on_data_read (get_on_data_read ());
    sock->set_on_data_write (get_on_data_write ());
    sock->set_on_data_hup (get_on_data_hup ());
    sock->set_on_data_err (get_on_data_err ());

    sock->set_addr (get_addr ());
    sock->set_on_err (get_on_err ());
    sock->set_on_hup (get_on_hup ());
    sock->set_on_read_available (get_on_read_available ());
    sock->set_on_write_finished (get_on_write_finished ());
    sock->set_on_read_finished (get_on_read_finished ());
    sock->set_on_connected (get_on_connected ());

    sock->status = SS_CONNECTED;
    
    // make the socket nonblocking
    if (sock->set_non_blocking ())
    {
        delete sock;
        return -1;
    }

    // register the socket
    if (smgr->register_socket (sock, POLLIN | POLLPRI | POLLOUT, get_sid()))
    {
        delete sock;
        return -1;
    }
    
    sock->clear_io ();
    sock->last_operation = OP_HANDSHAKE;

    debug_log ('s', "Socket fd=%d ACCEPTED, awaiting handshake.", 
            sock->get_fd());

    return 0; 

}


int tUNIXSocketNB::reconnect()
{
    if (role != AS_SOCK_CLIENT)
    {
        log(L_ERROR,"Disallowed server side reconnect() call.");
        return -1;
    }
   
    if (status != SS_DEFAULT && status != SS_BROKEN && status != SS_CONNECTED
            && status != SS_CONNECTING)
    {   
        log(L_ERROR,"Cannot reconnect socket from this state.");
        return -1;
    }

    if (status == SS_DEFAULT || status == SS_BROKEN)
    {
        status = SS_CONNECTING;
        
        // register the socket to ensure connection completion (needs POLLOUT)
        if (smgr->register_socket (this, POLLPRI | POLLIN | POLLOUT))
        {
            status = SS_ERRRESOURCES;
            err_handler();
            return -1;
        }
   
        debug_log ('s', "Sock=%d is going to be connected.",  get_fd() );

        if (connect (get_fd(), (struct sockaddr *) &addr, SUN_LEN (&addr)) < 0) 
        {
            if (errno == EINPROGRESS) 
                return 0;

            if (errno == EINTR || errno == EAGAIN)
            {
                status = SS_DEFAULT;
                return 0;
            }    
          
            log(L_ERROR, "Cannot connect on sock=%ld.", get_fd ());
            status = SS_BROKEN;
            err_handler();
            return -1;
        }

        status = SS_CONNECTED;
        debug_log ('s', "Sock=%d sucesfully connected.", get_fd ());
    }    

    if (status == SS_CONNECTED)
    {    
        smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
        clear_io ();
        last_operation = OP_HANDSHAKE;

        debug_log('s', "Socket fd=%d connected, awaiting handshake.", get_fd());
        return 0; 
    }
    
    return -1;
}


ssize_t tUNIXSocketNB::sock_recv(void *buffer, ssize_t len, unsigned int options)
{
    using namespace lib;
    die("Non-blocking sock_recv not implemented yet.");
    return -1;
}


ssize_t tUNIXSocketNB::sock_poke(void *buffer, ssize_t len)
{
    ssize_t r;
    ASSERT(status == SS_CONNECTED || status == SS_CONNECTING);
    ASSERT (buffer && len >= 0);

    r = recv (get_fd(), buffer, len, MSG_PEEK);
    if (r < 0)
         return -errno;

    return r;
}


static int safe_read (class tUNIXSocketNB *s, char *buffer, ssize_t bytes)
{
    int ret;

    ret = read (s->get_fd (), buffer, bytes);

    if (ret < 0)
    {
        switch (errno)
        {
            case EINTR:
            case EAGAIN:
                debug_log ('s', "Safe read waits for incoming data.");
                return 0;
            default:
                log (L_ERROR, "Read error on socket fd=%ld.", s->get_fd ());
                s->clear_io ();
                s->set_last_operation (OP_NO_OPERATION);
                s->set_status (SS_BROKEN);
                s->err_handler();
                return -1;
        }
    }

    return ret;
}


static int safe_write (class tUNIXSocketNB *s, char *buffer, ssize_t bytes)
{
    int ret;

    ret = write (s->get_fd (), buffer, bytes);

    if (ret < 0)
    {
        switch (errno)
        {
            case EINTR:
            case EAGAIN:
                return 0;
            default:
                log (L_ERROR, "Write error on socket fd=%ld.", s->get_fd ());
                s->clear_io ();
                s->set_last_operation (OP_NO_OPERATION);
                s->set_status (SS_BROKEN);
                s->err_handler();
                return -1;
        }
    }

    return ret;
}


// perform the handshake
int tUNIXSocketNB::handshake ()
{
    long r;
#if defined(HAVE_LINUX_CREDENTIALS) || defined(HAVE_LOCAL_CREDENTIALS)    

    if (auth_status == AU_NONE || auth_status == AU_AUTHENTICATING)
    {
        /* Connection authentication - send auth info */

#   if !defined(HAVE_LINUX_CREDENTIALS) \
        && ( defined(HAVE_STRUCT_FCRED) || defined(HAVE_STRUCT_SOCKCRED) )

        /* For BSD/OS and NetBSD we need to set this before the authentication
         * request or the whole process fscks up */
        long on = 1;
        if (setsockopt(get_fd (), 0, LOCAL_CREDS, &on, sizeof(on)) < 0)
            log (L_ERROR, "Can't do setsockopt to socket %li: %m", get_fd ()); 
#   endif

       r = send_auth_info ();

       switch (r)
        {
            case 1:     // okay
                auth_status = AU_AUTHENTICATION_SENT;
                break;
            case 0:     // retry
                clear_io ();
                smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
                last_operation = OP_HANDSHAKE;
                return 0;
            case -1:    // error
                log(L_ERROR,"UNIXSocketNB handshake failed for fatal reasons.");
                auth_status = AU_BROKEN;
                err_handler();
                return -1;
        }
    }

    if (auth_status == AU_AUTHENTICATION_SENT)
    {
        /* Connection authentication - receive auth info */
        r = receive_auth_info ();

        switch (r)
        {
            case 1:     // okay
                break;
            case 0:     // retry
                clear_io ();
                smgr->set_event_mask (this, POLLPRI | POLLIN);
                last_operation = OP_HANDSHAKE;
                return 0;
            case -1:    // error
                log(L_ERROR,"UNIXSocketNB handshake failed for fatal reasons.");
                auth_status = AU_BROKEN;
                err_handler();
                return -1;
        }
    }

#else
#   warning No unix domain socket authentication available - disabling it.
#endif

    return 1;
}


int tUNIXSocketNB::send_auth_info()
{
    auth_status = AU_AUTHENTICATING;

#if !defined(HAVE_LINUX_CREDENTIALS) && defined(HAVE_LOCAL_CREDENTIALS)

    char buf;
    struct iovec iov;
    struct msghdr msg;

#   ifdef HAVE_STRUCT_CMSGCRED
    /* Point to start of first structure */
    char cbuffer[sizeof(struct cmsghdr) + sizeof(struct cmsgcred)];
    struct cmsghdr *cmsg = (struct cmsghdr *) cbuffer;
#   endif

#   ifdef HAVE_STRUCT_SOCKCRED
    /* Point to start of first structure */
    char cbuffer[sizeof(struct cmsghdr) + sizeof(struct sockcred)];
    struct cmsghdr *cmsg = (struct cmsghdr *) cbuffer;
#   endif

    /* Send something, we do really care about ancilliary message data */
    buf = 0;
    iov.iov_base = &buf;
    iov.iov_len = 1;

    memset(&msg, 0, sizeof (msg));
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

#   ifdef HAVE_STRUCT_CMSGCRED
    /* Create control header, FreeBSD */
    msg.msg_control = cmsg;
    msg.msg_controllen = sizeof(cbuffer);
    memset(cmsg, 0, sizeof(cbuffer));
    cmsg->cmsg_len = sizeof(cbuffer);
    cmsg->cmsg_level = SOL_SOCKET;
    cmsg->cmsg_type = SCM_CREDS;
#   endif

    int r = sendmsg (get_fd (), &msg, 0);
    if (!r)
        return 0;
    if (r == -1)
    { 
        switch (errno)
        {
            case EINTR:
            case EAGAIN:        // retry
                return 0;
            default:            // error
                log (L_ERROR, "Can't send credentials on socket %ld: %m.", 
                        get_fd ());
                clear_io ();
                set_last_operation (OP_NO_OPERATION);
                status = SS_BROKEN;
                return -1;
        }
    }

    debug_log ('c', "Credentials sent on socket %d", get_fd());

#endif

    return 1;
}


#if defined(HAVE_LINUX_CREDENTIALS)

int tUNIXSocketNB::receive_auth_info()
{
    /* Linux style: use getsockopt(SO_PEERCRED) */
    struct ucred peercred;
    socklen_t so_len = sizeof(peercred);

    errno = 0;
    if (getsockopt(get_fd (), SOL_SOCKET, SO_PEERCRED, &peercred, &so_len) != 0 
            || so_len != sizeof(peercred))
    {
        /* Bad luck, no valid credentials structure available */
        log (L_ERROR, "getsockopt error on socket %ld: %m.", get_fd ());
        clear_io ();
        set_last_operation (OP_NO_OPERATION);
        status = SS_BROKEN;
        return -1;
    }

    auth_id = peercred.uid;
    debug_log ('s', "Socket %li authentication id: %li", get_fd (), auth_id); 

    /* Report success */
    return 1;
}

#elif defined(HAVE_LOCAL_CREDENTIALS)

int tUNIXSocketNB::receive_auth_info()
{
    /* BSD way to gain credentials structure */
    struct msghdr msg;
    struct iovec iov;
    char buf;
    tCred *cr;

    /* Compute size of input buffer (beware of NetBSD) */
    char cbuffer[ ALIGN (sizeof (struct cmsghdr)) + ALIGN (sizeof (tCred)) ];

    /* Point to start of first structure */
    struct cmsghdr *cmsg = (struct cmsghdr *) cbuffer;

    /* Set message and iovector */
    memset(&msg, 0, sizeof (msg));
    msg.msg_name = NULL;
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;
    msg.msg_control = (char *) cmsg;
    msg.msg_controllen = sizeof (cbuffer);
    memset(cmsg, 0, sizeof (cbuffer));

    /* Receive the info character and credentials struct */
    iov.iov_base = &buf;
    iov.iov_len = 1;

    /* Get it */
    int r = recvmsg(get_fd (), &msg, 0);
    if (!r)
        return 0;
    else if (r == -1)
    { 
        switch (errno)
        {
            case EINTR:
            case EAGAIN:        // retry
                return 0;
            default:            // error
                log (L_ERROR, "Recvmsg error: %m");
                goto err_receiving_credentials;
        }
    }

    if (cmsg->cmsg_len < sizeof (cbuffer) || cmsg->cmsg_type != SCM_CREDS)
    {
      err_receiving_credentials:
        log (L_ERROR, "Can't receive credentials on socket %ld.", get_fd ());
        clear_io ();
        set_last_operation (OP_NO_OPERATION);
        status = SS_BROKEN;
        return -1;
    }
    
    /* It seems to be okay, save the data */
    cr = (tCred *) CMSG_DATA(cmsg);
    auth_id = cr->cr_uid;
    debug_log ('s', "Socket %li authentication id: %li", get_fd (), auth_id); 

    return 1;
}

#else

int tUNIXSocketNB::receive_auth_info()
{
#   warning No unix domain socket authentication available - disabling it.
    return 1;
}


#endif


int tUNIXSocketNB::connect_again()
{
    ASSERT(status == SS_CONNECTED);
    ASSERT(last_operation == OP_HANDSHAKE || last_operation == OP_NO_OPERATION);
    ASSERT(auth_status != AU_BROKEN);

    if (auth_status == AU_NONE || auth_status == AU_AUTHENTICATING || 
            auth_status == AU_AUTHENTICATION_SENT)
        switch (handshake ())
        {
            case 0:  // retry
                return 0;
            case 1:  // okay
                break;
            case -1: // fatal error
                return -1;
        }

    auth_status = AU_AUTHENTICATED;
    smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
    clear_io ();
    last_operation = OP_NO_OPERATION;

    debug_log ('s', "Handshake successful on socket fd=%d", get_fd ());
    if (on_connected) 
        on_connected (this);

    return 0;
}


int tUNIXSocketNB::accept_again()
{
    ASSERT(status == SS_CONNECTED);
    ASSERT(last_operation == OP_HANDSHAKE || last_operation == OP_NO_OPERATION);
    ASSERT(auth_status != AU_BROKEN);

    if (auth_status == AU_NONE || auth_status == AU_AUTHENTICATING || 
            auth_status == AU_AUTHENTICATION_SENT)
        switch (handshake ())
        {
            case 0:  // retry
                return 0;
            case 1:  // okay
                break;
            case -1: // fatal error
                return -1;
        }

    auth_status = AU_AUTHENTICATED;
    smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
    clear_io ();
    last_operation = OP_NO_OPERATION;

    debug_log ('s', "Handshake successful on socket fd=%d", get_fd ());
    if (on_connected) 
        on_connected (this);

    return 0;
}


int tUNIXSocketNB::read_again()
{
    int ret;

    ASSERT(status == SS_CONNECTED);
    ASSERT(auth_status == AU_AUTHENTICATED);
    ASSERT((last_operation == OP_IO && r_io.last_op == IO_OP_READING) || 
            last_operation == OP_NO_OPERATION);
                
    ret = read (get_fd (), r_io.last_op_buf, r_io.last_op_bytes);
    debug_log ('s', "Read returned (%d)", ret);

    if (ret > 0)
    {
        r_io.last_op_bytes = ret;
        stop_r_io ();
        smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
        if (on_read_finished) 
            on_read_finished (this);
        return 0;
    } else 
    if (ret == 0)
    {
        r_io.last_op_bytes = ret;
        stop_r_io ();
        smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
        if (on_read_finished) 
            on_read_finished (this);
        return 0;
    } else
    {
        switch (errno)
        {
            case EINTR:
            case EAGAIN:
                debug_log ('s', "Read waits for incoming data.");
                r_io.last_op = IO_OP_READING;
                last_operation = OP_IO;
                smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
                return 0;
            default:
                log (L_ERROR, "Read error on socket fd=%ld.", get_fd ());
                last_operation = OP_NO_OPERATION;
                stop_r_io ();
                status = SS_ERROR;
                err_handler();
                return -1;
        }
    }
}


int tUNIXSocketNB::write_again()
{
    int ret;

    ASSERT(status == SS_CONNECTED);
    ASSERT(auth_status == AU_AUTHENTICATED);
    ASSERT((last_operation == OP_IO && w_io.last_op == IO_OP_WRITING) || 
            last_operation == OP_NO_OPERATION);
                
    ret = write (get_fd (), w_io.last_op_buf, w_io.last_op_bytes);
    debug_log ('s', "Write returned (%d)", ret);

    if (ret > 0)
    {
        w_io.last_op_bytes = ret;
        stop_w_io ();
        smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
        if (on_write_finished) 
            on_write_finished (this);
        return 0;
    } else 
    if (ret == 0)
    {
        w_io.last_op_bytes = ret;
        stop_w_io ();
        smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
        if (on_write_finished) on_write_finished (this);
        return 0;
    } else
    {
        switch (errno)
        {
            case EINTR:
            case EAGAIN:
                debug_log ('s', "Read waits for incoming data.");
                smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
                w_io.last_op = IO_OP_WRITING;
                last_operation = OP_IO;
                return 0;
            default:
                log (L_ERROR, "Write error on socket fd=%ld.", get_fd ());
                last_operation = OP_NO_OPERATION;
                stop_w_io ();
                status = SS_ERROR;
                err_handler();
                return -1;
        }
    }
}


int tUNIXSocketNB::atom_read_again()
{
    using messages::tPlainBuf;
    class tPlainBuf b (r_io.framebuf, sizeof(r_io.framebuf) );
    long ret;
    int32_t x,y;

    ASSERT(status == SS_CONNECTED);
    ASSERT(auth_status == AU_AUTHENTICATED);
    ASSERT((last_operation == OP_IO && r_io.last_op == IO_OP_ATOM_READING) || 
            last_operation == OP_NO_OPERATION);

    r_io.last_op = IO_OP_ATOM_READING;
    last_operation = OP_IO;
    switch (r_io.atom_op_state)
    {
        case AT_BEGINFRAME:
            ret = safe_read (this, r_io.framebuf +  r_io.already_processed, 
                             sizeof(r_io.framebuf) - r_io.already_processed);
            if (ret >= 0)
            {
                r_io.already_processed += ret;
                if (r_io.already_processed == sizeof(r_io.framebuf) ) 
                {
                     READ_INT32 (&b, r_io.atom_frame.magic);
                     READ_INT32 (&b, r_io.atom_frame.length);
                     READ_INT32 (&b, r_io.atom_frame.flags);
                     if (r_io.atom_frame.magic != V_MAGIC || 
                         r_io.atom_frame.length > r_io.last_op_bytes)
                     {
                        log (L_ERROR, "Error while processing start frame.");
                        r_io.last_op_bytes = -1;
                        clear_io ();
                        status = SS_BROKEN;
                        err_handler ();
                        return 0;
                     }
                     b.rewind();
                     r_io.already_processed = 0;
                     r_io.atom_op_state = AT_DATA;
                } else return 0;
            } else return -1; 
                
        case AT_DATA:
            ret = safe_read (this, 
                    r_io.last_op_buf + r_io.already_processed,
                    r_io.atom_frame.length - r_io.already_processed);
            if (ret >= 0)
            {
                r_io.already_processed += ret;
                if (r_io.already_processed == r_io.atom_frame.length)
                {
                     r_io.already_processed = 0;
                     r_io.atom_op_state = AT_ENDFRAME;
                } else return 0;
            } else return -1;    
            
        case AT_ENDFRAME:
            ret = safe_read (this, r_io.framebuf +  r_io.already_processed, 
                             sizeof(r_io.framebuf) - r_io.already_processed);
            if (ret >= 0)
            {
                r_io.already_processed += ret;
                if (r_io.already_processed == sizeof(r_io.framebuf) ) 
                {
                     READ_INT32 (&b, x);
                     READ_INT32 (&b, y);
                     if (r_io.atom_frame.magic != x || 
                             r_io.atom_frame.length != y  )
                     {
                        log (L_ERROR, "Error while processing end frame.");
                        r_io.last_op_bytes = -1;
                        clear_io ();
                        status = SS_BROKEN;
                        err_handler ();
                        return 0;
                     }
                     r_io.last_op_bytes = r_io.atom_frame.length;
                     stop_r_io ();
                     smgr->set_event_mask (this, POLLPRI | POLLIN | POLLOUT);
                     if (on_read_finished) 
                         on_read_finished (this);
                     return 0;
                } else return 0;
            } else return -1; 
            break;

        default:  
            log (L_ERROR, "UnixSocketNB: Unexpected state of atom-op.");
    }

    return -1;
}


int tUNIXSocketNB::atom_write_again()
{
    using messages::tPlainBuf;
    class tPlainBuf b (w_io.framebuf, sizeof(w_io.framebuf) );
    int ret;

    ASSERT(status == SS_CONNECTED);
    ASSERT(auth_status == AU_AUTHENTICATED);
    ASSERT((last_operation == OP_IO && w_io.last_op == IO_OP_ATOM_WRITING) || 
            last_operation == OP_NO_OPERATION);

    STORE_INT32 (&b, w_io.atom_frame.magic);
    STORE_INT32 (&b, w_io.atom_frame.length);
    STORE_INT32 (&b, w_io.atom_frame.flags);

    w_io.last_op = IO_OP_ATOM_WRITING;
    last_operation = OP_IO;
    switch (w_io.atom_op_state)
    {
        case AT_BEGINFRAME:
            ret = safe_write (this, w_io.framebuf +  w_io.already_processed, 
                              sizeof(w_io.framebuf) - w_io.already_processed);
            if (ret >= 0)
            {
                w_io.already_processed += ret;
                if (w_io.already_processed == sizeof(w_io.framebuf) ) 
                {
                     w_io.already_processed = 0;
                     w_io.atom_op_state = AT_DATA;
                } else return 0;
            } else return -1; 
                
        case AT_DATA:
            ret = safe_write (this, 
                    w_io.last_op_buf + w_io.already_processed, 
                    w_io.atom_frame.length - w_io.already_processed);
            if (ret >= 0)
            {
                w_io.already_processed += ret;
                if (w_io.already_processed == w_io.atom_frame.length)
                {
                     w_io.already_processed = 0;
                     w_io.atom_op_state = AT_ENDFRAME;
                } else return 0;
            } else return -1;    
            
        case AT_ENDFRAME:

            ret = safe_write (this, w_io.framebuf +  w_io.already_processed, 
                              sizeof(w_io.framebuf) - w_io.already_processed);
            if (ret >= 0)
            {
                w_io.already_processed += ret;
                if (w_io.already_processed == sizeof(w_io.framebuf) ) 
                {
                     w_io.last_op_bytes = w_io.atom_frame.length;
                     stop_w_io ();
                     smgr->set_event_mask (this, POLLPRI | POLLIN);
                     if (on_write_finished) 
                         on_write_finished (this);
                     return 0;
                } else return 0;
            } else return -1; 
            break;

        default:  
            log (L_ERROR, "UnixSocketNB: Unexpected state of atom-op.");
    }

    return -1;
}

}
