/**
 * @brief Echelog Spool Library -- BinStore implementation
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include "lib/lib.h"
#include "spool/binstore.h" 

#include <netinet/in.h>
#include <unistd.h>
 
namespace spool
{

using namespace lib;

/* magic constant */
static unsigned long C_MAGIC;

/* Initializes the pagefile structure */
tBinStore::tBinStore(int wr_thr)
{
    
    /* denote that we do not have assigned files yet */
    present = 0;

    /* set cached/write-through behaviour */
    write_through = wr_thr;

    /* obtain magic constant */
    C_MAGIC = htonl( 0xCA453E76 );

}

/* creates new BinStore and opens it for write only */
int tBinStore::create(char *data_fname, char *index_fname)
{
    ASSERT(data_fname);
    ASSERT(index_fname);

    access_mode = 2;
    
    /* first backup filenames */            
    if(!(data_filename = strdup( data_fname )) )
    {
         log( L_ERROR, "BINSTORE: Can't store binstore datafile name" 
                 " (\"%s\"): %m", data_fname );
         return -1;
    }

    /* create the data file */
    if(!(index_filename = strdup( index_fname )) )
    {
         log( L_ERROR, "BINSTORE: Can't store binstore indexfile name"
                 " (\"%s\"): %m", index_fname );
         return -1;
    }

    data_fd = s_open (data_fname, O_RDWR | O_CREAT | O_TRUNC , 0600);
    if ( data_fd < 0 ) 
    {
        log( L_ERROR, "BINSTORE: Can't create data file (\"%s\"): %m",
                data_fname );
        return -1;
    }

    /* create the index file */
    index_fd = s_open (index_fname, O_RDWR | O_CREAT | O_TRUNC , 0600);
    if ( index_fd < 0 )
    {
        log( L_ERROR, "BINSTORE: Can't create index file (\"%s\"): %m",
                index_fname );
        return -1;
    }

    /* ensure it is at least in disc caches */
    fsync(data_fd);
    fsync(index_fd);

    /* set the number of items held, cursor and mark that files are assigned */ 
    present = 1;
    last_id = 0;
    tail = 0;

    return 0;
}


/* opens BinStore for write only */
int tBinStore::open_w(char *data_fname, char *index_fname)
{
    off_t data_file_len, index_file_len;
    off_t data_read, index_read;
    tBinItemOffset new_item_offset;
    struct tBinItem item_desc;

    ASSERT(data_fname);
    ASSERT(index_fname);

    /* file will be opened for write access */
    access_mode = 2;
    
    /* first backup filenames */            
    if(!(data_filename = strdup( data_fname )) )
    {
         log( L_ERROR, "BINSTORE: Can't store binstore datafile name" 
                 " (\"%s\"): %m", data_fname );
         return -1;
    }

    if(!(index_filename = strdup( index_fname )) )
    {
         log( L_ERROR, "BINSTORE: Can't store binstore indexfile name"
                 " (\"%s\"): %m", index_fname );
         return -1;
    }

    /* open files */
    data_fd = s_open (data_fname, O_RDWR | O_CREAT , 0600);
    if ( data_fd < 0 ) 
    {
        log( L_ERROR, "BINSTORE: Can't open/create data file for writing "
                "(\"%s\"): %m", data_fname );
        return -1;
    }

    index_fd = s_open (index_fname, O_RDWR | O_CREAT , 0600);
    if ( index_fd < 0 )
    {
       log( L_ERROR, "BINSTORE: Can't open index file for writing (\"%s\"): %m",
                index_fname );
       return -1;
    }

    WRITELOCK_FILE (data_fd);
    WRITELOCK_FILE (index_fd);
    /* determine data file length */
    data_file_len = s_filelen (data_fd);

    /* determine index file length */
    index_file_len = s_filelen (index_fd); 

    /* set cursor and item count to zero */
    last_id = 0;
    tail = 0;

    /* we have alredy read 0 bytes from data and index file */
    data_read = index_read = 0;

    /* now proceed consistency check */
    while (1)
    {

       /* if there is record in index file */
       if ( index_read + static_cast<off_t>(sizeof(tBinItemOffset)) <=
               index_file_len )
       {
      
          /* read its start */
          if (s_read( index_fd, &new_item_offset, sizeof(tBinItemOffset))!= 
                    sizeof(tBinItemOffset) )
          {
              log( L_ERROR, "BINSTORE: Cannot read index file (\"%s\"): %m",
                      index_fname );
              return -1;
          }

          /* if start is somewhere where it could not be */
          if ( new_item_offset != data_read )
              break;

          index_read += sizeof(tBinItemOffset);

          /* ok, try tro read appropriate data */
          if ( data_read + static_cast<off_t>(sizeof(struct tBinItem)) <=
                  data_file_len )
          {
              
              if (s_read( data_fd, &item_desc, sizeof(struct tBinItem))!=
                     sizeof(struct tBinItem) )
              {
                 log( L_ERROR, "BINSTORE: Cannot read data file (\"%s\"): %m",
                      data_fname );
                 return -1;
              }

              data_read += sizeof(struct tBinItem);
              
              /* check on item */ 
              if ( item_desc.magic != C_MAGIC || item_desc.id != last_id + 1 || 
                   item_desc.length + data_read > data_file_len )
              {
                 data_read -= sizeof(struct tBinItem);
                 index_read -= sizeof(tBinItemOffset);
                 break;
              }

              data_read += item_desc.length;

              last_id ++;
             
          } else
          {
              index_read -= sizeof(tBinItemOffset);
              break;
          }
       } else /* index file is corrupted */ 
       {
          new_item_offset = data_read; 

          /* there is some item in data file */
          if ( data_read + static_cast<off_t>(sizeof(struct tBinItem)) <=
                  data_file_len )
          {
              
              if (s_read( data_fd, &item_desc, sizeof(struct tBinItem))!=
                     sizeof(struct tBinItem) )
              {
                 log( L_ERROR, "BINSTORE: Cannot read data file (\"%s\"): %m",
                      data_fname );
                 return -1;
              }

              data_read += sizeof(struct tBinItem);
              
              /* consistency check */ 
              if ( item_desc.magic != C_MAGIC || item_desc.id != last_id + 1 || 
                   item_desc.length + data_read > data_file_len )
              {
                 data_read -= sizeof(struct tBinItem);
                 break;
              }

              data_read += item_desc.length;

              last_id ++;

              /* add the correct offset to idx file */
              if (s_write( index_fd, &new_item_offset, sizeof(tBinItemOffset))!=
                     sizeof(tBinItemOffset) )
              {
                  log( L_ERROR, "BINSTORE: Cannot write into index file " 
                            "(\"%s\"): %m", index_fname );
                  return -1;
              }

              index_read += sizeof(tBinItemOffset);
              
          } else 
              break;
      }

      /* skip the data to next data header */
      if ( lseek ( data_fd, data_read, SEEK_SET ) != data_read )
      {
         log( L_ERROR, "BINSTORE: Can't seek in the data file (\"%s\") to "
                 "%ld: %m", data_filename, static_cast<unsigned long>
                 ( data_read ) );
         return -1;
      }

    }

    /* if index or data files were corrupted, then truncate them to their 
     * correct part */
    if (data_read < data_file_len)
        if ( ftruncate(data_fd, data_read) )
        {
              log( L_ERROR, "BINSTORE: Cannot correct length of data file "
                            "(\"%s\"): %m", data_fname );
              return -1; 
        }

    if (index_read < index_file_len)
        if ( ftruncate(index_fd, index_read) )
        {
              log( L_ERROR, "BINSTORE: Cannot correct length of index file "
                            "(\"%s\"): %m", index_fname );
              return -1; 
        }

    /* set the file cursors to end of both index and data file */
    if ( lseek (index_fd, 0, SEEK_END ) != index_read )
    {
          log( L_ERROR, "BINSTORE: Cannot seek to end of index file "
                        "(\"%s\") to %ld: %m", index_fname, 
                        static_cast<unsigned long>( index_read ) );
          return -1;
    }

    if ( lseek (data_fd, 0, SEEK_END ) !=data_read )
    {
          log( L_ERROR, "BINSTORE: Cannot seek to end of data file "
                        "(\"%s\") to %ld: %m", data_fname, 
                        static_cast<unsigned long>( data_read ));
          return -1;
    }

    /* ensure it is at least in disc caches */
    fsync(data_fd);
    fsync(index_fd);

    UNLOCK_FILE (index_fd);
    UNLOCK_FILE (data_fd);

    /* set the tail marker and mark that files are assigned */ 
    present = 1;
    tail = data_read;

    return 0;
}


/* opens BinStore for reading */
int tBinStore::open_r(char *data_fname, char *index_fname)
{
    off_t index_file_len;

    ASSERT(data_fname);
    ASSERT(index_fname);

    /* file will be opened for write access */
    access_mode = 1;
    
    /* first backup filenames */            
    if(!(data_filename = strdup (data_fname)))
    {
         log (L_ERROR, "BINSTORE: Can't store binstore datafile name" 
                 " (\"%s\"): %m", data_fname );
         return -1;
    }

    if(!(index_filename = strdup (index_fname)))
    {
         log (L_ERROR, "BINSTORE: Can't store binstore indexfile name"
                 " (\"%s\"): %m", index_fname );
         return -1;
    }

    /* open files */
    data_fd = s_open (data_fname, O_RDONLY, 0400);
    if (data_fd < 0) 
    {
        log (L_ERROR, "BINSTORE: Can't open data file for reading "
                "(\"%s\"): %m", data_fname );
        return -1;
    }

    index_fd = s_open (index_fname, O_RDONLY, 0400);
    if ( index_fd < 0 )
    {
       log( L_ERROR, "BINSTORE: Can't open index file for reading "
               "(\"%s\"): %m", index_fname );
       return -1;
    }

    /* determine index file length */
    index_file_len = s_filelen (index_fd); 

    /* set cursor and item count to zero */
    last_id = index_file_len / (sizeof (tBinItemOffset)) ;
    present = 1;

    return 0;
}

/* closes the Binstore, making sure all data are written */
int tBinStore::close()
{
    /* proceed only if files assigned */
    if (!present) return 0;

    /* close all files */
    if (s_close(data_fd) || s_close(index_fd) )
    {
        log( L_ERROR, "BINSTORE: Cannot close binstore files: %m" );
        return -1;
    } 

    xfree (data_filename);
    xfree (index_filename);

    present = 0;

    return 0;
}


/* adds a item to the end of the BinStore object */ 
tBinIdentificator tBinStore::add_item(void *buffer, int length)
{
   struct tBinItem item_desc;

   /* check if files are assigned and opened in write-only mode */
   if (!present || access_mode != 2) 
   {
      log (L_ERROR, "BINSTORE: Can't insert item into non-writeable binstore.");
      return 0;
   }

   WRITELOCK_FILE (data_fd);
   WRITELOCK_FILE (index_fd);

   /* fill in item descriptor */
   item_desc.magic = C_MAGIC;
   item_desc.id = last_id + 1;
   item_desc.length = length; 

   /* check if we can still insert the item */
   if (last_id + 1 < last_id)
   {
       log (L_ERROR, "BINSTORE: Limit on number inserted items reached.");
       return 0;
   }
   
   /* check if we can still insert the item */
   if ( tail + static_cast<off_t>(sizeof(struct tBinItem)) + 
           static_cast<off_t>(length) < tail ) 
   {
       log (L_ERROR, "BINSTORE: Limit on size of data file exceeded.");
       return 0;
   }

   /* write offset of it in index file */
   if (s_write (index_fd, &tail, sizeof(tBinItemOffset))!=
            sizeof(tBinItemOffset) )
   {
       log (L_ERROR, "BINSTORE: Cannot write into index file (\"%s\"): %m",
            index_filename );
       return 0;
   } 

   /* write data descriptor into data-file */
   if (s_write (data_fd, &item_desc, sizeof(struct tBinItem)) !=
            sizeof(struct tBinItem) )
   {
       log (L_ERROR, "BINSTORE: Cannot write into data file (\"%s\"): %m",
            data_filename);
       return 0;
   }

   /* and finally write all data */
   if (s_write (data_fd, buffer, length) != length)
   {
       log( L_ERROR, "BINSTORE: Cannot write into data file (\"%s\"): %m",
            data_filename );
       return 0;
   }

   /* update the binstore variables */
   last_id ++;
   tail += sizeof (struct tBinItem) + length;

   /* ensure it is at least in disc caches */
   if (write_through)
   {
      fsync (data_fd);
      fsync (index_fd);
   }

   UNLOCK_FILE (index_fd);
   UNLOCK_FILE (data_fd);

   return last_id;
}


/* Fills the buffer with data of requested item */
int tBinStore::get_item(tBinIdentificator item_id, void *buffer, int *buf_size)
{
   struct tBinItem item_desc;
   tBinItemOffset offset;

   /* check if files are associated and opened in read-only way */
   if (!present || access_mode != 1) 
   {
       log (L_ERROR, "BINSTORE: Cannot get item from readonly binstore.");
       return -1;
   }

   /* check if we have this item stored */
   if (item_id > last_id || item_id <= 0)
   {
       log (L_ERROR, "BINSTORE: Item not in Binstore.");
       return -1;
   }

   READLOCK_FILE (data_fd);
   READLOCK_FILE (index_fd);
    
   /* seek to find offset in index file */
   off_t index_ofs = (item_id-1) * sizeof (tBinIdentificator);
   if (lseek (index_fd, index_ofs, SEEK_SET) != index_ofs)
   {
       log (L_ERROR, "BINSTORE: Cannot seek in the index file (\"%s\") "
            "to %ld: %m", index_filename, index_ofs); 
       return -1;
   }

   /* read the offset */
   if (s_read (index_fd, &offset, sizeof (tBinItemOffset))!=
            sizeof (tBinItemOffset))
   {
       log (L_ERROR, "BINSTORE: Cannot read from index file (\"%s\"): %m",
            index_filename );
       return -1;
   } 

   /* seek to this offset in data file */
   if (lseek (data_fd, offset, SEEK_SET) != offset)
   {
       log (L_ERROR, "BINSTORE: Cannot seek in the data file (\"%s\"): %m",
            data_filename );
       return -1;
   }

   /* read the data descriptor */
   if (s_read (data_fd, &item_desc, sizeof (struct tBinItem)) != 
             sizeof (struct tBinItem))
   {
       log (L_ERROR, "BINSTORE: Cannot read from data file (\"%s\"): %m",
            data_filename );
       return -1;
   }

   /* check if the structure is ok */
   if (item_desc.magic != C_MAGIC || item_desc.id != item_id)
   {
       log (L_ERROR, "BINSTORE: Data or index file corrupted.");
       return -1;
   }

   /* read the item data to buffer */
   if (s_read (data_fd, buffer, min (*buf_size, item_desc.length)) != 
             min (*buf_size, item_desc.length))
   {
       log( L_ERROR, "BINSTORE: Cannot read from data file (\"%s\"): %m",
            data_filename );
       return -1;
   }
  
   /* store the item size */
   *buf_size = item_desc.length;
   
   UNLOCK_FILE (index_fd);
   UNLOCK_FILE (data_fd);
   
   return 0;

}


/* Get the item-id of last inserted item */
tBinIdentificator tBinStore::get_last_id()
{
   return last_id;
}

}
// namespace spool
