/**
 * @brief Echelog Spool Library -- Pagefile implementation
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */
#include "lib/lib.h"
#include "spool/pagefile.h" 
 
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>

namespace spool
{

using namespace lib;

/* buffer full of zeroes */
static char zero_filler[4096];
/* magic constant */
static unsigned long C_MAGIC;

/* Initializes the pagefile structure */
tPageFile::tPageFile(int wr_thr, int max_page_size, int max_items)
{
    /* set passed parameters */  
    write_through = wr_thr;
    pg_max_page_size = max_page_size;
    pg_max_items = max_items;
    
    /* denote that pagefile is not mapped into memory */
    present = 0;

    /* in the beginning is pagefile empty */
    pg_items = pg_nonextracted = 0;

    /* invalidate pointers into mapped pagefile */ 
    pg_data = pg_cursor = pg_limit = NULL;

    /* index of offsets to inserted items */
    pg_index = static_cast<int *>
                 ( xmalloc( (pg_max_items + 1)*sizeof(int) ) ); 

    /* obtain magic constant */
    C_MAGIC = htonl( 0xCA453E76 );

    /* pagefile is not closed for inserting items yet */
    closed = 0;

}


/* creates new pagefile and mmaps it into memory */
int tPageFile::create_pagefile(char *afilename, tIdentificator first)
{
    int written, block;

    ASSERT(afilename);
    
    /* first backup filename */            
    if(!(filename = strdup( afilename )) )
    {
         log( L_ERROR, "PAGEFILE: Can't store pagefile name (\"%s\"): %m",
                 filename );
         return -1;
    }

    /* create the pagefile */
    fd = s_open (filename, O_RDWR | O_CREAT | O_TRUNC , 0600);
    if ( fd < 0 ) 
    {
        log( L_ERROR, "PAGEFILE: Cannot create pagefile file (\"%s\"): %m",
                filename );
        return -1;
    }
   
    /* blank the pagefile with zeroes */ 
    for (written=0;written<pg_max_page_size;)
    {
        block = min ( pg_max_page_size - written, 4096 ); 
        if ( s_write( fd, zero_filler, block ) <  block)
        {
           log(L_ERROR, "PAGEFILE: Cannot write into pagefile (\"%s\"): %m", 
                   filename);
           return -1;
        }
        written += block;
    }

    /* ensure it is at least in disc caches */
    fsync(fd);

    /* perform the mmap of the pagefile */
    pg_data = static_cast<char *>
       (mmap( NULL, pg_max_page_size, PROT_READ | PROT_WRITE , MAP_SHARED,
              fd, 0));

    /* there were error mmaping the pagefile */
    if ((signed long)pg_data == -1)
    {
        log(L_ERROR, "PAGEFILE: Cannot map pagefile (\"%s\") into memory: %m",
                filename );
        return -1;
    }

    /* setting the parametrs, number of inserted and non-extracted items to 0 */
    pg_limit = pg_data + pg_max_page_size;
    pg_index[0] = pg_nonextracted = pg_items = 0;
    first_id = last_id = first_nonextracted_id = first;

    /* we have the pagefile loaded into memory, and set it not closed for
     * inserting  */
    present = 1;
    closed = 0;

    return 0;
}


/* loads existing pagefile into memory */
int tPageFile::assay_pagefile(char *afilename, tIdentificator first)
{
    struct stat stat_buf; 
    struct tItem c_item;
    tIdentificator current;
    int index;
    
    ASSERT(afilename);
    
    /* first backup filename */            
    if(!(filename = strdup( afilename )) )
    {
       log( L_ERROR, "PAGEFILE: Cannot store pagefile name (\"%s\"): %m",
               filename );
       return -1;
    }

    /*  open the pagefile */
    fd = s_open (filename, O_RDWR, 0600 );
    if ( fd < 0 ) 
    {
        log( L_ERROR, "PAGEFILE: Cannot open pagefile (\"%s\"): %m",
                filename );
        return -1;
    }

    /* determine pagefile length */
    if( fstat( fd, &stat_buf ) )
    {
        log( L_ERROR, "PAGEFILE: Cannot stat pagefile (\"%s\"): %m",
                filename );
        return -1;
    }

    if ( stat_buf.st_size > pg_max_page_size )
    {
        log( L_ERROR, "PAGEFILE: Pagefile length exceeds limit (\"%s\"): %m",
                filename );
        return -1;
    }
    
    /* perform mmap of the pagefile */
    pg_data = static_cast<char *> 
       (mmap( NULL, pg_max_page_size, PROT_READ | PROT_WRITE , MAP_SHARED,
              fd, 0));

    /* check if mmap succeded */
    if ((signed long)pg_data == -1)
    {
        log(L_ERROR, "PAGEFILE: Cannot map pagefile (\"%s\") into memory: %m",
                filename );
        return -1;
    }

    /* setting the parametrs, number of inserted and non-extracted items to 0 */
    pg_limit = pg_data + pg_max_page_size;
    index = pg_index[0] = pg_nonextracted = pg_items = 0;

    /* we will now traverse thru pagefile and check it if it is not 
     * corrupted and store into pg_index array offsets of items beginning */
    pg_cursor = pg_data;
    current = last_id = first_id = first;
    first_nonextracted_id = 0;
   
    /* traverse until item count limit or space limit reached 
     * -- this shouldn't happen unless file is corrupted */
    while ( pg_cursor + sizeof( struct tItem ) < pg_limit &&
            index < pg_max_items )
    {
       /* lets take new item */
       memcpy ( &c_item, pg_cursor, sizeof(struct tItem));

       /* item is invalid -- magic number is wrong */
       if (c_item.magic != C_MAGIC)
           break;

       /* next item does not match requested id - item id must be consecutive*/ 
       if (c_item.id != current)
       {
           /* it was first record - means no error - but we will take
            * care of skips inside interval */
           if ( current == first )
           {
               current = last_id = first_id = c_item.id;
           } else break;
       } 

       /* there can't be any legal item after missing item */
       if (c_item.status == MISSING ) break;

       /* look for first un extracted */
       if (c_item.status == PRESENT && first_nonextracted_id == 0)
           first_nonextracted_id = c_item.id;

       /* check if maximal number of items is reached */
       if ( index >= pg_max_items ) 
       {
           log ( L_ERROR, "PAGEFILE: Limit on maximal item number exceeded "
                   "in (\"%s\").", filename );
           break;
       }
     
       /* great - item has passed all checks and we accept him */
       pg_items ++; last_id ++; current ++; index ++;
       pg_cursor += sizeof( struct tItem ) + c_item.length; 
       pg_index[index] =  pg_cursor - pg_data;
       if ( c_item.status == PRESENT ) pg_nonextracted++;
           
    }
  
   /* we have the pagefile loaded into memory, and set it not closed for
    * inserting  */
    if (first_nonextracted_id == 0) 
        first_nonextracted_id = last_id;
    closed = 0; 
    present = 1; 
    return 0;
}


/* mmaps pagefile into memory assuming that the rest of
 * informations is already present in the structure */
int tPageFile::attach_pagefile()
{
    
    /* if is already mmaped do nothong */
    if (present) return 0;

    /* reopen the pagefile */
    fd = s_open (filename, O_RDWR, 0600 );
    if ( fd < 0 )
    {
        log( L_ERROR, "PAGEFILE: Cannot open pagefile (\"%s\"): %m",
                filename );
        return -1;
    }

    /* mmap pagefile */
    pg_data = static_cast<char *>
       (mmap( NULL, pg_max_page_size, PROT_READ | PROT_WRITE , MAP_SHARED,
              fd, 0));

    /* check for possible errors */
    if ((signed long)pg_data == -1)
    {
        log(L_ERROR, "PAGEFILE: Cannot map pagefile (\"%s\") into memory: %m",
              filename );
        return -1;
    } 
   
    /* now the pagefile is present in memory */
    present = 1;

    return 0;
}


/* unmaps the pagefile, saves it, frees all internal structures */
int tPageFile::close_pagefile()
{
    int err;

    /* unmap pagefile */
    err = detach_pagefile();

    /* free resources */
    s_close( fd );
    free( pg_index );
    present = 0;

    return err;
}


/* completly removes the file */
int tPageFile::destroy_pagefile()
{

    /* we cannot destroy currently mmaped pagefile */
    if (present)
    {
        log(L_ERROR, "PAGEFILE: Cannot destroy unmapped pagefile." );
        return -1;
    }

    /* try to remove the file or error */
    if (unlink(filename))
    {
        log(L_ERROR, "PAGEFILE: Cannot unlink pagefile (\"%s\"): %m",
                filename );
        return -1;
    }
    return 0;
}


/* unmaps the pagefile, but leaves other internal structures in
 * the memory - ready for fetch_pagefile() call */
int tPageFile::detach_pagefile ()
{
    /* if pagefile is not mmaped do nothing */
    if (!present) return 0;

    /* otherwise do the unmap */
    if (munmap (pg_data, pg_max_page_size))
    {
        log (L_ERROR, "PAGEFILE: Cannot unmap pagefile (\"%s\"): %m",
                filename );
        return -1;
    }

    if (s_close (fd))
    {
        log (L_ERROR, "PAGEFILE: Error closing pagefile (\"%s\"): %m",
                filename );
        return -1;
    }

    /* invalidate the pointers and mark that the pagefile is not present*/
    pg_data = pg_cursor = pg_limit = NULL;
    present = 0;

    return 0;
}


/* adds a item to the end of the queue - it must be ensured
 * that item will fit in pagefile - check is_to_change(int need)
 * before with need bytes */
tIdentificator tPageFile::insert_item(char *buffer, int length)
{
    struct tItem item;

    /* fill in item fields */
    item.magic = C_MAGIC;
    item.status = PRESENT;
    item.id = last_id;
    item.length = length;

    /* insert it into mmaped pagefile - no checks are done because
     * it must be ensured before passing data to this function
     * this function should never be called directly - use upper
     * layer - class tQueue insted of direct access thru tPagefile*/
    memcpy( pg_data + pg_index[last_id-first_id], &item, sizeof(struct tItem));
    memcpy( pg_data + pg_index[last_id-first_id] + sizeof(struct tItem),
            buffer, length );
    
    /* add the item offset to the index of items */
    pg_index[last_id-first_id+1] = pg_index[last_id - first_id] +
            length + sizeof(struct tItem);
    pg_items ++; pg_nonextracted ++; 

    /* synchro checks */
    synchro_pagefile();
    
    /* return id of the inserted item */
    return last_id ++;
}


/* marks item to extract if the whole segment is to extract
 * then the segment is removed */
void tPageFile::extract_item(tIdentificator item_id)
{
    struct tItem c_item;
    enum ITEMSTATUS status;

    /* checks for current status - removes only PRESENT items */
    status = get_status ( item_id );
    if ( status == EXTRACTED || status == MISSING )
        return;

    /* mark it as extracted */
    set_status( item_id, EXTRACTED );
    pg_nonextracted--;

    first_nonextracted_id++;
    for(;first_nonextracted_id<last_id;first_nonextracted_id++)
         if (get_status(first_nonextracted_id)==PRESENT) break;
                 
    /* if all items are "extracted" in pagefile then empty it */
    if (pg_nonextracted==0 && closed == 0) 
    {
        /* empty pagefile */
        first_nonextracted_id = first_id = last_id;
        pg_index[0] = 0; pg_items = 0;

        /* change the status of first item */
        memcpy ( &c_item, pg_data, sizeof( struct tItem ));
        c_item.status = MISSING; 
        c_item.id = first_nonextracted_id;
        memcpy ( pg_data, &c_item, sizeof( struct tItem ));

        /* synchro checks */
        synchro_pagefile();

    }

}


/* changes the status field in item described by item_id */
void tPageFile::set_status(tIdentificator item_id, enum ITEMSTATUS new_st)
{
    struct tItem c_item;
   
    /* check if the id is from range of this pagefile */
    if ( item_id < first_id ) return;
    if ( item_id >= last_id ) return;
   
    /* obtain the item and change its status */
    memcpy ( &c_item, pg_data + pg_index[item_id - first_id], 
            sizeof( struct tItem ));
    c_item.status = new_st;
    memcpy ( pg_data + pg_index[item_id - first_id], &c_item, 
            sizeof( struct tItem ));

    /* synchro checks */
    synchro_pagefile();
    
}


/* returns the status field in item described by item_id */
enum ITEMSTATUS tPageFile::get_status(tIdentificator item_id)
{
    struct tItem c_item;

    /* check if the id is from range of this pagefile */
    if ( item_id < first_id ) return MISSING;
    if ( item_id >= last_id ) return MISSING;

    /* obtain the item and return its address */
    memcpy ( &c_item, pg_data + pg_index[item_id - first_id], 
            sizeof( struct tItem ));
    return c_item.status; 
}

}
// namespace spool
