/**
 * @brief Echelog Spool Library -- Queue implementation
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include "lib/archdep.h"

#include <stdio.h>
#include <errno.h>

#include "lib/lib.h"
#include "spool/pagefile.h"
#include "spool/queue.h"

namespace spool
{

using namespace lib;

/* buffer full of zeroes */
static char zero_filler[4096];

tQueue::tQueue(int wr_thr, char *sp_dir, char *fname)
{
    int to_write, written, block;

    ASSERT(sp_dir);
    ASSERT(fname);

    /* assume the operation failed until end reached */
    present = 0;
    
    /* first store spooldir and index filename */
    if(!(filename = strdup( fname )) )
    {
         log( L_ERROR, "QUEUE: Can't store index filename (\"%s\"): %m",
                 fname );
         return;
    }
        
    if(!(spool_dir = strdup( sp_dir )) )
    {
         log( L_ERROR, "QUEUE: Can't store spooldir name (\"%s\"): %m",
                 sp_dir );
         return;
    }

    /* set writethrough parameter as passed */
    write_through = wr_thr;

    /* we have no pages at the beginning */
    pages = 0;

    /* open/create the index of pagefiles */
    fd = s_open (filename, O_RDWR | O_CREAT | O_EXCL , 0600);
    if ( fd < 0 && errno != EEXIST )
    {
        log( L_ERROR, "QUEUE: Cannot open index of pagefiles (\"%s\"): %m",
                filename );
        return;
    }

    to_write = MAX_PAGES * sizeof( tIdentificator );

    /* create of file failed - file can exist or it really failed */
    if ( fd < 0 ) // errno must be EEXIST - it means that file already exists
    {
        fd = s_open (filename, O_RDWR , 0600);
        if ( fd < 0 && errno != EEXIST )
        {
            log( L_ERROR, "QUEUE: Cannot create index of pagefiles (\"%s\"):"
	            " %m", filename );
            return;
        }

        /* read from index of pagefiles how many pagefiles we have */
        if ( s_read( fd, &pages, sizeof( tIdentificator ) ) <  
                static_cast<int>(sizeof( tIdentificator )))
        {
            log(L_ERROR, "QUEUE: Cannot read from index of pagefiles (\"%s\"):"
	            " %m", filename );
            return;
        }

        /* and prepare it for zeroing of the rest of index file */
        to_write -= ( pages + 1) * sizeof( tIdentificator );
        lseek ( fd, 0, SEEK_END );
    }

    /* blank it with zeroes */
    for (written=0;written<to_write;)
    {
        block = min ( to_write - written, 4096 );
        if ( s_write( fd, zero_filler, block ) <  block)
        {
           log(L_ERROR, "QUEUE: Cannot write into index of pagefiles (\"%s\"):"
	           " %m", filename );
           return;
        }
        written += block;
    }

    /* ensure it is at least in disc caches */
    fsync(fd);

    /* perform the mmap of index file */
    idx = static_cast<tIdentificator *>
       (mmap( NULL, MAX_PAGES * sizeof( tIdentificator ), 
              PROT_READ | PROT_WRITE , MAP_SHARED, fd, 0));

    /* alloc the array of pointers to pagefiles and zero it */
    pg_file = static_cast<class tPageFile **>
       (xmalloc( MAX_PAGES * sizeof(class tPageFile *) ) );
    memset( pg_file, 0, MAX_PAGES * sizeof(class tPageFile *) );

    /* check for errors */
    if ((signed long)idx == -1)
    {
        log(L_ERROR, "QUEUE: Cannot map index of pagefiles (\"%s\")"
	             " into memory: %m", filename );
        present = 0;
        return;
    }

    /* everything ok, index of pagefiles loaded */
    present = 1;
}


tQueue::~tQueue()
{
    /* if index of pagefiles still present, then unmap it */
    if (present) 
        close_queue();
}


int tQueue::detach_queue()
{
    int i,err;

    /* if index of pagefiles not present, we cannot detach it */
    if (!present)
    {
        log(L_ERROR, "QUEUE: Cannot detach queue of non-existing object.");
        return -1;
    }
    
    /* for each pagefile in queue we will free the structure */
    err = 0;
    for (i=1;i<=static_cast<int>(pages);i++)
        if (pg_file[i])
        {
            err = min( err, pg_file[i] -> close_pagefile() );
            delete pg_file[i];
        }
  
    /* and null the pointers to pagefiles */
    memset( pg_file, 0, MAX_PAGES * sizeof(class tPageFile *) );

    return err;
}


int tQueue::attach_queue()
{
    char filenamebuf[1000];
    int i;

    /* we operate only on existing objects */ 
    if (!present)
    {
        log(L_ERROR, "QUEUE: Cannot attach queue to non-existing object.");
        return -1;
    }

    /* for each pagefile stored */ 
    for(i=1;i<=static_cast<int>(pages);i++)
    {

        /* create new pagefile structure */
        pg_file[i] =  
            new tPageFile( write_through, MAX_PAGE_SIZE, MAX_ITEMS );
        if (! ( pg_file[i] ) )
        {
            log(L_ERROR, "QUEUE: Out of memory.");
            return -1;
        }
        
        /* load the proper pagefile in it */
        snprintf( filenamebuf, sizeof(filenamebuf), "%s/%ld.pag", 
                  spool_dir, idx[i] );
        if ( pg_file[i] -> assay_pagefile( filenamebuf, idx[i] ) )
        { 
            log(L_ERROR, "QUEUE: Cannot load pagefile (\"%s\").", filenamebuf);
            return -1;
        }

        /* for memory savings we hold only head and tail pagefile in memory*/
        if (i>2)
            if (pg_file[i-1] -> detach_pagefile()) return -1;
        if (i>1)
            pg_file[i-1] -> set_closed();

        /* check if end and beginning match */
        if (i>1)
          if ( pg_file[i-1] -> get_last_id() != idx[i] )
          {
              log(L_ERROR, "QUEUE: Index file of pagefiles corrupted or "
                      "overlap in pagefiles.");  
              return -1;
          }
    } 

    return 0;
}


/* all items with id < limit remove */
int tQueue::strip_queue(tIdentificator limit)
{
    int i,n,err;
    char filenamebuf[1000];

    /* if index of pagefiles not present, we cannot operate on it */
    if (!present)
    {
        log(L_ERROR, "QUEUE: Cannot free pg-chain of non-existing object.");
        return -1;
    }

    /* each pagefile which holds items with ids lesser than limit remove */
    err = 0;
    for(i=1;i<=static_cast<int>(pages);i++)
        if ( limit > pg_file[i] -> get_last_id()-1 )
        {
            err = min( err, pg_file[i] -> close_pagefile() );
            err = min( err, pg_file[i] -> destroy_pagefile() ); 
            delete pg_file[i];
        } else break;

    /* there were error */
    if (err)
    {
        log(L_ERROR, "QUEUE: Error occured while freeing queue-chain.");
        return -1;
    }   
 
    /* and shift the rest of pagefiles to the beginning */
    n = i;
    for(i=n;i<=static_cast<int>(pages);i++)
    {
        pg_file[i-n+1] = pg_file[i];
        idx[i-n+1] = idx[i]; 
    }

    /* decrease the number of pagefiles */
    pages = idx[0] = pages - n + 1;
    
    /* all pagefiles were removed - we must have at least one in queue */
    if ( pages == 0 )
    {
      
        /* create new pagefile structure */
        pg_file[1]=new tPageFile( write_through, MAX_PAGE_SIZE, MAX_ITEMS );
        if (! ( pg_file[1] ) )
        {
            log(L_ERROR, "QUEUE: Out of memory.");
            return -1;
        }

        /* create pagefile in it */
        snprintf( filenamebuf, sizeof(filenamebuf),
                 "%s/%ld.pag", spool_dir, limit );
        if ( pg_file[1] -> create_pagefile( filenamebuf, limit ) )
        {
            log(L_ERROR, "QUEUE: Cannot create pagefile (\"%s\").",filenamebuf);
            return -1;
        }
 
        /* fill in the index stuctures */
        pages = idx[0] = 1;
        idx[1] = limit ;

    }
    
    synchro_idx_file();

    /* first pagefile should be always mapped */
    if(pg_file[1] -> attach_pagefile())
    {
        log(L_ERROR, "QUEUE: Cannot attach first pagefile.");
        return -1;
    }

    /* extract all preceeding items */
    for(tIdentificator k=pg_file[1]->get_first_id();k<limit;k++)
        pg_file[1] -> extract_item(k); 

    return 0;
} 


int tQueue::open_queue(tIdentificator limit)
{

    /* we operate only on existing objects */
    if (!present)
    {
        log(L_ERROR, "QUEUE: Cannot attach queue to non-existing object.");
        return -1;
    }

    /* item id start at 1 */
    if (limit == 0) return -1;

    /* free old queue */
    if (detach_queue())
    {
        log(L_ERROR, "QUEUE: Error while detaching queue.");
        return -1;
    }
    
    /* attach new according to pagefile */
    if(attach_queue())
    {
        log(L_ERROR, "QUEUE: Error while attaching new queue.");
        return -1;
    }

    /* remove pagefiles holding only items with id < limit */
    if(strip_queue(limit))
    {
        log(L_ERROR, "QUEUE: Error while removing obsolete segment.");
        return -1;
    }

    return 0;
}


int tQueue::close_queue()
{

    /* we operate only on existing objects */
    if (!present)
    {
        return 0;
    }
  
    /* free the chain of pagefiles */
    if (detach_queue())
    {
        log(L_ERROR, "QUEUE: Error while detaching queue from memory.");
        return -1;
    }

    /* do the synchro checks */
    synchro_idx_file();

    /* unmap the index file */
    if ( munmap ( idx , MAX_PAGES * sizeof( tIdentificator ) ) )
    {
        log(L_ERROR, "QUEUE: Cannot unmap index file (\"%s\"): %m", filename );
        return -1;
    }

    idx = NULL;

    present = 0;
    
    return 0;
}


int tQueue::destroy_queue()
{
    int err,i;

    /* we operate only on existing objects */
    if (!present)
    {
        log(L_ERROR, "QUEUE: Cannot destroy non-existing object.");
        return -1;
    }

    err = 0;
    for(i=1;i<=static_cast<int>(pages);i++)
    {
        err = min( err, pg_file[i] -> close_pagefile() );
        err = min( err, pg_file[i] -> destroy_pagefile() );
        delete pg_file[i];
    } 

    /* there were error */
    if (err)
    {
        log(L_ERROR, "QUEUE: Error occured while destroying queue-chain.");
        return -1;
    }

    /* unmap the index file */
    if ( munmap ( idx , MAX_PAGES * sizeof( tIdentificator ) ) )
    {
        log(L_ERROR, "QUEUE: Cannot unmap index file (\"%s\"): %m", filename );
        return -1;
    }

    idx = NULL;

    /* try to remove the file or error */
    if (unlink(filename))
    {
        log(L_ERROR, "QUEUE: Cannot unlink queue index file (\"%s\"): %m",
                filename );
        return -1;
    }

    present = 0;
    return 0;
}


tIdentificator tQueue::insert_item(char *buffer, int length)
{
    char filenamebuf[1000];
    tIdentificator new_id;

    /* check if we operate on existing instance */
    if (!present)
    {
        log(L_ERROR, "QUEUE: Cannot insert item to non-existing queue.");
        return 0;
    } 

    /* if it does not fit in current pagefile -> create new */ 
    if ( pg_file[pages] -> is_space( length ) )
    {
        if ((new_id = pg_file[pages] -> insert_item( buffer, length )))
        return new_id;
    } else
    {
        /* creating new pagefile - check if limit on number 
         * of pagefiles reached */
        if ( pages >= MAX_PAGES )
        {
            log(L_ERROR, "QUEUE: Limit on number of pagefiles exceeded.");
            return 0;
        }

        /* close old and release it from memory */
        pg_file[pages] -> set_closed();
        if ( pages > 1 ) 
          if(pg_file[pages] -> detach_pagefile())
          {
             log(L_ERROR, "QUEUE: Error while detach previous tail (insert).");
             return 0;
          }

        /* create new pagefile structure */
        pg_file[pages + 1] =
            new tPageFile( write_through, MAX_PAGE_SIZE, MAX_ITEMS );
        if (! ( pg_file[pages + 1] ) )
        {
            log(L_ERROR, "QUEUE: Out of memory.");
            return 0;
        }
 
        /* fill in the new info */
        idx[pages+1] = pg_file[pages] -> get_last_id ();

        /* load the proper pagefile in it */
        snprintf( filenamebuf, sizeof(filenamebuf), 
                  "%s/%ld.pag", spool_dir, idx[pages+1]);
        if ( pg_file[pages+1] -> create_pagefile( filenamebuf, idx[pages+1] ) )
        {
            log(L_ERROR, "QUEUE: Cannot load pagefile (\"%s\").", filenamebuf);
            return 0;
        }

        /* update index file */
        pages++; idx[0] = pages;
        synchro_idx_file();

        /* insert item in it  */
        if ( pg_file[pages] -> is_space( length ) )
        {
            if ((new_id = pg_file[pages] -> insert_item( buffer, length ))) 
            return new_id;
        } else
        {
            log(L_ERROR, "QUEUE: Item too large.");
            return 0;
        }
        
    }

    return 0;
}


void tQueue::extract_item(tIdentificator item_id)
{

    /* check if we operate on existing instance */
    if (!present) 
    {
        log(L_ERROR, "QUEUE: Cannot extract item (%ld) from "
             "non-existing queue.", item_id);
        return;
    }

    /* if the item is in the first pagefile in chain then extract it and
       possibly if the first pagefile is empty - remove it from the queue */
    if ( pg_file[1] -> get_item( item_id ) )
    {
        pg_file[1] -> extract_item( item_id );
        if ( pg_file[1] -> to_remove() )
        {
            if(strip_queue( pg_file[1] -> get_last_id() ))
            {
                log(L_ERROR, "QUEUE: Error while destr prev. file (extract)"); 
            }
        }
    } else
    {
        log(L_ERROR, "QUEUE: Cannot extract item (%ld)from non-first pagefile.",
               item_id);
        return;
    } 
}

} // namespace spool
