/**
 * @file
 * @brief Contains implementation of tAggregatorTarget class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>, Kamil Toman <ktoman@email.cz>
 *
 */

#include "aggregatortarget.h"
#include "messages/msgbuf.h"
#include "messages/msgmgr.h"
#include <limits.h>
#include <stdlib.h>

using namespace lib;
using namespace messages;
using namespace mframes;


/* Perfect hash */
static inline long conv_id (char *id)
{
    int i, r;

    for (r = 0, i = 0; i < ID_LENGTH; i++)
        r = id[i] + (r << 8);

    xfree (id);
    return r;
}


/* Basic logstor initialization */
tAggregatorTarget::tAggregatorTarget ()
{
    m_aggregates = NULL;
    m_snapshots = NULL;
    m_aggregate_count = 0;
    m_output = NULL;

    m_aggregator_section = new cfitem [3];
    if (m_aggregator_section == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        m_agstate = AG_BROKEN;
        return;
    }

    int i = 0;
    m_aggregator_section[i].name = "Aggregator";
    m_aggregator_section[i].type = CT_SECTION;
    m_aggregator_section[i].data = NULL;
    i++;
    m_aggregator_section[i].name = "Aggregate";
    m_aggregator_section[i].type = CT_LIST;
    m_aggregator_section[i].data = NULL;
    i++;
    m_aggregator_section[i].name = NULL;
    m_aggregator_section[i].type = CT_STOP;
    m_aggregator_section[i].data = NULL;

    m_agstate = AG_CREATED;
}


/* Final cleanup */
tAggregatorTarget::~tAggregatorTarget ()
{
    ASSERT (m_agstate == AG_CREATED ||
            m_agstate == AG_BROKEN);

    if (m_agstate == AG_CREATED)
        delete [] m_aggregator_section;
}


/* Initialization of the filter */
bool tAggregatorTarget::init ()
{
    ASSERT (m_agstate == AG_CONFIGURED);

    m_output = tOFrame::get_actual_oframe ()->
                get_output_by_name (NULL);
    if (m_output == NULL)
    {
        log (L_ERROR, "Missing output configuration.");
        return true;
    }

    m_agstate = AG_INITIALIZED;

    return false;
}


/* Reverse method for init */
void tAggregatorTarget::cleanup ()
{
    ASSERT (m_agstate == AG_INITIALIZED);

    m_output = NULL;

    m_agstate = AG_CLEANED;
}


bool tAggregatorTarget::register_sections (lib::tMemPool& trash_pool)
{
    /* ensure the buffer is allocated */
    if (m_agstate == AG_BROKEN)
        return true;

    tList* list = new tList (trash_pool);
    if (list == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        return true;
    }
    m_aggregator_section[1].data = list;

    cf_register (m_aggregator_section);

    return false;
}


void tAggregatorTarget::unregister_sections ()
{
    cf_unregister (m_aggregator_section[0].name);

    delete static_cast<lib::tList *> (m_aggregator_section[1].data);
    m_aggregator_section[1].data = NULL;
}


bool tAggregatorTarget::process_config (lib::tMemPool& conf_mp,
        lib::tMemPool& trash_pool)
{
    ASSERT (m_agstate == AG_CREATED);

    tNode const* n = NULL;
    long acount = 0;

    /* add catch rules */
    tList* list = static_cast<tList*> (m_aggregator_section[1].data);
    m_aggregate_count = 0;
    acount = list->count ();
    if (acount > 0)
    {
        /* Allocate space for snapshots */
        m_snapshots = static_cast<tSnapshotList **>
             (conf_mp.alloc (sizeof (tSnapshotList *) * acount));
        memset (m_snapshots, 0, sizeof (tSnapshotList *) * acount);

        /* Allocate space for configured clauses */
        m_aggregates = static_cast<struct tAggregate*>
             (conf_mp.alloc (sizeof (struct tAggregate) * acount));
        memset (m_aggregates, 0, sizeof (struct tAggregate) * acount);

        /* split each rule into separate strings */
        FOR_ALL_LIST_ITEMS (n, list)
        {
            /* parse one line */
            if (parse_config_line (static_cast<char *> (n->data), 
                        conf_mp, trash_pool))
                goto aggregates_clean;
        }
    }

    m_agstate = AG_CONFIGURED;

    return false;

aggregates_clean:
    /* release all data*/
    for (int i = 0; i < m_aggregate_count; i++)
        delete m_snapshots[i];
    m_snapshots = NULL;
    m_aggregates = NULL;
    m_aggregate_count = 0;

    return true;
}


void tAggregatorTarget::cleanup_config ()
{
    ASSERT (m_agstate == AG_CLEANED ||
            m_agstate == AG_CONFIGURED);

    /* release all data*/
    for (int i = 0; i < m_aggregate_count; i++)
        delete m_snapshots[i];
    m_snapshots = NULL;
    m_aggregates = NULL;
    m_aggregate_count = 0;

    m_agstate = AG_CREATED;
}


/* integer comparator */
static int int_compator (const void *a, const void *b)
{
    return *static_cast<const int *> (a) - *static_cast<const int *> (b);
}


bool tAggregatorTarget::parse_config_line (char* line, tMemPool& conf_mp, 
        tMemPool& trash_pool)
{
    tList paramlist (trash_pool), ignore_list (trash_pool);
    char* snap_time = NULL;
    char *msg_name = NULL;
    int msg_handle = -1, msg_type = -1, i = 0, attr_count;
    long snap = -1;
    tNode const* n = NULL;

    /* Split the configuration line */
    if (s_strpsep (paramlist, line, trash_pool))
    {
        log (L_ERROR, "Error in config file: incomplete record in "
            "\"Aggregate\" clause number %d.", m_aggregate_count + 1);
        goto fail;
    }

    /* Parse (key, value) pairs */
    while (paramlist.count () > 0)
    {
        char *key_name =  static_cast<char *> (paramlist.remove_first ());
        char *key_value;

        if (paramlist.count () > 0)
            key_value = static_cast<char *> (paramlist.remove_first ());
        else
        {
            log (L_ERROR, "Error in config file: incomplete record in "
                "\"Aggregate\" clause number %d.", m_aggregate_count + 1);
            goto fail;
        }

        if (!strcmp("MessageTypeName", key_name))
        {
            /* Configure new clause */
            int type;
            msg_name = key_value;
            if ((msg_handle = gb_message_manager.get_handle_by_name (key_value))
                    == -1)
            {
                log (L_ERROR, "Error in config file: "
                        "\"Aggregate\" clause number %d: Unknown "
                        "message type \"%s\".", m_aggregate_count + 1, 
                        key_value);
                goto fail;
            }
            type = gb_message_manager.get_message_type_by_handle (msg_handle);
            if (msg_type != -1 || get_aggregate_handle (type) != -1)
            {
                log (L_ERROR, "Error in config file: "
                    "\"Aggregate\" clause number %d: "
                    "Multiple key_values of MessageTypeName '%s' "
                    "found.", m_aggregate_count + 1, key_value);
                goto fail;
            }
            msg_type = type;
        }
        else if (!strcmp("IgnoreAttribute", key_name))
            /* Setup ignored attributes */
            ignore_list.append (key_value);
        else if (!strcmp("SnapshotTimeLimit", key_name))
        {
            /* Setup snapshot limit */
            if (snap_time != NULL)
            {
                log (L_ERROR, "Error in config file: "
                    "\"Aggregate\" clause number %d: "
                    "Multiple key_values of SnapshotTimeLimit found.",
                    m_aggregate_count + 1);
                goto fail;
            }
            else
                snap_time = key_value;
        }
        else
        {
            log (L_ERROR, "Error in config file: "
                "\"Aggregate\" clause number %d: Unknown "
                "key %s.", m_aggregate_count + 1, key_name);
            goto fail;
        }
    }


    /* Test if mandatory msg_type is present */
    if (msg_type == -1)
    {
        log (L_ERROR, "Error in config file: "
                "\"Aggregate\" clause number %d: "
                "MessageTypeName must be specified.",
                m_aggregate_count + 1);
        goto fail;
    }

    /* Load snapshot limit */
    if (snap_time == NULL)
    {
        log (L_ERROR, "Error in config file: "
                "\"Aggregate\" clause number %d: "
                "SnapshotTimeLimit must be specified.",
                m_aggregate_count + 1);
        goto fail;
    }
    else
    {
        if (s_stol (snap_time, &snap, 1, INT_MAX))
        {
            log (L_ERROR, "Error in config file: "
                "\"Aggregate\" clause number %d: "
                "SnapshotTimeLimit must be a positive integer.",
                m_aggregate_count + 1);
            goto fail;
        }
    }

    /* Create a sorted array of ignored attribute handles */
    attr_count = ignore_list.count ();
    m_aggregates[m_aggregate_count].ignored_attrs = static_cast<int *>
        (conf_mp.alloc (sizeof (int) * attr_count));
    memset (m_aggregates[m_aggregate_count].ignored_attrs, 0, 
            sizeof (int) * attr_count);
    m_aggregates[m_aggregate_count].ignored_attr_count = attr_count;
    FOR_ALL_LIST_ITEMS (n, &ignore_list)
    {
        int handle;
        char *attr_name = static_cast<char *> (n->data);
        
        if ((handle = gb_message_manager.get_handle_of_attr (msg_handle,
                    attr_name)) == -1)
        {
            log (L_ERROR, "Error in config file: "
                    "\"Aggregate\" clause number %d: Unknown message "
                    "attribute \"%s\".", m_aggregate_count + 1, attr_name);
            goto fail;
        }
        if (gb_message_manager.is_key_attr_by_handle(msg_handle, handle))
        {
            log (L_ERROR, "Error in config file: "
                    "\"Aggregate\" clause number %d: Key attribute '%s' "
                    " can't be ignored.", m_aggregate_count + 1, attr_name);
            goto fail;
        }

        m_aggregates[m_aggregate_count].ignored_attrs[i++] = handle;
    }
    qsort (m_aggregates[m_aggregate_count].ignored_attrs, attr_count,
            sizeof (int), int_compator);

    /* Check for duplicates */
    if (attr_count > 1)
    {
        int last = m_aggregates[m_aggregate_count].ignored_attrs[0];
        for (i = 1; i < m_aggregates[m_aggregate_count].ignored_attr_count; i++)
        {
            int current = m_aggregates[m_aggregate_count].ignored_attrs[i];
            if (last == current)
            {
                log (L_ERROR, "Error in config file: "
                    "\"Aggregate\" clause number %d: "
                    "IgnoreAttribute '%s' specified more than once.",
                    m_aggregate_count + 1, 
                    gb_message_manager.get_attr_name (msg_handle, current));
                goto fail;
            }
            last = current;
        }
    }

    /* Fill in the rest */
    m_aggregates[m_aggregate_count].msg_type = msg_type;
    m_aggregates[m_aggregate_count].snapshot_time = snap;

    /* Make a new snapshot list for each msg_type */
    m_snapshots[m_aggregate_count] = 
        new tSnapshotList (&m_aggregates[m_aggregate_count], conf_mp);
    m_aggregate_count++;

    /* Okay, finished successfully */
    return false;

fail:
    return true;
}


/* Find aggregate handle for msg_type */
int tAggregatorTarget::get_aggregate_handle (int msg_type) const
{
    int i;

    for (i = 0; i < m_aggregate_count; i++)
        if (m_aggregates[i].msg_type == msg_type)
            return i;

    return -1;
}


/* Get last snapshot for (msg_type, agent, instance) */
tSnapshot *tAggregatorTarget::get_snapshot (int handle, long agent, 
        long instance)
{
    ASSERT (handle >= 0);

    tSnapshotList *l = m_snapshots[handle];
    const tNode *n;

    FOR_ALL_LIST_ITEMS(n, l)
    {
        tSnapshot *snapshot = static_cast<tSnapshot *> (n->data);
        ASSERT (snapshot);

        if (agent == snapshot->get_agent() && 
                instance == snapshot->get_instance ())
        {
            l->set_recent_node (const_cast<tNode *> (n));
            return snapshot;
        }
    }

    return NULL;
}


/* Add brand new snapshot for this msg_type */
void tAggregatorTarget::add_snapshot (int aggregate_handle, tSnapshot *snapshot)
{
    snapshot->set_last_sent (time (NULL));
    snapshot->mark_chunks_removed ();
    m_snapshots[aggregate_handle]->append (snapshot); 
}


/* Replace recent snapshot with new one */
void tAggregatorTarget::replace_recent_snapshot (int aggregate_handle, 
        tSnapshot *new_snapshot)
{  
    new_snapshot->mark_chunks_removed ();
    
    tSnapshotList *sl = m_snapshots[aggregate_handle];
    tSnapshot *recent_snapshot = static_cast<tSnapshot*>
        (sl->get_recent_node()->data);
    
    new_snapshot->set_last_sent (recent_snapshot->get_last_sent ());
    m_snapshots[aggregate_handle]->set_recent (new_snapshot); 
}


/* Send single snapshot message */
int tAggregatorTarget::send_snapshot (tSnapshot &snapshot)
{
    tMsgBuf message;

    if (message.prepare_to_write (diff_buffer, MAX_MESSAGE_SIZE))
        return -1;

    if (snapshot.put (message) < 0)
        return -1;
        
    if (message.close ())
        return -1;

    if (m_output->send_message (message))
        return -1;

    return 0;
}


/* Send all aged snapshots */
void tAggregatorTarget::resend_aged_snapshots ()
{
    time_t current_time = time (NULL);
    const tNode *n;

    for (int i = 0; i < m_aggregate_count; i++)
        FOR_ALL_LIST_ITEMS (n, m_snapshots[i])
        {
            tSnapshot *snapshot = static_cast<tSnapshot *> (n->data);
            ASSERT (snapshot);

            if (snapshot->get_last_sent() + m_aggregates[i].snapshot_time 
                    < current_time)
                if (send_snapshot (*snapshot) < 0)
                    log(L_ERROR, "Can't resend snapshot.");
        }
}


/* Send list of differences in one or more messages */
int tAggregatorTarget::send_changelist (tList& changelist, 
        tMsgHeader &old_header)
{
    tMsgBuf message;
    tPlainBuf *body;
    long length, subrecords;

    debug_log ('1', "Sending aggregated list");
    
    while (changelist.count() > 0)
    {
        /* Prepare a new message */
        if (message.prepare_to_write (diff_buffer, MAX_MESSAGE_SIZE))
            return -1;

        length = message.get_store_length ();
        subrecords = 0;
        body = message.get_body ();

        /* Write the body */
        while (length < MAX_MESSAGE_SIZE)
        {
            const tNode *c = changelist.get_first();
            if (!c)
                break;

            tMsgChunk *chunk = static_cast<tMsgChunk *> (c->data);
            long chunk_length = chunk->get_store_length();
                
            if (!chunk || chunk_length + length > MAX_MESSAGE_SIZE)
                break;

            if (chunk->put (body) <= 0)
                return -1;

            subrecords++;
            length += chunk_length;
            changelist.remove_first ();
        }

        /* Copy and adjust the old header */
        tMsgHeader *header = message.get_header ();
        header->copy_envelope (old_header);
        header->set_flag_incremental (true);
        header->set_subrecords (subrecords);

        /* Finish the message */
        if (message.close ())
            return -1;

        /* Send it */
        if (m_output->send_message (message))
            return -1;
    
        message.reset ();
    }

    return 0;
}


/* Ensures message passing to proper outputs */
bool tAggregatorTarget::process_msg (tMsgBuf& message)
{
    ASSERT (m_agstate == AG_INITIALIZED);
    ASSERT (m_output != NULL);

    int handle;
    bool ret = false;
    tList *diff;
    tMsgHeader *header = message.get_header ();
    ASSERT (header);

    /* Send old snapshots at first */
    resend_aged_snapshots ();

    if (!header->get_flag_aggregable () || header->get_flag_incremental () ||
            ((handle = get_aggregate_handle (header->get_msg_type ()) == -1)))
        /* Just pass non-aggregable messages */
        return forward_message (message);

    tSnapshot *old_snapshot, *snapshot;
    long agent = conv_id (header->get_agent_id ());
    long instance = conv_id (header->get_agent_instance_id ());

    /* Load a new snapshot */
    snapshot = new tSnapshot (m_aggregates[handle], agent, instance);
    if (snapshot == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        goto loading_snapshot_failed;
    }
    if (snapshot->init (message) < 0)
        goto loading_snapshot_failed;

    /* Get the old one */
    old_snapshot = get_snapshot (handle, agent, instance);
    if (!old_snapshot)
    {
        /* Add the new snapshot to configuration if there is none yet */
        add_snapshot (handle, snapshot);
        header->set_flag_aggregable (false);
        return forward_message (message);
    }
    
    /* Find the differences */
    diff = snapshot->find_differences (*old_snapshot);
    if (!diff)
    {
        /* There was a major problem, probably we're out of memory */
        log (L_ERROR, "Error aggregating message type %d.",
                header->get_msg_type ());
        /* Just pass the original message */
        forward_message (message);
        ret = true;
    }
    else
    {
        /* Send the list of differences */
        send_changelist (*diff, *header);
        delete diff;
    }

    /* Remember the new snapshot and purge the old */
    replace_recent_snapshot (handle, snapshot);
    delete old_snapshot;

    return ret;
    
  loading_snapshot_failed:
    /* On problems, try to at least pass the original message */
    forward_message (message);
    return true;
}
