/**
 * @file
 * @brief Contains declaration of tAggregatorTarget and tSnapshot classes
 *
 * @author Marek Zoth <marek.zoth@volny.cz>, Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_AGGREGATORTARGET_H
#define ECHELOG_AGGREGATORTARGET_H

#include "messages/msgchunk.h"
#include "messages/typelist.h"
#include "mframes/msgtarget.h"
#include "mframes/ioframe.h"


#define DEFAULT_SNAPSHOT_POOL_SIZE 16384


/// Aggregate clause configuration
struct tAggregate
{
    int msg_type;           ///< Message type
    int ignored_attr_count; ///< Number of ignored chunk attributes
    int *ignored_attrs;     ///< Sorted array of ignored chunk attribute handles
    int snapshot_time;      ///< Snapshot time limit
};


/// Single snapshot
class tSnapshot {

    protected:
        
        //-------------------------------
        // Owned components
        //-------------------------------
        
        /// Mempool for snapshot's contents
        lib::tMemPool snap_pool;
        /// Associate aggregate clause
        tAggregate& aggregate;
        /// Number of chunks stored in snapshot
        long chunk_count;
        /// Array of chunks stored in snapshot
        messages::tMsgChunk **chunks;
        /// Original message header
        messages::tMsgHeader msg_header;

        /// Agent id
        SIMPLE_R_PROPERTY (long, agent);
        /// Agent instance id
        SIMPLE_R_PROPERTY (long, instance);
        /// Timestamp when snapshot was sent 
        SIMPLE_PROPERTY (time_t, last_sent);

        //-------------------------------
        // Internal methods
        //-------------------------------

        /// Merges extra information from chunk b into a
        int tSnapshot::merge_chunks (messages::tMsgChunk &a, 
                messages::tMsgChunk &b);

    public: 
        
        //-------------------------------
        // Interface
        //-------------------------------
        
        /// Constructs empty snapshot for (agent,instance)
        tSnapshot (tAggregate &naggregate, long nagent, long ninstance)
            : snap_pool (DEFAULT_SNAPSHOT_POOL_SIZE), aggregate (naggregate),
              chunk_count (0), chunks (NULL), agent (nagent), 
              instance (ninstance), last_sent (0) {}
        /// Releases snapshot memory
        virtual ~tSnapshot();
        
        /// Loads initial state from snapshot message
        int init (messages::tMsgBuf& message);
        /// Writes snapshot to message
        int put (messages::tMsgBuf &message);
        /// Creates list of differences to another snapshot
        lib::tList * find_differences (tSnapshot &snapshot);

        /// Marks all chunks as removed
        inline void tSnapshot::mark_chunks_removed ()
        {
            for (int i = 0; i < chunk_count; i++)
                chunks[i]->set_status (messages::ST_REMOVED);
        }
};


/// List of snapshots
class tSnapshotList : public lib::tList {
    
    protected:
        
        /// Associated aggregate clause
        SIMPLE_R_EXT_RESOURCE (tAggregate *, aggregate);
        /// Recently accessed node
        SIMPLE_PROPERTY (lib::tNode *, recent_node);

    public:

        /// Constructs empty list of snapshots
        tSnapshotList (tAggregate *naggregate, lib::tMemPool &ipool) 
            : tList (ipool), aggregate (naggregate), recent_node (NULL) {}
        /// Set new data into recently accessed node
        void set_recent (tSnapshot *new_snapshot)
        {   ASSERT (recent_node); recent_node->data = new_snapshot; }
};




/**
 * @brief TODOC:
 */
class tAggregatorTarget : public mframes::tMsgTarget
{
    public:
        //-------------------------------
        // Interface
        //-------------------------------

        /**
         * @brief Constructor
         *
         * Performs basic initialization of the instance.
         */
        tAggregatorTarget ();

        /**
         * @brief Destructor
         *
         * Ensures final cleanup of the instance.
         */
        virtual ~tAggregatorTarget ();

        /**
         * @brief Initialization of the aggregator
         *
         * This method prepares aggregator for use with pass_msg().
         * It reads configuration and creates and prepares rules.
         *
         * @return
         * If succesful returns false. On failure return true.
         */
        virtual bool init ();

        /**
         * @brief Cleanup of the aggregator
         *
         * Releases and resets all allocated resources in init().
         * It is the reverse function for init().
         */
        virtual void cleanup ();

        /**
         * @brief Message passing cycle implementation
         *
         * Message is passed to each rule in order they were
         * in configuration. Rule will aggregator the message and ensures
         * the sending. Rule can modify the message which is then passed
         * to following rules.
         *
         * @param message
         * Message to be aggregatored and sent to other modules. Must be 
         * prepared for reading.
         *
         * @return
         * If the message was succesfully processed returns false.
         * If some error has occured returns true.
         */
        virtual bool process_msg (messages::tMsgBuf& message);

        //------------------------------------------
        // Implementation of tConfigurable interface
        //------------------------------------------

        /**
         * @brief TODOC:
         */
        virtual bool register_sections (lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void unregister_sections ();

        /**
         * @brief TODOC:
         */
        virtual bool process_config (lib::tMemPool& conf_mp,
                lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void cleanup_config ();

    private:
        //-------------------------------
        // Internal types
        //-------------------------------

        /**
         * @brief Description of the aggregator state
         */
        enum AGState
        {
            /// initalization failed during constructor
            AG_BROKEN,
            /// successfully created
            AG_CREATED,
            /// successfully configured
            AG_CONFIGURED,
            /// successfully initialized
            AG_INITIALIZED,
            /// successfully cleaned
            AG_CLEANED
        };

        //-------------------------------
        // Internal methods
        //-------------------------------

        /// Parse one configuration line (originally embraced section)
        bool parse_config_line (char* line, lib::tMemPool& conf_mp,
                lib::tMemPool& trash_pool);

        /// Find aggregate handle for msg_type
        int get_aggregate_handle (int msg_type) const;

        /// Get last snapshot for (msg_type, agent, instance)
        tSnapshot *get_snapshot (int aggregate_handle, long agent, 
                long instance);

        /// Add brand new snapshot for this msg_type
        void add_snapshot (int aggregate_handle, tSnapshot *snapshot);

        /// Replace recent snapshot with new one
        inline void replace_recent_snapshot (int aggregate_handle, 
                tSnapshot *new_snapshot);

        /// Forward incoming message to output
        inline bool forward_message (messages::tMsgBuf& message)
        {
            if (m_output->send_message (message))
                return true;
            return false;
        }

        /// Send all aged snapshots
        void resend_aged_snapshots ();
        /// Send single snapshot message
        int send_snapshot (tSnapshot &snapshot);
        /// Send list of differences in one or more messages
        int send_changelist (lib::tList& changelist, 
                messages::tMsgHeader& old_header);
        

        //-------------------------------
        // State attributes
        //-------------------------------

        /// State of the aggregator
        AGState m_agstate;

        //-------------------------------
        // Owned components
        //-------------------------------

        /// Output message buffer
        char diff_buffer[MAX_MESSAGE_SIZE];
        /// Desription of aggregator's configuration
        lib::cfitem* m_aggregator_section;
        /// Number of rules in m_rules array
        int m_aggregate_count;
        /// Array of rules to pass msgs to
        tAggregate *m_aggregates;
        /// Array of lists of snapshots
        tSnapshotList **m_snapshots;
        /// Output to pass messages to
        mframes::tOFrame::tOutput* m_output;
};

#endif //ifdef ECHELOG_FILTERTARGET_H

