/**
 * @file
 * @brief Contains implementation of tAggregatorTarget class
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "aggregatortarget.h"
#include "messages/msgbuf.h"
#include "messages/msgmgr.h"

using namespace lib;
using namespace messages;


tSnapshot::~tSnapshot ()
{
    delete[] chunks;
}


int tSnapshot::init (tMsgBuf& message)
{
    tMsgHeader *header = message.get_header ();
    chunk_count = header->get_subrecords ();
    chunks = new tMsgChunk *[chunk_count];
   
    for (int idx = 0; idx < chunk_count; idx++)
       chunks[idx] =  new tMsgChunk(header->get_msg_type (), snap_pool);

    if (!chunks)
    {
        log (L_ERROR, "Not enough memory.");
        chunk_count = 0;
        return -1;
    }

    /* Load all chunks from the message */
    for (int i = 0; i < chunk_count; i++)
    {
        if (chunks[i]->get (message.get_body ()) < 0)
        {
            char *agent_id = header->get_agent_id ();
            char *instance_id = header->get_agent_instance_id ();
            log (L_ERROR, "Chunk with invalid state in message (type %d, "
                    "agent %s, instance %s",  header->get_msg_type (), agent_id,
                    instance_id);
            xfree (agent_id);
            xfree (instance_id);

            return -1;
        }
        if (chunks[i]->get_status () != ST_ITEM)
        {
            /* This shouldn't happen, we're expecting correct snapshots */
            char *agent_id = header->get_agent_id ();
            char *instance_id = header->get_agent_instance_id ();
            log (L_ERROR, "Chunk with invalid state in message (type %d, "
                    "agent %s, instance %s", header->get_msg_type (), agent_id,
                    instance_id);
            xfree (agent_id);
            xfree (instance_id);
        }
        
        /* Mark all chunks as newly added */
        chunks[i]->set_status (ST_ADDED);
    }

    /* Remember the message envelope */
    msg_header.copy_envelope (*message.get_header ());

    return 0;
}


int tSnapshot::put (tMsgBuf &message)
{
    int i;
    tMsgHeader *header = message.get_header ();
    tPlainBuf *body = message.get_body ();

    /* Write all chunks to appropriate message buffer */
    for (i = 0; i < chunk_count; i++)
    {
        chunks[i]->set_status (ST_ITEM);
        if (chunks[i]->put (body) <= 0)
            return -1;
    }

    /* Copy and adjust the header */
    last_sent = time (NULL);
    header->copy_envelope (msg_header);
    header->set_msg_time (last_sent);
    header->set_subrecords (chunk_count);
    header->set_flag_aggregable (false);

    /* Reset chunk states */
    mark_chunks_removed ();

    return 0;
}


int tSnapshot::merge_chunks (tMsgChunk &a, tMsgChunk &b)
{
    int i, j;
    bool ignored;

    /* Check the usage */
    if (a.get_type() != b.get_type ())
        return -1;

    /* For all chunk attributes, try to merge information */
    for (i = 0; i < a.get_attr_count (); i++)
    {
        tEchelogType *at = a.get_attr(i), *bt = b.get_attr(i);
        tTYPE_RELATION r = type_order (*at, *bt);

        if (r == TP_EQUAL)
            /* Leech everything new from bt into at */
            at->merge (*bt);
        else if (r != TP_IDENTICAL)
        {
            /* Check if this is an ignored attribute */
            ignored = false;
            for (j = 0; j <= i && j < aggregate.ignored_attr_count; j++)
                if (i == aggregate.ignored_attrs[j])
                {
                    ignored = true;
                    break;
                }
            /* If not, mark this chunk as changed since at least one 
             * of its attributes differ */
            if (!ignored)
                a.set_status (ST_CHANGED);
        }
    }

    return 0;
}


tList* tSnapshot::find_differences (tSnapshot &snapshot)
{
    debug_log ('1', "Aggregating message");

    /* Create an empty list */
    tList *diff = new tList;
    if (!diff)
    {
        log (L_ERROR, "Not enough memory.");
        mark_chunks_removed ();
        return NULL;
    }

    /* For all chunks find their images */
    int i, j;
    for (i = 0; i < chunk_count; i++)
        for (j = 0; j < snapshot.chunk_count; j++)
        {
            if (chunks[i]->key_match (*snapshot.chunks[j]))
            {
                /* Mark both the original chunk and the image */
                chunks[i]->set_status (ST_ITEM);
                snapshot.chunks[j]->set_status (ST_ITEM);

                /* Try to merge extra bits from new snapshot */
                if (merge_chunks (*chunks[i], *snapshot.chunks[j]) < 0)
                {
                    log (L_ERROR, "Merging chunk of type %d failed.",
                            chunks[i]->get_type ());
                    goto merge_failed;
                }

                /* Add changed chunks to the list */
                if (chunks[i]->get_status() == ST_CHANGED)
                    diff->append (chunks[i]);
                break;
            }
        }

    /* Add all added chunks to the list */
    for (i = 0; i < chunk_count; i++)
        if (chunks[i]->get_status() == ST_ADDED)
            diff->append (chunks[i]);
    
    /* And all chunk which disappeared from the recent snapshot */
    for (i = 0; i < snapshot.chunk_count; i++)
        if (snapshot.chunks[i]->get_status() == ST_REMOVED)
            diff->append (snapshot.chunks[i]);

    return diff;

merge_failed:
    /* Restore the original state */
    mark_chunks_removed ();
    delete diff;
    return NULL;
}
