/**
 * @file
 * @brief Contains implementation of tAlertMasterTarget class
 *
 * @author Martin Macok <martin.macok@underground.cz>,
 * Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "alertmastertarget.h"
#include "messages/msgbuf.h"
#include "messages/msgmgr.h"
#include <sys/wait.h>

using namespace lib;
using namespace messages;
using namespace mframes;

/* Basic logstor initialization */
tAlertMasterTarget::tAlertMasterTarget ()
{
    m_prg = NULL;

    m_alertmaster_section = new cfitem [3];
    if (m_alertmaster_section == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        m_amstate = AMS_BROKEN;
        return;
    }

    int i = 0;
    m_alertmaster_section[i].name = "AlertMaster";
    m_alertmaster_section[i].type = CT_SECTION;
    m_alertmaster_section[i].data = NULL;
    i++;
    m_alertmaster_section[i].name = "Execute";
    m_alertmaster_section[i].type = CT_STRING;
    m_alertmaster_section[i].data = &m_prg;
    m_alertmaster_section[i].varf = NULL;
    m_alertmaster_section[i].varf_reset = NULL;
    m_alertmaster_section[i].flags = ITEM_REQUIRED;
    i++;
    m_alertmaster_section[i].name = NULL;
    m_alertmaster_section[i].type = CT_STOP;
    m_alertmaster_section[i].data = NULL;

    m_amstate = AMS_CREATED;
}

/* Final cleanup */
tAlertMasterTarget::~tAlertMasterTarget ()
{
    ASSERT (m_amstate == AMS_CREATED ||
            m_amstate == AMS_BROKEN);

    if (m_amstate == AMS_CREATED)
    {
        delete [] m_alertmaster_section;
    }
}

/* Initialization of the filter */
bool tAlertMasterTarget::init ()
{
    ASSERT (m_amstate == AMS_CONFIGURED);

    //place initialization here

    m_amstate = AMS_INITIALIZED;

    return false;
}

/* Reverse method for init */
void tAlertMasterTarget::cleanup ()
{
    ASSERT (m_amstate == AMS_INITIALIZED);

    //place cleanup here

    m_amstate = AMS_CLEANED;
}

bool tAlertMasterTarget::register_sections (lib::tMemPool& trash_pool)
{
    /* ensure the buffer is allocated */
    if (m_amstate == AMS_BROKEN)
        return true;

    cf_register (m_alertmaster_section);

    return false;
}

void tAlertMasterTarget::unregister_sections ()
{
    cf_unregister (m_alertmaster_section[0].name);
}

bool tAlertMasterTarget::process_config (lib::tMemPool& conf_mp,
        lib::tMemPool& trash_pool)
{
    ASSERT (m_amstate == AMS_CREATED);

    m_prg = s_strdup (m_prg, conf_mp);

    m_amstate = AMS_CONFIGURED;

    return false;
}

void tAlertMasterTarget::cleanup_config ()
{
    ASSERT (m_amstate == AMS_CLEANED ||
            m_amstate == AMS_CONFIGURED);

    m_prg = NULL;

    m_amstate = AMS_CREATED;
}

/* Ensures message passing to proper outputs */
bool tAlertMasterTarget::process_msg (tMsgBuf& message)
{
    ASSERT (m_amstate == AMS_INITIALIZED);

    return exec_message (message);
}

#define HL 80

/* this function is used only by forked childs */
void tAlertMasterTarget::do_exec_message (tMsgBuf& msg)
{
    FILE *f_popen = popen (m_prg, "w");

    if (!f_popen)
    {
        log(L_ERROR, "Child cannot popen() to send message. "
            "fork() or pipe() failed");
        return;
    }
        
    log(L_INFO, "Sending message to forked '%s'. ", m_prg);

    int fd_popen = fileno (f_popen);
    ASSERT (fd_popen!=-1);
    
    /* print Agent ID and Host ID from message header */
    tMsgHeader *header = msg.get_header ();
    char *hstr = header->get_host_str ();
    char *aptr = header->get_agent_id ();
    char head[HL];
    snprintf (head, HL, "HOST: %s AGENT: %s\n",
                        hstr, aptr);
    s_write (fd_popen, head, s_strnlen (head, HL));
    xfree(hstr);
    xfree(aptr);

    /* print the body of message */
    m_printer.print_message (fd_popen, msg);
    pclose (f_popen);      // smoke^Wclose the pipe and wait for prg exit ...
}

/* fork and daemonize child, parent exits immediately because it blocks */
void tAlertMasterTarget::daemonize_child() const
{
    setsid ();                             // see echelogd.cc daemonize()
    pid_t apid = fork ();
    if (apid < 0)              /* go buy more RAM */
    {
        log (L_ERROR, "Child fork failed");
        _exit (0);
    }
    else if (apid > 0)
        _exit (0); /* this is parent process, we don't need him */
                  /* and it's blocking parent anyway */
    s_close (1);
    s_close (0);
    sd_open ("/dev/null", O_RDONLY);
    sd_open ("/dev/null", O_WRONLY);
}


/* forks a children which popen()'s "prg" and feed it with text msg */
bool tAlertMasterTarget::exec_message (tMsgBuf& msg)
{
    pid_t child_pid = fork ();

    if (child_pid == -1) {                          // error, forking failed
        log (L_WARN, "Forking to run Exec program failed");
        return true;
    } else if (child_pid > 0) {              // we're proud of our childrens
        wait (NULL);       // this should wait very very small time
        return false;
    } else {                               // we're children, we do the work
        daemonize_child ();
        do_exec_message (msg);
        _exit (0);                               // we don't need you anymore
    }
}

