/**
 * @file
 * @brief Contains declaration of tAlertMasterTarget class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#ifndef ECHELOG_ALERTMASTERTARGET_H
#define ECHELOG_ALERTMASTERTARGET_H

#include "mframes/msgtarget.h"
#include "mframes/ioframe.h"
#include "messages/msgprinter.h"

/**
 * @brief TODOC:
 */
class tAlertMasterTarget : public mframes::tMsgTarget
{
    public:
        //-------------------------------
        // Interface
        //-------------------------------

        /**
         * @brief Constructor
         *
         * Performs basic initialization of the instance.
         */
        tAlertMasterTarget ();

        /**
         * @brief Destructor
         *
         * Ensures final cleanup of the instance.
         */
        virtual ~tAlertMasterTarget ();

        /**
         * @brief Initialization of the filter
         *
         * This method prepares filter for use with pass_msg().
         * It reads configuration and creates and prepares rules.
         *
         * @return
         * If succesful returns false. On failure return true.
         */
        virtual bool init ();

        /**
         * @brief Cleanup of the filter
         *
         * Releases and resets all allocated resources in init().
         * It is the reverse function for init().
         */
        virtual void cleanup ();

        /**
         * @brief Message passing cycle implementation
         *
         * Message is passed to each rule in order they were
         * in configuration. Rule will filter the message and ensures
         * the sending. Rule can modify the message which is then passed
         * to following rules.
         *
         * @param message
         * Message to be filtered and sent to other modules. Must be prepared
         * for reading.
         *
         * @return
         * If the message was succesfully processed returns false.
         * If some error has occured returns true.
         */
        virtual bool process_msg (messages::tMsgBuf& message);

        //------------------------------------------
        // Implementation of tConfigurable interface
        //------------------------------------------

        /**
         * @brief TODOC:
         */
        virtual bool register_sections (lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void unregister_sections ();

        /**
         * @brief TODOC:
         */
        virtual bool process_config (lib::tMemPool& conf_mp,
                lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void cleanup_config ();

    private:
        //-------------------------------
        // Internal types
        //-------------------------------

        /**
         * @brief Description of the filter state
         */
        enum AMSState
        {
            /// initalization failed during constructor
            AMS_BROKEN,
            /// successfully created
            AMS_CREATED,
            /// successfully configured
            AMS_CONFIGURED,
            /// successfully initialized
            AMS_INITIALIZED,
            /// successfully cleaned
            AMS_CLEANED
        };

        /* this function is used only by forked childs */
        void do_exec_message (messages::tMsgBuf& msg);

        /* fork and daemonize child, parent exits immediately because it
        blocks */
        void daemonize_child () const;

        /* forks a children which popen()'s "prg" and feed it with text msg */
        bool exec_message (messages::tMsgBuf& msg);

        //-------------------------------
        // State attributes
        //-------------------------------

        /// State of the filter
        AMSState m_amstate;

        //-------------------------------
        // Owned components
        //-------------------------------

        /// Desription of filter's configuration
        lib::cfitem* m_alertmaster_section;
        /// Printer we use
        messages::tMsgPrinter m_printer;
        /// program to execute
        const char* m_prg;
};

#endif //ifdef ECHELOG_FILTERTARGET_H

