/**
 * @brief Echelog Binary storage  
 *
 * This agent waits on specified UNIX-domain socket and binstores
 * incomming messages.
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */                                                                             

#include "binstor.h"
#include "mframes/msgprocessingmodule.h"

using namespace lib;
using namespace messages;
using namespace spool;
using namespace mframes;

struct cfitem tBinstorTarget::s_binstor_section[]={
    {name: "BinStor",            type: CT_SECTION,       data: NULL},
    {name: "Datafile",           type: CT_STRING,        data: &data_file,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "Indexfile",          type: CT_STRING,        data: &index_file,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: NULL,                 type: CT_STOP,           data: NULL}
};

class spool::tBinStore *bs;
char *index_file, *data_file;

/* Initialization of the filter */
bool tBinstorTarget::init ()
{
    bs = new tBinStore(1);
    if (!bs)
    {
        log (L_ERROR, "Not enough memory.");
        goto clean;
    }

    if (bs->open_w (data_file, index_file))
    {
        log (L_ERROR, "Cannot open/create BINSTORE files.");
        goto bin_clean;
    }

    return false;

bin_clean:    
    delete bs;
clean:    
    return true;
}

/* Reverse method for init */
void tBinstorTarget::cleanup ()
{
    bs->close ();
    delete bs;
}

bool tBinstorTarget::register_sections (lib::tMemPool& trash_pool)
{
    cf_register (s_binstor_section);
    return false;
}

void tBinstorTarget::unregister_sections ()
{
    cf_unregister (s_binstor_section[0].name);
}

bool tBinstorTarget::process_config (lib::tMemPool& conf_mp, 
        lib::tMemPool& trash_pool)
{
    data_file = s_strdup (data_file, conf_mp);
    index_file = s_strdup (index_file, conf_mp);

    return false;
}

void tBinstorTarget::cleanup_config ()
{
    data_file = NULL;
    index_file = NULL;
}

/* Ensures message passing to proper outputs */
bool tBinstorTarget::process_buffer (char* buf, int len)
{
    messages::tMsgBuf message;
    
    /* Check size of received buffer */
    if (len <= 0 && len > MAX_MESSAGE_SIZE)
        goto invalid_message;

    /* Try to load message header */
    if (message.prepare_to_read (buf, len))
        goto invalid_message;

    /* Check the length of message buffer against its declared length */
    if (message.get_declared_length() != len)
        goto invalid_message;

    if (message.close())
        log (L_ERROR, "Can't close opened message buffer.");

    /* Basic checks passed, store it to configured binstor */
    if (bs->add_item (buf, len)==0)
    {
            log(L_ERROR, "Error storing message to BIN.");
            return true;
    }

    return false;

  invalid_message:

    /* Oops, the message header is corrupted. Drop the message */
    log (L_ERROR, "Message with corrupted header received. Dropping it.");
    return true;
}

/// Name of the file, where the binstore's configuration is written
const char* const cfg_filename = ECHELOG_CONF_DIR  "/common.conf";

/// --===  PROGRAM ENTRY POINT ===--
int main (int argc, char *argv[])
{
    tModule::tCmdLine cmdline (argc, argv);
    tBinstorTarget target;
    tMsgProcessingModule binstor (cmdline, cfg_filename);

    if (binstor.init (&target, false))
        return 1;

    int ret = binstor.run ();
    binstor.cleanup ();

    return ret;
}
