/**
 * @brief Echelog BinCut tool 
 *
 * This tool extract messages within given time interval from input bin and
 * stores them into output bin.
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */                                                                             

#include "bincut.h"
#include "messages/msgbuf.h"
#include "messages/msgheader.h"
#include "mframes/simplemodule.h"

#include <time.h>
#include <sys/time.h>

#define DEFAULT_CONFIG_FILE ECHELOG_CONF_DIR  "/common.conf" 

using namespace lib;
using namespace messages;
using namespace mframes;

struct cfitem tBinCutModule::s_bincut_section[]={
    {name: "BinCut",             type: CT_SECTION,       data: NULL},
    {name: "InputDatafile",      type: CT_STRING,        
data: &tBinCutModule::s_input_data_file, varf: NULL, varf_reset: NULL, 
flags: ITEM_REQUIRED },
    {name: "InputIndexfile",     type: CT_STRING,        
data: &tBinCutModule::s_input_index_file, varf: NULL, varf_reset: NULL,
flags: ITEM_REQUIRED },
    {name: "OutputDatafile",     type: CT_STRING,
data: &tBinCutModule::s_output_data_file, varf: NULL, varf_reset: NULL,
flags: ITEM_REQUIRED },
    {name: "OutputIndexfile",    type: CT_STRING,
data: &tBinCutModule::s_output_index_file, varf: NULL, varf_reset: NULL,
flags: ITEM_REQUIRED },
    {name: "StartTime",          type: CT_STRING,
data: &tBinCutModule::s_starttime, varf: NULL, varf_reset: NULL,
flags: ITEM_REQUIRED },
    {name: "EndTime",            type: CT_STRING,
data: &tBinCutModule::s_endtime, varf: NULL, varf_reset: NULL,
flags: ITEM_REQUIRED },
    {name: NULL,                 type: CT_STOP,          data: NULL}
};

char *tBinCutModule::s_input_data_file, *tBinCutModule::s_input_index_file, 
     *tBinCutModule::s_output_data_file, *tBinCutModule::s_output_index_file,
     *tBinCutModule::s_starttime, *tBinCutModule::s_endtime;


bool tBinCutModule::process ()
{
    char buffer[MAX_MESSAGE_SIZE];
    int buf_size;
    class tMsgBuf message;
    long total=0,extracted=0;
    bool res=false;

    /* for each item */
    for (int i=1;i<=static_cast<int>(m_inbin->get_last_id());i++)
    {

        tModule::SigHandlingRet ret = tSimpleModule::handle_signals();
        
        if (ret == SHR_BREAK_ERR)
        {
            res = true;
            break;
        } 

        if (ret == SHR_BREAK_OK)
            break; 

        /* get next message */
        buf_size = MAX_MESSAGE_SIZE;
        if (m_inbin->get_item(i, buffer, &buf_size))
        {
             log (L_ERROR, "Error getting item from input BIN.");
             return true;
        }

        total++;

        /* Prepare the message buffer */
        if (message.prepare_to_read (buffer, buf_size))
        {
             log (L_ERROR, "Unable to parse message id=\"%d\" from input BIN", i);
             return true;
        }

        /* Look into messages header */
        tMsgHeader *header = message.get_header ();

        /* we are interested only in messages within give time interval */
        if (header->get_msg_time () < m_start_time ||
            header->get_msg_time () > m_end_time)
        {
            message.reset();
            continue;
        }

        /* put the message into output bin */
        if (m_outbin->add_item (buffer, buf_size)==0)
        {
            log (L_ERROR, "Error writing to output BINSTORE.");
            return true;
        }

        message.reset();

        extracted ++;

    }

    if (res)
        return true;
    else
    {
        log (L_INFO, "Total messages: %ld Extracted: %ld", total, extracted);
        return false;
    }
}
        
bool tBinCutModule::register_sections (lib::tMemPool& trash_pool)
{
    if (tSimpleModule::register_sections (trash_pool))
        return true;

    cf_register (s_bincut_section);

    return false;
}
    

void tBinCutModule::unregister_sections ()
{
    tSimpleModule::unregister_sections ();

    cf_unregister (s_bincut_section[0].name);
}

bool tBinCutModule::process_config (lib::tMemPool& conf_mp, lib::tMemPool&
        trash_pool)
{
    char buf1[32],buf2[32];
    struct tm *tmp_tm;
    int tmp_fd;

    if (tSimpleModule::process_config (conf_mp, trash_pool))
        return true;

    /* check whether input data file exists */
    if ((tmp_fd=s_open (s_input_data_file, O_RDONLY))< 0)
    {
        log (L_ERROR, "Cannot open input BIN files.");
        goto ancestor_clean;
    }

    s_close (tmp_fd);

    /* create input bin instance and open it for reading */
    m_inbin = new spool::tBinStore(0);
    if (!m_inbin)
    {
        log (L_ERROR, "Not enough memory.");
        goto ancestor_clean;
    }

    if (m_inbin->open_r (s_input_data_file, s_input_index_file))
    {    
        log (L_ERROR, "Cannot open input BIN files.");
        goto in_binstore_clean;
    }

    /* create output bin instance and open it for writing */
    m_outbin = new spool::tBinStore(0);
    if (!m_outbin)
    {
        log (L_ERROR, "Not enough memory.");
        goto in_binstore_create_clean;
    }

    if (m_outbin->create (s_output_data_file, s_output_index_file))
    {
        log (L_ERROR, "Cannot open output BINSTORE files.");
        goto out_binstore_clean;
    }   

    /* read starting time */
    if ((m_start_time=parsetime (s_starttime)) < 0)
    {
        log (L_ERROR, "Error in StartTime config file parameter.");
        goto out_binstore_create_clean;
    }

    /* read end time */
    if ((m_end_time=parsetime (s_endtime)) < 0)
    {
        log (L_ERROR, "Error in EndTime config file parameter.");
        goto out_binstore_create_clean;
    }

    /* print the extraction interval */
    tmp_tm = localtime (&m_start_time);
    strftime(buf1, sizeof(buf1), "%Y-%m-%d %H:%M:%S", tmp_tm);
    tmp_tm = localtime (&m_end_time);
    strftime(buf2, sizeof(buf2), "%Y-%m-%d %H:%M:%S", tmp_tm);

    if (m_start_time==0) strcpy (buf1, "start of log");
    if (m_end_time==0) strcpy (buf2, "start of log");
    if (m_start_time==0x7FFFFFFF) strcpy (buf1, "end of log");
    if (m_end_time==0x7FFFFFFF) strcpy (buf2, "end of log");

    log (L_INFO, "Extracting messages in interval <%s ; %s>.", buf1, buf2);

    return false;

out_binstore_create_clean:
    m_outbin->close ();

out_binstore_clean:
    delete m_outbin;

in_binstore_create_clean:
    m_inbin->close ();

in_binstore_clean:
    delete m_inbin;

ancestor_clean:
    tSimpleModule::cleanup_config ();

    return true;
}

void tBinCutModule::cleanup_config ()
{
    m_outbin->close ();
    delete m_outbin;
    m_inbin->close ();
    delete m_inbin;
    tSimpleModule::cleanup_config ();
}

/* --===  PROGRAM ENTRY POINT ===-- */
int main (int argc, char *argv[])
{
    tModule::tCmdLine cmdline (argc, argv);
    tBinCutModule bincut (cmdline, DEFAULT_CONFIG_FILE);

    if (bincut.init ())
        return 1;

    int ret = bincut.run ();
    bincut.cleanup();
   
    return ret;
}
