/**
 * @brief Echelog BinGrep tool 
 *
 * This tool searches through given bin and writes matching enties
 * to output html file
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */                                                                             

#include "bingrep.h"

#define DEFAULT_CONFIG_FILE ECHELOG_CONF_DIR  "/common.conf" 

using namespace lib;
using namespace messages;
using namespace spool;
using namespace mframes;

struct cfitem tBinGrepModule::s_bingrep_section[]={
    {name: "BinGrep",            type: CT_SECTION,       data: NULL},
    {name: "InputDatafile",      type: CT_STRING,        data: &input_data_file,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "InputIndexfile",     type: CT_STRING,        data: &input_index_file,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "OutputFile",         type: CT_STRING,        data: &output_file,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "AttrSearch",         type: CT_STRING,        data: &attr_search,
varf: NULL, varf_reset: NULL },
    {name: "Pattern",            type: CT_STRING,        data: &pattern,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "MsgName",            type: CT_STRING,        data: &msgname,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "Host",               type: CT_STRING,        data: &hostname,
varf: NULL, varf_reset: NULL },
    {name: "AgentName",          type: CT_STRING,        data: &agentname,
varf: NULL, varf_reset: NULL },
    {name: "Instance",           type: CT_STRING,        data: &instancenm,
varf: NULL, varf_reset: NULL },
    {name: NULL,                 type: CT_STOP,          data: NULL}
};

/* variables read from config file */
char *input_data_file, *input_index_file, *output_file,
     *attr_search, *pattern, *msgname, *hostname,
     *agentname, *instancenm;

class tMemPool list_mp(2048);

/* list of attributes we will search in */
class tList attrs_list(list_mp);
/* list of messages we will search in */
class tList msgs_list(list_mp);
/* list of hosts we will search in */
class tList hosts_list(list_mp);
/* list of instances we will search in */
class tList instances_list(list_mp);
/* list of agents we will search in */
class tList agents_list(list_mp);

/* flag - search in all attributes */
int all_attrs;
/* flag - search in all msgs */
int all_msgs;
/* flag - search in all hosts */
int all_hosts;
/* flag - search in all instances */
int all_instances;
/* flag - search in all agents */
int all_agents;

/* input bin */
class tBinStore *inbin;
/* regular expression tree */
regex_t    r_tree;

/* output instance to xml */
class tXMLOutput xml;
/* file descriptor of output file */
int xml_fd;

bool tBinGrepModule::process ()
{
    generate_output ();

    return false;
}
        
bool tBinGrepModule::register_sections (lib::tMemPool& trash_pool)
{
    if (tSimpleModule::register_sections (trash_pool))
        return true;

    cf_register (s_bingrep_section);

    return false;
}

void tBinGrepModule::unregister_sections ()
{
    tSimpleModule::unregister_sections ();

    cf_unregister (s_bingrep_section[0].name);
}

bool tBinGrepModule::process_config (lib::tMemPool& conf_mp, lib::tMemPool&
        trash_pool)
{
    if (tSimpleModule::process_config (conf_mp, trash_pool))
        return true;

    /* create output xml file */
    if (!strcmp (output_file, "stdout"))
        xml_fd = 1;
    else
        xml_fd = s_open (output_file, O_WRONLY | O_CREAT | O_TRUNC,
                         S_IRUSR | S_IWUSR);
    if (xml_fd < 0)
        die ("Can't create file \"%s\" for bingrep output.", output_file);

    /* setup the xml output object */
    xml.set_fd (xml_fd);
    xml.set_die_on_err (true);

    /* output the xml header */
    xml.print_header ("bingrep.xsl");

    /* check whether input data file exists */
    int tmp_fd = s_open (input_data_file, O_RDONLY);
    if (tmp_fd < 0)
        xml_die ("Cannot open specified input BIN files.");
    s_close (tmp_fd);

    /* create input bin instance and open it for reading */
    inbin = new tBinStore(0);
    if (!inbin)
        xml_die ("Error allocating memory for BIN object.");

    if (inbin->open_r (input_data_file, input_index_file))
        xml_die ("Cannot open specified input BIN files.");


    /* find which attributes we should search in */
    if (!attr_search || !strcmp (attr_search, "*"))
        all_attrs = 1;
    else
    {
        all_attrs = 0;

        if (s_strpsep (attrs_list, attr_search, conf_mp))
            xml_die ("Error in specified attributes to search in.");
    }

    /* find which messages we should search in */
    if (!msgname || !strcmp (msgname, "*"))
        all_msgs = 1;
    else
    {
        class tNode const *item;
        class tList tmp_list (conf_mp);

        all_msgs = 0;

        if (s_strpsep (tmp_list, msgname, conf_mp))
            xml_die ("Error in specified messages to search in.");

        FOR_ALL_LIST_ITEMS(item, &tmp_list)
        {
            char *n_msg = static_cast<char *> (item->data);
            int *msg_type = static_cast<int *> (conf_mp.alloc (sizeof (int)));

            /* get the message id */
            if (gb_message_manager.get_handle_by_name (n_msg) < 0)
                xml_die("Specified messages for search contain unknown "
                        "message type.");

            *msg_type = gb_message_manager.get_message_type_by_handle
                (gb_message_manager.get_handle_by_name (n_msg));
            msgs_list.append (msg_type);
        }

    }

    /* find which agents we should search in */
    if (!agentname || !strcmp (agentname, "*"))
        all_agents = 1;
    else
    {
        all_agents = 0;

        if (s_strpsep (agents_list, agentname, conf_mp))
            xml_die ("Error in specified agents to search in.");
    }

    /* find which instances we should search in */
    if (!instancenm || !strcmp (instancenm, "*"))
        all_instances = 1;
    else
    {
        all_instances = 0;

        if (s_strpsep (instances_list, instancenm, conf_mp))
            xml_die ("Error in specified instances to search in.");
    }

    /* find from which hosts we should search in */
    if (!hostname || !strcmp (hostname, "*"))
        all_hosts = 1;
    else
    {
        class tNode const *item;
        class tList tmp_list (conf_mp);

        all_hosts = 0;

        if (s_strpsep (tmp_list, hostname, conf_mp))
            xml_die ("Error in specified hosts to search in.");

        FOR_ALL_LIST_ITEMS(item, &tmp_list)
        {
            char *n_ip = static_cast<char *> (item->data);
            byte *ip_arr = static_cast<byte *> (conf_mp.alloc (4*sizeof (byte)));
            class tList ip_list (conf_mp);

            /* parse one ip */
            int n = (int)strlen(n_ip);
            int ai = 0;

            for (ai=0;ai<n;ai++)
                if (n_ip[ai]=='.') n_ip[ai]=' ';

            if (s_strpsep (ip_list, n_ip, conf_mp) || ip_list.count()!=4)
                xml_die("Error in specified hosts. Invalid IP.");

            int err = 0;
            
            for (int i=0;i<4;i++)
            {
                char *octet = static_cast<char *> (ip_list.remove_first ());
                long res;
                
                err = err || s_stol (octet, &res, 0, 255);
                ip_arr[i] = res;
            }

            if (err)
                xml_die("Error in specified hosts. Invalid IP.");

            hosts_list.append (ip_arr);
        }

    }

    int error;
    if ((error = regcomp (&r_tree, pattern, REG_EXTENDED)))
    {
        size_t bytes = regerror (error, &r_tree, NULL, 0);

        char *err_msg = static_cast<char *> (conf_mp.alloc (bytes));

        if (regerror (error, &r_tree, err_msg, bytes) != bytes)
            xml_die ("Not enough memory for allocating of error string.");

        xml_regexp_die (err_msg);
    }

    return false;

}

void tBinGrepModule::cleanup_config ()
{
    tSimpleModule::cleanup_config ();
}

/* --===  PROGRAM ENTRY POINT ===-- */
int main (int argc, char *argv[])
{
    tModule::tCmdLine cmdline (argc, argv);
    tBinGrepModule bingrep (cmdline, DEFAULT_CONFIG_FILE);

    if (bingrep.init ())
        return 1;

    int ret = bingrep.run ();
    bingrep.cleanup();
   
    return ret;
}
