/**
 * @brief Echelog BinGrep tool - Xml output  
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */                                                                             

#include "bingrep.h"

using namespace lib;
using namespace messages;

static class tMemPool pr_pool (2048);

void xml_cdie(char *message);

int good_attr(char *attr_name);

int match_attr (class tEchelogType &attr);

int print_chunk (class tXMLOutput &axml, class tMsgChunk &chunk, char *msgnam,
             char *hostnam, char *agentnam, char *instancenam, char *timebuf,
             tMSG_STATUS status);

void xml_info ();

int good_message(class tMsgHeader *header);

void grep_message (class tMsgBuf &message)
{
    class tMemPool chunk_mp (MAX_MESSAGE_SIZE);
    class tMsgChunk *chunk;
    char time_buf[32];

    /* get header information */
    class tMsgHeader *header =  message.get_header ();
    char *host_buf = header->get_host_str ();
    char *aptr = header->get_agent_id ();
    char *iptr = header->get_agent_instance_id ();
    int msg_type = header->get_msg_type ();

    /* get message name */
    int message_handle = gb_message_manager.get_handle_by_type (msg_type);
    char *message_name = gb_message_manager.get_message_name_by_handle
                                                       (message_handle);

    /* create time string for time of message */
    time_t msg_time = header->get_msg_time ();
    struct tm *tmp_tm = localtime (&msg_time);
    strftime(time_buf, sizeof (time_buf), "%Y-%m-%d %H:%M:%S", tmp_tm);

    /* Process all chunks */
    for (int j = 0; j < header->get_subrecords (); j++)
    {
         /* create new chunk */
         chunk = new tMsgChunk (msg_type, chunk_mp);

         /* load the chunk into the structure */ 
         if (chunk->get (message.get_body ()) <= 0)
             xml_cdie ("Message in temporary BIN is corrupted.");

         int num_of_attributes = chunk->get_attr_count();
         int found = 0;

         /* grep each attribute */
         for (int i = 0; i < num_of_attributes; i++)
         {
              char *name = gb_message_manager.get_attr_name (message_handle, i);

              /* check only attrs we are interested in */
              if (good_attr(name) && match_attr(*(chunk->get_attr(i))))
              {
                      found = 1;
                      break;
              }
                      
         }

         if (!found) continue;

         /* yeah we really found it */
         print_chunk(xml, *chunk, message_name, host_buf, aptr, iptr, time_buf,
                 chunk->get_status());

    }

    /* free allocated structures */
    xfree (host_buf);
    xfree (aptr);
    xfree (iptr);

}

void generate_output ()
{
    char buffer[MAX_MESSAGE_SIZE];
    int buf_size;
    class tMsgBuf message;

    /* print xml header */
    xml_info();

    /* grep all messages */
    for (int i=1;i<=static_cast<int>(inbin->get_last_id());i++)
    {

        /* get next message */
        buf_size = MAX_MESSAGE_SIZE;
        if (inbin->get_item(i, buffer, &buf_size))
            xml_cdie ("Error reading input BIN file.");

        /* Prepare the message buffer */
        if (message.prepare_to_read (buffer, buf_size))
            xml_cdie ("Message in input BIN is corrupted.");

        /* Look into messages header */
        tMsgHeader *header = message.get_header ();

        /* we are interested only in certain messages */
        if (!good_message (header))
        {
            message.reset();
            continue;
        }
 
        /* proceed grep on message */
        grep_message (message);

        /* free allocated structures */
        message.reset();

    }

    /* write end of xml file */
    xml.print_end_tag ("data");
    xml.print_end_tag ("bingrep");
    s_close (xml.get_fd ());
    
}

/* checks if this message holds interesting chunks */
int good_message(class tMsgHeader *header)
{
    class tNode const *item;
    byte *hptr = header->get_host_id ();
    char *aptr = header->get_agent_id ();
    char *iptr = header->get_agent_instance_id();
    int type_id = header->get_msg_type();
    int i, ok = 1;

    /* test whether we search for this host */
    if (!all_hosts)
    {
        int ip_match = 0;

        FOR_ALL_LIST_ITEMS(item, &hosts_list) 
        {
            byte *ip_arr = static_cast<byte *> (item->data);
            
            ip_match = 1;

            for(i=0;i<4;i++) if (ip_arr[i]!=hptr[i]) 
                ip_match = 0;

            if (ip_match) break;
        }

        if (!ip_match) ok = 0;
    }
    
    /* test if we are interested in this agent */
    if (!all_agents)
    {
        int ag_match = 0;

        FOR_ALL_LIST_ITEMS(item, &agents_list)
        {
            char *agent_str = static_cast<char *> (item->data);

            if (!strcmp (agent_str, aptr))
            {
                ag_match = 1;
                break;
            }
        }

        if (!ag_match) ok = 0;
    }

    /* test if we are interested in this instance */
    if (!all_instances)
    {
        int in_match = 0;

        FOR_ALL_LIST_ITEMS(item, &instances_list)
        {
            char *inst_str = static_cast<char *> (item->data);

            if (!strcmp (inst_str, iptr))
            {
                in_match = 1;
                break;
            }
        }

        if (!in_match) ok = 0;
    }

    /* test if we are interested in this message */
    if (!all_msgs)
    {
        int ms_match = 0;

        FOR_ALL_LIST_ITEMS(item, &msgs_list)
        {
            int *msgid = static_cast<int *> (item->data);

            if (*msgid>=0 && *msgid == type_id)
            {
                ms_match = 1;
                break;
            }
        }

        if (!ms_match) ok = 0;
    }

    xfree (hptr);
    xfree (aptr);
    xfree (iptr);

    return ok;
}

/* checks if this attribute is to be searched */
int good_attr(char *attr_name)
{
    class tNode const *item;

    /* test whether we search for this attribute */
    if (all_attrs) return 1;

    FOR_ALL_LIST_ITEMS(item, &attrs_list)
    {
        char *attr_nm = static_cast<char *> (item->data);

        if (!strcmp (attr_name, attr_nm))
            return 1;
    }

    return 0;
}


/* print one attribute */
int print_attr (class tXMLOutput &axml , class tEchelogType &attr, 
                const char *attr_name)
{
    int n = attr.get_num_of_components ();
    class tList list (pr_pool);
    int err = 0;
    
    /* simple type */
    if (n == 0)
    {
        attr.sprint (list, pr_pool, attr.get_plain_format());
        
        void *item = list.remove_first ();
        char *string = static_cast<char *> (item);

        if (!string)
            string = "";
        
        return axml.print_tag(attr_name, string);
    }
    else /* complex type */
    {
        err = min (err, axml.print_start_tag (attr_name));
        
        for (int i=0; i<n; i++)
        {
            err = min (err, print_attr (axml, *(attr.get_component_by_num (i)), 
                        attr.get_component_name_by_num (i)));
        }

        err = min (err, axml.print_end_tag (attr_name));

        return err;
    }
}

/* try to match one attribute */
int match_attr (class tEchelogType &attr)
{
    int n = attr.get_num_of_components ();
    class tMemPool match_pool (2048);
    class tList list (match_pool);
    regmatch_t pm;
    
    /* simple type */
    if (n == 0)
    {
        /* output attribute value */
        attr.sprint (list, pr_pool, attr.get_plain_format());
        
        void *item = list.remove_first ();
        char *string = static_cast<char *> (item);

        if (!string)
            string = "";

        /* try to analyze it */
        if (!(regexec (&r_tree, string, 1, &pm, 0)))
            return 1;
        else
            return 0;
        
    }
    else /* complex type */
    {
        /* try to find match at least in one component */
        for (int i=0; i<n; i++)
        {
            if (match_attr (*(attr.get_component_by_num (i))))
                return 1;
        }

        /* not found */
        return 0;
    }
}

char * status_to_str (tMSG_STATUS status)
{
    switch (status)
    {
        case ST_ADDED:
            return "Added";
        case ST_REMOVED:
            return "Removed";
        case ST_CHANGED:
            return "Changed";
        case ST_ITEM:
            return "Snapshot";
        default:
            return "Error";
    }
}


/* print chunk contents */
int print_chunk (class tXMLOutput &axml, class tMsgChunk &chunk, char *msgnam,
        char *hostnam, char *agentnam, char *instancenam, char *timebuf,
        tMSG_STATUS status)
{
     int message_handle = gb_message_manager.get_handle_by_type 
         (chunk.get_type ());
     char *mn = gb_message_manager.get_message_name_by_handle (message_handle);
     int num_of_attributes = chunk.get_attr_count();
     int err=0;

     /* print start of chunk */
     err = min (err, axml.print_start_tag (mn));

     /* header info */
     err = min (err, axml.print_tag ("m_status", status_to_str (status)));
     err = min (err, axml.print_tag ("m_host", hostnam));
     err = min (err, axml.print_tag ("m_agent", agentnam));
     err = min (err, axml.print_tag ("m_instance", instancenam));
     err = min (err, axml.print_tag ("m_time", timebuf));

     /* print each attribute */
     for (int i = 0; i < num_of_attributes; i++)
     {
         char *name = gb_message_manager.get_attr_name (message_handle, i);
         err = min (err, print_attr (axml, *(chunk.get_attr(i)), name)); 
     }

     /* print end of chunk */
     err = min (err, axml.print_end_tag (mn));

     pr_pool.free_pool ();

     return err;
}

void xml_die(char *message)
{
     xml.print_start_tag ("bingrep");
     xml.print_tag ("error", message);
     xml.print_end_tag ("bingrep");
     s_close (xml.get_fd ());
     exit(-1);
}

void xml_regexp_die(char *message)
{
     xml.print_start_tag ("bingrep");
     xml.print_start_tag ("error");
     xml.print_data("Error in regular expression in Pattern attribute: ");
     xml.print_data(message);
     xml.print_data(".\n");
     xml.print_end_tag ("error");
     xml.print_end_tag ("bingrep");
     s_close (xml.get_fd ());
     exit(-1);
}

void xml_cdie(char *message)
{
     xml.print_end_tag ("data");
     xml.print_tag ("error", message);
     xml.print_end_tag ("bingrep");
     s_close (xml.get_fd ());
     exit(-1);
}

void xml_info ()
{
    xml.print_start_tag ("bingrep");
    xml.print_tag ("sourcebin", input_data_file);
    xml.print_tag ("regexp", pattern);
    xml.print_tag ("message", (all_msgs==0?msgname:"unspecified"));
    xml.print_tag ("host", (all_hosts==0?hostname:"unspecified"));
    xml.print_tag ("agent", (all_agents==0?agentname:"unspecified"));
    xml.print_tag ("instance", (all_instances==0?instancenm:"unspecified"));
    xml.print_tag ("attrs", (all_attrs==0?attr_search:"unspecified"));
    xml.print_start_tag ("data");
}
