/**
 * @brief Echelog BinInfo tool - Xml output  
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */                                                                             

#include "bininfo.h"

using namespace lib;
using namespace messages;

static class tMemPool pr_pool (2048);

/* memory pool for summary lists */
class tMemPool list_mp (4096);

/* statistics for all hosts together */
struct tHostSummary allhosts = { {0,0,0,0}, NULL, LONG_MAX, 0, 0, 0};

/* list for message and host summary */
class lib::tList msgsum (list_mp), hostsum (list_mp);

void print_overall();
void print_hosts();

/* finds item describing message with msg_type in list */
struct tMsgSummary *get_msg_item (int msg_type, class tList &msglist)
{
    class tNode const *item;

    /* try to find it in the list */
    FOR_ALL_LIST_ITEMS(item, &msglist)
    {
        struct tMsgSummary *ms =
               static_cast<struct tMsgSummary *> (item->data);

        if (ms->msg_type == msg_type)
                return ms;
    }

    /* we did not found it - so we must create it */
    struct tMsgSummary *ms = static_cast<struct tMsgSummary *>
                (list_mp.alloc (sizeof (struct tMsgSummary)));

    ms->msg_type = msg_type;
    ms->first_time = LONG_MAX;
    ms->last_time = 0;
    ms->space = 0;
    ms->count = 0;

    msglist.append (ms);

    return ms;
}

/* finds item describing host in list */
struct tHostSummary *get_host_item (byte *host)
{
    class tNode const *item;
    int i;

    /* try to find it in the list */
    FOR_ALL_LIST_ITEMS(item, &hostsum)
    {
        struct tHostSummary *hs =
               static_cast<struct tHostSummary *> (item->data);

        for (i=0;i<4;i++)
            if (hs->host[i] != host[i]) break;

        if (i==4)
                return hs;
    }

    /* we did not found it - so we must create it */
    struct tHostSummary *hs = static_cast<struct tHostSummary *>
                (list_mp.alloc (sizeof (struct tHostSummary)));

    hs->ms = new tList (list_mp);
    memcpy (hs->host, host, sizeof (byte[4]));
    hs->first_time = LONG_MAX;
    hs->last_time = 0;
    hs->space = 0;
    hs->count = 0;

    hostsum.append (hs);

    return hs;
}

void update_lists (class tMsgBuf &message, int size)
{

    /* get header information */
    class tMsgHeader *header =  message.get_header ();
    byte *hptr = header->get_host_id ();
    int msg_type = header->get_msg_type ();
    time_t msg_time = header->get_msg_time ();

    /* find item with this message statistics */
    struct tMsgSummary *ms = get_msg_item (msg_type, msgsum);

    /* update this message statistics */
    ms->first_time = min (ms->first_time, msg_time);
    ms->last_time = max (ms->last_time, msg_time);
    ms->space += size;
    ms->count ++;
    allhosts.first_time = min (allhosts.first_time, msg_time);
    allhosts.last_time = max (allhosts.last_time, msg_time);
    allhosts.space += size;
    allhosts.count ++;

    /* find item with this host */
    struct tHostSummary *hs = get_host_item (hptr);

    /* update this host statistics */
    hs->first_time = min (hs->first_time, msg_time);
    hs->last_time = max (hs->last_time, msg_time);
    hs->space += size;
    hs->count ++;

    /* find item with this message statistics for this host */
    struct tMsgSummary *msh = get_msg_item (msg_type, *(hs->ms));

    /* update this message statistics */
    msh->first_time = min (msh->first_time, msg_time);
    msh->last_time = max (msh->last_time, msg_time);
    msh->space += size;
    msh->count ++;

    /* free allocated structures */
    xfree (hptr);

}

void create_output ()
{ 
    char buffer[MAX_MESSAGE_SIZE];
    int buf_size;
    class tMsgBuf message;

    /* proceed all messages */
    for (int i=1;i<=static_cast<int>(inbin->get_last_id());i++)
    {

        /* get next message */
        buf_size = MAX_MESSAGE_SIZE;
        if (inbin->get_item(i, buffer, &buf_size))
            xml_die ("Error reading input BIN file.");

        /* Prepare the message buffer */
        if (message.prepare_to_read (buffer, buf_size))
            xml_die ("Message in input BIN is corrupted.");

        /* proceed update of statistics for  message */
        update_lists (message, buf_size);

        /* free allocated structures */
        message.reset();
    }


    /* write beginning of xml file */
    xml.print_start_tag ("bininfo");
    xml.print_tag ("sourcebin", input_data_file);

    print_overall();
    print_hosts();

    /* write end of xml file */
    xml.print_end_tag ("bininfo");
}

void print_message_stat(struct tMsgSummary *ms, long tspace, long tcount)
{
    /* get message name */
    int message_handle = gb_message_manager.get_handle_by_type (ms->msg_type);
    char *message_name = gb_message_manager.get_message_name_by_handle
                                                       (message_handle);

    /* time strings */
    char buf1[32], buf2[32];
    struct tm *tmp_tm;
    tmp_tm = localtime (&(ms->first_time));
    strftime(buf1, sizeof(buf1), "%Y-%m-%d %H:%M:%S", tmp_tm);
    tmp_tm = localtime (&(ms->last_time));
    strftime(buf2, sizeof(buf2), "%Y-%m-%d %H:%M:%S", tmp_tm);

    /* space, space frac, mesasges, msgfrac */
    char *spc, *spcf, *msgs, *msgsf;
    s_asprintf (&spc, pr_pool, "%d", ms->space );
    s_asprintf (&spcf, pr_pool, "%.2lf", (double)ms->space / tspace *100);
    s_asprintf (&msgs, pr_pool, "%d", ms->count );
    s_asprintf (&msgsf, pr_pool, "%.2lf", (double)ms->count / tcount *100);

    /* print the output */
    xml.print_start_tag ("message_stat");
    xml.print_tag ("messagename", message_name);
    xml.print_tag ("mintime", buf1);
    xml.print_tag ("maxtime", buf2); 
    xml.print_tag ("space", spc);
    xml.print_tag ("space_frac", spcf);
    xml.print_tag ("messages", msgs);
    xml.print_tag ("messages_frac", msgsf);
    xml.print_end_tag ("message_stat"); 

    pr_pool.free_pool();
}

void print_overall()
{
    class tNode const *item;

    xml.print_start_tag ("overall");
    
    /* printout all message stats in list */
    FOR_ALL_LIST_ITEMS(item, &msgsum)
    {
        struct tMsgSummary *ms =
               static_cast<struct tMsgSummary *> (item->data);

        print_message_stat(ms, allhosts.space, allhosts.count);
    }

    /* time strings */
    char buf1[32], buf2[32];
    struct tm *tmp_tm;
    tmp_tm = localtime (&(allhosts.first_time));
    strftime(buf1, sizeof(buf1), "%Y-%m-%d %H:%M:%S", tmp_tm);
    tmp_tm = localtime (&(allhosts.last_time));
    strftime(buf2, sizeof(buf2), "%Y-%m-%d %H:%M:%S", tmp_tm);

    /* space, space frac, mesasges, msgfrac */
    char *spc, *msgs;
    s_asprintf (&spc, pr_pool, "%d", allhosts.space );
    s_asprintf (&msgs, pr_pool, "%d", allhosts.count );

    /* print the output */
    xml.print_start_tag ("message_stat");
    xml.print_tag ("messagename", "Total");
    xml.print_tag ("mintime", buf1);
    xml.print_tag ("maxtime", buf2);
    xml.print_tag ("space", spc);
    xml.print_tag ("space_frac", "100");
    xml.print_tag ("messages", msgs);
    xml.print_tag ("messages_frac", "100");
    xml.print_end_tag ("message_stat");

    pr_pool.free_pool();
    
    xml.print_end_tag ("overall");
}

void print_host(struct tHostSummary *hs)
{
    class tNode const *item;
    char *ipstr;

    xml.print_start_tag ("host_stat");

    s_asprintf (&ipstr, "%d.%d.%d.%d", (unsigned int)(hs->host[0]),
                (unsigned int)(hs->host[1]), (unsigned int)(hs->host[2]),
                (unsigned int)(hs->host[3]));
    
    xml.print_tag ("hostname", ipstr);
    
    /* printout all message stats in list */
    FOR_ALL_LIST_ITEMS(item, hs->ms)
    {
        struct tMsgSummary *ms =
               static_cast<struct tMsgSummary *> (item->data);

        print_message_stat(ms, hs->space, hs->count);
    }

    /* time strings */
    char buf1[32], buf2[32];
    struct tm *tmp_tm;
    tmp_tm = localtime (&(hs->first_time));
    strftime(buf1, sizeof(buf1), "%Y-%m-%d %H:%M:%S", tmp_tm);
    tmp_tm = localtime (&(hs->last_time));
    strftime(buf2, sizeof(buf2), "%Y-%m-%d %H:%M:%S", tmp_tm);

    /* space, space frac, mesasges, msgfrac */
    char *spc, *msgs;
    s_asprintf (&spc, pr_pool, "%d", hs->space );
    s_asprintf (&msgs, pr_pool, "%d", hs->count );

    /* print the output */
    xml.print_start_tag ("message_stat");
    xml.print_tag ("messagename", "Total");
    xml.print_tag ("mintime", buf1);
    xml.print_tag ("maxtime", buf2);
    xml.print_tag ("space", spc);
    xml.print_tag ("space_frac", "100");
    xml.print_tag ("messages", msgs);
    xml.print_tag ("messages_frac", "100");
    xml.print_end_tag ("message_stat");

    pr_pool.free_pool();
    
    xml.print_end_tag ("host_stat");
}

void print_hosts()
{
    class tNode const *item;

    xml.print_start_tag ("hosts");

    /* printout all host stats in list */
    FOR_ALL_LIST_ITEMS(item, &hostsum)
    {
        struct tHostSummary *hs =
               static_cast<struct tHostSummary *> (item->data);

        print_host (hs);
    }

    xml.print_end_tag ("hosts");
}

void xml_die(char *message)
{
    xml.print_start_tag ("bininfo");
    xml.print_tag ("error", message);
    xml.print_end_tag ("bininfo");
    s_close (xml.get_fd ());
    exit(-1);
}
