/**
 * @brief Echelog Binary Storage reader 
 *
 * This agent reads data from binary storage and sends them 
 * to specified UNIX-domain socket.
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */                                                                             

#include "binreader.h"

#include <signal.h>

using namespace lib;
using namespace mframes;
using namespace spool;

char *data_file;
char *index_file;

/* Common miniagents' configuration structure */
struct cfitem tBinReader::s_binreader_section[]={
    {name: "BinReader",           type: CT_SECTION,       data: NULL},
    {name: "Datafile",            type: CT_STRING,        data: &data_file,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "Indexfile",           type: CT_STRING,        data: &index_file,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: NULL,                  type: CT_STOP,          data: NULL}
};

volatile bool tBinReader::s_sigpipe_received = false;

/* Create a miniagent */
tBinReader::tBinReader (tCmdLine& cmdline, const char* cfg_filename)
    : tEchelogModule (cmdline, cfg_filename)
{
    m_inited = false;
    out_f = NULL;
}

bool tBinReader::init ()
{
    ASSERT (!m_inited);

    out_f = new tUNIXSocketOFrame();
    if (out_f == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        goto clean;
    }

    if (tEchelogModule::init (out_f, NULL, out_f))
        goto outf_clean;

    m_inited = true;

    return false;

outf_clean:
    delete out_f;
    out_f = NULL;
clean:
    return true;
}


void tBinReader::cleanup ()
{
    ASSERT (m_inited);

    tEchelogModule::cleanup ();
    delete out_f;
    out_f = NULL;

    m_inited = false;
}


bool tBinReader::dynamic_init ()
{
    if (tEchelogModule::dynamic_init ())
        goto clean;

    m_bs = new tBinStore (1);
    if (!m_bs)
    {
        log (L_ERROR, "Not enough memory.");
        goto clean;
    }

    if (m_bs->open_r (data_file, index_file))
    {
        log (L_ERROR, "Cannot open BINSTORE files.");
        goto binstore_clean;
    }

    m_output = tOFrame::get_actual_oframe ()->get_output_by_name (NULL);
    if (m_output == NULL)
    {
        log (L_ERROR, "Missing output configuration.");
        goto bsclose_clean;
    }

    m_actual_item = 1;

    return false;

bsclose_clean:
    m_bs->close ();
binstore_clean:
    delete m_bs;
    m_bs = NULL;
clean:
    return true;
}


void tBinReader::dynamic_cleanup ()
{
    m_bs->close ();
    delete m_bs;
    m_bs = NULL;
    tEchelogModule::dynamic_cleanup ();
}


bool tBinReader::register_sections (lib::tMemPool& trash_pool)
{
    if (tEchelogModule::register_sections (trash_pool))
        return true;

    cf_register (s_binreader_section);
    return false;
}


void tBinReader::unregister_sections ()
{
    cf_unregister (s_binreader_section[0].name);
    tEchelogModule::unregister_sections ();
}


bool tBinReader::process_config (lib::tMemPool& conf_mp, lib::tMemPool& trash_pool)
{
    /* call ancestor's processing */
    if (tEchelogModule::process_config (conf_mp, trash_pool))
        return true; 

    data_file = s_strdup (data_file, conf_mp);
    index_file = s_strdup (index_file, conf_mp);

    return false;
}


void tBinReader::cleanup_config ()
{
    data_file = NULL;
    index_file = NULL;

    /* call ancestor's cleanup */
    tEchelogModule::cleanup_config ();
}


void tBinReader::install_sig_handlers ()
{
    tReconfigurableModule::install_sig_handlers ();

    set_signal_handler (SIGPIPE, sig_handling_f);
}


/* A trivial signal transfering function */
void tBinReader::sig_handling_f (int signum)
{
    switch (signum)
    {
        case SIGPIPE:
            s_sigpipe_received = true;
            break;
    }
}


tModule::SigHandlingRet tBinReader::handle_signals ()
{
    SigHandlingRet ret = tEchelogModule::handle_signals ();

    if (ret == SHR_BREAK_OK || ret == SHR_BREAK_ERR)
        return ret;

    if (s_sigpipe_received)
        ret = on_sigpipe ();

    return ret;
}


tModule::SigHandlingRet tBinReader::on_sigpipe ()
{
    log (L_WARN, "Broken output channel, exiting.");
    s_sigpipe_received = false;

    return SHR_BREAK_OK;
}


int tBinReader::process ()
{
    char buffer[MAX_MESSAGE_SIZE];
    int buf_size;

    ASSERT (m_inited);

    if (m_actual_item > m_bs->get_last_id())
        return PR_BREAK_OK;

    buf_size = MAX_MESSAGE_SIZE;
    if (m_bs->get_item (m_actual_item, buffer, &buf_size))
    {
         return PR_BREAK_ERR;
    }                

    m_actual_item++;

    if(m_output->send_buffer (buffer, buf_size))
    {
         log (L_WARN, "Error writing data to output socket.");
         return PR_BREAK_ERR;
    }

    out_f->process (0);

    return PR_CONTINUE;
}


/// Name of the file, where the 's configuration is written
const char* const cfg_filename = ECHELOG_CONF_DIR  "/common.conf";


/* --===  PROGRAM ENTRY POINT ===-- */
int main (int argc, char *argv[])
{
    mframes::tModule::tCmdLine cmdline(argc, argv);
    tBinReader binreader (cmdline, cfg_filename);

    /* init an agent */
    if (binreader.init ())
        return 1;

    int ret = binreader.run ();
    binreader.cleanup ();

    return ret;
}
