/**
 * @file
 * @brief Echelog Binary reader utility
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */                                                                             

#ifndef ECHELOG_BINREADER_H
#define ECHELOG_BINREADER_H

#include "mframes/echelogmodule.h"
#include "magents/msgframe.h"
#include "spool/binstore.h"

/** 
 * This class implements a typical miniagent's functionality.
 * The agent periodically reads informational chunks
 * from a supplied parser and forms them into messages which 
 * are passed to the next module. It also covers the basic miniagent's
 * configuration and reconfiguration.
 *
 * This class should not have multiple instances in the program.
 */
class tBinReader : public mframes::tEchelogModule
{
    public:

        /** Create a miniagent and bind the supplied parsing code
         *
         * @param mp
         * A mempool which can be used for internal magent's structures
         *
         * @param p
         * A parser which should be used to gain magent's information
         *
         * @param id
         * Miniagent's message identificator
         */
        tBinReader (tCmdLine& cmdline, const char* cfg_filename);

        bool init ();

        virtual void cleanup ();

    protected:
        /// Initialization reconfiguration dependent part
        virtual bool dynamic_init ();

        /// Reverse method for dynamic_init()
        virtual void dynamic_cleanup ();

        /// Installs signal handlers
        virtual void install_sig_handlers ();

        /// Handles installed signals
        virtual SigHandlingRet handle_signals ();

        /// Handles signal SIGPIPE
        virtual SigHandlingRet on_sigpipe ();

        /// Returns true if SIGPIPE has been catched
        inline bool sigpipe_received ()
        { return s_sigpipe_received; }

        virtual int process ();

        virtual bool register_sections (lib::tMemPool& trash_pool);

        virtual void unregister_sections ();

        virtual bool process_config (lib::tMemPool& conf_mp,
                lib::tMemPool& trash_pool);

        virtual void cleanup_config ();

        /// Miniagent's message frame
        class mframes::tUNIXSocketOFrame* out_f;

        /// Flag to set when SIGPIPE has been catched
        volatile static bool s_sigpipe_received;

        /// Miniagent's internal state
        bool m_inited;

        /// item to proccess
        unsigned long m_actual_item;

        //// binstore object
        class spool::tBinStore *m_bs;

        /// Pointer to output where msgs are passed to
        mframes::tOFrame::tOutput* m_output;

    private:
        //-------------------------------
        // Internal methods
        //-------------------------------

        /// A trivial signal transfering function
        static void sig_handling_f (int signum);

        //-------------------------------
        // Owned components
        //-------------------------------

        static struct lib::cfitem s_binreader_section[];
};

#endif

