/**
 * @brief Echelog BinSelect tool 
 *
 * This tool outputs values of specified attributes within given time
 * interval in format suitable for rrdtool. 
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */                                                                             

#include "binselect.h"

#define DEFAULT_CONFIG_FILE ECHELOG_CONF_DIR  "/common.conf"

using namespace lib;
using namespace messages;
using namespace spool;
using namespace mframes;

struct cfitem tBinSelectModule::s_binselect_section[]={
    {name: "BinSelect",          type: CT_SECTION,       data: NULL},
    {name: "InputDatafile",      type: CT_STRING,        data: &input_data_file,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "InputIndexfile",     type: CT_STRING,        data: &input_index_file,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "OutputFile",         type: CT_STRING,        data: &output_file,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "StartTime",          type: CT_STRING,        data: &starttime,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "EndTime",            type: CT_STRING,        data: &endtime,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "MsgName",            type: CT_STRING,        data: &msgname,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "Host",               type: CT_STRING,        data: &hostname,
varf: NULL, varf_reset: NULL },
    {name: "AgentName",          type: CT_STRING,        data: &agentname,
varf: NULL, varf_reset: NULL },
    {name: "Instance",           type: CT_STRING,        data: &instancename,
varf: NULL, varf_reset: NULL },
    {name: "OutputAttributes",   type: CT_STRING,        data: &out_attrs,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: NULL,                 type: CT_STOP,          data: NULL}
};

/* placeholders for parameters */
char *starttime, *endtime, *hostname, *agentname, *instancename;
char *input_data_file, *input_index_file, *output_file, *msgname, *out_attrs;

/* we will output contents of messages within this interval */
time_t start_time, end_time;

/* we will output contents only for this host */
byte *host;

/* we will output contents only of this message */
int msg_type;

/* instance of input bin */
class tBinStore *inbin;

/* file descriptor of output file */
int out_fd;

class tMemPool list_mp(2048);

/* list of attributes we will output */
class tList attrs_list(list_mp);

static class tMemPool pr_pool (2048);
static int firstentry=1;
static time_t min_time=LONG_MAX, max_time=0;

/* checks if this message holds interesting chunks */
int good_message(class tMsgHeader *header)
{
    byte *hptr = header->get_host_id ();
    char *aptr = header->get_agent_id ();
    char *iptr = header->get_agent_instance_id();
    int type = header->get_msg_type();
    time_t msg_time = header->get_msg_time ();

    /* test if this message lies within given time interval */
    if (msg_time <= start_time || msg_time >= end_time )
        return 0; 

    /* test if we are interested in this message */
    if (msg_type != type)
        return 0;

    /* test if we are interested in this agent */
    if (agentname && strcmp (agentname, aptr))
        return 0;

    /* test if we are interested in this instance */
    if (instancename && strcmp (instancename, aptr))
        return 0;
    
    /* test whether we search for this host */
    if (host)
    {
        for(int i=0;i<4;i++) 
            if (host[i]!=hptr[i]) return 0;

    } else 
    {
        host = static_cast<byte *> (list_mp.alloc (4 * sizeof (byte)));
        for (int i=0;i<4;i++)
            host[i] = hptr[i];
    }

    if (msg_time < min_time) min_time = msg_time;
    if (msg_time > max_time) max_time = msg_time;

    xfree (hptr);
    xfree (aptr);
    xfree (iptr);

    return 1;
}

/* print one attribute */
void print_attr (class tEchelogType &attr)
{
    int n = attr.get_num_of_components ();
    class tList list (pr_pool);

    /* simple type */
    if (n == 0 && !(attr.is_null ()))
    {
        attr.sprint (list, pr_pool, attr.get_plain_format());

        void *item = list.remove_first ();
        char *string = static_cast<char *> (item);

        if (!string)
            string = "";

        if (s_write (out_fd, string, strlen (string)) < 0)
            die ("Error writing in output file.");

    } else 
    {
        if (s_write (out_fd, "U", 1) < 0)
            die ("Error writing in output file.");
    }

}

void print_chunk (class tMsgChunk &chunk, char *timestr)
{
    class tNode const *item;
    char *buf;

    if (firstentry)
    {
        s_asprintf (&buf, "Statistics for %d.%d.%d.%d.\n", 
                    (unsigned int)host[0], (unsigned int)host[1],
                    (unsigned int)host[2], (unsigned int)host[3]);
        if (s_write (out_fd, buf, strlen(buf)) < 0)
            die ("Error writing in output file.");

        xfree (buf);

        s_asprintf (&buf, "%d %d\n", start_time, end_time);    
        if (s_write (out_fd, buf, strlen(buf)) < 0)
            die ("Error writing in output file.");

        xfree (buf);
 
        firstentry = 0;
    }
    
    if (s_write (out_fd, timestr, strlen (timestr)) < 0)
        die ("Error writing in output file.");

    FOR_ALL_LIST_ITEMS(item, &attrs_list)
    {
        int *attr_handle = static_cast<int *> (item->data);

        if (s_write (out_fd, ":", 1) < 0)
            die ("Error writing in output file.");

        print_attr (*(chunk.get_attr(*attr_handle)));
    }

    if (s_write (out_fd, "\n", 1) < 0)
        die ("Error writing in output file.");

    pr_pool.free_pool ();

}

void out_message (class tMsgBuf &message)
{
    class tMemPool chunk_mp (MAX_MESSAGE_SIZE);
    class tMsgChunk *chunk;
    char *timestr;

    /* get header information */
    class tMsgHeader *header =  message.get_header ();

    /* create time string for time of message */
    time_t msg_time = header->get_msg_time ();
    s_asprintf (&timestr, "%d", (unsigned int) msg_time);

    /* Process all chunks */
    for (int j = 0; j < header->get_subrecords (); j++)
    {
         /* create new chunk */
         chunk = new tMsgChunk (msg_type, chunk_mp);

         /* load the chunk into the structure */
         if (chunk->get (message.get_body ()) <= 0)
             die ("Message in temporary BIN is corrupted.");

         /* print out the specified attributes */ 
         print_chunk (*chunk, timestr);

    }

    /* free allocated structures */
    xfree (timestr);
}

void generate_output ()
{
    char buffer[MAX_MESSAGE_SIZE];
    int buf_size;
    char *buf;
    class tMsgBuf message;

    /* look through all messages*/
    for (int i=1;i<=static_cast<int>(inbin->get_last_id());i++)
    {

        /* get next message */
        buf_size = MAX_MESSAGE_SIZE;
        if (inbin->get_item(i, buffer, &buf_size))
            die ("Error reading input BIN file.");

        /* Prepare the message buffer */
        if (message.prepare_to_read (buffer, buf_size))
            die ("Message in input BIN is corrupted.");

        /* Look into messages header */
        tMsgHeader *header = message.get_header ();

        /* we are interested only in certain messages */
        if (!good_message (header)) 
        {
            message.reset();
            continue;
        }

        /* proceed output for message */
        out_message (message);

        /* free allocated structures */
        message.reset();

    }

    if (firstentry)
    {
        char *err_buf = "No data relevant data found. Cannot build graph.\n";
        if (s_write (out_fd, err_buf, strlen(err_buf)) < 0)
            die ("Error writing in output file.");
    } else 
    {
        s_asprintf (&buf, "%d %d\n", min_time, max_time);
        if (s_write (out_fd, buf, strlen(buf)) < 0)
            die ("Error writing in output file.");

        xfree (buf);
    }
    /* close log */
    s_close (out_fd);
}

bool tBinSelectModule::process ()
{
    generate_output ();

    return false;
}
        
bool tBinSelectModule::register_sections (lib::tMemPool& trash_pool)
{
    if (tSimpleModule::register_sections (trash_pool))
        return true;

    cf_register (s_binselect_section);

    return false;
}

void tBinSelectModule::unregister_sections ()
{
    tSimpleModule::unregister_sections ();

    cf_unregister (s_binselect_section[0].name);
}

bool tBinSelectModule::process_config (lib::tMemPool& conf_mp, lib::tMemPool&
                trash_pool)
{
    if (tSimpleModule::process_config (conf_mp, trash_pool))
        return true;

    /* create output file */
    if (!strcmp (output_file, "stdout"))
        out_fd = 1;
    else
        out_fd = s_open (output_file, O_WRONLY | O_CREAT | O_TRUNC,
                         S_IRUSR | S_IWUSR);
    if (out_fd < 0)
        die ("Can't create output file \"%s\".", output_file);

    /* get the message id */
    if (gb_message_manager.get_handle_by_name (msgname) < 0)
    {
        die("Parameter Message holds unknown message type.");
    }

    msg_type = gb_message_manager.get_message_type_by_handle
        (gb_message_manager.get_handle_by_name (msgname));

    /* check whether input data file exists */
    int tmp_fd = s_open (input_data_file, O_RDONLY);
    if (tmp_fd < 0)
        die ("Cannot open specified input BIN files.");
    s_close (tmp_fd);

    /* create input bin instance and open it for reading */
    inbin = new tBinStore(0);
    if (!inbin)
        die ("Error allocating memory for BIN object.");

    if (inbin->open_r (input_data_file, input_index_file))
        die ("Cannot open specified input BIN files.");

    /* read starting time */
    if ((start_time=parsetime (starttime)) < 0)
        die ("Error in Start Time parameter. Specified time is invalid.");

    /* read ending time */
    if ((end_time=parsetime (endtime)) < 0)
        die ("Error in End Time parameter. Specified time is invalid.");

    /* get host */
    host = NULL;
    if (hostname && strcmp (hostname, "*"))
    {
        class tList iplist(conf_mp);

        for (int i=0;i<(int)strlen(hostname);i++)
            if (hostname[i]=='.') hostname[i]=' ';

        if (s_strpsep (iplist, hostname, conf_mp) || iplist.count()!=4)
            die("Error in Host parameter. Invalid IP.");

        host = static_cast<byte *> (conf_mp.alloc (4 * sizeof (byte)));

        int err = 0;

        for (int i=0;i<4;i++)
        {
            char *octet = static_cast<char *> (iplist.remove_first ());
            long res;

            err = err || s_stol (octet, &res, 0, 255);
            host[i] = res;
        }

        if (err)
            die("Error in Host parameter. Invalid IP.");

    }

    /* get agent */
    if (agentname && !strcmp (agentname, "*"))
        agentname = NULL;
    else agentname = strdup (agentname);

    /* get instance */
    if (instancename && !strcmp (instancename, "*"))
        instancename = NULL;
    else instancename = strdup (instancename); 

    /* find which attributes we will output */
    if (out_attrs && strcmp (out_attrs, "*"))
    {
        class tNode const *item;
        class tList tmp_list (conf_mp);
        int msg_h = gb_message_manager.get_handle_by_name (msgname);

        if (s_strpsep (tmp_list, out_attrs, conf_mp))
            die ("Error in specified attributes for output.");

        FOR_ALL_LIST_ITEMS(item, &tmp_list)
        {
            char *attr_name = static_cast<char *> (item->data);
            int handle = gb_message_manager.get_handle_of_attr (msg_h, attr_name);
            int *attr_h = static_cast<int *> (conf_mp.alloc (sizeof(int)));

            if (handle < 0)
                die ("Unknown attribute name %s within message %s.",
                        attr_name, msgname);

            *attr_h = handle;
            attrs_list.append (attr_h);

        }

    }

    return false;

}

void tBinSelectModule::cleanup_config ()
{
    tSimpleModule::cleanup_config ();
}

/* --===  PROGRAM ENTRY POINT ===-- */
int main (int argc, char *argv[])
{
    tModule::tCmdLine cmdline (argc, argv);
    tBinSelectModule binselect (cmdline, DEFAULT_CONFIG_FILE);

    if (binselect.init ())
        return 1;

    int ret = binselect.run ();
    binselect.cleanup();
   
    return ret;
}

