/**
 * @brief Echelog BinSnapshot tool 
 *
 * This tool restores snapshot from incremental messages.
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */                                                                             

#include "binsnapshot.h"

#define DEFAULT_CONFIG_FILE ECHELOG_CONF_DIR  "/common.conf"

using namespace lib;
using namespace messages;
using namespace spool;
using namespace mframes;

static char *snaptime, *hostname, *agentname, *instancenm;

time_t snap_time, time_used;
char *agent, *instance, *msgname;
byte *host;
int msg_type;

class tBinStore *outbin, *inbin;
char *input_data_file, *input_index_file;
char tmp_data_file[] = "/tmp/binsnapshot_dataXXXXXX";
char tmp_index_file[] = "/tmp/binsnapshot_idxXXXXXX";
char *output_file;
class tXMLOutput xml;
int xml_fd;

struct cfitem tBinSnapshotModule::s_binsnapshot_section[]={
    {name: "BinSnapshot",        type: CT_SECTION,       data: NULL},
    {name: "InputDatafile",      type: CT_STRING,        data: &input_data_file,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "InputIndexfile",     type: CT_STRING,        data: &input_index_file,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "OutputFile",         type: CT_STRING,        data: &output_file,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "Time",               type: CT_STRING,        data: &snaptime,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "MsgName",            type: CT_STRING,        data: &msgname,
varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "Host",               type: CT_STRING,        data: &hostname,
varf: NULL, varf_reset: NULL },
    {name: "AgentName",          type: CT_STRING,        data: &agentname,
varf: NULL, varf_reset: NULL },
    {name: "Instance",           type: CT_STRING,        data: &instancenm,
varf: NULL, varf_reset: NULL },
    {name: NULL,                 type: CT_STOP,          data: NULL}
};

bool tBinSnapshotModule::process ()
{
    generate_output ();

    return false;
}
        
bool tBinSnapshotModule::register_sections (lib::tMemPool& trash_pool)
{
    if (tSimpleModule::register_sections (trash_pool))
        return true;

    cf_register (s_binsnapshot_section);

    return false;
}

void tBinSnapshotModule::unregister_sections ()
{
    tSimpleModule::unregister_sections ();

    cf_unregister (s_binsnapshot_section[0].name);
}

bool tBinSnapshotModule::process_config (lib::tMemPool& conf_mp, lib::tMemPool&
        trash_pool)
{
    if (tSimpleModule::process_config (conf_mp, trash_pool))
        return true;

    /* create output xml file */
    if (!strcmp (output_file, "stdout"))
        xml_fd = 1;
    else
        xml_fd = s_open (output_file, O_WRONLY | O_CREAT | O_TRUNC,
                         S_IRUSR | S_IWUSR);
    if (xml_fd < 0)
        die ("Can't create file \"%s\" for snapshot.", output_file);

    /* setup the xml output object */
    xml.set_fd (xml_fd);
    xml.set_die_on_err (true);

    /* output the xml header */
    xml.print_header ("binsnapshot.xsl");

    /* get the message id */
    if (gb_message_manager.get_handle_by_name (msgname) < 0)
    {
        xml_die("Parameter Message holds unknown message type.");
    }

    msg_type = gb_message_manager.get_message_type_by_handle
        (gb_message_manager.get_handle_by_name (msgname));

    /* check whether input data file exists */
    int tmp_fd = s_open (input_data_file, O_RDONLY);
    if (tmp_fd < 0)
        xml_die ("Cannot open specified input BIN files.");
    s_close (tmp_fd);

    /* create names of temporary files */
    int d_fd = s_mkstemp (tmp_data_file);
    int i_fd = s_mkstemp (tmp_index_file);

    if (d_fd < 0 || i_fd < 0)
        xml_die ("Cannot create temporary BIN files.");

    if (s_close (d_fd) || s_close (i_fd))
        xml_die ("Error manipulating with temporary BIN files.");

    /* create input bin instance and open it for reading */
    inbin = new tBinStore(0);
    if (!inbin)
        xml_die ("Error allocating memory for BIN object.");

    if (inbin->open_r (input_data_file, input_index_file))
        xml_die ("Cannot open specified input BIN files.");

    input_data_file = strdup (input_data_file);
    input_index_file =  strdup (input_index_file);
    msgname = strdup (msgname);

    /* create temporary bin instance and open it for writing */
    outbin = new tBinStore(0);
    if (!outbin)
        xml_die ("Error allocating memory for BIN object.");

    if (outbin->create (tmp_data_file, tmp_index_file))
        xml_die ("Cannot create temporary BINSTORE files.");

    /* read starting time */
    if ((snap_time=parsetime (snaptime)) < 0)
        xml_die ("Error in Time parameter. Specified time is invalid.");

    /* get host */
    host = NULL;
    if (hostname && strcmp (hostname, "*"))
    {
        class tList iplist(conf_mp);

        for (int i=0;i<(int)strlen(hostname);i++)
            if (hostname[i]=='.') hostname[i]=' ';

        if (s_strpsep (iplist, hostname, conf_mp) || iplist.count()!=4)
            xml_die("Error in Host parameter. Invalid IP.");

        host = static_cast<byte *> (conf_mp.alloc (4 * sizeof (byte)));

        int err = 0;

        for (int i=0;i<4;i++)
        {
            char *octet = static_cast<char *> (iplist.remove_first ());
            long res;

            err = err || s_stol (octet, &res, 0, 255);
            host[i] = res;
        }

        if (err)
            xml_die("Error in specified hosts. Invalid IP.");

    }

    /* get agent */
    agent = NULL;
    if (agentname && strcmp (agentname, "*")) agent = strdup (agentname);

    /* get instance */
    instance = NULL;
    if (instancenm && strcmp (instancenm, "*")) instance = strdup (instancenm);

    if (snap_time==0)
        xml_die ("Error in Time parameter. Cannot be \"start\".");
    if (snap_time==0x7FFFFFFF)
        xml_die ("Error in Time parameter. Cannot be \"end\".");

    return false;
}

void tBinSnapshotModule::cleanup_config ()
{
    tSimpleModule::cleanup_config ();
}

/* --===  PROGRAM ENTRY POINT ===-- */
int main (int argc, char *argv[])
{
    tModule::tCmdLine cmdline (argc, argv);
    tBinSnapshotModule binsnapshot (cmdline, DEFAULT_CONFIG_FILE);

    if (binsnapshot.init ())
        return 1;

    int ret = binsnapshot.run ();
    binsnapshot.cleanup();
   
    return ret;
}

