/**
 * @brief Echelog BinSnapshot tool - Xml output  
 *
 * This tool restores snapshot from incremental messages.
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */                                                                             

#include "binsnapshot.h"

using namespace lib;
using namespace messages;
using namespace spool;

static class tMemPool pr_pool (2048);
static time_t k_up, k_lw, delta, upper, nearest, aggregable;
static char buffer[MAX_MESSAGE_SIZE];
static int buf_size;
static tMemPool chunk_mp (MAX_MESSAGE_SIZE);
static tList *snap= new tList(chunk_mp);

struct ch_desc {
        int active;
        class tMsgChunk *chunk;
};

int good_message(class tMsgHeader *header);
int print_chunk (class tXMLOutput &axml, class tMsgChunk &chunk);

/* finds nearest message */
void find_nearest()
{
    int msgs = 0;
    class tMsgBuf message;

    delta = k_up = LONG_MAX;
    k_lw = 0;

    /* find nearest state message item */
    for (int i=1;i<=static_cast<int>(inbin->get_last_id());i++) 
    {

        /* get next message */
        buf_size = MAX_MESSAGE_SIZE;    
        if (inbin->get_item(i, buffer, &buf_size))
            xml_die ("Error reading input BIN file.");

        /* Prepare the message buffer */
        if (message.prepare_to_read (buffer, buf_size))
            xml_die ("Message in input BIN is corrupted.");

        /* Look into messages header */
        tMsgHeader *header = message.get_header ();

        /* we are interested only in certain messages */
        if (!good_message (header)) 
        {
            message.reset();
            continue;
        }

        int dtime = header->get_msg_time () - snap_time;
        msgs ++;

        /* if nearer, improve the result */
        if (dtime > 0)
        {
            if (header->get_flag_incremental ()) 
            {
               if (dtime < delta) 
               {
                   delta = dtime;
                   nearest = header->get_msg_time ();
               }
            }
               else k_up = min (k_up, header->get_msg_time ());
        } else
        {
            dtime = snap_time - header->get_msg_time ();
            if (header->get_flag_incremental ())
            {
               if (dtime < delta)
               {
                   delta = dtime;
                   nearest = header->get_msg_time ();
               }
            }
               else k_lw = max (k_lw, header->get_msg_time ());
        }

        message.reset();
        
    }

    /* did we found any matching message ? */
    if (!msgs)
        xml_die ("There is no message in the input BIN from which could "
                 "the snapshot restored.");

    /* which state will be used for restoration */
    if ((k_up - snap_time) < min (snap_time - k_lw, delta)) 
    {
        time_used = k_up;
        upper = 1;
    } else 
    {
        if (snap_time - k_lw < delta) nearest = k_lw;
        time_used = nearest;
        upper = 0;
    }         
    
    delta = min (snap_time - k_lw, delta);

    aggregable = gb_message_manager.is_aggregable
                      (gb_message_manager.get_handle_by_type (msg_type));

}

/* get nearest snapshot and delta frames */
void put_frames()
{
    int msgs = 0;
    class tMsgBuf message;

    /* extract relevant messages */
    for (int i=1;i<=static_cast<int>(inbin->get_last_id());i++) 
    {

        /* get next message */
        buf_size = MAX_MESSAGE_SIZE;    
        if (inbin->get_item(i, buffer, &buf_size))
            xml_die ("Error reading input BIN files.");

        /* Prepare the message buffer */
        if (message.prepare_to_read (buffer, buf_size))
            xml_die ("Message in input BIN is corrupted.");

        /* Look into messages header */
        tMsgHeader *header = message.get_header ();

        /* we are interested only in certain messages */
        if (!good_message (header)) 
        {
            message.reset();
            continue;
        }

        int msgtime = header->get_msg_time ();

        if (upper && !(header->get_flag_incremental()) && msgtime == k_up)
        {
              /* put the message into output bin */
              if (msgs == 0 && outbin->add_item (buffer, buf_size)==0)
                  xml_die ("Error writing to temporary BIN files.");
              msgs ++;
        }

        if (!upper && msgtime >= k_lw && msgtime - snap_time <= delta )
        {

              if (!(header->get_flag_incremental()))
                  if (msgs++ > 0) 
                  {
                      message.reset();
                      continue;
                  }
              
              /* put the message into output bin */
              if (outbin->add_item (buffer, buf_size)==0)
                  xml_die ("Error writing to temporary BIN files.");

        }

        message.reset();
    }

    /* close bin instances */
    if (outbin->close())
        xml_die ("Error closing temporary BIN files.");
    delete outbin;

    if (inbin->close())
        xml_die ("Error closing input BIN files.");
    delete inbin;

    /* ambiguous state ? */
    if (msgs!=1)
        xml_die ("There are more relevant messages in the BIN for "
                 "restoring of specified state. Specify agent/instance/host "
                 "fields more precisely that they will not mix.");

}

void generate_snapshot_no_agr()
{
    class tMsgBuf message;
    class tMsgChunk *chunk;
    int bytes;
    
    /* create input bin instance and open it for reading */
    inbin = new tBinStore(0);
    if (!inbin)
        xml_die ("Error allocating memory for temporary BIN object.");

    if (inbin->open_r (tmp_data_file, tmp_index_file))
        xml_die ("Cannot open specified temporary BIN files.");

    /* write the document beginning */
    xml_info ();

    /* writeout the snapshot */
    for (int i=1;i<=static_cast<int>(inbin->get_last_id());i++)
    {

        /* get next message */
        buf_size = MAX_MESSAGE_SIZE;
        if (inbin->get_item(i, buffer, &buf_size))
            xml_die ("Error reading temporary BIN file.");

        /* Prepare the message buffers */
        if (message.prepare_to_read (buffer, buf_size))
            xml_die ("Message in temporary BIN is corrupted.");

        int msg_subrecords = message.get_header()->get_subrecords ();

        /* Process all chunks */
        for (int j = 0; j < msg_subrecords; j++)
        {
             /* create new chunk */
             chunk = new tMsgChunk (msg_type, chunk_mp);

             /* output the chunk to xml file */
             if ((bytes = chunk->get (message.get_body ()))>0)
                 print_chunk (xml, *chunk);
             else
                 xml_die ("Message in temporary BIN is corrupted.");

             /* free allocated memory */
             delete chunk;
             chunk_mp.free_pool ();
        } 

        /* free message structures */
        message.reset();

    } 
  
    /* print end of document */
    xml.print_end_tag ("data");
    xml.print_end_tag ("binsnapshot");

    s_close (xml.get_fd ());

    /* cleanup */
    inbin -> close();
    delete inbin;

    /* destroy the temporary bin files */
    unlink (tmp_data_file);
    unlink (tmp_index_file);
    
}

/* update current list of chunks with the new chunk */
void add_chunk (class tMsgChunk *chunk)
{
    class tMsgChunk *c;
    class tNode const *item;
    struct ch_desc *cd = NULL;

    /* is there a chunk with the same key-attrs ? */
    FOR_ALL_LIST_ITEMS(item, snap)
    {
        cd = static_cast<struct ch_desc *> (item->data);
        c = cd-> chunk;
        if (c->key_match (*chunk)) break;
    }

    /* no there is not -> we must create item for it */
    if (!item)
    {
        cd = static_cast<struct ch_desc *> 
             (chunk_mp.alloc (sizeof (struct ch_desc)));
        snap->append (cd); 
        cd -> active = 0;
        cd -> chunk = chunk;
    }

    /* update chunk information */
    switch (chunk->get_status())
    {
        case ST_ITEM:
        case ST_ADDED:
              cd -> active ++;
              cd -> chunk = chunk;
              break;
        case ST_REMOVED:
              cd -> active --;
              break;
        case ST_CHANGED:
              cd -> chunk = chunk; 
        default:;
    }

    /* if the chunk was removed, relly remove it */
    if (cd -> active == 0) 
        snap->remove_match (cd);
}

void generate_snapshot()
{
    int bytes;
    class tMsgBuf message;
    class tMsgChunk *chunk;
    class tNode const *item;
    struct ch_desc *cd;

    /* create input bin instance and open it for reading */
    inbin = new tBinStore(0);
    if (!inbin)
        xml_die ("Error allocating memory for temporary BIN object.");

    if (inbin->open_r (tmp_data_file, tmp_index_file))
        xml_die ("Cannot open specified temporary BIN files.");

    /* write the document beginning */
    xml_info ();

    /* writeout the snapshot */
    for (int i=1;i<=static_cast<int>(inbin->get_last_id());i++)
    {
        /* get next message */
        buf_size = MAX_MESSAGE_SIZE;
        if (inbin->get_item(i, buffer, &buf_size))
            xml_die ("Error reading temporary BIN file.");

        /* Prepare the message buffers */
        if (message.prepare_to_read (buffer, buf_size))
            xml_die ("Message in temporary BIN is corrupted.");

        int msg_subrecords = message.get_header()->get_subrecords ();

        /* Process all chunks */
        for (int j = 0; j < msg_subrecords; j++)
        {
             /* create new chunk */
             chunk = new tMsgChunk (msg_type, chunk_mp);

             /* output the chunk to xml file */
             if ((bytes = chunk->get (message.get_body ()))>0)
                 add_chunk (chunk);
             else
                 xml_die ("Message in temporary BIN is corrupted.");

        }

        /* free message structures */
        message.reset();

    }

    /* snapshot writeout */
    FOR_ALL_LIST_ITEMS(item, snap)
    {
        cd = static_cast<struct ch_desc *> (item->data);
        if (cd->active > 0)
            print_chunk (xml, *(cd->chunk));
    }

    /* print end of document */
    xml.print_end_tag ("data");
    xml.print_end_tag ("binsnapshot");

    s_close (xml.get_fd ());

    /* cleanup */
    inbin -> close();
    delete inbin;

    /* destroy the temporary bin files */
    unlink (tmp_data_file);
    unlink (tmp_index_file);

}

void generate_output ()
{

    /* finds nearest messages */
    find_nearest();

    /* get nearest snapshot and delta frames */
    put_frames();

    if (upper || !aggregable)
        generate_snapshot_no_agr();
    else
        generate_snapshot();
}

/* checks if this message holds interesting chunks */
int good_message(class tMsgHeader *header)
{
    byte *hptr = header->get_host_id ();
    char *aptr = header->get_agent_id ();
    char *iptr = header->get_agent_instance_id();
    int type = header->get_msg_type();
    int i, ok = 1;

    if (host!=NULL)
        for(i=0;i<4;i++) if (host[i]!=hptr[i]) ok = 0;

    if (agent!=NULL && strcmp(agent, aptr)) ok = 0;

    if (msg_type>=0 && msg_type!=type) ok = 0;

    if (instance!=NULL && strcmp(instance, iptr)) ok = 0;

    xfree (hptr);
    xfree (aptr);
    xfree (iptr);

    return ok;
}

/* print one attribute */
int print_attr (class tXMLOutput &axml , class tEchelogType &attr, 
                const char *attr_name)
{
    int n = attr.get_num_of_components ();
    class tList list (pr_pool);
    int err = 0;
    
    /* simple type */
    if (n == 0)
    {
        attr.sprint (list, pr_pool, attr.get_plain_format());
        
        void *item = list.remove_first ();
        char *string = static_cast<char *> (item);

        if (!string)
            string = "";
        
        return axml.print_tag(attr_name, string);
    }
    else /* complex type */
    {
        err = min (err, axml.print_start_tag (attr_name));
        
        for (int i=0; i<n; i++)
        {
            err = min (err, print_attr (axml, *(attr.get_component_by_num (i)), 
                        attr.get_component_name_by_num (i)));
        }

        err = min (err, axml.print_end_tag (attr_name));

        return err;
    }
}

/* print chunk contents */
int print_chunk (class tXMLOutput &axml, class tMsgChunk &chunk)
{
     int message_handle = gb_message_manager.get_handle_by_type 
         (chunk.get_type ());
     char *mn = gb_message_manager.get_message_name_by_handle (message_handle);
     int num_of_attributes = chunk.get_attr_count();
     int err=0;

     /* print start of chunk */
     err = min (err, axml.print_start_tag (mn));

     /* print each attribute */
     for (int i = 0; i < num_of_attributes; i++)
     {
         char *name = gb_message_manager.get_attr_name (message_handle, i);
         err = min (err, print_attr (axml, *(chunk.get_attr(i)), name)); 
     }

     /* print end of chunk */
     err = min (err, axml.print_end_tag (mn));

     pr_pool.free_pool ();

     return err;
}

void xml_die(char *message)
{
     xml.print_start_tag ("binsnapshot");
     xml.print_tag ("error", message);
     xml.print_end_tag ("binsnapshot");
     s_close (xml.get_fd ());
     exit(-1);
}

void xml_info ()
{
    char *ipstr, buf1[32], buf2[32];
    struct tm *tmp_tm;


    tmp_tm = localtime (&snap_time);
    strftime(buf1, sizeof(buf1), "%Y-%m-%d %H:%M:%S", tmp_tm);
    tmp_tm = localtime (&time_used);
    strftime(buf2, sizeof(buf2), "%Y-%m-%d %H:%M:%S", tmp_tm);

    if (host)
        s_asprintf (&ipstr, "%d.%d.%d.%d", (unsigned int)host[0], 
                    (unsigned int)host[1], (unsigned int)host[2],
                    (unsigned int)host[3]);
    else ipstr = "unspecified";

    xml.print_start_tag ("binsnapshot");
    xml.print_tag ("messagename", msgname);
    xml.print_tag ("sourcebin", input_data_file);
    xml.print_tag ("host", ipstr);
    xml.print_tag ("agent", (agent?agent:"unspecified"));
    xml.print_tag ("instance", (instance?instance:"unspecified"));
    xml.print_tag ("reqtime", buf1);
    xml.print_tag ("usedtime", buf2);
    xml.print_start_tag ("data");
     
}
