/**
 * @file
 * @brief System command handling
 *
 * This is the only module which is started by an administrator. It reads
 * the echelog configuration and and accordingly to this file it maintains
 * all other modules.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */                                                                             

#include "echelogd.h"
#include <errno.h>


#ifndef DOXY_SKIP
using namespace lib;
using namespace socks;


// Flag whether echelog configuration needs a deeper contemplation
bool gb_echelog_sane;

// Flag whether the reconfiguration is in progress
bool gb_reconfiguration_in_progress;
#endif

/**
 * @brief System queue of delayed commands
 */
static tListQueue sys_delayed;

/**
 * @brief Standard system command queue
 */
static tListQueue sys_commands;


/**
 * @brief Replaces the command's cargo with the module id.
 *
 * Old command ancillary buffer is released, space for module id
 * is allocated on standard heap for consistence.
 */
static void trans_cmd (tModule &mod, socks::tCommand *c)
{
    ASSERT (c);

    if (c->d_len)
        xfree (c->d_ptr);

    c->d_len = sizeof (long);
    c->d_ptr = xmalloc (c->d_len);
    *(static_cast<long *>(c->d_ptr)) = mod.get_id();
}


/**
 * @brief Removes all commands from queue.
 *
 * All commands from the specifified queue are removed and allocated 
 * space on the standard heap is freed.
 */
static void empty_sys_command_queue (tListQueue &q)
{
    tCommand *c;

    while (!q.is_empty ())
    {
        c = static_cast<tCommand *> (q.remove ());
        free_cmd (c);
    }
}


/**
 * @brief Tries to set a new active configuration.
 *
 * The function tries to replace an active configuration
 * with a new one in a transaction manner. If the system is not willing
 * to accept the new configuration the old active configuration
 * is preserved and echelog modules are left untouched.
 * If the new configuration appears valid, the function tries to load
 * it and to merge in meaningful modules' definitions from the old one.
 *
 * The supplied parameter <tt>config_name</tt> specifies a new configuration
 * filename the user wants to be used or NULL if the configuration file should 
 * remain the same as before. However, <tt>config_name</tt> parameter is
 * ignored if the deamon is recovering after a failure. In that case, the
 * <tt>DEFAULT_CACHED_CONFIG</tt> file containing the parsed configuration
 * (saved before the system crash) is used. If the configuration failed 
 * the deamon would try to load the last last working configuration from 
 * file name written down to <tt>DEFAULT_LOCK_CF</tt> file before the crash.
 *
 * The supplied parameter <tt>log_name</tt> may be used to denote a new file
 * where all echelogd information should be logged. This parameter is always
 * used---regardles the fact whether the new configuration would be used or
 * not.
 *
 * @return
 * The function returns zero on success. On failure (on configuration file
 * missing or invalid for instance), -1 is returned.
 */
static int exec_sys_reconfigure (const char *config_name = NULL, 
        const char *log_name = NULL)
{
    tConfiguration *new_conf = new tConfiguration;
    tConfiguration *old_conf = gb_eenv.current_configuration; 

    debug_log ('1', "Trying to configure echelog");

    /* Don't allow another configuration switch if the previous one 
     * has not yet finished */
    gb_reconfiguration_in_progress = true;

    /* Set a new file to log in, if desired */
    if (log_name)
    {
        xfree (gb_eenv.log_name);
        gb_eenv.log_name = s_strdup (log_name);

        /* But at first, place a notice where the output is redirected 
         * in the old log file */
        log (L_INFO, "The daemon's output is redirected to '%s'.", 
                gb_eenv.log_name);

        /* Set a new log */
        log_file (gb_eenv.log_name);
    }

    if (!old_conf->get_after_recovery())
    {
        if (config_name)
            /* Accept the user's configuration file */
            new_conf->set_config_name (config_name);
        else
            /* Used the configuration from the previous session */
            new_conf->set_config_name (old_conf->get_config_name ());
    }
    else
        /* Something was wrong, try to recover from the last saved
         * configuration file */
        new_conf->set_config_name (DEFAULT_CACHED_CONFIG);

    log (L_INFO, "Using configuration file '%s'.", new_conf->get_config_name());

    /* Eureca, try to setup a new configuration */
    if (new_conf->load () < 0)
        goto invalid_configuration;
    if (new_conf->merge (old_conf) < 0)
        goto invalid_configuration;

    /* Configuration accepted */
    log (L_INFO, "Echelog configured, applying configuration.");

    if (old_conf->get_after_recovery ())
    {
        /* Retrieve the original file from which the recovery 
         * parsed configuration has been created and use it
         * for future configuration reloads
         */
        new_conf->set_config_name (old_conf->get_config_name ());
    }
    
    /* Distribute the accepted configuration to all echelog modules */
    new_conf->apply ();

    /* Discard the old stuff to avoid confusion */
    debug_log ('1', "Flushing command queues.");
    empty_sys_command_queue (sys_delayed);
    empty_sys_command_queue (sys_commands);

    /* Switch the configurations */
    gb_eenv.current_configuration = new_conf;
    gb_eenv.previous_configuration = old_conf;

    /* Prepare old configuration definitions for erasure */
    old_conf->prepare_for_wipeout ();

    /* Release all temporary configuration structures and report the result */
    get_cfpool().free_pool ();
    return 0;


   invalid_configuration:
    log (L_ERROR, "Invalid configuration, configuration not applied.");

    /* Use the configuration name from DEFAULT_LOCK_CF */
    cf_set_config_filename (old_conf->get_config_name ());

    if (old_conf->get_after_recovery ())
    {
        /* Discard the old stuff to avoid confusion */
        debug_log ('1', "Flushing command queues.");
        empty_sys_command_queue (sys_delayed);
        empty_sys_command_queue (sys_commands);

        /* Reset the recovery flag to avoid looping and schedule the
         * restart of the initialization with the configuration file */
        old_conf->set_after_recovery (false);
        init_run_queue (old_conf->get_config_name ());
    }

    /* Remove new definitions and release inactive modules */
    delete new_conf;

    /* Configuration is clean now, allow another configuration switch */
    gb_reconfiguration_in_progress = false;

    /* Release all temporary configuration structures and report the result */
    get_cfpool().free_pool ();
    return -1; 
}


/**
 * @brief Starts all echelog modules
 *
 * This function schedules chain starts of all top-level echelog modules.
 * The commands are put to delayed system queue.
 */
static void start_all_modules ()
{
    long i;
    struct tNode const *n;
    tConfiguration *conf = gb_eenv.current_configuration;

    debug_log ('1', "Starting echelog modules.");

    /* This function is called only on user's request. Force a configuration
     * update consideration for each module.
     */
    for (i = 0; i < conf->get_mod_count(); i++)
    {
        tModule *mod = conf->get_module (i);
        ASSERT (mod);

        mod->set_reload_config (true);
    }

    /* If there are any configured modules */
    if (conf->get_level_count() > 0)
    {
        /* Schedule the start of all top-level echelog modules */
        FOR_ALL_LIST_ITEMS (n, conf->get_level(0))
        {
            tModule *mod = static_cast<tModule *> (n->data);
            tCommand *c = create_simple_command (CMD_CHAIN_START);
            ASSERT (mod);
            
            /* Add the parameter which module is on the top of the chain */
            trans_cmd (*mod, c);
            sys_delayed.insert (c);
        }
    }
}


/**
 * @brief Stops all echelog modules.
 *
 * This function schedules a stop of each echelog module. The commands
 * are placed into modules' command queues. The actual execution 
 * is determined by modules' local conditions.
 */
static void exec_stop_all_modules ()
{
    tConfiguration *conf = gb_eenv.current_configuration;
    long l, level_count = conf->get_level_count ();
    struct tNode const *n;

    debug_log ('1', "Stopping echelog modules.");

    /* Stop all modules, start up from miniagents */
    for (l = level_count - 1; l >= 0; l--)
    {
        FOR_ALL_LIST_ITEMS (n, conf->get_level(l))
        {
            tModule *mod = static_cast<tModule *> (n->data);
            mod->enqueue_cmd (create_simple_command (CMD_STOP));
        }
    }
}


/**
 * @brief Stops echelog completely.
 *
 * At first all modules are scheduled for a stop. Once this step is finished,
 * the echelog is HALTed.
 */
static void exec_sys_stop ()
{
    exec_stop_all_modules ();

    debug_log ('1', "Scheduling delayed halt command.");
    tCommand *c = create_simple_command (CMD_SYS_HALT);
    sys_delayed.insert (static_cast<void *> (c));
}


/**
 * @brief Stops the chain.
 *
 * The whole chain under the module <tt>mod</tt> is scheduled 
 * for a stop. The modules should stopped bottom-to-up.
 */
static int exec_chain_stop (tModule &mod)
{
    mod.for_all_inputs (exec_chain_stop);
    mod.enqueue_cmd (create_simple_command (CMD_STOP));
    return 0;
}


/**
 * @brief Starts the chain.
 *
 * The whole chain under the module <tt>mod</tt> is to be started.
 * The modules should be started up-to-bottom.
 */
static int exec_chain_start (tModule &mod)
{
    mod.enqueue_cmd (create_simple_command (CMD_START));
    mod.for_all_inputs (exec_chain_start);
    return 0;
}


/**
 * @brief Checks if the module is ready to be started.
 *
 * This is an ancillary routine for check_chain_start.
 */
static int check_if_chain_module_is_startable (tModule &mod)
{
    if (mod.get_mod_status() == MOD_EMPTY)
        return 0;

    return 1;
}


/**
 * @brief Checks if the chain can be started.
 *
 * Checks if the conditions for a chain start under the module <tt>mod</tt>
 * are satisfied, not satisfied or yet unknown and return 1, -1 or 0
 * respectively.
 */
static int check_chain_start (tModule &mod)
{
    if (check_if_chain_module_is_startable (mod))
        return 1;

    /* The chain is startable if all dependent chains are startable */
    if (mod.for_all_inputs (check_chain_start))
        return 1;
    
    /* Wait till the module state stabilizes */
    return 0;
}


/**
 * @brief Checks if the module state is consistent.
 *
 * This is an ancillary routine called by check_cmd_conditions
 * for all modules the started module <tt>mod</tt> is dependent of.
 */
static int check_chain_start_prereq (tModule &mod)
{
    if (mod.get_mod_status() != MOD_OK)
        return 1;

    /* Wait till the module state stabilizes */
    return 0;
}


// Makes echelog system state more acceptable
void exec_sanitize_modules ()
{
    tConfiguration *conf = gb_eenv.current_configuration;
    long l, level_count = conf->get_level_count ();
    struct tNode const *n;
    tCommand *c;

    debug_log ('1', "Sanitizing modules");

    /* Traverse the whole DAG, layer by layer */
    for (l = 0; l < level_count; l++)
    {
        /* Go over all items in each layer */
        FOR_ALL_LIST_ITEMS (n, conf->get_level(l))
        {
            tModule *mod = static_cast<tModule *> (n->data);

            /* Check module states only */
            switch (mod->get_mod_status())
            {
                case MOD_NEEDS_STOP:
                    /* 
                     * Disallow further socket communication 
                     * and schedule a stop command.
                     */
                    mod->set_conn_status (CS_BUSY);
                    mod->empty_cmd_queue ();
                    mod->enqueue_cmd (create_simple_command (CMD_STOP));
                    break;
                
                case MOD_NEEDS_RESTART:
                case MOD_NEEDS_START:
                    /*
                     * At first go through the chain and fix the rest
                     * of module states to avoid undesirable actions
                     * happen for chained modules in higher levels.
                     */
                    mod->clear_chain_flags (mod->get_reload_config ());

                    /* Then schedule chain restart for the locally 
                     * top-level chain module.
                     */
                    debug_log ('1', "%s: scheduling chain restart",
                            mod->get_name());
                    c = create_simple_command (CMD_CHAIN_RESTART);
                    trans_cmd (*mod, c);
                    sys_delayed.insert (c);
                    break;

                default:
                    /* Not final or sane module state, take no action */
                    break;
            }
        }
    }

    /* All modules should be fixed for now */
    gb_echelog_sane = true;
}


/**
 * @brief Checks whether environment condition are suitable for a command
 * execution.
 *
 * This function is used to examine the delayed command <tt>cmd</tt> from the
 * delayed system command queue and decides how to handle the command.
 * 
 * @return
 * If the conditions seem to be suitable for commands execution, 1 is returned.
 * If it can't be determined right now 0 is returned. If the command execution
 * makes no sence or would cause problems, -1 is returned.
 */
static int check_cmd_conditions (tCommand *cmd)
{
    long i;
    tModule *mod;
    tConfiguration *conf = gb_eenv.current_configuration;

    ASSERT (cmd);

    /* Do care only about system commands */
    if (cmd_is_system (cmd->command))
    {
        switch (cmd->command)
        {
            case CMD_SYS_START:
            case CMD_SYS_HALT:
                /*
                 * All modules must be empty (i.e. not started or
                 * properly stopped), broken (i.e. probably unstartable)
                 * or might be not maintaned by echelogd.
                 */
                for (i = 0; i < conf->get_mod_count(); i++)
                {
                    mod = conf->get_module (i);
                    if (mod->get_mod_status() != MOD_EMPTY &&
                            mod->get_mod_status() != MOD_BROKEN &&
                            mod->get_mod_status() != MOD_NOT_MAINTAINED)
                        /* They're not, wait */
                        return 0;
                }
                /* Okay, everything steady, run the command */
                return 1;

            case CMD_CHAIN_START:
                 /* Retrieve the chain-top module */
                mod = gb_eenv.current_configuration->get_module 
                    (*static_cast<long *> (cmd->d_ptr));
                ASSERT (mod);
                
                /* We need all modules the local chain-top module
                 * is dependent of to be okay. */
                if (mod->for_all_outputs (check_chain_start_prereq))
                    /* Something went wrong, the chain can't be started */
                    return -1;
                else
                    /* Okay, check if all chain modules are ready
                     * to be started */
                    return !check_chain_start (*mod);

            default:
                /*
                 * We define no special conditions for other commands.
                 */
                return 1;
        }
    }

    /* Allow non-system command execution */
    return 1;
}

#ifndef DOXY_SKIP

// Checks delayed commands in the system queue.
void check_sys_delayed ()
{
    tCommand *dc;

    debug_log ('3', "Examining delayed commands");

    /* Ensure there are some delayed commands in the queue */
    while (!sys_delayed.is_empty ())
    {
        dc = static_cast<tCommand *> (sys_delayed.peek ());

        debug_log ('3', "Considering execution of system command %x", 
                dc->command);

        /* Check the evironment */
        switch (check_cmd_conditions (dc))
        {
            case -1:
                /*
                 * The command would execution would make no sence
                 * or it might even cause problems. Discard it.
                 */
                dc = static_cast<tCommand *> (sys_delayed.remove ());
                log (L_WARN, "Discarding command %x: conditions for "
                        "execution not satisfied.", dc->command);
                free_cmd (dc);
                break;
                
            case 0:
                /*
                 * The decision can't be made right now
                 */
                return;

            case 1:
                /* 
                 * Okay, we may safely move the command to system queue
                 */
                debug_log ('3', "Okay, moved to system command queue");
                dc = static_cast<tCommand *> (sys_delayed.remove ());
                sys_commands.insert (dc);
                break;

            default:
                ASSERT (false &&
                    "I have four words for you: I LOVE THIS COMPANY! YEAAAAH!");
        }

    }
}


// Processes commands from the command queue.
int exec_sys_commands ()
{
    char *config_name, *log_name;
    tModule *mod;
    tCommand *c;

    debug_log ('3', "Executing commands");

    /* Process all queued commands */
    while (!sys_commands.is_empty ())
    {
        /* Remove the command from the queue and take proper action */
        c = static_cast<tCommand *> (sys_commands.remove ());
        debug_log('2', "Executing command %x", c->command);

        switch (c->command)
        {
            case CMD_SYS_START:
                /* 
                 * At first, retrieve the two command parameters---a new 
                 * configuration filename and a new log filename
                 */
                log (L_INFO, "Starting echelog.");
                cmd_retrieve_strings (&config_name, &log_name, 
                        static_cast<char *> (c->d_ptr), c->d_len);
                exec_sys_reconfigure (config_name, log_name);
                break;

            case CMD_SYS_STOP:
                /* 
                 * Execute user's command to stop echelogd and all modules 
                 * */
                log (L_INFO, "Stopping echelog.");
                exec_sys_stop ();
                break;

            case CMD_SYS_RESTART_ALL:
                /* 
                 * Execute user's command restart all modules 
                 * */
                log (L_INFO, "Restarting echelog.");
                exec_stop_all_modules ();
                start_all_modules ();
                break;

            case CMD_SYS_RECONFIGURE:
                /* 
                 * At first, retrieve the two command parameters-- a new 
                 * configuration filename and a new log filename
                 */
                log (L_INFO, "Reconfiguring echelog.");
                cmd_retrieve_strings (&config_name, &log_name, 
                        static_cast<char *> (c->d_ptr), c->d_len);
                exec_sys_reconfigure (config_name, log_name);
                break;

            case CMD_SYS_HALT:
                /*
                 * System is stopped, do the final cleanup and halt echelogd.
                 */
                log (L_INFO, "Halting echelog.");
                gb_eenv.current_configuration->cleanup ();
                free_cmd (c);

                /* Signalize the deamon termination */
                return 1;

            case CMD_CHAIN_STOP:
                /*
                 * Stop the chain under the module mod
                 */
                mod = gb_eenv.current_configuration->get_module 
                    (*static_cast<long *> (c->d_ptr));
                ASSERT (mod);
                debug_log ('3', "%s: chain stop", mod->get_name ());
                exec_chain_stop (*mod);
                break;
                
            case CMD_CHAIN_START:
                /*
                 * Start the chain under the module mod
                 */
                mod = gb_eenv.current_configuration->get_module 
                    (*static_cast<long *> (c->d_ptr));
                ASSERT (mod);
                debug_log ('3', "%s: chain start", mod->get_name ());
                exec_chain_start (*mod);
                break;

            case CMD_CHAIN_RESTART:
                /*
                 * Retart the chain under the module mod
                 */
                mod = gb_eenv.current_configuration->get_module 
                    (*static_cast<long *> (c->d_ptr));
                ASSERT (mod);
                {
                     /* That means, stop and subsequentely run all modules
                     * in the chain */
                    tCommand *c1 = create_simple_command (CMD_CHAIN_STOP),
                             *c2 = create_simple_command (CMD_CHAIN_START);

                    debug_log ('1', "%s: chain restart", mod->get_name ());
                    
                    /* Add the parameter which module is on the top of 
                     * the chain. */
                    trans_cmd (*mod, c1);
                    trans_cmd (*mod, c2);

                    /* And schedule the chain restart execution */
                    sys_delayed.insert (c1);
                    sys_delayed.insert (c2);
                }
                break;
                
            default:
                /* What the ... */
                log (L_ERROR, "This program executed an invalid instruction "
                        "and will be aborted.");
                free_cmd (c);

                /* Signalize the deamon termination */
                return 1;
        }

        /* Release just executed command from memory */
        free_cmd (c);
    }

    /* I know the law, I gotta do what I gotta do... */
    return 0;
}


// Receives command from module's socket.
void exec_sys_receive_cmd (tModule &mod, tCommand *c)
{
    const char *modname;
    tModule *cmod;
    tCommand *s_rcmd;

    ASSERT (c);

    /* Check if the originating module has a right to exec the command */
    if (!mod.has_exec_capability (c->command))
    {
        log (L_WARN, "Permission denied to execute a privileged "
              "command for module %s. Discarding command.", mod.get_name ());
        return ;
    }

    debug_log ('3', "Received command (%x) from module %s", c->command, 
            mod.get_name ());


    /* Call an appropriate handler */
    switch (c->command)
    {
        case CMD_PING:
            /*
             * Ping received, reply with pong
             */
            s_rcmd = create_simple_command (CMD_PONG);
            mod.enqueue_cmd (s_rcmd);
            free_cmd (c);
            break;

        case CMD_PONG:
            /*
             * Pong received, update module status
             */
            mod.exec_receive_pong ();
            free_cmd (c);
            break;

        case CMD_NOP:
            /* 
             * No operation
             */
            free_cmd (c);
            break;

        case CMD_TOGGLE:
            /*
             * Toggle dataflow path at specified module
             */
            modname = static_cast<char *> (c->d_ptr);
            if (!(cmod = gb_eenv.current_configuration->get_module (modname)))
            {
                /* Undefined module (probably a misspelled module name */
                log (L_ERROR, "Invalid module '%s' specified in command "
                        "received from module %s.", modname, mod.get_name());
                free_cmd (c);
                break;
            }
            else
            {
                char *path_str, *endptr;

                 /* At first, retrieve the command parameter---a new path */
                cmd_retrieve_strings (&path_str, &endptr, 
                        static_cast<char *> (c->d_ptr), c->d_len);

                if (!path_str)
                {
                    /* Undefined path (probably a misspelled pathname) */
                    log (L_ERROR, "Invalid toggle command received from "
                            "module %s.", cmod->get_name());
                    free_cmd (c);
                    break;
                }

                /* Toggle the dataflow path at specified FlowSwitch module */
                cmod->exec_receive_toggle (path_str);
            }
            free_cmd (c);
            break;

        case CMD_RECONFIGURE:
            /*
             * Reconfigure specified module
             */
            modname = static_cast<char *> (c->d_ptr);
            if (!(cmod = gb_eenv.current_configuration->get_module (modname)))
            {
                /* Undefined module (probably a misspelled module name */
                log (L_ERROR, "Invalid module '%s' specified in command "
                        "received from module %s.", modname, mod.get_name());
                free_cmd (c);
                break;
            }
            /* This is an user's request---force the configuration update 
             * consideration */
            cmod->set_reload_config (true);
            cmod->enqueue_cmd (c);
            break;

        case CMD_CHAIN_RESTART:
            /*
             * Restart all modules under specified top-chain module
             */
            modname = static_cast<char *> (c->d_ptr);
            if (!(cmod = gb_eenv.current_configuration->get_module (modname)))
            {
                /* Undefined module (probably a misspelled module name */
                log (L_ERROR, "Invalid module '%s' specified in restart "
                        "command received from module %s.", modname, 
                        mod.get_name());
                free_cmd (c);
                break;
            }
            /* This is an user's request---force the configuration update 
             * consideration */
            cmod->set_reload_config (true);
            cmod->set_mod_status (MOD_NEEDS_RESTART);

            /* Many module might be restarted, force echelog sanitization */
            gb_echelog_sane = false;
            free_cmd (c);
            break;

        case CMD_SYS_START:
            /*
             * Schedule a system start
             */
            sys_delayed.insert (c);
            break;

        case CMD_SYS_STOP:
        case CMD_SYS_RESTART_ALL:
            /*
             * Stop/restart all maintained modules immediately.
             */
            sys_commands.insert (c);
            break;

        case CMD_SYS_RECONFIGURE:
            /*
             * Try to reconfigure echelogd
             */
            if (gb_reconfiguration_in_progress)
            {
                /* Refuse the requests until the old reconfiguration 
                 * is still in progress */
                log (L_WARN, "Reconfiguration is in progress; discarding "
                        "command.");
                free_cmd (c);
            }
            else
                /* Otherwise, schedule a system reconfiguration */
                sys_delayed.insert (c);
            break;

        default:
            /* What ... programed that module? */
            log (L_WARN, "Unknown or unexpected command %x received "
                    "from module %s; discarding command.", 
                    c->command, mod.get_name ());
            free_cmd (c);
    }
}


// Process all modules in currently active configuration.
void process_modules (tConfiguration *conf)
{
    long i;
    debug_log ('2', "Processing modules");

    ASSERT (conf);

    /* Process each standalone module */
    for (i = 0; i < conf->get_mod_count (); i++)
    {
        tModule *mod = conf->get_module (i);
        ASSERT (mod);

        mod->process ();
    }
}


// Checks old modules and takes proper actions.
void process_old_modules (tConfiguration *conf)
{
    long i;
    bool all_finished = true;
    ASSERT (conf);

    debug_log ('1', "Processing old modules (active: %d)", 
            conf->get_active_modules ());

    for (i = 0; i < conf->get_mod_count (); i++)
    {
        tModule *mod = conf->get_module (i);
        ASSERT (mod);

        /* Process each module still used in old configuration */
        if (mod->get_mod_used ())
        {
            debug_log ('1', "%s: PROCESSING OLD (mod_state: %d)", 
                    mod->get_name(), mod->get_mod_status());

            /* Check if all modules has been properly finished */
            if (mod->get_mod_status() != MOD_EMPTY)
            {
                all_finished = false;
                mod->process ();
            }
        }
    }

    if (all_finished)
    {
        /* There is no active module in the old configuration */
        debug_log ('1', "Purging previous configuration");

        /* Purge it completely */
        delete conf;
        gb_eenv.previous_configuration = NULL;
        gb_reconfiguration_in_progress = false;

        return;
    }
}


// Initializes the echelogd command queue for start.
void init_run_queue (const char *config_file)
{
    debug_log ('2', "Initializing the running queue.");

    tCommand *c = create_simple_command (CMD_SYS_START);

    /* Add the configuration filename into the command, if given */
    if (config_file)
    {
        c->d_len = max (cmd_store_strings ((char **)&c->d_ptr, "", config_file,
                    "", MAX_CMD_DATA_LENGTH), 0);
    }

    /* Put the command directly to the system command queue */
    sys_commands.insert (static_cast<void *> (c));
}

#endif
