/**
 * @file
 * @brief Echelogd Configuration
 *
 * This is the only module which is started by an administrator. It reads
 * the echelog configuration and and accordingly to this file it maintains
 * all other modules.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */                                                                             

#include "echelogd.h"
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>


#ifndef DOXY_SKIP
using namespace lib;
#endif


/**
 * @brief Predefined program names array
 */
static const char *predefined_progs[PREDEFINED_PROGS_NUM] = { "FlowSwitch" };

/**
 * @brief Predefine module names array
 */
static const char *predefined_mods[PREDEFINED_PROGS_NUM] = { "echelogctl" };

/**
 * @brief List containing program definitions
 */
static lib::tList prog_li (get_cfpool ());

/**
 * @brief List containing module definitions
 */
static lib::tList mod_li (get_cfpool ());

/**
 * @brief List contating dataflow definitions
 */
static lib::tList dflow_li (get_cfpool ());

/**
 * @brief Echelog programs configuration section
 */
static struct cfitem echelog_programs[] = {
    {name: "Programs",            type: CT_SECTION,       data: NULL},
    {name: "Define",              type: CT_LIST,          data: &prog_li,
varf: NULL, varf_reset: NULL, flags: 0 },
    {name: NULL,                  type: CT_STOP,          data: NULL}
};

/**
 * @brief Echelog modules configuration section
 */
static struct cfitem echelog_modules[] = {
    {name: "Modules",             type: CT_SECTION,       data: NULL},
    {name: "Module",              type: CT_LIST,          data: &mod_li,
varf: NULL, varf_reset: NULL, flags: 0 },
    {name: NULL,                  type: CT_STOP,          data: NULL}
};

/**
 * @brief Echelog dataflow configuration section
 */
static struct cfitem echelog_dataflow[] = {
    {name: "Dataflow",            type: CT_SECTION,       data: NULL},
    {name: "Add",                 type: CT_LIST,          data: &dflow_li,
varf: NULL, varf_reset: NULL, flags: 0 },
    {name: NULL,                  type: CT_STOP,          data: NULL}
};


/**
 * @brief Simple functor for program name comparison
 *
 * Returns zero if the pointers refer to the same program, 
 * non-zero otherwise.
 */
static int prog_name_compare (const void *a, const void *b)
{
    const tProgramInfo * x = static_cast<const tProgramInfo *>
        (*static_cast<tProgramInfo * const *> (a));
    const tProgramInfo * y = static_cast<const tProgramInfo *>
        (*static_cast<tProgramInfo * const *> (b));

    return strcmp (x->name, y->name);
}


/**
 * @brief Simple functor for module name comparison
 *
 * Returns zero if the pointers refer to the same module, 
 * non-zero otherwise.
 */
static int mod_name_compare (const void *a, const void *b)
{
    const tModule * x = static_cast<const tModule *>
        (*static_cast<tModule * const *> (a));
    const tModule * y = static_cast<const tModule *>
        (*static_cast<tModule * const *> (b));

    return strcmp (x->get_name(), y->get_name());
}


/**
 * @brief Simple functor for integer comparison
 *
 * Returns zero if the pointers refer to the same long interger value, 
 * non-zero otherwise.
 */
static int ptr_to_long_compare (const void *a, const void *b)
{
    const long x = *static_cast<const long *>(a);
    const long y = *static_cast<const long *>(b);

    return x - y;
}


#ifndef DOXY_SKIP
// Checks the cross-dependencies between configuration sections
int tConfiguration::check_definitions () const
{
    long i;

    /* Check if all defined modules are used. In a single configuration
     * this would make no difference but it would require some more complex
     * checking later (on reconfiguration for example). */
    for (i = 0; i < mod_count; i++)
        if (!modules[i]->mod_used)
        {
            log (L_ERROR, "Module %s introduced but ill-defined "
                    "in dataflow section.", modules[i]->name);
            return -1;
        }

    return 0;
}


// Topologically sorts all configured modules.
int tConfiguration::topologically_sort ()
{
    long i, processed = 0, level = 0;
    bool found;

    /* Allocate an array of linear lists, one for each module */
    tList **l = static_cast<tList **>(get_cfpool().alloc 
            (sizeof (tList *) * mod_count));

    for (i = 0; i < mod_count; i++)
    {
        tNode const *n;

        /* Allocate a list */
        tList *ap = new tList (get_cfpool ());
        if (!ap)
            return -1;

        l[i] = ap;
        
        /* Append all module outputs to its linear dependency list */
        FOR_ALL_LIST_ITEMS (n, &modules[i]->outputs)
        {
            long *item = static_cast<long *>(get_cfpool().alloc(sizeof (long)));

            *item = (static_cast<tModule *>(n->data))->id;
            ap->append (item);
        }
    }


    /* Try to topologically sort the DAG */
    do {
        found = false;

        /* Find all modules independent of others */
        for (i = 0; i < mod_count; i++)
            if (l[i] && !l[i]->count ())
                modules[i]->level = level;

        /* For all modules */
        for (i = 0; i < mod_count; i++)
        {
            /* Remove such modules */
            if (modules[i]->level == level)
            {
                tNode const *n;

                /* From dependency lists of other modules */
                FOR_ALL_LIST_ITEMS (n, &modules[i]->inputs)
                {
                    long *ref = static_cast<long *>
                        (get_cfpool().alloc (sizeof (long)));
                    *ref = i;
                    
                    void * ptr = 
                        l[static_cast<tModule *>(n->data)->id]
                        ->remove_match (ref, ptr_to_long_compare);

                    ASSERT (ptr);
                }

                /* Remove the list and mark it as done */
                delete l[i];
                l[i] = NULL;

                /* Advance the processed modules counter
                 * and loop the algorithm */
                processed++;
                found = true;
            }
        }
            
        /* If there is a new level, advance the level counter */
        if (found)
            level++;
    }
    while (found);


    /* Check if all modules has been processed */
    if (processed != mod_count)
    {
        /* Obviously not, but then the algorithm hasn't found
         * any independent modules. The graph can't be a DAG. */
        log (L_ERROR, "Configuration parse error. "
                "The flowgraph contains a cycle.");
        for (i = 0; i < mod_count; i++)
            if (l[i])
                delete l[i];

        /* Report failure */
        return -1;
    }

    /* The constructors which should be called on each array item are often
     * problematic but I hope this is rather straightforward and the
     * compiler won't mess it up. */
    level_count = level;
    levels = new tList* [level_count];
    if (!levels)
        return -1;

    for (i = 0; i < level_count; i++)
	 levels[i] = new tList(pool);

    /* Make a list of all modules in each level. The level with the zero number
     * contains modules which can be started immediately. */
    for (i = 0; i < mod_count; i++)
    {
        tModule *mod = modules[i];

        if (mod->mod_status == MOD_NOT_MAINTAINED)
            continue;

        levels[mod->level]->append (mod);
    }

    debug_log ('2', "Dependencies scanned.");

    /* Report success */
    return 0;
}


// Configures and propagates modules' input nodes
int tConfiguration::config_input_nodes ()
{
    long l;
    tNode const *n;
    char *s;


    /* At first, configure the input nodes from the dependency-free level */
    FOR_ALL_LIST_ITEMS (n, (levels[0]))
    {
        tModule *mod =  static_cast<tModule *> (n->data);

        /* It makes no sence to put a flowswitch at the end of a chain */
        if (mod->mod_role != MOD_ROLE_NORMAL)
        {
            log (L_ERROR, "End-chain echelog module %s is of "
                    "an unsupported type.", mod->name);
            return -1;
        }

        /* Derive the input socket name from the module name */
        s = s_strconcat (get_cfpool (), ECHELOG_SOCK_DIR, "/", mod->name, NULL);
        mod->input_sock_name = s_strnreplace (s, "\"", "\\\"", 
                MAX_CONF_NAME_LEN);
    }

    /* Then configure modules in all subsequent levels */
    for (l = 1; l < level_count - 1; l++)
        FOR_ALL_LIST_ITEMS (n, (levels[l]))
        {
            tModule *mod =  static_cast<tModule *> (n->data);

            if (mod->mod_role == MOD_ROLE_FLOWSWITCH)
            {
                if (l == level_count - 1 || mod->outputs.count() == 0)
                {
                    /* It makes no sence to put a flowswitch at the end of 
                     * a chain */
                    log (L_ERROR, "End-chain echelog module %s is of "
                            "an unsupported type.", mod->name);
                    return -1;
                }

                /* Let each flowswitch default to the first (the most left)
                 * dataflow path */
                tNode const *c = mod->outputs.get_first ();
                tModule *pmod = static_cast<tModule *> (c->data);
                ASSERT (pmod && pmod->input_sock_name);

                /* Set the default */
                mod->input_sock_name = s_strdup (pmod->input_sock_name);
            }
            else
            {
                /* This is not a flowswitch---it makes its own input socket */
                s = s_strconcat (get_cfpool(), ECHELOG_SOCK_DIR, "/", 
                        mod->name, NULL);
                mod->input_sock_name = s_strnreplace (s, "\"", "\\\"",
                        MAX_CONF_NAME_LEN);
            }
        }

    debug_log ('2', "Chain connection points configured.");

    /* Report success */
    return 0;
}


// Parses and defines configured programs
int tConfiguration::parse_programs (lib::tList &list)
{
    long i;
    tList p (get_cfpool ());

    /* Allocate enough space for all program definitions---predefined 
     * programs + program section (one definition per line) */
    prog_count = list.count() + PREDEFINED_PROGS_NUM;
    programs = static_cast<tProgramInfo **> 
        (xrealloc (programs, prog_count * sizeof (tProgramInfo *)));

    /* Parse all program definition lines */
    for (i = PREDEFINED_PROGS_NUM; i < prog_count; i++)
    {
        /* Get the configuration line */
        char *str = static_cast<char *> (list.remove_first ());

        /* Split it to words and phrases */
        if ((s_strpsep (p, str, get_cfpool()) < 0) || p.count() != 2)
        {
            log (L_ERROR, "Configuration parse error. Invalid program "
                    "definition '%s'.", str);
            return -1;
        }

        /* Allocate a new program structure */
        programs[i] = static_cast<tProgramInfo *> 
                    (pool.alloc (sizeof (tProgramInfo)));

        /* Retrieve the program name and the pathname */
        const char *progname = const_cast<const char*>
            (static_cast<char *> (p.remove_first ()));
        const char *progpath = const_cast<const char*>
            (static_cast<char *>(p.remove_first()));
        if (!strlen (progname) || !strlen (progpath))
        {
            log (L_ERROR, "Configuration parse error. Invalid program "
                    "definition '%s'.", str);
            return -1;
        }

        /* Save their copies to the program structure */
        programs[i]->name = s_strdup (progname, pool);

        if (progpath[0] == '/')
            programs[i]->program_path = s_strdup (progpath, pool);
        else
            programs[i]->program_path = s_strconcat (pool, ECHELOG_BIN_DIR,
                                                     "/", progpath, NULL);

    }

    /* Sort all defined programs to allow faster searching by name */
    qsort (programs, prog_count, sizeof (tProgramInfo *), prog_name_compare);

    /* Check program definitions for duplicates (it's trivial on a sorted
     * list ;) */
    for (i = 1; i < prog_count; i++)
    {
        if (!strcmp (programs[i-1]->name, programs[i]->name))
        {
            log (L_ERROR, "Invalid program definition. Program '%s' defined "
                    "more than once.", programs[i]->name);
            return -1;
        }
    }

    debug_log ('2', "DEFINED PROGRAMS:");

    /* Check if all binaries exist and if they are executable */
    for (i = 0; i < prog_count; i++)
    {
        tProgramInfo *pi = programs[i];
        debug_log('2', "%s (%s)", pi->name, pi->program_path);

        if (pi->program_path && s_eaccess (pi->program_path, X_OK | F_OK))
        {
            log (L_ERROR, "Can't access and execute binary '%s'.", 
                    pi->program_path);
            return -1;
        }
    }

    /* Report success */
    return 0;
}


// Parses and defines configured modules
int tConfiguration::parse_modules (lib::tList &list)
{
    long i;
    tList p (get_cfpool ());

    /* Allocate enough space for all module definitions---predefined 
     * modules + module section (one definition per line) */
    mod_count = list.count() + PREDEFINED_MODS_NUM; 
    modules = static_cast<tModule **> 
        (xrealloc (modules, mod_count * sizeof (tModule *)));
    
    /* Clear out the allocated array. If the function is aborted
     * only non-NULL module will be deleted. */
    for (i = PREDEFINED_MODS_NUM; i < mod_count; i++)
        modules[i] = NULL;

    /* Parse all module definition lines */
    for (i = PREDEFINED_MODS_NUM; i < mod_count; i++)
    {
        tProgramInfo *mod_prog;
        tModuleRole role;
        const char *progname, *modname, *configname, *config;

        /* Get the configuration line */
        char *str = static_cast<char *> (list.remove_first ());

        /* Split it to words and phrases */
        if ((s_strpsep (p, str, get_cfpool ()) < 0) || p.count() != 3)
        {
            log (L_ERROR, "Configuration parse error. Invalid module "
                    "definition '%s'.", str);
            return -1;
        }

        /* Retrieve the module name */
        modname = const_cast<const char*>
            (static_cast<char *> (p.remove_first ()));
        if (!strlen (modname))
        {
            log (L_ERROR, "Configuration parse error. Invalid module "
                    "definition '%s'.", str);
            return -1;
        }

        /* Retrieve the module's binary name */
        progname = const_cast<const char*>
            (static_cast<char *> (p.remove_first ()));
        mod_prog = get_program  (progname);
        if (!mod_prog)
        {
            log (L_ERROR, "Configuration parse error. Program '%s' "
                    "is used but not defined in configuration.", progname);
            return -1;
        }

        if (!strcmp (progname, predefined_progs[0]))
            /* The first predefined program is a flowswitch */
            role = MOD_ROLE_FLOWSWITCH;
        else
            /* A users module, otherwise */
            role = MOD_ROLE_NORMAL;

        /* Retrieve the configuration name */
        configname = const_cast<const char*>
            (static_cast<char *> (p.remove_first ()));
        if (!configname || !strlen (configname))
        {
            log (L_ERROR, "Configuration parse error. Module '%s' "
                    "has no configuration file.", modules[i]->name);
            return -1;
        }

        if (configname[0] == '/')
            config = s_strdup (configname, pool);
        else
            config = s_strconcat (get_cfpool(), ECHELOG_CONF_DIR, "/",
                    configname, NULL);
        
        /* Allocate a new module and save parsed information to it */
        modules[i] = new tModule (pool, *this, modname, config);
        if (!modules[i])
            return -1;

        modules[i]->binary = mod_prog->program_path;
        modules[i]->mod_role = role;
        modules[i]->capability = CP_NORMAL;
    }

    /* Sort all defined modules to allow faster searching by name */
    qsort (modules, mod_count, sizeof (tModule *), mod_name_compare);

    debug_log ('2', "DEFINED MODULES:");
    
    /* Set new handlers for all defined modules */
    for (i = 0; i < mod_count; i++)
    {
        modules[i]->id = i;
        debug_log ('2', "%s (%s)", modules[i]->name, modules[i]->config);
    }

    /* Check module definitions for duplicates (it's trivial on a sorted
     * list ;) */
    for (i = 1; i < mod_count; i++)
    {
        if (!strcmp (modules[i-1]->name, modules[i]->name))
        {
            log (L_ERROR, "Invalid module definition. Module '%s' defined "
                    "more than once.", modules[i]->name);
            return -1;
        }

        /* Echelog control module is always the first predefined one */
        if (!strcmp (modules[i]->name, predefined_mods[0]))
            ctrl_mod_id = i;
    }

    /* Report success */
    return 0;
}


// Parses the dataflow configuration sections and define 
// the dependencies between echelog modules.
int tConfiguration::parse_dataflow (lib::tList &list)
{
    long i, dflow_lines_count = list.count ();
    tList p (get_cfpool ());


    /* Parse all dataflow definition lines */
    for (i = 0; i < dflow_lines_count; i++)
    {
        const char *modname, *delim;
        tModule *mod;

        /* Get the configuration line */
        char *str = static_cast<char *> (list.remove_first ());

        /* Split it to words and phrases */
        if ((s_strpsep (p, str, get_cfpool()) < 0) || 
                (!p.count()) || (p.count() == 2))
        {
            log (L_ERROR, "Configuration parse error. Invalid dataflow "
                    "entry definition line '%s'.", str);
            return -1;
        }

        /* Retrieve the module name */
        modname = const_cast<const char*>
                    (static_cast<char *> (p.remove_first ()));
        
        /* Check if such module exists */
        if (!(mod = get_module (modname)))
        {
            log (L_ERROR, "Configuration parse error. Module '%s' "
                    "is not defined.", modname);
            return -1;
        }
        else
            /* Activate the module */
            mod->mod_used = true;

        if (!p.count())
            /* The module has no output to other echelog modules */
            continue;
            
        /* Read and check the delimiter */
        delim = const_cast<const char*>(static_cast<char *>(p.remove_first ()));
        if (strcmp (delim, "->"))
        {
            log (L_ERROR, "Configuration parse error. '->' expected '%s' "
                    "found instead.", delim);
            return -1;
        }

        /* Check if there is at least one output module after '->' */
        if (!p.count ())
        {
            log (L_ERROR, "Configuration parse error. No modules found "
                    "after '->'.");
            return -1;
        }

        /* For all outputs */
        while (p.count() > 0)
        {
            tModule *omod;
            
            /* Check if the referenced modules exists */
            omod = get_module (const_cast<const char*>
                    (static_cast<char *>(p.remove_first())));
            if (!omod)
                return -1;

            /* And build up the local dependency trees */
            mod->outputs.append (omod);
            omod->inputs.append (mod);
        }
    }

    debug_log ('2', "Dataflow configured");

    /* Report success */
    return 0;
}


// Creates a configuration with predefined programs and modules only.
tConfiguration::tConfiguration ()
    : pool (4096), 
    levels (NULL), 
    active_modules (0), 
    after_recovery (false),
    config_name (NULL),
    prog_count (PREDEFINED_PROGS_NUM),
    mod_count (PREDEFINED_MODS_NUM),
    level_count(0)
{
    /* Allocate predefined program and module structures */
    programs = static_cast<tProgramInfo **> 
        (xmalloc (PREDEFINED_PROGS_NUM * sizeof (tProgramInfo *)));
    modules = static_cast<tModule **> 
        (xmalloc (PREDEFINED_MODS_NUM * sizeof (tModule *)));

    /* Add the 'flowswitch' pseudo program */
    programs[0] = static_cast<tProgramInfo *> 
        (pool.alloc (sizeof (tProgramInfo)));    
    programs[0]->name = s_strdup (predefined_progs[0], pool);
    programs[0]->program_path = NULL;

    /* The control module is somewhat special and is always defined */
    modules[0] = new tModule (pool, *this, predefined_mods[0], NULL);
    if (!modules[0])
        die ("Not enough memory.");

    modules[0]->capability = CP_SYS_COMMANDS;
    modules[0]->mod_status = MOD_NOT_MAINTAINED;
    modules[0]->mod_used = true;
    modules[0]->mod_role = MOD_ROLE_CONTROL;
    modules[0]->id = 0;

    /* Setup the control module */
    ctrl_mod_id = modules[0]->id;

    /* Create a reference dummy module */
    rfmod = new tModule (pool, *this, "", NULL);
    if (!rfmod)
        die ("Not enough memory.");
}


// Releases all used structures and forces inactive modules cleanups.
tConfiguration::~tConfiguration ()
{
    long i;
    
    /* Release all inactive modules */
    for (i = 0; i < mod_count; i++)
        if (modules[i])
            delete modules[i];

    if (levels)
        delete[] levels;

    /* And releases used heap space */
    pool.free_pool ();
    xfree (modules);
    xfree (programs);
}


#ifndef DOXY_SKIP
// Implements a property
const char *tConfiguration::get_config_name () const
{
    return config_name;
}


// Implements a property
void tConfiguration::set_config_name (const char *new_config_name)
{
    char *fc = const_cast<char *> (config_name);
    config_name = s_strdup (new_config_name);
    xfree (fc);
}
#endif


// Loads a new configuration
int tConfiguration::load ()
{
    int r = 0;

    /* register section for echelog system configuration */
    cf_register (echelog_programs);
    cf_register (echelog_modules);
    cf_register (echelog_dataflow);

    /* Read configuration file */
    cf_cleanup_registered_sections ();
    cf_clear_loaded_flags();

#define TRY_TO(x) if ((x)) { r = -1; goto load_cleanup; }    

    TRY_TO (cf_reread (config_name, get_cfpool()));
    TRY_TO (cf_check_required());
    TRY_TO (parse_programs (prog_li));
    TRY_TO (parse_modules (mod_li));
    TRY_TO (parse_dataflow (dflow_li));
    TRY_TO (check_definitions ());
    TRY_TO (topologically_sort ());
    TRY_TO (config_input_nodes ());

#undef TRY_TO

   load_cleanup:
    cf_unregister (echelog_programs[0].name);
    cf_unregister (echelog_modules[0].name);
    cf_unregister (echelog_dataflow[0].name);

    return r;
}


// Merges old configuration definitions with the newly loaded configuration
int tConfiguration::merge (tConfiguration *old_conf)
{
    long i, r = 0;

    /* For all newly defined modules */
    for (i = 0; i < get_mod_count(); i++)
    {
        tModule *mod = get_module (i);

        /* Try to retrieve an old module with the same name */
        tModule *old_mod = old_conf->get_module (mod->name);

        if (mod->mod_status == MOD_NOT_MAINTAINED)
        {
            /* If this is not a maintaine module, just copy 
             * the connection information */
            mod->conn_status = old_mod->conn_status;
            mod->mod_socket = old_mod->mod_socket;
            continue;
        }

        /* All other active modules (including special and pre-defined ones)
         * must have a valid configuration. */
        ASSERT (mod->config);

        if (old_mod)
        {
            /* If na older module of the same name exists, */
            if (mod->binary && strcmp (old_mod->binary, mod->binary))
            {
                /* but it's a different program force a module restart 
                 * with a new configuration */
                debug_log ('1', "%s: BINs differ.", mod->name);
                mod->config_timestamp = 0;
                mod->mod_status = MOD_NEEDS_RESTART;
                gb_echelog_sane = false;
            }
            else if (!strcmp (old_mod->config, mod->config))
                /* If the configs match let the module inherit
                 * the old timestamp */
                mod->config_timestamp = old_mod->config_timestamp;
        }
        else
        {
            /* This is a newly defined module */
            mod->mod_status = MOD_NEEDS_START;
            gb_echelog_sane = false;
        }

        /* Suppress configuration preprocessing if the system
         * is recovering after a crash. */
        if (!old_conf->after_recovery)
            if (mod->preprocess_configuration())
                return -1;

        /* Create a module overall configuration. This should be quite safe */
        if (mod->create_mod_config ())
            return -1;

        /* If there is nothing to match, continue with another module */
        if (!old_mod)
            continue;

        /* Otherwise, let the module inherit the old properties */
        mod->mod_socket = old_mod->mod_socket;
        mod->pid = old_mod->pid;
        mod->conn_status = old_mod->conn_status;
        mod->mod_status = old_mod->mod_status;
        mod->last_cmd = old_mod->last_cmd;
        mod->deadline = old_mod->deadline;

        if (old_mod->mod_status == MOD_OK)
        {
            /* If the old module seems to be okay */
            if (strcmp (old_mod->config, mod->config) || 
                    old_mod->config_timestamp != mod->config_timestamp ||
                    mod->outputs.count () != old_mod->outputs.count ())
            {
                /* But the configuration differ, force a reconfiguration */
                debug_log ('1', "%s: configurations differ, "
                        "forcing reconfiguration", mod->name);
                mod->mod_status = MOD_NEEDS_RECONFIG;
                mod->last_cmd = socks::CMD_NOP;
            }
            else
            {
                tNode const *n1, *n2;

                /* If the configuration appears the same, make sure
                 * that the dataflow is also the same as before */
                for (n1=old_mod->outputs.get_first(), 
                        n2=mod->outputs.get_first();
                        n1; n1 = n1->next, n2 = n2->next)
                {
                    tModule *n = static_cast<tModule *> (n1->data);
                    tModule *m = static_cast<tModule *> (n2->data);

                    if (strcmp (n->name, m->name))
                    {
                        /* Force a reconfiguration otherwise */
                        debug_log ('1', "%s: output modules differ, "
                                "forcing reconfiguration", mod->name);
                        mod->mod_status = MOD_NEEDS_RECONFIG;
                        break;
                    }
                }
            }
        }
        else
        {
            /* If the module is not okay it'd be better to restart it */
            debug_log ('1', "%s: improper module status, forcing restart",
                    mod->name);
            mod->mod_status = MOD_NEEDS_RESTART;
            gb_echelog_sane = false;
        }

        debug_log ('1', "%s: merge status %d", mod->name, mod->mod_status);
    }

    /* Report success */
    return r;
}


// Distributes the accepted configuration to all echelog modules
void tConfiguration::apply ()
{
    using namespace socks;

    long i, len;
    int fd;

    /* For all active modules enforce a new configuration file */
    for (i = 0; i < mod_count; i++)
        modules[i]->apply_new_configuration ();

    /* Update the lock file configuration filename information */
    if ((fd = s_open (DEFAULT_LOCK_CF, O_WRONLY | O_CREAT | O_TRUNC, 0600)) < 0)
        log (L_WARN, "Can't create run stat file '%s'.", DEFAULT_LOCK_CF);

    len = s_strnlen (config_name, MAX_CMD_DATA_LENGTH);
    if (s_write (fd, config_name, len) < 0)
        log (L_WARN, "Can't write to run stat file '%s'.", DEFAULT_LOCK_CF);

    if (s_close (fd) < 0)
        log (L_WARN, "Can't close run stat file '%s'.", DEFAULT_LOCK_CF);

    /* Prepare a preprocessed cached version for a potential recovery */
    if (cf_preprocess (config_name, DEFAULT_CACHED_CONFIG))
        log (L_WARN, "Can't write cached config file '%s'.", 
                DEFAULT_CACHED_CONFIG);
}


// Force a configuration cleanup
void tConfiguration::cleanup ()
{
    long i;

    /* Delete all inactive modules' cached configuration files and clean
     * modules' connections. */
    debug_log ('1', "Purging cached configuration files.");
    for (i = 0; i < mod_count; i++)
        modules[i]->configuration_cleanup ();

}


// Fix configuration structures and all inactive modules for a safe erasure
void tConfiguration::prepare_for_wipeout ()
{
    tConfiguration *new_conf = gb_eenv.current_configuration;
    tModule *mod, *nmod;
    long i;

    /* For all old configuration modules */
    for (i = 0; i < mod_count; i++)
    {
        mod = modules[i];

        /* Try to retrieve a new module with the same name */
        nmod = new_conf->get_module (mod->name);

        if (mod->mod_status == MOD_NOT_MAINTAINED)
        {
            /* Non maintained modules needn't to be stopped */
            if (!nmod)
                /* There is no successor. Purge it */
                mod->socket_cleanup ();
            else
            {
                /* Just clear to connection (let it handle the successor)
                 * and remove the module from the old configuration */
                if (mod->mod_socket)
                {
                    mod->mod_socket = NULL;
                    active_modules--;
                }
            }

            /* Mark module as inactive */
            mod->mod_used = false;
            continue;
        }

        /* Check if there is a successor module in the new configuration */
        if (!nmod)
        {
            /* No, stop the module when it's possible */
            if (mod->mod_status != MOD_EMPTY && 
                    mod->mod_status != MOD_STARTED &&
                    mod->mod_status != MOD_PASSING_AWAY &&
                    mod->mod_status != MOD_ZOMBIE)
            {
                /* The module shows signs of life. */
                using namespace socks;
                debug_log ('1', "%s: planned to STOP", mod->name);

                /* Freeze the conection, empty the command queue
                 * and schedule the module stop */
                mod->set_conn_status (CS_BUSY);
                mod->empty_cmd_queue ();
                mod->enqueue_cmd (create_simple_command (CMD_STOP));
                gb_echelog_sane = false;
            }
            else 
            {
                /* It suffices to clear out the PID */
                mod->mod_status = MOD_ZOMBIE;
                debug_log ('1', "%s: considered a zombie", mod->name);
            }
        }
        else
        { 
            /* There is one */
            mod->mod_status = MOD_EMPTY;
            mod->mod_used = false;

            /* Just clear to connection (let it handle the successor)
             * and remove the module from the old configuration */
            if (mod->mod_socket)
            {
                mod->mod_socket = NULL;
                active_modules--;

                debug_log ('1', "%s: active config (old: %d)", mod->name, 
                        active_modules);
            }
        }
    }
}


// Get a program definition by its name
tProgramInfo *tConfiguration::get_program (const char *name)
{
    tProgramInfo tmp, *tptr = &tmp;
    tmp.name = name;

    /* Use binary searching to find it */
    tProgramInfo **prog = static_cast<tProgramInfo **> 
        (bsearch (&tptr, programs, prog_count, sizeof(tProgramInfo *),
                  prog_name_compare));
    if  (!prog)
        return NULL;

    return *prog;
}


// Get a module by its name
tModule *tConfiguration::get_module (const char *name)
{
    ASSERT (name);

    /* Use binary searching to find it */
    rfmod->name = const_cast<char *> (name);
    tModule **mod = static_cast<tModule **> 
        (bsearch (&rfmod, modules, mod_count, sizeof(tModule *),
                  mod_name_compare));
    if  (!mod)
        return NULL;

    return *mod;
}

#endif
