/**
 * @file
 * @brief Command line parsing
 *
 * This module contains an interface which allows a user to start, stop, 
 * restart or reconfigure any particular module or the whole echelog 
 * system as such.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "echelogctl.h"
#include <stdio.h>
#include <unistd.h>
#include <errno.h>


#ifndef DOXY_SKIP
using namespace lib;
using namespace socks;
#endif


/// Replace a NULL string with an empty string ""
#define S_EMPTY(str) ((str) ? (str) : "")

/// Standard remote control panel help message
static const char *help_msg = 
    "Usage:\n"
    "  echelogctl [-h] {stop, [reconfiguration options] reconfigure restart}\n"
    "  echelogctl [-h] module <module_name> {restart, reconfigure, "
    "toggle <path_name>}\n\n"
    "Options for system recofiguration:\n"
    "  -C, --config\t\t\tSet a new configuration file for startup\n"
    "  -l, --log\t\t\tSet a new file to log in\n\n"
    "Global options:\n"
    "  -h, --help\t\t\tThis help\n";

/// Short options for echeloctl    
static char short_options[] = "hC:l:";

/// Long options for echeloctl    
static struct option long_options[] =
{
    {"config", 1, 0, 'C'},
    {"log", 1, 0, 'l'},
    {"help", 0, 0, 'h'},
    {0, 0, 0, 0}
};


/// The first argv[] field containing real commands (not options)
static char **cmd_list;

/// Number of unparsed user's commands
static int cmd_number;

/// This enumeration clarifies all returned token types
typedef enum {
    T_STOP,         ///< Keyword stop
    T_RESTART,      ///< Keyword restart
    T_RECONFIGURE,  ///< Keyword reconfigure
    T_TOGGLE,       ///< Keyword toggle
    T_MODULE,       ///< Keyword module
    T_STRING,       ///< Strign token
    T_EOS           ///< End of string
} tTOKEN_TYPE;

/// The structure holding simple tokens
struct tToken {
    tTOKEN_TYPE type;   ///< Type of token
    char *str;          ///< String value of scanned token
};

/// The structure hold the parsed information from the command line
struct tEchelogAct
{
    char *config;           ///< Configuration filename
    char *log;              ///< Log filename
    char *module;           ///< Module to take action on
    char *param;            ///< Further command parameters
    socks::tCMD_TYPE cmd;   ///< Command type
};



/**
 * @brief Prints an error message and exit 
 *
 * The function prints the standard error message, terminates logging
 * and exit()s with a non-zero status.
 */
static void cmdline_error_msg(const char *msg)
{
    fprintf (stderr, "Error: %s\nPlease consult your manual page or "
            "type 'echelogctl -h' for help\n", msg);
    log_close ();
    exit (1);
}


/** 
 * @brief Gets the next token from argv[] string 
 *
 * The analyzed lexical is filled in <tt>token</tt> structure.
 * Any lexical element is also returned as a string.
 */
void get_token (tToken &token)
{
    char *tok;
    
    /* If there is a command left on the command line */
    if (cmd_number-- >= 0)
        /* Advance the carry */
        tok = *cmd_list++;
    else { 
        /* We must be at the end of the command line */
        token.type = T_EOS; 
        return;
    }

    if (!tok) {
        /* We must be at the end of the command line */
        token.type = T_EOS;
        return;
    }

    token.str = tok;
    token.type = T_STRING;
    switch (tok[0])
    {
        case 's':
            if (!strcmp (tok, "stop"))
                token.type = T_STOP;
            break;
        case 'r':
            if (!strcmp (tok, "restart"))
                token.type =  T_RESTART;
            else if (!strcmp (tok, "reconfigure"))
                token.type =  T_RECONFIGURE;
            break;
        case 'm':
            if (!strcmp (tok, "module"))
                token.type =  T_MODULE;
            break;
        case 't':
            if (!strcmp (tok, "toggle"))
                token.type =  T_TOGGLE;
    }
}


/* 
 * @brief Parses user's command
 *
 * The parsed command and it's parameters are filled in <tt>act</tt>
 * structure. Error checking is performed.
 *
 * @note The function exit()s with a non-zero status on failure.
 */
static void parse_command (tEchelogAct &act)
{
    tToken token;

    /* Get the first keyword */
    get_token (token);
    if (token.type == T_MODULE)
    {
        /* Get module's name */
        get_token (token);
        if (token.type == T_EOS)
            cmdline_error_msg ("Incomplete commmand.");
        else
            act.module = token.str;

        /* Look at the desired action */
        get_token (token);
        if (token.type == T_RESTART)
            act.cmd = CMD_CHAIN_RESTART;
        else if (token.type == T_RECONFIGURE)
            act.cmd = CMD_RECONFIGURE;
        else if (token.type == T_TOGGLE)
        {
            /* Get toggle info */
            get_token (token);
            if (token.type == T_EOS)
                cmdline_error_msg ("Missing number in toggle clause.");
            else
                act.param = token.str;
            act.cmd = CMD_TOGGLE;
        }
        else
        {
            char *err;
            s_asprintf(&err, "Invalid command '%s' for module '%s'.", 
                    token.str, act.module);
            cmdline_error_msg (err);
        }
    }
    else
    {
        /* Look at the desired action for the whole system */
        switch (token.type)
        {
            case T_STOP:
                act.cmd = CMD_SYS_STOP;
                break;
            case T_RESTART:
                act.cmd = CMD_SYS_RESTART_ALL;
                break;
            case T_RECONFIGURE:
                act.cmd = CMD_SYS_RECONFIGURE;
                break;
            default:
                /* Invalid action */
                char *err;
                s_asprintf(&err, "Invalid action '%s'.", token.str);
                cmdline_error_msg (err);
        }
    }
}


// Processes control command line parameters
socks::tCommand *process_cmdline (int argc, char *argv[])
{
    tEchelogAct act;
    int c;

    /* Set defaults */
    act.config = act.log = act.module = NULL;
    
    /* Parse command line option at first */
    while (1)
    {
        c = getopt_long (argc, argv, short_options, long_options, NULL);
        if (c  == -1)
            break;

        switch (c)
        {
            case 'C':   /* config file */
                act.config = optarg;
                break;
            case 'l':   /* log file */
                act.log = optarg;
                break;
            case 'h':   /* help */
                fprintf (stderr, "%s", help_msg);
                exit (0);
                break;
            case ':':   /* missing arg */
                cmdline_error_msg("Missing parameter.");
            case '?':   /* unkown option */
            default:    /* invalid option */
                cmdline_error_msg("Unknown command line option used.");
        }
    }

    if (optind >= argc) 
        cmdline_error_msg ("No action to be done");
    else
    {
        /* Parse command structure */
        cmd_list = argv + optind;
        cmd_number = argc - optind;
        parse_command (act);

        if ((act.config || act.log) && act.cmd != CMD_SYS_RECONFIGURE)
            cmdline_error_msg ("Setting new config or log file "
                    "is allowed only on echelog reconfiguration.");
    }

    /* Form the resulting command */
    tCommand *cmd = static_cast<tCommand *> (xmalloc (MAX_CMD_LENGTH));
    cmd->command = act.cmd;
    if (cmd->command == CMD_SYS_RECONFIGURE)
        cmd->d_len = cmd_store_strings ((char **)&cmd->d_ptr, S_EMPTY
                (act.module), S_EMPTY(act.config), S_EMPTY(act.log),
                MAX_CMD_DATA_LENGTH);
    else
        cmd->d_len = cmd_store_strings ((char **)&cmd->d_ptr, S_EMPTY
                (act.module), S_EMPTY(act.param), "", MAX_CMD_DATA_LENGTH);

    /* Do a minor check */
    if (cmd->d_len < 0)
       cmdline_error_msg ("Supplied parameters don't fit in a single command.");

    /* And return the result */
    return cmd;
}
