/**
 * @file
 * @brief Echelogd Main
 *
 * This is the only module which is started by an administrator. It reads
 * the echelog configuration and and accordingly to this file it maintains
 * all other modules.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */                                                                             

#include "echelogd.h"
#include "socks/ctrl_unixsocket.h"

#include <stdio.h>
#include <unistd.h>


#ifndef DOXY_SKIP
using namespace lib;
using namespace socks;

// This structure contains all configuration related information.
tEchelogEnv gb_eenv;
#endif

/**
 * @brief Parsed (default) configuration file name
 */
static char * config_name = NULL;

/**
 * @brief Recovered configuration file name
 */
static char * recovery_config_name = NULL;

/**
 * @brief Standard error message
 */
static const char * const err_msg = 
    "Parameters error. Please consult your manual page\n";

/**
 * @brief Short options for echelogd
 */
static char short_options[] = "C:l:";

/**
 * @brief Long options for echelogd
 */
static struct option long_options[] =
{
    {"config", 1, 0, 'C'},
    {"log", 1, 0, 'l'},
    {0, 0, 0, 0}
};


/**
 * @brief Parses the echelogd command line.
 *
 * This is a simple parser of the daemon command line. Unlike others
 * it uses only plain getopt_long function. Only a configuration filename
 * and log filename settings are allowed.
 */
static void parse_cmdline (int argc, char **argv)
{
    int c, len;

    /* Set defaults */
    gb_eenv.log_name = s_strdup (DEFAULT_LOG);
    gb_eenv.current_configuration = NULL;
    
    /* Parse command line option at first */
    while (1)
    {
	c = getopt_long (argc, argv, short_options, long_options, NULL);
        if (c  == -1)
            break;

	switch (c)
	{
	    case 'C':   /* config file */
                len = min (strlen (optarg) + 1, MAX_CMD_DATA_LENGTH);
                config_name  = s_strdup_portion (optarg, len);
                config_name[len] = '\0';
		break;
	    case 'l':   /* log file */
                xfree (gb_eenv.log_name);
                gb_eenv.log_name = s_strdup (optarg);
		break;
            case ':':   /* missing arg */
		fprintf (stderr, "%s", err_msg);
		exit (1);
            case '?':   /* unkown option */
	    default:    /* invalid option */
		fprintf (stderr, "%s", err_msg);
		exit (1);
	}
    }

    /* No more parameters are allowed */
    if (optind < argc) 
    {
        fprintf (stderr, "%s", err_msg);
        exit (1);
    }

    /* Replace the empty configuration filename with something reasonable */
    if (!config_name || !strcmp (config_name, ""))
        config_name = s_strdup (DEFAULT_CONFIG);
}


/**
 * @brief Tests if echelog deamon is already listening on the control socket
 *
 * Tries to connect to echelogd on the default control socket. If the call ends
 * up successfully the echelogd is probably already running.
 *
 * @return 
 * The fuction returns 1 if another echelogd instance is currently
 * running, zero otherwise.
 */
static int test_presence ()
{
    tCtrlUNIXSocket s;

    if (!s_eaccess (DEFAULT_CTRL_SOCKET, F_OK) && 
            !s.create_client (DEFAULT_CTRL_SOCKET))
        return 1;

    return 0;
}


/**
 * @brief Tests if the system is restarted after a crash
 *
 * The echelogd creates a lock file which is purged upon successful
 * termination. If the file exists, if it seems to contain the last
 * configuration file name used by echelogd and if the cached parsed echelogd 
 * configuration is valid, the system is considered to be recovering after 
 * a crash.
 *
 * @return
 * The function returns true if echelogd is recovering from a system crash;
 * otherwise, false is returned.
 */
static bool test_recover ()
{
    int fd, len;

    /* Does the lock file exist?  */
    if(!s_access (DEFAULT_LOCK_CF, F_OK | R_OK))
    {
        /* Is it readable? */
        if ((fd = s_open (DEFAULT_LOCK_CF, O_RDONLY)) < 0)
            return false;

        /* Does it have a permitted length? */
        if ((len = min (s_filelen (fd), MAX_CMD_DATA_LENGTH)) <= 0)
            return false;

        /* Read the config filename of the previous session */
        recovery_config_name = static_cast<char *> (get_cfpool().alloc (len));
        s_read (fd, recovery_config_name, len);
        recovery_config_name[len] = '\0';
        s_close (fd);

        /* Try to parse the cached configuration */
        if (cf_preprocess (DEFAULT_CACHED_CONFIG, "/dev/null"))
            return false;
        
        /* Try to recover */
        return true;
    }

    /* Create a new lock file */
    if ((fd = s_open (DEFAULT_LOCK_CF, O_WRONLY | O_CREAT | O_EXCL, 0600)) < 0)
    {
        log (L_WARN, "Can't create run stat file '%s': %m.", DEFAULT_LOCK_CF);
        return false;
    }
    
    /* Write in the parsed configuration filename */
    len = s_strnlen (config_name, MAX_CMD_DATA_LENGTH);
    if (s_write (fd, config_name, len) < 0)
        log (L_WARN, "Can't write to run stat file '%s': %m.", DEFAULT_LOCK_CF);

    /* And flush it */
    if (s_close (fd) < 0)
    {
        log (L_WARN, "Can't close run stat file '%s': %m.", DEFAULT_LOCK_CF);
        return false;
    }

    /* This seems to be a normal startup */
    return false;
}


/**
 * @brief Setup default modules' log file
 *
 * The modules' log messages go to another log file than echelog messages
 * for administrator's convenience. If the creation fail all modules' messages
 * are sent to /dev/null.
 */
static void setup_modules_log ()
{
    gb_eenv.modules_fd = s_open (DEFAULT_MODULES_LOG, 
            O_WRONLY | O_CREAT | O_APPEND, 0600);

    if (gb_eenv.modules_fd < 0)
    {
        log (L_ERROR, "Can't open log file '%s'.", DEFAULT_MODULES_LOG);
        gb_eenv.modules_fd  = sd_open ("/dev/null", O_WRONLY);
    }
}


/** 
 * @brief Become a daemon
 *
 * This could be called five easy steps to create a unix deamon. Nothing
 * unexpectable should happen here. Since the program run in the background
 * with no control terminal all it's messages will be redirected to specified
 * file <tt>name</tt>.
 */
static void daemonize (const char *name)
{
    log (L_INFO, "Putting the program to the background.");

    /* Detach from a controlling terminal. */
    sd_detach_from_ctrl_terminal ();

    /* Init logging routines again. */
    log_init (name);

    /* Change the path to deamon's default directory to release 
     * the current directory (so it can be umounted etc.). */
    chdir (ECHELOG_INST_PREFIX);

    /* We don't know nothing about current umask. Drop it. */
    umask (0);

    /* Release inherited standard file descriptors and redirect
     * log error output to the specified log file. */
    log (L_INFO, "The daemon's output is redirected to '%s'.", gb_eenv.log_name);
    log_file (gb_eenv.log_name);

    /* Give up standard input and output (but leave them opened). */
    s_close (1);
    s_close (0);
    sd_open ("/dev/null", O_RDONLY);
    sd_open ("/dev/null", O_WRONLY);
}


/* --===  PROGRAM ENTRY POINT ===-- */
int main (int argc, char *argv[])
{
    /* Init logging routines */
    argv[0] = "echelogd";
    log_init (argv[0]);

    /* Process the control command line structure */
    parse_cmdline (argc, argv);

    /* Test if an another echelogd instance is running */
    if (test_presence ())
        die ("Echelogd is already running on this control socket. Exiting.");

    /* Set up logging for modules */
    setup_modules_log ();

    /* Set up basic configuration */
    gb_eenv.current_configuration = new tConfiguration;
    gb_eenv.previous_configuration = NULL;

    if (gb_eenv.current_configuration == NULL)
        die ("Not enough memory.");
    
    /* Check if we are recovering after a failure */
    if (test_recover ())
    {
        log (L_INFO, "System is recovering after a failure.");
        gb_eenv.current_configuration->set_config_name (recovery_config_name);
        gb_eenv.current_configuration->set_after_recovery (true);
    }
    else
        gb_eenv.current_configuration->set_config_name (config_name);

    /* Free temporary parsing structures */
    get_cfpool().free_pool ();
    
    /* Become a daemon */
    daemonize (argv[0]);

    /* Initialize the run queue */
    init_run_queue (config_name);
    xfree (config_name);

    /* Set up the control channel */
    setup_ctrl_channel ();

    /* You know the law: you gotta do what you gotta do */
    process ();
    
    /* Stopped? Okay, tidy up and quit */
    shutdown_ctrl_channel ();
    debug_log ('1', "Echelog halted.");
    log_close ();

    return 0;
}
