/** 
 * @file
 * @brief Main processing
 *
 * This is the only module which is started by an administrator. It reads
 * echelog configuration and and accordingly to this file it maintains
 * all other modules.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_ECHELOGD_H
#define ECHELOG_ECHELOGD_H

#include "cmd_param.h"
#include "echelogd_config.h"
#include "module.h"


/**
 * @brief Client connection information.
 *
 * This structure holds information about all yet-unauthenticated
 * connections to echelogd.
 */
struct tConnection
{
    socks::tSocket *conn_socket;    ///< General connection primitive
    tConnectionStatus status;       ///< Connection status
    struct timeval login_limit;     ///< Connection time limit
};


/**
 * @brief Configuration environment.
 *
 * This structure contains all echelogd configuration related information.
 */
struct tEchelogEnv {
    char *log_name;                         ///< Echelogd log filename
    int modules_fd;                         ///< Modules log filedescriptor
    tConfiguration *current_configuration;  ///< Active echelog configuration
    tConfiguration *previous_configuration; ///< Previous echelog configuration
};


/**
 * @brief The real instance of tEchelogEnv structure.
 */
extern tEchelogEnv gb_eenv;

/**
 * @brief Discrete time counter.
 */
extern struct timeval gb_current_time;

/**
 * @brief Flag whether echelog configuration needs a deeper contemplation.
 */
extern bool gb_echelog_sane;

/**
 * @brief Flag whether the reconfiguration is in progress.
 */
extern bool gb_reconfiguration_in_progress;

/**
 * @brief Sets up the control channel.
 * 
 * Creates a new controls socket and sets up the default handlers for newly
 * accepted connections.
 */
void setup_ctrl_channel ();

/**
 * @brief Cleans up the control channel.
 * 
 * Unregisters and releases the echelogd listening socket.
 */
void shutdown_ctrl_channel ();

/**
 * @brief Initializes the echelogd command queue for start.
 *
 * Schedules a system start. The function parameter <tt>config_file</tt>
 * will be packed into the <tt>CMD_SYS_START</tt> command to instruct echelog
 * to load the initial configuration file from there.
 */
void init_run_queue (const char *config_file);

/**
 * @brief Makes echelog system state more acceptable by fixing objectionable
 * echelog modules.
 *
 * Traverses through the echelog module dependancy graph adjusting module
 * states thorough all module chains. The goal is also to minimize
 * number of actions to make all modules behave reasonable.
 */
void exec_sanitize_modules ();

/**
 * @brief The main processing function.
 * 
 * This is the main processing function. It periodically checks echelog
 * modules, processes commands and handles socket I/O.
 *
 */
void process ();

/**
 * @brief Receives command from module's socket. 
 * 
 * Checks the originating module's capabilities and calls
 * appropriate command handler.
 */
void exec_sys_receive_cmd (tModule &mod, socks::tCommand *c);

/**
 * @brief Processes commands from the command queue. 
 *
 * Processes all commands in the standard command queue. 
 *
 * @return
 * Normally, the function returns zero. If a non-zero is returned the 
 * deamon is ready to be terminated.
 */
int exec_sys_commands ();

/** 
 * @brief Checks delayed commands in the system queue.
 *
 * Examines delayed system commands in a sequence and accordingly
 * to system state decides whether to move them to standard system
 * queue, to discard them or just to wait.
 */
void check_sys_delayed ();

/**
 * @brief Checks modules and takes proper actions. 
 *
 * Process all modules in currently active configuration.
 */
void process_modules (tConfiguration *conf);

/**
 * @brief Checks old modules and takes proper actions.
 *
 * Processes all active modules in the privious configuration and when they're
 * all finished it purges the rest of old configuration data and finishes
 * the reconfiguration process.
 */
void process_old_modules (tConfiguration *conf);

#endif
