/**
 * @file
 * @brief Echelogd configuration
 *
 * This is the only module which is started by an administrator. It reads
 * the echelog configuration and and accordingly to this file it maintains
 * all other modules.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_ECHELOGD_CONFIG_H
#define ECHELOG_ECHELOGD_CONFIG_H


#include "property.h"
#include "lib/lib.h"
#include "socks/socketmgr.h"
#include "socks/ctrl_unixsocket_nb.h"


// forward declaration
class tModule;


/**
 * @brief Maximal configuration file name length (in bytes).
 */
#define MAX_CONF_NAME_LEN 4096

/** 
 * @brief Number of standard predefined program items.
 */
#define PREDEFINED_PROGS_NUM 1

/** 
 * @brief Number of standard predefined module items.
 */
#define PREDEFINED_MODS_NUM 1


/**
 * @brief Information about module's binary.
 *
 * Structure holding information about a binary path
 * for a particular program.
 */
struct tProgramInfo {
    const char *name;           ///< Program name
    const char *program_path;   ///< Program path to binary
};


/**
 * @brief Configuration container.
 *
 * This class encapsulates the complete echelog configuration 
 * information and provides a simple interface to manipulate it.
 */
class tConfiguration {

    protected:

        /** 
         * @brief Memory pool for internal configuration stuff.
         */
        lib::tMemPool pool;

        /**
         * @brief Array of (pointers to) all defined programs.
         */
        tProgramInfo **programs;

        /**
         * @brief Array of (pointers to) all defined programs.
         */
        tModule **modules;

        /**
         * @brief Array of list of modules in each dependency level.
         */
	lib::tList** levels;

        /**
         * @brief <em>Property:</em> Active module counter.
         */
        SIMPLE_PROPERTY (long, active_modules);

        /**
         * @brief <em>Property:</em> Flag to signalize a recovery after.
         * a failure
         */
        SIMPLE_PROPERTY (bool, after_recovery);

        /**
         * @brief <em>Constant property:</em> Filename to read configuration
         * from.
         */
        CONST_PROPERTY (char *, config_name);

        /**
         * @brief <em>Read-only property:</em> Number of configured programs.
         */
        SIMPLE_R_PROPERTY (long, prog_count);

        /**
         * @brief <em>Read-only property:</em> Number of configured modules.
         */
        SIMPLE_R_PROPERTY (long, mod_count);

        /// <em>Read-only property:</em> Number of dependency layers.
        SIMPLE_R_PROPERTY (long, level_count);


        /**
         * @brief <em>Read-only property:</em> Module identificator of the
         * remote control module.
         */
        SIMPLE_R_PROPERTY (long, ctrl_mod_id);


        /**
         * @brief Referential module pointer.
         */
        tModule *rfmod;

    protected:

        /**
         * @brief Checks the cross-dependencies between configuration sections.
         *
         * Currently checks if all defined modules are used. In a single
         * configuration this would make no difference but it would require
         * some more complex checking later (on reconfiguration for example).
         */
        int check_definitions () const;

        /**
         * @brief Topologically sorts all configured modules.
         *
         * Try to topologically sort the DAG and Make a list of all modules in
         * each level. The level with the zero number contains modules which
         * can be started immediately.
         * 
         * @return
         * The function returns zero on success if there was an error 
         * in this configuration part, <tt>-1</tt> is returned.
         */
        int topologically_sort ();

        /**
         * @brief Configures and propagates modules' input nodes.
         * 
         * Configure modules in all echelog levels.  Let each flowswitch
         * default to the first (the most left) dataflow path.
         * 
         * @return
         * The function returns zero on success if there was an error 
         * in this configuration part, <tt>-1</tt> is returned.
         */
        int config_input_nodes ();

        /**
         * @brief Parses and defines configured programs.
         *
         * Allocates space for all program definitions, parses 
         * all program definition lines and sets the program
         * configuration structures.
         * 
         * @return
         * The function returns zero on success if there was an error 
         * in this configuration part, <tt>-1</tt> is returned.
         */
        int parse_programs (lib::tList &list);

        /**
         * @brief Parses and defines configured modules.
         *
         * Allocates space for all module definitions, parses 
         * all module definition lines and sets the module
         * configuration structures.
         *
         * @return
         * The function returns zero on success if there was an error 
         * in this configuration part, <tt>-1</tt> is returned.
         */
        int parse_modules (lib::tList &list);

        /**
         * @brief Parses the dataflow configuration sections and defines
         * the dependencies between echelog modules.
         *
         * Parses the dataflow, builds up the modules' local dependency
         * trees and activates modules
         * 
         * @return
         * The function returns zero on success if there was an error 
         * in this configuration part, <tt>-1</tt> is returned.
         */
        int parse_dataflow (lib::tList &list);

    public:


        /**
         * @brief Constructor--creates a configuration with predefined 
         * programs and modules only.
         */
        tConfiguration ();

        /**
         * @brief Destructor--releases all used structures and forces
         * inactive modules cleanups.
         */
        virtual ~tConfiguration ();

        /**
         * @brief Loads a new configuration.
         *
         * The function registers all sections needed for echelog
         * configuration, reads the configuration files and tries
         * to build up the configuration structures.
         * 
         * @return
         * The function returns zero on success if there was an error 
         * in this configuration part, <tt>-1</tt> is returned.
         */
        int load ();

        /**
         * @brief Merges old configuration definitions with the newly
         * loaded configuration.
         *
         * For all newly defined modules the functions tries to retrieve 
         * an old module with the same name and (if possible) to let 
         * the new module inherit the old properties.
         * 
         * @return
         * The function returns zero on success if there was an error 
         * in this configuration part, <tt>-1</tt> is returned.
         */
        int merge (tConfiguration *old_conf); 

        /**
         * @brief Distributes the accepted configuration to all echelog
         * modules.
         *
         * Enforces new configuration files to all active modules,
         * updates the lock file recovery information and prepares
         * a cached configuration.
         */
        void apply (); 

        /**
         * @brief Force a configuration cleanup.
         *
         * Delete all inactive modules' cached configuration files and clean
         * modules' connections.
         */
        void cleanup ();

        /**
         * @brief Fix configuration structures and all inactive 
         * modules for a safe erasure.
         *
         * The function prepares to stop all unused echelog modules.
         * If the module has no successor the module and is connection
         * are erased otherwise the successor inherits all viable resources.
         */
        void prepare_for_wipeout();


        /**
         * @brief Get a program definition by its name.
         *
         * The function returns a pointer to the program definition 
         * of the given <tt>name</tt> or NULL if there none such.
         */
        tProgramInfo *get_program (const char *name);

        /**
         * @brief Get a module by its name.
         *
         * The function returns a pointer to the module
         * of the given <tt>name</tt> or NULL if there none such.
         */
        tModule *get_module (const char *name);

        /**
         * @brief Get a module by its handle.
         */
        inline tModule *get_module (const long handle)
            { ASSERT (handle >= 0 && handle < mod_count);
              if (modules) return modules[handle]; 
              else return NULL; 
            }

        /**
         * @brief Get a program definition by its handle.
         */
        inline tProgramInfo *get_program (const long handle)
            { ASSERT (handle >= 0 && handle < prog_count);
                if (programs) return programs[handle];
                else return NULL;
            }

        /**
         * @brief Get a list of modules at certain level.
         */
        inline lib::tList *get_level (const long num)
            { ASSERT (num >= 0 && num < level_count);
              return levels[num];
            }

        /**
         * @brief Get a reference to configuration memory pool.
         */
        inline lib::tMemPool &get_pool () { return pool; }
        
};


#endif
