/**
 * @file
 * @brief Module Management
 *
 * This is the only module which is started by an administrator. It reads
 * the echelog configuration and and accordingly to this file it maintains
 * all other modules.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "echelogd.h"
#include <sys/wait.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>


#ifndef DOXY_SKIP
using namespace lib;
using namespace socks;

#define FAILURE 1
#endif


// Runs echelog module.
int tModule::run_module ()
{
    pid_t npid;
    char * const mod_env[] = { NULL };
#if DEBUG > 3
    char * mod_args[] = { NULL, "-C", NULL, "-Xt", NULL };
#else
    char * mod_args[] = { NULL, "-C", NULL, NULL };
#endif

    ASSERT (binary && mod_config);

    npid = fork ();
    if (npid == 0)
    {
        /* This is the child process. Adjust the run environment and exec 
         * the module. */
        if (s_close (2))
            _exit (FAILURE);
        if (s_close (1))
            _exit (FAILURE);
        if (s_close (0))
            _exit (FAILURE);
        if (s_open ("/dev/null", O_RDONLY) < 0)
            _exit (FAILURE);
        if (s_open ("/dev/null", O_WRONLY) < 0)
            _exit (FAILURE);
        if (dup2 (gb_eenv.modules_fd, 2) == -1)
            _exit (FAILURE);

        mod_args[0] = s_strdup (name);
        mod_args[2] = s_strdup (mod_config);

        /* Exec the module's binary with proper arguments and clean
         * environment */
        execve (binary, mod_args, mod_env);
        log (L_FATAL, "Exec failed");
        _exit (FAILURE);
    }
    else if (npid < 0)   /* Bad luck this time */
    {
        log (L_FATAL, "Fork failed.");
        return -FAILURE;
    }
    else
    {
        /* This is the parent process. */
        debug_log ('p', "Forking module %s, pid %i", name, npid);

        /* Advance the counter of configuration's active modules */
        rconf.set_active_modules (rconf.get_active_modules() + 1);
        pid = npid;

        /* Adjust the last executed command and its deadline */
        last_cmd = CMD_START;
        ADVANCE_TIMER (&deadline, MAX_SPAWN_TIME);
    }

    /* Return successfully */
    return 0;
}


// Kills echelog module.
int tModule::kill_module ()
{
    int r;

    debug_log ('3', "%s: trashing cmd queue up to next start cmd", name);

    /* 
     * Trash the command queue up to the next start command, the module
     * won't need it ;) 
     */
    while (cmds.count ())
    {
        tCommand *c = static_cast<tCommand *> (cmds.peek ());

        if (c->command == CMD_START)
            break;

        /* Release removed commands */
        c = static_cast<tCommand *> (cmds.remove ());
        free_cmd (c);
    }

    /* This shouldn't be called twice, it's just to be sure */
    if (pid != 1)
    {
        /* Does the process exist? */
        debug_log ('1', "%s: SIGQUIT", name);
        if ((r = s_test_signal_delivery (pid, NULL)))
        {
            /* Hmmm, weird. Trying to kill a dead module */
           module_disappeared:
            log (L_ERROR, "Module %s (pid %d) disappeared. Can't stop it "
                    "correctly.", name, pid);
            
            /* Quick-finish and cleanup the socket. */
            mod_status = MOD_EMPTY;
            socket_cleanup ();
            pid = 1;
            return -1; 
        }
        else
        {
            /* At first send the process a CONT signal to let it
             * terminate correctly if it's stopped accidentally. */
            if (kill (pid, SIGCONT) < 0)
                goto module_disappeared;
            
            /* Send it QUIT signal */
            if (kill (pid, SIGQUIT) < 0)
                /* This is also weird, the process is gone */
               goto module_disappeared;

            mod_status = MOD_PASSING_AWAY;

            /* Adjust the last executed command and its deadline */
            ADVANCE_TIMER (&deadline, MAX_TERMINATION_TIME);
            last_cmd = CMD_STOP;
        }
    }

    /* Report success */
        return 0;
}


// Finishes module and sets its process ID free.
int tModule::finish_module ()
{
    debug_log ('1', "%s: FINISH", name);

    /* Clear out the module status */
    mod_status = MOD_EMPTY;

    /* The process shouldn't be finished twice, it's just to be sure */
    if (pid != 1)
    {
        /* Test if the module is running or set it's PID free */
        switch (s_test_signal_delivery (pid, NULL))
        {
            case -1:
                /* This is weird, the process is gone. Do the cleanup. */
               module_disappeared: 
                log (L_ERROR, "Module %s (pid %d) disappeared. Can't stop it "
                        "correctly.", name, pid);
                socket_cleanup ();
                pid = 1;
                return -1; 

            case 0:
                /* The module is still not dead. Make it so. */
                if (kill (pid, SIGKILL) < 0)
                    goto module_disappeared;

                /* And clear the module's PID */
               restart_command:
                if (waitpid (pid, NULL, 0) < 0)
                    switch (errno)
                    {
                        case EINTR:
                            goto restart_command;

                        default:
                            goto module_disappeared;
                    }
                break;

            case 1:
                break;
        }
    }

    /* Release the module's connection */
    socket_cleanup ();

    /* Report success */
    pid = 1;
    return 0;
}


// Schedules start of echelog module.
void tModule::start_module ()
{
    if (mod_status == MOD_NOT_MAINTAINED || 
            mod_status == MOD_STARTED || mod_status == MOD_OK)
        return;

    log (L_INFO, "Starting module %s.", name);

    /* Remove all old commands from the command queue to avoid confusion */
    empty_cmd_queue ();

    if (mod_role == MOD_ROLE_FLOWSWITCH)
        /* This is not a real module thus can be started rather easily ;) */
        mod_status = MOD_OK;
    else
    {
        /* Try to run the module */
        if (run_module ())
        {
            /* Bad luck so early? It seems that the module can't be started */
            log (L_ERROR, "Can't start module %s.", name);
            mod_status = MOD_BROKEN;
            return;
        }

        /* Let the module identify itself */
        mod_status = MOD_STARTED;
    }
}


// Schedules stop of echelog module.
void tModule::stop_module ()
{
    if (mod_status == MOD_NOT_MAINTAINED || 
            mod_status == MOD_EMPTY || mod_status == MOD_BROKEN)
        return;

    log (L_INFO, "Stopping module %s.", name);

    if (mod_role == MOD_ROLE_FLOWSWITCH)
        /* This is not a real module thus can be stopped rather easily ;) */
        mod_status = MOD_EMPTY;
    else
    {
        /* Kill the module */
        if (kill_module ())
        {
            /* There was a problem, force a finish */
            mod_status = MOD_ZOMBIE;
            return;
        }

        /* Let the module wait for the module connection hangup */
        mod_status = MOD_PASSING_AWAY;
    }
}


// Reconfigures echelog module.
void tModule::reconfigure_module ()
{
    if (mod_status == MOD_NOT_MAINTAINED)
        return;

    log (L_INFO, "Reconfiguring module %s.", name);

    if (mod_role == MOD_ROLE_FLOWSWITCH)
    {
        /* The reconfiguration of a flowswitch is always successful */
        mod_status = MOD_OK;
        return ;
    }
    
    /* Does the process exist? */
    if (s_test_signal_delivery (pid, NULL))
    {
        /* It doesn't seem so. Force a restart */
       module_disappeared:
        debug_log ('1', "%s: forcing restart", name);
        mod_status = MOD_NEEDS_RESTART;
        gb_echelog_sane = false;
        return;
    }

    /* Try to awake an accidentally stopped module */
    if (kill (pid, SIGCONT) < 0)
        goto module_disappeared;

    /* And send a reconfiguration signal to it */
    if (kill (pid, SIGHUP) < 0)
        goto module_disappeared;

    /* Adjust the last executed command and its deadline */
    ADVANCE_TIMER (&deadline, MAX_RECONFIGURATION_TIME);
    last_cmd = CMD_RECONFIGURE;
}


// Toggles chain path at flowswitch module.
void tModule::toggle_module ()
{
    tNode const *c;

    if (mod_role != MOD_ROLE_FLOWSWITCH)
    {
        /* Create a new input/output socks description file */
        create_mod_config ();

        /* Schedule module reconfiguration / restart */
        if (mod_status == MOD_OK)
            mod_status = MOD_NEEDS_RECONFIG;
        else if (mod_status == MOD_STARTED)
        {
            mod_status = MOD_NEEDS_RESTART;
            gb_echelog_sane = false;
        }

        /* Stop the recursion at the first non-flowswitch module */
        return;
    }

    /* For all chained modules toggle the flags */
    FOR_ALL_LIST_ITEMS (c, &inputs)
    {
        tModule *pmod = static_cast<tModule *> (c->data);
        ASSERT (pmod);

        /* Recurse for all input modules */
        pmod->toggle_module ();
    }
}


// Schedules transfer of command to echelog module
void tModule::sched_cmd_transfer (socks::tCMD_TYPE command)
{
    debug_log ('6', "Scheduling command %u on socket %ld", command, 
            mod_socket->get_fd ());

    mod_socket->store_cmd (command);
    last_cmd = command;
    conn_status = CS_BUSY;

    /* Schedule the command header write */
    mod_socket->send_cmd ();
}


// Checks if conditions for command execution are satisfied or not
int tModule::check_cmd_conditions (socks::tCommand *cmd) const
{
    tNode const *c;
    ASSERT (cmd);

    debug_log ('3', "%s: considering execution of %x", name, cmd->command);

    if (mod_status == MOD_NOT_MAINTAINED)
    {
        /*
         * We know too little about external modules. At least, try to 
         * serialize the command sending. 
         */
            switch (conn_status)
            {
                case CS_OK:
                    return 1;
                case CS_BUSY:
                case CS_NOT_RESPONDING:
                    return 0;
                default:
                    return -1;
            }
        return 1;
    }

    switch (cmd->command)
    {
        case CMD_NOP:
            /* 
             * No operation is always allowed 
             */
            return 1;

        case CMD_PING:
        case CMD_PONG:
            /*
             * Hold or discard the command if the connection status
             * is not perfect 
             */
            switch (conn_status)
            {
                case CS_OK:
                    return 1;
                case CS_BUSY:
                case CS_NOT_RESPONDING:
                    return 0;
                default:
                    return -1;
            }

        case CMD_LOGIN_OK:
            /* 
             * Ensure that the module is all correct 
             */
            if (mod_status == MOD_OK)
                if (conn_status != CS_OK)
                    return 0;
                else
                    return 1;
            else
                return -1;

        case CMD_START:
            if (mod_status == MOD_OK || mod_status == MOD_STARTED)
                /* If the module has just been started or is already
                 * running, allow the command exection it'll have no effect */
                return 1;
            else if (mod_status == MOD_BROKEN)
                /* Oops, the module can't be started */
                return -1;
            else if (mod_status == MOD_EMPTY)
            {
                /*
                 * It seems to be startable, check all chained modules which
                 * the module is dependent on.
                 */
                FOR_ALL_LIST_ITEMS (c, &outputs)
                {
                    tModule *cmod = static_cast<tModule *> (c->data);

                    switch (cmod->mod_status)
                    {
                        case MOD_BROKEN:
                            return -1;
                        case MOD_OK:
                            break;
                        default:
                            return 0;
                    }
                }
                return 1;
            }
            break;

        case CMD_STOP:
            if (mod_status == MOD_EMPTY || mod_status == MOD_BROKEN)
                /* If the module has just been stopped or is already
                 * broken, allow the command exection it'll have no effect */
                return 1;
            else if (mod_status == MOD_STARTED || 
                    mod_status == MOD_PASSING_AWAY ||
                    mod_status == MOD_ZOMBIE)
                /* Wait till the module falls to a final state */
                return 0;
            else
            {
                if (mod_status == MOD_NEEDS_STOP)
                { 
                    /* 
                     * This state is used for modules from the old
                     * configuration which should be stopped. Wait 
                     * until all active module (in the new configuration)
                     * fall into final states.
                     */
                    FOR_ALL_LIST_ITEMS (c, &inputs)
                    {
                        tModule *cmod = static_cast<tModule *> (c->data);

                        if (cmod->mod_status != MOD_OK &&
                                cmod->mod_status != MOD_BROKEN &&
                                cmod->mod_status != MOD_EMPTY)
                            return 0;
                    }

                    /* We may safely stop it now */
                    return 1;
                }

                /*
                 * It seems to be stopable, check all chained modules which
                 * depends on this module.
                 */
                FOR_ALL_LIST_ITEMS (c, &inputs)
                {
                    tModule *cmod = static_cast<tModule *> (c->data);

                    if (cmod->mod_status != MOD_EMPTY &&
                            cmod->mod_status != MOD_BROKEN)
                        return 0;
                }
                /* We may safely stop it now */
                return 1;
            }

        case CMD_RECONFIGURE:
            /*
             * Reconfigure only working modules
             */
            if (mod_status == MOD_OK)
            {
                /* Check if all chained modules, which the module
                 * is dependent on, are okay. */
                FOR_ALL_LIST_ITEMS (c, &outputs)
                {
                    tModule *cmod = static_cast<tModule *> (c->data);
                    if (cmod->mod_status == MOD_BROKEN)
                        /* Reconfiguration makes no sence */
                        return -1;
                    else if (cmod->mod_status != MOD_OK)
                        return 0;
                }
                /* Okay, it may work */
                return 1;
            }
            else if (mod_status == MOD_BROKEN)
                /* It won't reconfigure itself, anyway */
                return -1;
            break;

        default:
            /* All other commands has no defined constraints */
            return 1;
    }

    /*
     * The default strategy is to wait till the module falls to a final
     * status which should be handled in the code above.
     */
    return 0;
}


// Processes module's command queue.
void tModule::process_cmd_queue ()
{
    tCommand *dc;
    
    /* Process only one command per round */
    if (!cmds.is_empty ())
    {
        dc =  static_cast<tCommand *> (cmds.peek ());

        switch (check_cmd_conditions (dc))
        {
            case -1:
                /* Something went wrong, discard the scheduled command */
                dc = static_cast<tCommand *> (cmds.remove ());
                log (L_WARN, "Discarding command %x for module %s: conditions "
                        "for execution not satisfied.", dc->command, name);
                free_cmd (dc);
                break;

            case 0:
                /* Just wait */
                return;
            
            case 1:
                /* Okay, we may safely remove the command from the queue
                 * and execute it. */
                dc = static_cast<tCommand *> (cmds.remove ());
                exec_cmd (dc);
                free_cmd (dc);
                break;

            default:
                /* If it was a quick and dirty os what is it now? */
                ASSERT(false && "Developers, developers, developers, ...eh...");
        }
    }
}


// Examine module status of the running module
void tModule::examine_running_module ()
{
    debug_log ('3', "%s: checking status", name);

    /* The module is running, what about the connection status? */
    if (conn_status == CS_OK)
    {
        debug_log ('4', "Timing: Cur=%d, dead=%d", gb_current_time.tv_sec,
                deadline.tv_sec);

        /* We process only standard standalone modules */
        if (mod_role == MOD_ROLE_NORMAL)
        {
            /* If the time limit is reached since reconfiguration or
             * since the last pong response has been receive, schedule
             * a new ping */
            if ((last_cmd == CMD_NOP || last_cmd == CMD_RECONFIGURE) &&
                    timercmp (&gb_current_time, &deadline, >=))
            {
                tCommand *c = create_simple_command (CMD_PING);
                cmds.insert (c);
            }
            else if (last_cmd != CMD_STOP &&
                    timercmp (&gb_current_time, &deadline, >=))
            {
                /* This is a bit unpleasant since this is supposed to be
                 * a running module. */
                log (L_WARN, "Module %s is not responding.", name);
                conn_status = CS_NOT_RESPONDING;
            }
        }

    }
    else if (conn_status == CS_NOT_RESPONDING)
        /* Be kind. At first, try to reconfigure the module */
        mod_status = MOD_NEEDS_RECONFIG;
    else if (conn_status == CS_BROKEN)
    {
        /* Umm, try it harder */
        mod_status = MOD_NEEDS_RESTART;
        gb_echelog_sane = false;
    }

    /* Try to execute the commands from the module's command queue */
    process_cmd_queue ();
}


// Execute command from module's command queue
void tModule::exec_cmd (socks::tCommand *c)
{
    switch (c->command)
    {
        case CMD_PING:
            /*
             * Schedule a ping command dispatch
             */
            debug_log ('t', "%s: ping", name);
            sched_cmd_transfer (CMD_PING);
            ADVANCE_TIMER (&deadline, MAX_PING_RESPONSE_TIME);
            break;

        case CMD_PONG:
            /*
             * Schedule a pong command dispatch
             */
            debug_log ('t',"%s: pong", name);
            sched_cmd_transfer (CMD_PONG);
            break;

        case CMD_LOGIN_OK:
            /*
             * Schedule the successful login confirmation
             */
            debug_log ('t',"%s: confirming registration", name);
            sched_cmd_transfer (CMD_LOGIN_OK);
            break;

        case CMD_NOP:
            /*
             * Adjust module's timing. Set no operations for
             * a time interval between two pings
             */
            debug_log ('4',"%s: adjusting timing", name);
            last_cmd = CMD_NOP;
            ADVANCE_TIMER (&deadline, MAX_PING_INTERVAL);
            break;
            
        case CMD_START:
            /*
             * If the user is willing to, update module's configuration
             * and start the module
             */
            if (reload_config)
            {
                preprocess_configuration ();
                apply_new_configuration ();
            }
            start_module ();
            break;
            
        case CMD_STOP:
            /*
             * Just stop the module
             */
            stop_module ();
            break;
            
        case CMD_RECONFIGURE:
            /*
             * If the user is willing to, update module's configuration
             * and reconfigure the module
             */
            if (reload_config)
            {
                preprocess_configuration ();
                apply_new_configuration ();
            }
            reconfigure_module ();
            break;
            
        default:
            /*
             * Ignore unknown or invalid commands
             */
            log (L_ERROR, "Trying to execute an invalid command for module %s",
                    name);
    }
}


// Preprocess module's configuration
int tModule::preprocess_configuration ()
{
    struct_stat st;

    if (mod_status == MOD_NOT_MAINTAINED)
        return 0;

    /* Check if the original configuration exists and if it's readable */
    if (s_eaccess (config, F_OK | R_OK))
    {
        log (L_ERROR, "Can't access configuration file '%s' for module %s", 
                config, name);
        return -1;
    }

    /* Check if the top configuration file changed from the last update */
    s_stat (config, &st);
    if (config_timestamp != st.st_mtime)
    { 
        debug_log ('2', "%s: preprocessing configuration",
                name);

        /* Check updated configuration for parse errors */
        if (cf_preprocess (config, anc_config))
        {
            /* The file contains syntax errors, remove it */
            cf_report_error ();
            if (unlink (anc_config))
                log (L_ERROR, "Can't remove ancillary file '%s'", anc_config);
            return -1;
        }

        /* Remember a new modification time */
        s_stat (config, &st);
        config_timestamp = st.st_mtime;
    }
    
    /* Report success */
    return 0;
}


// Create module's overall and dependency configuration file.
int tModule::create_mod_config ()
{
    FILE *f;
    int fd;

    if (mod_status == MOD_NOT_MAINTAINED)
        return 0;
    
    debug_log ('3', "%s: creating mod configuration", name);
    
    /* Try to create a new overall configuration file */
    if ((fd = s_open (mod_config, O_WRONLY | O_CREAT | O_TRUNC, 0600)) < 0)
    {
      err_open_cache_file:
        log (L_ERROR, "Can't create cached configuration file '%s' "
                "for module %s: %m", mod_config, name);
        return -1;
    }

    if (!(f = fdopen (fd, "w")))
        goto err_open_cache_file;

    /* Escape the '.inc' configuration filename */
    char *escaped_inc_config = s_strnreplace (inc_config, "\"", "\\\"", 
            MAX_CONF_NAME_LEN);

    /* Prepend the include to the system section */
    fprintf (f, "include \"%s\"\n\n[System]\n", escaped_inc_config);

    /* Write in the input connection point */
    if (inputs.count () > 0)
        fprintf (f, "Input %s\n", input_sock_name);

    /* And all module's outputs */
    if (outputs.count() > 0)
    {
        tNode const *n;

        FOR_ALL_LIST_ITEMS (n, &outputs)
        {
            fprintf (f, "Output ");

            tModule *omod = (static_cast<tModule *>(n->data));
            fprintf (f, "%s \"%s\" ", omod->name, omod->input_sock_name);

            debug_log ('2', "Output[%s] = %s", name, omod->name);

            fprintf (f, "\n");
        }
    }

    /* Tidy up and close the overall configuration file */
    xfree (escaped_inc_config);
    if (fclose (f))
    {
        log (L_ERROR, "Error closing cached configuration file %s.",mod_config);
        return -1;
    }

    /* Report success */
    return 0;
}


// Apply module's updated configuration 
void tModule::apply_new_configuration ()
{
    if (mod_socket)
        mod_socket->set_app_specific_data (this);

    /* Okay, reset the forcing update flag */
    reload_config = false;
    
    if (mod_status == MOD_NOT_MAINTAINED)
        return;

    ASSERT (config && anc_config && inc_config);

    if (!s_eaccess (anc_config, F_OK | R_OK))
    {
        /* 
         * If there's a new cached '.anc' file, move it
         * to '.inc' file included from the cached configuration.
         * Thus it will be used on next module reconfiguration 
         * or restart.
         */
        log (L_INFO, "New configuration detected for module %s.", name);
        
        if (s_move_file (inc_config, anc_config) < 0)
            log (L_ERROR, "Error caching configuration to file '%s'.", 
                    inc_config);
    }
    else if (s_eaccess (inc_config, F_OK))
    {
        /*
         * Create at least an empty '.inc' file to avoid confusing
         * log messages from maintained echelog modules.
         */
        debug_log ('1', "%s: creating empty inc configuration file", name);
        int fd = s_open (mod_config, O_WRONLY | O_CREAT | O_TRUNC, 0600);
        s_close (fd);
    }
}


// Set all external module's resources free
void tModule::configuration_cleanup ()
{
    debug_log ('1', "%s: configuration cleanup", name);
    
    if (mod_status != MOD_NOT_MAINTAINED)
    {
        /* Delete all module's cached configuration files */
        if (!s_eaccess (mod_config, F_OK) && unlink (mod_config))
            log (L_ERROR, "Can't remove old cached configuration file "
                    "'%s'.", mod_config);

        if (!s_eaccess (anc_config, F_OK) && unlink (anc_config))
            log (L_ERROR, "Can't remove old ancillary configuration file "
                    "'%s'.", anc_config);

        if (!s_eaccess (inc_config, F_OK) && unlink (inc_config))
            log (L_ERROR, "Can't remove old cached configuration file "
                    "'%s'.", inc_config);
    }

    /* If there is still an open connection clean it up */
    socket_cleanup ();
}


// Constructor---Creates a new empty inactive module
tModule::tModule (lib::tMemPool &cpool, tConfiguration &conf, 
        const char *modname, const char *cconfig) 
:   pool (cpool),
    rconf (conf),
    mod_socket (NULL), 
    conn_status (CS_EMPTY), 
    mod_status (MOD_EMPTY),
    reload_config (false),
    id (-1), 
    capability (CP_NONE), 
    mod_used (false),
    mod_role (MOD_ROLE_NORMAL),
    binary (NULL), 
    config_timestamp (0),
    input_sock_name (NULL),
    pid (1),
    level (-1), 
    cmds (cpool), 
    inputs (cpool), 
    outputs (cpool),
    last_cmd (CMD_NOP)
{
    /* Assign it a new name */
    name = s_strdup (modname, pool);

    if (cconfig)
    {
        /* Create a standard cache configuration filename scheme */
        config = s_strdup (cconfig, pool);
        mod_config = s_strconcat (pool, ECHELOG_CACHE_DIR, "/", name, NULL);
        inc_config = s_strconcat (pool, mod_config, ".inc", NULL);
        anc_config = s_strconcat (pool, mod_config, ".anc", NULL);
    }
    else
        /* This is a dummy module */
        config = mod_config = inc_config = anc_config = NULL;

    /* Clear the deadline to avoid undesirable effect in configuration */
    timerclear (&deadline);
}


// Destructor---Releases all unused module's resources
tModule::~tModule ()
{
    /* Release the rest of used heap memory */
    if (input_sock_name)
        xfree (input_sock_name);

    if (mod_config)
    {
        /* 
         * If this is not a dummy module and if the module doesn't have
         * successor in current configuration clean out also the old 
         * configuration stuff.
         */
        if (!gb_eenv.current_configuration->get_module (name))
            configuration_cleanup ();
    }
}


// Checks if the module has a capability to execute the specified command.
bool tModule::has_exec_capability (const socks::tCMD_TYPE command) const
{
    /* Privileged commands execution is allowed only to privileged modules */
    if (cmd_is_privileged (command) && capability != CP_SYS_COMMANDS)
        return false;

    /* If the module has no real rights deny the execution */
    if (capability == CP_NONE)
        return false;

    /* Normal commands are allowed to all but selected modules */
    return true;
}


// Schedules module's registration confirmation
void tModule::exec_register_module (socks::tCtrlUNIXSocketNB *s)
{
    debug_log ('2', "%s: registering connection (fd=%ld)", name, s->get_fd());

    /* Associate the socket with the module and reset the module status */
    mod_socket = s;
    conn_status = CS_OK;
    if (mod_status != MOD_NOT_MAINTAINED)
        mod_status = MOD_OK;

    /* And send schedule the login confirmation message to the module */
    cmds.insert (create_simple_command (CMD_LOGIN_OK));
}


// Command sent notifier function
void tModule::exec_cmd_sent_notification ()
{
    debug_log ('3', "%s: command sent notification", name);

    /* Don't manipulate the connection in an uknown state */
    if (conn_status == CS_BUSY)
    {
        /* Reset the busy flag */
        conn_status = CS_OK;

        if (!cmd_needs_confirmation (last_cmd))
        {
            /* If the command doesn't need a confirmation */
            if (!cmds.count())
            {
                /* Set no operation to enforce a holding pattern */
                tCommand *c = create_simple_command (CMD_NOP);
                enqueue_cmd (c);
            }
            
            /* Process the next command from the module's command queue */
            process_cmd_queue ();
        }
    }
}


// Executes pong handler routine
void tModule::exec_receive_pong ()
{
    if (mod_status == MOD_OK || 
            (mod_status == MOD_NEEDS_RECONFIG &&  
            last_cmd == CMD_PING))
    {
        /* Reset the module to the default running state */
        set_conn_status (CS_OK);
        set_mod_status (MOD_OK);
        cmds.insert (create_simple_command (CMD_NOP));
    }
}


// Executes toggle handler routine
void tModule::exec_receive_toggle (const char *pathname)
{
    tNode const *c;
    tModule *mod;

    /* At first, test if this is a valid flowswitch at all */
    if (mod_status != MOD_OK || mod_role != MOD_ROLE_FLOWSWITCH)
    {
        log (L_WARN, "Can't toggle path on %s. Module is not a FlowSwitch.",
                name);
        return ;
    }

    if (!(mod = gb_eenv.current_configuration->get_module (pathname)))
    {
        /* Undefined path (probably a misspelled pathname) */
        log (L_ERROR, "Invalid path '%s' specified in toggle "
                "command received from module %s.", pathname, name); 
        return ;
    }

    /* If so, set a new path */
    bool found = false;
    FOR_ALL_LIST_ITEMS (c, &outputs)
    {
        tModule *pmod = static_cast<tModule *> (c->data);

        if (pmod == mod) /* The module mod is output module of this */
        {
            xfree (input_sock_name);
            input_sock_name = s_strdup (pmod->input_sock_name);
            found = true;
            break;
        }
    }

    if (!found)
    {
        /* Undefined path (probably a misspelled pathname) */
        log (L_ERROR, "Invalid path '%s' specified in toggle "
                "command received from module %s.", pathname, name); 
        return ;
    }

    /* And recalculate the path for all the subsequent modules.
     * Adjust module states accordingly. */
    toggle_module ();
}


// Cleans up the socket and removes module from active configuration
void tModule::socket_cleanup ()
{
    /* Check if there still exists an active connection */
    if (mod_socket)
    {
        debug_log ('2', "%s: socket cleanup (fd=%ld)", name, 
                mod_socket->get_fd());

        /* Clear out app_specific_data to avoid memory corruption
         * on socket delete */
        mod_socket->set_app_specific_data (NULL);

        /* Unregister and release the socket */
        if (mod_socket->get_smgr()->unregister_socket (mod_socket) < 0)
            log (L_ERROR, "Can't unregister socket (fd=%ld).",
                    mod_socket->get_fd());
        if (mod_socket->get_created_by_smgr ())
            delete mod_socket;

        /* Disassociate the socket */
        mod_socket = NULL;

        /* Remove the module from the active configuration */
        if (mod_status != MOD_NOT_MAINTAINED)
            rconf.set_active_modules (rconf.get_active_modules() - 1);
    }

    /* Reset the connection status */
    conn_status = CS_EMPTY;
}


// Checks module's status and processes its commmand queue
void tModule::process ()
{
    switch (mod_status)
    {
        case MOD_NOT_MAINTAINED:
        case MOD_EMPTY:
        case MOD_BROKEN:
        case MOD_NEEDS_START:
        case MOD_NEEDS_STOP:
        case MOD_NEEDS_RESTART:
            /* 
             * Do nothing but process the waiting commands in the queue.
             */
            process_cmd_queue ();
            break;

        case MOD_OK:
            /*
             * Process the command queue and take proper actions
             * if the running module does not behave as expected.
             */
            process_cmd_queue ();
            examine_running_module ();
            break;

        case MOD_STARTED:
            /*
             * If the just started module doesn't respond in the specified
             * time interval, consider it broken and stop it.
             */
            if (timercmp (&gb_current_time, &deadline, >))
            {
                log (L_WARN, "Module %s is not responding. Stopping it.", name);
                kill_module ();
                mod_status = MOD_ZOMBIE;
                log (L_ERROR, "Removing broken module '%s' from active "
                        "configuration.", name);
                empty_cmd_queue ();
            }
            else
                process_cmd_queue ();
            break;

        case MOD_PASSING_AWAY:
            /*
             * If the time limit is reached and the module still hasn't
             * terminated use a little bit of violence 
             */
            if (timercmp (&gb_current_time, &deadline, >))
            {
                finish_module ();
                process_cmd_queue ();
            }
            break;

        case MOD_NEEDS_RECONFIG:
            /*
             * Reconfigure or restart a not respoding module.
             */
            if (last_cmd != CMD_RECONFIGURE)
                /* We're about to reconfigure it */
                reconfigure_module ();
            else if (last_cmd == CMD_PING)
            {
                /*
                 * If the module doesn't respond even after reconfiguration
                 * try to restart it.
                 */
                if (timercmp (&gb_current_time, &deadline, >=))
                {
                    gb_echelog_sane = false;
                    mod_status = MOD_NEEDS_RESTART;
                }
            }
            else
            {
                /* Fine, it's much better now */
                tCommand *c = create_simple_command (CMD_PING);
                cmds.insert (c);
                conn_status = CS_OK;
                process_cmd_queue ();
            }
            break;

        case MOD_ZOMBIE:
            /* 
             * Finish him!
             */
            finish_module ();
            process_cmd_queue ();
            break;

        default:
            /* What a jerk! */
            log (L_FATAL, "%s: weird module status %x.", name, mod_status);
            ASSERT (false);
    }
}


// Put command to module's commmand queue
void tModule::enqueue_cmd (socks::tCommand *c)
{
    ASSERT (c);
    cmds.insert (c);
}


// Remove all commands from the module's command queue
void tModule::empty_cmd_queue ()
{
    tCommand *c;

    debug_log ('2', "%s: trashing cmd queue", name);

    while (cmds.count ())
    {
        c = static_cast<tCommand *> (cmds.remove ());
        free_cmd (c);
    }
}


// Clears module flag states for a stop
void tModule::clear_chain_flags (bool reload_config_flag)
{
    tNode const *n;

    if (mod_status == MOD_NEEDS_START)
    {
        /* A new module, okay, mark it as empty. Clear out the module's
         * command queue to avoid confusion. */
        set_mod_status (MOD_EMPTY);
        empty_cmd_queue ();
    }
    else if (mod_status != MOD_BROKEN && 
            mod_status != MOD_EMPTY &&
            mod_status != MOD_PASSING_AWAY && 
            mod_status != MOD_ZOMBIE) 
    {
        /* This should be a final state. Forbid the socket I/O and
         * mark the module to be stopped. */
        debug_log ('3', "%s: clearing reconfig/restart flag", name);
        mod_status = MOD_NEEDS_STOP;
        conn_status = CS_BUSY;

        /* Trash the module's command queue. It'll be of no use, anyway */
        empty_cmd_queue ();
    }

    /* If the users is willing to restart, propage the
     * forcing reload flag through the chain */
    if (reload_config_flag)
        reload_config = true;

    /* Recursively clean all dependent modules */
    FOR_ALL_LIST_ITEMS (n, &inputs)
    {
        tModule *cmod = static_cast<tModule *>(n->data);
        cmod->clear_chain_flags (reload_config);
    }
}


// Call functor <tt>func</tt> on all of module's inputs.
int tModule::for_all_inputs (int (*func) (tModule &mod))
{
    int r = 0;
    tNode const *n;

    FOR_ALL_LIST_ITEMS (n, &inputs)
        r += func (*static_cast<tModule *>(n->data));

    /* Return the accumulator */
    return r;
}


// Call functor <tt>func</tt> on all of module's outputs.
int tModule::for_all_outputs (int (*func) (tModule &mod))
{
    int r = 0;
    tNode const *n;

    FOR_ALL_LIST_ITEMS (n, &outputs)
        r += func (*static_cast<tModule *>(n->data));

    /* Return the accumulator */
    return r;
}
