/**
 * @file
 * @brief Module Management
 *
 * This is the only module which is started by an administrator. It reads
 * the echelog configuration and and accordingly to this file it maintains
 * all other modules.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_ECHELOGD_MODULE_H
#define ECHELOG_ECHELOGD_MODULE_H

#include "property.h"
#include "lib/lib.h"
#include "socks/socketmgr.h"
#include "socks/ctrl_unixsocket_nb.h"
#include "echelogd_config.h"

#include <time.h>
#include <sys/time.h>

/**
 * @brief Advances ::timeval time counter by a certain number of seconds.
 */
#define ADVANCE_TIMER(t,n_sec) (*(t)) = gb_current_time, (t)->tv_sec += (n_sec)


/** 
 * @brief Possible connection states.
 */
typedef enum {
    CS_EMPTY = 0,       ///< No connection
    CS_NEW,             ///< New connection
    CS_IDENT,           ///< Client announces its id
    CS_OK,              ///< Connection established and working properly
    CS_BUSY,            ///< Ctrl command is being transmitted
    CS_NOT_RESPONDING,  ///< Connected client is not responding
    CS_BROKEN,          ///< Connection broken
} tConnectionStatus;


/** 
 * @brief Possible module states.
 */
typedef enum {
    MOD_OK = 0,         ///< Module okay
    MOD_EMPTY,          ///< Empty module
    MOD_STARTED,        ///< Module just started
    MOD_PASSING_AWAY,   ///< Module is being killed
    MOD_ZOMBIE,         ///< Module is undead
    MOD_NEEDS_START,    ///< Module needs to be started
    MOD_NEEDS_STOP,     ///< Module needs to be stopped
    MOD_NEEDS_RECONFIG, ///< Module needs reconfiguration
    MOD_NEEDS_RESTART,  ///< Module needs to be restarted
    MOD_BROKEN,         ///< Broken module
    MOD_NOT_MAINTAINED  ///< Module is not maintained by overseer
} tModuleStatus;


/** 
 * @brief Possible module roles.
 */
typedef enum {
    MOD_ROLE_NORMAL,    ///< Classic echelog module
    MOD_ROLE_CONTROL,   ///< Echelog control module
    MOD_ROLE_FLOWSWITCH ///< Echelog flowswitch
} tModuleRole;


/**
 * @brief Possible module capabilities.
 */
typedef enum {
    CP_NONE = 0,                ///< No capability
    CP_NORMAL = 10,             ///< Can execute common commands
    CP_SYS_COMMANDS = 100       ///< Can execute common and privileged commands
} tCAPABILITY;


/**
 * @brief Main interface for control of modules
 *
 * This class defines a single echelog module and provides a powerful
 * interface to operate with it.
 */
class tModule {

        /**
         * @brief We suppose a configuration to define most aspects of modules
         */
        friend class tConfiguration;

    protected:

        /**
         * @brief Memory pool for internal module's stuff.
         */
        lib::tMemPool &pool;
        
        /**
         * @brief Reference to a configuration the module belongs to.
         */
        tConfiguration &rconf;

        /**
         * @brief <em>External resource:</em> General associated socket.
         */
        SIMPLE_EXT_RESOURCE (socks::tCtrlUNIXSocketNB *, mod_socket);

        /**
         * @brief <em>Property:</em> Socket connection status.
         */
        SIMPLE_PROPERTY (tConnectionStatus, conn_status);

        /**
         * @brief <em>Property:</em> Module main status.
         */
        SIMPLE_PROPERTY (tModuleStatus, mod_status);

        /**
         * @brief Flag if the configuration update should be considered.
         */
        SIMPLE_PROPERTY (bool, reload_config);

        /**
         * @brief <em>Constant read-only property:</em> Module's name.
         */
        SIMPLE_R_CONST_PROPERTY (char *, name);

        /**
         * @brief <em>Read-only property:</em> Module's main identificator.
         */
        SIMPLE_R_PROPERTY (long, id);

        /// <em>Read-only property:</em> Module's capability.
        SIMPLE_R_PROPERTY (tCAPABILITY, capability);

        /** 
         * @brief <em>Read-only property:</em> Flag if the module is actively
         * used in configuration.
         */
        SIMPLE_R_PROPERTY (bool, mod_used);

        /**
         * @brief Module's role.
         */
        tModuleRole mod_role;

        /**
         * @brief Path to module's program binary.
         */
        const char *binary;

        /**
         * @brief Specified configuration file.
         */
        const char *config;

        /**
         * @brief Timestamp of specified configuration file
         */
        time_t config_timestamp;

        /**
         * @brief Module's echelogd overall and dependency configuration file.
         */
        const char *mod_config;

        /**
         * @brief Active module configuration file.
         */
        const char *inc_config;

        /**
         * @brief A candidate for the active module configuration file.
         */
        const char *anc_config;

        /**
         * @brief Defined module's input socket name.
         */
        char *input_sock_name;

        /**
         * @brief Process ID.
         */
        pid_t pid;

        /**
         * @brief Dependency level.
         */
        long level;

        /**
         * @brief Module's command queue.
         */
        lib::tListQueue cmds;

        /**
         * @brief List of all module's inputs.
         */
        lib::tList inputs;

        /**
         * @brief List of all module's outputs.
         */
        lib::tList outputs;

        /**
         * @brief Last executed command.
         */
        socks::tCMD_TYPE last_cmd;

        /**
         * @brief Command expiration time.
         */
        struct timeval deadline;

    protected:
        
        /**
         * @brief Runs echelog module.
         *
         * <tt>fork()s</tt> and <tt>exec()s</tt> the module's binary. The
         * program binary name is changed to a module's name in <tt>ps</tt>
         * listing for administrator's convenience. Module's standard error
         * output is redirected to <tt>gb_eenv.modules_fd</tt>, the environment
         * is cleared.
         *
         * @note If the module's binary <tt>exec()</tt> failed, the program is
         * terminated with a non-zero status.
         *
         * @return The function returns zero upon a successful completion.
         * If the <tt>fork()</tt> syscall failed, <tt>-1</tt> is returned.
         *
         */
        int run_module ();

        /**
         * @brief Kills echelog module.
         * 
         * Tries to send a quit signal to echelog module. The module
         * is awaken before it's stopped. After this, the module still
         * needs to be <tt>finish()ed</tt>.
         *
         * @return The function returns zero upon a successful completion.
         * If an error occured, <tt>-1</tt> is returned.
         */
        int kill_module ();

        /**
         * @brief Finishes module and sets its process ID free.
         * 
         * Tests if the module is still running and terminates it if so.
         * The module's PID entry is cleared, to module's connection is closed
         * and unregistered.
         *
         * @return The function returns zero upon a successful completion.
         * If an error occured, <tt>-1</tt> is returned.
         */
        int finish_module ();


        /**
         * @brief Schedules start of echelog module.
         *
         * If the module has not already started, the function starts
         * the module and set it's status appropriately.
         */
        void start_module ();

        /**
         * @brief Schedules stop of echelog module.
         *
         * If the module has not been already stopped, try to stop it.
         * Usually, fhe function sends a quit signal to module's program.
         */
        void stop_module ();

        /**
         * @brief Reconfigures echelog module.
         * 
         * This function tries to sent a module a reconfiguration signal.
         * If the module's program has been accidentally stopped it's awaken
         * at first.
         */
        void reconfigure_module ();

        /**
         * @brief Toggles chain path at FlowSwitch module.
         *
         * Adjusts all modules affected by the flowswitch toggle and
         * marks them for reconfiguration or restarts them accordingly.
         */
        void toggle_module ();

        /**
         * @brief Schedules transfer of command to echelog module.
         *
         * The command is scheduled for a transfer to the module.
         * The connection is marked busy to avoid clashes.
         */
        void sched_cmd_transfer (socks::tCMD_TYPE command);

        /**
         * @brief Checks if conditions are satisfied or not in order 
         * to execute a command waiting in module's queue.
         *
         * This function is used to examine a delayed command <tt>cmd</tt>
         * from the module's command queue and decides how to handle the
         * command.
         * 
         * @return If the conditions seem to be suitable for commands
         * execution, <tt>1</tt> is returned. If it can't be determined right
         * now <tt>0</tt> is returned. If the command execution makes no sence
         * or would cause problems, <tt>-1</tt> is returned.
         */
        int check_cmd_conditions (socks::tCommand *cmd) const;

        /**
         * @brief Processes module's command queue.
         *
         * Processes all commands from the module's command queue. Each
         * command is examined by <tt>check_cmd_conditions</tt> function
         * in order to find out whether to execute, delay or discard 
         * the command.
         */
        void process_cmd_queue ();

        /**
         * @brief Examine module status of the running module.
         *
         * The running module may have its own problems. It may
         * loose a connection to its subsequent module or it may
         * segfault on an unexpected input or whatever. It's a good
         * idea to periodically check the module and to take proper
         * actions if its not responding.
         */
        void examine_running_module ();

        /**
         * @brief Execute a command from module's command queue.
         *
         * This can be used for a factual execution of a command
         * removed from module's command queue.
         */
        void exec_cmd (socks::tCommand *c);

        /**
         * @brief Preprocess module's configuration
         *
         * The function checks if the top-configuration file changed since
         * the last update and if so it tries to make a new parsed
         * configuration file.
         *
         * @return
         * The function returns zero upon successful creation of
         * <tt>'.anc'</tt> cached file; on error <tt>-1</tt> is returned.
         */
        int preprocess_configuration ();

        /**
         * @brief Create module's overall and dependency configuration file.
         * 
         * In compliance with echelog configuration create the basic
         * module's configuration file containing an <tt>include</tt> of the
         * parsed module's configuration file and the definitions
         * of system connection nodes.
         *
         * @return
         * The function returns zero upon successful creation of the
         * cached configuration file; on error <tt>-1</tt> is returned.
         */
        int create_mod_config ();

        /**
         * @brief Apply module's updated configuration.
         * 
         * If there's a new cached configuration file, owerwrite the current
         * active configuration. Thus it will be used on next module
         * reconfiguration or restart.
         *
         * @return
         * The function returns zero upon successful creation of the
         * cached configuration file; on error <tt>-1</tt> is returned.
         */
        void apply_new_configuration ();

        /**
         * @brief Set all external module's resources free.
         * 
         * The function deletes all module's cached configuration files
         * and if there is still an open connection it's also cleaned up.
         */
        void configuration_cleanup ();

    public:

        /**
         * @brief Constructor---Creates a new empty inactive module.
         */
        tModule (lib::tMemPool &cpool, tConfiguration &conf, 
                const char *modname, const char *cconfig);
        /**
         * @brief Destructor---Releases all unused module's resources.
         */
        virtual ~tModule ();


        /**
         * @brief Checks if the module has a capability to execute 
         * the specified <tt>command</tt>.
         *
         * Privileged commands execution is allowed only to privileged modules.
         * Normal commands are allowed to all but selected modules.
         *
         * @return
         * The function returns true if the rights to execute the 
         * <tt>command</tt> are granted, false otherwise.
         */
        bool has_exec_capability (const socks::tCMD_TYPE command) const;

        /**
         * @brief Schedules module's registration confirmation.
         *
         * The function associates the socket with the module
         * and schedules the login confirmation message to the module.
         */
        void exec_register_module (socks::tCtrlUNIXSocketNB *s);

        /**
         * @brief Command sent notifier function.
         *
         * This is a part of socket serialization handling. If the
         * connection is valid the busy flag is reset and thus the
         * next command may be scheduled.
         */
        void exec_cmd_sent_notification ();

        /**
         * @brief Executes pong handler routine.
         *
         * Accordingly to module state the routine either ignores
         * the received pong or resets the module to the default running
         * state.
         */
        void exec_receive_pong ();

        /**
         * @brief Executes toggle handler routine.
         *
         * If the module is a valid flowswitch and if the <tt>pathname</tt>
         * denotes a valid path, set a new dataflow.
         */
        void exec_receive_toggle (const char *pathname);

        /**
         * @brief Cleans up the socket and removes module from active
         * configuration.
         *
         * If there still exists an active connection, disassociate it
         * and release all used resources.
         */
        void socket_cleanup ();


        /**
         * @brief Checks module's status and processes its command queue.
         * 
         * The function examines the actual module state and tries to
         * prepare the best pattern what to do. 
         */
        void process ();

        /**
         * @brief Put command to module's command queue.
         *
         * Enqueue the command <tt>c</tt> to module's command queue.
         */
        void enqueue_cmd (socks::tCommand *c);

        /**
         * @brief Flush the module's command queue.
         *
         * Remove all commands from the module's command queue.
         */
        void empty_cmd_queue ();

        /**
         * @brief Clears module flag states for a stop.
         *
         * This routine is used to clear out states of all modules in the chain
         * under this module (including it). The final module states
         * are cleared out to minimize action needed to make the system sane.
         * The <tt>reload_config_flag</tt> is set if the user is willing 
         * to consider a configuration update.
         */
        void clear_chain_flags (bool reload_config_flag);

        /**
         * @brief Call functor <tt>func</tt> on all of module's inputs.
         *
         * The function calls the functor <tt>func</tt> on all of module's
         * inputs and sums the results.
         */
        int for_all_inputs (int (*func) (tModule &mod));

        /**
         * @brief Call functor <tt>func</tt> on all of module's outputs.
         *
         * The function calls the functor <tt>func</tt> on all of module's
         * outputs and sums the results.
         */
        int for_all_outputs (int (*func) (tModule &mod));
};


#endif
