/**
 * @file
 * @brief Echelogd Processing
 *
 * This is the only module which is started by an administrator. It reads
 * the echelog configuration and and accordingly to this file it maintains
 * all other modules.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */                                                                             

#include "echelogd.h"


#ifndef DOXY_SKIP
using namespace lib;
using namespace socks;

// Discrete time counter
struct timeval gb_current_time;
#endif


// The listening echelogd control channel
static tCtrlUNIXSocketNB *ctrl_channel;

// Echelogd main socket manager
static tSocketMgr smgr;

// List of incoming connections
static tList new_clients;


/**
 * @brief Gets adminitrator UID.
 *
 * Translates ECHELOG_ADMIN into system UID or restricts 
 * the administrator to root only.
 */
static uid_t get_admin_uid ()
{
    uid_t admin_uid;

    /* Translate the admin's name */
    if (s_name2uid (ECHELOG_ADMIN, admin_uid))
    {
        log (L_WARN, "Can't determine admin's UID. Privileged commands can "
                "be only run by root.");
        return 0;
    }

    /* Okay, allow privileged commands from this user account */
    return admin_uid;
}


/**
 * @brief Retrieves socket capabilities.
 *
 * Returns a capability which can be granted to a unix domain socket
 * <tt>sock</tt>. The mechanism is implemented via unix domain socket
 * authentication (currently implemented only on Linux and *BSD).
 */
static tCAPABILITY retrieve_capability (tUNIXSocketNB *sock)
{

#if defined(__SUNOS__)
    /* 
     * Unix domain socket authentication is not supported on SunOS.
     * Disable all checks. Later, this could be implemented via doors 
     * mechanism. 
     */
    return CP_SYS_COMMANDS;

#else

    /* Cache systems UIDs */
    static uid_t admin_uid = get_admin_uid ();
    static uid_t my_uid = getuid ();
    uid_t auth_uid = sock->get_auth_id ();

    if (!auth_uid || auth_uid == admin_uid)
        /* If the client socket end is owned by root or by an admin grant
         * all capabilities. */
        return CP_SYS_COMMANDS;
    else if (auth_uid == my_uid)
        /* If the module is run under the same user is echelogd core
         * grant normal module capabilities. */
        return CP_NORMAL;
    else
        /* Grant no capabilities otherwise */
        return CP_NONE;

#endif
}


/**
 * @brief Simple functor for socket reference comparison
 *
 * Returns zero if the pointers refer to the same socket, 
 * non-zero otherwise 
 */
static int conn_socket_comp (const void *a, const void *b)
{
    const tConnection *x = static_cast<const tConnection *> (a);
    const tConnection *y = static_cast<const tConnection *> (b);

    if (x->conn_socket == y->conn_socket)
        return 0;

    return 1;
}


/**
 * @brief Default socket module handler for incoming data.
 *
 * Tries to receive a command from the socket to the control's 
 * input buffer.
 */
static void on_read (tSocketNB *sock)
{
    tCtrlUNIXSocketNB *s = static_cast<tCtrlUNIXSocketNB *> (sock);

    s->receive_cmd ();
}


/**
 * @brief Default module socket handler for available data on socket.
 *
 * Tries to receive a command from the socket. Then it retrieves
 * associated module and if a valid commands has been received 
 * it calls the proper handler.
 */
static void on_read_notify (tSocketNB *sock)
{
    tCtrlUNIXSocketNB *s = static_cast<tCtrlUNIXSocketNB *> (sock);
    tModule *mod = static_cast<tModule *>(s->get_app_specific_data ());
    tCommand *c = s->get_cmd ();
    ASSERT (mod);

    /* Check if the command has been retrieved successfully */
    if (!c)
    {
        log (L_WARN, "Invalid control packet received from module %s. "
                "Discarding it.", mod->get_name());
        return ;
    }

    /* Launch the handler */
    exec_sys_receive_cmd (*mod, c);
}


/**
 * @brief Default module socket write completion notifier.
 *
 * It retrieves associated module and lets it know that the command 
 * has been sent successfully.
 */
static void on_write_notify (tSocketNB *sock)
{
    tModule *mod = static_cast<tModule *>
        (sock->get_app_specific_data ());
    ASSERT (mod);

    /* Launch the module handler */
    mod->exec_cmd_sent_notification ();
}


/**
 * @brief Default module socket hangup handler
 *
 * It catches up the socket event, adjusts the module status appropriately
 * and cleans the socket up.
 */
static void on_hup (tSocketNB* sock)
{
    tModule *mod = static_cast<tModule *>
        (sock->get_app_specific_data ());
    ASSERT (mod);

    switch (mod->get_mod_status())
    {
        case MOD_PASSING_AWAY:
            /* 
             * Okay, this hangup is what we expect. Just finish the module 
             */
            mod->set_mod_status (MOD_ZOMBIE);
        case MOD_NOT_MAINTAINED:
        case MOD_NEEDS_STOP:
        case MOD_NEEDS_RESTART:
            /*
             * This is a bit unusuall but it should make no harm
             */
            debug_log ('1', "Closing connection for module %s", 
                    mod->get_name ());
            mod->set_conn_status (CS_EMPTY);
            break;

        default:
            /*
             * Something went wrong, the control connection shouldn't break
             * while the module is in this state. The system may need
             * a sanitization.
             */
            log (L_WARN, "Module %s error: broken connection.", mod->get_name());
            mod->set_conn_status (CS_BROKEN);
            mod->set_mod_status (MOD_NEEDS_RESTART);
            gb_echelog_sane = false;
    }

    mod->socket_cleanup ();
}


/**
 * @brief Default module socket error handler
 *
 * It catches up the socket error event, adjusts the module status
 * appropriately and cleans the socket up.
 */
static void on_err (tSocketNB* sock)
{
    tModule *mod = static_cast<tModule *>
        (sock->get_app_specific_data ());
    ASSERT (mod);

    /* This probably means that some data were still unread from
     * the socket when a hangup occured. */
    log (L_WARN, "Module %s error: connection reset.", mod->get_name());

    switch (mod->get_mod_status ())
    {
        case MOD_PASSING_AWAY:
            /*
             * We don't care what an undead is trying to say to us.
             */
            mod->set_mod_status (MOD_ZOMBIE);
        case MOD_NOT_MAINTAINED:
        case MOD_NEEDS_STOP:
        case MOD_NEEDS_RESTART:
            /*
             * This also shouldn't cause any problems if the connection
             * is properly cleaned up
             */
            mod->set_conn_status (CS_EMPTY);
            break;

        default:
            /*
             * Something is very wrong here. Better restart the module.
             */
            mod->set_conn_status (CS_BROKEN);
            mod->set_mod_status (MOD_NEEDS_RESTART);
    }

    /* This was an error not a hangup, it'd be better to force 
     * a system sanitization */
    gb_echelog_sane = false;

    /* Okay, cleanup the module's connection */
    mod->socket_cleanup ();
}


/**
 * @brief Default socket hangup handler for newly accepted connections.
 *
 * Since we generally don't care about a random (unidentified) connection it's
 * rather safe to just cleanup the connection.
 */
static void on_login_hup (tSocketNB *sock)
{
    if (sock->get_role() == AS_SOCK_SERVER)
    {
        tSocketMgr *m = sock->get_smgr();
        if (m) m->cleanup ();
        delete ctrl_channel;
        die ("Server listening socket broken. Aborting.");
    }
    else
    {
        /* Unregister and release the socket and associated buffers */
        tSocketMgr *m = sock->get_smgr();
        if (m) 
            if (m->unregister_socket (sock) < 0)
            log (L_ERROR, "Can't unregister socket (fd=%ld).", sock->get_fd());
        log (L_INFO,"Client connection closed.");
    }
}


/**
 * @brief Default socket error handler for newly accepted connections.
 *
 * Since we generally don't care about a random (unidentified) connection it's
 * rather safe to just cleanup the connection.
 */
static void on_login_err (tSocketNB *sock)
{
    if (sock->get_role() == AS_SOCK_SERVER)
    {
        tSocketMgr *m = sock->get_smgr();
        if (m) m->cleanup ();
        delete ctrl_channel;
        die ("Server listening socket broken. Aborting.");
    }
    else
    {
        /* Unregister and release the socket and associated buffers */
        tSocketMgr *m = sock->get_smgr();
        if (m)
            if (m->unregister_socket (sock) < 0)
            log (L_ERROR, "Can't unregister socket (fd=%ld).", sock->get_fd());
        log (L_INFO,"Error on client connection socket cleared.");
    }
}


// Default login connection handler for newly accepted connections.
static void on_login_connect (tSocketNB* sock)
{
    debug_log ('2', "Client connection completed on socket %li.", 
            sock->get_fd ());
}

 
/**
 * @brief Default socket module handler for incoming data on newly accepted
 * connections.
 *
 * When data arrives on a newly accepted socket connection the handler 
 * tries to read an identification request.
 */
static void on_login_read (tSocketNB *sock)
{
    if (sock->get_role() == AS_SOCK_CLIENT && sock->get_app_specific_data ())
    {
        tCtrlUNIXSocketNB *s = static_cast<tCtrlUNIXSocketNB *> (sock);
        tConnection *c = static_cast<tConnection *>(s->get_app_specific_data());

        debug_log ('6', "Login READ");

        if (c->status == CS_NEW)
        {
            /* Adjust the notification status and begin the command receival */
            c->status = CS_IDENT;
            s->receive_cmd ();
        }
    }
}


/**
 * @brief Default socket module handler for available data on newly accepted
 * connections.
 *
 * Tries to identify and authenticate a module requesting a registration.
 * If the module uses the same protocol version, if it does have required
 * capabilities and is not already logged in, it is registered and a
 * confirmation reply is sent to it.
 *
 * Otherwise the module registration is refused and the connection is closed.
 */
static void on_login_read_notify (tSocketNB *sock)
{
    tCtrlUNIXSocketNB *s = static_cast<tCtrlUNIXSocketNB *> (sock);
    tConnection *c = static_cast<tConnection *> (s->get_app_specific_data ());

    debug_log ('6', "Login READ FINISHED");

    if (sock->get_role() == AS_SOCK_CLIENT && c)
    {
        /* Check if the connection is in a know state, ignore it otherwise */
        if (c->status == CS_IDENT)
        {
            if (s->test_received_cmdtype (CMD_IDENTIFY))
            {
                /* Invalid agent information */
                s->clear_io ();
                s->set_status (SS_BROKEN);

                /* Used ancilliary structures are released on unregistration */
                s->err_handler ();
                return;
            }

            /* The complete command received */
            c->status = CS_OK;

            tCommand *cmd = s->get_cmd ();
            char *modname = static_cast<char *> (cmd->d_ptr);
            tModule *mod = gb_eenv.current_configuration->get_module (modname);

            /* New module, try register it */
            if (mod)
            {
                if (mod->get_conn_status() != CS_EMPTY &&
                        mod->get_conn_status() != CS_BROKEN)
                {
                    log (L_WARN, "Module %s already connected to echelogd. "
                            "Refusing new connection.", mod->get_name());
                    s->get_smgr()->unregister_socket (s);
                    free_cmd (cmd);
                    return;
                }

                /* Check if the socket has enough capabilities the module
                 * requires to have */
                if (mod->get_capability() <= retrieve_capability (s))
                {
                    void *item_removed = new_clients.remove_match 
                        (c, conn_socket_comp);
                    ASSERT (item_removed);

                    /* Free old info and set mod association */
                    xfree (item_removed);
                    s->set_app_specific_data (mod);

                    /* Set up the socket for normal communication */
                    s->set_on_hup (on_hup);
                    s->set_on_err (on_err);
                    s->set_on_read_available (on_read);
                    s->set_on_read_finished (on_read_notify);
                    s->set_on_write_finished (on_write_notify);
                    s->set_on_register (NULL);
                    s->set_on_unregister (NULL);
                    s->set_on_connected (NULL);

                    /* Acknowledge the registration command and send schedule
                     * a confirmation message to the module */
                    mod->exec_register_module (s);
                }
                else
                {
                    /* Unsatisfied capabilites for module */
                    log (L_WARN, "Connection for module %s refused: "
                            "permission denied.", modname);
                    s->get_smgr()->unregister_socket (s);
                }
            }
            else
            {
                /* Undefined module */
                log (L_WARN, "Unknown module '%s'. Removing connnection.", 
                        modname);
                s->get_smgr()->unregister_socket (s);
            }

            /* Release the executed command */
            free_cmd (cmd);
        }
    }
}

/**
 * @brief Default socket module handler for a registration of unidentified 
 * accepted connections.
 *
 * Registers the connection and associates a notification structure. 
 * The connection is timestamped and inserted to waiting clients queue.
 */
static void on_login_register (tSocket* sock)
{
    if (sock->get_role() == AS_SOCK_CLIENT)
    {
        debug_log ('s', "REGISTERING SOCKET %i", sock->get_sid ());

        /* Allocate a new notification structure */
        tConnection *new_item = static_cast<tConnection *> 
            (xmalloc (sizeof (tConnection)));

        /* Initialize and associate it with the socket */
        new_item->conn_socket = sock;
        new_item->status = CS_NEW;
        ADVANCE_TIMER (&new_item->login_limit, MAX_PENDING_CONNECTION_TIME);

        /* Append it to waiting clients queue */
        new_clients.append (static_cast<void *> (new_item));
        sock->set_app_specific_data (new_item);
    }
}


/**
 * @brief Default socket module handler for an unregistration of unidentified 
 * accepted connections.
 *
 * Unregisters the connection, cleans up the socket and releases all allocated
 * resources.
 */
static void on_login_unregister (tSocket *sock)
{
    void *dptr;
    ASSERT (sock);

    /* Cleanup the socket */
    if (sock->close_socket())
        log(L_ERROR,"Error closing socket.");

    /* If the socket holds a reference to associated notification structure, 
     * clean it */
    if ((dptr = sock->get_app_specific_data ()))
    {
        tConnection c;

        /* Remove the connection from waiting clients queue */
        c.conn_socket = sock;
        void *item_removed = new_clients.remove_match (&c, conn_socket_comp);
        ASSERT (item_removed);
    }

    /* Release the socket and associated notification structure */
    delete sock;
}


/**
 * @brief Closes all pending connections.
 *
 * Each client (after its connection is correctly accepted) must authenticate
 * itself within certain time limit. The clients are waiting in a queue
 * with an increasing time of arrival. If the limit for the unauthenticated
 * connection is exceeded the connection is closed and all associated
 * structures released.
 */
static void close_pending_connections ()
{
    if (new_clients.count() > 0)
    {
        tConnection *item = static_cast<tConnection *> 
            (new_clients.get_first()->data);

        /* If the oldest pending connection reached the time limit */
        if (timercmp (&gb_current_time, &item->login_limit, >))
        {
            ASSERT (item->status == CS_NEW || item->status == CS_IDENT);
            tSocket *s = item->conn_socket;

            /* Clear it out and release associated resources */
            log (L_WARN, "The connection %li login timed out. Removing it.", 
                    s->get_sid ());
            s->get_smgr()->unregister_socket (s);
        }
    }
}

#ifndef DOXY_SKIP
// Sets up the control channel
void setup_ctrl_channel ()
{
    /* There should be no ctrl socket file */
    if (!s_eaccess (DEFAULT_CTRL_SOCKET, F_OK) && unlink (DEFAULT_CTRL_SOCKET))
        log (L_WARN,"Can't set control socket free.");

    /* Create a new control socket */
    ctrl_channel = new tCtrlUNIXSocketNB;
    if (!ctrl_channel || ctrl_channel->create_server(DEFAULT_CTRL_SOCKET, smgr))
    {
        delete ctrl_channel;
        die ("Can't setup the control channel '%s'.", DEFAULT_CTRL_SOCKET);
    }

    /* Setup all neccessary events */
    ctrl_channel->set_on_hup (on_login_hup);
    ctrl_channel->set_on_err (on_login_err);
    ctrl_channel->set_on_connected (on_login_connect);
    ctrl_channel->set_on_read_available (on_login_read);
    ctrl_channel->set_on_read_finished (on_login_read_notify);
    ctrl_channel->set_on_register (on_login_register);
    ctrl_channel->set_on_unregister (on_login_unregister);
}


// Cleans up the control channel 
void shutdown_ctrl_channel ()
{
    if (smgr.unregister_socket (ctrl_channel) < 0)
        log (L_ERROR, "Can't unregister socket (fd=%ld).",
                ctrl_channel->get_fd());
    delete ctrl_channel;
}


// Main processing function
void process ()
{
    long io_time;
    struct timeval time_limit, dtime, ptime;

    debug_log ('1', "process ()");

    /* Calculate the quantum timer */
    dtime.tv_sec = DEFAULT_ECHELOGD_IO_QUANTUM / 1000;
    dtime.tv_usec = (DEFAULT_ECHELOGD_IO_QUANTUM - dtime.tv_sec * 1000) * 1000;

    /* Get the system time */
    gettimeofday (&gb_current_time, NULL);

    /* Init echelod system states */
    gb_echelog_sane = true;
    gb_reconfiguration_in_progress = false;

    while (true)
    {
        /* Check if the system state is acceptable */
        if (!gb_echelog_sane)
            exec_sanitize_modules ();

        gettimeofday (&gb_current_time, NULL);

        /* Check delayed system commands */
        check_sys_delayed ();

        gettimeofday (&gb_current_time, NULL);

        /* Process system commands from the queue */
        if (exec_sys_commands ())
            break;
        
        /* Recalulate time quantum timer */
        gettimeofday (&gb_current_time, NULL);
        timeradd (&gb_current_time, &dtime, &time_limit);
        io_time = DEFAULT_ECHELOGD_IO_QUANTUM;

        /* Poll sockets and handle non-blocking sockets io */
        do 
        {
            smgr.process (io_time);

            /* Check if there is time left from the quantum */
            gettimeofday (&gb_current_time, NULL);
            timersub (&time_limit, &gb_current_time, &ptime);
            io_time =  ptime.tv_sec * 1000 + ptime.tv_usec / 1000;
        }
        while (io_time > 0);

        /* Clear out pending connections */
        close_pending_connections ();

        gettimeofday (&gb_current_time, NULL);

        /* Check modules and timeouts */
        process_modules (gb_eenv.current_configuration);

        /* Process active modules from older configuration, if needed */
        if (gb_reconfiguration_in_progress)
            process_old_modules (gb_eenv.previous_configuration);
    }

    /* Halting echelog, remove the lock file */
    if(s_access (DEFAULT_LOCK_CF, F_OK) || (unlink (DEFAULT_LOCK_CF) < 0))
        log (L_WARN, "Can't remove persistent stat file '%s'.", 
                DEFAULT_LOCK_CF);

    return;
}

#endif
