/**
 * @file
 * @brief Contains implementation of tFilterCond class and child classes
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "filtercond.h"
#include "messages/msgmgr.h"
#include "messages/typelist.h"
#include <string.h>
#include <regex.h>

using namespace lib;
using namespace messages;

//----------------------------------------------------------
// Declaration & Implementation of the tFilterCondSS class
//----------------------------------------------------------

/**
 * @brief Substring Condition
 *
 * This class is inherited from filters::tFilterCond class and implements
 * test_chunk() method to test substring existence of value in chunk's tested
 * attribute.
 *
 * @note
 * It is possible to use this condition for VARCHAR attributes only.
 */
class tFilterCondSS : public tFilterCond
{
    public:
        /**
         * @brief Condition evaluation method
         *
         * Redefined to behave as a "Substring" condition.
         * All parameters and return value has the same meaning as
         * described in documentation of filters::tFilterCond::test_chunk().
         */
        virtual bool test_chunk (const messages::tMsgChunk& chunk) const;

    protected:
        /**
         * @brief Destructor
         */
        virtual ~tFilterCondSS ();

        /**
         * @brief Constructor
         *
         * All parameters have the same meaning as they are described
         * in documentaion of filters::tFilterCond::tFilterCond().
         */
        tFilterCondSS (int msg_handle, char* attr_name,
            tList& value, tMemPool& conf_mp);

        /**
         * @brief Condition validity test
         *
         * This class has its own meaning of validity, therefore
         * this method is redefined here.
         */
        virtual bool is_invalid ();

        /// Factory method's friendship declaration
        friend tFilterCond* tFilterCond::
            create_condition (const char* type,
            char* attr_name, lib::tList& value, int msg_handle,
            lib::tMemPool& conf_mp);
};

/* Basic initialization */
tFilterCondSS::tFilterCondSS (int msg_handle, char* attr_name,
    tList& value, tMemPool& conf_mp)
    : tFilterCond (msg_handle, attr_name, value, conf_mp)
{
}

/* Final cleanup */
tFilterCondSS::~tFilterCondSS ()
{
}

/* Tests validity of condition */
bool tFilterCondSS::is_invalid ()
{
    if (mp_value == NULL)
        return true;

    /* this operation is supported only for varchar type */
    const char* type_name = gb_message_manager.get_type_of_attr (m_msg_handle,
                                m_attr_name);
    if (m_comps.count () > 0)
        type_name = gb_type_manager.get_component_type_by_path (type_name,
                        m_comps);
    if (strcmp (type_name, "VARCHAR") &&
        strcmp (type_name, "USERNAME") &&
        strcmp (type_name, "HOSTNAME"))
    {
        log (L_ERROR, "Substring condition is now supported "
            "only for 'VARCHAR' compatible types.");
        return true;
    }

    return false;
}

/* Test function for this type of condition */
bool tFilterCondSS::test_chunk (const messages::tMsgChunk& chunk) const
{
    tEchelogType* test_value = chunk.get_attr (m_attr_name);
    if (m_comps.count () > 0)
        test_value = test_value->get_component_by_path (m_comps);

    const char* haystack = static_cast<tVarchar*>
        (test_value)->get_direct_value ();
    const char* needle = static_cast<tVarchar*>
        (mp_value)->get_direct_value ();

    if (strstr (haystack, needle))
        return true;

    return false;
}

//----------------------------------------------------------
// Declaration & Implementation of the tFilterCondRE class
//----------------------------------------------------------
/**
 * @brief Substring Condition
 *
 * This class is inherited from filters::tFilterCond class and implements
 * test_chunk() method to test substring existence of value in chunk's tested
 * attribute.
 *
 * @note
 * It is possible to use this condition for VARCHAR attributes only.
 */
class tFilterCondRE : public tFilterCondSS
{
    public:
        /**
         * @brief Condition evaluation method
         *
         * Redefined to behave as a "GlobaMatching" condition.
         * All parameters and return value has the same meaning as
         * described in documentation of filters::tFilterCond::test_chunk().
         */
        virtual bool test_chunk (const messages::tMsgChunk& chunk) const;

    private:
        /**
         * @brief Destructor
         */
        virtual ~tFilterCondRE ();

        /**
         * @brief Constructor
         *
         * All parameters have the same meaning as they are described
         * in documentaion of filters::tFilterCond::tFilterCond().
         */
        tFilterCondRE (int msg_handle, char* attr_name,
            tList& value, tMemPool& conf_mp);

        /**
         * @ brief Compiled Regex
         *
         * This is the compiled form of the regex
         */
        regex_t regex;

        /// Factory method's friendship declaration
        friend tFilterCond* tFilterCond::
            create_condition (const char* type,
            char* attr_name, lib::tList& value, int msg_handle,
            lib::tMemPool& conf_mp);
};

/* Basic initialization */
tFilterCondRE::tFilterCondRE (int msg_handle, char* attr_name,
    tList& value, tMemPool& conf_mp)
    : tFilterCondSS (msg_handle, attr_name, value, conf_mp)
{
    const char* needle = static_cast<tVarchar*>(mp_value)->get_direct_value ();
    int ret;
  
    /* build regex */
    if ((ret = regcomp(&regex, needle, REG_EXTENDED | REG_NOSUB))) 
    {
        char buf[MAX_LOG_LINE_LENGTH];
        regerror(ret, &regex, buf, MAX_LOG_LINE_LENGTH);
        log (L_ERROR, "bad regexp '%s': %s", needle, buf);
        mp_value = NULL;
        return ;
    }
}

/* Final cleanup */
tFilterCondRE::~tFilterCondRE ()
{
    regfree (&regex);
}

/* Test function for this type of condition */
bool tFilterCondRE::test_chunk (const messages::tMsgChunk& chunk) const
{
    tEchelogType* test_value = chunk.get_attr (m_attr_name);
    if (m_comps.count () > 0)
        test_value = test_value->get_component_by_path (m_comps);

    const char* haystack = static_cast<tVarchar*>
        (test_value)->get_direct_value ();

    /* execute regex with the value */
    if (!regexec(&regex, haystack, 0, NULL, 0))
        return true;
      
    return false;
}


//----------------------------------------------------------
// Declaration & Implementation of the tFilterCondEQ class
//----------------------------------------------------------

/**
 * @brief Equal Condition
 *
 * This class is inherited from filters::tFilterCond class and implements
 * test_chunk() method to test equality of value and chunk's tested
 * attribute.
 */
class tFilterCondEQ : public tFilterCond
{
    public:
        /**
         * @brief Condition evaluation method
         *
         * Redefined to behave as a "Equal" condition.
         * All parameters and return value has the same meaning as
         * described in documentation of filters::tFilterCond::test_chunk().
         */
        virtual bool test_chunk (const messages::tMsgChunk& chunk) const;

    private:
        /**
         * @brief Destructor
         */
        virtual ~tFilterCondEQ ();

        /**
         * @brief Constructor
         *
         * All parameters have the same meaning as they are described
         * in documentaion of filters::tFilterCond::tFilterCond().
         */
        tFilterCondEQ (int msg_handle, char* attr_name,
            tList& value, tMemPool& conf_mp);

        /// Factory method's friendship declaration
        friend tFilterCond* tFilterCond::
            create_condition (const char* type,
            char* attr_name, lib::tList& value, int msg_handle,
            lib::tMemPool& conf_mp);
};

/* Basic initialization */
tFilterCondEQ::tFilterCondEQ (int msg_handle, char* attr_name,
    tList& value, tMemPool& conf_mp)
    : tFilterCond (msg_handle, attr_name, value, conf_mp)
{
}

/* Final cleanup */
tFilterCondEQ::~tFilterCondEQ ()
{
}

/* Test function for this type of condition */
bool tFilterCondEQ::test_chunk (const messages::tMsgChunk& chunk) const
{
    tEchelogType* test_value = chunk.get_attr (m_attr_name);
    if (m_comps.count () > 0)
        test_value = test_value->get_component_by_path (m_comps);
    tTYPE_RELATION rel = mp_value->order (*test_value);
    if (rel == TP_IDENTICAL || rel == TP_EQUAL)
        return true;

    return false;
}

//----------------------------------------------------------
// Declaration & Implementation of the tFilterCondLE class
//----------------------------------------------------------

/**
 * @brief Less or Equal Condition
 *
 * This class is inherited from filters::tFilterCond class and implements
 * test_chunk() method to test if the value greater or equal than chunk's
 * tested attribute.
 */
class tFilterCondLE : public tFilterCond
{
    public:
        /**
         * @brief Condition evaluation method
         *
         * Redefined to behave as a "Less or Equal" condition.
         * All parameters and return value has the same meaning as
         * described in documentation of filters::tFilterCond::test_chunk().
         */
        virtual bool test_chunk (const messages::tMsgChunk& chunk) const;

    private:
        /**
         * @brief Destructor
         */
        virtual ~tFilterCondLE ();

        /**
         * @brief Constructor
         *
         * All parameters have the same meaning as they are described
         * in documentaion of filters::tFilterCond::tFilterCond().
         */
        tFilterCondLE (int msg_handle, char* attr_name,
            tList& value, tMemPool& conf_mp);

        /// Factory method's friendship declaration
        friend tFilterCond* tFilterCond::
            create_condition (const char* type,
            char* attr_name, lib::tList& value, int msg_handle,
            lib::tMemPool& conf_mp);
};

/* Basic initialization */
tFilterCondLE::tFilterCondLE (int msg_handle, char* attr_name,
    tList& value, tMemPool& conf_mp)
    : tFilterCond (msg_handle, attr_name, value, conf_mp)
{
}

/* Final cleanup */
tFilterCondLE::~tFilterCondLE ()
{
}

/* Test function for this type of condition */
bool tFilterCondLE::test_chunk (const messages::tMsgChunk& chunk) const
{
    tEchelogType* test_value = chunk.get_attr (m_attr_name);
    if (m_comps.count () > 0)
        test_value = test_value->get_component_by_path (m_comps);
    tTYPE_RELATION rel = mp_value->order (*test_value);
    if (rel == TP_GREATER || rel == TP_EQUAL || rel == TP_IDENTICAL)
        return true;

    return false;
}

//----------------------------------------------------------
// Declaration & Implementation of the tFilterCondGE class
//----------------------------------------------------------

/**
 * @brief Greater or Equal Condition
 *
 * This class is inherited from filters::tFilterCond class and implements
 * test_chunk() method to test if the value less or equal than chunk's
 * tested attribute.
 */
class tFilterCondGE : public tFilterCond
{
    public:
        /**
         * @brief Condition evaluation method
         *
         * Redefined to behave as a "Greater or Equal" condition.
         * All parameters and return value has the same meaning as
         * described in documentation of filters::tFilterCond::test_chunk().
         */
        virtual bool test_chunk (const messages::tMsgChunk& chunk) const;

    private:
        /**
         * @brief Destructor
         */
        virtual ~tFilterCondGE ();

        /**
         * @brief Constructor
         *
         * All parameters have the same meaning as they are described
         * in documentaion of filters::tFilterCond::tFilterCond().
         */
        tFilterCondGE (int msg_handle, char* attr_name,
            tList& value, tMemPool& conf_mp);

        /// Factory method's friendship declaration
        friend tFilterCond* tFilterCond::
            create_condition (const char* type,
            char* attr_name, lib::tList& value, int msg_handle,
            lib::tMemPool& conf_mp);
};

/* Basic initialization */
tFilterCondGE::tFilterCondGE (int msg_handle, char* attr_name,
    tList& value, tMemPool& conf_mp)
    : tFilterCond (msg_handle, attr_name, value, conf_mp)
{
}

/* Final cleanup */
tFilterCondGE::~tFilterCondGE ()
{
}

/* Test function for this type of condition */
bool tFilterCondGE::test_chunk (const messages::tMsgChunk& chunk) const
{
    tEchelogType* test_value = chunk.get_attr (m_attr_name);
    if (m_comps.count () > 0)
        test_value = test_value->get_component_by_path (m_comps);
    tTYPE_RELATION rel = mp_value->order (*test_value);
    if (rel == TP_LESS || rel == TP_EQUAL || rel == TP_IDENTICAL)
        return true;

    return false;
}

//----------------------------------------------------------
// Declaration & Implementation of the tFilterCondLT class
//----------------------------------------------------------

/**
 * @brief Less Than Condition
 *
 * This class is inherited from filters::tFilterCond class and implements
 * test_chunk() method to test if the value greater than chunk's
 * tested attribute.
 */
class tFilterCondLT : public tFilterCond
{
    public:
        /**
         * @brief Condition evaluation method
         *
         * Redefined to behave as a "Less Than" condition.
         * All parameters and return value has the same meaning as
         * described in documentation of filters::tFilterCond::test_chunk().
         */
        virtual bool test_chunk (const messages::tMsgChunk& chunk) const;

    private:
        /**
         * @brief Destructor
         */
        virtual ~tFilterCondLT ();

        /**
         * @brief Constructor
         *
         * All parameters have the same meaning as they are described
         * in documentaion of filters::tFilterCond::tFilterCond().
         */
        tFilterCondLT (int msg_handle, char* attr_name,
            tList& value, tMemPool& conf_mp);

        /// Factory method's friendship declaration
        friend tFilterCond* tFilterCond::
            create_condition (const char* type,
            char* attr_name, lib::tList& value, int msg_handle,
            lib::tMemPool& conf_mp);
};

/* Basic initialization */
tFilterCondLT::tFilterCondLT (int msg_handle, char* attr_name,
    tList& value, tMemPool& conf_mp)
    : tFilterCond (msg_handle, attr_name, value, conf_mp)
{
}

/* Final cleanup */
tFilterCondLT::~tFilterCondLT ()
{
}

/* Test function for this type of condition */
bool tFilterCondLT::test_chunk (const messages::tMsgChunk& chunk) const
{
    tEchelogType* test_value = chunk.get_attr (m_attr_name);
    if (m_comps.count () > 0)
        test_value = test_value->get_component_by_path (m_comps);
    tTYPE_RELATION rel = mp_value->order (*test_value);
    if (rel == TP_GREATER)
        return true;

    return false;
}

//----------------------------------------------------------
// Declaration & Implementation of the tFilterCondGT class
//----------------------------------------------------------

/**
 * @brief Greater Than Condition
 *
 * This class is inherited from filters::tFilterCond class and implements
 * test_chunk() method to test if the value less than chunk's
 * tested attribute.
 */
class tFilterCondGT : public tFilterCond
{
    public:
        /**
         * @brief Condition evaluation method
         *
         * Redefined to behave as a "Greater Than" condition.
         * All parameters and return value has the same meaning as
         * described in documentation of filters::tFilterCond::test_chunk().
         */
        virtual bool test_chunk (const messages::tMsgChunk& chunk) const;

    private:
        /**
         * @brief Destructor
         */
        virtual ~tFilterCondGT ();

        /**
         * @brief Constructor
         *
         * All parameters have the same meaning as they are described
         * in documentaion of filters::tFilterCond::tFilterCond().
         */
        tFilterCondGT (int msg_handle, char* attr_name,
            tList& value, tMemPool& conf_mp);

        /// Factory method's friendship declaration
        friend tFilterCond* tFilterCond::
            create_condition (const char* type,
            char* attr_name, lib::tList& value, int msg_handle,
            lib::tMemPool& conf_mp);
};

/* Basic initialization */
tFilterCondGT::tFilterCondGT (int msg_handle, char* attr_name,
    tList& value, tMemPool& conf_mp)
    : tFilterCond (msg_handle, attr_name, value, conf_mp)
{
}

/* Final cleanup */
tFilterCondGT::~tFilterCondGT ()
{
}

/* Test function for this type of condition */
bool tFilterCondGT::test_chunk (const messages::tMsgChunk& chunk) const
{
    tEchelogType* test_value = chunk.get_attr (m_attr_name);
    if (m_comps.count () > 0)
        test_value = test_value->get_component_by_path (m_comps);
    tTYPE_RELATION rel = mp_value->order (*test_value);
    if (rel == TP_LESS)
        return true;

    return false;
}

//----------------------------------------------------------
// Implementation of the tFilterCond class
//----------------------------------------------------------

/* Factory method for condition creation */
tFilterCond* tFilterCond::create_condition (const char* type,
    char* attr_name, tList& value, int msg_handle,
    tMemPool& conf_mp)
{
    ASSERT (type != NULL);

    tFilterCond* ret = NULL;
    bool recognized = false;

    /* check if the type is exacly 1-2 characters */
    int len = strlen (type);
    if (len != 1 && len != 2)
    {
        log (L_ERROR, "Error in condition type '%s':"
            "condition type must be 1-2 characters long", type);
        return ret;
    }

    if (!strcmp (type, "ss"))
    {
        ret = new tFilterCondSS (msg_handle, attr_name, value, conf_mp);
        recognized = true;
        goto test;
    }

    if (!strcmp (type, "~"))
    {
        ret = new tFilterCondRE (msg_handle, attr_name, value, conf_mp);
        recognized = true;
        goto test;
    }

    if (!strcmp (type, "=="))
    {
        ret = new tFilterCondEQ (msg_handle, attr_name, value, conf_mp);
        recognized = true;
        goto test;
    }

    if (!strcmp (type, ">="))
    {
        ret = new tFilterCondGE (msg_handle, attr_name, value, conf_mp);
        recognized = true;
        goto test;
    }

    if (!strcmp (type, "<="))
    {
        ret = new tFilterCondLE (msg_handle, attr_name, value, conf_mp);
        recognized = true;
        goto test;
    }

    if (!strcmp (type, ">"))
    {
        ret = new tFilterCondGT (msg_handle, attr_name, value, conf_mp);
        recognized = true;
        goto test;
    }

    if (!strcmp (type, "<"))
    {
        ret = new tFilterCondLT (msg_handle, attr_name, value, conf_mp);
        recognized = true;
        goto test;
    }

test:
    if (ret == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        return NULL;
    }

    if (!recognized)
    {
        log (L_ERROR, "Unrecognized condition type '%s'", type);
        delete ret;
        return NULL;
    }

    if (ret->is_invalid ())
    {
        log (L_ERROR, "Invalid condition specifified.");
        delete ret;
        return NULL;
    }

    return ret;
}

/* Basic initialization */
tFilterCond::tFilterCond (int msg_handle, char* attr_name,
    tList& value, tMemPool& conf_mp) : m_comps (conf_mp)
{
    ASSERT (attr_name != NULL);
    ASSERT (value.count () >= 1);
    ASSERT (value.get_first ()->data != NULL);

    /* initialize attributes */
    m_msg_handle = msg_handle;

    /* parse attr_name to attribute name and components */
    m_attr_name = attr_name;
    char* tmp = m_attr_name;
    while ((tmp = strchr (tmp, '.')) != NULL)
    {
        *tmp = '\0';
        m_comps.append (++tmp);
    }

    /* try to convert string value to value of proper type */
    const char* type_name =
        gb_message_manager.get_type_of_attr (m_msg_handle, m_attr_name);
    if (type_name == NULL)
    {
        log (L_ERROR, "Attribute '%s' is not present in message '%s'",
            m_attr_name,
            gb_message_manager.get_message_name_by_handle (m_msg_handle));
        mp_value = NULL;
        return;
    }
    if (m_comps.count () > 0)
    {
        type_name = gb_type_manager.get_component_type_by_path (type_name,
                        m_comps);
        if (type_name == NULL)
        {
            log (L_ERROR, "Invalid component specified "
                "for attribute '%s'", m_attr_name);
            mp_value = NULL;
            return;
        }
    }
    mp_value = gb_type_manager.get_type_instance (conf_mp, type_name);
    int count = mp_value->sscan (value);
    if (!count)
    {
        log (L_ERROR, "Error in specification of condition value, "
            "cannot interpret '%s' and following as type '%s'",
            static_cast<char*> (value.get_first ()->data), type_name);
        mp_value = NULL;
        return;
    }

    for (int i = 0; i < count; i++)
        value.remove_first ();
}

/* Final cleanup */
tFilterCond::~tFilterCond ()
{
}

/* Tests validity of condition */
bool tFilterCond::is_invalid ()
{
    if (mp_value == NULL)
        return true;
    else
        return false;
}
