/**
 * @file
 * @brief Contains declaration of tFilterCond class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 */

#ifndef ECHELOG_FILTERCOND_H
#define ECHELOG_FILTERCOND_H

#include "messages/msgchunk.h"

/**
 * @brief Simple Condition Filter's condition class
 *
 * This class covers all common functionality of conditions for Simple
 * Condition Filter. There are several types of conditions that could be
 * created. Conditions are created by create_condition() factory method
 * and then tests for specific chunks could be performed by test_chunk().
 *
 * Each condition of Simple Condition Filter compares or checks submitted
 * value against specified attribute of the tested message chunk. Type
 * of the value must be the same as type of the attribute.
 *
 * This class is an interface for conditions, which are derived
 * from this class. Specific conditions should implement test_chunk()
 * method.
 */
class tFilterCond
{
    public:
        //-------------------------------
        // Factory method
        //-------------------------------

        /**
         * @brief Factory method
         *
         * Creates new instance of condition of given type.
         * Other parameters than described here are simply passed to
         * condition's constructor and have the the same meaning as
         * parameters there with the same name, see tFilterCond().
         *
         * @param type
         * Type of the instance to be created. Possible values
         * are: "ss", ">=", "<=", ">", ">".
         *
         * @return
         * Non-NULL on success, NULL in case of failure.
         */
        static tFilterCond* create_condition (const char* type,
            char* attr_name, lib::tList& value, int msg_handle,
            lib::tMemPool& conf_mp);

        //-------------------------------
        // Interface
        //-------------------------------

        /**
         * @brief Evaluation method
         *
         * This function evaluates the condition for given chunk.
         * Value stored inside the condition is checked against
         * value in chunk's attribute. Name of the attribute to be
         * checked and type of check was specified in create_condition().
         *
         * @param
         * Chunk to be tested.
         *
         * @return
         * If the condition is true for given chunk returns true.
         * Otherwise returns false.
         */
        virtual bool test_chunk (const messages::tMsgChunk& chunk) const = 0;

        /**
         * @brief Destructor
         */
        virtual ~tFilterCond ();

    protected:
        //-------------------------------
        // Internal methods
        //-------------------------------

        /**
         * @brief Constructor
         *
         * Performs basic initialization and scans and creates
         * value from the string on the head of the value list.
         *
         * @param msg_handle
         * Handle of the message tested chunks will belong to.
         *
         * @param attr_name
         * Name of the attribute to be tested in chunk.
         *
         * @param value
         * List which contain the value in string form on the head.
         *
         * @param conf_mp
         * Memory pool where value will be allocated. It should be
         * the memory pool where your configuration items are stored.
         * Pool is not freed here, freeing of pool is responsibility
         * of condition owner.
         */
        tFilterCond (int msg_handle, char* attr_name,
            lib::tList& value, lib::tMemPool& conf_mp);

        /**
         * @brief Checks the validity of the condition
         *
         * Checks if the creation process of the condition value
         * was successful ie. mp_value isn't NULL, so value was
         * successfuly scanned from string form from configuration,
         * other initialization in done constructor was succesful.
         *
         * @return
         * If the condition is invalid returns true.
         * Otherwise returns false.
         */
        virtual bool is_invalid ();

        //-------------------------------
        // Owned components
        //-------------------------------

        /// Value which is compared to attribute
        messages::tEchelogType* mp_value;

        /// Handle of the message tested chunks belongs to
        int m_msg_handle;

        /// List of parsed component names of attribute
        lib::tList m_comps;

        //-------------------------------
        // Used only components
        //-------------------------------

        /// Name of the attribute to check against value
        char* m_attr_name;
};

#endif //ifdef ECHELOG_FILTERCOND_H

