/**
 * @file
 * @brief Contains implementation of tFilterHeaderCond class
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "filterheadercond.h"
#include "messages/msgmgr.h"
#include "messages/typelist.h"
#include <string.h>
#include <regex.h>

using namespace lib;
using namespace messages;


//----------------------------------------------------------
// Declaration & Implementation of the tFilterHeaderCond class
//----------------------------------------------------------

tFilterHeaderCond::tFilterHeaderCond ()
    : msg_handle (-1), m_conds (NULL)
{
}


tFilterHeaderCond::~tFilterHeaderCond ()
{
    cleanup ();
}


tHeaderOp tFilterHeaderCond::parse_operator (tList &list)
{
    
    /* Check if there is operator and operand */
    if (list.count() < 2)
    {
        log (L_ERROR, "Error in rule configuration, "
                "incomplete header condition configuration.");
        return OP_UNKNOWN;
    }

    char *op = static_cast<char*> (list.remove_first ());
    ASSERT (op != NULL);

    if (!strcmp (op, "ss"))
        return OP_SUBSTRING;
    else if (!strcmp (op, "~"))
        return OP_REGEXP;
    else if (!strcmp (op, "=="))
        return OP_EQ;
    else if (!strcmp (op, "!="))
        return OP_NEQ;
    else if (!strcmp (op, ">="))
        return OP_GEQ;
    else if (!strcmp (op, "<="))
        return OP_LEQ;
    else if (!strcmp (op, ">"))
        return OP_GE;
    else if (!strcmp (op, "<"))
        return OP_LE;

    return OP_UNKNOWN;
}


bool tFilterHeaderCond::parse_msg_handle (tList &list)
{
    if (list.count() < 1)
    {
        log (L_ERROR, "Error in rule configuration, "
                "incomplete header condition configuration.");
        return true;
    }

    char *tmp_str = static_cast<char*> (list.remove_first ());
    ASSERT (tmp_str != NULL);

    /* get msg handle of message with name tmp_str */
    msg_handle = gb_message_manager.get_handle_by_name(tmp_str);
    if (msg_handle == -1)
    {
        log (L_ERROR, "Error in rule configuration, "
            "unrecognized name of the message '%s'.", tmp_str);
        return true;
    }
    
    m_msg_type = gb_message_manager.get_message_type_by_handle (msg_handle);

    return false;
}


void *tFilterHeaderCond::create_uni_operand (tHeaderOp op, char *str, 
        tMemPool &conf_mp)
{
    if (op == OP_REGEXP)
    {
        regex_t *preg = static_cast<regex_t *> 
            (conf_mp.alloc (sizeof (regex_t)));

        int ret;
        if ((ret = regcomp(preg, str, REG_EXTENDED | REG_NOSUB))) 
        {
            char buf[MAX_LOG_LINE_LENGTH];
            regerror(ret, preg, buf, MAX_LOG_LINE_LENGTH);
            log (L_ERROR, "bad regexp '%s': %s", str, buf);
            return NULL;
        }

        return preg;
    }

    return s_strdup (str, conf_mp);
}


bool tFilterHeaderCond::init (tList& list, tMemPool& conf_mp)
{
    tMemPool init_mp(512);
    tList msg (init_mp);

    m_conds = new tList (conf_mp);
    if (!m_conds)
    {
        log (L_ERROR, "Not enough memory.");
        return true;
    }

    bool always_satisfied = false;
    char* rule_str;
    while ((rule_str = static_cast<char*> (list.remove_first ())) != NULL)
    {
        msg.reset ();
        if (s_strdsep (msg, rule_str, ".", conf_mp))
            return true;

        char *msg_str = static_cast<char*> (msg.remove_first ());

        if (!msg.count())
        {
            /* The string can plain MSG or MSGALL */
            if (!strcmp (msg_str, "MSGALL"))
            {
                /* this rule should apply to all msgs, 
                 * no other conditions apply */
                always_satisfied = true;
                return false;
            }
            else if (!strcmp (msg_str, "MSG"))
            {
                /* get message information by name */
                if (parse_msg_handle (list))
                    return true;

                m_conds->append 
                    (tHeaderCond::create_new (OP_EQ, REC_TYPE, &m_msg_type));
            }
            else
                break;
        }
        else
        {
            if (strcmp (msg_str, "MSG"))
            {
                /* Not MSG specified rule, quit */
                break;
            }

            char *val_str;
            char *type_str = static_cast<char*> (msg.remove_first ());
            if (!strcmp (type_str, "type"))
            {
                tHeaderOp op = parse_operator (list);
                if (op != OP_EQ)
                    goto invalid_operator;

                /* get message information by name */
                if (parse_msg_handle (list))
                    return true;

                m_conds->append 
                    (tHeaderCond::create_new (OP_EQ, REC_TYPE, &m_msg_type));
            }
            else if (!strcmp (type_str, "msg_time"))
            {
                tHeaderOp op = parse_operator (list);
                if (op != OP_EQ && op != OP_NEQ && 
                        op != OP_LEQ && op != OP_GEQ &&
                        op != OP_LE && op != OP_GE)
                    goto invalid_operator;

                val_str = static_cast<char*> (list.remove_first ());
                ASSERT (val_str);
                const char *time_format = "%Y-%m-%d %H:%M:%S";
                /* try getting time value */
                struct tm tm_val;
                int32_t *value = static_cast<int32_t *> 
                    (conf_mp.alloc (sizeof (int32_t)));
                *value = 0;
                if (strptime (val_str, time_format, &tm_val))
                {
                    *value = mktime (&tm_val);
                    if ( *value == -1 ) /* shouldn't happen */
                        goto invalid_value;
                }

                m_conds->append 
                    (tHeaderCond::create_new (op, REC_TIME, value));
            }
            else if (!strcmp (type_str, "subrecords"))
            {
                tHeaderOp op = parse_operator (list);
                if (op != OP_EQ && op != OP_NEQ && 
                        op != OP_LEQ && op != OP_GEQ &&
                        op != OP_LE && op != OP_GE)
                    goto invalid_operator;
                
                val_str = static_cast<char*> (list.remove_first ());
                ASSERT (val_str);

                int32_t *value = static_cast<int32_t *> 
                    (conf_mp.alloc (sizeof (int32_t)));
                long v;
                if (s_stol (val_str, &v, INT32_MIN, INT32_MAX))
                    goto invalid_value;
                *value = v;

                m_conds->append 
                    (tHeaderCond::create_new (OP_EQ, REC_TYPE, value));
            }
            else if (!strcmp (type_str, "agent_id"))
            {
                tHeaderOp op = parse_operator (list);
                if (op != OP_EQ && op != OP_NEQ && 
                        op != OP_LEQ && op != OP_GEQ &&
                        op != OP_LE && op != OP_GE &&
                        op != OP_SUBSTRING && op != OP_REGEXP)
                    goto invalid_operator;
                
                val_str = static_cast<char*> (list.remove_first ());
                ASSERT (val_str);
                void *value = create_uni_operand (op, val_str, conf_mp);
                if (!value)
                    goto invalid_value;

                m_conds->append 
                    (tHeaderCond::create_new (op, REC_AGENT_ID, value));
            }
            else if (!strcmp (type_str, "agent_instance_id"))
            {
                tHeaderOp op = parse_operator (list);
                if (op != OP_EQ && op != OP_NEQ && 
                        op != OP_LEQ && op != OP_GEQ &&
                        op != OP_LE && op != OP_GE &&
                        op != OP_SUBSTRING && op != OP_REGEXP)
                    goto invalid_operator;
                
                val_str = static_cast<char*> (list.remove_first ());
                ASSERT (val_str);
                void *value = create_uni_operand (op, val_str, conf_mp);
                if (!value)
                    goto invalid_value;

                m_conds->append 
                   (tHeaderCond::create_new (op, REC_AGENT_INSTANCE_ID, value));
            }
            else if (!strcmp (type_str, "host_id"))
            {
                tHeaderOp op = parse_operator (list);
                if (op != OP_EQ && op != OP_NEQ &&
                        op != OP_LEQ && op != OP_GEQ &&
                        op != OP_LE && op != OP_GE &&
                        op != OP_REGEXP)
                    goto invalid_operator;
                
                val_str = static_cast<char*> (list.remove_first ());
                ASSERT (val_str);
                void *value = create_uni_operand (op, val_str, conf_mp);
                if (!value)
                    goto invalid_value;

                m_conds->append 
                   (tHeaderCond::create_new (op, REC_HOST_ID, value));
            }
            else if (!strcmp (type_str, "flags"))
            {
                tHeaderOp op = parse_operator (list);
                if (op != OP_EQ && op != OP_NEQ)
                    goto invalid_operator;
                
                val_str = static_cast<char*> (list.remove_first ());
                ASSERT (val_str);

                int32_t *value = static_cast<int32_t *> 
                    (conf_mp.alloc (sizeof (int32_t)));
                if (!strcasecmp (val_str, "BASIC"))
                    *value = 0;
                else if (!strcasecmp (val_str, "INCREMENTAL"))
                    *value |= HDR_FLAG_INCREMENTAL;
                else if (!strcasecmp (val_str, "AGGREGABLE"))
                    *value |= HDR_FLAG_AGGREGABLE;
                else
                    goto invalid_value;

                m_conds->append 
                   (tHeaderCond::create_new (op, REC_FLAGS, value));
            }
            else
            {
              invalid_operator:
                 log (L_ERROR, "Error in rule configuration, "
                     "unknown condition 'MSG.%s'.", type_str);
                 return true;
              invalid_value:
                 log (L_ERROR, "Error in rule configuration, "
                     "unknown value in condition 'MSG.%s'.", val_str);
                 return true;
            }
        }
    }

    /* Put back the look-ahead token */
    if (rule_str)
        list.prepend (rule_str);

    /* Check for mandatory condition MSG.type */
    if (!always_satisfied && msg_handle == -1)
    {
        log (L_ERROR, "Error in rule configuration, "
            "mandatory condition MSG.type or MSGALL unspecified.");
        return true;
    }

    return false;
}


void tFilterHeaderCond::cleanup ()
{
    if (m_conds)
    {
        for (int i = m_conds->count (); i > 0; i--)
            delete static_cast<tHeaderCond*> (m_conds->remove_first ());
    }
    m_conds = NULL;
}


bool tFilterHeaderCond::test_header (const messages::tMsgHeader& header) const
{
    tNode const *n = NULL;
    const tHeaderCond *cond;

    FOR_ALL_LIST_ITEMS (n, m_conds)
    {
        cond = static_cast<tHeaderCond *> (n->data);
        ASSERT (cond);

        if (!cond->test_header (header))
            return false;
    }
    
    return true;
}
