/**
 * @file
 * @brief Contains declaration of tFilterHeaderCond class
 *
 * @author Kamil Toman <ktoman@email.cz>
 */

#ifndef ECHELOG_FILTERHEADERCOND_H
#define ECHELOG_FILTERHEADERCOND_H

#include "messages/msgchunk.h"
#include "headercond.h"

/**
 * @brief Simple Condition Filter's header condition class
 *
 * This class is an analogy of <tt>tFilterCond</tt> class. You can
 * create a set of conditions and then you can test it against 
 * a message header.
 */
class tFilterHeaderCond
{
        //-------------------------------
        // Properties
        //-------------------------------

        /// Message handle of examined message
        SIMPLE_R_PROPERTY (int32_t, msg_handle);

    public:

        //-------------------------------
        // Interface
        //-------------------------------
        
        tFilterHeaderCond ();

        /**
         * Parse and initialize all header conditions
         *
         * All conditions are parsed and set up from input list
         * of string values.
         *
         * On success the function returns false, on failure, true
         * is returned.
         */
        bool init (lib::tList& list, lib::tMemPool& conf_mp);

        /// Reverse method to init()
        void cleanup ();
        
        /**
         * @brief Evaluation method
         *
         * This function evaluates the condition for given header.
         * Values stored inside conditions are checked against header
         * values.
         *
         * @param
         * Header to be tested.
         *
         * @return
         * If the condition is true for given header returns true.
         * Otherwise returns false.
         */
        bool test_header (const messages::tMsgHeader& header) const;

        /**
         * @brief Test if the header filter is always satisfied 
         *
         * The function returns true if the filter is always satisfied,
         * false otherwise.
         */
        bool is_always_satisfied () const
            { return msg_handle == -1; }
        
        /**
         * @brief Destructor
         */
        ~tFilterHeaderCond ();

    protected:

        //-------------------------------
        // Internal methods
        //-------------------------------

        /// Get the message handle by message name
        bool parse_msg_handle (lib::tList &list);

        /// Parse Operator
        tHeaderOp parse_operator (lib::tList &list);

        /// Returns true if the condition is satisfied
        bool tFilterHeaderCond::test_cond (const messages::tMsgHeader& header,
            const tHeaderCond &cond) const;

        /// Parses string comparator operand (string or regexp)
        void *create_uni_operand (tHeaderOp op, char *str, 
                lib::tMemPool &conf_mp);
        
        //-------------------------------
        // Owned components
        //-------------------------------

        /// Message type
        int32_t m_msg_type;

        /// Message type for comparison
        lib::tList *m_conds;

};

#endif //ifdef ECHELOG_FILTERHEADERCOND_H

