/**
 * @file
 * @brief Contains implementation of tFilterRule class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "filterrule.h"
#include "messages/msgmgr.h"
#include "filtercond.h"
#include "filterheadercond.h"


using namespace lib;
using namespace messages;

/* Basic initialization */
tFilterRule::tFilterRule ()
{
    m_initialized = false;
    m_outs_count = 0;
    m_outs = NULL;
}

/* Final cleanup */
tFilterRule::~tFilterRule ()
{

}

/* Initialization & validation of rule */
bool tFilterRule::init (tList& list, tMemPool& conf_mp)
{
    char *tmp_str;

    /* cleanup previous initialization */
    if (m_initialized)
        cleanup ();

    /* at least 3 strings must be in rule config */
    if (list.count () < 3)
    {
        log (L_ERROR, "Error in rule configuration, "
            "incomplete rule configuration.");
        return true;
    }

    /* parse header conditions */
    if (init_header_conds (list, conf_mp))
        return true;

    if (!m_header_cond.is_always_satisfied())
    {
        /* parse chunk conditions */
        if (init_conds (list, conf_mp))
            return true;
    }

    /* at least one string must be after conditions */
    if (list.count () < 1)
    {
        log (L_ERROR, "Error in rule configuration, "
            "incomplete rule configuration.");
        goto conds_clean;
    }

    tmp_str = static_cast<char*> (list.remove_first ());
    ASSERT (tmp_str != NULL);

    /* what's your command my lord */
    if (!strcmp (tmp_str, "DISCARD"))
    {
        if (list.count () > 0)
            log (L_WARN, "Rest of the rule configuration after DISCARD "
                "is ignored.");
        m_action = RA_DISCARD;
        m_initialized = true;

        return false;
    }
    if (!strcmp (tmp_str, "PASS"))
        m_action = RA_PASS;
    else if (!strcmp (tmp_str, "PASS_DISCARD"))
        m_action = RA_PASS_DISCARD;
    else
    {
        log (L_ERROR, "Error in rule configuration, "
            "unrecognized rule action %s", tmp_str);
        goto conds_clean;
    }

    /* init outputs */
    if (init_outs (list))
        goto conds_clean;

    m_initialized = true;

    return false;

conds_clean:
    cleanup_conds ();
    m_initialized = false;

    return true;
}

/* Initialization of the outputs */
bool tFilterRule::init_outs (const tList& list)
{
    tNode const* n = NULL;
    int i = 0;

    mframes::tOFrame* of = mframes::tOFrame::get_actual_oframe ();
    m_outs_count = list.count();

    /* at least one output must be specified */
    if (m_outs_count < 1)
    {
        log (L_ERROR, "Error in rule configuration, missing output names.");
        goto fail;
    }

    /* create new outs array */
    m_outs = new mframes::tOFrame::tOutput* [m_outs_count];
    if (m_outs == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        goto fail;
    }
    FOR_ALL_LIST_ITEMS (n, &list)
    {
        ASSERT (n->data != NULL);

        /* get out handle by name */
        m_outs[i] =
            of->get_output_by_name (static_cast<char*> (n->data));
        if (m_outs[i] == NULL)
        {
            log (L_ERROR, "Invalid output name \"%s\" specified "
                "in rule configuration.",
                static_cast<char*> (n->data));
            goto outs_clean;
        }
        i++;
    }

    return false;

outs_clean:
    delete [] m_outs;
    m_outs = NULL;
fail:
    m_outs_count = 0;

    return true;
}

/* Initialization of the conditions */
bool tFilterRule::init_conds (tList& list, tMemPool& conf_mp)
{
    /* parse the list for conditions */
    void* tmp_cond = NULL;
    const char* tmp_cond_type = NULL;
    char* tmp_cond_attr = NULL;
    tNode const* n;
    while ((n = list.get_first ()) != NULL)
    {
        /* check what's on the head */
        const char* tmp_str = static_cast<char*> (n->data);
        if (strcmp (tmp_str, "COND"))
            break;

        /*OK COND is on the head, so next three are mandatory */
        list.remove_first ();
        if (list.count () < 3)
        {
            log (L_ERROR, "Error in rule configuration, "
                "incomplete condition configuration.");
            goto conds_clean;
        }
        tmp_cond_attr = static_cast<char*> (list.remove_first ());
        tmp_cond_type = static_cast<char*> (list.remove_first ());
        debug_log ('1', "Trying to create condition of type (%s) "
            "attribute (%s) and value starting with (%s)", tmp_cond_type,
            tmp_cond_attr,
            static_cast<char*> (list.get_first ()->data));
        tmp_cond = tFilterCond::create_condition (tmp_cond_type,
            tmp_cond_attr, list, m_msg_handle, conf_mp);
        if (tmp_cond == NULL)
        {
            log (L_ERROR, "Invalid condition in rule configuration.");
            goto conds_clean;
        }
        m_conds.append (tmp_cond);
    }

    return false;

conds_clean:
    for (int i = m_conds.count (); i > 0; i--)
        delete static_cast<tFilterCond*> (m_conds.remove_first ());

    return true;
}

/* Initialization of the conditions */
bool tFilterRule::init_header_conds (tList& list, tMemPool& conf_mp)
{
    if (m_header_cond.init (list, conf_mp))
        return true;

    if (!m_header_cond.is_always_satisfied())
    {
        m_msg_handle = m_header_cond.get_msg_handle ();
        m_msg_type = gb_message_manager.get_message_type_by_handle 
            (m_msg_handle);
    }

    return false;
}

/* Cleanup of rule */
void tFilterRule::cleanup ()
{
    if (m_initialized)
    {
        if (m_action == RA_PASS)
            cleanup_outs ();
        cleanup_conds ();
        m_initialized = false;
    }
}

/* Cleanup outputs, reverse to init_outs */
void tFilterRule::cleanup_outs ()
{
    delete [] m_outs;
    m_outs = NULL;
    m_outs_count = 0;
}

/* Cleanup conditions, reverse to init_conds */
void tFilterRule::cleanup_conds ()
{
    for (int i = m_conds.count (); i > 0; i--)
        delete static_cast<tFilterCond*> (m_conds.remove_first ());
}

/* Cleanup conditions, reverse to init_conds */
void tFilterRule::cleanup_header_conds ()
{
    m_header_cond.cleanup ();
}

/* Passes matching message chunks to outputs */
bool tFilterRule::pass_msg (tMsgBuf& in_msg,
    tMsgBuf& out_msg, bool& new_created, tMsgBuf& tmp_msg)
{
    ASSERT (m_initialized);

    static tMemPool chunk_mp (MAX_MESSAGE_SIZE);

    /* if this rule applies to all messages handle it specially */
    if (m_header_cond.is_always_satisfied())
    {
        if (m_action != RA_DISCARD)
        {
            /* RA_PASS or RA_PASS_DISCARD */
            debug_log ('1', "Passing all chunks to outputs.");

            /* just pass the message to the outputs */
            bool ret = pass_msg (in_msg);

            /* return the same message */
            new_created = false;
            return ret;
        }

        /* Just discard the whole message */
        return false;
    }

    /* Get input message id */
    tMsgHeader* in_header = in_msg.get_header ();

    /* Check if this rule applies to this message type */
    if (!m_header_cond.test_header (*in_header))
    {
        new_created = false;
        return false;
    }

    /* rewind all messages */
    in_msg.rewind ();
    out_msg.rewind ();
    tmp_msg.rewind ();

    /* Load one chunk at a time */
    chunk_mp.free_pool ();
    tMsgChunk* chunk = new tMsgChunk (m_msg_type, chunk_mp);
    if (chunk == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        return true;
    }

    int in_msg_subrecords = in_header->get_subrecords ();
    int out_msg_subrecords = 0;
    int tmp_msg_subrecords = 0;
    debug_log ('1', "I'm gonna to parse %d chunks from msg.",
        in_msg_subrecords);
    for (int i = 0; i < in_msg_subrecords; i++)
    {
        int bytes = chunk->get (in_msg.get_body ());
        if (bytes > 0)
        {
            debug_log ('1', "Chunk %d successfully read: %d bytes.",
                i, bytes);
            /* test the chunk if it pass the conditions */
            tNode const* n = NULL;
            const tFilterCond* tmp_cond;
            bool ok = true;
            FOR_ALL_LIST_ITEMS (n, &m_conds)
            {
                tmp_cond = static_cast<tFilterCond*> (n->data);
                if (!tmp_cond->test_chunk (*chunk))
                {
                    debug_log ('1', "Chunk %d didn't passed conditions.", i);
                    ok = false;
                    break;
                }
            }

            if (ok)
            {
                chunk->put (tmp_msg.get_body ());
                tmp_msg_subrecords++;
                debug_log ('1', "Chunk %d written to temp message.", i);
            }
            else
            {
                chunk->put (out_msg.get_body ());
                out_msg_subrecords++;
                debug_log ('1', "Chunk %d written to output message.", i);
            }
        }
        else if (!bytes)
        {
            log (L_WARN, "Zero size chunk in message.");
        }
        else
        {
            log (L_WARN, "Invalid chunk in message.");
        }
    }
    delete chunk;
    chunk_mp.free_pool ();

    /* update out&tmp msg subrecord count */
    debug_log ('1', "%d chunks passed all conditions.", out_msg_subrecords);
    out_msg.get_header ()->set_subrecords (out_msg_subrecords);
    tmp_msg.get_header ()->set_subrecords (tmp_msg_subrecords);

    /* close the out&tmp message */
    if (out_msg.close ())
    {
        log (L_ERROR, "Cannot close the message therefore can't pass "
            "message to next module.");
        return true;
    }
    if (tmp_msg.close ())
    {
        log (L_ERROR, "Cannot close the message therefore can't pass "
            "message to next module.");
        return true;
    }

    bool ret = false;

    if (m_action == RA_PASS || m_action == RA_PASS_DISCARD)
    {
        /* msg will be passed if it is not empty or
           if the in_msg was empty and no condition was specified */
        if (tmp_msg_subrecords > 0 ||
            in_msg_subrecords == 0 && m_conds.count () == 0)
        {
            debug_log ('1', "Temp msg will be passed.");
            ret = pass_msg (tmp_msg);
        }
    }

    switch (m_action)
    {
        case RA_PASS:
            new_created = false;
            break;
        case RA_DISCARD:
        case RA_PASS_DISCARD:
            /* if in_mgs had zero chunks do not return new msg */
            if (in_msg_subrecords > 0)
            {
                debug_log ('1', "New msg will be processed by "
                    "following rules.");
                new_created = true;
            }
            else
            {
                /* Empty input message, check number of chunk conditions */
                if (m_conds.count() > 0)
                    /* Non empty set of chunk conditions, pass it */
                    new_created = false;
                else
                    /* Empty message accepted by header conditions, discard */
                    new_created = true;
            }
            break;
    }

    return ret;
}

/* Passes message to all outputs of this rule */
bool tFilterRule::pass_msg (const tMsgBuf& msg)
{
    using namespace mframes;

    bool ret = false;

    /* pass the mesage to all my outputs */
    for (int i = 0; i < m_outs_count; i++)
    {
        if (m_outs[i]->send_message (msg))
        {
            log (L_ERROR, "Cannot pass the message to the output: %s",
                m_outs[i]->get_name ());
            ret = true;
        }
    }

    return ret;
}
