/**
 * @file
 * @brief Contains declaration of tFilterRule class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#ifndef ECHELOG_FILTERRULE_H
#define ECHELOG_FILTERRULE_H

#include "mframes/ioframe.h"
#include "filterheadercond.h"

/**
 * @brief Simple Condition Filter's rule class
 *
 * An instance of this class represents one rule from scond's configuration.
 * It is configured by init() method and behaves according to this
 * configuration. Rule creates conditions for itself and uses them in
 * message passing mechanism. Message is filtered by chunks and they
 * are divided to two groups. First group contains matched
 * chunks (those that satisfies all conditions), second the other chunks.
 * According to action required in rule's configuration these groups
 * forms new %messages and they are sent to outputs or returned from
 * pass_msg() as a new message to process by following rules.
 *
 * Rule configuration is a list of strings like this:
 * @verbatim
   MSG "msg_name"
   COND "attribute_name" == "value"
   PASS "output1" "output2"
   @endverbatim
 * Basic schema of rule configuration:
 * @verbatim
   (   (MSG "msg_name" |
        MSG.type == "msg_name" |
        MSG.msg_time (== | != | <= | >= | < | > ) "value[%Y-%m-%d %H:%M:%S]" |
        MSG.subrecords (== | != | <= | >= | < | > ) value |
        MSG.agent_id (== | != | <= | >= | < | > | ss | ~) "value" |
        MSG.agent_instance_id (== | != | <= | >= | < | > | ss | ~) "value" |
        MSG.host_id (== | != | <= | >= | < | > | ss | ~) "value" |
        MSG.flags (== | != ) (BASIC | INCREMENTAL | AGGREGABLE))+
       (COND "attribute_name" "cond_type" "value")*
    ( ((PASS | PASS_DISCARD) "output1" ("outputN")*) |
        DISCARD
    )
   )|
   ( MSGALL
     (PASS | PASS_DISCARD | DISCARD) "output1" ("outputN")*
   )
   @endverbatim
 * The parenthesis "(" ")" and asterisk "*" and colon "|" are used only to
 * describe the schema as a regular expression.
 * How to configure (write) a rule is described in more details in scond's
 * user's guide.
 *
 * Typical life-cycle of this class is creation, initialization by
 * init(), %messages passing by pass_msg(), then cleanup().
 */
class tFilterRule
{
    public:
        //-------------------------------
        // Interface
        //-------------------------------

        /**
         * @brief Constructor
         */
        tFilterRule ();

        /**
         * @brief Destructor
         */
        virtual ~tFilterRule ();

        /**
         * @brief Initialization of the rule
         *
         * Prepares rule for message passing.
         *
         * @param list
         * List of strings that holds rule configuration
         *
         * @param conf_mp
         * Memory pool where some configuration items could be stored.
         * It isn't freed it is a responsibility of rule's owner.
         */
        bool init (lib::tList& list, lib::tMemPool& conf_mp);

        /**
         * @brief Cleanup of the rule
         *
         * Performs all cleanup of allocated resources and resets
         * all attributes except of those allocated in configuration
         * memory pool.
         */
        void cleanup ();

        /**
         * @brief Filters and passes the message
         *
         * Input message is filtered by chunks and divided into
         * two groups those that fulfils conditions and those that
         * doesn't. From this groups new %messages are formed and
         * optionally sent to other modules. One of them could be
         * returned as an output message, in this case new_created
         * flag will be set to true and owner of the rule should pass
         * this new message to other rules.
         *
         * @param in_msg
         * Input message, must be prepared for reading.
         *
         * @param out_msg
         * Output message, must be prepared for writing, could be
         * modified, is closed.
         *
         * @param new_created
         * Set to true if new message is present in out_msg.
         *
         * @param tmp_msg
         * Message for teporary usage. Must be prepared for writing.
         * Could be modified, is closed.
         */
        bool pass_msg (messages::tMsgBuf& in_msg,
            messages::tMsgBuf& out_msg, bool& new_created,
            messages::tMsgBuf& tmp_msg);

    private:
        //-------------------------------
        // Internal types
        //-------------------------------

        /**
         * @brief Action to performed with filtered message
         */
        enum RuleAction
        {
            /// Create new message from chunks that matches conditions
            /// and pass it to outputs
            RA_PASS,
            /// Create new message from chunks that don't match conditions
            /// and return it as a new message
            RA_DISCARD,
            /// Do RA_PASS and RA_DISCARD together
            RA_PASS_DISCARD
        };

        //-------------------------------
        // Internal methods
        //-------------------------------

        /**
         * @brief Creates and initializes m_outs array
         *
         * Reads strings from list and interprets them as
         * names of outputs. m_outs array is allocated and filled
         * with appropriate output handles.
         *
         * @return
         * On success returns false. Otherwise true.
         */
        bool init_outs (const lib::tList& list);

        /**
         * @brief Creates and initializes conditions
         *
         * According to configuration in list the conditions
         * are created and initalized, used config string values from
         * list are removed.
         *
         * @param list
         * List with configuration string values.
         *
         * @param conf_mp
         * Memory pool where configuration items could be allocated. This pool
         * is not freed in rule, it must be freed by rule's owner itself.
         *
         * @return
         * On success returns false. Otherwise true.
         */
        bool init_conds (lib::tList& list, lib::tMemPool& conf_mp);

        /**
         * @brief Creates and initalizes header conditions
         *
         * According to configuration in list the header conditions
         * are created and initalized, used config string values from
         * list are removed.
         *
         * @param list
         * List with configuration string values.
         *
         * @param conf_mp
         * Memory pool where configuration items could be allocated. This pool
         * is not freed in rule, it must be freed by rule's owner itself.
         *
         * @return
         * On success returns false. Otherwise true.
         */
        bool init_header_conds (lib::tList& list, lib::tMemPool& conf_mp);

        /// Reverse method for init_outs()
        void cleanup_outs ();
        /// Reverse method for init_conds()
        void cleanup_conds ();
        /// Reverse method for init_header_conds()
        void cleanup_header_conds ();

        /**
         * @brief Passes messages to all outputs
         *
         * Message is sent to each output identified by handle
         * in array m_outs. If some error has occured while sending
         * a message to some output, tries to send message to all
         * others before leaves.
         *
         * @param msg
         * Message to be sent.
         *
         * @return
         * If successfully sent to all outputs returns false.
         * In case of error returns true.
         */
        bool pass_msg (const messages::tMsgBuf& msg);

        //-------------------------------
        // State attributes
        //-------------------------------

        /// Flag to set if the rule was successfuly initialized
        bool m_initialized;

        //-------------------------------
        // Owned components
        //-------------------------------

        /// Action to be performed with message (PASS,DISCARD ..)
        RuleAction m_action;
        /// Type of the message this rule applies to
        int32_t m_msg_type;
        /// Handle of the message this rule applies to
        int32_t m_msg_handle;
        /// List of conditions to be evaluated
        lib::tList m_conds;
        /// List of header conditions to be evaluated
        tFilterHeaderCond m_header_cond;
        /// Number of output handles in m_outs
        int m_outs_count;
        /// Array of pointers outputs to pass filtered part of message to
        mframes::tOFrame::tOutput** m_outs;
};

#endif //ifdef ECHELOG_FILTERRULE_H

