/**
 * @file
 * @brief Contains implementation of tFilterTarget class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "filtertarget.h"
#include "messages/msgbuf.h"

using namespace lib;
using namespace messages;

/* Basic scondfilter initialization */
tFilterTarget::tFilterTarget ()
{
    m_rules = NULL;
    m_rules_count = 0;
    m_rules_tmp_config = NULL;
    mp_cfpool = NULL;

    m_rule_section = new cfitem [3];
    if (m_rule_section == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        m_scfstate = SCF_BROKEN;
        return;
    }

    int i = 0;
    m_rule_section[i].name = "Rules";
    m_rule_section[i].type = CT_SECTION;
    m_rule_section[i].data = NULL;
    i++;
    m_rule_section[i].name = "Rule";
    m_rule_section[i].type = CT_LIST;
    m_rule_section[i].data = NULL;
    m_rule_section[i].varf = NULL;
    m_rule_section[i].varf_reset = NULL;
    m_rule_section[i].flags = ITEM_REQUIRED;
    i++;
    m_rule_section[i].name = NULL;
    m_rule_section[i].type = CT_STOP;
    m_rule_section[i].data = NULL;

    m_scfstate = SCF_CREATED;
}

/* Final cleanup */
tFilterTarget::~tFilterTarget ()
{
    ASSERT (m_scfstate == SCF_CREATED ||
            m_scfstate == SCF_BROKEN);

    if (m_scfstate == SCF_CREATED)
    {
        delete [] m_rule_section;
    }
}

/* Initialization of the filter */
bool tFilterTarget::init ()
{
    ASSERT (m_scfstate == SCF_CONFIGURED);

    int i = 0;
    for (; i < m_rules_count; i++)
        if (m_rules[i].init (m_rules_tmp_config[i], *mp_cfpool))
        {
            log (L_ERROR, "Configuration error in \"Rule\" value.");
            goto rules_clean;
        }

    /* delete the temporary shit */
    delete [] m_rules_tmp_config;
    m_rules_tmp_config = NULL;
    mp_cfpool = NULL;

    m_scfstate = SCF_INITIALIZED;

    return false;

rules_clean:
    for (i--; i >= 0; i--)
        m_rules[i].cleanup ();

    return true;
}

/* Reverse method for init */
void tFilterTarget::cleanup ()
{
    ASSERT (m_scfstate == SCF_INITIALIZED);

    for (int i = 0; i < m_rules_count; i++)
        m_rules[i].cleanup ();

    m_scfstate = SCF_CLEANED;
}

bool tFilterTarget::register_sections (lib::tMemPool& trash_pool)
{
    /* ensure the buffer is allocated */
    if (m_scfstate == SCF_BROKEN)
        return true;

    tList* list = new tList (trash_pool);
    if (list == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        return true;
    }
    m_rule_section[1].data = list;

    cf_register (m_rule_section);

    return false;
}

void tFilterTarget::unregister_sections ()
{
    cf_unregister (m_rule_section[0].name);

    delete static_cast<lib::tList *> (m_rule_section[1].data);
    m_rule_section[1].data = NULL;
}

bool tFilterTarget::process_config (lib::tMemPool& conf_mp, 
        lib::tMemPool& trash_pool)
{
    ASSERT (m_scfstate == SCF_CREATED);

    tNode const* n = NULL;
    int i = 0;

    /* ensure that at least one rule is specified */
    tList* list = static_cast<tList*> (m_rule_section[1].data);
    m_rules_count = list->count ();
    if (m_rules_count < 1)
    {
        log (L_ERROR, "At least one rule must be specified in"
            "configuration.");
        goto fail;
    }

    /* create rules array */
    m_rules = new tFilterRule [m_rules_count];
    if (m_rules == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        goto fail;
    }

    /* create temporary rules config */
    m_rules_tmp_config = new tList [m_rules_count];
    if (m_rules_tmp_config == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        goto rules_clean;
    }

    /* split each rule into separate strings */
    FOR_ALL_LIST_ITEMS (n, list)
    {
        if (s_strpsep (m_rules_tmp_config[i],
            static_cast<char *> (n->data), conf_mp))
        {
            log (L_ERROR, "Configuration error in \"Rule\" value.");
            goto rules_tmp_clean;
        }
        i++;
    }

    /* save the conf pool */
    mp_cfpool = &conf_mp;

    m_scfstate = SCF_CONFIGURED;

    return false;

rules_tmp_clean:
    delete [] m_rules_tmp_config;
    m_rules_tmp_config = NULL;
rules_clean:
    delete [] m_rules;
    m_rules = NULL;
fail:
    return true;
}

void tFilterTarget::cleanup_config ()
{
    ASSERT (m_scfstate == SCF_CLEANED ||
            m_scfstate == SCF_CONFIGURED);

    delete [] m_rules;
    m_rules = NULL;
    if (m_rules_tmp_config != NULL)
    {
        delete [] m_rules_tmp_config;
        m_rules_tmp_config = NULL;
    }
    m_rules_count = 0;
    mp_cfpool = NULL;

    m_scfstate = SCF_CREATED;
}

/* Ensures message passing to proper outputs */
bool tFilterTarget::process_msg (tMsgBuf& message)
{
    ASSERT (m_scfstate == SCF_INITIALIZED);

    static char out_msg_buff [MAX_MESSAGE_SIZE];
    static char tmp_msg_buff [MAX_MESSAGE_SIZE];
    tMsgBuf message2;
    tMsgBuf message3;
    tMsgBuf* in = &message;
    tMsgBuf* out = &message2;
    tMsgBuf* tmp = &message3;
    bool ret = false;
    bool new_created = false;
    int i;

    /* Prepare the message buffers */
    if (out->prepare_to_write (out_msg_buff, MAX_MESSAGE_SIZE))
    {
    	log (L_WARN, "Unable to prepare a message for writing, "
    	    "aborting message passing.");
    	ret = true;
        goto exit;
    }
    if (tmp->prepare_to_write (tmp_msg_buff, MAX_MESSAGE_SIZE))
    {
    	log (L_WARN, "Unable to prepare a message for writing, "
    	    "aborting message passing.");
    	ret = true;
        goto exit;
    }

    /* copy the header to out & tmp message */
    out->get_header()->copy_envelope (*in->get_header());
    tmp->get_header()->copy_envelope (*in->get_header());

    /* pass message through each rule */
    for (i = 0; i < m_rules_count; i++)
    {
        if (m_rules[i].pass_msg (*in, *out, new_created, *tmp))
        {
            log (L_ERROR, "Cannot pass message through rule nr. %d", i);
            ret = true;
            break;
        }
        else
        {
            /* if new message was created, continue passing the new */
            if (new_created)
            {
                /* out is closed so close only in */
                in->close ();

                /* we're not interested newly created empty messages */
                if (out->get_header ()->get_subrecords () == 0)
                    break;

                /* switch the buffers */
                tMsgBuf* exch;
                exch = in; in = out; out = exch;

                /* Prepare the message buffers */
                const char * buff = in->get_buffer ();
                in->reset ();
                if (in->prepare_to_read (buff, MAX_MESSAGE_SIZE))
                {
                    log (L_WARN, "Unable to prepare a message for reading, "
                        "aborting message passing.");                 	
                    ret = true;
                    break;
                }
                buff = out->get_buffer ();
                out->reset ();
                if (out->prepare_to_write (buff, MAX_MESSAGE_SIZE))
                {
                    log (L_WARN, "Unable to prepare a message for writing, "
                        "aborting message passing.");                 	
                    ret = true;
                    break;
                }

                new_created = false;
            }
            in->rewind ();
            out->rewind ();
            tmp->rewind ();
        }
    }

exit:
    return ret;
}
