/**
 * @file
 * @brief Contains declaration of tFilterTarget class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#ifndef ECHELOG_FILTERTRAGET_H
#define ECHELOG_FILTERTARGET_H

#include "mframes/msgtarget.h"
#include "filterrule.h"

/**
 * @brief Simple Condition Filter's filter class
 *
 * This class implements the filters::tFilter interface and
 * includes all Simple Condition Filter's specific functionality.
 * Reads the filter's configuration, parses and interprets it to
 * create specified rules. Also the message passing schema is
 * implemented here in pass_msg().
 *
 * Filters configuration is a list of strings, each contains
 * strings which forms new list of strings as descibed in
 * filters::tSCondRule documentation.
 *
 * Typical life-cycle of this class is creation, initialization by
 * init(), %messages passing by pass_msg(), then cleanup().
 */
class tFilterTarget : public mframes::tMsgTarget
{
    public:
        //-------------------------------
        // Interface
        //-------------------------------

        /**
         * @brief Constructor
         *
         * Performs basic initialization of the instance.
         */
        tFilterTarget ();

        /**
         * @brief Destructor
         *
         * Ensures final cleanup of the instance.
         */
        virtual ~tFilterTarget ();

        /**
         * @brief Initialization of the filter
         *
         * This method prepares filter for use with pass_msg().
         * It reads configuration and creates and prepares rules.
         *
         * @return
         * If succesful returns false. On failure return true.
         */
        virtual bool init ();

        /**
         * @brief Cleanup of the filter
         *
         * Releases and resets all allocated resources in init().
         * It is the reverse function for init().
         */
        virtual void cleanup ();

        /**
         * @brief Message passing cycle implementation
         *
         * Message is passed to each rule in order they were
         * in configuration. Rule will filter the message and ensures
         * the sending. Rule can modify the message which is then passed
         * to following rules.
         *
         * @param message
         * Message to be filtered and sent to other modules. Must be prepared
         * for reading.
         *
         * @return
         * If the message was succesfully processed returns false.
         * If some error has occured returns true.
         */
        virtual bool process_msg (messages::tMsgBuf& message);

        //------------------------------------------
        // Implementation of tConfigurable interface
        //------------------------------------------

        /**
         * @brief TODOC:
         */
        virtual bool register_sections (lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void unregister_sections ();

        /**
         * @brief TODOC:
         */
        virtual bool process_config (lib::tMemPool& conf_mp,
                lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void cleanup_config ();

    private:
        //-------------------------------
        // Internal types
        //-------------------------------

        /**
         * @brief Description of the filter state
         */
        enum SCFState
        {
            /// initalization failed during constructor
            SCF_BROKEN,
            /// successfully created
            SCF_CREATED,
            /// successfully configured
            SCF_CONFIGURED,
            /// successfully initialized
            SCF_INITIALIZED,
            /// successfully cleaned
            SCF_CLEANED
        };

        //-------------------------------
        // Internal methods
        //-------------------------------

        //-------------------------------
        // State attributes
        //-------------------------------

        /// State of the filter
        SCFState m_scfstate;

        //-------------------------------
        // Owned components
        //-------------------------------

        /// Desription of filter's configuration
        lib::cfitem* m_rule_section;
        /// Number of rules in m_rules array
        int m_rules_count;
        /// Array of rules to pass msgs to
        tFilterRule* m_rules;

        //-------------------------------
        // Used-only components - temporary
        //-------------------------------

        lib::tMemPool* mp_cfpool;
        lib::tList* m_rules_tmp_config;

};

#endif //ifdef ECHELOG_FILTERTARGET_H

