/**
 * @file
 * @brief Contains implementation of tHeaderCond class
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "filterheadercond.h"
#include "messages/msgmgr.h"
#include "messages/typelist.h"
#include <string.h>
#include <regex.h>

using namespace lib;
using namespace messages;


//----------------------------------------------------------
// Declaration & Implementation of the tHeaderCond class
//----------------------------------------------------------
tHeaderCond::tHeaderCond (tHeaderOp new_op, tHeaderRec new_rec, 
        void *new_operand)
    : op (new_op), rec (new_rec), operand (new_operand)
{
}


tHeaderCond::~tHeaderCond ()
{
    if (op == OP_REGEXP)
    {
        regex_t *preg = static_cast<regex_t *>(operand);
        if (preg)
            regfree (preg);
    }
}


bool tHeaderCond::test_header (const messages::tMsgHeader& header) const
{
    bool ret = false;

    char *tmp_str = NULL;
    switch (rec)
    {
        case REC_TYPE:
            ret = test_number (header.get_msg_type());
            break;
        case REC_TIME:
            ret = test_number (header.get_msg_time());
            break;
        case REC_SUBRECORDS:
            ret = test_number (header.get_subrecords());
            break;
        case REC_FLAGS:
            ret = test_number (header.get_flags());
            break;

        case REC_AGENT_ID:
            tmp_str = header.get_agent_id();
            ret = test_string (tmp_str);
            break;
        case REC_AGENT_INSTANCE_ID:
            tmp_str = header.get_agent_instance_id();
            ret = test_string (tmp_str);
            break;
        case REC_HOST_ID:
            tmp_str = header.get_host_str();
            ret = test_string (tmp_str);
            break;

        default:
            ASSERT ("Nonsential header condition.");
    }

    if (tmp_str)
        xfree (tmp_str);

    return ret;
}


bool tHeaderCond::test_number (int32_t num) const
{
    ASSERT (operand);

    int32_t num2 = *(static_cast<int32_t *>(operand));
    switch (op)
    {
        case OP_EQ:
            return num == num2;
        case OP_NEQ:
            return num != num2;
        case OP_GEQ:
            return num >= num2;
        case OP_LEQ:
            return num <= num2;
        case OP_LE:
            return num < num2;
        case OP_GE:
            return num > num2;

        default:
            ASSERT ("Nonsential numeric header condition.");
    }
    return false;
}


bool tHeaderCond::test_string (const char *str) const
{
    ASSERT (operand);

    if (op == OP_SUBSTRING)
    {
        char *needle = static_cast<char *>(operand);
        if (strstr (str, needle) != NULL)
            return true;
    }
    else if (op == OP_REGEXP)
    {
        regex_t *preg = static_cast<regex_t *>(operand);

        /* execute regex with the value */
        if (!regexec(preg, str, 0, NULL, 0))
            return true;
    }

    const char *str2 = (static_cast<const char *>(operand));
    int ret = strcmp (str, str2);

    switch (op)
    {
        case OP_EQ:
            return ret == 0;
        case OP_NEQ:
            return ret != 0;
        case OP_GEQ:
            return ret >= 0;
        case OP_LEQ:
            return ret <= 0;
        case OP_LE:
            return ret < 0;
        case OP_GE:
            return ret > 0;

        default:
            ASSERT ("Nonsential string header condition.");
    }

    return false;
}
