/**
 * @file
 * @brief Contains declaration of tHeaderCond class
 *
 * @author Kamil Toman <ktoman@email.cz>
 */

#ifndef ECHELOG_HEADERCOND_H
#define ECHELOG_HEADERCOND_H

/// Header comparison operators
enum tHeaderOp
{
    OP_UNKNOWN = 0,     ///< error, uknown operation
    OP_EQ,              ///< <tt>==</tt>, equality operator
    OP_NEQ,             ///< <tt>!=</tt>, non-equality operator
    OP_GEQ,             ///< <tt>&gt;=</tt>, greater-or-equal operator
    OP_LEQ,             ///< <tt>&lt;=</tt>, less-or-equal operator
    OP_LE,              ///< <tt>&lt;</tt>, less-than operator
    OP_GE,              ///< <tt>&gt;</tt>, greater-to operator
    OP_SUBSTRING,       ///< <tt>ss</tt>, substring operator
    OP_REGEXP           ///< <tt>~</tt>, match operator (extended regexp)
};

/// <tt>tHeaderRec</tt> specifies which header record is to be accessed
enum tHeaderRec
{
    REC_TYPE = 0,               ///< message type record
    REC_TIME,                   ///< message timestamp record
    REC_SUBRECORDS,             ///< message number of subrecords
    REC_AGENT_ID,               ///< message agent id record
    REC_AGENT_INSTANCE_ID,      ///< message agent instance id record
    REC_HOST_ID,                ///< messsage host id record
    REC_FLAGS                   ///< messsage flags record
};

/**
 * @brief Simple Condition Filter's header condition class
 *
 * This class is analogical to <tt>tFilterCond</tt> class but it
 * operates on header records.
 *
 * There are several types of conditions that could be
 * created. All conditions can be desribed by this regular expression:
 * @verbatim
 * (MSG "msg_name" |
 *  MSG.type == "msg_name" |
 *  MSG.msg_time (== | != | <= | >= | < | > ) "value[%Y-%m-%d %H:%M:%S]" |
 *  MSG.subrecords (== | != | <= | >= | < | > ) value |
 *  MSG.agent_id (== | != | <= | >= | < | > | ss | ~) "value" |
 *  MSG.agent_instance_id (== | != | <= | >= | < | > | ss | ~) "value" |
 *  MSG.host_id (== | != | <= | >= | < | > | ss | ~) "value" |
 *  MSG.flags (== | != ) (BASIC | INCREMENTAL | AGGREGABLE))+
 *
 *  All conditions are tested in sequence against specified header records
 *  by function test_header().
 */
class tHeaderCond
{
    private:

        tHeaderOp op;   ///< operation to be performed
        tHeaderRec rec; ///< header record to operate on
        void *operand;  ///< comparison value

        /// Tests condition against <tt>num</tt>. Returns true on success.
        bool test_number (int32_t num) const;
        /// Tests condition against <tt>str</tt>. Returns true on success.
        bool test_string (const char *str) const;

    public:

        /**
         * @brief Constructor
         *
         * Performs basic initialization of the instance.
         */
        tHeaderCond (tHeaderOp new_op, tHeaderRec new_rec, 
            void *new_operand);

        /**
         * @brief Destructor
         *
         * Ensures final cleanup of the instance.
         */
        ~tHeaderCond ();

        /// Creates a new instance of <tt>tHeaderCond</tt>.
        static tHeaderCond *create_new (tHeaderOp op, tHeaderRec rec, 
                void *operand)
            {
                tHeaderCond *hc = new tHeaderCond (op, rec, operand);
                if (!hc)
                    lib::die ("Not enough memory");

                return hc;
            }

        /// Test condition against given record field. Returns true on success.
        bool test_header (const messages::tMsgHeader& header) const;
};


#endif
