/**
 * @file
 * @brief Contains implementation of tLogStorTarget class
 *
 * @author Ales Privetivy <ales@ucw.cz>, Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "logstortarget.h"
#include "messages/msgbuf.h"
#include "messages/msgmgr.h"

using namespace lib;
using namespace messages;
using namespace mframes;

/* Basic logstor initialization */
tLogStorTarget::tLogStorTarget ()
{
    m_rules = NULL;
    m_rules_cnt = 0;
    m_output = NULL;

    m_logstor_section = new cfitem [3];
    if (m_logstor_section == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        m_lsstate = LS_BROKEN;
        return;
    }

    int i = 0;
    m_logstor_section[i].name = "Logstor";
    m_logstor_section[i].type = CT_SECTION;
    m_logstor_section[i].data = NULL;
    i++;
    m_logstor_section[i].name = "Catch";
    m_logstor_section[i].type = CT_LIST;
    m_logstor_section[i].data = NULL;
    i++;
    m_logstor_section[i].name = NULL;
    m_logstor_section[i].type = CT_STOP;
    m_logstor_section[i].data = NULL;

    m_lsstate = LS_CREATED;
}

/* Final cleanup */
tLogStorTarget::~tLogStorTarget ()
{
    ASSERT (m_lsstate == LS_CREATED ||
            m_lsstate == LS_BROKEN);

    if (m_lsstate == LS_CREATED)
    {
        delete [] m_logstor_section;
    }
}

/* Initialization of the filter */
bool tLogStorTarget::init ()
{
    ASSERT (m_lsstate == LS_CONFIGURED);

    m_output = tOFrame::get_actual_oframe ()->
                get_output_by_name (NULL);

    m_lsstate = LS_INITIALIZED;

    return false;
}

/* Reverse method for init */
void tLogStorTarget::cleanup ()
{
    ASSERT (m_lsstate == LS_INITIALIZED);

    m_output = NULL;

    m_lsstate = LS_CLEANED;
}

bool tLogStorTarget::register_sections (lib::tMemPool& trash_pool)
{
    /* ensure the buffer is allocated */
    if (m_lsstate == LS_BROKEN)
        return true;

    tList* list = new tList (trash_pool);
    if (list == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        return true;
    }
    m_logstor_section[1].data = list;

    cf_register (m_logstor_section);

    return false;
}

void tLogStorTarget::unregister_sections ()
{
    cf_unregister (m_logstor_section[0].name);

    delete static_cast<lib::tList *> (m_logstor_section[1].data);
    m_logstor_section[1].data = NULL;
}

bool tLogStorTarget::process_config (lib::tMemPool& conf_mp,
        lib::tMemPool& trash_pool)
{
    ASSERT (m_lsstate == LS_CREATED);

    tNode const* n = NULL;
    int i = 0;

    /* add catch rules */
    tList* list = static_cast<tList*> (m_logstor_section[1].data);
    m_rules_cnt = list->count ();
    if (m_rules_cnt > 0)
    {
        m_rules = static_cast<struct tRule*>
             ( conf_mp.alloc( sizeof (struct tRule) * m_rules_cnt ));
        memset( m_rules, 0, sizeof (struct tRule) * m_rules_cnt );

        /* split each rule into separate strings */
        FOR_ALL_LIST_ITEMS (n, list)
        {
            /* parse one line */
            if (parse_config_line (static_cast<char *> (n->data), i, conf_mp))
            {
                log (L_ERROR, "Error in \"Catch\" clause.");
                goto rules_clean;
            }
            i++;
        }
    }

    m_lsstate = LS_CONFIGURED;

    return false;

rules_clean:
    /* release all rules data and close log-files */
    for(;i >= 0;i--)
        s_close(m_rules[i].logfd);
    m_rules = NULL;
    m_rules_cnt = 0;

    return true;
}

void tLogStorTarget::cleanup_config ()
{
    ASSERT (m_lsstate == LS_CLEANED ||
            m_lsstate == LS_CONFIGURED);

    if (m_rules_cnt > 0)
    {
        for(int i = 0;i < m_rules_cnt; i++)
            s_close(m_rules[i].logfd);
        m_rules = NULL;
        m_rules_cnt = 0;
    }

    m_lsstate = LS_CREATED;
}

/* Ensures message passing to proper outputs */
bool tLogStorTarget::process_msg (tMsgBuf& message)
{
    ASSERT (m_lsstate == LS_INITIALIZED);

    tMsgHeader *header = message.get_header ();
    byte *hptr = header->get_host_id ();
    char *aptr = header->get_agent_id ();
    char *iptr = header->get_agent_instance_id();
    int type = header->get_msg_type();
    int i;

    for (i = 0; i < m_rules_cnt; i++)
        if (match_rule (i, hptr, aptr, iptr, type)) break;

    xfree (hptr);
    xfree (aptr);
    xfree (iptr);

    if (i < m_rules_cnt)
    {
        WRITELOCK_FILE (m_rules[i].logfd);
        int res = m_printer.print_message (m_rules[i].logfd, message);
        UNLOCK_FILE (m_rules[i].logfd);

        if (res)
        {
            log (L_ERROR, "Error during message printing to logfile.");
            goto fail;
        }
        
    }
    else
        if (m_output && m_output->send_message (message))
        {
            log (L_ERROR, "Can't pass messages to next module");
            goto fail;
        }

    return false;
fail:
    return true;
}

bool tLogStorTarget::parse_config_line (char* line, int crule,
    tMemPool& conf_mp)
{
    tList paramlist (conf_mp);
    tList iplist (conf_mp);
    char *host, *agent, *type_name, *fname, *instance;
    int fd, type;

    if (s_strpsep (paramlist, line, conf_mp))
    {
        log (L_ERROR, "Error in config file: "
            "CATCH clause number %d.", crule+1);
        goto fail;
    }

    host = agent = type_name = fname = NULL;
    while (paramlist.count () > 0)
    {
        char *attr_name = static_cast<char *> (paramlist.remove_first ());
        char *value;

        if (paramlist.count () > 0)
            value = static_cast<char *> (paramlist.remove_first ());
        else
        {
            log (L_ERROR, "Error in config file: "
                "CATCH clause number %d.", crule+1);
            goto fail;
        }

        if (!strcmp("Host", attr_name))
            host = value;
        else if (!strcmp("Agent", attr_name))
            agent = value;
        else if (!strcmp("MessageType", attr_name))
            type_name = value;
        else if (!strcmp("Instance", attr_name))
            instance = value;
        else if (!strcmp("Logfile", attr_name))
            fname = value;
        else
        {
            log (L_ERROR, "Error in config file: "
                "CATCH clause number %d: Unknown "
                "attribute %s.", crule+1, attr_name);
            goto fail;
        }
    }

    if (host && !strcmp(host,"*")) host = NULL;
    if (agent && !strcmp(agent,"*")) agent = NULL;
    if (type_name && !strcmp(type_name,"*")) type_name = NULL;
    if (instance && !strcmp(instance,"*")) instance = NULL;
    if (!fname)
    {
        log (L_ERROR, "Error in config file: "
            "CATCH clause number %d: Logfile name "
            "must be specified.", crule + 1);
        goto fail;
    }
    if (type_name)
    {
        if (gb_message_manager.get_handle_by_name (type_name) == -1)
        {
            log (L_ERROR, "Error in config file: "
                "CATCH clause number %d: Unknown "
                "message type \"%s\".", crule + 1, type_name);
            goto fail;
        }
        type = gb_message_manager.get_message_type_by_handle (
                gb_message_manager.get_handle_by_name (type_name));
    } else type = -1;

    if (host)
    {
        int i;

        for (i = 0;i < (int)strlen (host); i++)
            if (host[i]=='.') host[i]=' ';

        if (s_strpsep (iplist, host, conf_mp) || iplist.count ()!= 4)
        {
            log (L_ERROR, "Error in config file: "
                "CATCH clause number %d: Parsing host "
                "attribute.", crule + 1);
            goto fail;
        }

        m_rules[crule].hostptr = static_cast<byte *>
                               (conf_mp.alloc(4*sizeof(byte)));
        for (i=0; i < 4; i++)
        {
            long num;

            if (s_stol (static_cast<char *> (iplist.remove_first()),
                        &num, 0, 255))
            {
                log (L_ERROR, "Error in config file: "
                    "CATCH clause number %d: "
                    "Bad format of IP in Host attribute.",
                    crule + 1);
                goto host_clean;
            }
            m_rules[crule].hostptr[i] = num;
        }
    } else m_rules[crule].hostptr = NULL;


    fd = s_open ( fname, O_WRONLY | O_CREAT | O_APPEND,
                S_IRUSR | S_IWUSR);

    if (fd < 0)
    {
        log (L_ERROR, "Can't open logfile \"%s\" to log in.", fname);
        goto host_clean;
    }

    m_rules[crule].agentstr = agent;
    m_rules[crule].messagetype = type_name;
    m_rules[crule].logfilename = fname;
    m_rules[crule].msg_type = type;
    m_rules[crule].instance = instance;
    m_rules[crule].logfd = fd;

    return false;

host_clean:
    m_rules[crule].hostptr = NULL;
fail:
    return true;
}

int tLogStorTarget::match_rule (int rule, byte *hptr, char *aptr,
    char *iptr, int type)
{
    if (m_rules[rule].hostptr!=NULL)
        for(int i=0;i<4;i++)
            if (m_rules[rule].hostptr[i]!=hptr[i]) return 0;

    if (m_rules[rule].agentstr!=NULL && strcmp(m_rules[rule].agentstr, aptr))
        return 0;

    if (m_rules[rule].msg_type>=0 && m_rules[rule].msg_type!=type)
        return 0;

    if (m_rules[rule].instance!=NULL && strcmp(m_rules[rule].instance, iptr))
        return 0;

    return 1;
}
