/**
 * @file
 * @brief Contains declaration of tLogStorTarget class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#ifndef ECHELOG_LOGSTORTARGET_H
#define ECHELOG_LOGSTORTARGET_H

#include "mframes/msgtarget.h"
#include "mframes/ioframe.h"
#include "messages/msgprinter.h"


struct tRule
{
    byte *hostptr;
    char *agentstr;
    char *messagetype;
    char *instance;
    int msg_type;
    char *logfilename;
    int logfd;
};


/**
 * @brief TODOC:
 */
class tLogStorTarget : public mframes::tMsgTarget
{
    public:
        //-------------------------------
        // Interface
        //-------------------------------

        /**
         * @brief Constructor
         *
         * Performs basic initialization of the instance.
         */
        tLogStorTarget ();

        /**
         * @brief Destructor
         *
         * Ensures final cleanup of the instance.
         */
        virtual ~tLogStorTarget ();

        /**
         * @brief Initialization of the filter
         *
         * This method prepares filter for use with pass_msg().
         * It reads configuration and creates and prepares rules.
         *
         * @return
         * If succesful returns false. On failure return true.
         */
        virtual bool init ();

        /**
         * @brief Cleanup of the filter
         *
         * Releases and resets all allocated resources in init().
         * It is the reverse function for init().
         */
        virtual void cleanup ();

        /**
         * @brief Message passing cycle implementation
         *
         * Message is passed to each rule in order they were
         * in configuration. Rule will filter the message and ensures
         * the sending. Rule can modify the message which is then passed
         * to following rules.
         *
         * @param message
         * Message to be filtered and sent to other modules. Must be prepared
         * for reading.
         *
         * @return
         * If the message was succesfully processed returns false.
         * If some error has occured returns true.
         */
        virtual bool process_msg (messages::tMsgBuf& message);

        //------------------------------------------
        // Implementation of tConfigurable interface
        //------------------------------------------

        /**
         * @brief TODOC:
         */
        virtual bool register_sections (lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void unregister_sections ();

        /**
         * @brief TODOC:
         */
        virtual bool process_config (lib::tMemPool& conf_mp,
                lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void cleanup_config ();

    private:
        //-------------------------------
        // Internal types
        //-------------------------------

        /**
         * @brief Description of the filter state
         */
        enum LSState
        {
            /// initalization failed during constructor
            LS_BROKEN,
            /// successfully created
            LS_CREATED,
            /// successfully configured
            LS_CONFIGURED,
            /// successfully initialized
            LS_INITIALIZED,
            /// successfully cleaned
            LS_CLEANED
        };

        //-------------------------------
        // Internal methods
        //-------------------------------

        bool parse_config_line (char* line, int crule,
            lib::tMemPool& conf_mp);

        int match_rule (int rule, byte *hptr, char *aptr,
            char *iptr, int type);

        //-------------------------------
        // State attributes
        //-------------------------------

        /// State of the filter
        LSState m_lsstate;

        //-------------------------------
        // Owned components
        //-------------------------------

        /// Desription of filter's configuration
        lib::cfitem* m_logstor_section;
        /// Number of rules in m_rules array
        int m_rules_cnt;
        /// Array of rules to pass msgs to
        tRule *m_rules;
        /// Output to pass messages to
        mframes::tOFrame::tOutput* m_output;
        /// Printer we use
        messages::tMsgPrinter m_printer;
};

#endif //ifdef ECHELOG_FILTERTARGET_H

