/**
 * @file
 * @brief Echelog Library -- getlog
 *
 * @module getlog log top level buffer
 * 
 * This agent periodically reads specified logs and passes information
 * stored there as echelog messages.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */


#include "getlog.h"
#include <string.h>

using namespace lib;


/** Advance to next file cache and return true on rotation */
inline bool tBuffer::next_buffer()
{
    bool wrapped = false;

    current_buffer++;
    if ((current_buffer - buffers) >= num_of_buffers)
    {
        current_buffer = buffers;
        wrapped = true;
    }

    return wrapped;
}


/** Get a single line from a file cache (preserve the order specified
 * in getlog's configuration */
void tBuffer::get_line (tBufferResponse &r)
{
    for (;;)
    {
        current_buffer->get_line (r);
        switch (r.state)
        {
            case BR_EMPTY:
                /* check if we're not reading the last buffer */
                if (next_buffer ())
                    return;     /* last buffer, bail out */
                break;
                
            case BR_EIO:                /* io error */
                r.str_length = s_asprintf (&r.str, 
                        "Can't read file '%s', skipping it.", 
                        current_buffer->get_bname ()) - 1;
                next_buffer ();
                return;

            case BR_CORRUPTED:
                r.str_length = s_asprintf (&r.str,
                        "Corrupted records in file %s, skipping extra "
                        "characters", current_buffer->get_bname ());
                return;

            case BR_ROTATED:
                r.str_length = s_asprintf (&r.str,
                        "File '%s' has been altered, expecting LOG ROTATION",
                         current_buffer->get_bname ());
                return;
                
            case BR_SHORTENED:
                r.str_length = s_asprintf (&r.str,
                        "File '%s' has been shortened, suspecting LOG EDITING",
                         current_buffer->get_bname ());
                return;

            default:            /* pass information and bail out */
                return; 
        }
    }
}


/** Load state files for all file caches */
void tBuffer::load_state_files ()
{
    for (int i = 0; i < num_of_buffers; i++)
        buffers[i].load_state ();
}


/** Save state files for all file caches */
void tBuffer::save_state_files ()
{
    for (int i = 0; i < num_of_buffers; i++)
        if (buffers[i].save_state ())
            log (L_ERROR, "Can't write state for file '%s' to dir '%s'",
                    buffers[i].get_bname(), state_dir);
}


/** Register filenames for all file caches */
void tBuffer::register_files (const char *conf_state_dir, lib::tList &filenames)
{
    struct tNode const *c;
    int i;

    num_of_buffers = filenames.count ();
    if (num_of_buffers <= 0)
        die ("No log files to read.");

   /* check conf state dir write access */
    struct_stat st;
    sd_stat (conf_state_dir, &st);
    sd_eaccess (&st, F_OK | R_OK | W_OK | X_OK);
    state_dir = s_strdup (conf_state_dir);

    /* for each file cache register the filename 
     * given in getlog configuration */
    buffers = new tFileBuffer [num_of_buffers];
    if (!buffers)
        die ("Not enough memory.");
    i = 0;
    FOR_ALL_LIST_ITEMS (c, &filenames)
        buffers[i++].register_file (state_dir, static_cast<char *> (c->data));

    /* set current buffer to the very first */
    current_buffer = buffers;
}


/** Unregister all log files */
void tBuffer::unregister_files ()
{
    delete[] buffers;
    buffers = NULL;
    xfree (state_dir);
    state_dir = NULL;
}
