/**
 * @file
 * @brief Echelog Library -- getlog
 *
 * @module getlog log file buffering module
 *
 * This agent periodically reads specified logs and passes information
 * stored there as echelog messages.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/archdep.h"
#include <string.h>
#include <sys/param.h>


#include "messages/plainbuf.h"
#include "messages/testbuf.h"
#include "getlog.h"

using namespace lib;


/** Create an empty cache for a single file */
tFileBuffer::tFileBuffer () : 
    bfd (-1), bpos (0), cache_pos (cache_start), 
    cache_end(cache_start - 1), needs_skip (false), 
    st_filename (NULL), report_new_file (true), bname_len (0), bname (NULL)
{
    /* also, clear log state information */
    memset (static_cast<void *> (&bst), 0, sizeof (struct_stat));
    memset (static_cast<void *> (&old_bst), 0, sizeof (struct_stat));
}


/** Flush and free all cache data */
tFileBuffer::~tFileBuffer ()
{
    /* Close the log file if needed */
    if (bfd >= 0 && s_close (bfd) < 0)
        log (L_ERROR, "Can't close file '%s'", bname);
    bfd = -1;
    xfree (st_filename);
    st_filename = NULL;
    xfree (bname);
    bname = NULL;
    bname_len = 0;
}


/** Atomic load of state variables from a buffer buf */
int tFileBuffer::load_state_info (char *buf, long length)
{
    using messages::tPlainBuf;
    int32_t inode, size, nbpos;
    char *filename;
    tPlainBuf b(buf, length);

    /* state fmt: inode, size, position, filename */
    READ_INT32 (&b, inode);
    READ_INT32 (&b, size);
    READ_INT32 (&b, nbpos);
    READ_COMMON (cstring, &b, filename);

    /* check if the filenames match */
    if (strcmp (bname, filename))
        return -1;      /* someone doing tricks with state files? */

    /* save in values from previous session */
    bst.st_ino = inode;
    bst.st_size = size;
    bpos = nbpos;
    
    /* successful return */
    return 0;
}


/** Open and load the cache state from previous session */
void tFileBuffer::load_state ()
{
    int r, fd;
    struct_stat st;
    char *st_buf;

    debug_log ('2', "Trying '%s' as a state file", st_filename);

    /* check if there is a state file */
    if (s_eaccess (st_filename, F_OK))
    {
        debug_log ('2', "No state file");
        return;
    }

    /* if so, check whether we can read/update it */
    fd = s_open (st_filename, O_RDWR, S_IRUSR | S_IWUSR);
    if (fd == -1)
    {
        log (L_WARN, "Can't open state file '%s'", st_filename);
        return;
    }
    r = s_fstat (fd, &st);
    if (r < 0 || !st.st_size)
    {
        log (L_WARN, "Bad file state file '%s', skipping", st_filename);
        return;
    }

    /* ok, try to read the previous session */
    st_buf = static_cast<char *> (xmalloc (st.st_size));
    sd_read (fd, st_buf, static_cast<off_t> (st.st_size));
    r = load_state_info (st_buf, st.st_size);
    xfree (st_buf);
    if (r)
    {
        log (L_WARN, "Corrupted state file '%s', skipping", st_filename);
        return;
    }

    debug_log ('2', "State loaded (%s)", bname);
    report_new_file = false;

    /* close it for now */
    sd_close (fd);
}


/** Save the cache state of the current session to disk */
int tFileBuffer::save_state ()
{
    using messages::tPlainBuf;
    
    tPlainBuf b;
    int sfd;

    /* recreate the state file */
    sfd = s_open (st_filename, O_WRONLY | O_CREAT | O_TRUNC | O_SYNC, 
            S_IRUSR | S_IWUSR);
    if (sfd == -1)
        return -1;

    /* state fmt: inode, size, position, filename */
    STORE_INT32 (&b, bst.st_ino);
    STORE_INT32 (&b, bst.st_size);
    STORE_INT32 (&b, bpos);
    STORE_COMMON (cint32, &b, bname_len);
    STORE_COMMON (string, &b, bname);

    /* write and sync it */
    if (s_write (sfd, static_cast<void *>(const_cast<char *>(b.get_buffer ())), 
                b.get_store_length ()) < 0)
        return -1;

    /* finish */
    if (s_close (sfd))
        return -1;

    return 0;
}


/** Register a log file and do the basic state initialization */
void tFileBuffer::register_file (const char *state_dir, const char *filename)
{
    ASSERT (filename);
    log (L_INFO, "Registering log file '%s'", filename);

    if (filename[0] != '/')
        die ("Configuration error: log name '%s' is not an absolute pathname.",
                filename);

    /* check read access */
    sd_stat (filename, &bst);
    sd_eaccess (&bst, F_OK | R_OK);

    /* register it and create the state filename */
    bname = s_strdup (filename);
    bname_len = strlen (filename);

    char *tmp = s_strnreplace (filename, "/", "_", bname_len); 
    st_filename = s_strconcat (state_dir, "/", tmp, DEFAULT_ST_EXT, NULL);
    xfree (tmp);
}


/** Fetch data from disk to getlog's cache */
tBufferState tFileBuffer::fetch (char *buf)
{
    struct_stat st;

    if (report_new_file)
    {
        report_new_file = false;
        return BR_NEW;
    }   

    /* we might need to reopen the log file */
    if (bfd < 0) 
    {
        debug_log ('2', "Reopening file '%s'", bname);
        if ((bfd = s_open (bname, O_RDONLY)) < 0)
            return BR_EIO;
    }

    /* state fmt: inode, size, position, filename */
    if (s_fstat (bfd, &st))
    {
        s_close (bfd), bfd = -1;
        return BR_EIO;
    }

    if (st.st_ino != bst.st_ino) /* inode changed, we expect a log rotation */
    {
        /* 2003/01/13, bdolez@ant-computing.com, patch for rotate */
        /* a log rotation is expected, check the file ${bname}.1 and
         * if such file exists try to extract additional information from it
         * (it is a common practice that some logging tools write a few records
         * into the old file before they notice a rotation). 
         */
        char        tmp[MAXPATHLEN + ROTATE_SUFFIX_LEN + 1];
        struct_stat rotated_st; 
        int         fd = -1;

        /* Build new filename ${bname}.1 */
        strncpy(tmp, bname, MAXPATHLEN); 
        tmp[MAXPATHLEN] = 0;
        strncat(tmp, ROTATE_SUFFIX, sizeof(ROTATE_SUFFIX)); 

        log (L_INFO , "Trying to open rotated file '%s'", tmp);

        /* 
         * try to reopen the file and if the inode has not changed and 
         * if there is something new to read use it as primary source
         */
        if ((fd = s_open (tmp, O_RDONLY)) >= 0 && !s_fstat(fd, &rotated_st) 
            && rotated_st.st_ino == bst.st_ino 
            && rotated_st.st_size > bst.st_size) 
        {
            /* close new log file and use old rotated log */
            s_close (bfd);
            bfd = fd;
            st = rotated_st;

            /* continue processing the original log file */
        }
        else  /* nothing to be read from the old rotated log */
        {
            /* close unused rotated file */
            if (fd >= 0)
                s_close(fd), fd = -1;

            /* get the new stat info and set position to zero */
            old_bst = bst;
            bst = st;
            bpos = 0;
            s_close (bfd), bfd = -1;

            /* report log rotation */
            return BR_ROTATED;
        }
    } 

    /* the same log file, we need to seek to proper place on reopen */
    if ((st.st_size < bst.st_size) || (s_seek (bfd, bpos) == -1))  
    {
        /* the file is smaller than it was before, this is suspicious -
         * someone is probably editing the log */
        old_bst = st;
        bst = st;
        bpos = 0;
        s_seek (bfd, bpos);
        return BR_SHORTENED;
    }
    else /* save current file size */
        bst.st_size = st.st_size;

    if (bst.st_size == bpos)    /* no new data */
    {
        s_close (bfd), bfd = -1;
        return BR_EMPTY;
    }

    /* seems okay, load next chunk */
    ssize_t bytes_to_read = min (CACHE_SIZE - (buf - cache_start), 
            st.st_size - bpos);
    ssize_t r = s_read (bfd, buf, bytes_to_read);
    debug_log ('2', "Disk read (%s): %li bytes", bname, r);

    if (!r || r > bytes_to_read)     /* nothing read or a nonsence */
        return BR_EIO;
    else if (r < bytes_to_read)
    {
        /* log shortened during the read or we suddently couldn't read it */
        old_bst = bst;
        bpos = 0;
        s_seek (bfd, bpos);

        /* rather be paranoid */
        return BR_SHORTENED;
    }
        
    /* okay, update cache delimiters */
    bpos += r;
    cache_end = buf + r - 1;

    /* if we read the whole file we should close it
       to go on with log rotator */
    if (bpos == st.st_size)
        s_close (bfd), bfd = -1;

    /* okay, data successfully fetched to cache */
    return BR_OK;
}


/** Skip all extra characters till a newline character is found */
inline tBufferState tFileBuffer::skip ()
{
    char *c = cache_pos;
    tBufferState sr;

    /* we expect the skipping successful */
    needs_skip = false;
    sr = BR_OK;

    /* skip everything till '\n' is found */
    do
    {
        while (c <= cache_end)
        {
            if (*c != '\n') /* sync on newline */
                goto found_newline;
            c++;
        }
        /* try next chunk */
        c = cache_start;
        sr = fetch(c);
    }
    while (sr == BR_OK);

    /* set the flag again if the file ends before we are able to sync */
    if (sr == BR_EMPTY)
        needs_skip = true;
    /* let the skipping disabled if buffer was rotated or shortened */
    
    found_newline:
    /* sync on the character after newline */
    cache_pos = c+1;

    /* return status */
    return sr;
}


/** Get a single line from file cache */
void tFileBuffer::get_line (tBufferResponse &r)
{
    static int bytes = 0;
    register char *c = cache_pos + bytes;

    /* re-sync on newline */
    if (needs_skip && ((r.state = skip ()) != BR_OK))
        return; /* something unusual happened */

    /* accumulate line characters */
    do
    {
        if (c > cache_end)      /* we ran out of cache chunk */
        {
            if (cache_pos != cache_start && bytes)   /* shift cache data */
                memmove (cache_start, cache_pos, bytes);
            /* load the rest of data */
            if ((r.state = fetch (cache_start + bytes)) != BR_OK)
                return;
            /* adjust cache markers */
            cache_pos = cache_start;
            c = cache_pos + bytes;
        }
        c++, bytes++;
    }
    while (*c != '\n' && bytes <= MAX_LOG_LINE_LENGTH);

    if (bytes <= MAX_LOG_LINE_LENGTH)
    {
        /* everything seemed okay */
        r.state = BR_OK;
        needs_skip = false;
    }
    else
    {   
        /* report garbage on input */
        r.state = BR_CORRUPTED;
        bytes = MAX_LOG_LINE_LENGTH;
        needs_skip = true;
    }

    /* set string portion to be returned */
    r.str = cache_pos;
    r.str_length = bytes;

    /* advance in-cache position */
    cache_pos = c+1, bytes = 0;
}
