/**
 * @file
 * @brief Echelog Miniagent -- getlog
 *
 * This agent periodically reads specified logs and passes information
 * stored there as echelog messages.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_MAGENT_GETLOG_FILEBUFFER_H
#define ECHELOG_MAGENT_GETLOG_FILEBUFFER_H

#include "lib/lib.h"
#include <sys/types.h>
#include <sys/stat.h>


#define CACHE_SIZE 32768
#define DEFAULT_ST_EXT ".st"
#define DEFAULT_ST_EXT_SIZE 3


/// Buffer state information
typedef enum {
    BR_OK = 0,           //< Buffer request properly finished
    BR_NEW,              //< Buffer snapshot
    BR_EMPTY,            //< Buffer is empty
    BR_ROTATED,          //< Possible buffer rotation
    BR_SHORTENED,        //< Buffer has been shortened
    BR_CORRUPTED,        //< Buffer of an unexpected format
    BR_EIO,              //< I/O error while manipulating buffer
} tBufferState;


/// Buffer response structure
struct tBufferResponse {
    tBufferState state; //< Buffer response type
    char *str;          //< Pointer to the beggining of a string/string portion
    long str_length;    //< Zero / string portion length
};


/// Single log file cache
class tFileBuffer {

    private:

        struct_stat bst;        //< Current log file stat information
        struct_stat old_bst;    //< Previous log file stat information
        int bfd;                //< Log file's descriptor
        long bpos;              //< Current log file position

        char cache_start[CACHE_SIZE];   //< File cache
        char *cache_pos;                //< Current in-cache position
        char *cache_end;                //< Actual end of file cache
        bool needs_skip;                /**< True when skipping extra 
                                         * characters on corrupted lines */
                                        
        char *st_filename;      //< State file's name
        bool report_new_file;   //< Flag for a new log file

        long bname_len;                         //< Log file name length
        SIMPLE_R_PROPERTY (char *, bname);      //< Log file name

    private:

        /** Atomic load of state variables from a buffer buf */
        int load_state_info (char *buf, long length);
        /** Fetch data from disk to getlog's cache */
        tBufferState fetch (char *buf);
        /** Skip all extra characters till a newline character is found */
        inline tBufferState skip ();

    public:
        /** Create an empty cache for a single file */
        tFileBuffer ();
        /** Flush and free all cache data */
        virtual ~tFileBuffer ();
        /** Create and initializes a new cache */
        tFileBuffer (const char *state_dir, const char *filename)
            { tFileBuffer (); register_file (state_dir, filename); }

        /** Register a log file and do the basic state initialization */
        void register_file (const char *state_dir, const char *filename);
        /** Open and load the cache state from previous session */
        void load_state ();
        /** Save the cache state of the current session to disk */
        int save_state ();
        /** Get a single line from file cache */
        void get_line (tBufferResponse &r);
};

#endif
