/**
 * @file
 * @brief Echelog Miniagent - getlog
 *
 * @module main getlog module
 *
 * This agent periodically reads specified logs and passes information
 * stored there as echelog messages.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */                                                                             

#include "getlog.h"

using namespace lib;
using namespace messages;
using namespace magents;


/* form and send messages */
bool tMagentGetlog::process_messages (bool& on_time)
{
    using namespace messages;

    bool portion_finished = false;
    tBufferResponse r;

    on_time = true;
    debug_log ('1', "Preparing messages");

    while (!portion_finished)
    {
        /* get another information chunk */
        buffer.get_line (r);

        switch (r.state)
        {
            case BR_OK:
                /* if the new chunk won't fit in current message send it */
                if (MAX_MESSAGE_SIZE <= (prepare_chunk (r) 
                                         + mp_mf->get_store_length())) 
                    send_last_message ();

                /* add a new message chunk */
                mp_mf->add_chunk (*chunk);
                break;

            case BR_NEW:
                /* send an empty snapshot for a new log */
                send_last_non_empty_message ();
                send_noninc_snapshot ();
                break;

            case BR_EMPTY:
                /* send the current message */
                send_last_message ();

                /* we're done for now */
                portion_finished = true;
                break;
                
            case BR_CORRUPTED:
                /* send what you can and report a corrupted log */
                send_alert (r.str);
                break;
                
            case BR_EIO:
            case BR_ROTATED:
            case BR_SHORTENED:
                /* send what you can and report a problem */
                send_alert (r.str);
                send_noninc_snapshot ();
                break;

            default:
                /* This should never happen */
                die ("Unknown response from i/o buffer");
        }

        /* alarm too early? */
        if (s_sigalrm_received)
        {
            /* and report that magent is overworked */
            send_last_non_empty_message ();
            mp_mf->send_alert ("GetLog overworked");

            on_time = false;
            /* return to schedule routine */
            return false;
        }
        if (sighup_received ())
                log(L_INFO, "Received SIGHUP, scheduling configuration reload");
        if (sigquit_received () ||
                sigint_received () ||
                sockerr_received () ||
                sigpipe_received ())
            {
                on_time = false;      /* not finished on time */
                break;
            }
    }
        
    send_last_non_empty_message ();    

    /* fine, finished on time */
    return false;
}


/// Name of the file, where the 's configuration is written
const char* const cfg_filename = DEFAULT_MODULE_CONFIG;


/* --===  PROGRAM ENTRY POINT ===-- */
int main (int argc, char *argv[])
{
    mframes::tModule::tCmdLine cmdline(argc, argv);
    tMagentGetlog getlog (cmdline, cfg_filename);

    /* init an agent */
    if (getlog.init ("glog"))
        return 1;

    int ret = getlog.run();
    getlog.cleanup ();

    return ret;
}
