/**
 * @file
 * @brief Echelog Miniagent -- netstat
 *
 * This agent periodically executes netstat and collected information 
 * passes as echelog messages.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_MAGENT_NETSTAT_NETSTATPARSER_H
#define ECHELOG_MAGENT_NETSTAT_NETSTATPARSER_H

#include "magents/runbuffer.h"
#include "magents/simplelex.h"
#include "magents/parser.h"
#include "messages/typelist.h"

#define MAX_PARSED_COLUMNS 7

/** @brief Chunk parser states */
typedef enum {
    P_NONE,     ///< Parsing has yet not started
    P_UDP4,     ///< Parsing udp4 info
    P_TCP4      ///< Parsing tcp4 info
} tPARSE_STATE;

/** @brief Column delimiters */
typedef struct {
    long start;     ///< The first column
    long end;       ///< The last column
} tColumn;

/** 
 * This class can be used for one-by-line parsing of 'netstat' output.
 * It follows the sematics of the generic tParser and
 * thus may be safely used with the tMagent class.
 */
class tNetstatParser : public magents::tParser {

    protected:
        lib::tMemPool pool;
        lib::tMemPool *parse_pool;      ///< mempool used for lexical analysis 
        lib::tList *scan_list;          ///< list for scan strings
        magents::tRunBuffer buffer;     ///< infobuffer instance for simplelex
        magents::tSimpleLex *lex;       ///< simple lexical analyzer
        messages::tMsgChunk *chunk;     ///< current info chunk
        tPARSE_STATE state;             ///< current chunk parser state
        bool recovering_from_error;     /**< true when recovering from 
                                         * a parse error */
                                        
        /// All NetstatMessage's attributes 
        struct {
            long msgid;
            messages::tVarchar *protocol;
            messages::tHost *local_address;
            messages::tHost *remote_address;
            messages::tInt32 *local_port;
            messages::tInt32 *remote_port;
            messages::tVarchar *state;
        } info;

        messages::tMsgChunk *error_chunk;       ///< error info chunk
        messages::tVarchar *error_info;         ///< error line

        tColumn cols[MAX_PARSED_COLUMNS];       ///< column delimiters
        bool report_empty_chunk;                //< non empty chunk parsed

    protected:

        /** @brief Parse address records */
        int parse_address (messages::tHost &address, messages::tInt32 &port,
                magents::tToken &t);

    public:
        /** @brief Construct a netstat output parser
         *
         * @param pool
         * Memory pool for internal parser structures 
         */
        tNetstatParser ();

        /** Destroy the chunk and release bound objects */
        virtual ~tNetstatParser ();

        /*
         * @brief TODOC:
         */
        virtual bool init ();

        /*
         * @brief TODOC:
         */
        virtual void cleanup ();

        /** @brief Init the parser
         *
         * This implements tParser::init() routine which is 
         * called by tMagent.
         */
        virtual int start ();

        /** @brief Stop the parser
         *
         * This implements tParser::stop() routine which is 
         * called by tMagent.
         */
        virtual void stop ();


        /** @brief Clean up the parser
         *
         * This implements tParser::cleanup() routine which is 
         * called by tMagent.
         */
        virtual void parse_cleanup ();


        /** @brief Get another info chunk
         *
         * This implements tParser::get_chunk() routine which is 
         * called by tMagent.
         */
        virtual void get_chunk (magents::tParserResponse &r);
        
        /** @brief Get a chunk containing error information
         *
         * This implements tParser::get_error_info() routine which is 
         * called by tMagent.
         */
        virtual void get_error_info (magents::tParserResponse &r);
        
        /** @brief Get miscellaneous information
         *
         * This implements tParser::get_misc_info() routine which is 
         * called by tMagent.
         */
        virtual void get_misc_info (magents::tParserResponse &r);
};

#endif

