/**
 * @file
 * @brief Echelog Miniagent -- netstat
 *
 * @module netstat output parser
 * 
 * This agent periodically executes netstat and collected information 
 * passes as echelog messages.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "netstatparser.h"
#include "messages/msgchunk.h"

using namespace lib;
using namespace magents;


#if defined (__LINUX__)
#    define NETSTAT "/bin/netstat"
#    define ADDR_PORT_SEPARATOR ':'
char * const netstat_args[] = { "netstat", "-antu", NULL };
#elif defined (__BSD__)
#    define NETSTAT "/usr/bin/netstat"
#    define ADDR_PORT_SEPARATOR '.'
char * const netstat_args[] = { "netstat", "-anf", "inet", NULL };
#endif

char * const netstat_envp[] = { NULL };


/* Init the parser */
int tNetstatParser::start ()
{
    using namespace lib;

    int r = buffer.init (NETSTAT, netstat_args, netstat_envp);

    if (!r)	/* If the buffer initialization went ok */
    {
        /* Skip over two header lines */
        if (lex->skip_to ('\n') == -1 || lex->skip_to ('\n') == -1)
            return -1;
        report_empty_chunk = true;
    }

    return r;
}


/* Get another info chunk */
void tNetstatParser::get_chunk (tParserResponse &r)
{
    using namespace lib;
    using namespace messages;

    tToken t;

    if (recovering_from_error)
    {
        /* Resync on the next line */
        lex->skip_to ('\n');
        recovering_from_error = false;
    }

    if (buffer.peek_byte () < 0)
    {
        /* Nothing to be parsed */
        if (report_empty_chunk)
        {
            r.chunk = chunk;
            r.state = PR_EMPTY;
            report_empty_chunk = false;
        }
        else
            r.state = PR_END;
        return;
    }

    debug_log ('1', "===========================");

#define GET_STRING() \
    do {\
        if (lex->get_token (t) < 0 || !t.len) \
        {\
            r.state = PR_ERROR;\
            recovering_from_error = true;\
            return;\
        }\
        debug_log ('1', "str='%s'", t.str);\
    }\
    while(0);


    /* The format of each line should be:
     * Protocol Recv-Q Send-Q  LocalAddress ForeignAddress [state]
     */

    /* Get protocol */
    GET_STRING();
    info.protocol->set_direct_value (t.str, t.len);

    /* Skip over Recv-Q and Send-Q */
    GET_STRING();
    GET_STRING();

    /* Parse local adress and port */
    GET_STRING();
    if (parse_address (*info.local_address, *info.local_port, t) < 0)
    {
        r.state = PR_ERROR;
        recovering_from_error = true;
        return;
    }

    /* Parse remote address and port */
    GET_STRING();
    if (parse_address (*info.remote_address, *info.remote_port, t) < 0)
    {
        r.state = PR_ERROR;
        recovering_from_error = true;
        return;
    }

    /* Get state (if any) */
    if (lex->get_to (t, '\n') < 0)
    {
        r.state = PR_ERROR;
        recovering_from_error = true;
        return;
    }
    char *state_str = s_trim (t.str, *parse_pool);
    debug_log ('1', "str='%s'", state_str);
    if (!t.len || !*state_str)
        info.state->set_null ();
    else
        info.state->set_direct_value (state_str);
 
    r.chunk = chunk;
    r.state = PR_OK;
    report_empty_chunk = false;
}

