/**
 * @file
 * @brief Echelog Miniagent -- netstat
 *
 * @module netstat output parser
 * 
 * This agent periodically executes netstat and collected information 
 * passes as echelog messages.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "netstatparser.h"
#include "messages/msgchunk.h"
#include <ctype.h>


/* Define the address and port separator character */
#if defined (__LINUX__)
#    define ADDR_PORT_SEPARATOR ':'
#elif defined (__BSD__) || defined (__SUNOS__)
#    define ADDR_PORT_SEPARATOR '.'
#endif

char * const netstat_envp[] = { NULL };
const unsigned long gbmp_size = 32768;

using namespace lib;
using namespace magents;


int tNetstatParser::parse_address (messages::tHost &address, 
        messages::tInt32 &port, tToken &t)
{
    char *addr_str = t.str, *port_str = &t.str[t.len - 1];

    /* Find the beginning of an address number */
    while (addr_str < port_str && isspace (*addr_str))
        addr_str++;

    /* Discard blank space at the end of a port number */
    while (port_str > addr_str && isspace (*port_str))
        *port_str = '\0', port_str--;

    /* Find the beginning of the port number */
    while (port_str > t.str && *port_str != ADDR_PORT_SEPARATOR)
        port_str--;
    if (port_str > t.str)
        *port_str++ = '\0';
    else
        return -1;
    debug_log ('1', "addr_str='%s', port_str='%s'", addr_str, port_str);

    scan_list->reset ();

    /* Fill in the address */
    if (!strcmp (addr_str, "*"))
        address.ip.set_value (0);
    else
    {
        scan_list->prepend (addr_str);
        if (!address.ip.sscan (*scan_list))
            return -1;
    }
    address.fqdn.set_null ();

    /* Fill in the port number */
    if (!strcmp (port_str, "*"))
        port.set_null ();
    else
    {
        long port_num;

        if (s_stol (port_str, &port_num, 0, 65535))
            return -1;
        port.set_value (port_num);
    }

    return 0;
}


/* Construct a netstat output parser */
tNetstatParser::tNetstatParser ()
    : pool (gbmp_size), state (P_NONE), recovering_from_error (false)
{ 
    error_chunk = NULL;
    chunk = NULL;
    lex = NULL;
    scan_list = NULL;
    parse_pool = NULL;
}


/* Destroy the chunk and release bound objects */
tNetstatParser::~tNetstatParser ()
{
}


bool tNetstatParser::init ()
{
    using namespace lib;
    using namespace messages;

    /* Create an internal pool for parsing purposes */
    parse_pool = new tMemPool (MAX_MESSAGE_SIZE);
    if (!parse_pool)
    {
        log (L_ERROR, "Not enough memory.");
        goto fail;
    }

    /* Create a scan list */
    scan_list = new tList (*parse_pool);
    if (!scan_list)
    {
        log (L_ERROR, "Not enough memory.");
        goto pool_clean;
    }

    /* Create a frame structures */
    lex = new tSimpleLex (buffer, *parse_pool);
    chunk = new tMsgChunk ("NetstatMessage", pool);
    error_chunk = new tMsgChunk ("SimpleLog", pool);
    error_info = static_cast<tVarchar *>(error_chunk->get_attr ("LINE"));
    if (!lex || !chunk || !error_chunk || !error_info)
    {
        log (L_ERROR, "Not enough memory.");
        goto mem_clean;
    }

    /* Bind all attributes to the message chunk */
#define BIND_ATTR(type, name) static_cast<type *>(chunk->get_attr(#name))
    info.protocol       = BIND_ATTR(tVarchar, PROTOCOL);
    info.local_address  = BIND_ATTR(tHost, LOCAL_ADDRESS);
    info.remote_address = BIND_ATTR(tHost, REMOTE_ADDRESS);
    info.local_port     = BIND_ATTR(tInt32, LOCAL_PORT);
    info.remote_port    = BIND_ATTR(tInt32, REMOTE_PORT);
    info.state          = BIND_ATTR(tVarchar, STATE);

    /* We will only add snapshot chunks */
    chunk->set_status (ST_ITEM);
    error_chunk->set_status (ST_ITEM);
    report_empty_chunk = true;

    return false;

mem_clean:
    delete error_chunk;
    error_chunk = NULL;
    delete chunk;
    chunk = NULL;
    delete lex;
    lex = NULL;
    delete scan_list;
    scan_list = NULL;
pool_clean:
    delete parse_pool;
    parse_pool = NULL;
fail:
    return true;

}

void tNetstatParser::cleanup ()
{
    buffer.cleanup ();
    delete chunk;
    chunk = NULL;
    delete error_chunk;
    error_chunk = NULL;
    delete lex;
    lex = NULL;
    delete scan_list;
    scan_list = NULL;
    delete parse_pool;
    parse_pool = NULL;
}

/* Get a chunk containing error information */
void tNetstatParser::get_error_info (tParserResponse &r)
{
    using namespace lib;
    using namespace messages;

    long length;
    char *str;

    /* Check if there is an error output */
    if ((length = s_filelen (buffer.get_temp_fd ())) <= 0)
    {
        r.state = PR_END;
        return;
    }

    /* Yes, get it */
    length = min (length, MAX_MESSAGE_SIZE - MAX_MESSAGE_SIZE / 4);
    str = static_cast<char *> (parse_pool->alloc (length));

    if ((s_seek (buffer.get_temp_fd(), 0) < 0)
        || (s_read (buffer.get_temp_fd (), str, length) < 0))
    {
        log (L_ERROR, "Can't get detailed information about parse error");
        r.state = PR_END;
        return;
    }

    /* Remove information about this error */
    if (ftruncate (buffer.get_temp_fd (), 0))
        die ("Can't truncate the temporary file");
        
    error_info->set_direct_value (str, length);
    r.chunk = error_chunk;
    r.state = PR_OK;
}


/* Get miscellaneous information */
void tNetstatParser::get_misc_info (tParserResponse &r)
{
    using namespace lib;
    using namespace messages;

    long length;

    /* Send program's stderr output (if any) */
    if ((length = s_filelen (buffer.get_temp_fd ())) >= 0)
    {
        get_error_info (r);
        return;
    }
        
    r.state = PR_END;
    return;
}


/* Clean up the parser */
void tNetstatParser::parse_cleanup ()
{
    parse_pool->free_pool ();
}


/* Stop the parser */
void tNetstatParser::stop ()
{
    buffer.cleanup ();
}
