/**
 * @file
 * @brief Echelog Miniagent -- ps
 *
 * @module ps output parser
 * 
 * This agent periodically executes ps and collected information 
 * passes as echelog messages.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "psparser.h"
#include "messages/msgchunk.h"
#include <sys/time.h>

#define PS "/bin/ps"

/* Define parameters for 'ps' to ensure the same output */
#if defined (__LINUX__)
char * const ps_args[] = { "ps", "-ewwo", 
                           "euser=,euid=,ruser=,ruid=,pid=,ppid=,pcpu=,vsz=,"\
                                "rss=,tty=,state=,stime=,args=", NULL };
#elif defined (__BSD__)
char * const ps_args[] = { "ps", "-awwo", 
                          "user=,uid=,ruser=,ruid=,pid=,ppid=,cpu=,vsz=,rss=,"\
                                "tty=,state=,start=,command=", NULL };
#elif defined (__SUNOS__)
char * const ps_args[] = { "ps", "-e", "-o", "user=", "-o", "uid=", 
                           "-o", "ruser=", "-o", "ruid=", "-o", "pid=", 
                           "-o", "ppid=", "-o", "pcpu=", "-o", "vsz=", 
                           "-o", "rss=", "-o", "tty=", "-o", "s=", 
                           "-o", "stime=", "-o", "args=", NULL };
#endif

char * const ps_envp[] = { NULL };

//                   3:04PM     22:54   Fri4AM   Jan 16   7Dec01   
char * st_fmt[] = { "%I:%M%p", "%H:%M", "%a%I%p", "%b%d", "%d%b%y",
                          NULL } ;

const unsigned long gbmp_size = 32768;

using namespace magents;

/* Construct a ps output parser */
tPsParser::tPsParser ()
    : pool (gbmp_size), recovering_from_error (false)
{
    error_chunk = NULL;
    chunk = NULL;
    lex = NULL;
    parse_pool = NULL;
}


/* Destroy the chunk and release bound objects */
tPsParser::~tPsParser ()
{
}

bool tPsParser::init ()
{
    using namespace lib;
    using namespace messages;

    /* Create an internal pool for parsing purposes */
    parse_pool = new tMemPool (MAX_MESSAGE_SIZE);
    if (!parse_pool)
    {
        log (L_ERROR, "Not enough memory.");
        goto fail;
    }

    /* Create a frame structures */
    lex = new tSimpleLex (buffer, *parse_pool);
    chunk = new tMsgChunk ("ProcessMessage", pool);
    error_chunk = new tMsgChunk ("SimpleLog", pool);
    error_info = static_cast<tVarchar *>(error_chunk->get_attr ("LINE"));
    if (!lex || !chunk || !error_chunk || !error_info)
    {
        log (L_ERROR, "Not enough memory.");
        goto mem_clean;
    }

    /* Bind all attributes to the message chunk */
#define BIND_ATTR(type, name) static_cast<type *>(chunk->get_attr(#name))
    info.eowner     = BIND_ATTR(tUser, EOWNER);
    info.owner      = BIND_ATTR(tUser, OWNER);
    info.pid        = BIND_ATTR(tInt32, PID);
    info.ppid       = BIND_ATTR(tInt32, PPID);
    info.cpu        = BIND_ATTR(tVarchar, CPU);
    info.memsize    = BIND_ATTR(tInt32, MEMSIZE);
    info.rss        = BIND_ATTR(tInt32, RSS);
    info.tty        = BIND_ATTR(tVarchar, TTY);
    info.state      = BIND_ATTR(tVarchar, STATE);
    info.start_time = BIND_ATTR(tUnixTime, STARTTIME);
    info.cmdline    = BIND_ATTR(tVarchar,CMDLINE);

    /* We will only add snapshot chunks */
    chunk->set_status (ST_ITEM);
    error_chunk->set_status (ST_ITEM);
    report_empty_chunk = true;

    return false;

mem_clean:
    delete error_chunk;
    error_chunk = NULL;
    delete chunk;
    chunk = NULL;
    delete lex;
    lex = NULL;
    delete parse_pool;
    parse_pool = NULL;
fail:
    return true;
}

void tPsParser::cleanup ()
{
    buffer.cleanup ();
    delete error_chunk;
    delete chunk;
    delete lex;
    delete parse_pool;
    error_chunk = NULL;
    chunk = NULL;
    lex = NULL;
    parse_pool = NULL;
}

/* Init the parser */
int tPsParser::start ()
{
    using namespace lib;

    int r = buffer.init (PS, ps_args, ps_envp);

    if (!r)	/* If the buffer initialization went ok */
    {
        int c;

	/* Skip the empty header line (if any) */
        lex->skip_blanks ();
        if ((c = buffer.peek_byte ()) == '\n')
            lex->skip_to('\n');
        report_empty_chunk = true;
    }

    return r;
}


/* Get another info chunk */
void tPsParser::get_chunk (tParserResponse &r)
{
    using namespace lib;
    using namespace messages;

    tToken t;
    struct tm st_tm;           // start_time
    struct timeval st_timeval; // start_time - use mempool?

#define GET_UINT(x) \
    do {\
        if (lex->get_token (t) < 0) \
        {\
            r.state = PR_ERROR;\
            recovering_from_error = true;\
            return;\
        }\
        debug_log ('1', "str='%s'", t.str);\
        if (t.type != T_NUMBER)\
        {\
            r.state = PR_ERROR;\
            recovering_from_error = true;\
            return;\
        }\
        if (t.sign == true)\
            (x)->set_null ();\
        else\
            (x)->set_value (t.num);\
    } while (0)

#define GET_STRING(x) \
    do {\
        if (lex->get_token (t) < 0) \
        {\
            r.state = PR_ERROR;\
            recovering_from_error = true;\
            return;\
        }\
        debug_log ('1', "str='%s'", t.str);\
        if (!t.len)\
            (x)->set_null();\
        else\
            (x)->set_direct_value (t.str, t.len);\
    } while(0);

    if (recovering_from_error)
    {
        /* Resync on the next line */
        lex->skip_to ('\n');
        recovering_from_error = false;
    }

    if (buffer.peek_byte () < 0)
    {
        /* Nothing to be parsed */
        if (report_empty_chunk)
        {
            r.chunk = chunk;
            r.state = PR_EMPTY;
            report_empty_chunk = false;
        }
        else
            r.state = PR_END;
        return;
    }

    debug_log ('1', "---------------------------");
    // Effective owner
    GET_STRING (&info.eowner->username);
    GET_UINT (&info.eowner->uid);

    // Real owner
    GET_STRING (&info.owner->username);
    GET_UINT (&info.owner->uid);

    // Simple columns
    GET_UINT (info.pid);
    GET_UINT (info.ppid);
    GET_STRING (info.cpu);
    GET_UINT (info.memsize);
    GET_UINT (info.rss);
    GET_STRING (info.tty);
    GET_STRING (info.state);

    // Get start time
    if (lex->get_token (t) < 0)
    {
        r.state = PR_ERROR;
        return;
    }
    // Get start time
    debug_log ('1', "Time: %s", t.str);
    reset_tm(&st_tm);
    if (t.len > 0 && (gettimeofday(&st_timeval,NULL) == 0) &&
        ( str2timef_rel(t.str, st_fmt, localtime((const time_t *)
                &st_timeval.tv_sec), -1, &st_tm, 0) != NULL))
        info.start_time->set_value(mktime(&st_tm)); 
    else
        info.start_time->set_null();

    // Command
    lex->skip_blanks ();
    if (lex->get_upto (t, '\n') < 0)
    {
        recovering_from_error = true;
        r.state = PR_ERROR;
        return;
    }
    if (!t.len)
        info.cmdline->set_null();
    else
        info.cmdline->set_direct_value (t.str, t.len);
    debug_log ('1', "Args: %s", t.str);


    // Eat up a newline character
    if (lex->get_token (t) >= 0 && t.type != T_NEWLINE)
    {
        r.state = PR_ERROR;
        recovering_from_error = true;
        return;
    }

    r.chunk = chunk;
    r.state = PR_OK;
    report_empty_chunk = false;

    // exclude the spawned 'ps' from the gathered data, skip to the next chunk
    if (!info.pid->is_null() && info.pid->get_value() == buffer.get_pid ())
        return get_chunk (r);
}


/* Get a chunk containing error information */
void tPsParser::get_error_info (tParserResponse &r)
{
    using namespace lib;
    using namespace messages;

    long length;
    char *str;

    /* Check if there is an error output */
    if ((length = s_filelen (buffer.get_temp_fd ())) <= 0)
    {
        r.state = PR_END;
        return;
    }

    /* Yes, get it */
    length = min (length, MAX_MESSAGE_SIZE - MAX_MESSAGE_SIZE / 4);
    str = static_cast<char *> (parse_pool->alloc (length));

    if ((s_seek (buffer.get_temp_fd(), 0) < 0)
        || (s_read (buffer.get_temp_fd (), str, length) < 0))
    {
        log (L_ERROR, "Can't get detailed information about parse error.");
        r.state = PR_END;
        return;
    }

    /* Remove information about this error */
    if (ftruncate (buffer.get_temp_fd (), 0))
        die ("Can't truncate the temporary file.");
        
    error_info->set_direct_value (str, length);
    r.chunk = error_chunk;
    r.state = PR_OK;
}


/* Get miscellaneous information */
void tPsParser::get_misc_info (tParserResponse &r)
{
    using namespace lib;
    using namespace messages;

    long length;

    /* Send program's stderr output (if any) */
    if ((length = s_filelen (buffer.get_temp_fd ())) >= 0)
    {
        get_error_info (r);
        return;
    }
        
    r.state = PR_END;
    return;
}


/* Clean up the parser */
void tPsParser::parse_cleanup ()
{
    parse_pool->free_pool ();
}


/* Stop the parser */
void tPsParser::stop ()
{
    buffer.cleanup ();
}
