/**
 * @file
 * @brief Echelog Miniagent -- ps
 *
 * This agent periodically executes ps and collected information 
 * passes as echelog messages.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_MAGENT_PS_PSPARSER_H
#define ECHELOG_MAGENT_PS_PSPARSER_H

#include "magents/runbuffer.h"
#include "magents/simplelex.h"
#include "magents/parser.h"
#include "messages/typelist.h"


/** 
 * This class can be used for one-by-line parsing of 'ps' output.
 * It follows the sematics of the generic tParser and
 * thus may be safely used with the tMagent class.
 */
class tPsParser : public magents::tParser {
    protected:
        lib::tMemPool pool;
        lib::tMemPool *parse_pool;      //< mempool used for lexical analysis 
        magents::tRunBuffer buffer;     //< infobuffer instance for simplelex
        magents::tSimpleLex *lex;       //< simple lexical analyzer
        messages::tMsgChunk *chunk;     //< current info chunk
        bool recovering_from_error;     /**< true when recovering from 
                                         * a parse error */
                                        
        /// All ProcessMessage's attributes 
        struct {
            long msgid;
            messages::tUser *owner, *eowner;
            messages::tInt32 *pid, *ppid, *memsize, *rss;
            messages::tVarchar *cpu, *tty, *state, *cmdline;
            messages::tUnixTime *start_time;
        } info;

        messages::tMsgChunk *error_chunk;       //< error info chunk
        messages::tVarchar *error_info;         //< error line
        bool report_empty_chunk;                //< non empty chunk parsed

    public:
        /** @brief Construct a ps output parser
         *
         * @param pool
         * Memory pool for internal parser structures 
         */
        tPsParser ();

        /** Destroy the chunk and release bound objects */
        virtual ~tPsParser ();

        /*
         * @brief TODOC:
         */
        virtual bool init ();

        /*
         * @brief TODOC:
         */
        virtual void cleanup ();

        /** @brief Init the parser
         *
         * This implements tParser::init() routine which is 
         * called by tMagent.
         */
        virtual int start ();

        /** @brief Stop the parser
         *
         * This implements tParser::stop() routine which is 
         * called by tMagent.
         */
        virtual void stop ();


        /** @brief Clean up the parser
         *
         * This implements tParser::cleanup() routine which is 
         * called by tMagent.
         */
        virtual void parse_cleanup ();


        /** @brief Get another info chunk
         *
         * This implements tParser::get_chunk() routine which is 
         * called by tMagent.
         */
        virtual void get_chunk (magents::tParserResponse &r);
        
        /** @brief Get a chunk containing error information
         *
         * This implements tParser::get_error_info() routine which is 
         * called by tMagent.
         */
        virtual void get_error_info (magents::tParserResponse &r);
        
        /** @brief Get miscellaneous information
         *
         * This implements tParser::get_misc_info() routine which is 
         * called by tMagent.
         */
        virtual void get_misc_info (magents::tParserResponse &r);
};

#endif

