/**
 * @file
 * @brief Echelog Miniagent -- utmp
 *
 * @module utmp output parser
 * 
 * This agent periodically collects information from utmp and passes
 * it as echelog messages.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/archdep.h"
#include <sys/types.h>
#include <errno.h>

#include "eutmpparser.h"
#include "messages/msgchunk.h"

#if defined (__SUNOS__)
#       include <utmpx.h>
#       define UT_NAMESIZE 32
#       define UT_HOSTSIZE 257
#       define UT_LINESIZE 32
#       define _utmp utmpx
#       define _getutent getutxent
#       define _setutent setutxent
#       define _endutent endutxent
#       undef HAVE_STUPID_UTMP
#else
#       include <utmp.h>
#       define _utmp utmp
#       define _getutent getutent
#       define _setutent setutent
#       define _endutent endutent
#       define UT_ADDRV6SIZE 4
#endif

#ifdef HAVE_STUPID_UTMP
#	define ut_user ut_name
#endif


using namespace lib;
using namespace magents;


static inline bool is_utype (struct _utmp *ut)
{
#ifdef HAVE_STUPID_UTMP
    return true;
#else
    if (ut->ut_type == USER_PROCESS)
        return true;

    return false;
#endif
}
const unsigned long gbmp_size = 32768;


/* Construct a ps output parser */
tUtmpParser::tUtmpParser () : pool (gbmp_size)
{
    parse_pool = NULL;
    chunk = NULL;
    error_chunk = NULL;
    error_info = NULL;
}

/* Destroy the chunk and release bound objects */
tUtmpParser::~tUtmpParser ()
{

}

bool tUtmpParser::init ()
{
    using namespace messages;

    /* Create an internal pool for parsing purposes */
    parse_pool = new tMemPool (MAX_MESSAGE_SIZE);
    if (!parse_pool)
    {
        log (L_ERROR, "Not enough memory.");
        goto fail;
    }

    /* Create a frame structures */
    chunk = new tMsgChunk ("SessionMessage", pool);
    error_chunk = new tMsgChunk ("SimpleLog", pool);
    error_info = static_cast<tVarchar *>(error_chunk->get_attr ("LINE"));
    if (!chunk || !error_chunk || !error_info)
    {
        log (L_ERROR, "Not enough memory.");
        goto mem_clean;
    }

#define BIND_ATTR(type, name) \
    static_cast<type *>(chunk->get_attr(#name))

    /* Bind all attributes to the message chunk */
    info.user       = BIND_ATTR(tUser, USER);
    info.host       = BIND_ATTR(tHost, HOST);
    info.tty        = BIND_ATTR(tVarchar, TTY);
    info.session    = BIND_ATTR(tInt32, SESSTYPE);
    info.idle       = BIND_ATTR(tInt32, IDLESEC);
    info.start_time = BIND_ATTR(tUnixTime, STARTTIME);

    /* We will only add snapshot chunks */
    chunk->set_status (ST_ITEM);
    error_chunk->set_status (ST_ITEM);
    report_empty_chunk = true;

    return false;

mem_clean:
    delete chunk;
    chunk = NULL;
    delete error_chunk;
    error_chunk = NULL;
    delete parse_pool;
    parse_pool = NULL;
fail:
    return true;
}

void tUtmpParser::cleanup ()
{
    delete error_chunk;
    error_chunk = NULL;
    delete chunk;
    chunk = NULL;
    delete parse_pool;
    parse_pool = NULL;
}

/* Init the parser */
int tUtmpParser::start ()
{
    /* Place the cursor at the beginning of utmp file */
    _setutent ();
    report_empty_chunk = true;

    return 0;
}


/* Get another info chunk */
void tUtmpParser::get_chunk (tParserResponse &r)
{
    using namespace lib;
    using namespace messages;

    struct _utmp *ut;
    uid_t user_uid;

    /* Skip up to usable records */
    do {
        ut = _getutent ();
    } while (ut && !is_utype (ut));

    debug_log ('1', "Reading the next chunk");
    
    if (ut)
    {
        report_empty_chunk = false;
        /* Get user's name and id */
        info.user->username.set_direct_value (
                s_strndup (ut->ut_user, UT_NAMESIZE, *parse_pool));
        if (ut->ut_user && !s_name2uid (ut->ut_user, user_uid))
            info.user->uid.set_value (user_uid);
        else
            info.user->uid.set_null ();

        /* Get user's terminal */
        info.tty->set_direct_value (
                s_strndup (ut->ut_line, UT_LINESIZE, *parse_pool));

        /* Get user's remote host */
        if (ut->ut_host && s_strnlen (ut->ut_host, UT_HOSTSIZE > 1))
            info.host->fqdn.set_direct_value (
                    s_strndup (ut->ut_host, UT_HOSTSIZE, *parse_pool));
        else
            info.host->set_null ();

#ifndef HAVE_STUPID_UTMP
        info.session->set_value (ut->ut_session);
        info.start_time->set_value (ut->ut_tv.tv_sec);
#endif

        r.chunk = chunk;
        r.state = PR_OK;
    }
    else if (report_empty_chunk)
    {
        r.chunk = chunk;
        r.state = PR_EMPTY;
        report_empty_chunk = false;
    }
    else
        r.state = PR_END;
}


/* Get a chunk containing error information */
void tUtmpParser::get_error_info (tParserResponse &r)
{
    r.state = PR_END;
}


/* Get miscellaneous information */
void tUtmpParser::get_misc_info (tParserResponse &r)
{
    r.state = PR_END;
}


/* Clean up the parser */
void tUtmpParser::parse_cleanup ()
{
    parse_pool->free_pool ();
}


/* Stop the parser */
void tUtmpParser::stop ()
{
    _endutent ();
}
