/**
 * @file
 * @brief Echelog Miniagent -- utmp
 *
 * This agent periodically collects information from utmp and passes 
 * it as echelog messages.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_MAGENT_UTMP_UTMPPARSER_H
#define ECHELOG_MAGENT_UTMP_UTMPPARSER_H

#include "magents/parser.h"
#include "messages/typelist.h"


/**
 * This class can be used for utmp information retrieval.
 * It follows the sematics of the generic tParser and
 * thus may be safely used with the tMagent class.
 */
class tUtmpParser : public magents::tParser {

    protected:
        lib::tMemPool pool;
        lib::tMemPool *parse_pool;       //< mempool used for lexical analysis 
        messages::tMsgChunk *chunk;      //< current info chunk

        /// All SessionMessage's attributes 
        struct {
            messages::tUser *user;
            messages::tHost *host;
            messages::tInt32 *idle, *session;
            messages::tVarchar *tty;
            messages::tUnixTime *start_time;
        } info;

        messages::tMsgChunk *error_chunk;       //< error info chunk
        messages::tVarchar *error_info;         //< error line
        bool report_empty_chunk;                //< non empty chunk parsed

    public:
        /** @brief Construct a w output parser
         *
         * @param pool
         * Memory pool for internal parser structures 
         */
        tUtmpParser ();

        /** Destroy the chunk and release bound objects */
        virtual ~tUtmpParser ();

        /*
         * @brief TODOC:
         */
        virtual bool init ();

        /*
         * @brief TODOC:
         */
        virtual void cleanup ();

        /** @brief Init the parser
         *
         * This implements tParser::init() routine which is 
         * called by tMagent.
         */
        virtual int start ();

        /** @brief Stop the parser
         *
         * This implements tParser::stop() routine which is 
         * called by tMagent.
         */
        virtual void stop ();


        /** @brief Clean up the parser
         *
         * This implements tParser::cleanup() routine which is 
         * called by tMagent.
         */
        virtual void parse_cleanup ();


        /** @brief Get another info chunk
         *
         * This implements tParser::get_chunk() routine which is 
         * called by tMagent.
         */
        virtual void get_chunk (magents::tParserResponse &r);
        
        /** @brief Get a chunk containing error information
         *
         * This implements tParser::get_error_info() routine which is 
         * called by tMagent.
         */
        virtual void get_error_info (magents::tParserResponse &r);
        
        /** @brief Get miscellaneous information
         *
         * This implements tParser::get_misc_info() routine which is 
         * called by tMagent.
         */
        virtual void get_misc_info (magents::tParserResponse &r);
};

#endif

