/**
 * @file
 * @brief Echelog Miniagent -- w
 *
 * @module w output parser
 * 
 * This agent periodically executes w and collected information 
 * passes as echelog messages.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#include "lib/archdep.h"
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <pwd.h>

#include "wparser.h"
#include "messages/msgchunk.h"

#define SYS_W "/usr/bin/w"


/* Define parameters for 'w' to ensure the same output */
#if defined (__LINUX__)
char * const w_args[] = { "w", "-f", NULL };
#elif defined (__BSD__)
char * const w_args[] = { "w", NULL };
#elif defined (__SUNOS__)
char * const w_args[] = { "w", NULL };
#endif

char * const w_envp[] = { NULL };
const unsigned long gbmp_size = 32768;


using namespace lib;
using namespace magents;

/* parsers' configuration structure */
long tWParser::s_send_uptime_records;
lib::cfitem tWParser::s_wmagent_section[]=
{
    {
        name: "WMagent",
        type: CT_SECTION,
        data: NULL
    },
    {
        name: "SendUptimeRecords",
        type: CT_INT,
        data: &tWParser::s_send_uptime_records,
        varf: NULL,
        varf_reset: NULL,
        flags: ITEM_REQUIRED
    },
    {
        name: NULL,
        type: CT_STOP,
        data: NULL
    }
};

/* Construct a w output parser */
tWParser::tWParser ()
   : pool (gbmp_size),
     cur_column (-1), first_run_after_init (true),
     recovering_from_error (false)
{
    error_chunk = NULL;
    up_chunk = NULL;
    chunk = NULL;
    lex = NULL;
    parse_pool = NULL;
}


/* Destroy the chunk and release bound objects */
tWParser::~tWParser ()
{
}

bool tWParser::init ()
{
    using namespace messages;

    /* Create an internal pool for parsing purposes */
    parse_pool = new tMemPool (MAX_MESSAGE_SIZE);
    if (!parse_pool)
    {
        log (L_ERROR, "Not enough memory.");
        goto fail;
    }

    /* Create a frame structures */
    lex = new tSimpleLex (buffer, *parse_pool);
    chunk = new tMsgChunk ("SessionMessage", pool);
    if (!lex || !chunk)
    {
        log (L_ERROR, "Not enough memory.");
        goto parse_clean;
    }
    debug_log ('2', "ID: %li", chunk->get_type ());
    up_chunk = new tMsgChunk ("UptimeMessage", pool);
    if (!up_chunk)
    {
        log (L_ERROR, "Not enough memory.");
        goto lex_clean;
    }
    debug_log ('2', "ID: %li", up_chunk->get_type ());
    error_chunk = new tMsgChunk ("SimpleLog", pool);
    if (!error_chunk)
    {
        log (L_ERROR, "Not enough memory.");
        goto upchunk_clean;
    }
    debug_log ('2', "ID: %li", error_chunk->get_type ());
    error_info = static_cast<tVarchar *>(error_chunk->get_attr ("LINE"));
    if (!error_info)
    {
        log (L_ERROR, "Not enough memory.");
        goto errchunk_clean;
    }

#define BIND_ATTR_COMMON(chunk, type, name) \
    static_cast<type *>(chunk->get_attr(#name))
#define BIND_ATTR(type, name) BIND_ATTR_COMMON(chunk, type, name)
#define BIND_ATTR_UP(type, name) BIND_ATTR_COMMON(up_chunk, type, name)

    /* Bind all attributes to the up_chunk */
    up_info.uptime  = BIND_ATTR_UP(tInt32, UPTIMESEC);
    up_info.users   = BIND_ATTR_UP(tInt16, USERS);
    up_info.load1   = BIND_ATTR_UP(tInt16, LOAD1);
    up_info.load5   = BIND_ATTR_UP(tInt16, LOAD5);
    up_info.load15  = BIND_ATTR_UP(tInt16, LOAD15);

    /* Bind all attributes to the message chunk */
    info.user       = BIND_ATTR(tUser, USER);
    info.host       = BIND_ATTR(tHost, HOST);
    info.tty        = BIND_ATTR(tVarchar, TTY);
    info.session    = BIND_ATTR(tInt32, SESSTYPE);
    info.idle       = BIND_ATTR(tInt32, IDLESEC);
    info.start_time = BIND_ATTR(tUnixTime, STARTTIME);

    /* We will only add snapshot chunks */
    up_chunk->set_status (ST_ITEM);
    chunk->set_status (ST_ITEM);
    error_chunk->set_status (ST_ITEM);
    report_empty_chunk = true;

    return false;

errchunk_clean:
    delete error_chunk;
    error_chunk = NULL;
upchunk_clean:
    delete up_chunk;
    up_chunk = NULL;
lex_clean:
    delete chunk;
    chunk = NULL;
    delete lex;
    lex = NULL;
parse_clean:
    delete parse_pool;
    parse_pool = NULL;
fail:
    return true;
}

void tWParser::cleanup ()
{
    buffer.cleanup ();
    delete error_chunk;
    delete up_chunk;
    delete chunk;
    delete lex;
    delete parse_pool;
    error_chunk = NULL;
    up_chunk = NULL;
    chunk = NULL;
    lex = NULL;
    parse_pool = NULL;
}

/* Convert a float value to a short (multiplied by 100) */
short tWParser::conv_to_short (const tToken &t)
{
    char *nstart, *nend, *end;
    short int ret = -1;
    double n;

    /* trim for number value checking */
    nstart = t.str;
    nend = t.str + t.len - 1;
    while (nstart < t.str + t.len 
            && *nstart == ' ' || *nstart == '\t' || *nstart == '\n') 
        nstart++;
    while (nend > t.str && *nend == ' ' || *nend == '\t' || *nend == '\n') 
        nend--;

    /* Try to read it as float */
    n = strtod (nstart, &end);
    if ((end == nend + 1) && (n != ULONG_MAX || errno != ERANGE))
        ret = static_cast<short int> (n * 100);

    return ret;
}


/* Fill info about the column named field_name */
int tWParser::get_field (const char *field_name, bool right_aligned)
{
    long scol, ecol;
    tToken t;

    /* Find next field */
    do {
        lex->skip_blanks ();
        scol = lex->get_column ();
        if (lex->get_token (t) < 0)
            return 1;
        ecol = lex->get_column ();
        col_format[++cur_column].redundant = true;
    } while (strcasecmp (t.str, field_name));

    /* OK, fill in the user column format description */
    col_format[cur_column].redundant = false;
    col_format[cur_column].right_aligned = right_aligned;
    if (right_aligned)
        col_format[cur_column].col = ecol;
    else
        col_format[cur_column].col = scol;

    debug_log ('1', "Field '%s' %s at column: %li", field_name, 
            right_aligned ? "right aligned" : "left aligned", 
            col_format[cur_column].col);

    return 0;
}


/* Skip over all redundant tokens */
inline void tWParser::skip_redundant ()
{
    tToken t;

    ASSERT (cur_column < MAX_COLUMNS);
    while (col_format[++cur_column].redundant)
        lex->get_token (t);

    lex->skip_blanks ();
}


/* Init the parser */
int tWParser::start ()
{
    tToken t;
    short days_up, n;
    int r;
    char *up_fmt[] = UPTIME_FMT;
    struct tm up_tm;
    
    /* Try to init the underlying buffer */
    if ((r = buffer.init (SYS_W, w_args, w_envp)) != 0)
        return r;

    /* Parse the uptime record of the following format (one line):
     * <time> up [<num_of_days> day(s),] <time_interval>, \
     *    <num> user,  load average: <float>, <float>, <float>
     */

    /* Skip to the first token after the date and the keyword 'up' */
    do 
    {
        if (lex->get_token (t) < 0)
            return 1;

    } while (strcmp (t.str, "up"));
    debug_log ('1', "==================================");

    /* Try to get the uptime record */
    if (!lex->get_token (t))
    {
        debug_log ('1', "str: '%s'", t.str);
        if (t.type == T_NUMBER)
        {
            tToken t2;

            /* Okay it's a number but it must not be negative */
            if (t.sign != false)
                return 1;

            /* Resolve the format from the next word before comma */
            lex->skip_blanks();
            if (lex->get_to (t2, ',') < 0 || t2.len < 1)
                return 1;
            t2.str[--t2.len] = 0;

	    debug_log ('1', "str: '%s'", t2.str);

            if (!strcasecmp (t2.str, "day") || !strcasecmp (t2.str, "days")
                    || !strcasecmp (t2.str, "day(s)"))
            {
                /* Fine, this should the number of uptime days */
                days_up = t.num;

                /* Get the time interval */
                lex->skip_blanks();
                if (lex->get_to (t, ',') < 0)
                    return 1;
                debug_log ('1', "str: '%s'", t.str);

                /* And cut off the trailing ',' */
                t.str[--t.len] = 0;
            }
            else
            {
                /* Uptime days record is missing */
                days_up = 0;

                t.str = s_strconcat (*parse_pool, t.str, t2.str, NULL);
                t.len += t2.len;

                debug_log ('1', "str: '%s'", t.str);
            }
        }
        else
        {
            /* The machine is up less than one day 
             * and the complete record is missing */
            days_up = 0;

            /* Umm, we have to cut off the trailing ',' */
            t.str[--t.len] = 0;
        }
    }
    else
        return 1;

    /* Okay we should have uptime in t */
    reset_tm(&up_tm);
    if((r=str2timef_p(t.str, up_fmt, &up_tm, 0)) < 0)
        r=0;

    /* Save the result to up_chunk */
    up_info.uptime->set_value (r + days_up * 86400);

    /* Get the number of users currently logged on */
    if ((lex->get_token (t) < 0) || t.type != T_NUMBER || t.sign != false)
        return 1;

    up_info.users->set_value (static_cast<short int> (t.num));

    /* Skip the kidding */
    if (lex->skip_to (':') < 0)
        return 1;

    /* Get load1 */
    if ((lex->get_upto (t, ',') < 0))
        return 1;
    debug_log ('1', "str: '%s'", t.str);
    if ((n = conv_to_short (t)) < 0)
        return 1;
    up_info.load1->set_value (static_cast<short int> (n));

    /* Get load5 */
    lex->skip_to (',');
    if ((lex->get_upto (t, ',') < 0))
        return 1;
    debug_log ('1', "str: '%s'", t.str);
    if ((n = conv_to_short (t)) < 0)
        return 1;
    up_info.load5->set_value (static_cast<short int> (n));

    /* Get load15 */
    lex->skip_to (',');
    if ((lex->get_token (t) < 0))
        return 1;
    debug_log ('1', "str: '%s'", t.str);
    if ((n = conv_to_short (t)) < 0)
        return 1;
    up_info.load15->set_value (static_cast<short int> (n));

    /* Get to the next line */
    lex->skip_to ('\n');
    first_run_after_init = true;

    /* Attempt to init a 'w' parser. Pray to god ;) 
     * At least the order of columns seems to be the same everywhere
     * and all extra columns consist of single tokens.
     * It should suffice to skip redundant columns...
     */

    cur_column = -1;
    if (get_field ("user"))
        return 1;
    if (get_field ("tty"))
        return 1;
    if (get_field ("login@", true))
        return 1;
    if (get_field ("idle", true))
        return 1;
    if (lex->skip_to ('\n'))
        return 1;
   
    report_empty_chunk = true;
    return 0;
}


/* Get another info chunk */
void tWParser::get_chunk (tParserResponse &r)
{
    using namespace lib;
    using namespace messages;

    tToken t;
    uid_t user_uid;
    int idle_res;
    long bcol, ecol;
    struct tm st_tm;           // start_time
    struct timeval st_timeval; // start_time - use mempool?
    char *login_fmt[] = LOGIN_FMT;
    char *idle_fmt[] = IDLE_FMT;

    debug_log ('1', "Reading the next chunk");
    
    if (s_send_uptime_records && first_run_after_init)
    {
        r.chunk = up_chunk;
        r.state = PR_OK;
        first_run_after_init = false;
        return;
    }
    if (recovering_from_error)
    {
        /* Resync on the next line */
        lex->skip_to ('\n');
        recovering_from_error = false;
    }

    if (buffer.peek_byte () < 0)
    {
        /* Nothing to be parsed */
        if (report_empty_chunk)
        {
            r.chunk = chunk;
            r.state = PR_EMPTY;
            report_empty_chunk = false;
        }
        else
            r.state = PR_END;
        return;
    }

    r.state = PR_ERROR;
    cur_column = -1;

    /* Get username and user's id (single token) */
    bcol = lex->get_column ();
    skip_redundant ();
    ecol = lex->get_column ();
    if (bcol <= col_format[cur_column].col && 
            ecol > col_format[cur_column].col)
        info.user->set_null ();
    else
    {
        if (lex->get_token (t))
        {
            recovering_from_error = true;
            return;
        }
        debug_log ('1', "str='%s'", t.str);
        if (t.len > 0)
        {
            info.user->username.set_direct_value (t.str);
            if (!s_name2uid (t.str, user_uid))
                info.user->uid.set_value (user_uid);
            else
                info.user->uid.set_null ();
        }
        else
            info.user->set_null ();
    }

    /* Get tty (single token) */
    bcol = lex->get_column ();
    skip_redundant ();
    ecol = lex->get_column ();
    if (bcol <= col_format[cur_column].col && 
            ecol > col_format[cur_column].col)
        info.tty->set_null ();
    else
    {
        if (lex->get_token (t))
        {
            recovering_from_error = true;
            return;
        }
        debug_log ('1', "str='%s'", t.str);
        if (t.len > 0 && strcmp (t.str, "?"))
            info.tty->set_direct_value (t.str);
        else
            info.tty->set_null ();
    }

    /* Get login time (right aligned) */
    bcol = lex->get_column ();
    skip_redundant ();
    ecol = lex->get_column ();
    if (bcol <= col_format[cur_column].col && 
            ecol > col_format[cur_column].col)
        info.start_time->set_null ();
    else 
    {
        if (lex->get_upto_col (t, col_format[cur_column].col))
        {
            recovering_from_error = true;
            return;
        }
        debug_log ('1', "str='%s'", t.str);
        reset_tm(&st_tm);
        if (t.len > 0 && (gettimeofday(&st_timeval,NULL) == 0) &&
                (str2timef_rel(t.str, login_fmt, localtime((const time_t *)
                 &st_timeval.tv_sec), -1, &st_tm, 0) != NULL))
            info.start_time->set_value(mktime(&st_tm));
        else
            info.start_time->set_null ();
    }
    
    /* Get idle */
    bcol = lex->get_column ();
    skip_redundant ();
    ecol = lex->get_column ();
    if (bcol <= col_format[cur_column].col && 
            ecol > col_format[cur_column].col)
        info.idle->set_null ();
    else
    {
        if (lex->get_upto_col (t, col_format[cur_column].col))
        {
            recovering_from_error = true;
            return;
        }
        debug_log ('1', "str='%s'", t.str);
        reset_tm(&st_tm);
        if (t.len > 0 && 
                ((idle_res=str2timef_p(t.str, idle_fmt, &st_tm, 0)) >= 0))
            info.idle->set_value(idle_res);
        else
            info.idle->set_null ();
    }

    if (lex->skip_to ('\n'))
    {
        recovering_from_error = true;
        return;
    }

    r.chunk = chunk;
    r.state = PR_OK;
    report_empty_chunk = false;
}


/* Get a chunk containing error information */
void tWParser::get_error_info (tParserResponse &r)
{
    using namespace lib;
    using namespace messages;

    long length;
    char *str;

    /* Check if there is an error output */
    if ((length = s_filelen (buffer.get_temp_fd ())) <= 0)
    {
        r.state = PR_END;
        return;
    }

    /* Yes, get it */
    length = min (length, MAX_MESSAGE_SIZE - MAX_MESSAGE_SIZE / 4);
    str = static_cast<char *> (parse_pool->alloc (length));

    if ((s_seek (buffer.get_temp_fd(), 0) < 0)
        || (s_read (buffer.get_temp_fd (), str, length) < 0))
    {
        log (L_ERROR, "Can't get detailed information about parse error");
        r.state = PR_END;
        return;
    }

    /* Remove information about this error */
    if (ftruncate (buffer.get_temp_fd (), 0))
        die ("Can't truncate the temporary file");
        
    error_info->set_direct_value (str, length);
    r.chunk = error_chunk;
    r.state = PR_OK;
}


/* Get miscellaneous information */
void tWParser::get_misc_info (tParserResponse &r)
{
    using namespace lib;
    using namespace messages;

    long length;

    /* Send program's stderr output (if any) */
    if ((length = s_filelen (buffer.get_temp_fd ())) >= 0)
    {
        get_error_info (r);
        return;
    }

    r.state = PR_END;
    return;
}


/* Clean up the parser */
void tWParser::parse_cleanup ()
{
    parse_pool->free_pool ();
}


/* Stop the parser */
void tWParser::stop ()
{
    buffer.cleanup ();
}

bool tWParser::register_sections (lib::tMemPool& trash_pool)
{
    if (tParser::register_sections (trash_pool))
        return true;

    cf_register(s_wmagent_section);
    return false;
}

void tWParser::unregister_sections ()
{
    cf_unregister(s_wmagent_section[0].name);
    tParser::unregister_sections ();
}
