/**
 * @file
 * @brief Echelog Miniagent -- w
 *
 * This agent periodically executes w and collected information 
 * passes as echelog messages.
 *
 * @author Kamil Toman <ktoman@email.cz>
 *
 */

#ifndef ECHELOG_MAGENT_W_WPARSER_H
#define ECHELOG_MAGENT_W_WPARSER_H

#include "magents/runbuffer.h"
#include "magents/simplelex.h"
#include "magents/parser.h"
#include "messages/typelist.h"

struct tColumn
{
    bool redundant;     //< true if we don't care about this column
    bool right_aligned; //< true if right aligned
    long col;           //< starting column / ending column (if right aligned)
};

/* The 'w' output should not have more than eight columns */
#define MAX_COLUMNS 8

                //   3:04pm     Fri 4pm    7Dec01
#define LOGIN_FMT { "%I:%M%p" , "%a%I%p", "%d%b%y", NULL }

#if defined (__LINUX__)
                //   3:04m     12:03    12.00s   100d
#define IDLE_FMT { "%H:%Mm" , "%M:%S", "%S.00s", "%dd", NULL } // TODO
#elif defined (__BSD__)
                //   3:04     12
#define IDLE_FMT { "%H:%M" , "%M", "%-", NULL } // TODO
#elif defined (__SUNOS__)
                //   3:04     12
#define IDLE_FMT { "%H:%M" , "%M", NULL } // TODO
#endif

//                    10:31   "17 min"
#define UPTIME_FMT { "%H:%M", "%Mm",    NULL } // TODO


/**
 * This class can be used for one-by-line parsing of 'w' output.
 * It follows the sematics of the generic tParser and
 * thus may be safely used with the tMagent class.
 */
class tWParser : public magents::tParser {
    protected:
        lib::tMemPool pool;
        lib::tMemPool *parse_pool;       //< mempool used for lexical analysis 
        magents::tRunBuffer buffer;      //< infobuffer instance for simplelex
        magents::tSimpleLex *lex;        //< simple lexical analyzer
        messages::tMsgChunk *chunk;      //< current info chunk
        messages::tMsgChunk *up_chunk;   //< uptime records message chunk
        tColumn col_format[MAX_COLUMNS]; //< description of each 'w' column
        long cur_column;                 //< current column
        bool first_run_after_init;       //< true for just initialized parser
        bool recovering_from_error;      /**< true when recovering from 
                                          * a parse error */

        /// All UptimeMessage's attributes
        struct {
            messages::tInt32 *uptime;
            messages::tInt16 *users, *load1, *load5, *load15;
        } up_info;
                                        
        /// All SessionMessage's attributes 
        struct {
            messages::tUser *user;
            messages::tHost *host;
            messages::tInt32 *idle, *session;
            messages::tVarchar *tty;
            messages::tUnixTime *start_time;
        } info;

        messages::tMsgChunk *error_chunk;       //< error info chunk
        messages::tVarchar *error_info;         //< error line
        bool report_empty_chunk;                //< non empty chunk parsed

        /// Convert a float value to a short (multiplied by 100)
        short conv_to_short (const magents::tToken &t);
        /// Fill info about the column named field_name
        int get_field (const char *field_name, bool right_aligned = false);
        /// Skip over all redundant tokens
        inline void skip_redundant ();
        
        /// true if the user want uptime information in the header to be sent
        static long s_send_uptime_records;

        static lib::cfitem s_wmagent_section[];

    public:
        /** @brief Construct a w output parser
         *
         * @param pool
         * Memory pool for internal parser structures 
         */
        tWParser ();

        /** Destroy the chunk and release bound objects */
        virtual ~tWParser ();

        /*
         * @brief TODOC:
         */
        virtual bool init ();

        /*
         * @brief TODOC:
         */
        virtual void cleanup ();

        /** @brief Init the parser
         *
         * This implements tParser::init() routine which is 
         * called by tMagent.
         */
        virtual int start ();

        /** @brief Stop the parser
         *
         * This implements tParser::stop() routine which is 
         * called by tMagent.
         */
        virtual void stop ();


        /** @brief Clean up the parser
         *
         * This implements tParser::cleanup() routine which is 
         * called by tMagent.
         */
        virtual void parse_cleanup ();


        /** @brief Get another info chunk
         *
         * This implements tParser::get_chunk() routine which is 
         * called by tMagent.
         */
        virtual void get_chunk (magents::tParserResponse &r);
        
        /** @brief Get a chunk containing error information
         *
         * This implements tParser::get_error_info() routine which is 
         * called by tMagent.
         */
        virtual void get_error_info (magents::tParserResponse &r);
        
        /** @brief Get miscellaneous information
         *
         * This implements tParser::get_misc_info() routine which is 
         * called by tMagent.
         */
        virtual void get_misc_info (magents::tParserResponse &r);

        bool register_sections (lib::tMemPool& trash_pool);
        void unregister_sections ();
};

#endif

