/**
 * @file
 * @brief Contains implementation of tMsgSniffPrinter class
 *
 * @author Kamil Toman <ktoman@email.cz>, Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "msgprinter.h"
#include <stdio.h>
#include "messages/msgmgr.h"

using namespace lib;
using namespace messages;

tMsgSniffPrinter::tMsgSniffPrinter () : m_chunk_mp (MAX_MESSAGE_SIZE),
    m_tmp_pool (4096), m_tmp_list (m_tmp_pool)
{
    m_logfile = NULL;
    m_inited = false;
}

tMsgSniffPrinter::~tMsgSniffPrinter ()
{
}

bool tMsgSniffPrinter::init (FILE* dest)
{
    m_logfile = dest;
    m_inited = true;

    return false;
}

void tMsgSniffPrinter::cleanup ()
{
    ASSERT (m_inited);
    m_logfile = NULL;
    m_inited = false;
}

bool tMsgSniffPrinter::print_message (tMsgBuf& msg)
{
    ASSERT (m_inited);

    tMsgChunk *chunk;
    int i, msg_type, msg_subrecords;
    bool status = false; // ok

    /* Print basic information */
    print_header (msg);

    /* Get message characteristics */
    msg_type = msg.get_header()->get_msg_type ();
    msg_subrecords = msg.get_header()->get_subrecords ();

    /* Create a new chunk */
    chunk = new tMsgChunk (msg_type, m_chunk_mp);
    if (!chunk)
    {
        log (L_ERROR, "Not enough memory.");
        goto fail;
    }

    /* Load one chunk at a time */
    for (i = 0; i < msg_subrecords; i++)
    {
        int bytes = chunk->get (msg.get_body ());
        if (bytes > 0)
            print_chunk (*chunk, i);      // print each chunk
        else if (!bytes)
            fprintf (m_logfile, "NULL message chunk\n");
        else
        {
            log (L_WARN, "Invalid chunk");
            /* print status */
            fprintf (m_logfile, "[Chunk #%i]:  INVALID CHUNK!\n", i);
            status = true;        // invalid chunk
        }

    }
    fprintf (m_logfile,
            "============================================================\n");

    /* Close the message */
    if (msg.close ())
    {
    	log (L_WARN, "The message reading failed");
        status = true;      // unknown error
    }

    /* Let the user see the message */
    fflush (m_logfile);

    /* Destroy what's left */
    delete chunk;
    m_chunk_mp.free_pool ();

    return status;

fail:
    return true;
}

/* Print header information */
void tMsgSniffPrinter::print_header (tMsgBuf &message)
{
    tMsgHeader *header = message.get_header ();
    char *hstr = header->get_host_str ();
    char *aptr = header->get_agent_id ();
    char *inst_aptr = header->get_agent_instance_id ();
    time_t tim = header->get_msg_time ();
    char buf[32];

    strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S",
            localtime (&tim));

    fprintf (m_logfile, "Header information:\n"
            "-------------------\n"
            "Message total length:\t%i\n"
            "Message header length:\t%i\n"
            "Message type:\t%i\n"
            "Message agent:\t%s\n"
            "Message agent instance:\t%s\n"
            "Message host:\t%s\n"
            "Message flags: %s%s\n"
            "Message time:\t%s\n"
            "Message subrecords #: %i\n"
            "Body length:\t%i\n"
            "------------------------------\n",
            header->get_total_msg_length (), header->get_len(),
            header->get_msg_type(), aptr, inst_aptr, hstr,
            header->get_flag_incremental()? "INCREMENTAL " : "",
            header->get_flag_aggregable()? "AGGREGABLE" : "",
            buf, header->get_subrecords(), message.get_max_body_len());

   xfree (hstr);
   xfree (aptr);
   xfree (inst_aptr);
}

/* print the list of printouts */
void tMsgSniffPrinter::printout (tList &list)
{
    char *string;
    long n;

    for (n = list.count(); n > 0; n--)
    {
        void *item = list.remove_first ();
        string = static_cast<char *> (item);

        if (!string)
            string = "(null)";

        fprintf (m_logfile, "%s", string);

        if (n > 1)
            fprintf (m_logfile, ", ");
    }

    fprintf (m_logfile, "\n");
}

/* convert tMSG_STATUS to string */
const char * tMsgSniffPrinter::str_status (tMSG_STATUS status)
{
    switch(status) {
        case ST_UNKNOWN:
            return "ST_UNKNOWN";
        case ST_NOCHANGE:
            return "ST_NOCHANGE";
        case ST_ADDED:
            return "ST_ADDED";
        case ST_REMOVED:
            return "ST_REMOVED";
        case ST_CHANGED:
            return "ST_CHANGED";
        case ST_ITEM:
            return "ST_ITEM";
        default:
            return "???";
    }
}

/* print chunk contents */
void tMsgSniffPrinter::print_chunk (tMsgChunk &chunk, int num)
{
     int i, num_of_attributes, message_handle;

     message_handle = gb_message_manager.get_handle_by_type (chunk.get_type());
     num_of_attributes = chunk.get_attr_count();

     /* print status */
     fprintf (m_logfile, "[Chunk #%i] status: %s, length: %i\n",
             num, str_status(chunk.get_status()), chunk.get_store_length ());
     
     /* and then each type in its default format */
     for (i = 0; i < num_of_attributes; i++)
     {
         char *name = gb_message_manager.get_attr_name (message_handle, i);

         fprintf (m_logfile, "%-12s: ", name);
         chunk.get_attr(i)->sprint (m_tmp_list, m_tmp_pool);
         printout (m_tmp_list);
     }

     m_tmp_pool.free_pool ();
}
