/**
 * @file
 * @brief Contains implementation of tMsgSniffTarget class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#include "msgsnifftarget.h"

using namespace lib;
using namespace messages;

/* Basic scondfilter initialization */
tMsgSniffTarget::tMsgSniffTarget ()
{
    m_log_name = NULL;
    m_routput_name = NULL;
    m_logfile = NULL;
    m_raw_logfd = -1;
    m_output = NULL;

    m_msgsniff_section = new cfitem [4];
    if (m_msgsniff_section == NULL)
    {
        log (L_ERROR, "Not enough memory.");
        m_msstate = MS_BROKEN;
        return;
    }

    int i = 0;
    m_msgsniff_section[i].name = "MsgSniff";
    m_msgsniff_section[i].type = CT_SECTION;
    m_msgsniff_section[i].data = NULL;
    i++;
    m_msgsniff_section[i].name = "RawLogFile";
    m_msgsniff_section[i].type = CT_STRING;
    m_msgsniff_section[i].data = NULL;
    i++;
    m_msgsniff_section[i].name = "LogFile";
    m_msgsniff_section[i].type = CT_STRING;
    m_msgsniff_section[i].data = NULL;
    m_msgsniff_section[i].varf = NULL;
    m_msgsniff_section[i].varf_reset = NULL;
    m_msgsniff_section[i].flags = ITEM_REQUIRED;
    i++;
    m_msgsniff_section[i].name = NULL;
    m_msgsniff_section[i].type = CT_STOP;
    m_msgsniff_section[i].data = NULL;

    m_msstate = MS_CREATED;
}

/* Final cleanup */
tMsgSniffTarget::~tMsgSniffTarget ()
{
    ASSERT (m_msstate == MS_CREATED ||
            m_msstate == MS_BROKEN);

    if (m_msstate == MS_CREATED)
    {
        delete [] m_msgsniff_section;
    }
}

/* Initialization of the filter */
bool tMsgSniffTarget::init ()
{
    ASSERT (m_msstate == MS_CONFIGURED);

    if (!strcmp (m_log_name, "stdout"))
        m_logfile = stdout;
    else
    {
        m_logfile = fdopen (
            sd_open (m_log_name,
                O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR),
            "w");
        if (!m_logfile)
        {
            log (L_ERROR, "Can't open file to log in.");
            goto fail;
        }
    }
    log (L_INFO, "Log file: %s", m_log_name);

    /* initialize printer */
    if (m_printer.init (m_logfile))
    {
        log (L_ERROR, "Initialization of message printer has failed.");
        goto log_close;
    }

    if (m_routput_name)
        m_raw_logfd = sd_open (m_routput_name,
                    O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR);

    /* save my output */
    m_output = mframes::tOFrame::get_actual_oframe ()->get_output_by_name
(NULL);
    m_msstate = MS_INITIALIZED;

    return false;

log_close:
    fclose (m_logfile);
    m_logfile = NULL;
fail:
    return true;
}

/* Reverse method for init */
void tMsgSniffTarget::cleanup ()
{
    ASSERT (m_msstate == MS_INITIALIZED);

    m_output = NULL;
    if (m_raw_logfd != -1)
        sd_close (m_raw_logfd);
    m_raw_logfd = -1;
    m_printer.cleanup ();
    if (m_logfile != stdout)
        fclose (m_logfile);
    m_logfile = NULL;

    m_msstate = MS_CLEANED;
}

bool tMsgSniffTarget::register_sections (lib::tMemPool& trash_pool)
{
    /* ensure the buffer is allocated */
    if (m_msstate == MS_BROKEN)
        return true;

    m_msgsniff_section[1].data = &m_routput_name;
    m_msgsniff_section[2].data = &m_log_name;
    cf_register (m_msgsniff_section);

    return false;
}

void tMsgSniffTarget::unregister_sections ()
{
    cf_unregister (m_msgsniff_section[0].name);
    m_msgsniff_section[1].data = NULL;
    m_msgsniff_section[2].data = NULL;
}

bool tMsgSniffTarget::process_config (lib::tMemPool& conf_mp, 
        lib::tMemPool& trash_pool)
{
    ASSERT (m_msstate == MS_CREATED);

    if (m_routput_name != NULL)
        m_routput_name = s_strdup (m_routput_name, conf_mp);
    m_log_name = s_strdup (m_log_name, conf_mp);

    m_msstate = MS_CONFIGURED;

    return false;
}

void tMsgSniffTarget::cleanup_config ()
{
    ASSERT (m_msstate == MS_CLEANED ||
            m_msstate == MS_CONFIGURED);

    m_log_name = NULL;
    m_routput_name = NULL;

    m_msstate = MS_CREATED;
}

/* Ensures message passing to proper outputs */
bool tMsgSniffTarget::process_msg (tMsgBuf& message)
{
    ASSERT (m_msstate == MS_INITIALIZED);

    bool ret = false;

    /* print message */
    if (m_printer.print_message (message))
    {
        log (L_ERROR, "Error in message printing.");
        ret = true;
    }

    /* dump raw data to disk if desired */
    if (m_raw_logfd >= 0)
    {
        if (s_write (m_raw_logfd,
                     message.get_buffer (), get_read_len ()) < 0)
        {
            log (L_ERROR, "Can't dump raw data to disk");
            ret = true;
        }
    }

    /* pass it if desired */
    if (m_output != NULL)
        if (m_output->send_message (message))
        {
            log (L_ERROR, "Can't pass the message.");
            ret = true;
        }

    return ret;
}


