/**
 * @file
 * @brief Contains declaration of tMsgSniffTarget class
 *
 * @author Marek Zoth <marek.zoth@volny.cz>
 *
 */

#ifndef ECHELOG_MSGSNIFFTARGET_H
#define ECHELOG_MSGSNIFFTARGET_H

#include "mframes/msgtarget.h"
#include "mframes/ioframe.h"
#include "msgprinter.h"

/**
 * @brief TODOC:
 */
class tMsgSniffTarget : public mframes::tMsgTarget
{
    public:
        //-------------------------------
        // Interface
        //-------------------------------

        /**
         * @brief Constructor
         *
         * Performs basic initialization of the instance.
         */
        tMsgSniffTarget ();

        /**
         * @brief Destructor
         *
         * Ensures final cleanup of the instance.
         */
        virtual ~tMsgSniffTarget ();

        /**
         * @brief Initialization of the filter
         *
         * This method prepares filter for use with pass_msg().
         * It reads configuration and creates and prepares rules.
         *
         * @return
         * If succesful returns false. On failure return true.
         */
        virtual bool init ();

        /**
         * @brief Cleanup of the filter
         *
         * Releases and resets all allocated resources in init().
         * It is the reverse function for init().
         */
        virtual void cleanup ();

        /**
         * @brief Message passing cycle implementation
         *
         * Message is passed to each rule in order they were
         * in configuration. Rule will filter the message and ensures
         * the sending. Rule can modify the message which is then passed
         * to following rules.
         *
         * @param message
         * Message to be filtered and sent to other modules. Must be prepared
         * for reading.
         *
         * @return
         * If the message was succesfully processed returns false.
         * If some error has occured returns true.
         */
        virtual bool process_msg (messages::tMsgBuf& message);

        //------------------------------------------
        // Implementation of tConfigurable interface
        //------------------------------------------

        /**
         * @brief TODOC:
         */
        virtual bool register_sections (lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void unregister_sections ();

        /**
         * @brief TODOC:
         */
        virtual bool process_config (lib::tMemPool& conf_mp, 
                lib::tMemPool& trash_pool);

        /**
         * @brief TODOC:
         */
        virtual void cleanup_config ();

    private:
        //-------------------------------
        // Internal types
        //-------------------------------

        /**
         * @brief Description of the msgsniff's state
         */
        enum MSState
        {
            /// initalization failed during constructor
            MS_BROKEN,
            /// successfully created
            MS_CREATED,
            /// successfully configured
            MS_CONFIGURED,
            /// successfully initialized
            MS_INITIALIZED,
            /// successfully cleaned
            MS_CLEANED
        };

        //-------------------------------
        // State attributes
        //-------------------------------

        /// State of the MsgSniff
        MSState m_msstate;

        //-------------------------------
        // Owned components
        //-------------------------------

        /// Description of msgsniff's configuration
        lib::cfitem* m_msgsniff_section;
        /// Name of the file to log to
        char* m_log_name;
        /// Raw data output file name
        char* m_routput_name;
        /// file to log to
        FILE * m_logfile;
        /// file descriptor to write raw data to
        int m_raw_logfd;
        /// Pointer to output where msgs are passed to
        mframes::tOFrame::tOutput* m_output;
        /// Printer used to print messages
        tMsgSniffPrinter m_printer;
};

#endif //ifdef ECHELOG_MSGSNIFFTARGET_H

