/**
 * @brief Echelog Receiver config file reading 
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include "messages/std_conf.h"
#include "receiver.h"

using namespace lib;
using namespace socks;

/* placeholders for read attributes */
char *pk_file, *cert_file, *root_file, *server_name, *dh_file;
char *state_dir, *output_name;
int r_port;

/* pool for parsing, freed after each parsing */
static class lib::tMemPool conf_mp(4096);
/* temporary list of parsed hosts to accept */
static class lib::tList accepted_hosts( conf_mp );

/* object taking care of qoutas and connections */ 
class tHostManager *host_mgr;
/* outcomming UD socket */
class tUNIXSocket *out_sock;
/* listenning socket for openssl connections */
class tSSLSocket *in_sock;
/* socket manager does all the nasty work with sockets */
class tSocketMgr *smgr;
/* settings for openssl connection */
class tSSLContext *ctx;

/* declaration of [Receiver] section fields */
static struct cfitem receiver_section[]={
    {name: "Receiver",          type: CT_SECTION,    data: NULL},
    {name: "PrivateKeyFile",    type: CT_STRING,     data: &pk_file,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "CertificateFile",   type: CT_STRING,     data: &cert_file, 
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "RootCertFile",      type: CT_STRING,     data: &root_file,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "DHParamsFile",      type: CT_STRING,     data: &dh_file,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "ReceiverHostname",  type: CT_STRING,     data: &server_name,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "RemotePort",        type: CT_INT,        data: &r_port,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "AcceptHost",        type: CT_LIST,       data: &accepted_hosts,
     varf: NULL, varf_reset: NULL },
    {name: "StateDir",          type: CT_STRING,     data: &state_dir,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: NULL,                type: CT_STOP,       data: NULL}
};

/* declaration of [System] section */
static struct cfitem system_section[]={
    {name: "System",            type: CT_SECTION,       data: NULL},
    {name: "Output",            type: CT_STRING,     data: &output_name,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: NULL,                type: CT_STOP,          data: NULL}
};


/* Set up of receiver after updating configuration */
void test_and_update_config (void)
{
    class tList out_list(conf_mp);
    
    /* create new object taking care of qoutas and connections */
    host_mgr = new tHostManager( state_dir );
    if (!host_mgr || host_mgr -> set_allowed_hosts ( accepted_hosts ) )
        die ("Unrecoverable errors in configuration, exitting");

    if (s_strpsep (out_list, output_name, conf_mp) || out_list.count ()!=2)
        die("Error in Output parameter, exiting.");

    out_list.remove_first ();
    output_name = static_cast<char *> (out_list.remove_first ()); 

    /* create new outgoing UD socket */
    out_sock = new tUNIXSocket;

    if( !out_sock || out_sock -> create_client( output_name ) )
        die("Can't connect to chained module via %s", output_name);

    out_sock->set_on_data_hup (unix_sock_hup_f);
    out_sock->set_on_data_err (unix_sock_err_f);
    out_sock->set_on_data_read (unix_sock_read_f);
  
    if ( smgr->register_socket( out_sock ))
    {
        die("Failed to register outgoing unix socket (%s).", output_name );
    }
 
    /* Release parsing data */
    get_cfpool().free_pool ();

    log (L_INFO, "Successfully configured");
}

/* Load module's configuration */
void load_config (int argc, char *argv[])
{
    using namespace messages;
    using namespace socks;

    gb_default_config_filename = DEFAULT_CONFIG_FILE;

    /* Load receiver configuration */
    cf_register(receiver_section);
    cf_register(system_section);
    /* Configure echelog types and messages */
    if (cf_read_std_configuration (argc, argv))
        die ("Unable to load configuration file '%s'",
             cf_get_config_filename ());

    log (L_INFO, "Using configuration file '%s'.", cf_get_config_filename ());

    /* create new instance holding openssl parameters */
    ctx = new tSSLContext;
    if ( !ctx || ctx->set_key_files_with_dh( pk_file, cert_file, 
                                             root_file, dh_file ))
    {
        die("Cannot load files for SSL handshake.");
    }

    /* create new soket manager instance */
    if ( !( smgr = new tSocketMgr ) )
    {
        die("Not enough memory for estabilishing socket manager.");
    }
     
    /* estabilish new listenning socket for openssl instances */
    in_sock = new tSSLSocket; 
    if (!in_sock)
        die ("Not enough memory.");

    in_sock->set_on_hup (ssl_sock_tcp_hup);
    in_sock->set_on_err (ssl_sock_tcp_err);
    in_sock->set_on_connected (ssl_sock_connected);
    in_sock->set_on_read_available (ssl_sock_read_available);
    in_sock->set_on_read_finished (ssl_sock_read_finished);
    in_sock->set_on_write_finished (ssl_sock_write_finished);
    in_sock->set_on_ssl_shutdown_finished (ssl_sock_shutdown_finished);
    
    if ( !in_sock || in_sock -> create_server(server_name,r_port,ctx,smgr))
    {
        die("Failed to estabilish server (%s:%d).", server_name, r_port);
    }    
    
    /* now set up things which change after reload */
    test_and_update_config ();

}

/* Reload getlog configuration part */
void reread_configuration ()
{
    using namespace messages;

    /* Announce this */
    log (L_INFO, "Reloading echelog configuration");

    /* release old object taking care of qoutas and connections */
    host_mgr -> write_all_states();
    delete host_mgr;

    /* release outgoing UD socket */
    if (smgr->unregister_socket (out_sock) < 0)
            log (L_ERROR, "Can't unregister socket (fd=%ld)", 
                    out_sock->get_fd());
    if (out_sock->close_socket())
        log (L_ERROR, "Error closing socket.");
    delete out_sock;
    
    /* Remove old configuration data - there should be none - but for the case*/
    conf_mp.free_pool ();

    log (L_INFO, "Using configuration file '%s'.", cf_get_config_filename ());

    /* Read new configuration */
    if (cf_reread_configuration (conf_mp))
         die ("Unable to load configuration file '%s'",
             cf_get_config_filename ());

    /* Reload new configuration */
    test_and_update_config ();
}

/* quick shutdown routine */
void bring_down()
{
    /* release all socket held in socket manager */
    smgr -> cleanup_created();
    smgr -> cleanup();
    
    /* remove listenning socket */
    if (in_sock -> close_socket())
        log (L_ERROR, "Error closing socket.");
    delete in_sock;

    /* remove outgoing socket */
    if (out_sock -> close_socket())
        log (L_ERROR, "Error closing socket.");
    delete out_sock;

    /* write all state info */
    host_mgr -> write_all_states();

    /* Exit */
    log (L_INFO, "Terminated with non-zero status.");
    log_close();
    exit(1);
}
