/**
 * @brief Echelog Receiver handlers for events on sockets 
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include "receiver.h"

using namespace socks;
using namespace lib;

/* remove client connection */
static void remove_socket(class tSocket *sock)
{
     /* get information about the socket */
     struct tRecAppData *data = static_cast<struct tRecAppData *>
                                ( sock -> get_app_specific_data() );
     
     /* write the state information */
     log (L_INFO,"Removing socket (%ld).", sock->get_fd());
     if (data && data->handle >= 0)
     {
         host_mgr -> write_state (data -> handle );
         host_mgr-> set_connected (data->handle, 0);
     }

     /* unregister, close and destroy the socket */
     sock->get_smgr ()->unregister_socket (sock);
     sock->close_socket ();
     delete sock;
}

/* event handler for error on outgoing UD socket - exits receiever */
void unix_sock_err_f(class tSocket *sock)
{
      log(L_WARN,"Outgoing connection broken; shutting down.");
      bring_down();
}

/* event handler for hangup on outgoing UD socket - exits receiever */
void unix_sock_hup_f(class tSocket *sock)
{
      log(L_WARN,"Outgoing connection closed; shutting down.");
      bring_down();
}

/* event handler for incomming data on outgoing UD socket - exits receiever */
void unix_sock_read_f(class tSocket *sock)
{
      log(L_WARN,"Unexpected data on outgoing connection: shutting down.");
      bring_down();
}

/* event handler for error on incomming connection */
void ssl_sock_tcp_err(class tSocketNB *sock)
{
      /* listenning socket error */
      if (sock->get_role() == AS_SOCK_SERVER)
      {
          log(L_WARN,"Listenning socket encountered error; shutting down.");
          bring_down();
      } 
      else 
      /* error on incomming connection */
      {
           log(L_WARN,"Socket (%ld) error: Releasing socket.", sock->get_fd());
           remove_socket (sock);
      }    
}

/* event handler for hangup on incomming connection */
void ssl_sock_tcp_hup(class tSocketNB *sock)
{
      /* listenning socket hangup - it cant happen - but who knows */
      if (sock->get_role() == AS_SOCK_SERVER)
      {
          log(L_WARN,"Listenning socket encountered error; shutting down.");
          bring_down();
      }
      else
      /* hangup on incomming connection */
      {
          log(L_WARN,"Other side ended communication (%ld). Releasing socket.",
                     sock->get_fd() );
          remove_socket (sock);
      }
}

/* connected succesfully */
void ssl_sock_connected(class tSocketNB *sock)
{
      class tSSLSocket *s = static_cast<class tSSLSocket *>(sock);
      struct tRecAppData* app_data;
      struct tHostInfo *h;
      char *hostname;
      int handle;
      enum tQuitStatus quit = QS_OK;

      ASSERT ( sock -> get_last_operation() == OP_NO_OPERATION );

      /* lets dig out handle of the host */
      hostname = sock->get_address();
      if ( (handle = host_mgr -> get_host_handle ( hostname )) == -1 )
      { 
           log( L_WARN, "Host %s wants to connect but is not known.",
                         hostname );
           quit = QS_UNKNOWN;
      }

      /* check if it is not connected twice */
      if ( quit == QS_OK && host_mgr -> get_connected ( handle ) )
      {
           log( L_WARN, "Host %s wants to connect but is already connected.",
                         hostname );
           quit = QS_ALREADY_CONNECTED;
      }

      /* read this host state information */
      if ( quit == QS_OK && host_mgr -> read_state ( handle ) )
      {
           log( L_WARN, "Host %s wants to connect: there were error reading "
                        " its state.", hostname );
           remove_socket (sock);
           return;
      }
      
      /* create socket identification data */
      app_data = static_cast<struct tRecAppData *>
                  (xmalloc(sizeof(struct tRecAppData)));

      if ( !app_data )
      {
           log( L_WARN, "Host %s wants to connect: not enough memory - "
                        "refusing. ", hostname );
           remove_socket (sock);
           return;
      }

      sock -> set_app_specific_data( app_data );

      /* check the host identity via common name (if requested) */ 
      if (quit == QS_OK)
      {
          host_mgr -> set_connected ( handle, 1 );

          h = host_mgr -> get_host_info (handle);

          /* process certificate common name verify check */
          if ( h->common_name && s->verify_common_name(h->common_name)!= 0)
          {
               log( L_WARN, "Host %s wants to connect: common name doest not "
                            "match - refusing. ", hostname );
               quit = QS_INV_CERT;
          } else
          /* process quota checks */
          if ( h->period > 0 )
          {
               if (time(NULL) - h->start_time > h->period )
               {
                   h->start_time = time(NULL);
                   h->flow = 0;
               } else 
               if ( h->flow > h->limit ) quit = QS_QUOTA_EXC;
             
               host_mgr -> write_state ( handle );
          }
      }

      /* host is ok ... send him invitation */
      if (quit==QS_OK)
      {
          app_data -> handle = handle;
          app_data -> state_info.set_state( CS_START );
          app_data -> state_info.set_arg ( host_mgr -> get_state( handle ) );  
          app_data -> phase = RS_EMPTY;
      } else
      /* STRANGER ! WE DO NOT LIKE STRANGERS ! BEGGARS ... ERRM */
      {
          app_data -> handle = handle;
          app_data -> state_info.set_state( CS_QUIT );
          app_data -> state_info.set_arg ( quit );
          app_data -> phase = RS_QUIT;
      }

      /* send the status message */
      app_data -> state_info.store_message();
      sock -> at_write ( app_data -> state_info.get_msg_buf(),
                         STATE_MESSAGE_SIZE);

}

/* there are incomming data on the ssl-socket */
void ssl_sock_read_available(class tSocketNB *sock)
{
      /* get this socket data */
      struct tRecAppData *data = static_cast<struct tRecAppData *>
                                 ( sock -> get_app_specific_data() );

      ASSERT ( sock -> get_last_operation() == OP_NO_OPERATION );
      /* in which phase we are */
      switch ( data -> phase )
      {
          case RS_NEGOTIATE:
              sock -> at_read ( data -> state_info.get_msg_buf(), 
                                STATE_MESSAGE_SIZE );
              break;
          case RS_RECEIVE:
              sock -> at_read ( data -> data_buffer, 
                                sizeof( data -> data_buffer ));
              break;
          default:
              remove_socket (sock);
      }
 
}

/* read operation finished on incomming socket */
void ssl_sock_read_finished(class tSocketNB *sock)
{
      /* get this socket data */
      struct tRecAppData *data = static_cast<struct tRecAppData *>
                                 ( sock -> get_app_specific_data() );
      struct tHostInfo *h;
      long len;

      /* in which phase we are */
      switch ( data -> phase )
      {
          case RS_NEGOTIATE:
              data -> state_info.load_message();

              /* we await request status message */
              if ( sock -> get_r_io().last_op_bytes !=  STATE_MESSAGE_SIZE ||
                   data -> state_info.get_state() != CS_REQ)
              {
                  log(L_INFO, "Connection closed with (%s): violated protocol "
                              "- init handshake.", sock -> get_address());
                  remove_socket (sock);
                  break;
              } 

              /* set as requested and store the information */
              host_mgr->set_state (data->handle, data->state_info.get_arg ());
              if (host_mgr->write_state (data->handle))        
              {
                  log(L_INFO, "Connection closed with (%s): cannot "
                              " save state info.", sock -> get_address());
                  remove_socket (sock);
                  break;
              }

              /* enter phase where we will receive message */
              data -> phase = RS_RECEIVE;

              h = host_mgr->get_host_info(data->handle);

              log (L_INFO, "Connection with \"%s\" resumed at %ld.",
                        h->hostname, h->received );
              
              break;
          case RS_RECEIVE:

              /* check if we got the message */
              len = sock -> get_r_io().last_op_bytes;
              if ( len < 0 )
              {
                  log(L_INFO, "Connection closed with (%s): bad data sent.",
                              sock -> get_address());
                  remove_socket (sock);
                  break;
              }

              /* pass the received message to outgoing socket */
              if (out_sock->at_write (data->data_buffer, len)!= len)
                  bring_down();

              debug_log ( '0', "Received this item: %ld", 
                               host_mgr->get_state(data->handle));

              
              /* message received, go to next */
              data->state_info.set_state( CS_ACCEPTED );
              data->state_info.set_arg (host_mgr->get_state(data->handle));
              host_mgr->set_state (data->handle,
                                   host_mgr->get_state (data->handle)+1);
              data->phase = RS_COMMIT;

              /* perform quota check */
              h = host_mgr->get_host_info(data->handle);
              if (h->period > 0)
              {
                 h->flow += sock->get_r_io().last_op_bytes;
                 if (time(NULL) - h->start_time > h->period )
                 {
                     h->start_time = time(NULL);
                     h->flow = sock->get_r_io().last_op_bytes;
                 } else
                 if ( h->flow > h->limit )
                 {    
                     data->state_info.set_state( CS_QUIT );
                     data->state_info.set_arg( QS_QUOTA_EXC );
                     data->phase = RS_QUIT;
                 }
              }
              
              /* store curreny state */
              host_mgr->write_state ( data->handle );
              data -> state_info.store_message();

              /* send state message */
              sock->at_write (data->state_info.get_msg_buf(),
                              STATE_MESSAGE_SIZE );
              
              break;
          default:
              /* communication error */
              remove_socket (sock);
      }

}

/* write finished on connected ssl-socket */
void ssl_sock_write_finished(class tSocketNB *sock)
{
      /* get this socket info data */
      struct tRecAppData *data = static_cast<struct tRecAppData *>
                                 ( sock -> get_app_specific_data() );

      /* in which phase we are ? */
      switch ( data -> phase )
      {    
          /* transfer info sent, await reply */
          case RS_EMPTY:
              data -> phase = RS_NEGOTIATE;
              if ( sock -> get_w_io().last_op_bytes != STATE_MESSAGE_SIZE )
              {
                  log(L_INFO, "Connection closed with (%s): Failed to write "
                              "state message.", sock -> get_address());
                  remove_socket (sock);
                  break;
              }
              break;
          /* receipt commited, enter receiving phase */
          case RS_COMMIT:
              data -> phase = RS_RECEIVE;
              if ( sock -> get_w_io().last_op_bytes != STATE_MESSAGE_SIZE )
              {
                  log(L_INFO, "Connection closed with (%s): Failed to write "
                              "state message.", sock -> get_address());
                  remove_socket (sock);
                  break;
              }
              break;
          /* after sending quit command or in any other state release socket */
          case RS_QUIT:
          default:
              remove_socket (sock);
      }

          
}

/* communication ended by the otherside */
void ssl_sock_shutdown_finished(class tSSLSocket *sock)
{
      log(L_INFO, "Comunication ended with other side.");
      remove_socket (sock);
}
