/**
 * @brief Echelog Receiver support for storing state of transfer and quotas 
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include "messages/testbuf.h"
#include "receiver.h"

#include <string.h>
#include <errno.h>
#include <unistd.h>

using namespace lib;

/* variable used for parameter list processing in walk_f() */
static int walk_index;
/* array where all information about hosts is stored */
static struct tHostInfo *a_h;
/* memory pool for allocation made by hostmanager */
static class tMemPool *w_mp;
/* directory where the state files are stored */
static char *path;

/* this function parses one AcceptHost clause from config file */
static int walk_f(void *a)
{
    class tList paramlist(*w_mp);
    char *host, *limit, *period, *cn, *filename;
    long period_l, limit_l;

    /* split the string to parts */
    if (s_strpsep (paramlist, static_cast<char *>(a), *w_mp))
       die ("Error in config file: AcceptHost clause number %d.", walk_index+1);

    /* all parameters are initially not known */
    host = limit = period = cn = NULL;

    /* for every attribute store its value to appropriate variable*/
    while (paramlist.count () > 0)
    {
        char *attr_name = static_cast<char *> (paramlist.remove_first ());
        char *value;

        if (paramlist.count () > 0)
            value = static_cast<char *> (paramlist.remove_first ());
        else
            die ("Error in config file: AcceptHost clause number %d.", 
                    walk_index+1); 

        if (!strcmp ("Hostname", attr_name))
            host = value;
        else if (!strcmp ("CommonName", attr_name))
            cn = value;
        else if (!strcmp ("Limit", attr_name))
            limit = value;
        else if (!strcmp ("Period", attr_name))
            period = value;
        else
            die ("Error in config file: AcceptHost clause number %d: Unknown "
                 "attribute %s.", walk_index+1, attr_name);
    }

    /* host must be defined */
    if (!host)
            die ("Error in config file: AcceptHost clause number %d: Missing "
                 "Hostname attribute.", walk_index+1);

    /* try to convert limit to number */
    if (!limit) limit_l = 0; else limit_l = max (0, atoi(limit));

    /* try to convert period to number */
    if (!period) period_l = 0; else period_l = max (0, atoi(period));
            
    /* create state file filename */
    filename = static_cast<char *>
        (w_mp->alloc (strlen (host) + strlen (path) + 2));
    memcpy (filename, path, strlen (path));
    filename [strlen (path)] = '/';
    memcpy (filename + strlen (path) + 1, host, strlen (host) + 1);

    /* store the information, we had parsed */
    a_h[walk_index].hostname = host;
    a_h[walk_index].filename = filename;
    a_h[walk_index].common_name = cn;
    a_h[walk_index].connected = 0;
    a_h[walk_index].period = period_l;
    a_h[walk_index++].limit = limit_l;

    return 0;
}

/* constructor for hostmanager - prepares the instance */
tHostManager::tHostManager ( char *st_dir )
{
     int len = strlen( st_dir ) + 1;

     /* create new mempool for hostmanager allocations */
     mp = new tMemPool( 2048 );
     if (!mp)
        die ("Not enough memory.");

     /* store its own copy of directory name, where state information is */
     state_dir = static_cast<char *>( mp->alloc( len ) );
     memcpy ( state_dir, st_dir, len );
}

tHostManager::~tHostManager()
{
     /* free the pool and destroy it */
     mp->free_pool();
     delete mp;
}

/* checks if file with state info is on his place, otherwise create it */
int tHostManager::check_file ( int handle )
{

     int fd = s_open (allowed_hosts[handle].filename, O_RDONLY, 0600);

     /* file not exists */
     if ( fd < 0 && ( errno == EACCES || errno == ENOENT ) )
     {
         /* lets fill in default state and create the file */
         allowed_hosts[handle].received = 0;
         allowed_hosts[handle].start_time = time(NULL);
         allowed_hosts[handle].flow = 0;
         return write_state( handle );
     } else
     /* error opening file */
     if( fd < 0 )
     {
         log ( L_INFO, "Cannot access file %s (reading state): %m",
                        allowed_hosts[handle].filename );
         return -1;
     } else
     /* file is on its place, will only read the information */
     {
         s_close( fd );
         
         /* if we cant read the file, fill in default state and create file */
         if ( read_state( handle ) )
         {
              allowed_hosts[handle].received = 0;
              allowed_hosts[handle].start_time = time(NULL);
              allowed_hosts[handle].flow = 0;
              return write_state( handle );
         } else return 0;
     }

}

/* this function takes list of AcceptHost clauses, parses it and stores
 * the inforamtion about accepted host into hostmanager instance */
int tHostManager::set_allowed_hosts( class lib::tList &hl )
{
     int i;
  
     /* determine number of clauses and allocate appropriate array */
     allowed_hosts_cnt = hl.count();
     allowed_hosts = static_cast<struct tHostInfo*>
         ( mp->alloc( sizeof (struct tHostInfo) * allowed_hosts_cnt ));
     memset( allowed_hosts, 0, sizeof (struct tHostInfo) * allowed_hosts_cnt );
     
     /* fill additional parameters for walk_f() */
     w_mp = mp;
     walk_index = 0;
     a_h = allowed_hosts;
     path = state_dir;
     
     /* process all the clauses */
     hl.walk(walk_f);

     /* load all state info */
     for (i=0;i<allowed_hosts_cnt;i++)
         if (check_file(i)) break;
         
     /* did we loaded all state files correctly */
     if (i<allowed_hosts_cnt)
         return -1; else return 0;
}

/* find handle for host according to hostname */
int tHostManager::get_host_handle ( char *hostname )
{
     int i;
  
     /* we simply go through array until match */
     for (i=0;i<allowed_hosts_cnt;i++)
         if (!strcasecmp( hostname, allowed_hosts[i].hostname) ) break;
 
     /* did find the host ? */
     if ( i == allowed_hosts_cnt ) 
         return -1;
     else
         return i;
}

/* read from file state to internal hostmanger structures */
int tHostManager::read_state ( int handle )
{
     char buffer[REC_STATE_BUFF_SIZE];
     using messages::tPlainBuf;
     tPlainBuf b(buffer, REC_STATE_BUFF_SIZE);
     int32_t state, s_time, received;
  
     int fd = s_open (allowed_hosts[handle].filename, O_RDONLY, 0600); 

     if ( fd < 0 ) 
     {
         log ( L_ERROR, "Cannot open file %s for reading state: %m",
                       allowed_hosts[handle].filename );
         return -1;
     } 
     
     if( s_read(fd, buffer, REC_STATE_BUFF_SIZE) != REC_STATE_BUFF_SIZE )
     {
         log ( L_ERROR, "Cannot read file %s (reading state): %m",
                        allowed_hosts[handle].filename );
         return -1;
     }

     if (s_close(fd))
     {
         log ( L_ERROR, "Cannot close file %s (reading state): %m",
                        allowed_hosts[handle].filename );
         return -1;
     }

     READ_INT32 ( &b, state );
     READ_INT32 ( &b, s_time );
     READ_INT32 ( &b, received );

     if ( s_time > time(NULL) )
     {
         log (L_ERROR, "Clock skew detected in stored time for %s.",
                       allowed_hosts[handle].filename );
     }

     allowed_hosts[handle].flow = received;
     allowed_hosts[handle].start_time = s_time;
     allowed_hosts[handle].received = (spool::tIdentificator)state;

     return 0;
}

/* try to store state information */
int tHostManager::write_state ( int handle )
{
     char buffer[REC_STATE_BUFF_SIZE];
     using messages::tPlainBuf;
     tPlainBuf b(buffer, REC_STATE_BUFF_SIZE);
     int32_t state;
  
     int fd = s_open (allowed_hosts[handle].filename, 
                      O_WRONLY | O_CREAT | O_TRUNC, 0600); 

     if ( fd < 0 ) 
     {
         log ( L_INFO, "Cannot open file %s for writing state: %m",
                       allowed_hosts[handle].filename );
         return -1;
     } 

     state = (long) allowed_hosts[handle].received;
     STORE_INT32 ( &b, state );
     STORE_INT32 ( &b, allowed_hosts[handle].start_time );
     STORE_INT32 ( &b, allowed_hosts[handle].flow );
     
     if( s_write(fd, buffer, REC_STATE_BUFF_SIZE) != REC_STATE_BUFF_SIZE )
     {
         log ( L_INFO, "Cannot write file %s (reading state): %m",
                        allowed_hosts[handle].filename );
         return -1;
     }

     if (s_close(fd))
     {
         log ( L_INFO, "Cannot close file %s (reading state): %m",
                        allowed_hosts[handle].filename );
         return -1;
     }

     return 0;
}

/* write all state information */
void tHostManager::write_all_states()
{
     for(int i=0;i<allowed_hosts_cnt;i++)
         if (get_connected( i )) write_state( i );
}
