/**
 * @brief Echelog sender config file reading 
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include "sender.h"
#include "messages/std_conf.h"

using namespace lib;
using namespace socks;
using namespace messages;
using namespace spool;

/* placeholders for read attributes */
char *pk_file, *cert_file, *root_file, *server_name, *cn, *common_name;
char *input_name, *sp_dir, *idx_file;
int r_port, wr_thrg=1, ta_secs=10, compressed=1;
long flags;

/* pool for parsing, freed after each parsing */
static class lib::tMemPool conf_mp(4096);

/* listenning UD socket */
class tUNIXSocket *in_sock;
/* outcomming openssl socket */
class tSSLSocket *out_sock;
/* socket manager does all the nasty work with sockets */
class tSocketMgr *smgr;
/* spooler object */
class tQueue *queue;
/* settings for openssl connection */
class tSSLContext *ctx;

/* declaration of [Sender] section fields */
static struct cfitem sender_section[]={
    {name: "Sender",            type: CT_SECTION,    data: NULL},
    {name: "PrivateKeyFile",    type: CT_STRING,     data: &pk_file,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "CertificateFile",   type: CT_STRING,     data: &cert_file, 
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "RootCertFile",      type: CT_STRING,     data: &root_file,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "ReceiverHostname",  type: CT_STRING,     data: &server_name,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "ReceiverCommonname",type: CT_STRING,     data: &cn,
     varf: NULL, varf_reset: NULL },          
    {name: "RemotePort",        type: CT_INT,        data: &r_port,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "TryAgainSec",       type: CT_INT,        data: &ta_secs,
     varf: NULL, varf_reset: NULL },
    {name: "SpoolDir",          type: CT_STRING,     data: &sp_dir,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "IndexFile",         type: CT_STRING,     data: &idx_file,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: "WriteThrough",      type: CT_INT,        data: &wr_thrg,
     varf: NULL, varf_reset: NULL },
    {name: "SendCompressed",    type: CT_INT,        data: &compressed,
     varf: NULL, varf_reset: NULL },
    {name: NULL,                type: CT_STOP,       data: NULL}
};

/* declaration of [System] section */
static struct cfitem system_section[]={
    {name: "System",            type: CT_SECTION,    data: NULL},
    {name: "Input",             type: CT_STRING,     data: &input_name,
     varf: NULL, varf_reset: NULL, flags: ITEM_REQUIRED },
    {name: NULL,                type: CT_STOP,       data: NULL}
};


/* Set up of sender after updating configuration */
void test_and_update_config (void)
{

    /* create instance holding parameters for openssl connection */
    ctx = new tSSLContext;

    if ( !ctx || ctx->set_key_files( pk_file, cert_file, root_file ))
    {
        die("Cannot load files for SSL handshake.");
    }

    /* create outcomming openssl socket */
    out_sock = new tSSLSocket;
    if (!out_sock)
        die ("Not enough memory");

    out_sock->set_on_hup (ssl_sock_tcp_hup);
    out_sock->set_on_err (ssl_sock_tcp_err);
    out_sock->set_on_connected (ssl_sock_connected);
    out_sock->set_on_read_available (ssl_sock_read_available);
    out_sock->set_on_read_finished (ssl_sock_read_finished);
    out_sock->set_on_write_finished (ssl_sock_write_finished);
    out_sock->set_on_ssl_shutdown_finished (ssl_sock_shutdown_finished);

    if(!out_sock || out_sock->create_client (server_name, r_port, ctx, smgr))
        die("Can't create SSL connection stuctures.");

    /* set compression if desired */
    if(compressed)
        flags = FL_ZLIB_COMPRESSED; else flags = 0;

    /* set common name for verification */
    if(cn)
        common_name = s_strdup(cn);
    else
        common_name = NULL;
 
    /* Release parsing data */
    get_cfpool().free_pool ();

    log (L_INFO, "Sucessfully configured");
}

/* Load module's configuration */
void load_config (int argc, char *argv[])
{

    gb_default_config_filename = DEFAULT_CONFIG_FILE;
    wr_thrg=1;ta_secs=10;compressed=1;

    cf_register(sender_section);
    cf_register(system_section);

    /* Configure echelog types and messages */
    if (cf_read_std_configuration (argc, argv))
        die ("Unable to load configuration file '%s'",
                    cf_get_config_filename ());
    
    log (L_INFO, "Using configuration file '%s'.", cf_get_config_filename ());

    /* create new soket manager instance */
    if ( !( smgr = new tSocketMgr ) )
    {
        die("Not enough memory for estabilishing socket manager.");
    }
    
    /* create listenning socket for incomming UD connections */
    in_sock = new tUNIXSocket; 
    if ( !in_sock || in_sock->create_server(input_name))
    {
        die("Failed to estabilish listenning socket for incomming data.");
    }    

    in_sock->set_on_data_hup (unix_sock_hup_f);
    in_sock->set_on_data_err (unix_sock_err_f);
    in_sock->set_on_data_read (unix_sock_read_f);
  
    if ( smgr->register_socket( in_sock, POLLPRI | POLLIN ))
    {
        die("Failed to register listenning socket for incomming data.");
    }

    /* create the spooler instance */
    queue = new tQueue( (wr_thrg?1:0), sp_dir, idx_file );
    if (!queue)
        die ("Not enough memory.");

    if (queue->open_queue (1))
         die ("Cannot open/create spool-structures.");

    /* check whether renumberation is needed */
    if (queue->get_last_id () > MAX_ITEM_ID)
    {
         log (L_INFO, "Transfer identificators too large. Spool is gooing to "
                      "shifted to smaller values.");
        
         /* close the queue */
         if (queue->close_queue ())
             die ("Error trying to close the spool.");
         delete queue;
         
         /* try to shift the spool */
         if (shift_to_start (sp_dir, ".spooldir.tmp", idx_file))
             die ("Error when trying to shift spool.");
         
         /* and palce ther tag */
         if (tag_renumber (sp_dir))
             die ("Cannot make tag file indicating renumberation.");

         /* create the spooler instance */
         queue = new tQueue( (wr_thrg?1:0), sp_dir, idx_file );
         if (!queue)
             die ("Not enough memory.");
         if (queue->open_queue (1))
             die ("Cannot open/create spool-structures.");
         
    }

    /* now set up things which change after reload */ 
    test_and_update_config ();
}

/* Reload getlog configuration part */
void reread_configuration ()
{
    using namespace messages;

    /* Announce this */
    log (L_INFO, "Reloading echelog configuration");

    /* Close the connection to receiver */
    if (out_sock->get_smgr_handle () >= 0)
        if (smgr->unregister_socket(out_sock) < 0)
            log (L_ERROR, "Can't unregister socket (fd=%ld)",
                    out_sock->get_fd());
    if (out_sock->close_socket())
        log (L_ERROR, "Can't close socket (fd=%ld)", out_sock->get_fd ());
    delete out_sock;
    
    /* release old instance holding parameters for openssl connections */
    delete ctx;

    /* release memory for common name */
    if (common_name) xfree(common_name);
    
    /* Remove old configuration data */
    conf_mp.free_pool ();

    log (L_INFO, "Using configuration file '%s'.", cf_get_config_filename ());

    /* set default values for config file parameters */
    wr_thrg=1;ta_secs=10;compressed=1;

    /* Read new configuration */
    if (cf_reread_configuration (conf_mp))
         die ("Unable to load configuration file '%s'",
                      cf_get_config_filename ());

    /* Reload new configuration */
    test_and_update_config ();
}

/* quick shutdown routine */
void bring_down()
{
    /* shutdown the spooler */
    queue->close_queue ();

    /* release all socket held in socket manager */
    smgr -> cleanup(); 

    /* remove listenning socket */
    delete in_sock;
    
    /* remove outgoing socket */
    delete out_sock;

    /* Exit */
    log (L_INFO, "Terminated with non-zero status.");
    log_close();
    exit(1);
}
