/**
 * @brief Echelog Receiver 
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include "sender.h"
#include "statusmsg.h"

using namespace socks;
using namespace lib;
using namespace spool;

class tStateInfo stat_info;
enum tRState phase;
tIdentificator id_to_send;
static int renumbered;

/* remove client connection */
static void remove_socket(class tSocket *sock)
{
     log (L_INFO,"Removing socket (%ld).", sock->get_fd());
     sock->get_smgr ()->unregister_socket (sock);
     sock->close_socket ();
     delete sock;
}

/* return error string according to error value */
static const char * get_error_string(enum tQuitStatus err)
{
     switch (err)
     {
         case QS_OK:
                return "No error.";
         case QS_UNKNOWN:
                return "Receiver has not this sender on list for accepting.";
         case QS_ALREADY_CONNECTED:
                return "Another connection from this host already exists.";
         case QS_INV_CERT:
                return "Receiver rejected our certificate.";
         case QS_QUOTA_EXC:
                return "Quota exceeded.";
         default:
                return "Unknown error.";
     }                                   
}     

/* event handler for error listenning socket and incomming connections */
void unix_sock_err_f(class tSocket *sock)
{
      /* error on listenning socket - shutdown sender */
      if (sock->get_role() == AS_SOCK_SERVER)
      {
          log(L_WARN,"Listenning socket encountered error; shutting down.");
          bring_down();
      } else
      /* error on client connection - remove the nasty socket */    
      {
          log(L_WARN,"Socket (%ld) error: Releasing the socket.",
                     sock->get_fd());
          remove_socket (sock);
      }
}

/* event handler for hangup of listenning socket and incomming connections */
void unix_sock_hup_f(class tSocket *sock)
{
      /* hangup on listenning socket - this should not happen */
      if (sock->get_role() == AS_SOCK_SERVER)
      {
          log(L_WARN,"Listenning socket encountered error; shutting down.");
          bring_down();
      }
      else
      /* hangup on client connection - remove the socket */
      {
          log(L_WARN,"Other side ended communication (%ld). Releasing socket.",
                     sock->get_fd() );
          remove_socket (sock);
      }
}

/* data arrived to the socket */
void unix_sock_read_f(class tSocket *sock)
{
      /* listenning socket? socket manager should never call this for
       * listenning sockets */
      if (sock->get_role() == AS_SOCK_SERVER)
      {
          log(L_WARN,"Unexpected data on listenning socket: shutting down.");
          bring_down();
      } 
      /* message arrived */
      else 
      {
          int len;
          char buffer[MAX_MESSAGE_SIZE];

          /* read it */
          if((len=sock->at_read(buffer, MAX_MESSAGE_SIZE))<0)
          {
              log(L_WARN,"Error receiving data from unix socket.");
              unix_sock_err_f (sock);
          }

          if (queue->get_last_id () > MAX_ITEM_ID)
          {
              log(L_WARN,"Transfer item identifiers became too large. Spool "
                         "shift needed. Exit.");
              bring_down();
          }

          tIdentificator id = queue->insert_item (buffer, len);

          if (id <= 0)
          {
              log(L_WARN,"Error inserting item into the spool. Exit.");
              bring_down();
          }

          debug_log('0', "Received item %ld", id);

          /* wait till we can write to ssl socket */
          out_sock->get_smgr()->set_event_mask (out_sock, 
                                                POLLPRI | POLLIN | POLLOUT);
      }
}

/* handler of errors on outgoing openssl connection */
void ssl_sock_tcp_err(class tSocketNB *sock)
{
      switch(sock->get_status())
      {
          case SS_CONNECTING:
              /* error occured during connecting */
              log(L_WARN,"Connection failed: will try later.");
          case SS_DEFAULT:
              /* no connection estabilished yet */
              if (sock->get_status()==SS_DEFAULT)
                  log(L_WARN,"Connection not exists. Will try renew it.");
          case SS_BROKEN:
              /* connection broken down - close old socket and create new one */
              if (out_sock->get_smgr_handle()>=0)
                  out_sock->get_smgr()->unregister_socket (out_sock);
              delete out_sock;

              /* create new socket for outgoing connection */
              out_sock = new tSSLSocket;

              /* set its handlers */
              out_sock->set_on_hup (ssl_sock_tcp_hup);
              out_sock->set_on_err (ssl_sock_tcp_err);
              out_sock->set_on_connected (ssl_sock_connected);
              out_sock->set_on_read_available (ssl_sock_read_available);
              out_sock->set_on_read_finished (ssl_sock_read_finished);
              out_sock->set_on_write_finished (ssl_sock_write_finished);
              out_sock->set_on_ssl_shutdown_finished (
                                               ssl_sock_shutdown_finished);

              /* fill in its structures and prepare it to connect */
              if(!out_sock || out_sock->create_client (server_name, r_port,
                                                       ctx, smgr))
                  die("Can't continue connecting.");

              /* set alarms after which we will try to connect */
              alarm(ta_secs);
                      
              break;
          /* bad state of outgoing socket - shutdown of sender */
          default:
              log(L_WARN,"Critical error on outgoing ssl connection.");
              bring_down();
     }
}

/* handler of hangups on outgoing openssl connection */
void ssl_sock_tcp_hup(class tSocketNB *sock)
{
      /* we were striped out by server, we will try later */
      log(L_WARN,"Server hangup, will try connect again.");

      /* connection broken down - close old socket and create new one */
      out_sock->get_smgr ()->unregister_socket (out_sock);
      delete out_sock;

      /* create new socket for outgoing connection */
      out_sock = new tSSLSocket;

      /* set its handlers */
      out_sock->set_on_hup (ssl_sock_tcp_hup);
      out_sock->set_on_err (ssl_sock_tcp_err);
      out_sock->set_on_connected (ssl_sock_connected);
      out_sock->set_on_read_available (ssl_sock_read_available);
      out_sock->set_on_read_finished (ssl_sock_read_finished);
      out_sock->set_on_write_finished (ssl_sock_write_finished);
      out_sock->set_on_ssl_shutdown_finished (ssl_sock_shutdown_finished);

      /* fill in its structures and prepare it to connect */
      if(!out_sock || out_sock->create_client (server_name, r_port, ctx, smgr))
          die("Can't continue connecting.");

       /* set alarms after which we will try to connect */
      alarm(ta_secs);

}

/* we were connected - hurrah */
void ssl_sock_connected(class tSocketNB *sock)
{
      class tSSLSocket *s = static_cast<class tSSLSocket *>(sock);
      
      /* verify common name if requested */
      if (common_name && s->verify_common_name(common_name)!=0)
      {
          log(L_INFO, "Connection closed with (%s) because common name "
                      "does not match.", sock -> get_address());
          sock -> set_status (SS_DEFAULT);
          ssl_sock_tcp_err (sock);
          return;
      }

      /* schedule reding of init status message */
      sock->at_read (stat_info.get_msg_buf(), STATE_MESSAGE_SIZE);
      phase = RS_EMPTY;
}

/* we do not expect data for arrival unless we request it */
void ssl_sock_read_available(class tSocketNB *sock)
{
      /* reconnect the socket */
      log(L_INFO, "Connection closed with (%s) because of communication "
                  "error.", sock -> get_address());
      sock -> set_status (SS_DEFAULT);
      ssl_sock_tcp_err (sock);
}

/* we just finished reading operation */
void ssl_sock_read_finished(class tSocketNB *sock)
{
      enum tCommitState state;
      spool::tIdentificator arg;

      /* in which phase are we ? */
      switch(phase)
      {
          /* we received receiver transfer start request */
          case RS_EMPTY:
                stat_info.load_message();
                state = stat_info.get_state();
                arg = stat_info.get_arg();

                /* we expect only start and quit messages */
                if (sock -> get_r_io().last_op_bytes !=  STATE_MESSAGE_SIZE ||
                   (state != CS_START && state != CS_QUIT))
                {
                    log(L_INFO, "Connection closed with (%s): violated protocol"
                                " - init handshake.", sock -> get_address());
                    sock -> set_status (SS_DEFAULT);
                    ssl_sock_tcp_err (sock);
                    break;
                }
                
                /* ermm ... we should quit ... we will try to reconnect later */
                if (state == CS_QUIT)
                {    
                    log(L_INFO, "Connection closed with (%s): %s",
                           sock -> get_address(),
                           get_error_string((enum tQuitStatus)arg));
                    sock -> set_status (SS_DEFAULT);
                    ssl_sock_tcp_err (sock);
                    break;
                }

                /* we were accepted - so test if receiver requests valid
                 * transfer start id */
                id_to_send = arg;          
                renumbered = 0;
                if (id_to_send==0)
                {
                    /* request new transfer */
                    if (queue->get_first_id()==1)
                    {
                        id_to_send = 1;
                        log (L_INFO, "Starting new transfer from id = 1.");
                    } else
                        die ("Receiver requests new transfer, but there are "
                             "still some unsent data. Exit.");
                } else
                /* did we shifted the spool ? */    
                if (istag_renumber (sp_dir))
                {
                        id_to_send = 1;
                        log (L_INFO, "Starting new transfer from id = 1 "
                                     "(due to spool shift).");
                        renumbered = 1;
                } else
                /* we will try to continue in old transfer */
                if (id_to_send < queue->get_first_nonextracted_id() ||
                    id_to_send > queue->get_last_id())
                       die ("Receiver requests to continue from id = %ld, but "
                            "this id lies out of stored range (%ld-%ld). Exit.",
                            id_to_send, queue->get_first_nonextracted_id(),
                            queue->get_last_id());
                else
                /* receiver request valid id ... update spooler */
                {
                    if (queue->strip_queue(id_to_send))
                        die ("Error stripping the queue.");
                    if (queue->get_status(id_to_send)!=PRESENT && 
                        id_to_send!=queue->get_last_id() )
                        die ("Receiver requests to continue from id = %ld, but "
                             "this item was already sent. Exit.", id_to_send);
                    else 
                        log( L_INFO, "Transfer succesfully restored at id=%ld.",
                             id_to_send);
                }
               
                
                /* send confirmation of start id */
                stat_info.set_arg(id_to_send);
                stat_info.set_state (CS_REQ);
                stat_info.store_message();
                
                /* send the confirmation message */
                sock -> at_write (stat_info.get_msg_buf(), STATE_MESSAGE_SIZE);

                /* enter negotiate phase */ 
                phase = RS_NEGOTIATE;
                break;
          /* we received commit message of last sent message */
          case RS_COMMIT:
                /* load its data */
                stat_info.load_message();
                state = stat_info.get_state();
                arg = stat_info.get_arg();
               
                /* is it really confirmation or quota refusal */
                if (sock -> get_r_io().last_op_bytes !=  STATE_MESSAGE_SIZE ||
                  (state!=CS_ACCEPTED && state!=CS_QUIT && state!=CS_REFUSED) ||
                  ((state==CS_ACCEPTED || state==CS_REFUSED) &&
                   id_to_send!=arg ))
                {
                    log(L_INFO, "Connection closed with (%s): violated protocol"
                                " - init handshake.", sock -> get_address());
                    sock -> set_status (SS_DEFAULT);
                    ssl_sock_tcp_err (sock);
                    break;
                }

                /* it is quota refusal */
                if (state==CS_QUIT)
                {
                    log(L_INFO, "Connection closed with (%s): %s",
                           sock -> get_address(),
                           get_error_string((enum tQuitStatus)arg));
                    sock -> set_status (SS_DEFAULT);
                    ssl_sock_tcp_err (sock);
                    break;
                }

                /* great last message commited */
                if (state==CS_ACCEPTED)
                {
                    queue->extract_item(id_to_send);
                    debug_log( '0', "Extracted item %ld.", id_to_send);
                    id_to_send++; 

                    /* now we are sure taht receiver knows about spool shift
                     * and we can remove the tag */
                    if (renumbered && untag_renumber(sp_dir))
                    {
                        log (L_ERROR, "Error removing tagfile.");
                        bring_down();        
                    }

                    renumbered = 0;
                } 

                /* go to sending next message */
                phase = RS_SEND;
                try_write();
                break;
          /* in other phases we are not eager for incomming data */
          default:
                log(L_INFO, "Connection closed with (%s): violated protocol "
                            "- init handshake.", sock -> get_address());
                sock -> set_status (SS_DEFAULT);
                ssl_sock_tcp_err (sock);
      }
}

/* write operation finished on outgoing openssl socket */
void ssl_sock_write_finished(class tSocketNB *sock)
{
      /* in which phase we are ? */
      switch(phase)
      {
          /* we just sent start id of transfer */
          case RS_NEGOTIATE:
              /* did we send it correctly / */
              if (sock->get_r_io().last_op_bytes !=  STATE_MESSAGE_SIZE) 
              {
                  log(L_INFO, "Connection closed with (%s): violated protocol "
                              "- init handshake.", sock -> get_address());
                  sock -> set_status (SS_DEFAULT);
                  ssl_sock_tcp_err (sock);
                  break;
              }

              /* yep, try to send message */
              phase = RS_SEND;
              try_write();
              break;
          /* we just write whole message to receiver */
          case RS_SEND:
              /* wait for commit of receipt */
              phase = RS_COMMIT;
              sock->at_read (stat_info.get_msg_buf(), STATE_MESSAGE_SIZE);
              break;
          /* we do not expect write finish in other states */
          default:
              log(L_INFO, "Connection closed with (%s): violated protocol - "
                          "init handshake.", sock -> get_address());
              sock -> set_status (SS_DEFAULT);
              ssl_sock_tcp_err (sock);
      }

}

/* communication ended by the otherside - try to reconnect */
void ssl_sock_shutdown_finished(class tSSLSocket *sock)
{
      log (L_INFO, "Comunication ended with other side.");
      sock->set_status (SS_DEFAULT);
      ssl_sock_tcp_err (sock);
}

/* try to send next message (if there is one) */
void try_write()
{
      struct tItem item_desc;

      /* are really connected ? */
      if (out_sock->get_status()==SS_CONNECTED &&
          out_sock->get_ssl_status()==SL_CONNECTED &&
          out_sock->get_last_operation()==OP_NO_OPERATION &&
          phase==RS_SEND)
      {
          /* do we have anything spooled ? */
          if (id_to_send<queue->get_last_id())
          {
              /* we do not have this message - strange */
              if (queue->get_item(id_to_send)==NULL || 
                  queue->get_item_description(id_to_send, &item_desc)==NULL )
              {
                  id_to_send++;
              } else 
              
              /* write the message */
              out_sock->at_write( queue->get_item(id_to_send), item_desc.length,
                                  flags);
          }           
      } 
}
