/**
 * @brief Echelog Receiver - Support for spool shift, when item identifiers
 * become to large.
 *
 * @author Ales Privetivy <ales@ucw.cz>
 *
 */

#include "lib/lib.h"
#include "spool/queue.h"

#include <libgen.h>

using namespace lib;
using namespace spool;

int shift_to_start(char *old_spool_dir, char *temp_spool_dir, 
                   char *idx_file)
{
    class tQueue *old_queue, *new_queue;
    char *idx_file_name, *new_idx_file;
    tIdentificator id;
    int err = 0;
    
    ASSERT (old_spool_dir);
    ASSERT (temp_spool_dir);
    ASSERT (idx_file);

    /* create the temporary spooldir */
    if (mkdir (temp_spool_dir, 0700))
    {
        log (L_ERROR, "Cannot create temporary spool directory \"%s\": %m",
                      temp_spool_dir);
        return -1;
    }     
    
    /* create new spool instance */
    old_queue = new tQueue (0, old_spool_dir, idx_file);
    if (!old_queue)
        die ("Not enough memory.");

    if (old_queue->open_queue (1))
    {    
        log (L_ERROR, "Cannot open spool in directory \"%s\".", old_spool_dir);
        err = -1;
    }

    /* find out temporary index file name */
    idx_file_name = basename(idx_file);
    int base_len = strlen (idx_file_name);
    int temp_len = strlen (temp_spool_dir);
    new_idx_file = static_cast<char *> (xmalloc (base_len + temp_len + 2));
    if (!new_idx_file)
        die ("Not enough memory.");
    memcpy (new_idx_file, temp_spool_dir, temp_len);
    memcpy (new_idx_file + temp_len + 1, idx_file_name, base_len);
    new_idx_file [temp_len] = '/'; new_idx_file [base_len + temp_len + 1] = 0;

    /* create new spool instance */
    new_queue = new tQueue (0, temp_spool_dir, new_idx_file);
    if (!new_queue)
        die ("Not enough memory.");
    if (new_queue->open_queue (1))
    {
        log (L_ERROR, "Cannot open spool in directory \"%s\".", temp_spool_dir);
        err = -1;
    }
    xfree (new_idx_file);

    /* copy the old spool to the beginning of new one */
    id = old_queue->get_first_nonextracted_id ();
    while (err == 0 && id < old_queue->get_last_id ())
    {
          struct tItem item_desc;

          /* we do not have this item - strange */
          if (old_queue->get_item (id) == NULL ||
              old_queue->get_item_description(id, &item_desc)==NULL )
          {
              log (L_ERROR, "Spool \"%s\" corrupted.", old_spool_dir);
              err = -1;
              break;
          } else
          if (new_queue->insert_item (old_queue->get_item (id), 
              item_desc.length) <= 0)
          {
              log (L_ERROR, "Can't insert into spool \"%s\".", temp_spool_dir);
              err = -1;
              break;
          }
          
          /* remove the item */
          old_queue->extract_item (id);
          id ++;
    }
    
    if (old_queue->close_queue ())
    {
        log (L_ERROR, "Error closing spool \"%s\".", old_spool_dir);
        err = -1;
    }

    if (new_queue->close_queue ())
    {
        log (L_ERROR, "Error closing spool \"%s\".", temp_spool_dir);
        err = -1;
    }

    if (err == 0)
    {        
        if (remove_directory (old_spool_dir))
        {
            log (L_ERROR, "Failed to remove old spool \"%s\".", old_spool_dir);
            err = -1;
        }    

        if (move_directory (temp_spool_dir, old_spool_dir))
        {
            log (L_ERROR, "Failed to move new spool \"%s\".", temp_spool_dir);
            err = -1;
        }

        return err;
    } else
    {
        if (remove_directory (temp_spool_dir))
        {
            log (L_ERROR, "Failed to remove temp spool \"%s\".", temp_spool_dir);
            err = -1;
        }    
        
        return -1;
    }        

}

int tag_renumber(char *spooldir)
{
    char tag[] = ".renumbered";
    int tag_len = strlen (tag);
    int dir_len = strlen (spooldir);
    char *tagfile;
    int fd;

    /* find out tagfile name */
    tagfile = static_cast<char *> (xmalloc (dir_len + tag_len + 2));
    memcpy (tagfile, spooldir, dir_len);
    memcpy (tagfile + dir_len + 1, tag, tag_len);
    tagfile [dir_len] = '/'; tagfile [dir_len + tag_len + 1] = 0;
    
    /* create the tagfile */
    if ((fd = s_open (tagfile, O_CREAT | O_TRUNC, 0600)) < 0)
    {
        log (L_ERROR, "Cannot create tagfile \"%s\": %m", tagfile);
        xfree (tagfile);
        return -1;
    }  

    if (s_close (fd))
    {
        log (L_ERROR, "Cannot close tagfile \"%s\": %m", tagfile);
        xfree (tagfile);
        return -1;
    }

    xfree (tagfile);
    return 0;
}

int untag_renumber(char *spooldir)
{
    char tag[] = ".renumbered";
    int tag_len = strlen (tag);
    int dir_len = strlen (spooldir);
    char *tagfile;

    /* find out tagfile name */
    tagfile = static_cast<char *> (xmalloc (dir_len + tag_len + 2));
    memcpy (tagfile, spooldir, dir_len);
    memcpy (tagfile + dir_len + 1, tag, tag_len);
    tagfile [dir_len] = '/'; tagfile [dir_len + tag_len + 1] = 0;

    /* unlink the tagfile */
    if (unlink (tagfile))
    {
        log (L_ERROR, "Cannot unlink tagfile \"%s\": %m", tagfile);
        xfree (tagfile);
        return -1;
    }

    xfree (tagfile);
    return 0;
}

int istag_renumber(char *spooldir)
{
    char tag[] = ".renumbered";
    int tag_len = strlen (tag);
    int dir_len = strlen (spooldir);
    char *tagfile;
    int fd;

    /* find out tagfile name */
    tagfile = static_cast<char *> (xmalloc (dir_len + tag_len + 2));
    memcpy (tagfile, spooldir, dir_len);
    memcpy (tagfile + dir_len + 1, tag, tag_len);
    tagfile [dir_len] = '/'; tagfile [dir_len + tag_len + 1] = 0;

    /* tyr to open the tagfile */
    if ((fd = s_open (tagfile, 0, 0600)) < 0)
    {
        xfree (tagfile);
        return 0;
    }

    if (s_close (fd))
        log (L_ERROR, "Cannot close tagfile \"%s\": %m", tagfile);

    xfree (tagfile);
    return 1;
}
